//================================================================================================
// EQX game engine
//
// 3D file converter, based on the libE3D library.
// This may also be used as a simple framework to learn libE3D.
// Example use (convert torus.hrc to an .obj file):
// e3c -v torus.hrc a.obj
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-20 19:06:21
//
// This sample code is freely modifiable and distributable
//
// EQUINOX-3D and its components are Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================================================
#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>

#include <float.h>
#include <math.h>
#include <string.h>

#include <ELanguage.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/Types.h>
#include <E3D/Plugins.h>

#include <E3D/UI/Prefs.h>


#include "common.h"

// The main 3D Scene
//
E3dScene*  E3d_Scene = NULL;


// Singleton for the application
//
typedef struct
{
	EBool  Verbose;
} E3c;


static E3c  _Tool;

// We might as well use the command-line argument parser in the library
//
#define ROffset(field)  (void*)(EPtrOffset(E3c, field))
static EResource	_ArgResources[]=
{
	{ "Verbose",   EResBOOLEAN,    ROffset(Verbose),  .Semantic= "Verbose mode.",   .UI= "v" },
	{ }
 };
#undef ROffset



//========================================================================================
// Function, used by libE3D to print messages in the GUI (EQUINOX-3D's StatusBar).
// We are a command-line tool, so we'll just print the message to stdout.
//========================================================================================
E3dID E3dp_PrintMessage(const int PType, const int PRemoveInMillisedonds, const char* PMessage, ...)
{
	if(_Tool.Verbose)
	{
		va_list  LAL;
		va_start(LAL, PMessage);           // PMessage is the last argument before the varargs
			vprintf(PMessage, LAL);
		va_end(LAL);
		printf("\n");fflush(stdout);
	}

	return(0);
}


//========================================================
// Main
//========================================================
int main(const int PArgs_Count, const char* const* PArgs)
{
	E_MallocInit();


	E3dp_PreferencesInit(PArgs_Count, PArgs);


// Initialize paths
//
	E3d_LaunchPath   = E_LaunchPath(PArgs[0]);
	E_ExecutableName = EStrDup(PArgs[0]);

// Initialize working versions of paths (the ones saved in the preferences may be NULL, indicating "use default"
// (relative to the launch location)
//
	if(E3dp_Prefs.ResourcePath) E3dp_Prefs.WorkingResourcePath = EStrDup(E3dp_Prefs.ResourcePath);
	else
	{
		E3dp_Prefs.WorkingResourcePath = EStrDup(E3d_LaunchPath);
	}


// Parse arguments and load/save 3D files
//
	E3c*  LTool = &_Tool;EClr_p(LTool);

	if(PArgs_Count > 2)
	{
		const int  LNextArg = ERes_GetArguments(PArgs, PArgs_Count, LTool, _ArgResources);

		if(LNextArg < PArgs_Count + 2)
		{
// Initialize lib2D and libE3D
//
		Ep_Init();
		E3d_Init();

// Load 3D I/O plugins
//
			if(E3d_LaunchPath)
			{
				char  LTmpStr[MAXPATHLEN];sprintf(LTmpStr, "%s/3DIO", E3d_LaunchPath);
				E3d_3DIOReadStatus = E3d_PluginsLoad(LTmpStr, "Formats.list");
			}

// Create the main Scene
//
			E3dScene*  LScene = E3dScene_New("default");if(LScene == NULL) return(-1);
			E3d_Scene = LScene;


// Get the extension from the output file name and figure out what file format we should save, and whether we can/should save a Scene, a hierarchy, or a single Model
//
			const char*  LOutFileName = PArgs[PArgs_Count - 1];const char*  LOutExtension = EStr_GetFileNameExtension(LOutFileName);if(LOutExtension)
			{
				E3dIOModule*  LSceneOutIOModule     = NULL;
				E3dIOModule*  LHierarchyOutIOModule = NULL;
				E3dIOModule*  LModelOutIOModule     = NULL;
				if(LOutExtension)
				{
					LSceneOutIOModule     = E3d_IOModuleByFileNameExtension(LOutExtension,  E3d_3DSceneFileFormats,     E3d_3DSceneFileFormats_Count,     EWRITE);
					LHierarchyOutIOModule = E3d_IOModuleByFileNameExtension(LOutExtension,  E3d_3DHierarchyFileFormats, E3d_3DHierarchyFileFormats_Count, EWRITE);
					LModelOutIOModule     = E3d_IOModuleByFileNameExtension(LOutExtension,  E3d_3DModelFileFormats,     E3d_3DModelFileFormats_Count,     EWRITE);
				}

				if(LSceneOutIOModule == NULL && LHierarchyOutIOModule == NULL && LModelOutIOModule == NULL)
				{
					printf("No saving-capable 3D file format was found with extension '%s'.\n", LOutExtension);fflush(stdout);
					return(1);
				}

// Is this a Scene file?
//
				const char*  LInFileName = PArgs[LNextArg];
				E3dScene*  LNewScene;int  LResult = E3d_SceneReadFromFile(LInFileName, E3dFileFormatAUTOMATIC, NULL, &LNewScene, NULL, FALSE);if(LResult == E_SUCCESS)
				{
					if(LNewScene)
					{
						if(LTool->Verbose) { printf("Loaded Scene from %s\n", LInFileName);fflush(stdout); }
						if(LSceneOutIOModule)
						{
							const int  LSaveResult = E3d_SceneWriteToFile(LOutFileName,  NULL, 0,  LScene, LSceneOutIOModule, NULL);if(LSaveResult == E_SUCCESS)
							{
								if(LTool->Verbose) { printf("Saved hierarchy to %s\n", LOutFileName);fflush(stdout); }
							}
							else { printf("Could not save %s: %s\n", LOutFileName, E_ResultToString(LSaveResult));fflush(stdout); }
						}
						else
						{
							printf("Could not find Scene-supporting output file format with extension '%s'\n", LOutExtension);fflush(stdout);
						}
					}
				}
				else if(LResult == E_UNSUPPORTED_FORMAT)
				{
// No, it's not a Scene file. Is it a hierarchy file (.hrc, etc.)?
//
					E3dNode*  LNode;LResult = E3d_HierarchyReadFromFile(LInFileName, E3dFileFormatAUTOMATIC, NULL, &LNode);if(LResult == E_SUCCESS)
					{
						if(LNode)
						{
							if(LTool->Verbose)
							{
								if(LNode->Child) printf("Loaded hierarchy from %s\n", LInFileName);
								else printf("Loaded Model from %s\n", LInFileName);
								fflush(stdout);
							}
							LNode->FileName = EStrDup(LInFileName);

							E3dNodeHrc_UpdateMaterialInheritance(LNode);
							E3dNodeHrc_UpdateConnections(LNode);
							E3dScene_NodeHrcAdd(LScene, LNode, 0);    // Add the Model or hierarchy to the Scene

							if(LHierarchyOutIOModule)
							{
								const int  LSaveResult = E3d_HierarchyWriteToFile(LOutFileName,  NULL, 0,  LNode, LHierarchyOutIOModule, NULL);if(LSaveResult == E_SUCCESS)
								{
									if(LTool->Verbose) { printf("Saved hierarchy to %s\n", LOutFileName);fflush(stdout); }
								}
								else { printf("Could not save %s: %s\n", LOutFileName, E_ResultToString(LSaveResult));fflush(stdout); }
							}
							else
							{
								if(LNode->Child) { printf("The loaded file has a hierarchy, but I could not a find hierarchy-supporting output file format with extension '%s'\n", LOutExtension);fflush(stdout); }
								else
								{
// If the loaded Model doesn't have children, a single-Model save format will do
//
									if(LModelOutIOModule)
									{
										const int  LSaveResult = E3dNode_SaveToFile(LOutFileName,  NULL, 0,  LNode, LModelOutIOModule, NULL);if(LSaveResult == E_SUCCESS)
										{
											if(LTool->Verbose) { printf("Saved Model to %s\n", LOutFileName);fflush(stdout); }
										}
										else { printf("Could not save %s: %s\n", LOutFileName, E_ResultToString(LSaveResult));fflush(stdout); }
									}
									else { printf("Could not a find either a Model, or a hierarchy-supporting output file format with extension '%s'\n", LOutExtension);fflush(stdout); }
								}
							}
						}
					}
					else if(LResult == E_UNSUPPORTED_FORMAT)
					{
// Nope, not a hierarchy file. Is it a single Model/Object file (.stl, .obj etc.)?
//
						LResult = E3dNode_LoadFromFile(LInFileName, E3dFileFormatAUTOMATIC, NULL, &LNode);if(LResult == E_SUCCESS)
						{
							if(LNode)
							{
								if(LTool->Verbose) { printf("Loaded Model from %s\n", LInFileName);fflush(stdout); }
								LNode->FileName = EStrDup(LInFileName);

								E3dNodeHrc_UpdateMaterialInheritance(LNode);
								E3dScene_NodeHrcAdd(LScene, LNode, 0);  // Add the Model or hierarchy to the Scene

								if(LSceneOutIOModule)
								{
								}
								else
								{
									printf("Could not find scene-supporting output file format with extension '%s'\n", LOutExtension);fflush(stdout);
								}
							}
						}
						else
						{
							printf("Input file '%s' was not recognized.\n", LInFileName);fflush(stdout);
						}
					}
				}
				else
				{
					printf("Error reading input file: %s\n", E_ResultToString(LResult));fflush(stdout);
				}
			}
			else
			{
				printf("No output file extension was given (required to determine the saving format).\n");fflush(stdout);
			}
		}
		else goto PrintUsage;
	}
	else
	{
PrintUsage:
		printf("%s3D file converter v0.3.0%s - %s\n", ECyanCode, ENormalCode, _CRString);fflush(stdout);

		char  LTmpStr[1024];LTmpStr[0] = 0;char*  LPtr = LTmpStr;

		strcpy(LPtr, "Convert a 3D file or files to another file format. This is also a sample application to show how to use libE3D.\n");LPtr += strlen(LPtr);
		strcpy(LPtr, "Usage: e3c [");LPtr += strlen(LPtr);
		LPtr = EStr_GetColorCode(LPtr, EStrCOLOR_GREEN);
		strcpy(LPtr, "options");LPtr += strlen(LPtr);
		LPtr = EStr_GetColorCode(LPtr, EStrCOLOR_NORMAL);
		strcpy(LPtr, "] <input file(s)> <output file>\n\nOptions:\n");LPtr += strlen(LPtr);

		printf("%s", LTmpStr);

		ERes_ArgsPrintUsage(_ArgResources, 23);
		fflush(stdout);
	}


	return(0);
}
