//================================================================================================
// EQX Fusion game engine rendering sample
//
// It initializes libE3D, loads a 3D hierarchy from a file renders it
// User controls:
// - Right mouse button: orbit camera around the origin
// - Scroll wheel:       dolly camera
// - R:                  cycle trhough different renderers
// - G:                  toggle X-Z grid (ground plane)
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 22:35:27
//
// This sample code is freely modifiable and distributable
//
// EQUINOX-3D and its components are Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================================================

#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/Plugins.h>
#include <E3D/Camera.h>
#include <E3D/Scene.h>

#include <E3D/Rendering.h>  // Rendering system
#include <E3D/GL/Draw.h>    // OpenGL back-end
#include <E3D/GL/Window.h>  // OpenGL back-end

#include <E3D/UI/Prefs.h>
#include <E3D/UI/Window.h>


// Include the Emscripten library only if targetting WebAssembly
#ifdef __EMSCRIPTEN__
#include <emscripten/emscripten.h>
#define GLFW_INCLUDE_ES3

enum
{
	BrCHROME = 1, BrFIREFOX
};

// TODO: we need to detect this, because some things behave a bit differently between browsers
static int  _Browser = BrCHROME;
//static int  _Browser = BrFIREFOX;
#endif


#include <GLFW/glfw3.h>

#include "common.h"

#include <Xe/XeGL.h>
XeFontSize*  E3dp_SchematicsFontSize = NULL;

#ifdef EWASM  // WebAssembly
//static const char*  _DefaultAssetFileName = "fs/data/test.eqx";
//static const char*  _DefaultAssetFileName = "fs/data/DamagedHelmet.glb";
//static const char*  _DefaultAssetFileName = "fs/data/Helmet.eqx";
static const char*  _DefaultAssetFileName = "fs/data/beveledcubes.eqx";
#else
static const char*  _DefaultAssetFileName = "data/test.eqx";
#endif

static char*  _AssetFileName = NULL;


static uint8_t  _LoadingStatus = 0;

// The Scene we load assets into
//
E3dScene*  _Scene = NULL;


// The main 3D Scene used by libE3D
//
E3dScene*  E3d_Scene = NULL;

static E3dRenderer*   _Renderers[8];
static int            _NRenderers = 0;     // Number of Renderers we created
static int            _RendererIndex = 3;  // We create 4 different renderers. Let's default to the 3rd one

// Mouse movement modes
enum
{
	PmORBIT = 1,  // Orbit the camera around its aim point (press/drag with the left button)
	PmTRACK = 2
};


static E3dGPUContext*  _GPUContext = NULL;
static E3dWindow*      _3DWindow = NULL;

static int                 _PointerMode = 0;
static E3dCameraTransform  _CameraTransformBase;  // Camera transform at button press

static E3dCamera*     _Camera         = NULL;
static E3dModel*      _CameraModel    = NULL;
static E3dCoord3      _CameraAimPoint = { 0.0, 0.0, 0.0 };

static int            _BasePtrX = 0, _BasePtrY = 0;


static GLFWwindow*    _GLFWWindow = NULL;


//========================================================================================
// Function, used by libE3D to print messages in the GUI (EQUINOX-3D's StatusBar).
// We are a command-line tool, so we'll just print the message to stdout.
//========================================================================================
E3dID E3dp_PrintMessage(const int PType, const int PRemoveInMillisedonds, const char* PMessage, ...)
{
	va_list  LAL;va_start(LAL, PMessage);    // PMessage is the last argument before the varargs
		vprintf(PMessage, LAL);
	va_end(LAL);
	printf("\n");fflush(stdout);

	return(0);
}


//========================================
// Handle GLFW Errors
//========================================
static void error_callback(const int PError, const char* PDescription)
{
	fprintf(stderr, "Error: %s\n", PDescription);
}


static void key_callback(GLFWwindow* PWindow, const int PKey, const int PScanCode, const int PAction, const int PModifiers)
{
	if(PAction != GLFW_PRESS) return;
	if(_3DWindow == NULL) return;

// Cycle through Renderers with the 'R' key
//
	switch(PKey)
	{
		case GLFW_KEY_R:
			_RendererIndex++;if(_RendererIndex >= _NRenderers) _RendererIndex = 0;
		break;

		case GLFW_KEY_G:  // Toggle the X-Z grid with the 'G' key
			if(_3DWindow->GridSettings.GridXZOn) _3DWindow->GridSettings.GridXZOn = FALSE;else _3DWindow->GridSettings.GridXZOn = TRUE;
		break;

		case GLFW_KEY_ESCAPE:  // Exit when Esc or Q is pressed
		case GLFW_KEY_Q:
			glfwSetWindowShouldClose(PWindow, GLFW_TRUE);
		break;
	}

/*
	if((PKey == GLFW_KEY_ENTER && PModifiers == GLFW_MOD_ALT) || (PKey == GLFW_KEY_F11 && PModifiers == GLFW_MOD_ALT))
	{
		if(glfwGetWindowMonitor(PWindow))
		{
//			glfwSetWindowMonitor(PWindow, NULL, windowed_xpos, windowed_ypos,  windowed_width, windowed_height, 0);
		}
		else
		{
			GLFWmonitor*  monitor = glfwGetPrimaryMonitor();if(monitor)
			{
				const GLFWvidmode*  mode = glfwGetVideoMode(monitor);
				glfwGetWindowPos(PWindow, &windowed_xpos, &windowed_ypos);
				glfwGetWindowSize(PWindow, &windowed_width, &windowed_height);
				glfwSetWindowMonitor(PWindow, monitor, 0, 0, mode->width, mode->height, mode->refreshRate);
			}
		}
	}
*/
}


static void force_exit(void)
{
#ifdef __EMSCRIPTEN__
  emscripten_force_exit(EXIT_FAILURE);
#else
  exit(EXIT_FAILURE);
#endif
}



static void _SceneUpdate(E3dScene* PScene)
{
// Add a point light to the Scene
//
	{
		E3dModel*  LLightModel;E_New(E3dModel, LLightModel);
		E_SetXYZ_s(LLightModel->TransformSet.Translation,   -5.0, 10.0, 4.0);  // Move the Model to a position
//		E_SetXYZ_s(LLightModel->TransformSet.Translation,   -5.0, 5.0, 20.0);  // Move the Model to a position
//		E_SetXYZ_s(LLightModel->TransformSet.Translation,   -50.0, 50.0, 200.0);  // Move the Model to a position
		E3dModel_AddObject(LLightModel, E3dLight_class);                       // Create a Light and add it to LLightModel
		E3dLight*  LLight = (E3dLight*)LLightModel->Objects[0];
		Ec_SetRGBf32(&LLight->Color,  1.0, 1.0, 1.0);
		LLight->DiffuseFactor = 0.0;
		E3dNodeHrc_UpdateMatrices((E3dNode*)LLightModel, FALSE);               // Update the matrices of the Light Model. It's only a single Model (not a hierarchy), so we don't need to call E3dNodeHrc_UpdateConnections()
		E3dScene_NodeHrcAdd(PScene, (E3dNode*)LLightModel, 0);                 // Add the Model to the Scene
	}

	EIndex  LRC = 0;const EIndex  LRN = PScene->RootNodes_Count;for(;LRC < LRN;LRC++)
	{
		E3dNode*  LRootNode = PScene->RootNodes[LRC];

		E3dNodeHrc_UpdateForDisplay(LRootNode, E3dUF_ALL);  // Build OpenGL vertex arrays, set up Materials for rendering, etc. for each Model in the hierarchy. This may be done selectively (only for a specific Model, Mesh or PolygGroup) when you change an object.
		E3dScene_MaterialsAddFromNodeHrc(PScene, LRootNode);
		E3dNodeHrc_UpdateMaterialInheritance(LRootNode);
		E3dNodeHrc_RemoveMappersFromTextures(LRootNode);

		const EIndex  LTxReadResult = E3dScene_NodeHrcReadTextures(PScene, LRootNode, E3dp_Prefs.TexturePath, E3dp_Prefs.CurrentFrame);if(LTxReadResult >= 0)
		{
			if(LTxReadResult > 0) { printf("Loaded %lld texture(s)\n", (long long)LTxReadResult);fflush(stdout); }
		}
		else { printf("Error reading texture(s): %s", E_ResultToString(LTxReadResult));fflush(stdout); }
//		E3dScene_Read2DTextures(PScene, FALSE, FALSE);
	}

// Indicate to the renderers that the Scene changed
//
	{ EIndex  LC = 0;for(;LC < _NRenderers;LC++) _Renderers[LC]->SceneChangeFlags |= E3dSceneCHG_ALL; }
}


//========================================================
// Load a Node / Node hierarchy or a Scene from a file
//========================================================
static int _LoadAsset(E3dScene* PScene, const char* PFileName)
{
	{
		const int  LReadResult = E3d_SceneReadFromFile(PFileName, E3dFileFormatAUTOMATIC, NULL, &PScene, NULL, FALSE);if(LReadResult == E_SUCCESS)
		{
			printf("Loaded scene from %s\n", PFileName);
			fflush(stdout);
			_SceneUpdate(PScene);
			return(LReadResult);
		}
	}

	{
		E3dNode*  LNode = NULL;const int  LReadResult = E3d_HierarchyReadFromFile(PFileName, E3dFileFormatAUTOMATIC, NULL, &LNode);if(LReadResult == E_SUCCESS)
		{
			if(LNode)
			{
				if(LNode->Child) printf("Loaded hierarchy from %s\n", PFileName);
				else printf("Loaded Model from %s\n", PFileName);
				fflush(stdout);

				LNode->FileName = EStrDup(PFileName);
				E3dScene_NodeHrcAdd(PScene, LNode, 0);          // Add the Model or hierarchy to the Scene

				_SceneUpdate(PScene);
			}
		}
		return(LReadResult);
	}

	return(E_NO_DATA);
}


static void _LoadPlugins(void)
{
#ifdef EWASM  // WebAssembly
	_StaticLoadPlugins();
#else
// Load image I/O plugins
//
/*
#ifdef EWASM  // WebAssembly
		{
			char  LTmpStr[MAXPATHLEN];sprintf(LTmpStr, "fs/ImageIO");
			E3d_3DIOReadStatus = E3d_PluginsLoad(LTmpStr, "Formats.list");
		}
#else
*/
		if(E3d_LaunchPath)
		{
			{ char  LTmpStr[MAXPATHLEN];sprintf(LTmpStr, "%s/ImageIO", E3d_LaunchPath);EpImage_IOReadStatus = E3d_PluginsLoad(LTmpStr, "Formats.list"); }
			{	char  LTmpStr[MAXPATHLEN];sprintf(LTmpStr, "%s/3DIO", E3d_LaunchPath);E3d_3DIOReadStatus = E3d_PluginsLoad(LTmpStr, "Formats.list"); }
			{	char  LTmpStr[MAXPATHLEN];sprintf(LTmpStr, "%s/Plugins", E3d_LaunchPath);E3d_PluginsLoad(LTmpStr, "Plugins.list"); }
//printf("%s() %s:%d  [%s] %d\n", __FUNCTION__, __FILE__, __LINE__, E3d_LaunchPath, E3d_3DIOReadStatus);fflush(stdout);
		}
#endif  // EWASM
}


//static pthread_t  _LoadingThread;

static void* _LoadTestScene(void* PP)
{
	E3dScene*  LScene = (E3dScene*)PP;

	_LoadingStatus = 1;

	const int  LReadResult = _LoadAsset(LScene, _AssetFileName);if(LReadResult == E_SUCCESS)
	{
		E3dScene_UnselectAll(LScene, E3dSF_MODELS | E3dSF_OBJECTS, FALSE);
	}
	else
	{
		printf("Error reading input file: %s %s%s%s\n", _AssetFileName, EYellowCode, E_ResultToString(LReadResult), ENormalCode);fflush(stdout);
	}

	_LoadingStatus = 2;

	return(NULL);
}




static EIndex  _FrameNumber = 0;
static EIndex  _MaxFrameNumber = 2;  // < 0 : don't print frame numbers at all

extern void            E3dWindow_DrawGrid(E3dWindow* P3DWindow,  E3dCamera* PCamera, E3dModel* PCameraModel,  const E3dDrawContext* PDrawContext, const EBool PUsingLighting);
extern E3dDrawContext  E3dp_NormalDrawContext;

//================================
// Render a frame
//================================
static void _RenderFrame(void)
{
	int  LWindowSizeX, LWindowSizeY;glfwGetFramebufferSize(_GLFWWindow, &LWindowSizeX, &LWindowSizeY);
	E3glViewport(0, 0,  LWindowSizeX, LWindowSizeY);

	const E3dCoordinate  LAspectRatio = (E3dCoordinate)LWindowSizeX / (E3dCoordinate)LWindowSizeY;
	_3DWindow->Size.X = LWindowSizeX;_3DWindow->Size.Y = LWindowSizeY;_3DWindow->AspectRatio = LAspectRatio;
	E3dCamera_UpdateProjectionMatrix(_Camera, LAspectRatio);

// Call libE3D to draw the loaded Scene
//
#ifndef EWASM
	int  LCharsPrinted = 0;
#endif
	if(_MaxFrameNumber >= 0 && _FrameNumber <= _MaxFrameNumber)
	{
#ifdef EWASM
		printf("\n%s() %s:%d  Frame %5lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)_FrameNumber);fflush(stdout);
#else
		LCharsPrinted = _FrameNumber > 0 ? printf("Rendering frame %s%5lld%s", ECyanCode, (long long)_FrameNumber, ENormalCode) : 0;fflush(stdout);
#endif
	}

	E3dRenderer*  LRenderer = _Renderers[_RendererIndex];

//if(E3d_Verbosity & E3dVbOPENGL) XeGLM_CheckStatusNL(FALSE);

// libE3D allows you to either use a libE3D Renderer, or get down to the "metal" and do everything yourself (e.g. direct OpenGL or Vulkan calls).
// And, everything in-between!
// Here's how you call the Renderer and just let it draw the Scene, optional grids, coordinate system, etc.
//

	if(LRenderer) LRenderer->Class->Render(_3DWindow, LRenderer,  E3d_Scene,  _Camera, _CameraModel);
	else
	{
	  glClearColor(0.0, 0.7, 0.9, 1.0);

		E3glClearDepth(1);
		E3glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);E3glFlush();
		E3glEnable(GL_DEPTH_TEST);

		{
			E3glDepthMask(GL_FALSE);     // No Z-write
			E3glDisable(GL_DEPTH_TEST);  // No Z-test
			E3dDrawContext  LDrawContext = E3dp_NormalDrawContext;E3d_DrawContextUpdateFromRenderer(LRenderer, &LDrawContext);
			E3dWindow_DrawGrid(_3DWindow,  _Camera, _CameraModel,  &LDrawContext, FALSE);
		}
//printf("%s() %s:%d  %sNo Renderer%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
	}


// Or, you can manually draw hierarches with the OpenGL module (a good "in-between"): (example not working yet)
//
/*
		E3dMatrix  LWorldToViewerMatrix;E3dGL_WindowSetViewTransform(L3DWindow, LRenderer, 0, 1,  LWorldToViewerMatrix);

		E3glMatrixMode(GL_MODELVIEW);
		E3dGL_LoadMatrix(LWorldToViewerMatrix);
		if(LNode) E3dGL_HierarchyDraw(L3DWindow, LRenderer, 0, &E3dp_NormalDrawContext, E3dDF_SOLID|E3dDF_Z_BUFFER|E3dDF_LIGHTING,  LNode, LWorldToViewerMatrix,  E3dOPAQUE, FALSE);
*/


// Or, you can manually draw each Model, Mesh, Spline, etc. with the OpenGL module: (example coming)
//


// You can even access the GPU vertex arrays that libE3D generated for a particular Mesh/PolyGroup and do your own shader setup and E3glDrawElements() calls.
// (example coming)
// Or, you can use them with Vulkan...
//


	glfwSwapBuffers(_GLFWWindow);

#ifndef EWASM
	if(_MaxFrameNumber >= 0 && _FrameNumber <= _MaxFrameNumber) { char  LT[64];memset(LT, 8, LCharsPrinted);LT[LCharsPrinted] = 0;printf("%s", LT);fflush(stdout); }
#endif

	_FrameNumber++;
	if(_MaxFrameNumber >= 0 && _FrameNumber > _MaxFrameNumber) E3d_Verbosity = 0;  // Mute debug printouts after a number of frames
	glfwPollEvents();

//#ifdef EWASM
	if(_FrameNumber > 1 && _LoadingStatus < 3) { _LoadTestScene(_Scene);E3d_Scene = _Scene;_LoadingStatus = 3;EIndex  LC = 0;for(;LC < _NRenderers;LC++) _Renderers[LC]->SceneChangeFlags |= E3dSceneCHG_ALL; }
/*
#else
	if(_LoadingStatus == 2) { E3d_Scene = _Scene;_LoadingStatus = 3;for(;LC < _NRenderers;LC++) _Renderers[LC]->SceneChangeFlags |= E3dSceneCHG_ALL; }
#endif
*/
//force_exit();
}

//================================
// Pointer move callback
//================================
static void cursor_position_callback(GLFWwindow* PWindow,  double PPtrX, double PPtrY)
{
	const int  LPtrXDiff = PPtrX - _BasePtrX;
	const int  LPtrYDiff = PPtrY - _BasePtrY;

	switch(_PointerMode)
	{
		case PmORBIT:  // Orbit the Camera around the aim point
			{
				E3dModel*  LCameraModel = _CameraModel;
				E3dCoordinate  LLon = _CameraTransformBase.Lon - ((E3dCoordinate)LPtrXDiff) * E3dp_Prefs.OrbitStepLongitude;
				E3dCoordinate  LLat = _CameraTransformBase.Lat  + ((E3dCoordinate)LPtrYDiff) * E3dp_Prefs.OrbitStepLatitude;
				if(LLat < -89.0) LLat = -89.0;else if(LLat > 89.0) LLat = 89.0;
				if(LLon < 0.0) LLon += 360.0;else if(LLon > 360.0) LLon -= 360.0;

//printf("%s() %s:%d  % 7.2f,% 7.2f,% 7.2f\n", __FUNCTION__, __FILE__, __LINE__, XYZ_s(LCameraModel->TransformSet.Translation));fflush(stdout);
				E3d_Orbit(&_CameraModel->TransformSet.Translation, &_CameraAimPoint,  LLon, LLat);
				const E3dCoord3  LUpVector = { 0.0, 1.0, 0.0 };E3dCamera_LookAt(NULL, LCameraModel,  &LCameraModel->TransformSet.Translation, &_CameraAimPoint, &LUpVector,  0.0);
			}
		break;

		case PmTRACK:
			{
				E3dModel*  LCameraModel = _CameraModel;
				E3dCoord3  LMove;E3d_Track(LCameraModel->LocalToWorldMatrix, _CameraTransformBase.AimDistance, _Camera->FieldOfView, (E3dCoordinate)_3DWindow->Size.Y, E3dp_Prefs.TrackStep,  LPtrXDiff, LPtrYDiff,  &LMove);

				LCameraModel->TransformSet.Translation.X = _CameraTransformBase.Position.X + LMove.X;
				LCameraModel->TransformSet.Translation.Y = _CameraTransformBase.Position.Y + LMove.Y;
				LCameraModel->TransformSet.Translation.Z = _CameraTransformBase.Position.Z + LMove.Z;

				_CameraAimPoint.X = _CameraTransformBase.AimPoint.X + LMove.X;
				_CameraAimPoint.Y = _CameraTransformBase.AimPoint.Y + LMove.Y;
				_CameraAimPoint.Z = _CameraTransformBase.AimPoint.Z + LMove.Z;
				const E3dCoord3  LUpVector = { 0.0, 1.0, 0.0 };E3dCamera_LookAt(NULL, LCameraModel,  &LCameraModel->TransformSet.Translation, &_CameraAimPoint, &LUpVector,  0.0);
			}
		break;
	}
}


//================================
// Mouse button callback
//================================
static void mouse_button_callback(GLFWwindow* PWindow, const int PButton, const int PAction, const int PModifiers)
{
	double  LPtrX, LPtrY;glfwGetCursorPos(PWindow, &LPtrX, &LPtrY);

	switch(PButton)
	{
		case GLFW_MOUSE_BUTTON_LEFT:
			{
				if(PAction == GLFW_PRESS)
				{
					if(_CameraModel == NULL) return;
					_PointerMode = PmORBIT;
					_BasePtrX = LPtrX;
					_BasePtrY = LPtrY;

					E3d_CalcOrbit(&_CameraModel->TransformSet.Translation, &_CameraAimPoint,  &_CameraTransformBase.Lon, &_CameraTransformBase.Lat, &_CameraTransformBase.AimDistance);
//printf("%s() %s:%d  %lf %lf\n", __FUNCTION__, __FILE__, __LINE__, _CameraTransformBase.Lon, _CameraTransformBase.Lat);
				}
				else if(_PointerMode == PmORBIT) _PointerMode = 0;
			}
		break;

		case GLFW_MOUSE_BUTTON_MIDDLE:
			{
				if(PAction == GLFW_PRESS)
				{
					if(_CameraModel == NULL) return;
					_PointerMode = PmTRACK;
					_BasePtrX = LPtrX;
					_BasePtrY = LPtrY;

					_CameraTransformBase.Position.X = _CameraModel->LocalToWorldMatrix[M30];
					_CameraTransformBase.Position.Y = _CameraModel->LocalToWorldMatrix[M31];
					_CameraTransformBase.Position.Z = _CameraModel->LocalToWorldMatrix[M32];
					_CameraTransformBase.AimPoint = _CameraAimPoint;
//printf("%s() %s:%d  %lf %lf\n", __FUNCTION__, __FILE__, __LINE__, _CameraTransformBase.Lon, _CameraTransformBase.Lat);
				}
				else if(_PointerMode == PmTRACK) _PointerMode = 0;
			}
		break;
	}
}


//================================
// Scroll wheel callback
//================================
static void scroll_callback(GLFWwindow* PWindow,  const double PX, const double PY)
{
#ifdef EWASM
	const float  LMult = _Browser == BrCHROME ? -0.1 : -3.0;
#else
	const float  LMult = 5.0;
#endif
	const E3dCoordinate  LAimDistance = E3d_Dolly(_CameraTransformBase.AimDistance, PY * E3dp_Prefs.DollyStep3 * E3dp_Prefs.DollyStep * LMult);if(LAimDistance > 0.0)
	{
		const E3dCoord3  LCameraPosition = _CameraModel->TransformSet.Translation;
		const E3d_Sub3i_vv(LDiff,  LCameraPosition, _CameraAimPoint);const E3dCoordinate  LDistSq = E3d_Dot3_vv(LDiff, LDiff);

		const E3dCoordinate  LDistInv = LDistSq > 0.0 ? 1.0 / E3d_Sqrt(LDistSq) : 1.0;
		const E3d_Mul3i_vs (LAimDirection,  LDiff, -LDistInv);
		const E3dCoord3  LNewPosition = { _CameraAimPoint.X + LAimDirection.X * -LAimDistance, _CameraAimPoint.Y + LAimDirection.Y * -LAimDistance, _CameraAimPoint.Z + LAimDirection.Z * -LAimDistance };

		const E3dCoord3  LUpVector = { 0.0, 1.0, 0.0 };E3dCamera_LookAt(NULL, _CameraModel,  &LNewPosition, &_CameraAimPoint, &LUpVector,  0.0);

		_CameraTransformBase.AimDistance = LAimDistance;
//printf("%s() %s:%d  % 7.2f,% 7.2f,% 7.2f\n", __FUNCTION__, __FILE__, __LINE__, XYZ_s(_CameraModel->TransformSet.Translation));fflush(stdout);
	}
}


#ifndef EWASM
static void _PrintHelp(void)
{
	printf("%slibE3D Fusion engine %s%s sample code. libE3D & EQUINOX-3D are (TM) and (C) 1994-2021 by Gabor Nagy.\nYou may freely modify and redistribute this sampe code, as long as the copyright notices are left intact.\n", ECyanCode, FeVERSION, ENormalCode);fflush(stdout);
	printf("User controls:\n");fflush(stdout);
	printf("%sLeft mouse button%s: Orbit camera around the origin\n", EGreenCode, ENormalCode);fflush(stdout);
	printf("%sScroll wheel%s:      Dolly camera\n", EGreenCode, ENormalCode);fflush(stdout);
	printf("%sR%s:                 Cycle trhough different renderers\n", EGreenCode, ENormalCode);fflush(stdout);
	printf("%sG%s:                 Toggle X-Z grid (ground plane)\n", EGreenCode, ENormalCode);fflush(stdout);
	printf("%sQ or Esc%s:          Exit\n", EGreenCode, ENormalCode);fflush(stdout);
	printf("You can specify the name of a 3D file name to load, in the command line. If none is given, the default file (%s) is loaded.\n", _DefaultAssetFileName);fflush(stdout);
}
#endif


static void _RendererSetUp(E3dRenderer* PRenderer, E3dModel* PCameraModel, E3dCamera* PCamera, const EcRGBAf32Color* PBackgroundColor)
{
	E_AssignToMember(PCamera, PRenderer,Camera); E_AssignToMember(PCameraModel, PRenderer,CameraModel);PRenderer->BackgroundRGBAfColor = *PBackgroundColor;
	ERes_NamedResourceSet1i(PRenderer, "SSAA", 4);
//	E3dRenderer_SetSSAA(PRenderer, 4);
	PRenderer->ShadowsOn = FALSE;
}


//========================================
// Read the preferences of a plugin
//========================================
void App_ReadSettings(const char* PName, EResource* PResources, void* PStructAddress)
{
	char  LTmpStr[MAXPATHLEN + 1];sprintf(LTmpStr, "%s/.fusionengine/%s", (char*)getenv("HOME"), PName);
	if(!ERes_ReadASCII(LTmpStr, PResources, PStructAddress))
	{
		E3d_Log(ELogWARNING, "Reading plugin setup file", "Could not open file: %s in either %s/.fusionengine", PName, getenv("HOME"));
	}
}


//========================================
// Save the preferences of a plugin
//========================================
void App_SaveSettings(const char* PName, EResource* PResources, void* PStructAddress)
{
	char  LTmpStr[MAXPATHLEN + 1];sprintf(LTmpStr, "%s/.fusionengine/%s", (char*)getenv("HOME"), PName);
	if(!ERes_SaveASCII(LTmpStr, "FusionEngine(TM) plugin settings\n", PResources, PStructAddress))
	{
		E3d_Log(ELogWARNING, "Saving plugin setup file", "Could not open file: %s in either %s/.fusionengine", PName, getenv("HOME"));
	}
}


//========================================================
// Main
//========================================================
int main(const int PArgs_Count, const char* const* PArgs)
{
	E_MallocInit();

	{
		uint32_t  LN = 0;const char**  glfwExtensions;glfwExtensions = glfwGetRequiredInstanceExtensions(&LN);
		printf("%s() %s:%d  %sExtensions%s %d\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode, LN);fflush(stdout);
		uint32_t LC = 0;for(;LC < LN;LC++)
		{
			printf("%s() %s:%d  %s%s%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, glfwExtensions[LC], ENormalCode);fflush(stdout);
		}
	}

#ifdef EWASM
	printf("%s() %s:%d  FusionEngine build %s\n", __FUNCTION__, __FILE__, __LINE__, FeVERSION);fflush(stdout);
#else
	_PrintHelp();

// Initialize paths
//
	E3d_LaunchPath   = E_LaunchPath(PArgs[0]);
	E_ExecutableName = EStrDup(PArgs[0]);
#endif

	E3dp_PreferencesInit(PArgs_Count, PArgs);

#ifndef EWASM
// Initialize working versions of paths (the ones saved in the preferences may be NULL, indicating "use default" (relative to the launch location)
//
	if(E3dp_Prefs.ResourcePath) E3dp_Prefs.WorkingResourcePath = EStrDup(E3dp_Prefs.ResourcePath);
	else
	{
		E3dp_Prefs.WorkingResourcePath = EStrDup(E3d_LaunchPath);
	}
#endif

	E3dGL_GLRenderersRegister();

// Create the main 3D Scene. We create an empty Scene first, so we can start rendering while assets are being loaded.
//
	E3dScene*  LEmptyScene = E3dScene_New("emtpy");if(LEmptyScene == NULL) return(-1);
	E3dScene*  LScene = E3dScene_New("default");if(LScene == NULL) return(-1);
	_Scene = LScene;

// Create a Camera with a Model that instantiates it (rather than just directly hacking view transforms for OpenGL).
// This will allow us to use any renderer, as well as doing stereoscopic rendering (e.g for VR).
//
	E3dModel*   LCameraModel;E_New(E3dModel, LCameraModel);
	E3dCamera*  LCamera = (E3dCamera*)E3dModel_AddObject(LCameraModel, E3dCamera_class);

	E_Assign(LCameraModel, _CameraModel);
	E_Assign(LCamera, _Camera);

	{
		const E3dCoord3  LPosition = { 0.0, 12.0, 42.0 }, LAimPoint = { 0.0, 0.0, 0.0 }, LUpVector = { 0.0, 1.0, 0.0 };
		E3dCamera_LookAt(LCamera, LCameraModel,  &LPosition, &LAimPoint, &LUpVector,  0.0);
		E3d_CalcOrbit(&_CameraModel->TransformSet.Translation, &_CameraAimPoint,  &_CameraTransformBase.Lon, &_CameraTransformBase.Lat, &_CameraTransformBase.AimDistance);
	}


// Read a file.
// Make sure you load triangulated geometry, as the engine doesn't have the complex triangulator algorithm used by EQUINOX-3D.
// No game or engine we know of supports loading concave n-gons with holes, as it would take longer to load/process 3D content.
//
#ifdef EWASM  // WebAssembly
	const char*  LInFileName = _DefaultAssetFileName;
#else
		const char*  LInFileName;if(PArgs_Count > 1) LInFileName = PArgs[1];else LInFileName = _DefaultAssetFileName;
#endif  // EWASM
	_AssetFileName = EStrDup(LInFileName);


	E3d_Verbosity = E3dVbALL;  // Turn on libE3D messages, so we can find potential issues with loading assets

// Initialize lib2D and libE3D
//
	Ep_Init();
	E3d_Init();

	_LoadPlugins();

// Create a 3DWindow and an OpenGL renderer. Note that you can create multiple renderers and switch between them!
// These Renderers are built into the (optional, but statically linked) OpenGL module of libE3D. Others, like the Cg, GLSL, or ray-tracer are plugins
//
	_GPUContext = (E3dGPUContext*)E3dGLContext_New();
	{ char  LTitle[256];sprintf(LTitle, "libE3D Fusion engine %s - sample app", FeVERSION);_3DWindow = E3dWindow_New(LTitle);E_Assign(_3DWindow, _3DWindow); }
	_3DWindow->GPUContext = _GPUContext;
printf("%s() %s:%d  E3dWindow created  GLContext %p\n", __FUNCTION__, __FILE__, __LINE__, _GPUContext);fflush(stdout);
	_Renderers[_NRenderers] = E3dRenderer_New(E3dWireAndZRenderer_class);_NRenderers++;     // Wireframe renderer with Z-buffer
	_Renderers[_NRenderers] = E3dRenderer_New(E3dHiddenLineRenderer_class);_NRenderers++;   // Hidden line renderer
//	_Renderers[_NRenderers] = E3dRenderer_New(E3dShadedSolidRenderer_class);_NRenderers++;  // Solid renderer with shading, using plain OpenGL
	{	E3dRendererClass*  LClass = E3dRenderer_ClassFindByName("GLSL");if(LClass) { _Renderers[_NRenderers] = E3dRenderer_New(LClass);_NRenderers++; } }        // Advanced GPU renderer with GLSL shaders
	_RendererIndex = _NRenderers - 1;
printf("%s() %s:%d  Renderers created: %d\n", __FUNCTION__, __FILE__, __LINE__, _NRenderers);fflush(stdout);

// Set the Camera on the Renderer.
// AssignToMember allows us to track/debug object references precisely, with the powerful, optional object-tracker / garbage collector.
// This tracker is completely compiled out in release builds, so there is zero overhead.
// When active, it can tell you things, like object X of class Y is referenced from data member M of object Z, where it was created (source file / line) and where the assignment happened (source file / line)
//
	EcRGBAf32Color  LBackgroundColor = { 0.4, 0.4, 0.4, 1.0 };
//	EcRGBAf32Color  LBackgroundColor = { 0.0, 0.0, 0.0, 1.0 };
	{ EcRGBAiColor  LBGColor;Ec_RGBAf32ToRGBAi(&LBackgroundColor, &LBGColor);E3dp_UpdateGridColors(&LBGColor); }
	{ EIndex  LC = 0;for(;LC < _NRenderers;LC++) _RendererSetUp(_Renderers[LC], LCameraModel, LCamera, &LBackgroundColor); }

//	_3DWindow->GridSettings.GridXZOn = FALSE;  // By default, the X-Z grid (ground plane) is shown. Here's how to turn it off

	glfwSetErrorCallback(error_callback);


// Load assets
	E3d_Verbosity |= E3dVbOPENGL;

#ifdef EWASM  // WebAssembly
//		pthread_create(&_LoadingThread, 0, _LoadTestScene, LScene);
#else
//		_LoadPTestScene(LScene);
#endif



  // Start GLFW
  if(!glfwInit())
  {
    fprintf(stderr, "Error: GLFW Initialization failed.");
    force_exit();
  }

	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);

  // Create the display window

//	glfwWindowHint(GLFW_STENCIL_BITS, 4);
//	glfwWindowHint(GLFW_SAMPLES, 4);  // 4x MSAA
	glfwWindowHint(GLFW_RESIZABLE, GL_TRUE);

	_GLFWWindow = glfwCreateWindow(800, 450, _3DWindow->Name, NULL, NULL);if(!_GLFWWindow)
	{
    fprintf(stderr, "Error: GLFW Window Creation Failed");
		glfwTerminate();
		force_exit();
	}

//glfwSetFramebufferSizeCallback(_GLFWWindow, reshape);
	glfwSetKeyCallback(_GLFWWindow, key_callback);
	glfwSetMouseButtonCallback(_GLFWWindow, mouse_button_callback);
	glfwSetScrollCallback(_GLFWWindow, scroll_callback);
	glfwSetCursorPosCallback(_GLFWWindow, cursor_position_callback);

  // Set the _GLFWWindow as the drawing destination
	glfwMakeContextCurrent(_GLFWWindow);


// Now, we can initialize the OpenGL back-end of libE3D.
// This is not done in E3d_Init, because the OpenGL module doesn't even need to be linked to simple libE3D applications (such as 3D file format converters).
// Also, it requires a valid OpenGL context, so we do this after calling glfwMakeContextCurrent().
//
	E3dGL_Init();

// Make the empty Scene active, until we load assets
	E3d_Scene = LEmptyScene;

	// Run the main loop
#ifdef EWASM
  emscripten_set_main_loop(_RenderFrame, 0, FALSE);
#else
  // Display the _GLFWWindow until ESC is pressed
  while(!glfwWindowShouldClose(_GLFWWindow))
  {
    _RenderFrame();
  }
  // Clean up
  glfwDestroyWindow(_GLFWWindow);
  glfwTerminate();
  exit(EXIT_SUCCESS);
#endif
}
