//================================================================
// Show/hide things
//
// Plugin for EQUINOX-3D
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-20 19:03:30
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Label.h>
#include <Xe/PushButton.h>

#include <E3D/E3D.h>

#include <E3D/UI/Panel.h>


static EguiItem          _MenuButtons[16];static int  _MenuButtons_Count = 0;
static EScriptFunction*  _ScriptFunctions[32];static int  _ScriptFunctions_Count = 0;


// Operations
//
enum
{
	CR_HIDE_TOGGLE_SELECTED = 1,
	CR_HIDE_SELECTED,
	CR_HIDE_UNSELECTED,
	CR_UNHIDE_SELECTED,
	CR_UNHIDE_ALL
};


#define M_ObjectGetChangeFlagsByClass(mObject)\
{\
	if(EObject_IsSubClass(mObject, E3dGeometry))\
	{\
		if(EObject_IsSubClass(mObject, E3dPointSet))\
		{\
			LChanged |= E3dChgPOINTSET;\
		}\
		else if(EObject_IsSubClass(mObject, E3dSpline))\
		{\
			LChanged |= E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY;\
		}\
		else if(EObject_IsSubClass(mObject, E3dMesh))\
		{\
			E3dMesh*	lMesh = (E3dMesh*)mObject;\
\
			if(lMesh->PolyGroups) LChanged |= E3dChgMESHSHAPE;\
			else LChanged |= E3dChgWIREFRAME;\
		}\
		else if(EObject_IsSubClass(mObject, E3dFace))\
		{\
			LChanged |= E3dChgFACE;\
		}\
	}\
	else if(EObject_IsSubClass(LObject, E3dLight)) LChanged |= E3dChgLIGHT;\
	else if(EObject_IsSubClass(LObject, E3dCamera))\
	{\
		E3dCamera*	lCamera = (E3dCamera*)mObject;\
\
		if(E3dp_Main3DWindow && lCamera == E3dp_Main3DWindow->PerspectiveView->VCamera) LChanged |= E3dChgCAMERA;\
	}\
}


static EChangeFlags _ModelGetChangeFlagsByObjects(E3dModel* PModel)
{
	E3dObject**   LObjects = PModel->Objects;
	EChangeFlags	LChanged = 0;

	EIndex  LObjC = 0;const EIndex  LObjN = PModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
	{
		E3dObject*	LObject = LObjects[LObjC];

		M_ObjectGetChangeFlagsByClass(LObject);
	}

	return(LChanged);
}


//========================================
// Object display/hide
//========================================
static EChangeFlags _HideToggleSelection(void)
{
	E3dScene*     LScene = E3d_Scene;
	EIndex        LNodesHid = 0, LNodesUnhid = 0,
	              LObjectsHid = 0, LObjectsUnhid = 0,
	              LPolyGroupsHid = 0, LPolyGroupsUnhid = 0;
	EChangeFlags  LChanged = 0;


	EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*  LNode = LScene->RootNodes[LC];for(;LNode;LNode = LNode->Next)
		{
			switch(LNode->Selection)
			{
				case E3dNodeSelNODE:
				case E3dNodeSelBRANCH:
				case E3dNodeSelBRANCH_ROOT:
					if(LNode->Visible)
					{
						LNode->Visible = FALSE;LNodesHid++;
					}
					else
					{
						LNode->Visible = TRUE;LNodesUnhid++;
					}

					{
						E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);
						if(LModel) LChanged |= _ModelGetChangeFlagsByObjects(LModel);
					}
				break;
			}
		}
	}


	{
		E3dObject**  LSelectedObjects;const EIndex  LNumOfSelectedObjects = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemALL, E3dSELECTED|E3dEXCLUDE_MODEL_SELECTION, FALSE);if(LSelectedObjects)
		{
			for(LC = 0;LC < LNumOfSelectedObjects;LC++)
			{
				E3dObject*  LObject = LSelectedObjects[LC];

				if(EObject_IsSubClass(LObject, E3dMesh))
				{
					if(LObject->Selection == E3dNodeSelOBJECT)
					{
						if(LObject->Visible)
						{
							LObject->Visible = FALSE;LObjectsHid++;
						}
						else
						{
							LObject->Visible = TRUE;LObjectsUnhid++;
						}
						M_ObjectGetChangeFlagsByClass(LObject);
					}
					else
					{
						E3dMesh*        LMesh = (E3dMesh*)LObject;
						E3dPolyGroup**  LPolyGroups = LMesh->PolyGroups;
						EIndex   LGCnt = 0;const EIndex  LGNum = LMesh->PolyGroups_Count;for(;LGCnt < LGNum;LGCnt++)
						{
							E3dPolyGroup*  LPolyGroup = LPolyGroups[LGCnt];

							if(LPolyGroup->Selected)
							{
								if(LPolyGroup->Visible)
								{
									LPolyGroup->Visible = FALSE;LPolyGroupsHid++;
								}
								else
								{
									LPolyGroup->Visible = TRUE;LPolyGroupsUnhid++;
								}
								M_ObjectGetChangeFlagsByClass(LObject);
							}
						}
					}
				}
				else
				{
					if(LObject->Selection == E3dNodeSelOBJECT)
					{
						if(LObject->Visible)
						{
							LObject->Visible = FALSE;LObjectsHid++;
						}
						else
						{
							LObject->Visible = TRUE;LObjectsUnhid++;
						}
						M_ObjectGetChangeFlagsByClass(LObject);
					}
				}
			}
			EFree(LSelectedObjects);
		}
	}

	if(LNodesHid > 0 || LNodesUnhid > 0 || LObjectsHid || LObjectsUnhid || LPolyGroupsHid || LPolyGroupsUnhid)
	{
		E3dCallbackStruct  LCBS;

		LCBS.Reason = E3dCR_HIDE_UNHIDE;

		E3dScene_CallCallbacks((ECoreObject*)LScene,  LScene->Callbacks, LScene->Callbacks_Count,  &LCBS);
	}

	return(LChanged);
}


//========================================
// Object display/hide
//========================================
static EChangeFlags _HideUnhide(const int POperation)
{
	E3dScene*     LScene = E3d_Scene;
	EIndex        LNodesHid = 0, LNodesUnhid = 0,
	              LObjectsHid = 0, LObjectsUnhid = 0,
	              LPolyGroupsHid = 0, LPolyGroupsUnhid = 0;
	EChangeFlags  LChanged = 0;


	EIndex  LRC = 0;const EIndex  LRN = LScene->RootNodes_Count;for(;LRC < LRN;LRC++)
	{
		E3dNode*  LNode = LScene->RootNodes[LRC];for(;LNode;LNode = LNode->Next)
		{
			switch(POperation)
			{
				case CR_HIDE_SELECTED:
					switch(LNode->Selection)
					{
						case E3dNodeSelNODE:
						case E3dNodeSelBRANCH:
						case E3dNodeSelBRANCH_ROOT:
							if(LNode->Visible)
							{
								LNode->Visible = FALSE;LNodesHid++;

								{
									E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);
									if(LModel) LChanged |= _ModelGetChangeFlagsByObjects(LModel);
								}
							}
						break;

						default:
							{
								E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
								{
									E3dObject**  LObjects = LModel->Objects;
									EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
									{
										E3dObject*  LObject = LObjects[LObjC];

										if(LObject->Selection == E3dNodeSelOBJECT)
										{
											if(LObject->Visible)
											{
												LObject->Visible = FALSE;LObjectsHid++;M_ObjectGetChangeFlagsByClass(LObject);
											}
										}
										else
										{
											if(EObject_IsSubClass(LObject, E3dMesh))
											{
												E3dMesh*        LMesh = (E3dMesh*)LObject;
												E3dPolyGroup**  LPolyGroups = LMesh->PolyGroups;
												EIndex  LGC = 0;const EIndex  LGN = LMesh->PolyGroups_Count;for(;LGC < LGN;LGC++)
												{
													E3dPolyGroup*  LPolyGroup = LPolyGroups[LGC];if(LPolyGroup->Selected)
													{
														if(LPolyGroup->Visible)
														{
															LPolyGroup->Visible = FALSE;LPolyGroupsHid++;M_ObjectGetChangeFlagsByClass(LObject);
														}
													}
												}
											}
										}
									}
								}
							}
						break;
					}
				break;

				case CR_HIDE_UNSELECTED:
					switch(LNode->Selection)
					{
						case E3dNodeSelNONE:
							if(LNode->Visible)
							{
								LNode->Visible = FALSE;LNodesHid++;

								{
									E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);
									if(LModel) LChanged |= _ModelGetChangeFlagsByObjects(LModel);
								}
							}
						break;

						case E3dNodeSelOBJECT:
							{
								E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
								{
									E3dObject**   LObjects = LModel->Objects;
									EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
									{
										E3dObject*  LObject = LObjects[LObjC];

										switch(LObject->Selection)
										{
											case E3dObjectSelNONE:
												if(LObject->Visible)
												{
													LObject->Visible = FALSE;LObjectsHid++;M_ObjectGetChangeFlagsByClass(LObject);
												}
											break;

											case E3dObjectSelPOLYGROUP:
												if(EObject_IsSubClass(LObject, E3dMesh))
												{
													E3dMesh*        LMesh = (E3dMesh*)LObject;
													E3dPolyGroup**  LPolyGroups = LMesh->PolyGroups;

													EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
													{
														E3dPolyGroup*  LPolyGroup = LPolyGroups[LPGC];

														if(!LPolyGroup->Selected)
														{
															if(LPolyGroup->Visible)
															{
																LPolyGroup->Visible = FALSE;LPolyGroupsHid++;M_ObjectGetChangeFlagsByClass(LObject);
															}
														}
													}
												}
											break;
										}
									}
								}
							}
						break;
					}
				break;

				case CR_UNHIDE_SELECTED:
					if(LNode->Selection != E3dNodeSelNONE)
					{
						E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);

						const EBool  LNodeSelected = E3dNode_IsSelected(LNode);
						if(LNodeSelected && !LNode->Visible)
						{
							LNode->Visible = TRUE;LNodesUnhid++;

							if(LModel) LChanged |= _ModelGetChangeFlagsByObjects(LModel);
						}

						if(LModel)
						{
							E3dObject**  LObjects = LModel->Objects;

							EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
							{
								E3dObject*  LObject = LObjects[LObjC];

								if(LNodeSelected || LObject->Selection)
								{
									if(!LObject->Visible)
									{
										LObject->Visible = TRUE;LObjectsUnhid++;M_ObjectGetChangeFlagsByClass(LObject);
									}

// If all Nodes of this Object are hidden, un-hide the first one
//
									if(LObject->Models_Count)
									{
										EIndex  LC = 0;for(;LC < LObject->Models_Count;LC++)
										{
											if(LObject->Models[LC]->Visible) break;
										}
										if(LC == LObject->Models_Count) LObject->Models[0]->Visible = TRUE;
									}

									if(EObject_IsSubClass(LObject, E3dMesh))
									{
										E3dMesh*        LMesh = (E3dMesh*)LObject;
										E3dPolyGroup**  LPolyGroups = LMesh->PolyGroups;
										EIndex  LGCnt = 0;const EIndex  LGNum = LMesh->PolyGroups_Count;for(;LGCnt < LGNum;LGCnt++)
										{
											E3dPolyGroup*  LPolyGroup = LPolyGroups[LGCnt];

											if(LObject->Selection == E3dObjectSelPOLYGROUP)
											{
												if(LPolyGroup->Selected)
												{
													if(!LPolyGroup->Visible)
													{
														LPolyGroup->Visible = TRUE;LPolyGroupsUnhid++;M_ObjectGetChangeFlagsByClass(LObject);
													}
												}
											}
											else if(!LPolyGroup->Visible)
											{
												LPolyGroup->Visible = TRUE;LPolyGroupsUnhid++;M_ObjectGetChangeFlagsByClass(LObject);
											}
										}
									}
								}
							}
						}
					}
				break;

				case CR_UNHIDE_ALL:
					{
						E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);

						if(!LNode->Visible)
						{
							LNode->Visible = TRUE;LNodesUnhid++;

							if(LModel) LChanged |= _ModelGetChangeFlagsByObjects(LModel);
						}

						if(LModel)
						{
							E3dObject**  LObjects = LModel->Objects;
							EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
							{
								E3dObject*  LObject = LObjects[LObjC];

								if(!LObject->Visible)
								{
									LObject->Visible = TRUE;LObjectsUnhid++;M_ObjectGetChangeFlagsByClass(LObject);
								}

								if(EObject_IsSubClass(LObject, E3dMesh))
								{
									E3dMesh*        LMesh = (E3dMesh*)LObject;
									E3dPolyGroup**  LPolyGroups = LMesh->PolyGroups;
									EIndex  LGCnt = 0;const EIndex  LGNum = LMesh->PolyGroups_Count;for(;LGCnt < LGNum;LGCnt++)
									{
										E3dPolyGroup*  LPolyGroup = LPolyGroups[LGCnt];if(!LPolyGroup->Visible)
										{
											LPolyGroup->Visible = TRUE;LPolyGroupsUnhid++;
											M_ObjectGetChangeFlagsByClass(LObject);
										}
									}
								}
							}
						}
					}
				break;
			}
		}
	}

	if(LNodesHid > 0 || LNodesUnhid > 0 || LObjectsHid || LObjectsUnhid || LPolyGroupsHid || LPolyGroupsUnhid)
	{
		E3dCallbackStruct  LCBS;

		LCBS.Reason = E3dCR_HIDE_UNHIDE;
		E3dScene_CallCallbacks((ECoreObject*)LScene, LScene->Callbacks, LScene->Callbacks_Count, &LCBS);
	}

	return(LChanged);
}


//========================================
// Menu callback
//========================================
static void _MCB_ModelDisplay(Widget PW, EPointer PClientData, EPointer PCallData)
{
	EChangeFlags  LChanged = 0;

	switch((EPointerInt)PClientData)
	{
		case CR_HIDE_TOGGLE_SELECTED:  LChanged = _HideToggleSelection();break;
		default:                       LChanged = _HideUnhide((EPointerInt)PClientData);break;
	}

//{ char	str[256];printf("%s() [%s]\n", __FUNCTION__, E3d_SceneChangeFlagsToString(LChanged, str));fflush(stdout); }
	if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged | E3dChgSCHEMATICS);
}


static void _MCB_ToggleHideSelectedEdges(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	EIndex     LTotalEdgesChanged = 0;

	E3dObject**  LSelectedObjects;const EIndex  LNumOfSelectedObjects = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LSelectedObjects)
	{
		EIndex  LC = 0;for(;LC < LNumOfSelectedObjects;LC++)
		{
			E3dMesh*  LMesh = (E3dMesh*)(LSelectedObjects[LC]);
			EIndex    LMeshEdgesChanged = 0;
			EIndex  LEC = 0;const EIndex  LEN = LMesh->Edges_Count;E3dEdge*  LEdge = LMesh->Edges;for(;LEC < LEN;LEC++, LEdge++)
			{
				if(LEdge->Flags & E3dEdgeSELECTED)
				{
					if(LEdge->Flags & E3dEdgeVISIBLE) LEdge->Flags &= ~E3dEdgeVISIBLE;
					else LEdge->Flags |= E3dEdgeVISIBLE;
					LMeshEdgesChanged++;
				}
			}
			if(LMeshEdgesChanged)
			{
				E3dMesh_EdgeFlagsUpdate(LMesh);
				LTotalEdgesChanged += LMeshEdgesChanged;
			}
		}
		EFree(LSelectedObjects);
	}

	if(LTotalEdgesChanged) E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
}


static void _MCB_HideSelectedEdges(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	EIndex     LTotalEdgesChanged = 0;

	E3dObject**  LSelectedObjects;const EIndex  LNumOfSelectedObjects = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LSelectedObjects)
	{
		EIndex  LC = 0;for(;LC < LNumOfSelectedObjects;LC++)
		{
			E3dMesh*  LMesh = (E3dMesh*)(LSelectedObjects[LC]);
			EIndex    LMeshEdgesChanged = 0;
			EIndex  LEC = 0;const EIndex  LEN = LMesh->Edges_Count;E3dEdge*  LEdge = LMesh->Edges;for(;LEC < LEN;LEC++, LEdge++)
			{
				if(LEdge->Flags & E3dEdgeSELECTED)
				{
					if(LEdge->Flags & E3dEdgeVISIBLE)
					{
						LEdge->Flags &= ~E3dEdgeVISIBLE;
						LMeshEdgesChanged++;
					}
				}
			}
			if(LMeshEdgesChanged)
			{
				E3dMesh_EdgeFlagsUpdate(LMesh);
				LTotalEdgesChanged += LMeshEdgesChanged;
			}
		}
		EFree(LSelectedObjects);
	}

	if(LTotalEdgesChanged) E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
}


static EChangeFlags _Hide_SelectedPoints(E3dScene* PScene)
{
	EIndex  LTotalPointsChanged = 0;

	E3dObject**  LSelectedObjects;const EIndex  LNumOfSelectedObjects = E3dScene_GetObjects(PScene, &LSelectedObjects, E3dItemMESH|E3dItemSPLINE, E3dSELECTED, FALSE);if(LSelectedObjects)
	{
		EIndex  LC = 0;for(;LC < LNumOfSelectedObjects;LC++)
		{
			E3dObject*  LObject = LSelectedObjects[LC];if(EObject_IsSubClass(LObject, E3dMesh))
			{
				E3dMesh*  LMesh = (E3dMesh*)LObject;
				EIndex    LPointsChanged = 0;
				EIndex  LVC = 0;const EIndex  LVN = LMesh->Vertices_Count;E3dVertex*  LVertex = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LVertexSize))
				{
					if(LVertex->Flags & E3dVtxSELECTED)
					{
						if(LVertex->Flags & E3dVtxVISIBLE)
						{
							LVertex->Flags &= ~E3dVtxVISIBLE;
							LPointsChanged++;
						}
					}
				}
				LTotalPointsChanged += LPointsChanged;
			}
			else if(EObject_IsSubClass(LObject, E3dSpline))
			{
printf("%s() %s:%d  %sIMPLEMENT ME%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
			}
		}
		EFree(LSelectedObjects);
	}

	if(LTotalPointsChanged) return(E3dChgWIREFRAME);
	else return(0);
}


static void _MCB_Hide_SelectedPoints(Widget PW, EPointer PClientData, EPointer PCallData)
{
	EChangeFlags  LSceneChangeFlags = _Hide_SelectedPoints(E3d_Scene);
	if(LSceneChangeFlags) E3dWindows_RedrawByChangeFlags(LSceneChangeFlags);
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Display", "Toggle hide/un-hide selected", '\0', NULL, NULL, FALSE, NULL, _MCB_ModelDisplay, (EPointer)CR_HIDE_TOGGLE_SELECTED);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Display", "Hide selected",       '\0', NULL, NULL, FALSE, NULL, _MCB_ModelDisplay, (EPointer)CR_HIDE_SELECTED);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Display", "Hide unselected",     '\0', NULL, NULL, FALSE, NULL, _MCB_ModelDisplay, (EPointer)CR_HIDE_UNSELECTED);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Display", "Un-hide selected",    '\0', NULL, NULL, FALSE, NULL, _MCB_ModelDisplay, (EPointer)CR_UNHIDE_SELECTED);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Display", "Un-hide all",         '\0', NULL, NULL, FALSE, NULL, _MCB_ModelDisplay, (EPointer)CR_UNHIDE_ALL);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Display", "Toggle hide/un-hide selected Edges", '\0', NULL, NULL, FALSE, NULL, _MCB_ToggleHideSelectedEdges, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Display", "Hide selected Edges",                '\0', NULL, NULL, FALSE, NULL, _MCB_HideSelectedEdges, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Display", "Hide selected Points",               '\0', NULL, NULL, FALSE, NULL, _MCB_Hide_SelectedPoints, (EPointer)0);
	E3dScene_ScriptFunctionRegister("Hide_SelectedPoints", "Hide selected points", "Hide selected Vertices on selected Meshes.", _Hide_SelectedPoints,  _ScriptFunctions + _ScriptFunctions_Count++);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		{ int  LC = 0;for(;LC < _MenuButtons_Count;LC++) { XeItemFree(_MenuButtons[LC]); } _MenuButtons_Count = 0;  }
		{ int  LC = 0;for(;LC < _ScriptFunctions_Count;LC++) { E3dScene_ScriptFunctionRemove(_ScriptFunctions[LC]); } _ScriptFunctions_Count = 0; }
	}

	return(E_SUCCESS);
}
