//================================================================
// EQUINOX-3D(TM) plugin
//
// - Edit Mesh
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 20:41:11
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/E3D.h>
#include <E3D/Math.h>
#include <E3D/Matrix.h>
#include <E3D/Scene.h>

#include <Xe/Label.h>
#include <Xe/Matrix.h>
#include <Xe/PushButton.h>
#include <Xe/Scale.h>
#include <Xe/Separator.h>
#include <Xe/Shell.h>
#include <Xe/TextField.h>
#include <Xe/ToggleButton.h>
#include <Xe/Dialog.h>

#include <Color/Panel.h>
#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage

#include "Edit.h"

static EPluginLock*  _PolyGroup_DialogLock = NULL;


typedef struct
{
	Widget  SemanticW, NameW, ShowW, RemoveW;
} E3dVertexAttributeGUI;

EList_t(E3dVertexAttributeGUI);  // E3dVertexAttributeGUIList


// PolyGroup dialog-box
//
E3dPolyGroup*         _Edit_SelectedPolyGroup = NULL;

static Widget         _PolyGroup_DialogW = NULL;
static Widget         _PolyGroupNameW = NULL,
                      _PolyGroupFollowSelectionBW = NULL, _PolyGroupApplyChangesToAllSelectedW = NULL,
                      _PolyGroupFacetedBW = NULL, _PolyGroupSmoothingAngleW = NULL, _PolyGroupVectorDrawScaleScaleW = NULL,
                      _PolyGroupCastsShadowsBW, _PolyGroupReceivesShadowsBW,
                      _PolyGroupDoubleSidedBW, _PolyGroupWireframeW, _SharpPolygonsW,
                      _PolyGroupGetUVLabelW, _PolyGroupGetUVFromSTMatrixW,
                      _PolyGroupGetUVFromSTWs[E3dMAX_NUM_OF_2DTEXTUREMAPPERS],
                      _PolyGroupGetTangentFromSTWs[E3dMAX_NUM_OF_2DTEXTUREMAPPERS],
                      _PolyGroupRemoveTriangleStripsW,
                      _PolyGroupStatsLabelW = NULL, _PolyGroupStatsValuesLabelW = NULL;

static const int      _PolyGroup_VAStartIndex = 0;

static Widget         _PolyGroup_VertexAttributesMatrixW = NULL;
EListI_s(static, E3dVertexAttributeGUI, _PolyGroup_VertexAttributesGUI);



static EBool          _PolyGroupFollowSelection = TRUE, _PolyGroupApplyChangesToAllSelected = TRUE;



EBool                 _Edit_PolyGroupSceneCallbackAdded = FALSE;


static E3dCoordinate  _PolygonPackingSpacing = 0.01;

static E3dLock  _PolyGroupMeshLock=
{
	"Edit PolyGroup",  // Name
	NULL               // LockProc
};


enum
{
	CR_NAME = XeCR_CUSTOM0,
	CR_TOGGLE_FOLLOW_SELECTION,
	CR_TOGGLE_APPLY_CHANGES_TO_SELECTED,
	CR_TOGGLE_FACETED,
	CR_CHANGE_VECTOR_SCALE,
	CR_TOGGLE_CASTS_SHADOWS,
	CR_TOGGLE_RECEIVES_SHADOWS,
	CR_TOGGLE_DOUBLE_SIDED,
	CR_REMOVE_STRIPS,
	CR_TANGENT_FROM_SHAPE,
};



static void  _PolyGroupUpdateStatsLabel(const E3dPolyGroup* PPolyGroup);


// Inertia tensor
//
typedef struct
{
	E3dCoordinate  XX, YY, ZZ, XY, YZ, XZ;
} E3dInertia;


#define _SubExpressions(w0, w1, w2,  f1, f2, f3,  g0, g1, g2) do\
{\
	E3dCoordinate  temp0 = w0 + w1;\
	f1 = temp0 + w2;\
	E3dCoordinate  temp1 = w0 * w0;\
	E3dCoordinate  temp2 = temp1 + w1 * temp0;\
	f2 = temp2 + w2 * f1;\
	f3 = w0 * temp1 + w1 * temp2 + w2 * f2;\
	g0 = f2 + w0 * (f1 + w0);g1 = f2 + w1 * (f1 + w1); g2 = f2 + w2 * (f1 + w2);\
} while(0)



#define _ProcessTriangle() do\
{\
	const E3dCoordinate  LX0 = LVertex0->Position.X, LY0 = LVertex0->Position.Y, LZ0 = LVertex0->Position.Z;\
	const E3dCoordinate  LX1 = LVertex1->Position.X, LY1 = LVertex1->Position.Y, LZ1 = LVertex1->Position.Z;\
	const E3dCoordinate  LX2 = LVertex2->Position.X, LY2 = LVertex2->Position.Y, LZ2 = LVertex2->Position.Z;\
\
/* Get edges and cross product of edges	*/\
	const E3dCoordinate  a1 = LX1 - LX0,  b1 = LY1 - LY0,  c1 = LZ1 - LZ0,  a2 = LX2 - LX0,  b2 = LY2 - LY0, c2 = LZ2 - LZ0;\
	const E3dCoordinate  d0 = b1 * c2 - b2 * c1,  d1 = a2 * c1 - a1 * c2, d2 = a1 * b2 - a2 * b1;\
\
/* Compute integral terms		*/\
	E3dCoordinate  f1x, f2x, f3x,  f1y, f2y, f3y,  f1z, f2z, f3z,   g0x, g1x, g2x,  g0y, g1y, g2y,  g0z, g1z, g2z;\
\
	_SubExpressions(LX0, LX1, LX2,  f1x, f2x, f3x,  g0x, g1x, g2x);\
	_SubExpressions(LY0, LY1, LY2,  f1y, f2y, f3y,  g0y, g1y, g2y);\
	_SubExpressions(LZ0, LZ1, LZ2,  f1z, f2z, f3z,  g0z, g1z, g2z);\
\
/* Update integrals  */\
	intg[0] += d0 * f1x;\
	intg[1] += d0 * f2x; intg[2] += d1 * f2y;\
	intg[4] += d0 * f3x; intg[5] += d1 * f3y;\
	intg[7] += d0 * (LY0 * g0x + LY1 * g1x + LY2 * g2x);\
	intg[8] += d1 * (LZ0 * g0y + LZ1 * g1y + LZ2 * g2y);\
	intg[9] += d2 * (LX0 * g0z + LX1 * g1z + LX2 * g2z);\
	intg[3] += d2 * f2z;\
	intg[6] += d2 * f3z;\
} while(0)


static int _E3dMesh_VolumeProperties(const E3dMesh* PMesh, E3dCoordinate* PVolume, E3dCoord3* PCenterOfMass, E3dInertia* PInertia)
{
	const E3dCoordinate  mult[10] = { 1.0/6.0, 1.0/24.0, 1.0/24.0, 1.0/24.0, 1.0/60.0, 1.0/60.0, 1.0/60.0, 1.0/120.0, 1.0/120.0, 1.0/120.0 };
	E3dCoordinate        intg[10] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 ,0.0 };  // order: 1, x, y, z, x^2, y^2, z^2, xy, yz, zx

	*PVolume = 0.0;
	E3dPolyGroup* const*  LPolyGroups = PMesh->PolyGroups;
	const E3dVertex*      LVertices = PMesh->Vertices;const E3dVertexSize  LMeshVertexSize = PMesh->VertexSize;
	EIndex  LGC = 0;const EIndex  LGN = PMesh->PolyGroups_Count;for(;LGC < LGN;LGC++)
	{
		const E3dPolyGroup*  LPolyGroup = LPolyGroups[LGC];const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;
		const E3dPolygon*    LPolygon = LPolyGroup->Polygons;

		EIndex  LC = 0;const EIndex  LN = LPolyGroup->Polygons_Count;for(LC = 0;LC < LN;LC++, LPolygon++)
		{
			const E3dVertexNode*  LVertexNodes = LPolygon->VertexNodes;

			if(LPolygon->NumOfVertices == 3)
			{
				const E3dVertex*  LVertex0;const E3dVertex*  LVertex1;const E3dVertex*  LVertex2;
				{ const E3dVertexNode*  LVertexNode = LVertexNodes;LVertex0 = EStruct_ByIndex(LVertices, LMeshVertexSize, LVertexNode->VertexID);EStruct_Inc(LVertexNode, LVertexNodeSize); }
				{ const E3dVertexNode*  LVertexNode = LVertexNodes;LVertex1 = EStruct_ByIndex(LVertices, LMeshVertexSize, LVertexNode->VertexID);EStruct_Inc(LVertexNode, LVertexNodeSize); }
				{ const E3dVertexNode*  LVertexNode = LVertexNodes;LVertex2 = EStruct_ByIndex(LVertices, LMeshVertexSize, LVertexNode->VertexID); }
				_ProcessTriangle();
			}
			else if(LPolygon->ITriangles_Count)
			{
				const E3dITriangle*  LITriangle = LPolygon->ITriangles;if(LITriangle)
				{
					EIndex  lTC = 0;const EIndex  lTN = LPolygon->ITriangles_Count;for(;lTC < lTN;lTC++, LITriangle++)
					{
						const E3dVertexNodeIndex*  LVertexNodeIndices = LITriangle->VertexNodes;
						const E3dVertex*       LVertex0;const E3dVertex*  LVertex1;const E3dVertex*  LVertex2;

						{ const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndices[0]);LVertex0 = EStruct_ByIndex(LVertices, LMeshVertexSize, LVertexNode->VertexID); }
						{ const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndices[1]);LVertex1 = EStruct_ByIndex(LVertices, LMeshVertexSize, LVertexNode->VertexID); }
						{ const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndices[2]);LVertex2 = EStruct_ByIndex(LVertices, LMeshVertexSize, LVertexNode->VertexID); }
						_ProcessTriangle();
					}
				}
			}
		}
	}

	{ int  LC = 0;for(;LC < 10;LC++) intg[LC] *= mult[LC]; }

	const E3dCoordinate  LMass = intg[0];

	if(LMass <= 0.0) return(E_NO_DATA);

	*PVolume = LMass;   // Assuming density = 1.0

	E3dCoord3  LCenterOfMass;
	LCenterOfMass.X = intg[1] / LMass;
	LCenterOfMass.Y = intg[2] / LMass;
	LCenterOfMass.Z = intg[3] / LMass;

// Center of mass
//
	if(PCenterOfMass)
	{
		*PCenterOfMass = LCenterOfMass;
	}

// inertia tensor relative to center of mass
//
	if(PInertia)
	{
		PInertia->XX = intg[5] + intg[6] - LMass * (LCenterOfMass.Y * LCenterOfMass.Y + LCenterOfMass.Z * LCenterOfMass.Z);
		PInertia->YY = intg[4] + intg[6] - LMass * (LCenterOfMass.Z * LCenterOfMass.Z + LCenterOfMass.X * LCenterOfMass.X);
		PInertia->ZZ = intg[4] + intg[5] - LMass * (LCenterOfMass.X * LCenterOfMass.X + LCenterOfMass.Y * LCenterOfMass.Y);
		PInertia->XY = -(intg[7] - LMass * LCenterOfMass.X * LCenterOfMass.Y);
		PInertia->YZ = -(intg[8] - LMass * LCenterOfMass.Y * LCenterOfMass.Z);
		PInertia->XZ = -(intg[9] - LMass * LCenterOfMass.Z * LCenterOfMass.X);
	}

	return(E_SUCCESS);
}


//========================================
// Update Mesh stats label
//========================================
static void _Mesh_UpdateStatsLabel(const E3dMesh* PMesh, Widget PStatsValLW)
{
	E3dCoordinate  LVolume = 0.0;
	const EIndex   LVN = PMesh->Vertices_Count;

	if(LVN)
	{
		if(_E3dMesh_VolumeProperties(PMesh, &LVolume, NULL, NULL) == E_SUCCESS)
		{
//printf("%s() %s:%d Volume = %f\n", __FUNCTION__, __FILE__, __LINE__, LVolume);fflush(stdout);
		}
	}

	E3dPolyGroup**  LPolyGroups = PMesh->PolyGroups;
	char            LTmpStr[256];
	float           LAverageStripLength = 0.0;
	EIndex          LNumOfPolygons, LNumOfTriangles, LNumOfTriangleStrips;

	E3dCoordinate  LSurfaceArea = 0.0;

	EIndex  LMaxVertexNum = 0;

	LNumOfPolygons = LNumOfTriangles = LNumOfTriangleStrips = 0;
	EIndex  LPGC = 0;const EIndex  LPGN = PMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
	{
		const E3dPolyGroup*  LPolyGroup = LPolyGroups[LPGC];

		LNumOfPolygons  += LPolyGroup->Polygons_Count;
		LNumOfTriangles += E3dPolyGroup_CountTriangles(LPolyGroup);

		if(LVN) LSurfaceArea += E3dMeshPolyGroup_Area(PMesh, LPolyGroup);

		LNumOfTriangleStrips += LPolyGroup->TriangleStrips_Count;
		if(LPolyGroup->TriangleStrips_Count)
		{
			E3dTriangleStrip*  LTriangleStrip = LPolyGroup->TriangleStrips;

			EIndex  LC = 0;const EIndex  LNumOfStrips = LPolyGroup->TriangleStrips_Count;for(;LC < LNumOfStrips;LC++, LTriangleStrip++)
			{
				if(LMaxVertexNum < LTriangleStrip->NumOfVertices) LMaxVertexNum = LTriangleStrip->NumOfVertices;
				LAverageStripLength += (float)(LTriangleStrip->NumOfVertices);
			}
		}
	}


	EIndex  LNumOfSelectedVertices = 0;

	{
		EIndex  LVC = 0;E3dVertex*  LVertex = PMesh->Vertices;const E3dVertexSize  LMeshVertexSize = PMesh->VertexSize;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LMeshVertexSize))
		{
			if(LVertex->Flags & E3dVtxSELECTED) LNumOfSelectedVertices++;
		}
	}


	if(LNumOfTriangleStrips)
	{
		LAverageStripLength /= (float)LNumOfTriangleStrips;
		sprintf(LTmpStr, "%lld (%lld selected)\n%lld (%lld selected %lld visible)\n%lld\n%lld (%lld selected)\n%lld\n%lld\n%lld vertices\n%.2f vertices\n%.6f\n%.6f",
		       (long long)LVN, (long long)LNumOfSelectedVertices,
		       (long long)PMesh->Edges_Count, (long long)E3dMesh_NumOfSelectedEdges(PMesh), (long long)E3dMesh_NumOfVisibleEdges(PMesh),
		       (long long)LPGN,
		       (long long)LNumOfPolygons, (long long)E3dMesh_NumOfSelectedPolygons(PMesh),
		       (long long)LNumOfTriangles,
		       (long long)LNumOfTriangleStrips, (long long)LMaxVertexNum, LAverageStripLength,
		       LSurfaceArea, LVolume);
	}
	else
	{
		sprintf(LTmpStr, "%lld (%lld selected)\n%lld (%lld selected %lld visible)\n%lld\n%lld (%lld selected)\n%lld\n0\n-\n-\n%.6f\n%.6f",
		        (long long)LVN, (long long)LNumOfSelectedVertices,
		        (long long)PMesh->Edges_Count, (long long)E3dMesh_NumOfSelectedEdges(PMesh), (long long)E3dMesh_NumOfVisibleEdges(PMesh),
		        (long long)LPGN, (long long)LNumOfPolygons, (long long)E3dMesh_NumOfSelectedPolygons(PMesh), (long long)LNumOfTriangles, LSurfaceArea, LVolume);
	}

	XeString  LXeStr = XeString_NewVa(TRUE,  XeFONT, XeFontDEFAULT,  XeALIGNMENT, XeSTRING_ALIGN_RIGHT,  XeTEXT, LTmpStr,  0);if(LXeStr)
	{
		EArg  Args[2];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNlabelString, LXeStr);
		Xe_SetValues(PStatsValLW,  Args, ArgCnt);

		XeString_Free(LXeStr);
	}
}


typedef struct
{
	Widget         StatsValuesLabelW,
	               WireframeW, ShowAllEdgesW, ShowAllVerticesW,
	               EdgeHardnessW, UseCustomEdgeColorW;

	XeDrawer       ColorDrawer;
	Widget         ColorSwatchW;
	EcColorPanel*  ColorPanel;

	Widget                     VertexAttributesMatrixW;
	E3dVertexAttributeGUIList  VertexAttributesGUI;
} E3dMeshResourcesGUI;


//========================================
// Mesh dialog callback handler
//========================================
static void _DCB_MeshWireframe(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*        LPanel = (E3dObjectPanel*)PClientData;
	E3dMesh*               LMesh = (E3dMesh*)LPanel->Object;
	const E3dDisplayFlags  LDisplayFlags = ((XeToggleButtonCallbackStruct*)PCallData)->Set ? E3dDM_Wireframe : 0;

	if(LPanel->ApplyChangesToAllSelected)
	{
		E3dScene*  LScene = E3d_Scene;
		E3dObject**  LSelectedObjects;const EIndex  LN = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LSelectedObjects)
		{
			EIndex  LC = 0;for(;LC < LN;LC++)
			{
				E3dMesh*  LTMesh = (E3dMesh*)LSelectedObjects[LC];LTMesh->DisplayFlags = LDisplayFlags;
			}
			EFree(LSelectedObjects);
		}
		else LMesh->DisplayFlags = LDisplayFlags;
	}
	else LMesh->DisplayFlags = LDisplayFlags;

	EBool  LMeshVisible = FALSE;

	if(LMesh->Visible)
	{
		EIndex  LC = 0;const EIndex  LN = LMesh->Models_Count;for(;LC < LN;LC++)
		{
			if(LMesh->Models[LC]->Visible) { LMeshVisible = TRUE;break; }
		}
	}

//	E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
	if(LMeshVisible) E3dWindows_RedrawByChangeFlags(E3dChgHIDE_UNHIDE|E3dChgMESHSHAPE);
	else E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
}


//========================================
// Mesh dialog callback handler
//========================================
static void _DCB_EdgeColorEnable(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dMesh*         LMesh = (E3dMesh*)LPanel->Object;

	LMesh->UseCustomEdgeColor = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
}


static void _DCB_MeshShowAllEdges(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dMesh*         LMesh = (E3dMesh*)LPanel->Object;
	const EBool      LV = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	if(LPanel->ApplyChangesToAllSelected)
	{
		E3dScene*  LScene = E3d_Scene;
		E3dObject**  LSelectedObjects;const EIndex  LN = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LSelectedObjects)
		{
			EIndex  LC = 0;for(;LC < LN;LC++)
			{
				E3dMesh*  LTMesh = (E3dMesh*)LSelectedObjects[LC];LTMesh->ShowAllEdges = LV;
			}
			EFree(LSelectedObjects);
		}
		else LMesh->ShowAllEdges = LV;
	}
	else LMesh->ShowAllEdges = LV;

	E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
}


static void _DCB_ShowAllVertices(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dMesh*         LMesh = (E3dMesh*)LPanel->Object;
	const EBool      LV = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	if(LPanel->ApplyChangesToAllSelected)
	{
		E3dScene*  LScene = E3d_Scene;
		E3dObject**  LSelectedObjects;const EIndex  LN = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LSelectedObjects)
		{
			EIndex  LC = 0;for(;LC < LN;LC++)
			{
				E3dMesh*  LTMesh = (E3dMesh*)LSelectedObjects[LC];LTMesh->ShowAllVertices = LV;
			}
			EFree(LSelectedObjects);
		}
		else LMesh->ShowAllVertices = LV;
	}
	else LMesh->ShowAllVertices = LV;

	E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
}


//================================================================
// Return hardness value of the first selected Edge of a Mesh
//================================================================
static float _MeshGetEdgeHardness(E3dMesh* PMesh)
{
	EIndex  LC = 0;const EIndex  LN = PMesh->Edges_Count;const E3dEdge*  LEdge = PMesh->Edges;for(;LC < LN;LC++, LEdge++)
	{
		if(LEdge->Flags & E3dEdgeSELECTED) return(LEdge->Hardness);
	}
	return(-1.0);  // No selected Edges found
}


static void _DCB_EdgeHardness(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dMesh*         LMesh = (E3dMesh*)LObject;
	const float      LEdgeHardness = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	EIndex  LC = 0;const EIndex  LN = LMesh->Edges_Count;E3dEdge*  LEdge = LMesh->Edges;for(;LC < LN;LC++, LEdge++)
	{
		if(LEdge->Flags & E3dEdgeSELECTED) LEdge->Hardness = LEdgeHardness;
	}

	E3dMesh_EdgeFlagsUpdate(LMesh);

	EChangeFlags  LChanged = E3dObject_CallOutputs((E3dBaseObject*)LObject, E3dChgMESHSHAPE);

	E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME|LChanged);
}


//========================================
// Color ColorPanel callback
//========================================
static void _CB_ColorChanged(EcColorPanel* PColorPanel, EPointer PClientData, EPointer PCallData)
{
	EcColorPanelCallbackStruct*  LCPCBR = (EcColorPanelCallbackStruct*)PCallData;
	int                          LColorIndex;

	if(LCPCBR->NumOfColors > 0) LColorIndex = LCPCBR->ColorIDs[0];
	else return;

	E3dObjectPanel*  LObjectPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LObjectPanel->Object;E3dMesh*  LMesh = E_DynamicCast(LObject, E3dMesh);if(LMesh)
	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LObjectPanel, E3dMesh_class);if(LDrawer)
		{
			E3dMeshResourcesGUI*  LResourcesGUI = (E3dMeshResourcesGUI*)LDrawer->ResourcesGUI;

			const EBool  LColorWasInactive = LMesh->EdgeRGBAfColor.A < 0.0 ? TRUE : FALSE;

			LMesh->EdgeRGBAfColor = PColorPanel->RGBAfColors[LColorIndex];

// If the user changed the Spline's color for the first time (it was "inactive" before), set it to "active"
//
			if(LColorWasInactive)
			{
//printf("%s() %s:%d %sACTIVATE%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout);
				LMesh->EdgeRGBAfColor.A = 1.0;
				EcColorPanel_SetRGBAfColor(LResourcesGUI->ColorPanel, 0, &LMesh->EdgeRGBAfColor);
			}

//printf("%s() %s:%d  <%s>  RGBAfColor % 7.2f % 7.2f % 7.2f % 7.2f\n", __FUNCTION__, __FILE__, __LINE__, LSpline->Name, RGBA_s(LMesh->EdgeRGBAfColor));fflush(stdout);

			EcRGBAiColor  LColor;Ec_RGBAf32ToRGBAi(&LMesh->EdgeRGBAfColor, &LColor);LColor.A = 0xFFFF;  // LMesh->EdgeRGBAfColor may have a 0 alpha, which would affect the drawing of this Widget
			Xe_SetValuesVa(LResourcesGUI->ColorSwatchW,  XeNbackgroundColor, &LColor,  NULL);
		}

		E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
	}
}


#define _MeshVAStartIndex  E3dMesh_DefaultVertexAttributes_Count
//#define _MeshVAStartIndex  0


static void  _UpdateVertexAttributesGUI(const EResourceList* PAttributes, const EIndex PVAStartIndex, Widget PParentW, E3dVertexAttributeGUIList* PVertexAttributesGUI,  XeCallbackFunc PCallback, EPointer PCallbackClientData);

//========================================
// Mesh dialog callback handler
//========================================
static void _DCB_MeshVertexAttribute(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LObjectPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LObjectPanel->Object;E3dMesh*  LMesh = E_DynamicCast(LObject, E3dMesh);if(LMesh)
	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LObjectPanel, E3dMesh_class);if(LDrawer)
		{
//printf("%s() %s:%d  %sGREEN%s  %d\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode, (int)LMesh->VertexAttributes.Count);fflush(stdout);
			E3dMeshResourcesGUI*  LResourcesGUI = (E3dMeshResourcesGUI*)LDrawer->ResourcesGUI;

			EIndex  LVAC = _MeshVAStartIndex;const int  LVAN = LMesh->VertexAttributes.Count;EResource*  LVAtr = LMesh->VertexAttributes.Items + _MeshVAStartIndex;E3dVertexAttributeGUI*  LGUI = LResourcesGUI->VertexAttributesGUI.Items;for(;LVAC < LVAN;LVAC++, LVAtr++, LGUI++)
			{
				if(PW == LGUI->NameW)
				{
printf("%s() %s:%d  %sGREEN%s  %d\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode, (int)LVAC);fflush(stdout);
					char*  LText = XeTextField_GetString(PW);
					EFreeChk(LVAtr->Name);LVAtr->Name = LText;
					return;
				}
				else if(PW == LGUI->ShowW)
				{
					const EBool  LV = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
					if(LV) LVAtr->UI = (char*)E3dUI_Show;
					else LVAtr->UI = NULL;
					break;
				}
				else if(PW == LGUI->RemoveW)
				{
					E3dMesh_VertexAttributeDelByIndex(LMesh, LVAC);
					_UpdateVertexAttributesGUI(&LMesh->VertexAttributes, _MeshVAStartIndex, LResourcesGUI->VertexAttributesMatrixW, &LResourcesGUI->VertexAttributesGUI, _DCB_MeshVertexAttribute, LObjectPanel);
					break;
				}
			}
		}
	}

	E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
}


//========================================
// PolyGroup dialog callback handler
//========================================
static void _DCB_PolyGroupVertexAttribute(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;

	EIndex  LVAC = _PolyGroup_VAStartIndex;const int  LVAN = LPolyGroup->VertexAttributes.Count;EResource*  LVAtr = LPolyGroup->VertexAttributes.Items + _PolyGroup_VAStartIndex;E3dVertexAttributeGUI*  LGUI = _PolyGroup_VertexAttributesGUI.Items;for(;LVAC < LVAN;LVAC++, LVAtr++, LGUI++)
	{
		if(PW == LGUI->NameW)
		{
			char*  LText = XeTextField_GetString(PW);
			EFreeChk(LVAtr->Name);LVAtr->Name = LText;
			return;
		}
		else if(PW == LGUI->ShowW)
		{
			const EBool  LV = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
			if(LV) LVAtr->UI = (char*)E3dUI_Show;
			else LVAtr->UI = NULL;
			E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
			break;
		}
		else if(PW == LGUI->RemoveW)
		{
			if(E3dPolyGroup_VertexAttributeDelByIndex(LPolyGroup, LVAC) == E_SUCCESS)
			{
//				_PolyGroup_UpdateDialog(LPolyGroup);
				_UpdateVertexAttributesGUI(&LPolyGroup->VertexAttributes, _PolyGroup_VAStartIndex, _PolyGroup_VertexAttributesMatrixW, &_PolyGroup_VertexAttributesGUI,  _DCB_PolyGroupVertexAttribute, NULL);
				E3dWindows_RedrawByChangeFlags(E3dChgVERTEX_ATTRIBUTE|E3dChgWIREFRAME);
			}
			break;
		}
	}
}


static void _UpdateVertexAttributesGUI(const EResourceList* PAttributes, const EIndex PVAStartIndex, Widget PParentW, E3dVertexAttributeGUIList* PVertexAttributesGUI,  XeCallbackFunc PCallback, EPointer PCallbackClientData)
{
// Add new GUI if needed
//
	{
		EIndex  LVAC = PVAStartIndex + PVertexAttributesGUI->Count;const int  LVAN = PAttributes->Count;for(;LVAC < LVAN;LVAC++)
		{
			E3dVertexAttributeGUI*  LGUI = EList_StructAdd_p(PVertexAttributesGUI);

			EArg  Args[32];int  ArgCnt;
			EXeStart;
			EXeSetArg(XeNheightAdjustable, FALSE);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			LGUI->SemanticW = XeWidget_ManagedNew("", XeLabelWidget_class, PParentW,  Args, ArgCnt);

			Widget  LW;

			{
				EXeStart;
				EXeSetArg(XeNhighlightThickness, 1);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LGUI->NameW = LW = XeWidget_ManagedNew("NameW", XeTextField_class, PParentW,  Args, ArgCnt);
				XeOverrideTranslations(LW, XeStd_TextWidgetTransTable);
				XeAddCallback(LW, XeNvalueChangedCallback, PCallback, PCallbackClientData);
			}

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNheightAdjustable, FALSE);
			XeToolTip8("Visualize this VertexAttribute.");
			LGUI->ShowW = LW = XeWidget_ManagedNew("Show", XeToggleButton_class, PParentW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, PCallback, PCallbackClientData);

			EXeStart;
			EXeSetArg(XeNspacingLeft, 6);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);EXeSetArg(XeNheightAdjustable, FALSE);
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNlabelType, XeIMAGE);EXeSetImages3(_Image_Close);
			XeToolTip8("Remove this Vertex Attribute.");
			LGUI->RemoveW = LW = XeWidget_ManagedNew("x", XePushButton_class, PParentW,  Args, ArgCnt);
			XeAddCallback(LW, XeNactivateCallback, PCallback, PCallbackClientData);
		}
	}

	{
		const int        LGUIN = PVertexAttributesGUI->Count;
		XeChangeManage*  LChildren;ENewN(LGUIN * 4, LChildren);XeChangeManage*  LChild = LChildren;

		const int        LAttrN = PAttributes->Count;
		Widget*          LTraverseWs;ENewN(LAttrN - PVAStartIndex, LTraverseWs);

		if(LAttrN > PVAStartIndex) XeWidget_Manage(PParentW);
		else XeWidget_Unmanage(PParentW);

		EIndex  LC = PVAStartIndex;const EResource*  LAttr = PAttributes->Items + PVAStartIndex;E3dVertexAttributeGUI*  LGUI = PVertexAttributesGUI->Items;for(;LC < LAttrN;LC++, LAttr++, LGUI++)
		{
			XeString  LXeStr = XeString_SimpleNew(LAttr->Semantic);if(LXeStr)
			{
				Xe_SetValuesVa(LGUI->SemanticW,  XeNlabelString, LXeStr,  NULL);XeString_Free(LXeStr);
			}

			XeTextField_SetString(LGUI->NameW, LAttr->Name, FALSE);

			LChild->W = LGUI->SemanticW;LChild->Manage = TRUE;LChild++;
			LChild->W = LGUI->NameW;LChild->Manage = TRUE;LChild++;
			LChild->W = LGUI->ShowW;LChild->Manage = TRUE;LChild++;
			LChild->W = LGUI->RemoveW;LChild->Manage = TRUE;LChild++;

			LTraverseWs[LC - PVAStartIndex] = LGUI->NameW;
		}

		Xe_TraversalRing(LTraverseWs, LAttrN - PVAStartIndex);
		EFree(LTraverseWs);

		{
			const EIndex  LN = LGUIN + PVAStartIndex;for(;LC < LN;LC++, LGUI++)
			{
				LChild->W = LGUI->SemanticW;LChild->Manage = FALSE;LChild++;
				LChild->W = LGUI->NameW;LChild->Manage = FALSE;LChild++;
				LChild->W = LGUI->ShowW;LChild->Manage = FALSE;LChild++;
				LChild->W = LGUI->RemoveW;LChild->Manage = FALSE;LChild++;
			}
		}

		if(LChild - LChildren > 0) XeWidgets_ChangeManage(LChildren, LChild - LChildren);
		EFree(LChildren);
	}
}


//================================================================
// GUIProc for E3dMesh
//================================================================
void _Mesh_GUI(ECoreObject* PObject, void* PPanel, void* PDrawer, const int PAction)
{
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dMeshResourcesGUI*     LResourcesGUI;
	E3dMesh*                 LMesh = (E3dMesh*)PObject;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dMeshResourcesGUI*)LDrawer->ResourcesGUI;
				_Mesh_UpdateStatsLabel(LMesh, LResourcesGUI->StatsValuesLabelW);
				XeToggleButton_SetState(LResourcesGUI->WireframeW, LMesh->DisplayFlags == E3dDM_Wireframe, FALSE);
				XeToggleButton_SetState(LResourcesGUI->ShowAllVerticesW, LMesh->ShowAllVertices, FALSE);
				XeToggleButton_SetState(LResourcesGUI->ShowAllEdgesW, LMesh->ShowAllEdges, FALSE);

				{
					const float  LEdgeHardness = _MeshGetEdgeHardness(LMesh);
					Xe_WidgetSetSensitive(LResourcesGUI->EdgeHardnessW, LEdgeHardness >= 0.0);  // Composite Widgets need this called (rather than just SetValues)
			    XeScale_SetValue_f32(LResourcesGUI->EdgeHardnessW,  LEdgeHardness,  FALSE);
			  }

				EcRGBAiColor  LColor;Ec_RGBAf32ToRGBAi(&LMesh->EdgeRGBAfColor, &LColor);LColor.A = 0xFFFF;  // LMesh->EdgeRGBAfColor may have a 0 alpha, which would affect the drawing of this Widget
				Xe_SetValuesVa(LResourcesGUI->ColorSwatchW,  XeNbackgroundColor, &LColor,  NULL);

				EcColorPanel_SetRGBAfColor(LResourcesGUI->ColorPanel, LResourcesGUI->ColorPanel->DrawColor0, &LMesh->EdgeRGBAfColor);
				XeToggleButton_SetState(LResourcesGUI->UseCustomEdgeColorW, LMesh->UseCustomEdgeColor, FALSE);

				_UpdateVertexAttributesGUI(&LMesh->VertexAttributes, _MeshVAStartIndex, LResourcesGUI->VertexAttributesMatrixW, &LResourcesGUI->VertexAttributesGUI,  _DCB_MeshVertexAttribute, PPanel);
			}
			else
			{
				Widget  LTopMatrixW = LDrawer->Drawer.DrawerW, LW;

				ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;
				EList_Init_s(LResourcesGUI->VertexAttributesGUI);

				EArg  Args[32];int  ArgCnt;

				EXeStart;
				EXeSetArg(XeNspacingLeft, Xe_DrawerIndent);
				Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

				EXeStart;
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNmarginSizeX, 0);
				EXeSetArg(XeNset, LMesh->DisplayFlags == E3dDM_Wireframe);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeToolTip8("Always draw this Mesh in wireframe mode, to make editing Objects inside of it easier.");
				LResourcesGUI->WireframeW = LW = XeWidget_ManagedNew("Wireframe", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MeshWireframe, (EPointer)PPanel);

				{
					EXeStart;
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginBottom, 4);
					EXeSetArg(XeNspacingX, 12);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
//					EXeSetArg(XeNmarginRight, 8);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					XeWidget_ManagedNew("Show:", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LMesh->ShowAllVertices);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Show all Vertices of the new Mesh, including hidden ones.");
					LResourcesGUI->ShowAllVerticesW = LW = XeWidget_ManagedNew("All Vertices", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_ShowAllVertices, (EPointer)PPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LMesh->ShowAllEdges);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Show all Edges of the new Mesh, including hidden ones.");
					LResourcesGUI->ShowAllEdgesW = LW = XeWidget_ManagedNew("All Edges", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MeshShowAllEdges, (EPointer)PPanel);
				}

// VertexAttributes
//
				{
					EXeStart;
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					XeWidget_ManagedNew("Custom Vertex attributes", XeLabelWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginBottom, 4);
					EXeSetArg(XeNcolumns, 4);
					EXeSetArg(XeNradioBehavior, TRUE);EXeSetArg(XeNradioAlwaysOne, FALSE);
					EXeSetArg(XeNspacingX, 6);EXeSetArg(XeNspacingY, 0);
					EXeSetArg(XeNspacingRight, 6);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("VertexAttributesMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);
					LResourcesGUI->VertexAttributesMatrixW = LMatrixW;

					_UpdateVertexAttributesGUI(&LMesh->VertexAttributes, _MeshVAStartIndex, LResourcesGUI->VertexAttributesMatrixW, &LResourcesGUI->VertexAttributesGUI, _DCB_MeshVertexAttribute, PPanel);
				}

// Stats
//
				{
					EXeStart;
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNspacingX, 8);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					char      LTmpStr[256];sprintf(LTmpStr, "Vertices:\nEdges:\nPolyGroups:\nPolygons:\nTriangles:\nTriangle-strips:\nLongest strip:\nAverage strip:\nSurface area:\nVolume:");
					XeString  LXeStr = XeString_NewVa(TRUE,
					                                    XeFONT, XeFontDEFAULT,
					                                    XeALIGNMENT, XeSTRING_ALIGN_LEFT,
					                                    XeTEXT, LTmpStr,
					                                    0);
					if(LXeStr) EXeSetArg(XeNlabelString, LXeStr);

					XeWidget_ManagedNew("", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);
					if(LXeStr) XeString_Free(LXeStr);


					EXeStart;
					EXeSetArg(XeNspacingRight, 6);
					LResourcesGUI->StatsValuesLabelW = XeWidget_ManagedNew("", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

					_Mesh_UpdateStatsLabel((E3dMesh*)PObject, LResourcesGUI->StatsValuesLabelW);
				}

// Selected Edge info
//
				{
					const float  LEdgeHardness = _MeshGetEdgeHardness(LMesh);
					EXeStart;
					EXeSetArg(XeNspacingTop, 4);
					EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					EXeSetArg(XeNsensitive, LEdgeHardness >= 0.0);
					XeToolTip8("Set 'hardness' value for selected Edges.\nThis is used for subdivision surfaces to keep some\nEdges creased and for bevelling / filleting Edges.");
			    LResourcesGUI->EdgeHardnessW = LW = XeScaleNew_Coord(LTopMatrixW, "Edge hardness",  Args, ArgCnt);
			    XeScale_SetValueAndRanges_f32(LW,  LEdgeHardness,  0.0, 1.0,  0.0, 1.0,  FALSE);
			    XeAddCallback(LW, XeNvalueChangedCallback, _DCB_EdgeHardness, (EPointer)PPanel);
				}

				EXeStart;
				EXeSetArg(XeNshadowThickness, 2);
				EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeWidget_ManagedNew("", XeSeparatorWidget_class, LTopMatrixW,  Args, ArgCnt);

// Color Drawer
//
				{
					EXeStart;
					EXeSetArg(XeNmarginTop, 4);EXeSetArg(XeNmarginBottom, 2);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNspacingX, 4);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("ColorMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					XeDrawer*  LColorDrawer = &LResourcesGUI->ColorDrawer;XeDrawer_NewVa(LMatrixW, LColorDrawer, "Edge color", 0, 0);

					LResourcesGUI->ColorSwatchW = EcColorSwatch_NewRGBAf32(NULL, LMatrixW, NULL, &LMesh->EdgeRGBAfColor, XeCB_DrawerOpenClose, (EPointer)LColorDrawer);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNspacingLeft, 4);EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNset, LMesh->UseCustomEdgeColor);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Use the color specified here, for the Edges of this Mesh.\nWhen this is OFF, use the default wireframe color.");
					LResourcesGUI->UseCustomEdgeColorW = LW = XeWidget_ManagedNew("Enable", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_EdgeColorEnable, (EPointer)PPanel);


					EcColorPanel*  LColorPanel;
					LResourcesGUI->ColorPanel = LColorPanel = EcColorPanel_NewVa(LTopMatrixW, "Color",  XeNcolorChangedCallback, _CB_ColorChanged,  XeNcolorChangedCallbackClientData, PPanel,
					 XeNalphaOn, TRUE,
					 XeNcolorComponentType, EcFLOAT_0_1,
					 XeNRGBAfColor, &LMesh->EdgeRGBAfColor,
					 XeNintensityOn, TRUE,  XeNcreateColorWheel, TRUE,  NULL);

					LColorDrawer->DrawerW = LColorPanel->TopMatrixW;
				}
			}
		break;

		case E_FREE:
			{
				LResourcesGUI = (E3dMeshResourcesGUI*)LDrawer->ResourcesGUI;
				EList_Free_s(LResourcesGUI->VertexAttributesGUI);
printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
			}
		break;
	}
}


//========================================
// Update PolyGroup stats label
//========================================
static void _PolyGroupUpdateStatsLabel(const E3dPolyGroup* PPolyGroup)
{
	char            LTmpStr[256];
	char            LAreaStr[256];
	float           LAverageStripLength = 0.0;
	EIndex          LNumOfEdges;
	const EIndex    LNumOfTriangles = E3dPolyGroup_CountTriangles(PPolyGroup);
	const E3dMesh*  LMesh = (E3dMesh*)PPolyGroup->Geometry;

	if(LMesh->PolyGroups_Count > 1) LNumOfEdges = PPolyGroup->Edges_Count;
	else LNumOfEdges = LMesh->Edges_Count;

	EStr_DoubleToString(E3dMeshPolyGroup_Area(LMesh, PPolyGroup), FALSE, LAreaStr);

	if(PPolyGroup->TriangleStrips_Count)
	{ 
		sprintf(LTmpStr, "Edges\nPolygons:\nTriangles:\nTriangle-strips:\nLongest strip:\nAverage strip:\nSurface area:");
		XeString  LXeStr = XeString_NewVa(TRUE, XeFONT, XeFontDEFAULT, XeALIGNMENT, XeSTRING_ALIGN_LEFT,
		                                    XeTEXT, LTmpStr,
                                        0);

		EArg  Args[16];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNlabelString, LXeStr);
		Xe_SetValues(_PolyGroupStatsLabelW,  Args, ArgCnt);
		XeString_Free(LXeStr);


		const E3dTriangleStrip*  LTriangleStrip = PPolyGroup->TriangleStrips;
		const EIndex             LNumOfStrips   = PPolyGroup->TriangleStrips_Count;
		EIndex                   LMaxVertexNum = 0;

		EIndex  LC = 0;for(;LC < LNumOfStrips;LC++, LTriangleStrip++)
		{
		 if(LMaxVertexNum < LTriangleStrip->NumOfVertices) LMaxVertexNum = LTriangleStrip->NumOfVertices;
		 LAverageStripLength += (float)(LTriangleStrip->NumOfVertices);
		}

		LAverageStripLength /= (float)LNumOfStrips;

		sprintf(LTmpStr, "%lld\n%lld (%lld selected)\n%6lld (%lld selected)\n%6lld\n%lld vertices\n%.2f vertices\n%s",
		 (long long)LNumOfEdges,
		 (long long)E3dMeshPolyGroup_NumOfSelectedEdges(LMesh, PPolyGroup),
		 (long long)(PPolyGroup->Polygons_Count),
		 (long long)E3dPolyGroup_NumOfSelectedPolygons(PPolyGroup),
		 (long long)LNumOfTriangles,
		 (long long)(PPolyGroup->TriangleStrips_Count),
		 (long long)LMaxVertexNum,
		 LAverageStripLength,
		 LAreaStr);
	}
	else
	{
		sprintf(LTmpStr, "Edges\nPolygons:\nTriangles:\nSurface area:");
		XeString  LXeStr = XeString_NewVa(TRUE, XeFONT, XeFontDEFAULT, XeALIGNMENT, XeSTRING_ALIGN_LEFT,
		                                    XeTEXT, LTmpStr,
		                                    0);

		EArg  Args[2];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNlabelString, LXeStr);
		Xe_SetValues(_PolyGroupStatsLabelW,  Args, ArgCnt);
		XeString_Free(LXeStr);


		sprintf(LTmpStr, "%lld (%lld selected)\n%lld (%lld selected)\n%lld\n%s",
		 (long long)LNumOfEdges,
		 (long long)E3dMeshPolyGroup_NumOfSelectedEdges(LMesh, PPolyGroup),
		 (long long)(PPolyGroup->Polygons_Count),
		 (long long)E3dPolyGroup_NumOfSelectedPolygons(PPolyGroup),
		 (long long)LNumOfTriangles,
		 LAreaStr);
	}

	{
		XeString  LXeStr = XeString_NewVa(TRUE, XeFONT, XeFontDEFAULT, XeALIGNMENT, XeSTRING_ALIGN_RIGHT,  XeTEXT, LTmpStr,  0);if(LXeStr)
		{
			EArg  Args[2];int  ArgCnt;

			EXeStart;
			EXeSetArg(XeNlabelString, LXeStr);
			Xe_SetValues(_PolyGroupStatsValuesLabelW,  Args, ArgCnt);

			XeString_Free(LXeStr);
		}
	}
}


//========================================
// Scene changed callback
//========================================
void _EditCB_ScenePolyGroup(ECoreObject* PScn, EPointer PClientData, EPointer PCallData)
{
	switch(((E3dCallbackStruct*)PCallData)->Reason)
	{
		case E3dCR_GEOMETRY_CHANGED:
			{
				E3dSceneObjectCallbackStruct*  LCBS = (E3dSceneObjectCallbackStruct*)PCallData;

				if(LCBS->SubObject == NULL || LCBS->SubObject == (EObject*)_Edit_SelectedPolyGroup)
				{
					_PolyGroup_UpdateDialog(_Edit_SelectedPolyGroup);
				}
			}
		break;

		case E3dCR_SELECTION_CHANGED:
			if(_PolyGroupFollowSelection)
			{
				E3dItem  LSelectedItem;

				switch(E3dScene_GetSingleSelection(E3d_Scene, &LSelectedItem, E3dItemPOLYGROUP, 0))
				{
					case E3dItemPOLYGROUP:
						{
							E3dPolyItem*  LPolyItem = &LSelectedItem.PolyItem;

							_Edit_PolyGroupPopupDialog(LPolyItem->PolyGroup, NULL, FALSE);
						}
					break;
				}
			}
		break;

		case E3dCR_POLYGON_SELECTION:
			if(_PolyGroupFollowSelection)
			{
				E3dItem  LSelectedItem;

				switch(E3dScene_GetSingleSelection(E3d_Scene, &LSelectedItem, E3dItemPOLYGROUP, 0))
				{
					case E3dItemPOLYGROUP:
						{
							E3dPolyItem*  LPolyItem = &LSelectedItem.PolyItem;

							_Edit_PolyGroupPopupDialog(LPolyItem->PolyGroup, NULL, FALSE);
						}
					break;
				}
			}
		break;
	}
}


//================================================================
// Create geometric tangents for a PolyGroup
//================================================================
static void _PolyGroupCreateTangents(E3dPolyGroup* PPolyGroup)
{
	int  LOffset = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_Tangent, 0);if(LOffset < 0)
	{
		E3dPolyGroup_VertexAttributeAdd(PPolyGroup,  NULL, E3dVtxAttribute_Tangent, 0,  NULL);
		LOffset = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_Tangent, 0);
	}

	{
		const E3dVertexNodeSize  LVertexNodeSize =  PPolyGroup->VertexNodeSize;
		EIndex  LPC = 0;const EIndex  LPN = PPolyGroup->Polygons_Count;E3dPolygon*  LPolygon = PPolyGroup->Polygons;for(LPC = 0;LPC < LPN;LPC++, LPolygon++)
		{
			const int  LNormalOffset = E3dVertexNodeAttribute_Offset(&PPolyGroup->VertexAttributes,  E3dVtxAttribute_Normal, 0);

			if((PPolyGroup->Flags & E3dUSE_VERTEX_NORMALS) && LNormalOffset >= 0)
			{
				EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertexNodes;E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;const E3dCoord3*  LVertexNormal = (E3dCoord3*)((uint8_t*)LPolygon->VertexNodes + LNormalOffset);E3dCoord3*  LTangent = (E3dCoord3*)((uint8_t*)LVertexNode + LOffset);for(;LVC < LVN;LVC++, EStruct_Inc(LVertexNormal, LVertexNodeSize), EStruct_Inc(LTangent, LVertexNodeSize))
				{
					_E3d_GenerateTangent_pp(LVertexNormal, LTangent);
				}
			}
			else
			{
				EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertexNodes;E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;E3dCoord3*  LTangent = (E3dCoord3*)((uint8_t*)LVertexNode + LOffset);for(;LVC < LVN;LVC++, EStruct_Inc(LVertexNode, LVertexNodeSize), EStruct_Inc(LTangent, LVertexNodeSize))
				{
					_E3d_GenerateTangent_vp(LPolygon->Normal, LTangent);
				}
			}
		}
	}
}


//================================================================
// Create geometric tangents from the way texture N is mapped
//================================================================
static EBool _PolyGroupCreateTangentsFromTextureMapping(E3dPolyGroup* PPolyGroup, const int PSTIndex)
{
	const int  LSTxOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_ST, PSTIndex);if(LSTxOffsetInVtxNode >= 0)
	{
		E3dPolyGroup_VertexAttributeAdd(PPolyGroup,  NULL, E3dVtxAttribute_Tangent, 0,  NULL);

		{
			const int                LTangentOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_Tangent, 0);
			const E3dVertexNodeSize  LVertexNodeSize = PPolyGroup->VertexNodeSize;

			EIndex  LPC = 0;const EIndex  LNumOfPolygons = PPolyGroup->Polygons_Count;E3dPolygon*  LPolygon = PPolyGroup->Polygons;for(;LPC < LNumOfPolygons;LPC++, LPolygon++)
			{
				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dST*          LST = (E3dST*)EOffsetToPtr(LVertexNode, LSTxOffsetInVtxNode);
				E3dCoord3*      LTangent = (E3dCoord3*)EOffsetToPtr(LVertexNode, LTangentOffsetInVtxNode);
				EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertexNodes;for(LVC = 0;LVC < LVN;LVC++)
				{
					*LTangent = LST->Tangent;
					EStruct_Inc(LST, LVertexNodeSize);
					EStruct_Inc(LTangent, LVertexNodeSize);
				}
			}
		}
		return(TRUE);
	}
	else return(FALSE);
}


static void _PolyGroupDialogClose(void)
{
	Xe_WidgetUnmap(_PolyGroup_DialogW);

	if(_Edit_PolyGroupSceneCallbackAdded)
	{
		E3dScene_CallbackRemove(E3d_Scene, _EditCB_ScenePolyGroup, (EPointer)0);
		_Edit_PolyGroupSceneCallbackAdded = FALSE;
	}

	if(_Edit_SelectedPolyGroup)
	{
		E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;
		E3dObject_LockDel((E3dObject*)LPolyGroup->Geometry, &_PolyGroupMeshLock);

		E_DeleteChk(_Edit_SelectedPolyGroup);
	}
}


static int _PolyGroupDialogLockDel(EPlugin* PPlugin, EPluginLock* PLock)
{
	_PolyGroupDialogClose();

	return(E_SUCCESS);
}


//========================================================================================
// Return TRUE, if PPolyGroup's attributes are being edited (the dialog box is displayed)
//========================================================================================
EBool _PolyGroup_IsEdited(E3dPolyGroup* PPolyGroup)
{
	if(_Edit_SelectedPolyGroup == PPolyGroup) return(TRUE);
	return(FALSE);
}


//========================================
// PolyGroup dialog callback handler
//========================================
static void _DCB_PolyGroupClose(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_PolyGroupDialogClose();

	E3dWindows_RedrawByChangeFlags(E3dChgLOCK_UNLOCK);
}


//========================================
// PolyGroup dialog callback handler
//========================================
static void _DCB_PolyGroup(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;

	switch((EPointerInt)PClientData)
	{
		case CR_NAME:
			{
				char*  LText = XeTextField_GetString(_PolyGroupNameW);

				EObject_SetName(_Edit_SelectedPolyGroup, LText, FALSE);
			}
		break;

		case CR_TOGGLE_FOLLOW_SELECTION:
			_PolyGroupFollowSelection = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
		break;

		case CR_TOGGLE_APPLY_CHANGES_TO_SELECTED:
			_PolyGroupApplyChangesToAllSelected = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
		break;

		case CR_TOGGLE_FACETED:
			{
				EChangeFlags  LChanged = 0;
				const EBool   LSet = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

				if(_PolyGroupApplyChangesToAllSelected)
				{
					E3dObject**  LSelectedObjects;const EIndex  LNumOfSelectedObjects = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LSelectedObjects)
					{
						EIndex  LC = 0;for(;LC < LNumOfSelectedObjects;LC++)
						{
							E3dMesh*      LMesh = (E3dMesh*)(LSelectedObjects[LC]);
							EChangeFlags  LMeshChange = 0;
							EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
							{
								E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[LPGC];
								if(LMesh->Selection != E3dObjectSelPOLYGROUP || LPolyGroup->Selected)
								{
									if(LSet)
									{
										LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;
										LMeshChange |= E3dUF_POLYGON_NORMAL;
									}
									else
									{
										LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;

										E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);  // In case the PolyGroup had no vertex normals before
										LPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;

										E3dMeshPolyGroup_RefreshNormals(LMesh, LPolyGroup, LPolyGroup->VertexNormalType, LPolyGroup->SmoothingAngle);

										LMeshChange |= E3dUF_VERTEX_NORMAL;
									}
								}
							}

							if(LMeshChange)
							{
								E3dObject_UpdateForDisplay((E3dObject*)LMesh, LMeshChange, LScene);
								LChanged |= E3dChgVERTEX_ATTRIBUTE;
							}
						}

						EFree(LSelectedObjects);
					}
				}
				else
				{
					E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;
					E3dMesh*       LMesh = E_DynamicCast(LPolyGroup->Geometry, E3dMesh);if(LMesh)
					{
						if(((XeToggleButtonCallbackStruct*)PCallData)->Set)
						{
							LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;
							E3dObject_UpdateForDisplay((E3dObject*)LMesh, E3dUF_POLYGON_NORMAL, LScene);
						}
						else
						{
							LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;

							LPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;

							E3dMeshPolyGroup_RefreshNormals(LMesh, _Edit_SelectedPolyGroup, LPolyGroup->VertexNormalType, LPolyGroup->SmoothingAngle);
							E3dObject_UpdateForDisplay((E3dObject*)LMesh, E3dUF_VERTEX_NORMAL, LScene);
							LChanged |= E3dChgVERTEX_ATTRIBUTE;
						}
					}
				}

				if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);
			}
		break;

		case CR_CHANGE_VECTOR_SCALE:
			{
				EChangeFlags  LChanged = 0;

				if(_PolyGroupApplyChangesToAllSelected)
				{
					E3dObject**  LSelectedObjects;const EIndex  LNumOfSelectedObjects = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LSelectedObjects)
					{
						EIndex  LC = 0;for(;LC < LNumOfSelectedObjects;LC++)
						{
							E3dMesh*      LMesh = (E3dMesh*)(LSelectedObjects[LC]);
							EChangeFlags  LMeshChange = 0;
							EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
							{
								E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[LPGC];
								if(LMesh->Selection != E3dObjectSelPOLYGROUP || LPolyGroup->Selected)
								{
									LPolyGroup->VectorDrawScale = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
								}
							}

							if(LMeshChange)
							{
								LChanged |= E3dChgSELECTION;
							}
						}

						EFree(LSelectedObjects);
					}
				}
				else
				{
					E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;

					LPolyGroup->VectorDrawScale = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
					LChanged |= E3dChgSELECTION;
				}

				if(LChanged)
				{
					E3dWindows_RedrawByChangeFlags(LChanged);
				}
			}

			E3dWindows_RedrawByChangeFlags(E3dChgSELECTION);
		break;

		case CR_TOGGLE_CASTS_SHADOWS:
			if(((XeToggleButtonCallbackStruct*)PCallData)->Set) _Edit_SelectedPolyGroup->CastsShadows = TRUE;
			else _Edit_SelectedPolyGroup->CastsShadows = FALSE;

			E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
		break;

		case CR_TOGGLE_RECEIVES_SHADOWS:
			if(((XeToggleButtonCallbackStruct*)PCallData)->Set) _Edit_SelectedPolyGroup->ReceivesShadows = TRUE;
			else _Edit_SelectedPolyGroup->ReceivesShadows = FALSE;

			E3dWindows_RedrawByChangeFlags(E3dChgLIGHT);
		break;

		case CR_TOGGLE_DOUBLE_SIDED:
			if(((XeToggleButtonCallbackStruct*)PCallData)->Set) _Edit_SelectedPolyGroup->DoubleSided = TRUE;
			else _Edit_SelectedPolyGroup->DoubleSided = FALSE;

			E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
		break;

		case CR_REMOVE_STRIPS:
			{
				E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;
				E3dPolyGroup_FreeTriangleStrips(LPolyGroup);

				_PolyGroupUpdateStatsLabel(LPolyGroup);
				E3dObject_UpdateForDisplay((E3dObject*)LPolyGroup->Geometry, E3dUF_REMOVE_STRIPS, LScene);
				E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
			}
		break;

		case CR_TANGENT_FROM_SHAPE:
			{
				E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;
				_PolyGroupCreateTangents(LPolyGroup);

				_PolyGroupUpdateStatsLabel(_Edit_SelectedPolyGroup);
				E3dObject_UpdateForDisplay((E3dObject*)LPolyGroup->Geometry, E3dUF_REMOVE_STRIPS, LScene);
				E3dWindows_RedrawByChangeFlags(E3dChgVERTEX_ATTRIBUTE);
			}
		break;
	}
}


//========================================
// PolyGroup dialog callback handler
//========================================
static void _DCB_PolyGroupGetUVFromSTx(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	const int  LSTIndex = (EPointerInt)PClientData;

	E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;
	if(E3dPolyGroup_STXtoUV(LPolyGroup, LSTIndex) == E_NO_DATA)
	{
		E3dp_PrintMessage(ELogWARNING, 5000, "This PolyGroup has no texture coordinate set (ST) %d", LSTIndex);
	}
	else
	{
		_PolyGroupUpdateStatsLabel(LPolyGroup);
		E3dObject_UpdateForDisplay((E3dObject*)LPolyGroup->Geometry, E3dUF_REMOVE_STRIPS, LScene);
		E3dWindows_RedrawByChangeFlags(E3dChgVERTEX_ATTRIBUTE);
	}
}


//========================================
// PolyGroup dialog callback handler
//========================================
static void _DCB_PolyGroupGetTangentFromSTx(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	const int  LSTIndex = (EPointerInt)PClientData;

	E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;

	if(_PolyGroupCreateTangentsFromTextureMapping(LPolyGroup, 0))
	{
		_PolyGroupUpdateStatsLabel(_Edit_SelectedPolyGroup);
		E3dObject_UpdateForDisplay((E3dObject*)LPolyGroup->Geometry, E3dUF_REMOVE_STRIPS, LScene);
		E3dWindows_RedrawByChangeFlags(E3dChgVERTEX_ATTRIBUTE);
	}
	else E3dp_PrintMessage(ELogWARNING, 5000, "This PolyGroup has no texture coordinate set (ST) %d", LSTIndex);
}


static inline void _PolyGroup_SmoothingAngle(E3dPolyGroup* PPolyGroup, const float PAngle)
{
	PPolyGroup->SmoothingAngle = PAngle;

// Turn on vertex normals automatically if the user changes the SmoothingAngle
//
	E3dPolyGroup_VertexAttributeAdd(PPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);
	PPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;
	PPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;
	E3dMeshPolyGroup_RefreshNormals(E_DynamicCast(PPolyGroup->Geometry, E3dMesh), PPolyGroup, PPolyGroup->VertexNormalType, PPolyGroup->SmoothingAngle);
}


static void _DCB_PolyGroup_SmoothingAngle(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	EChangeFlags  LChanged = 0;
	EListAI(, E3dObject*, LMeshes);
	const float   LAngle = ((XeScaleCallbackStruct*)PCallData)->Value.f32;

	if(_PolyGroupApplyChangesToAllSelected)
	{
		E3dObject**  LSelectedObjects;const EIndex  LNumOfSelectedObjects = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);
		if(LSelectedObjects)
		{
			EIndex  LC = 0;for(;LC < LNumOfSelectedObjects;LC++)
			{
				E3dMesh*  LMesh = (E3dMesh*)(LSelectedObjects[LC]);
				EIndex    LMeshChanged = 0;
				EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
				{
					E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[LPGC];
					if(LMesh->Selection != E3dObjectSelPOLYGROUP || LPolyGroup->Selected)
					{
						_PolyGroup_SmoothingAngle(LPolyGroup, LAngle);
						LMeshChanged++;
					}
				}

				if(LMeshChanged)
				{
					EList_PointerAddAChk(LMeshes, 16, LMesh);
					LChanged |= E3dChgVERTEX_ATTRIBUTE;
				}
			}

			EFree(LSelectedObjects);
		}
	}
	else
	{
		if(_Edit_SelectedPolyGroup)
		{
			E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;
			_PolyGroup_SmoothingAngle(_Edit_SelectedPolyGroup, LAngle);
			EList_PointerAddAChk(LMeshes, 16, LPolyGroup->Geometry);

			LChanged |= E3dChgVERTEX_ATTRIBUTE;
		}
	}


	if(LMeshes)
	{
		EIndex  LC = 0;for(;LC < LMeshes_Count;LC++)
		{
			E3dObject_UpdateForDisplay(LMeshes[LC], E3duf_NORMALS, LScene);
		}
		EFree(LMeshes);
	}


	if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// PolyGroup dialog callback handler
//========================================
static void _DCB_PolyGroupWireframe(Widget LW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	EChangeFlags  LChanged = 0;
	const EBool   LSet = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	if(_PolyGroupApplyChangesToAllSelected)
	{
		E3dObject**  LSelectedObjects;const EIndex  LNumOfSelectedObjects = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LSelectedObjects)
		{
			EIndex  LC = 0;for(;LC < LNumOfSelectedObjects;LC++)
			{
				E3dMesh*  LMesh = (E3dMesh*)(LSelectedObjects[LC]);
				EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
				{
					E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[LPGC];
					if(LMesh->Selection != E3dObjectSelPOLYGROUP || LPolyGroup->Selected)
					{
						if(LSet) LPolyGroup->DisplayFlags = E3dDM_Wireframe;
						else LPolyGroup->DisplayFlags = 0;
					}
				}
			}

			EFree(LSelectedObjects);
		}
	}

	if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);
}


static inline EIndex _PolyGroup_SharpPolygons(E3dPolyGroup* PPolyGroup, const EBool PSet)
{
	const EIndex  LNumOfSelectedPolygons = E3dPolyGroup_NumOfSelectedPolygons(PPolyGroup);

	if(LNumOfSelectedPolygons)
	{
		EIndex  LPC = 0;const EIndex  LPN = PPolyGroup->Polygons_Count;E3dPolygon*  LPolygon = PPolyGroup->Polygons;
		if(PSet) for(;LPC < LPN;LPC++, LPolygon++) LPolygon->Flags |= E3dPolygonSHARP;
		else for(;LPC < LPN;LPC++, LPolygon++) LPolygon->Flags &= ~E3dPolygonSHARP;

		return(1);
	}
	return(0);
}


static void _DCB_PolyGroup_SharpPolygons(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	EChangeFlags  LChanged = 0;
	const EBool   LSet = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
	EListAI(, E3dObject*, LMeshes);

	if(_PolyGroupApplyChangesToAllSelected)
	{
		E3dObject**  LSelectedObjects;const EIndex  LNumOfSelectedObjects = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);
		if(LSelectedObjects)
		{
			EIndex  LC = 0;for(;LC < LNumOfSelectedObjects;LC++)
			{
				E3dMesh*  LMesh = (E3dMesh*)(LSelectedObjects[LC]);
				EBool     LMeshChanged = FALSE;
				EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
				{
					E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[LPGC];
					if(LMesh->Selection != E3dObjectSelPOLYGROUP || LPolyGroup->Selected)
					{
						LMeshChanged += _PolyGroup_SharpPolygons(LPolyGroup, LSet);
					}
				}

				if(LMeshChanged)
				{
					EList_PointerAddAChk(LMeshes, 16, LMesh);
					LChanged |= E3dChgVERTEX_ATTRIBUTE;
				}
			}

			EFree(LSelectedObjects);
		}
	}
	else
	{
		if(_Edit_SelectedPolyGroup)
		{
			E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;
			if(_PolyGroup_SharpPolygons(LPolyGroup, LSet))
			{
				EList_PointerAddAChk(LMeshes, 16, LPolyGroup->Geometry);

				LChanged |= E3dChgVERTEX_ATTRIBUTE;
			}
		}
	}


	if(LMeshes)
	{
		EIndex  LC = 0;for(;LC < LMeshes_Count;LC++)
		{
			E3dObject_UpdateForDisplay(LMeshes[LC], E3duf_SHAPE, LScene);
		}
		EFree(LMeshes);
	}


	if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Update PolyGroup dialog
//========================================
void _PolyGroup_UpdateDialog(const E3dPolyGroup* PPolyGroup)
{
	XeTextField_SetString(_PolyGroupNameW, PPolyGroup->Name, FALSE);

	_PolyGroupUpdateStatsLabel(PPolyGroup);

//printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
	if(PPolyGroup->TriangleStrips_Count) XeWidget_Manage(_PolyGroupRemoveTriangleStripsW);
	else XeWidget_Unmanage(_PolyGroupRemoveTriangleStripsW);

	XeToggleButton_SetState(_PolyGroupFacetedBW, ((PPolyGroup->Flags & E3dUSE_VERTEX_NORMALS) == 0), FALSE);
	XeScale_SetValue_f32(_PolyGroupSmoothingAngleW, PPolyGroup->SmoothingAngle, FALSE);
	XeScale_SetValue_Coord(_PolyGroupVectorDrawScaleScaleW, PPolyGroup->VectorDrawScale, FALSE);

	XeToggleButton_SetState(_PolyGroupCastsShadowsBW, PPolyGroup->CastsShadows, FALSE);
	XeToggleButton_SetState(_PolyGroupReceivesShadowsBW, PPolyGroup->ReceivesShadows, FALSE);
	XeToggleButton_SetState(_PolyGroupDoubleSidedBW, PPolyGroup->DoubleSided, FALSE);

	XeToggleButton_SetState(_PolyGroupWireframeW, PPolyGroup->DisplayFlags == E3dDM_Wireframe, FALSE);

	{
		EBool  LHasST = FALSE;

		XeChangeManage  LChildren[E3dMAX_NUM_OF_2DTEXTUREMAPPERS];

		{
			XeChangeManage*  LChild = LChildren;

			EIndex  LC = 0;for(;LC < E3dMAX_NUM_OF_2DTEXTUREMAPPERS;LC++)
			{
				const EBool  LHasThisST = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_ST, LC) >= 0;
				LChild->W = _PolyGroupGetUVFromSTWs[LC];LChild->Manage = LHasThisST;LChild++;
				if(LHasThisST) LHasST = TRUE;
			}
			if(LChild - LChildren > 0) XeWidgets_ChangeManage(LChildren, LChild - LChildren);
		}

		{
			XeChangeManage*  LChild = LChildren;

			EIndex  LC = 0;for(;LC < E3dMAX_NUM_OF_2DTEXTUREMAPPERS;LC++)
			{
				const EBool  LHasThisST = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_ST, LC) >= 0;
				LChild->W = _PolyGroupGetTangentFromSTWs[LC];LChild->Manage = LHasThisST;LChild++;
			}
			if(LChild - LChildren > 0) XeWidgets_ChangeManage(LChildren, LChild - LChildren);
		}

		{
			if(LHasST) XeWidget_Manage(_PolyGroupGetUVFromSTMatrixW);
			else XeWidget_Unmanage(_PolyGroupGetUVFromSTMatrixW);
		}
	}

	_UpdateVertexAttributesGUI(&PPolyGroup->VertexAttributes, _PolyGroup_VAStartIndex, _PolyGroup_VertexAttributesMatrixW, &_PolyGroup_VertexAttributesGUI,  _DCB_PolyGroupVertexAttribute, NULL);
}


//========================================
// Lay out current texture
//========================================
static void _DCB_PolyGroupTextureLayout(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dPolyGroup*  LPolyGroup = _Edit_SelectedPolyGroup;
	E3dMesh*       LMesh = E_DynamicCast(LPolyGroup->Geometry, E3dMesh);

	const char*  LVASemantic = NULL;int  LVAIndex = 0;
	switch((EPointerInt)PClientData)
	{
		case 0:   LVASemantic = E3dVtxAttribute_ST;LVAIndex = 0;break;
		case 1:   LVASemantic = E3dVtxAttribute_ST;LVAIndex = 1;break;
		case 16:  LVASemantic = E3dVtxAttribute_UV;LVAIndex = 0;break;
	}

	const E3dCoordinate  LSpacing = _PolygonPackingSpacing;
	if(E3dMeshPolyGroup_TextureLayout(LMesh, LPolyGroup,  LVASemantic, LVAIndex,  LSpacing, LSpacing,  LSpacing, LSpacing,  NULL) == E_SUCCESS)
	{
		E3dObject_UpdateForDisplay((E3dObject*)LMesh, E3dUF_VERTEX_ST, E3d_Scene);
//	E3dObject_UpdateForDisplay((E3dObject*)LMesh, E3dUF_ALL, E3d_Scene);
		E3dWindows_RedrawByChangeFlags(E3dChgVERTEX_ATTRIBUTE|E3dChgTEXTURE_MAPPING);
		_PolyGroup_UpdateDialog(LPolyGroup);
	}
}


//========================================
// PolyGroup dialog callback handler
//========================================
static void _DCB_PolyGroupPackingSpacing(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_PolygonPackingSpacing = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
}


//========================================
// Edit information on a PolyGroup
//========================================
void _Edit_PolyGroupPopupDialog(const E3dPolyGroup* PPolyGroup, const E3dObjectComponentClass* POpenFolderForThisComponentClass, const EBool PRaiseDialogBox)
{
// Is the PolyGroup-dialog already up?
//
	if(_Edit_SelectedPolyGroup)
	{
		if(_Edit_SelectedPolyGroup != PPolyGroup)
		{
			E3dObject_LockDel((E3dObject*)_Edit_SelectedPolyGroup->Geometry, &_PolyGroupMeshLock);

			E3dObject_LockAdd((E3dObject*)PPolyGroup->Geometry, &_PolyGroupMeshLock);

			E_Assign(PPolyGroup, _Edit_SelectedPolyGroup);
		}

		_PolyGroup_UpdateDialog(PPolyGroup);
		if(PRaiseDialogBox) XeShell_Raise(_PolyGroup_DialogW);
		return;
	}


	E_Assign(PPolyGroup, _Edit_SelectedPolyGroup);

	if(_PolyGroup_DialogLock == NULL) _PolyGroup_DialogLock = E_PluginLockAdd(_EditInfoPlugin, _PolyGroupDialogLockDel);

	if(_PolyGroup_DialogW == NULL)
	{
		Widget    LW;
		EArg      Args[32];int  ArgCnt;
		char      LTmpStr[256];
		Widget    LIndentWs[8];int  LIndentN;      // For indenting the ScaleWidgets
		Widget    LTraverseWs[8];int  LTraverseN;  // For TextField traversal


		const XeCoordinate  LButtonMarginLeft = 0, LButtonMarginRight = 2;

		E3dObject_LockAdd((E3dObject*)PPolyGroup->Geometry, &_PolyGroupMeshLock);

		EXeStart;
		EXeSetArg(XeNallowShellResize, TRUE);
		EXeSetArg(XeNbuttonsStretch, FALSE);
		EXeSetArg(XeNworkAreaType, XeMATRIX);
		EXeSetArg(XeNwmDecorations, XeWMDecorCLOSE|XeWMDecorBORDER|XeWMDecorRESIZEH|XeWMDecorTITLE);
		EXeSetArg(XeNtitle, "Edit PolyGroup parameters");EXeSetArg(XeNiconName, "PolyGroup");
		Widget  LDialogW = XeDialog_New(XeDIALOG_TEMPLATE, "PolyGroup", E3dp_TopShellW,  Args, ArgCnt);if(LDialogW == NULL) return;

		_PolyGroup_DialogW = LDialogW;

		XeAddCallback(LDialogW, XeNwmCallback, _DCB_PolyGroupClose, (EPointer)0);

		Widget  LTopMatrixW = XeDialog_GetChild(LDialogW, XeDIALOG_WORK_AREA);

		EXeStart;
		EXeSetArg(XeNorientation, XeVERTICAL);
		EXeSetArg(XeNmarginSizeX, 2);EXeSetArg(XeNmarginSizeY, 4);
		EXeSetArg(XeNspacingY, 0);
		EXeSetArg(XeNheightAdjustable, FALSE);
		Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

		const int  LLeftMargin = 6;

		{
			EXeStart;
			EXeSetArg(XeNspacingLeft, LLeftMargin);
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNspacingX, 2);
			EXeSetArg(XeNheightAdjustable, FALSE);
			Widget  LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			XeWidget_ManagedNew("Name", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNstringUTF8, PPolyGroup->Name);
			EXeSetArg(XeNheightAdjustable, FALSE);
			_PolyGroupNameW = LW = XeWidget_ManagedNew("PolyGroupName", XeTextField_class, LMatrixW,  Args, ArgCnt);
			XeOverrideTranslations(_PolyGroupNameW, XeStd_TextWidgetTransTable);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup, (EPointer)CR_NAME);
		}

		{
			EXeStart;
			EXeSetArg(XeNspacingLeft, LLeftMargin);
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNspacingX, 8);
			EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNwidthAdjustable, FALSE);
			EXeSetArg(XeNheightAdjustable, FALSE);
			Widget  LMatrixW = XeWidget_ManagedNew("FollowSelApplyToAllMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNmarginSizeX, 0);
			EXeSetArg(XeNset, _PolyGroupFollowSelection);
			EXeSetArg(XeNheightAdjustable, FALSE);
			_PolyGroupFollowSelectionBW = LW = XeWidget_ManagedNew("Follow selection", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup, (EPointer)CR_TOGGLE_FOLLOW_SELECTION);

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNset, _PolyGroupApplyChangesToAllSelected);
			EXeSetArg(XeNheightAdjustable, FALSE);
			XeToolTip8("Apply changes to all selected PolyGroups.");
			_PolyGroupApplyChangesToAllSelectedW = LW = XeWidget_ManagedNew("Apply to all", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup, (EPointer)CR_TOGGLE_APPLY_CHANGES_TO_SELECTED);
		}

		EXeStart;
		EXeSetArg(XeNspacingTop, 2);
		EXeSetArg(XeNshadowThickness, 2);
		EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
		EXeSetArg(XeNheightAdjustable, FALSE);
		XeWidget_ManagedNew("", XeSeparatorWidget_class, LTopMatrixW,  Args, ArgCnt);

		EXeStart;
		EXeSetArg(XeNspacingLeft, LLeftMargin);
		EXeSetArg(XeNspacingTop, 2);
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNset, ((PPolyGroup->Flags&E3dUSE_VERTEX_NORMALS) == 0));
		EXeSetArg(XeNheightAdjustable, FALSE);
		_PolyGroupFacetedBW = LW = XeWidget_ManagedNew("Faceted", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup, (EPointer)CR_TOGGLE_FACETED);

		LTraverseN = 0;
		LIndentN = 0;
		EXeStart;
		EXeSetArg(XeNtextFieldDigits, 6);EXeSetArg(XeNdecimalPoints, 3);
		EXeSetArg(XeNincrement, 10);
		EXeSetArg(XeNorientation, XeHORIZONTAL);
		EXeSetArg(XeNheightAdjustable, FALSE);
		LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = _PolyGroupSmoothingAngleW = LW = XeScale_Newf32(LTopMatrixW, "Smoothing angle",  Args, ArgCnt);
		XeScale_SetValueAndRange_f32(LW,  PPolyGroup->SmoothingAngle,  0.0, 180.0,  FALSE);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup_SmoothingAngle, (EPointer)NULL);

		EXeStart;
		EXeSetArg(XeNtextFieldDigits, 6);EXeSetArg(XeNdecimalPoints, 3);
		EXeSetArg(XeNincrement, 10);
		EXeSetArg(XeNorientation, XeHORIZONTAL);
		EXeSetArg(XeNheightAdjustable, FALSE);
		LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = _PolyGroupVectorDrawScaleScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Normal scale",  Args, ArgCnt);
		XeScale_SetValueAndRanges_Coord(LW,  PPolyGroup->VectorDrawScale,  0.001, 1000.0,  0.001, 100.0,  FALSE);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup, (EPointer)CR_CHANGE_VECTOR_SCALE);

		Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, LLeftMargin);

		{
			EXeStart;
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNspacingLeft, LLeftMargin);
			EXeSetArg(XeNmarginSizeX, 0);
			EXeSetArg(XeNmarginSizeY, 5);
			EXeSetArg(XeNspacingX, 12);
			EXeSetArg(XeNheightAdjustable, FALSE);
			Widget  LMatrixW = XeWidget_ManagedNew("PGShMatrix", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNmarginSizeX, 0);
			EXeSetArg(XeNset, PPolyGroup->CastsShadows);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			_PolyGroupCastsShadowsBW = LW = XeWidget_ManagedNew("Casts shadows", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup, (EPointer)CR_TOGGLE_CASTS_SHADOWS);

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNmarginSizeX, 0);
			EXeSetArg(XeNset, PPolyGroup->ReceivesShadows);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			_PolyGroupReceivesShadowsBW = LW = XeWidget_ManagedNew("Receives shadows", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup, (EPointer)CR_TOGGLE_RECEIVES_SHADOWS);
		}


		{
			EXeStart;
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNspacingLeft, LLeftMargin);
			EXeSetArg(XeNspacingX, 12);
			EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNheightAdjustable, FALSE);
			Widget  LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNmarginSizeX, 0);
			EXeSetArg(XeNset, (EBool)(PPolyGroup->DoubleSided));
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			_PolyGroupDoubleSidedBW = LW = XeWidget_ManagedNew("Double-sided", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup, (EPointer)CR_TOGGLE_DOUBLE_SIDED);

			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNmarginSizeX, 0);
			EXeSetArg(XeNset, PPolyGroup->DisplayFlags == E3dDM_Wireframe);
			EXeSetArg(XeNheightAdjustable, FALSE);
			XeToolTip8("Always draw this PolyGroup in wireframe mode, to make editing Objects inside of it easier.");
			_PolyGroupWireframeW = LW = XeWidget_ManagedNew("Wireframe", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroupWireframe, (EPointer)NULL);


			EXeStart;
			EXeSetArg(XeNshadowThickness, 0);
			EXeSetArg(XeNhighlightThickness, 0);
			EXeSetArg(XeNmarginSizeX, 0);
			EXeSetArg(XeNset, PPolyGroup->DisplayFlags == E3dDM_Wireframe);
			EXeSetArg(XeNheightAdjustable, FALSE);
			XeToolTip8("Mark selected Polygons as \"sharp\".\nSubdivision surfaces will only \"smooth\" the Edges of such Polygons.");
			_SharpPolygonsW = LW = XeWidget_ManagedNew("Sharp Polys", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroup_SharpPolygons, (EPointer)NULL);
		}


		EXeStart;
		EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
		EXeSetArg(XeNheightAdjustable, FALSE);
		_PolyGroupRemoveTriangleStripsW = LW = XeWidget_New("Remove triangle-strips", XePushButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNactivateCallback, _DCB_PolyGroup, (EPointer)CR_REMOVE_STRIPS);
		if(PPolyGroup->TriangleStrips_Count) XeWidget_Manage(_PolyGroupRemoveTriangleStripsW);


		{
			const EBool  LHasST = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_ST, 0) >= 0;

			EXeStart;
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNspacingLeft, LLeftMargin);
			EXeSetArg(XeNmarginTop, 4);
			EXeSetArg(XeNmarginSizeY, 4);
			EXeSetArg(XeNmarginSizeX, 0);
			EXeSetArg(XeNspacingX, 4);
			EXeSetArg(XeNselfResizePolicyX, XeGROW);EXeSetArg(XeNselfResizePolicyY, XeGROW);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			Widget  LMatrixW;
			if(LHasST) LMatrixW = XeWidget_ManagedNew("GrabUVMatrixW",  XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);
			else LMatrixW = XeWidget_New("GrabUVMatrixW",  XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);
			_PolyGroupGetUVFromSTMatrixW = LMatrixW;

			EXeStart;
			EXeSetArg(XeNwidthAdjustable, FALSE);
			_PolyGroupGetUVLabelW = XeWidget_ManagedNew("Get UVs from", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

			{
				char  LName[32];
				int  LC = 0;for(;LC < E3dMAX_NUM_OF_2DTEXTUREMAPPERS;LC++)
				{
					EXeStart;
					EXeSetArg(XeNmarginLeft, LButtonMarginLeft);EXeSetArg(XeNmarginRight, LButtonMarginRight);EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					sprintf(LName, "ST%d", LC);
					if(E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_ST, LC) >= 0) _PolyGroupGetUVFromSTWs[LC] = LW = XeWidget_ManagedNew(LName, XePushButton_class, LMatrixW,  Args, ArgCnt);
					else _PolyGroupGetUVFromSTWs[LC] = LW = XeWidget_New(LName, XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_PolyGroupGetUVFromSTx, (EPointer)(EPointerInt)LC);
				}
			}
		}

		{
			EXeStart;
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNspacingLeft, LLeftMargin);
			EXeSetArg(XeNspacingTop, 8);
			EXeSetArg(XeNmarginSizeX, 0);
			EXeSetArg(XeNspacingX, 0);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			Widget  LMatrixW = XeWidget_ManagedNew("PGPackingMatrix",  XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

			EXeStart;
			XeWidget_ManagedNew("Lay out", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNspacingLeft, 4);
			EXeSetArg(XeNmarginLeft, LButtonMarginLeft);EXeSetArg(XeNmarginRight, LButtonMarginRight);EXeSetArg(XeNmarginSizeY, 0);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			XeToolTip8("Lay out Polygons of this PolyGroup in ST0 coordinate space, based on their surface area.\nThis ensures a uniform pixel density on the surface of the Polygons, which is necessary\nfor rendering optimal light maps and other types of pre-computed textures.\nSelect 'Freehand' mapping for the light map texture.\nIf there is no ST0 set on the PolyGroup, it will be added.");
			LW = XeWidget_ManagedNew("ST0", XePushButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNactivateCallback, _DCB_PolyGroupTextureLayout, (EPointer)0);

			EXeStart;
			EXeSetArg(XeNspacingLeft, 4);
			EXeSetArg(XeNmarginLeft, LButtonMarginLeft);EXeSetArg(XeNmarginRight, LButtonMarginRight);EXeSetArg(XeNmarginSizeY, 0);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			XeToolTip8("Lay out Polygons of this PolyGroup in ST1 coordinate space, based on their surface area.\nThis ensures a uniform pixel density on the surface of the Polygons, which is necessary\nfor rendering optimal light maps and other types of pre-computed textures.\nSelect 'Freehand' mapping for the light map texture.\nIf there is no ST1 set on the PolyGroup, it will be added.");
			LW = XeWidget_ManagedNew("ST1", XePushButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNactivateCallback, _DCB_PolyGroupTextureLayout, (EPointer)1);

			EXeStart;
			EXeSetArg(XeNspacingLeft, 4);
			EXeSetArg(XeNmarginLeft, LButtonMarginLeft);EXeSetArg(XeNmarginRight, LButtonMarginRight);EXeSetArg(XeNmarginSizeY, 0);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			XeToolTip8("Lay out Polygons of this PolyGroup in UV coordinate space, based on their surface area.\nThis ensures a uniform pixel density on the surface of the Polygons, which is necessary\nfor rendering optimal light maps and other types of pre-computed textures.\nSelect UV mapping for the light map texture.\nIf there is no UV set on the PolyGroup, it will be added.");
			LW = XeWidget_ManagedNew("UV", XePushButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNactivateCallback, _DCB_PolyGroupTextureLayout, (EPointer)16);

			EXeStart;
			EXeSetArg(XeNspacingLeft, 6);
			EXeSetArg(XeNtextFieldDigits, 7);EXeSetArg(XeNdecimalPoints, 4);
			EXeSetArg(XeNincrement, 10);
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNheightAdjustable, FALSE);
			LTraverseWs[LTraverseN++] = LW = XeScale_Newf32(LTopMatrixW, "Spacing",  Args, ArgCnt);
			XeScale_SetValueAndRanges_Coord(LW,  _PolygonPackingSpacing,  0.0001, 0.02,  0.001, 0.02,  FALSE);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PolyGroupPackingSpacing, (EPointer)0);
		}

		{
			EXeStart;
			EXeSetArg(XeNspacingLeft, LLeftMargin);EXeSetArg(XeNspacingRight, 6);
			EXeSetArg(XeNspacingTop, 8);
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNspacingX, 4);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			Widget  LMatrixW = XeWidget_ManagedNew("PGInfoMatrix",  XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNwidthAdjustable, FALSE);
			EXeSetArg(XeNheightAdjustable, FALSE);
			LW = XeWidget_ManagedNew("Tangents from", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNmarginSizeY, 0);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			XeToolTip8("Generate vertex tangents from the Mesh shape.");
			LW = XeWidget_ManagedNew("Shape", XePushButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNactivateCallback, _DCB_PolyGroup, (EPointer)CR_TANGENT_FROM_SHAPE);

			{
				char  LName[32];
				int  LC = 0;for(;LC < E3dMAX_NUM_OF_2DTEXTUREMAPPERS;LC++)
				{
					EXeStart;
					EXeSetArg(XeNmarginLeft, LButtonMarginLeft);EXeSetArg(XeNmarginRight, LButtonMarginRight);EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					sprintf(LName, "ST%d", LC);
					if(E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_ST, LC) >= 0) _PolyGroupGetTangentFromSTWs[LC] = LW = XeWidget_ManagedNew(LName, XePushButton_class, LMatrixW,  Args, ArgCnt);
					else _PolyGroupGetTangentFromSTWs[LC] = LW = XeWidget_New(LName, XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_PolyGroupGetTangentFromSTx, (EPointer)(EPointerInt)LC);
				}
			}
		}

		{
			EXeStart;
			EXeSetArg(XeNspacingTop, 6);
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNspacingLeft, LLeftMargin);
			EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 5);
			EXeSetArg(XeNspacingX, 16);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			Widget  LMatrixW = XeWidget_ManagedNew("PGInfoMatrix",  XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

			EXeStart;
			sprintf(LTmpStr, "Edges\nPolygons:\nTriangles:\nTriangle-strips:\nLongest strip:\nAverage strip:\nSurface area:");
			XeString  LXeStr = XeString_NewVa(TRUE,  XeFONT, XeFontDEFAULT,  XeALIGNMENT, XeSTRING_ALIGN_LEFT,  XeTEXT, LTmpStr,  0);if(LXeStr)
			{
				EXeSetArg(XeNlabelString, LXeStr);
			}

			_PolyGroupStatsLabelW = XeWidget_ManagedNew("", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);
			if(LXeStr) XeString_Free(LXeStr);

			EXeStart;
			_PolyGroupStatsValuesLabelW = XeWidget_ManagedNew("", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);
		}

		{
/*
			EXeStart;
			EXeSetArg(XeNmarginSizeY, 0);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			EXeSetArg(XeNheightAdjustable, FALSE);
			XeToolTip8("Lay out Polygons of this PolyGroup in UV coordinate space, based on their surface area.\nThis ensures a uniform pixel density on the surface of the Polygons, which is necessary\nfor rendering optimal light maps and other types of pre-computed textures.\nSelect UV mapping for the light map texture.");
			LW = XeWidget_ManagedNew("Lay out UVs", XePushButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNactivateCallback, _DCB_PolyGroupTextureLayout, (EPointer)0);
*/
		}


// VertexAttributes
//
		{
			EXeStart;
			EXeSetArg(XeNspacingLeft, LLeftMargin);
			EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
			XeWidget_ManagedNew("Vertex attributes", XeLabelWidget_class, LTopMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNspacingLeft, LLeftMargin);
			EXeSetArg(XeNorientation, XeHORIZONTAL);
			EXeSetArg(XeNmarginBottom, 4);
			EXeSetArg(XeNcolumns, 4);
			EXeSetArg(XeNradioBehavior, TRUE);EXeSetArg(XeNradioAlwaysOne, FALSE);
			EXeSetArg(XeNspacingX, 6);EXeSetArg(XeNspacingY, 0);
			EXeSetArg(XeNspacingRight, 6);
			EXeSetArg(XeNheightAdjustable, FALSE);
			Widget  LMatrixW = XeWidget_ManagedNew("VertexAttributesMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);
			_PolyGroup_VertexAttributesMatrixW = LMatrixW;

			_UpdateVertexAttributesGUI(&PPolyGroup->VertexAttributes, _PolyGroup_VAStartIndex, _PolyGroup_VertexAttributesMatrixW, &_PolyGroup_VertexAttributesGUI,  _DCB_PolyGroupVertexAttribute, NULL);
		}



		Xe_TraversalRing(LTraverseWs, LTraverseN);

		_PolyGroupUpdateStatsLabel(PPolyGroup);

		XeShell_Raise(_PolyGroup_DialogW);
		XeShell_CenterToWidget(_PolyGroup_DialogW, E3dp_TopShellW);
	}
	else
	{
		if(XeShell_DisplayStatus(_PolyGroup_DialogW) != XeDISPLAYED)
		{
			E3dObject_LockAdd((E3dObject*)PPolyGroup->Geometry, &_PolyGroupMeshLock);
			_PolyGroup_UpdateDialog(PPolyGroup);
			XeShell_Raise(_PolyGroup_DialogW);
		}

		_UpdateVertexAttributesGUI(&PPolyGroup->VertexAttributes, _PolyGroup_VAStartIndex, _PolyGroup_VertexAttributesMatrixW, &_PolyGroup_VertexAttributesGUI,  _DCB_PolyGroupVertexAttribute, NULL);
	}

	if(!_Edit_PolyGroupSceneCallbackAdded)
	{
		E3dScene_CallbackAdd(E3d_Scene, E3dChgSELECTION|E3dChgMESHSHAPE, _EditCB_ScenePolyGroup, (EPointer)0);_Edit_PolyGroupSceneCallbackAdded = TRUE;
	}
}
