//================================================================
// EQUINOX-3D(TM) plugin
//
// - Edit selected Object (Geometry, Light etc.)
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-01 23:27:57
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/E3D.h>
#include <E3D/Math.h>
#include <E3D/Matrix.h>
#include <E3D/Scene.h>


#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage

#include <Xe/Label.h>
#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/Separator.h>
#include <Xe/Shell.h>
#include <Xe/PushButton.h>
#include <Xe/TextField.h>
#include <Xe/ToggleButton.h>

#include <Xe/Dialog.h>

#include "Edit.h"

#include "../EQXExt.h"


static EPluginLock*  _ObjectDialogLock = NULL;

static void          _CB_ObjectSelection(ECoreObject* LScn, EPointer PClientData, EPointer PCallData);
static EBool         _ObjectSelectionCallbackAdded = FALSE;


static E3dObjectPanel*  _ObjectPanel = NULL;
static Widget           _ObjectDialogW = NULL;


//========================================
// Update PointSet stats label
//========================================
static void _PointSetUpdateStatsLabel(E3dPointSet* PPointSet, Widget PStatsValLW)
{
	char  LTmpStr[256];sprintf(LTmpStr, "%lld", (long long)(PPointSet->Points_Count));

	XeString  LXeStr = XeString_NewVa(TRUE,  XeFONT, XeFontDEFAULT,  XeALIGNMENT, XeSTRING_ALIGN_RIGHT,  XeTEXT, LTmpStr,  0);if(LXeStr)
	{
		EArg  Args[2];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNlabelString, LXeStr);
		Xe_SetValues(PStatsValLW,  Args, ArgCnt);

		XeString_Free(LXeStr);
	}
}


typedef struct
{
 Widget  StatsValuesLabelW;
} E3dPointSetResourcesGUI;


//================================================================
// GUIProc for E3dPointSet
//================================================================
void _PointSet_GUI(ECoreObject* PObject, void* PPanel, void* PDrawer, const int PAction)
{
	XeComponentClassDrawer*   LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dPointSetResourcesGUI*  LResourcesGUI;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dPointSetResourcesGUI*)(LDrawer->ResourcesGUI);
				_PointSetUpdateStatsLabel((E3dPointSet*)PObject, LResourcesGUI->StatsValuesLabelW);
			}
			else
			{
				Widget  LTopMatrixW = LDrawer->Drawer.DrawerW;
				char    LTmpStr[256];

				ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

				EArg    Args[16];int  ArgCnt;

				EXeStart;
				EXeSetArg(XeNspacingLeft, Xe_DrawerIndent);
				Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

				EXeStart;
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNmarginSizeX, 4);
				EXeSetArg(XeNspacingX, 8);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				EXeSetArg(XeNheightAdjustable, FALSE);
				Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

				EXeStart;

				sprintf(LTmpStr, "Points:");
				XeString  LXeStr = XeString_NewVa(TRUE,
														XeFONT, XeFontDEFAULT,  XeALIGNMENT, XeSTRING_ALIGN_LEFT,  XeTEXT, LTmpStr,
														0);
				if(LXeStr) EXeSetArg(XeNlabelString, LXeStr);

				XeWidget_ManagedNew("", XeLabelWidget_class, LMatrixW, Args, ArgCnt);
				if(LXeStr) XeString_Free(LXeStr);

				EXeStart;
				LResourcesGUI->StatsValuesLabelW = XeWidget_ManagedNew("", XeLabelWidget_class, LMatrixW, Args, ArgCnt);

				_PointSetUpdateStatsLabel((E3dPointSet*)PObject, LResourcesGUI->StatsValuesLabelW);
			}
		break;
	}
}


//========================================================================
// Callback for Object changes to update the appropriate ObjectPanel
//========================================================================
static EChangeFlags _ObjectChanged(EInterface* PInterface, EInterfaceCallbackStruct* PCBS)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PInterface->ClientData;

//printf("%s() %s:%d  <%s>\n", __FUNCTION__, __FILE__, __LINE__, LPanel->Object ? LPanel->Object->Name : "");fflush(stdout);
	if(LPanel->Object) E3dObjectPanel_Update(LPanel, LPanel->Object, NULL);

// We don't change the Object, so return 0
//
	return(0);
}


static EChangeFlags _ObjectDoneEditing(E3dObjectPanel* PPanel)
{
	EChangeFlags  LChanged = 0;

	if(PPanel->Object)
	{
		if(E3dObject_LockDel(PPanel->Object, &E3dObject_EditLock)) LChanged |= E3dChgLOCK_UNLOCK;

		if(EObject_IsSubClass(PPanel->Object, E3dSpline))
		{
			E3dSpline*    LSpline = (E3dSpline*)(PPanel->Object);
			E3dSplineCV*  LCVs = LSpline->CVs;

			EIndex  LC = 0;const EIndex  LN = LSpline->CVs_Count;for(;LC < LN;LC++) LCVs[LC].CVFlags &= ~E3dCV_ACTIVE;
			LChanged |= E3dChgSELECTION;
		}

		if(EObject_IsSubClass(PPanel->Object, E3dGeometry))
		{
			E3dBaseObject_RemoveOutput((E3dBaseObject*)PPanel->Object, &PPanel->InputInterface);
		}
		E_Delete(PPanel->Object);
		E_DeleteChk(PPanel->SubObject);
	}

	return(LChanged);
}


//========================================
// Close Object dialog-box
//========================================
static void _ObjectDialogClose(E3dObjectPanel* PPanel, const EBool PRedraw)
{
	if(PPanel && PPanel->Object)
	{
		E3dObject*  LObject = PPanel->Object;

		{
			EIndex  LC = 0;for(;LC < LObject->Components_Count;LC++)
			{
				E3dObjectComponent*  LInfo = LObject->Components[LC];
				if(LInfo->Class && LInfo->Class->GUI) LInfo->Class->GUI(LObject, LInfo, PPanel, NULL, E_CLOSE);
			}
		}

		EChangeFlags	LChanged = 0;

		Xe_WidgetUnmap(_ObjectDialogW);
		LChanged |= _ObjectDoneEditing(PPanel);

		if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);

		E3dScene_CallbackRemove(E3d_Scene, _CB_ObjectSelection, (EPointer)0);_ObjectSelectionCallbackAdded = FALSE;
	}
}


//========================================
// Object dialog-box callback handler
//========================================
static void _DCB_ObjectDialogClose(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_ObjectDialogClose((E3dObjectPanel*)PClientData, TRUE);

	E_PluginLockDel(_EditInfoPlugin, &_ObjectDialogLock, FALSE);
}


static void _EH_CloseDialogBox(Widget PW, XeEvent* PEvent, EPointer PClientData, EBool* PCdr)
{
	switch(PEvent->Type)
	{
		case XeKeyPress:
			{
				const XeKeyEvent*  LEvent = (const XeKeyEvent*)PEvent;
//printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
				if((LEvent->XeKeySym & 0xFFFF) == EKeyESCAPE)  // Esc key
				{
printf("%s() %s:%d  %04llx\n", __FUNCTION__, __FILE__, __LINE__, (unsigned long long)(LEvent->XeKeySym & 0xFFFF));fflush(stdout);
					_ObjectDialogClose((E3dObjectPanel*)PClientData, TRUE);
					E_PluginLockDel(_EditInfoPlugin, &_ObjectDialogLock, FALSE);
				}
			}
		break;

		case XeKeyRelease:
		break;
	}
}


static int _ObjectDialogLockDel(EPlugin* PPlugin, EPluginLock* PLock)
{
	_ObjectDialogClose(_ObjectPanel, FALSE);

	return(E_SUCCESS);
}


//========================================================
// Common dialog-box creator for Mesh, Light etc.
//========================================================
static void _ObjectPopupDialog(E3dModel* PModel, E3dObject* PObject, EObject* PSubObject, EIndex PPointIndex, E3dObjectComponentClass* POpenFolderForThisComponentClass)
{
	char   LTitle[256];
	char*  LClassName = PObject->Class->Name;

	if(strncmp("E3d", LClassName, 3) == 0) LClassName += 3;
	if(strlen(LClassName) < 32) sprintf(LTitle, "%s attributes", LClassName);
	else sprintf(LTitle, "Attributes");

	if(_ObjectDialogLock == NULL) _ObjectDialogLock = E_PluginLockAdd(_EditInfoPlugin, _ObjectDialogLockDel);

	E3dObjectPanel*  LPanel;

	if(_ObjectDialogW == NULL)
	{
		EArg  Args[32];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNallowShellResize, TRUE);
		EXeSetArg(XeNbuttonsStretch, FALSE);
		EXeSetArg(XeNminSizeX, 260);EXeSetArg(XeNminSizeY, 100);
		EXeSetArg(XeNworkAreaType, XeMATRIX);
		EXeSetArg(XeNcreateButtons, FALSE);
		EXeSetArg(XeNwmDecorations, XeWMDecorCLOSE|XeWMDecorBORDER|XeWMDecorRESIZEH|XeWMDecorTITLE);
		EXeSetArg(XeNtitle, LTitle);EXeSetArg(XeNiconName, LClassName);
		Widget  LDialogW = XeDialog_New(XeDIALOG_TEMPLATE, LClassName, E3dp_TopShellW,  Args, ArgCnt);if(LDialogW == NULL) return;

		_ObjectDialogW = LDialogW;

		Widget  LTopMatrixW = XeDialog_GetChild(LDialogW, XeDIALOG_WORK_AREA);
		EXeStart;
		EXeSetArg(XeNorientation, XeVERTICAL);
		EXeSetArg(XeNmarginSizeX, 4);EXeSetArg(XeNmarginSizeY, 2);
		EXeSetArg(XeNspacingY, 2);
		Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

		_ObjectPanel = LPanel = E3dObjectPanel_Create(LTopMatrixW, ~0, PObject, PSubObject, PPointIndex,  POpenFolderForThisComponentClass);
		XeEventHandler_Add(LPanel->NameW, XeKeyPressMask|XeKeyReleaseMask, TRUE, _EH_CloseDialogBox, (EPointer)LPanel);
		XeAddCallback(LDialogW, XeNwmCallback, _DCB_ObjectDialogClose, (EPointer)LPanel);

		XeShell_Raise(LDialogW);
		XeShell_CenterToWidget(LDialogW, E3dp_TopShellW);

		if(E3dObject_LockAdd(PObject, &E3dObject_EditLock)) E3dWindows_RedrawByChangeFlags(E3dChgLOCK_UNLOCK);

// Initialize InputInterface for _Objecthanged()
//
		{
			EInterface*  LInput = &LPanel->InputInterface;  // E_InterfaceInit() was called in E3dObjectPanel_Create
			LInput->Name         = EStrDup("Follow selection");
			LInput->InputChanged = _ObjectChanged;
			LInput->ClientData   = (EPointer)LPanel;  // We use ClientData, rather than Object, because E3dObjectPanel is not an EObject-derivative
//			LInput->Object       = (EObject*)LPanel;
			LInput->Internal     = TRUE;
		}

		if(EObject_IsSubClass(PObject, E3dGeometry)) E3dBaseObject_AddOutput((E3dBaseObject*)PObject, &LPanel->InputInterface);
	}
	else
	{
// Dialog-box is already up.
//
		Widget        LDialogW = _ObjectDialogW;
		EChangeFlags  LChanged = 0;

		LPanel = _ObjectPanel;

		XeDialog_SetValuesVa(LDialogW,  XeNtitle, LTitle,  XeNiconName, LClassName,  NULL);

		if(LPanel->Object == PObject)
		{
//printf("%s() %s:%d <%s> <%s>\n", __FUNCTION__, __FILE__, __LINE__, PObject->Name, PSubObject->Name);fflush(stdout);

// Same Object
//
			E3dObjectPanel_UpdateSubSub(LPanel, PObject, PSubObject, PPointIndex,  POpenFolderForThisComponentClass);
			XeShell_Raise(LDialogW);
			return;
		}
		else if(LPanel->Object)
		{
// A different Object was being edited on this Panel
//
			LChanged |= _ObjectDoneEditing(LPanel);
		}

		E3dObjectPanel_UpdateSubSub(LPanel, PObject, PSubObject, PPointIndex, POpenFolderForThisComponentClass);
		if(E3dObject_LockAdd(PObject, &E3dObject_EditLock)) LChanged |= E3dChgLOCK_UNLOCK;

		if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);

		if(EObject_IsSubClass(PObject, E3dGeometry)) E3dBaseObject_AddOutput((E3dBaseObject*)PObject, &LPanel->InputInterface);
	}

// Prevent the dialog box stealing keyboard focus, by only raising it, if it's not visible
//
	if(XeShell_DisplayStatus(_ObjectDialogW) != XeDISPLAYED) XeShell_Raise(_ObjectDialogW);

	if(Eqx_Prefs.SelectAndFocusOnName)
	{
		XeDisplay*  LDisplay = Xe_GetDisplay(LPanel->NameW);
		if(!E_IsKeyDown(LDisplay, 0))
		{
			Xe_SetFocus(LPanel->NameW, TRUE, TRUE);
			XeTextField_SelectAll(LPanel->NameW);
		}
	}

	if(!_ObjectSelectionCallbackAdded)
	{
		E3dScene_CallbackAdd(E3d_Scene, E3dChgSELECTION, _CB_ObjectSelection, (EPointer)0);_ObjectSelectionCallbackAdded = TRUE;
	}
}


//========================================
// Edit Object attributes
//========================================
void _Object_Edit(E3dModel* PModel, E3dObject* PObject, const EIndex PLODIndex,  EObject* PSubObject, const EIndex PPointIndex,  E3dObjectComponentClass* POpenFolderForThisComponentClass)
{
//	E3dObject*  LLOD = PObject;if(PLODIndex) LLOD = (E3dObject*)(((E3dGeometry*)PObject)->LODs[PLODIndex - 1]);

	if(EObject_IsSubClass(PObject, E3dGeometry))
	{
//		E3dGeometry*  LGeometry = (E3dGeometry*)PObject;LGeometry->AutoLOD = FALSE;
		_ObjectPopupDialog(PModel, PObject, PSubObject, PPointIndex,  POpenFolderForThisComponentClass);
	}
	else _ObjectPopupDialog(PModel, PObject, NULL, -1,  POpenFolderForThisComponentClass);
}


//========================================
// Scene changed callback
// Called from Edit.c
//========================================
static void _CB_ObjectSelection(ECoreObject* PScn, EPointer PClientData, EPointer PCallData)
{
	E3dCallbackStruct*  LCBS = (E3dCallbackStruct*)PCallData;
	E3dObjectPanel*     LPanel = _ObjectPanel;

	if(LPanel && LPanel->FollowSelection)
	{
		switch(LCBS->Reason)
		{
			case E3dCR_SPLINE_SEGMENT_SELECTED:
			 {
				 E3dSceneObjectCallbackStruct*  LCBSO = (E3dSceneObjectCallbackStruct*)PCallData;
				 E3dObject*                     LObject = LCBSO->Object;
				 E3dModel*                      LModel = LObject->Models[0];

				 if(EObject_IsSubClass(LObject, E3dFace)) _Object_Edit(LModel, LObject, 0,  LCBSO->SubObject, LCBSO->PointIndex,  NULL);
				 else _Object_Edit(LModel, LObject, 0,  NULL, LCBSO->PointIndex,  NULL);
			 }
			break;

			case E3dCR_SELECTION_CHANGED:
				{
					E3dItem	LSelectedItem;

					switch(E3dScene_GetSingleSelection(E3d_Scene, &LSelectedItem, E3dItemALL, 0))
					{
						case E3dItemOBJECT:
						case E3dItemCAMERA:
						case E3dItemLIGHT:
						case E3dItemPOINTSET:
						case E3dItemSPLINE:
						case E3dItemFACE:
						case E3dItemMESH:
							{
								E3dObjItem*  LObjItem = &LSelectedItem.ObjItem;
								_Object_Edit((E3dModel*)(LObjItem->Node), LObjItem->Object, 0,  NULL, 0,  NULL);
							}
						break;

						case E3dItemPOLYGROUP:
							if(_Edit_SelectedPolyGroup)
							{
								E3dPolyItem*  LPolyItem = &LSelectedItem.PolyItem;

								_Edit_PolyGroupPopupDialog(LPolyItem->PolyGroup, NULL, FALSE);
							}
						break;
					}
				}
			break;

			case E3dCR_POINTS_SELECTION:
				{
					E3dSceneObjectCallbackStruct*  LCBSO = (E3dSceneObjectCallbackStruct*)PCallData;
					E3dObject*                     LObject = LCBSO->Object;if(LObject)
					{
						E3dModel*  LModel = LObject->Models[0];

						_Object_Edit(LModel, LObject, 0,  NULL, 0,  NULL);
					}
				}
			break;
		}
	}
}


//========================================================================================
// Return TRUE, if PNode's attributes are being edited (the dialog box is displayed)
//========================================================================================
static EBool _Object_IsEdited(E3dObject* PObject)
{
	if(_ObjectPanel && _ObjectPanel->Object == PObject) return(TRUE);
	return(FALSE);
}


void E3dObjectEdit_Init(void)
{
	E3dObject_EditFunc = _Object_Edit;
	E3dObject_IsEditedFunc  = _Object_IsEdited;
}


//========================================
// Clean up EditObject
//========================================
void E3dObjectEdit_Cleanup(void)
{
	E3dScene*  LScene = E3d_Scene;

	E3dObject_EditFunc = NULL;
	E3dObject_IsEditedFunc = NULL;

	if(_ObjectPanel) { E3dObjectPanel_Free(_ObjectPanel);_ObjectPanel = NULL; }

	if(_Edit_PolyGroupSceneCallbackAdded)
	{
		E3dScene_CallbackRemove(LScene, _EditCB_ScenePolyGroup, (EPointer)0);_Edit_PolyGroupSceneCallbackAdded = FALSE;
	}

//	XeWidget_Free(_ObjectDialogW);  // Don't free top-level shells
}
