//================================================================================================
// Plugin for EQUINOX-3D
//
// Example to show how to create custom Node.
// This plugin visualizes the CIE  RGB <-> color spectrum curves.
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 03:53:32
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================================================
#include <assert.h>
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/Matrix.h>
#include <E3D/Scene.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>


#include <Xe/Matrix.h>
#include <Xe/PushButton.h>
#include <Xe/Scale.h>
#include <Xe/ToggleButton.h>

#include <Color/Panel.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>


char*  E_PluginInfo = "Visualize BRDFs - Compare ShadeSurface, SampleDirection and SampleDirections functions of a Shader,\nfor debugging.";

static EPlugin*    _Plugin = NULL;

static EguiItem    _MenuButtons[16];static int  _NumOfMenuButtons = 0;
static EguiItem    _ToolbarButtons[16];static int  _NumOfToolbarButtons = 0;

#define E3dMAX_NEAREST_POINTS  65536

EpIconDef3(BRDFVis)

typedef struct
{
	E3dKDTreePointCore;
	float  Intensity;
} E3dBRDFSamplePoint;


typedef struct
{
	E3dMesh*        Mesh;
	E3dCoordinate*  UnitSpherePoints;EIndex  NumOfUnitSpherePoints;
	E3dCoordinate*  RotatedUnitSpherePoints;  // The sphere Mesh is oriented, so its "north/south axis" aligns with the mirror reflection vector, to get the best resolution
	EBool           GrabUnitSphere;           // Sphere Mesh topology updated, cache Vertex positions as "unit sphere"
} E3dVisGeo;


// Direction of sampling. BRDFs are reciprocal by definition (B meaning bi-directional), but this allows us to visualize either way
enum
{
	E3dSAMPLE_LIGHTDIRS = 1,    // Fixed viewing direction, sample incoming light - this is how shader importance sampling works (SampleDirection() and SampleDirections() functions)
	E3dSAMPLE_VIEWDIRS          // Fixed light direction, sample viewing directions
};


// How to reconstruct the BRDF surface from importance-sampling.
enum
{
	E3dGAUSSIAN_SPLATS = 1,  // Density estimation, using gaussian splats on the hemisphere and kd-trees for look-up (a-la photon mapping)
	E3dSPHERICAL_HARMONICS,
	E3dWAVELET
};


typedef struct
{
	E3dNodeComponentCore;

	E3dCoordinate      BlackRadius, Scale;

	int                Dimensions;

	E3dMaterial*       Material;

	EBool              OrientSphere;

	E3dVisGeo          ShadeSurface;
	E3dVisGeo          SampleDirection;
	E3dVisGeo          SampleDirections;

	int                SamplesSqrt;
	float              SplatRadius;        // For E3dGAUSSIAN_SPLATS integration
	uint8_t            SamplingDirection;  // E3dSAMPLE_LIGHTDIRS or E3dSAMPLE_VIEWDIRS
	uint8_t            IntegrationMethod;  // E3dGAUSSIAN_SPLATS etc.

	E3dSamplingCache   SamplingCache;

	E3dKDTreePointMap           PointMap;       // Sample points are stored in a KD-tree, so we can do fast nearest neighbour queries for density estimation (to integrate importance-sampled directions)
	E3dNearestKDTreePointList*  NearestPoints;  // For collecting nearest neighbours
} E3dInfo;


#define ROffset(field)  (void*)(EPtrOffset(E3dInfo, field))
static EResource  _Resources[]=
{
	{ .Name= "Black radius",      .Type= E3dCOORDINATE_TYPE,  .Address= ROffset(BlackRadius) },
	{ .Name= "Scale",             .Type= E3dCOORDINATE_TYPE,  .Address= ROffset(Scale) },

	{ .Name= "Material",          .Type= EResOBJECT,          .Address= ROffset(Material) },

	{ .Name= "ShadeSurface",      .Type= EResOBJECT,          .Address= ROffset(ShadeSurface.Mesh) },
	{ .Name= "SampleDirection",   .Type= EResOBJECT,          .Address= ROffset(SampleDirection.Mesh) },
	{ .Name= "SampleDirections",  .Type= EResOBJECT,          .Address= ROffset(SampleDirections.Mesh) },

	{ .Name= "SamplesSqrt",       .Type= EResUINT32,          .Address= ROffset(SamplesSqrt) },
	{ .Name= "SamplingDirection", .Type= EResUINT8,           .Address= ROffset(SamplingDirection) },
	{ .Name= "IntegrationMethod", .Type= EResUINT8,           .Address= ROffset(IntegrationMethod) },
	{ .Name= "SplatRadius",       .Type= EResFLOAT32,         .Address= ROffset(SplatRadius) },

	{ }
};
#undef ROffset


typedef struct
{
	Widget  SampleViewDirsW, SampleLightDirsW,  BlackRadiusW, ScaleW, SamplesSqrtW, SplatRadiusW, OrientSphereW;
} E3dResourcesGUI;


static E3dNodeComponentClass*  _NodeComponentClass = NULL;


static void _CheckMesh(E3dInfo* PInfo, E3dVisGeo* PVisGeo)
{
	const EIndex  LDimensions = PInfo->Dimensions;

	if(PVisGeo->GrabUnitSphere)
	{
		E3dMesh*  LMesh = PVisGeo->Mesh;

		{
			const E3dVertex*  LVertex = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;const EIndex  LN = LMesh->Vertices_Count;
			EResizeNT(E3dCoordinate, LN * LDimensions,  PVisGeo->UnitSpherePoints);PVisGeo->NumOfUnitSpherePoints = LN;
			EResizeNT(E3dCoordinate, LN * LDimensions,  PVisGeo->RotatedUnitSpherePoints);

			{
				EIndex	LC = 0;E3dCoordinate*  LUnitSpherePoint = PVisGeo->UnitSpherePoints;for(;LC < LN;LC++, EStruct_Inc(LVertex, LVertexSize), LUnitSpherePoint += LDimensions)
				{
					LUnitSpherePoint[0] = LVertex->Position_3[0];
					LUnitSpherePoint[1] = LVertex->Position_3[1];
					LUnitSpherePoint[2] = LVertex->Position_3[2];

					E3d_Normalize3_a(LUnitSpherePoint);
				}
			}
		}

		PVisGeo->GrabUnitSphere = FALSE;
	}

	if(PInfo->OrientSphere)
	{
		E3dMatrix  LMatrix;E3dMatrix_LoadIdentity(LMatrix);

// Orient the sphere, so its "north pole" is along the mirror reflection vector
//
		{
			E3dpWindow*   L3DWindow = E3dp_Main3DWindow;
			E3dRenderer*  LRenderer = L3DWindow->RendererSet->Renderer;
			if(LRenderer->VisibleLights)
			{
				const E3dCoord3  LPosition = { 0.0, 0.0, 0.0 };
				E3dCoord3        LNormal = { 0.0, 1.0, 0.0 };
				const E3dCoord3  L1 = { 0.0, 1.0, 0.0 };

				E3dRLightInstance*  LLight = LRenderer->VisibleLights;

				{
					E3d_Sub3i_vv  (LMirrorReflectionDir,  LPosition, LLight->Position);E3dCoordinate  LDist;E3d_NormalizeGetLength3_v(LMirrorReflectionDir, LDist);

					E3d_Reflect(&LMirrorReflectionDir, &LNormal);

//printf("%s() %s:%d  %f %f %f\n", __FUNCTION__, __FILE__, __LINE__, XYZ_s(LMirrorReflectionDir));fflush(stdout);

					E3d_VectorsGetRotMatrix(&L1, &LMirrorReflectionDir, LMatrix);
				}
			}
		}

		if(PVisGeo->RotatedUnitSpherePoints)
		{
			EIndex	LC = 0;const EIndex  LN = PVisGeo->NumOfUnitSpherePoints;const E3dCoordinate*  LUnitSpherePoint = PVisGeo->UnitSpherePoints;E3dCoordinate*  LRotatedUnitSpherePoint = PVisGeo->RotatedUnitSpherePoints;for(;LC < LN;LC++, LUnitSpherePoint += LDimensions, LRotatedUnitSpherePoint += LDimensions)
			{
				LRotatedUnitSpherePoint[0] = LUnitSpherePoint[0];
				LRotatedUnitSpherePoint[1] = LUnitSpherePoint[1];
				LRotatedUnitSpherePoint[2] = LUnitSpherePoint[2];

				E3dMatrix_Transform3x3_a(LMatrix, LRotatedUnitSpherePoint);
				if(LRotatedUnitSpherePoint[1] < 0.0) LRotatedUnitSpherePoint[1] = 0.0;
				E3d_Normalize3_a(LRotatedUnitSpherePoint);
			}
		}
	}
}


//================================================
// Guassian filter with precomputed constants
//================================================
static inline E3dCoordinate _GaussianFilterPrep(const E3dCoordinate PDistanceSq, const E3dCoordinate POneOverTwoRadiusSq)
{
	const E3dCoordinate  LAlpha = 0.918, LBeta = 1.953, LK = PDistanceSq * POneOverTwoRadiusSq,
	                     LDenom = 0.8581521110346772563,  //  = 1.0 - exp(-LBeta)
	                     LNumerator = 1.0 - exp(-LBeta * LK);

//	return(1.0 - LAlpha * (1.0 - LNumerator / LDenom));  // Inverted gaussian : it distributes better, with less "sparkling" around corners
	return(LAlpha * (1.0 - LNumerator / LDenom));
}


static E3dModel* _SceneGetFirstVisibleLightModel(E3dScene* PScene)
{
	EIndex  LRC = 0;const EIndex  LRN = PScene->RootNodes_Count;for(;LRC < LRN;LRC++)
	{
		E3dNode*  LNode = PScene->RootNodes[LRC];for(;LNode;LNode = LNode->Next)
		{
			if(LNode->Visible)
			{
				E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
				{
					EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
					{
						E3dObject*  LObject = LModel->Objects[LObjC];if(LObject->Visible && EObject_IsSubClass(LObject, E3dLight)) return(LModel);
					}
				}
			}
		}
	}
	return(NULL);
}


static EChangeFlags _Update(E3dNode* PChangedBranchRootNode, E3dNode* PNode, E3dNodeComponent* PInfo, const EChangeFlags PChanged, const EBool PCheckOnly)
{
	if(PCheckOnly) return(0);

	E3dInfo*      LInfo = (E3dInfo*)PInfo;
	E3dModel*     LModel = (E3dModel*)PNode;
	E3dMesh*      LShadeSurfaceMesh     = NULL;
	E3dMesh*      LSampleDirectionMesh  = NULL;
	E3dMesh*      LSampleDirectionsMesh = NULL;
	EChangeFlags  LTriggeredChanges = 0;

	if(LInfo == NULL) return(0);

	if(PChanged == 0)
	{
// Just check consistency
//
		if(LModel->Objects_Count == 1)
		{
			LShadeSurfaceMesh     = LInfo->ShadeSurface.Mesh;
			LSampleDirectionMesh  = LInfo->SampleDirection.Mesh;
			LSampleDirectionsMesh = LInfo->SampleDirections.Mesh;

			LTriggeredChanges |= E3dChgMESHSHAPE;

			E3dObjectComponentClass*	LClass = E3dObjectComponent_ClassFindByName("Sphere");if(LClass)
			{
				{
					E3dObjectComponent*  LSphereInfo = E3dObject_ComponentNew(LClass);if(LSphereInfo)
					{
						ERes_NamedResourceSet1f(LInfo, "Radius", 1.0);
						E3dObject_CustomUpdate((E3dObject*)LShadeSurfaceMesh, LSphereInfo, E3dUF_ALL - E3duf_CALLS, &LTriggeredChanges);
					}
					LInfo->ShadeSurface.GrabUnitSphere = TRUE;
				}

				{
					E3dObjectComponent*  LSphereInfo = E3dObject_ComponentNew(LClass);if(LSphereInfo)
					{
						ERes_NamedResourceSet1f(LInfo, "Radius", 1.0);
						E3dObject_CustomUpdate((E3dObject*)LSampleDirectionMesh, LSphereInfo, E3dUF_ALL - E3duf_CALLS, &LTriggeredChanges);
					}
					LInfo->SampleDirection.GrabUnitSphere = TRUE;
				}

				{
					E3dObjectComponent*  LSphereInfo = E3dObject_ComponentNew(LClass);if(LSphereInfo)
					{
						ERes_NamedResourceSet1f(LInfo, "Radius", 1.0);
						E3dObject_CustomUpdate((E3dObject*)LSampleDirectionsMesh, LSphereInfo, E3dUF_ALL - E3duf_CALLS, &LTriggeredChanges);
					}
					LInfo->SampleDirections.GrabUnitSphere = TRUE;
				}
			}
		}

		return(0);
	}
	else
	{
		LShadeSurfaceMesh     = LInfo->ShadeSurface.Mesh;
		LSampleDirectionMesh  = LInfo->SampleDirection.Mesh;
		LSampleDirectionsMesh = LInfo->SampleDirections.Mesh;
	}

	if(LShadeSurfaceMesh     && LShadeSurfaceMesh->Vertices_Count     != LInfo->ShadeSurface.NumOfUnitSpherePoints)     LInfo->ShadeSurface.GrabUnitSphere     = TRUE;
	if(LSampleDirectionMesh  && LSampleDirectionMesh->Vertices_Count  != LInfo->SampleDirection.NumOfUnitSpherePoints)  LInfo->SampleDirection.GrabUnitSphere  = TRUE;
	if(LSampleDirectionsMesh && LSampleDirectionsMesh->Vertices_Count != LInfo->SampleDirections.NumOfUnitSpherePoints) LInfo->SampleDirections.GrabUnitSphere = TRUE;

	if(PChanged & (E3dChgLIGHT|E3dChgATTRIBUTES) && LShadeSurfaceMesh && LSampleDirectionMesh && LSampleDirectionsMesh)
	{
		_CheckMesh(LInfo, &LInfo->ShadeSurface);
		_CheckMesh(LInfo, &LInfo->SampleDirection);
		_CheckMesh(LInfo, &LInfo->SampleDirections);

		{
			E3dpWindow*      L3DWindow = E3dp_Main3DWindow;
			E3dRenderer*     LRenderer = L3DWindow->RendererSet->Renderer;
			E3dSurfacePoint  LPoint;
			E_SetXYZ_s(LPoint.Position,  0.0, 0.0, 0.0);
			E_SetXYZ_s(LPoint.Normal,    0.0, 1.0, 0.0);
			E_SetXYZ_s(LPoint.Tangent,   1.0, 0.0, 0.0);

			if(LRenderer->VisibleLights)
			{
				E3dScene*  LScene = E3d_Scene;
				E3dModel*  LSceneGetFirstVisibleLightModel = _SceneGetFirstVisibleLightModel(LScene);
				E3dCoord3  LLightPosition;E_SetXYZ_s(LLightPosition,  LSceneGetFirstVisibleLightModel->LocalToWorldMatrix[M30], LSceneGetFirstVisibleLightModel->LocalToWorldMatrix[M31], LSceneGetFirstVisibleLightModel->LocalToWorldMatrix[M32]);

				E3dMaterial*  LSurfaceMaterial = LInfo->Material;if(LSurfaceMaterial)
				{
					E3dShader*    LShader = LSurfaceMaterial->Shader;
					int           LNumOfSpectralSamples = LRenderer->NumOfSpectralSamples;
					E3d_Sub3i_vv  (LPointToLight,  LLightPosition, LPoint.Position);E3dCoordinate  LLightDistance;E3d_NormalizeGetLength3_v(LPointToLight, LLightDistance);

					const E3dShaderClass*  LShaderClass = LShader->Class;

					{
// Visualize the Shader's ShadeSurface() function
//
						E3dVisGeo*  LVisGeo = &LInfo->ShadeSurface;
						E3dMesh*    LMesh   = LVisGeo->Mesh;
						if(LMesh->Vertices_Count == LVisGeo->NumOfUnitSpherePoints)
						{
							const E3dCoordinate*  LUnitSpherePoint = LVisGeo->RotatedUnitSpherePoints;
							E3dVertex*            LVertex = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;const EIndex  LDimensions = LInfo->Dimensions;
							ERandomState          LRandomState[5];E_RandomInit(42, LRandomState);

							E3dShaderLight  LShaderLight;
							E3dCoordinate   LNdotL;

							if(LInfo->SamplingDirection == E3dSAMPLE_VIEWDIRS)
							{
								LNdotL = E3d_Dot3_vv(LPoint.Normal, LPointToLight);
								LShaderLight.PointToLightXYZ = LPointToLight;
								LShaderLight.PointToLight.W = LLightDistance;

								LShaderLight.NdotL = LNdotL;
								LShaderLight.SolidAngle = 0.0;

								ESetN(LShaderLight.Power, LNdotL, LNumOfSpectralSamples);  // ShadeSurface expects NdotL to be already applied to LShaderLight.Power.
								LShaderLight.Shadowed = FALSE;
								LShaderLight.ImportanceSampled = FALSE;
							}

							EIndex  LC = 0;const EIndex  LN = LMesh->Vertices_Count;for(;LC < LN;LC++, LUnitSpherePoint += LDimensions, EStruct_Inc(LVertex, LVertexSize))
							{
/*
								{
									E3dCamera*       LCamera = LRenderer->Camera;
									const E3dCoord3  LCameraPosition = { LCameraModel->LocalToWorldMatrix[M30], LCameraModel->LocalToWorldMatrix[M31], LCameraModel->LocalToWorldMatrix[M32] };
									E3d_Sub3i_vv     (LPointToViewer,  LCameraPosition, LPoint.Position);E3dCoordinate  LDist;E3d_NormalizeGetLength3_v(LPointToViewer, LDist);
									LPoint.PointToViewerXYZ = LPointToViewer;
									LPoint.PointToViewer.W = LDist;
								}
*/

								if(LInfo->SamplingDirection == E3dSAMPLE_VIEWDIRS)
								{
									LPoint.PointToViewer.X = LUnitSpherePoint[0];
									LPoint.PointToViewer.Y = LUnitSpherePoint[1];
									LPoint.PointToViewer.Z = LUnitSpherePoint[2];
									LPoint.PointToViewer.W = 1.0;
								}
								else
								{
									LPoint.PointToViewer.X = LPointToLight.X;
									LPoint.PointToViewer.Y = LPointToLight.Y;
									LPoint.PointToViewer.Z = LPointToLight.Z;
									LPoint.PointToViewer.W = LLightDistance;

									LShaderLight.PointToLight.X = LUnitSpherePoint[0];
									LShaderLight.PointToLight.Y = LUnitSpherePoint[1];
									LShaderLight.PointToLight.Z = LUnitSpherePoint[2];
									LShaderLight.PointToLight.W = 1.0;
									LNdotL = E3d_Dot3_vv(LPoint.Normal, LShaderLight.PointToLight);
									LShaderLight.NdotL = LNdotL;
									LShaderLight.SolidAngle = 0.0;

									ESetN(LShaderLight.Power, LNdotL, LNumOfSpectralSamples);  // ShadeSurface expects NdotL to be already applied to LShaderLight.Power.
									LShaderLight.Shadowed = FALSE;
									LShaderLight.ImportanceSampled = FALSE;
								}


//							if(LShaderLight.NdotL >= 0.0 && LPoint.PointToViewer.Y >= 0.0)
								if(LShaderLight.NdotL >= 0.0)
								{
//								E3dShaderResult  LShaderResult;E3dM_ShaderResultInit_s(LShaderResult, LSurfaceMaterial, 0, LNumOfSpectralSamples,  NULL, NULL,  NULL, TRUE, FALSE);
									E3dShaderResult  LShaderResult;E3dShaderResult_Init(&LShaderResult, LSurfaceMaterial, 0, LNumOfSpectralSamples,  NULL, NULL, LRandomState,  NULL, TRUE, FALSE);

									LShader->Class->ShadeSurface(&LPoint, LSurfaceMaterial, LSurfaceMaterial->Shader, NULL, LShader,  &LShaderLight, 1,  &LShaderResult);

									if(LShaderResult.Color[0] < 0.0) LShaderResult.Color[0] = 0.0;
									E3dCoordinate  LI = LInfo->BlackRadius + LShaderResult.Color[0] * LInfo->Scale;

									LVertex->Position.X = LUnitSpherePoint[0] * LI;
									LVertex->Position.Y = LUnitSpherePoint[1] * LI;
									LVertex->Position.Z = LUnitSpherePoint[2] * LI;
								}
								else
								{
									E3dCoordinate  LI = LInfo->BlackRadius;

									LVertex->Position.X = LUnitSpherePoint[0] * LI;
									LVertex->Position.Y = LUnitSpherePoint[1] * LI;
									LVertex->Position.Z = LUnitSpherePoint[2] * LI;
								}
							}
						}
					}

					{
//const ETime  LSDStart_ms = E_GetMilliseconds();printf("%s() %s:%d  SampleDirections START\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
// Visualize the Shader's SampleDirections() function
//
						E3dVisGeo*  LVisGeo = &LInfo->SampleDirections;
						E3dMesh*    LMesh   = LVisGeo->Mesh;

//E3dMesh_PolyGroupsFree(LMesh);  // We use point visualization for now...

						E3dMesh_RenderDataFree(LMesh);

						{
							ERandomState       LRandomState[5];E_RandomInit(42, LRandomState);
							const EIndex       LN = LInfo->SamplesSqrt, LNN = LN * LN;
							const EIndex       LNSamplesMult = 2;  // Allocate this many times more samples, because some shaders have multiple lobes that they need to sample independently (e.g. diffuse + specular)

							E3dSamplingCache*  LSamplingCache = &LInfo->SamplingCache;

//						const EIndex       LBRDFCDSMaxSizeSqrt = E3dBRDF_CDF_MAX_SIZE_SQRT;
							const EIndex       LBRDFCDSMaxSizeSqrt = 8, LBRDFCDSMaxSize = LBRDFCDSMaxSizeSqrt * LBRDFCDSMaxSizeSqrt;

							E3dShaderLight*    LSamples;

							if(LInfo->SamplingDirection == E3dSAMPLE_VIEWDIRS)
							{
								ENewN(LNN * LNSamplesMult, LSamples);
								if(LSamplingCache->ShaderLightCacheSize == 0) E3dSamplingCache_Build(LSamplingCache,  LBRDFCDSMaxSize);
							}
							else
							{
								ENewN(LNN * LNSamplesMult, LSamples);
								if(LSamplingCache->ShaderLightCacheSize == 0) E3dSamplingCache_Build(LSamplingCache,  LNN * LBRDFCDSMaxSize);
							}

//{ const ETime  LNow_ms = E_GetMilliseconds();printf("%s() %s:%d  SampleDirections 1 %lld ms\n", __FUNCTION__, __FILE__, __LINE__, (long long)(LNow_ms - LSDStart_ms));fflush(stdout); }

//printf("%s() %s:%d  sizeof(E3dShaderLight) %d\n", __FUNCTION__, __FILE__, __LINE__, sizeof(E3dShaderLight));fflush(stdout);

// SampleDirections() generates point-to-light vectors, based on the surface normal etc., and the view direction...
// BRDFs are reciprocal, so we can just plug in the light direction here.
							LPoint.PointToViewerXYZ = LPointToLight;

							if(LInfo->SamplingDirection == E3dSAMPLE_VIEWDIRS)
							{
// Fixed light direction, sample viewing directions
//
								if(LMesh->Vertices_Count != LNN)
								{
									E3dMesh_VerticesFree(LMesh);
									E3dMesh_VerticesAdd(LMesh, LNN);
								}

								const E3dVertexSize  LVertexSize = LMesh->VertexSize;

printf("%s() %s:%d  %sGREEN%s  LN %d\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode, LN);fflush(stdout);

								EIndex  LC = 0;E3dShaderLight*  LSample = LSamples;E3dVertex*  LVertex = LMesh->Vertices;for(;LC < LNN;LC++)
								{
									const EIndex  LCount = LShaderClass->SampleDirections(&LPoint, LSurfaceMaterial, LSurfaceMaterial->Shader, NULL,  LN, LSamplingCache, 0.0, LRandomState,  0, LNumOfSpectralSamples,  LSample);if(LCount > 0)
									{
//										const float  LNdotL = E3d_Dot3_vv(LSample->PointToLight, LPoint.Normal);

										const E3dCoordinate  LW = LSample->PointToLight.W * LSample->Power[0] / 4.0;
//									const E3dCoordinate  LW = LSample->Power[0];
//									const E3dCoordinate  LW = 1.0;
										LVertex->Position.X = LSample->PointToLight.X * LW;
										LVertex->Position.Y = LSample->PointToLight.Y * LW;
										LVertex->Position.Z = LSample->PointToLight.Z * LW;
										LVertex->Flags = E3dVtxDEFAULT;
										EStruct_Inc(LVertex, LVertexSize);
									}
								}
								LMesh->Vertices_Count = EStruct_GetIndex(LMesh->Vertices, LMesh->VertexSize, LVertex);
							}
							else
							{
// Fixed viewing direction, sample incoming light directions
//
								LPoint.PointToViewerXYZ = LPointToLight;
								LPoint.PointToViewer.W  = LLightDistance;

								const EIndex  LCount = LShaderClass->SampleDirections(&LPoint, LSurfaceMaterial, LSurfaceMaterial->Shader, NULL,  LN, LSamplingCache, 0.0, LRandomState,  0, LNumOfSpectralSamples,  LSamples);

printf("%s() %s:%d  %sSample light dirs%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout);
if(LCount == 0) { printf("%s() %s:%d  %sLCount = 0%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout); }

//printf("%s() %s:%d  %lld  %lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LN, (long long)LCount);fflush(stdout);

// To integrate the importance-sampled BRDF response, we build a KD-tree and do density estimation.
// We do the density estimation via. gaussian splat monte carlo integration (similar to photon mapping).
//
								if(LInfo->IntegrationMethod == E3dGAUSSIAN_SPLATS)
								{
printf("%s() %s:%d    %sGaussian splats%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout);
									E3dKDTreePointMap*  LPointMap = &LInfo->PointMap;

									if(PChanged & (E3dChgMATERIALTYPE|E3dChgMATERIALPARAM|E3dChgLIGHT|E3dChgCAMERA)) LPointMap->Points_Count = 0;

									if(LPointMap->Points_Count == 0)
									{
//										if(LCount > LPointMap->Points_CountAllocated) 
										LPointMap->TmpPoints = (E3dKDTreePoint*)ERealloc(LPointMap->TmpPoints, LPointMap->BytesPerPoint * LCount);LPointMap->Points_CountAllocated = LCount;
										LPointMap->Balanced = FALSE;

//printf("%s() %s:%d Realloc\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
										LPointMap->Points_Count = LCount;

										{
											EIndex  LC = 0;const E3dShaderLight*  LSample = LSamples;E3dBRDFSamplePoint*  LSamplePoint = (E3dBRDFSamplePoint*)LPointMap->TmpPoints;for(;LC < LCount;LC++, LSample++, LSamplePoint++)
											{
												LSamplePoint->Position = LSample->PointToLightXYZ;
												LSamplePoint->Intensity = LSample->PointToLight.W * LSample->Power[0] / 4.0;
//												LSamplePoint->Intensity = 1.0;
												LSamplePoint->Plane = 0;
											}

//printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
											{
												const EBool  LShowProgress = LN > 256;

												if(LShowProgress)
												{
													E3dProgressIndicator_SetMessage("Building KD-tree for BRDF sample points");
													E3dProgressIndicator_Set(0.5, NULL, NULL);
													XeApp_Sync(E3dp_XeApp);XeApp_Sync(E3dp_XeApp);
												}
												E3d_KDTreePointMapBuildTree(LPointMap,  NULL, NULL,  LShowProgress);
												if(LShowProgress)
												{
													E3dProgressIndicator_Set(-1.0, NULL, NULL);
													XeApp_Sync(E3dp_XeApp);
												}
											}
										}
									}


// Now, sample the BRDF at the given points (Sphere Mesh Vertices)
//
									{
										if(LInfo->NearestPoints) { E3d_NearestKDTreePointsFree(LInfo->NearestPoints);LInfo->NearestPoints = NULL; }
										LInfo->NearestPoints = E3d_NearestKDTreePointsNew(E3dMAX_NEAREST_POINTS);

										E3dNearestKDTreePointList*  LNearestPoints = LInfo->NearestPoints;
										const float                 LSplatRadius = LInfo->SplatRadius;
										const float                 LSplatRadiusSq = LSplatRadius * LSplatRadius;
										float*                      LDistanceSq = LNearestPoints->DistanceSq;
										const E3dBRDFSamplePoint**  LPoints = (const E3dBRDFSamplePoint**)LNearestPoints->Points;
										const ETime                 LStart_ms = E_GetMilliseconds();
										ETime                       LLastProgressUpdate_ms = 0;
										EBool                       LShowProgress = FALSE;
										const double                LNormalizer = 5.0 * E3dPI * E3dPI / (double)LNN;

										if(LMesh->Vertices_Count == LVisGeo->NumOfUnitSpherePoints)
										{
											const E3dCoordinate*  LUnitSpherePoint = LVisGeo->RotatedUnitSpherePoints;
											E3dVertex*            LVertex = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;const EIndex  LDimensions = LInfo->Dimensions;

// See SphereSurfaceSamplingWithGaussianSplat00.odp
//
											const E3dCoordinate  LIntersectedArcX = (-LSplatRadiusSq + 2.0) / 2.0;                                 // X of Pi
											const E3dCoordinate  LIntersectedArcLength = E3d_ACos(LIntersectedArcX);                               // Length of arc on the unit circle that is enclosed in the splat
//printf("%s() %s:%d  %f %f\n", __FUNCTION__, __FILE__, __LINE__, LIntersectedArcX, LIntersectedArcLength);fflush(stdout);
											const E3dCoordinate  LIntersectedArcLengthSq = LIntersectedArcLength * LIntersectedArcLength;
											const E3dCoordinate  LIntersectedArea = 2.0 * E3dPI * LIntersectedArcLengthSq;   // Area on the surface of the unit circle that is enclosed in the splat
//											const E3dCoordinate  LOneOverTwoRadiusSq = 1.0 / (2.0 * LSplatRadiusSq);
											const E3dCoordinate  LOneOverTwoRadiusSq = 1.0 / (2.0 * LIntersectedArcLengthSq);

//printf("%s() %s:%d  %lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LMesh->Vertices_Count);fflush(stdout);
											EIndex  LC = 0;const EIndex  LVN = LMesh->Vertices_Count;const double  LVNF = (double)LVN;for(;LC < LVN;LC++, LUnitSpherePoint += LDimensions, EStruct_Inc(LVertex, LVertexSize))
											{
//printf("%s() %s:%d  %lld of %lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LC, (long long)LMesh->Vertices_Count);fflush(stdout);
												LNearestPoints->QueryPosition[0] = LUnitSpherePoint[0];
												LNearestPoints->QueryPosition[1] = LUnitSpherePoint[1];
												LNearestPoints->QueryPosition[2] = LUnitSpherePoint[2];
												LDistanceSq[0] = LSplatRadiusSq;
												LNearestPoints->Points_Count = 0;
												LNearestPoints->HaveHeap = 0;

												E3d_NearestPointsJensen_iterative(LPointMap, LNearestPoints);const EIndex  LPN = LNearestPoints->Points_Count;if(LPN)
												{
//if(LPN < 10) { printf("%s() %s:%d  %lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LPN);fflush(stdout); }
													EIndex  LPC = 1;float  LNWTotal = 0.0;for(;LPC <= LPN;LPC++)
													{
														const E3dBRDFSamplePoint*  LNearestPoint = LPoints[LPC];

														{
//															const E3dCoordinate  LNW = _GaussianFilterPrep(LDistanceSq[LPC], LOneOverTwoRadiusSq);
															const float  LGreatCircleDistance = E3d_ArcLength((E3dCoord3*)LUnitSpherePoint, &LNearestPoint->Position);const E3dCoordinate  LNW = _GaussianFilterPrep(LGreatCircleDistance * LGreatCircleDistance, LOneOverTwoRadiusSq);

															LNWTotal += LNearestPoint->Intensity * LNW;
														}
													}

//													const float  LWW = LNWTotal * LNormalizer / LSplatRadiusSq;
													const float  LWW = LNWTotal * LNormalizer / LIntersectedArea;
													LVertex->Position.X = LUnitSpherePoint[0] * LWW;
													LVertex->Position.Y = LUnitSpherePoint[1] * LWW;
													LVertex->Position.Z = LUnitSpherePoint[2] * LWW;
													LVertex->Flags = E3dVtxDEFAULT;
												}
												else
												{
													LVertex->Position.X = 0.0;
													LVertex->Position.Y = 0.0;
													LVertex->Position.Z = 0.0;
													LVertex->Flags = E3dVtxDEFAULT;
												}

												if((LC & 0xFF) == 0)
												{
													const ETime  LNow_ms = E_GetMilliseconds();
													if(!LShowProgress)
													{
														const ETime  LElapsed_ms = LNow_ms - LStart_ms;if(LElapsed_ms > 500)
														{
															E3dProgressIndicator_SetMessage("Integrating samples");
															XeApp_Sync(E3dp_XeApp);
															LShowProgress = TRUE;
														}
													}

													if(LShowProgress && (LNow_ms - LLastProgressUpdate_ms) >= 100)
													{
														E3dProgressIndicator_Set((double)LC / LVNF, NULL, NULL);XeApp_Sync(E3dp_XeApp);
														LLastProgressUpdate_ms = LNow_ms;
													}
												}
											}
										}



/*
										{
											const double  LCountF = (double)LCount;
											EIndex  LC = 0;const E3dShaderLight*  LSample = LSamples;E3dVertex*  LVertex = LMesh->Vertices;for(;LC < LCount;LC++, LSample++, EStruct_Inc(LVertex, LVertexSize))
											{
												const E3dCoordinate  LW = LSample->PointToLight.W * LSample->Power[0] / 4.0;

												LNearestPoints->QueryPosition[0] = LSample->PointToLightXYZ.X;
												LNearestPoints->QueryPosition[1] = LSample->PointToLightXYZ.Y;
												LNearestPoints->QueryPosition[2] = LSample->PointToLightXYZ.Z;
												LDistanceSq[0] = LSplatRadiusSq;
												LNearestPoints->Points_Count = 0;
												LNearestPoints->HaveHeap = 0;

												float  LNWTotal = 0.0;

												E3d_NearestPointsJensen_iterative(LPointMap, LNearestPoints);if(LNearestPoints->Points_Count)
												{
//assert(LNearestPoints->Points_Count <= E3dMAX_NEAREST_POINTS);
// See SphereSurfaceSamplingWithGaussianSplat00.odp
//
													const E3dCoordinate  LIntersectedArcX = (-LSplatRadiusSq + 2.0) / 2.0;                                 // X of Pi
													const E3dCoordinate  LIntersectedArcLength = E3d_ASin(LIntersectedArcX);                               // Length of arc on the unit circle that is enclosed in the splat
													const E3dCoordinate  LIntersectedArea = 2.0 * E3dPI * LIntersectedArcLength * LIntersectedArcLength;   // Area on the surface of the unit circle that is enclosed in the splat
													const E3dCoordinate  LInvA = 10000.0 / LIntersectedArea,
																							 LOneOverTwoRadiusSq = 1.0 / (2.0 * LSplatRadiusSq);

													EIndex  LPC = 1;const EIndex  LPN = LNearestPoints->Points_Count;for(;LPC <= LPN;LPC++)
													{
														const E3dBRDFSamplePoint*  LNearestPoint = LPoints[LPC];

														{
															const E3dCoordinate  LNW = _GaussianFilterPrep(LDistanceSq[LPC], LOneOverTwoRadiusSq) * LInvA;
															LNWTotal += LNearestPoint->Intensity * LNW;
														}
													}
												}

												const float  LWW = LW * LNWTotal * LNormalizer;
												LVertex->Position.X = LSample->PointToLight.X * LWW;
												LVertex->Position.Y = LSample->PointToLight.Y * LWW;
												LVertex->Position.Z = LSample->PointToLight.Z * LWW;
												LVertex->Flags = E3dVtxDEFAULT;

												if((LC & 0xFF) == 0)
												{
													const ETime  LNow_ms = E_GetMilliseconds();
													if(!LShowProgress)
													{
														const ETime  LElapsed_ms = LNow_ms - LStart_ms;if(LElapsed_ms > 500) { E3dProgressIndicator_SetMessage("Integrating samples");XeApp_Sync(E3dp_XeApp);LShowProgress = TRUE; }
													}

													if(LShowProgress && (LNow_ms - LLastProgressUpdate_ms) >= 100) { E3dProgressIndicator_Set((double)LC / LCountF, NULL, NULL);XeApp_Sync(E3dp_XeApp);LLastProgressUpdate_ms = LNow_ms; }
												}
											}
										}
*/


										if(LShowProgress)
										{
											E3dProgressIndicator_Set(-1.0, NULL, NULL);
											XeApp_Sync(E3dp_XeApp);
										}
									}
								}
								else
								{
									EIndex  LC = 0;const E3dShaderLight*  LSample = LSamples;E3dVertex*  LVertex = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;for(;LC < LCount;LC++, LSample++, EStruct_Inc(LVertex, LVertexSize))
									{
										const float  LNdotL = E3d_Dot3_vv(LSample->PointToLight, LPoint.Normal);

//								const E3dCoordinate  LW = LSample->PointToLight.W * LSample->Power[0] / 4.0;
										const E3dCoordinate  LW = LNdotL * LSample->Power[0];
//								const E3dCoordinate  LW = 1.0;
										LVertex->Position.X = LSample->PointToLight.X * LW;
										LVertex->Position.Y = LSample->PointToLight.Y * LW;
										LVertex->Position.Z = LSample->PointToLight.Z * LW;
										LVertex->Flags = E3dVtxDEFAULT;
									}
								}
							}

							EFree(LSamples);
						}
					}

				}
			}
		}


		E3dNode_UpdateForDisplay(PNode, E3dUF_ALL);
	}



	return(LTriggeredChanges);
}


static void _VisGeoInit(E3dVisGeo* PVisGeo)
{
	PVisGeo->Mesh           = NULL;
	PVisGeo->GrabUnitSphere = FALSE;
	PVisGeo->UnitSpherePoints = NULL;PVisGeo->NumOfUnitSpherePoints = 0;
	PVisGeo->RotatedUnitSpherePoints = NULL;
}


static void _VisGeoFree(E3dVisGeo* PVisGeo)
{
	E_DeleteChk(PVisGeo->Mesh);
	EFreeChkClr(PVisGeo->UnitSpherePoints);PVisGeo->NumOfUnitSpherePoints = 0;
	EFreeChkClr(PVisGeo->RotatedUnitSpherePoints);
	PVisGeo->GrabUnitSphere = TRUE;
}

#define E3dPOINT_ALLOC_INCREMENT  16

//========================================
// Constructor
//========================================
static int _Init(E3dNode* PNode, E3dNodeComponent* PInfo)
{
	E3dInfo*  LThis = (E3dInfo*)PInfo;

	LThis->BlackRadius = 0.0;
	LThis->Scale       = 1.0;


	LThis->SamplingDirection = E3dSAMPLE_LIGHTDIRS;
//	LThis->IntegrationMethod = 0;
	LThis->IntegrationMethod = E3dGAUSSIAN_SPLATS;
	LThis->SamplesSqrt       = 256;

	LThis->SplatRadius = 0.05;

	LThis->Material = NULL;

	_VisGeoInit(&LThis->ShadeSurface);
	_VisGeoInit(&LThis->SampleDirection);
	_VisGeoInit(&LThis->SampleDirections);

	LThis->OrientSphere = TRUE;

	LThis->Dimensions = 3;

	{
		E3dKDTreePointMap*  LPointMap = &LThis->PointMap;
		E3d_KDTreePointMapInit(LPointMap);
		LPointMap->BytesPerPoint = sizeof(E3dBRDFSamplePoint);

		LThis->NearestPoints = NULL;

		LPointMap->TmpPoints = NULL;
		LPointMap->Points_Count = 0;
		LPointMap->Points_CountAllocated = 0;
	}
	LThis->NearestPoints = NULL;

	E3dSamplingCache_Init(&LThis->SamplingCache);

	return(E_SUCCESS);
}


//========================================
// Destructor
//========================================
static int _Free(E3dNode* PNode, E3dNodeComponent* PInfo)
{
	E3dInfo*  LThis = (E3dInfo*)PInfo;

	E_DeleteChk(LThis->Material);

	_VisGeoFree(&LThis->ShadeSurface);
	_VisGeoFree(&LThis->SampleDirection);
	_VisGeoFree(&LThis->SampleDirections);

	E3dSamplingCache_FreeContents(&LThis->SamplingCache);

	E3d_KDTreePointMapFreeContents(&LThis->PointMap);
	if(LThis->NearestPoints) { E3d_NearestKDTreePointsFree(LThis->NearestPoints);LThis->NearestPoints = NULL; }

	return(E_SUCCESS);
}


//========================================
// Copy-constructor
//========================================
static int _Copy(const E3dNode* PSrcNode, const E3dNodeComponent* PSrcComponent,  E3dNode* PDstNode, E3dNodeComponent* PDstComponent)
{
	E3dInfo*  LSInfo = (E3dInfo*)PSrcComponent;
	E3dInfo*  LDInfo = (E3dInfo*)PDstComponent;

	_Init(PDstNode, PDstComponent);

#define _CopyField(mField) LDInfo->mField = LSInfo->mField
	_CopyField(BlackRadius);
	_CopyField(Scale);
	_CopyField(OrientSphere);

	_CopyField(SamplingDirection);
	_CopyField(IntegrationMethod);
	_CopyField(SamplesSqrt);
	_CopyField(SplatRadius);
#undef _CopyField
	if(LSInfo->Material) E_AssignToMember(LSInfo->Material, LDInfo, Material);

	return(E_SUCCESS);
}


static void _DCB_SampleLightDirs(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*  LPanel = (E3dNodePanel*)PClientData;
	E3dNode*       LNode = LPanel->Node;
	E3dInfo*       LInfo = (E3dInfo*)E3dNodeComponent_ByClass(LNode, _NodeComponentClass);

	if(((XeToggleButtonCallbackStruct*)PCallData)->Set)
	{
		LInfo->SamplingDirection = E3dSAMPLE_LIGHTDIRS;

		E3dSamplingCache_FreeContents(&LInfo->SamplingCache);

		_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);
		E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
	}
}


static void _DCB_SampleViewDirs(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*  LPanel = (E3dNodePanel*)PClientData;
	E3dNode*       LNode = LPanel->Node;
	E3dInfo*       LInfo = (E3dInfo*)E3dNodeComponent_ByClass(LNode, _NodeComponentClass);

	if(((XeToggleButtonCallbackStruct*)PCallData)->Set)
	{
		LInfo->SamplingDirection = E3dSAMPLE_VIEWDIRS;

		E3dSamplingCache_FreeContents(&LInfo->SamplingCache);

		_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);
		E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
	}
}


static void _DCB_BlackRadius(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*  LPanel = (E3dNodePanel*)PClientData;
	E3dNode*       LNode = LPanel->Node;
	E3dInfo*       LInfo = (E3dInfo*)E3dNodeComponent_ByClass(LNode, _NodeComponentClass);

	LInfo->BlackRadius = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);
	E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
}


static void _DCB_Scale(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*  LPanel = (E3dNodePanel*)PClientData;
	E3dNode*       LNode = LPanel->Node;
	E3dInfo*       LInfo = (E3dInfo*)E3dNodeComponent_ByClass(LNode, _NodeComponentClass);

	LInfo->Scale = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);
	E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
}


static void _DCB_SamplesSqrt(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*  LPanel = (E3dNodePanel*)PClientData;
	E3dNode*       LNode = LPanel->Node;
	E3dInfo*       LInfo = (E3dInfo*)E3dNodeComponent_ByClass(LNode, _NodeComponentClass);

	LInfo->SamplesSqrt = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	E3dKDTreePointMap*  LPointMap = &LInfo->PointMap;LPointMap->Points_Count = 0;

	E3dSamplingCache_FreeContents(&LInfo->SamplingCache);

	_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);
	E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
}


static void _DCB_SplatRadius(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*  LPanel = (E3dNodePanel*)PClientData;
	E3dNode*       LNode = LPanel->Node;
	E3dInfo*       LInfo = (E3dInfo*)E3dNodeComponent_ByClass(LNode, _NodeComponentClass);

	LInfo->SplatRadius = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);
	E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_OrientSphere(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*  LPanel = (E3dNodePanel*)PClientData;
	E3dNode*       LNode = LPanel->Node;
	E3dInfo*       LInfo = (E3dInfo*)E3dNodeComponent_ByClass(LNode, _NodeComponentClass);

	LInfo->OrientSphere = (EBool)(((XeToggleButtonCallbackStruct*)PCallData)->Set);

	_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dSceneCHG_ALL, FALSE);
	E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
}


static void _GUI(E3dNode* PNode, E3dNodeComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dNodePanel*            LPanel = (E3dNodePanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dInfo*                 LInfo = (E3dInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LInfo)
			{
				if(LDrawer->ResourcesGUI)
				{
					E3dResourcesGUI*  LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

					XeScale_SetValue_Coord(LResourcesGUI->BlackRadiusW,      LInfo->BlackRadius,  FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->ScaleW,            LInfo->Scale,        FALSE);
					XeScale_SetValue(LResourcesGUI->SamplesSqrtW,            LInfo->SamplesSqrt,  FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->SplatRadiusW,      LInfo->SplatRadius,  FALSE);
					XeToggleButton_SetState(LResourcesGUI->OrientSphereW,    LInfo->OrientSphere, FALSE);
					XeToggleButton_SetState(LResourcesGUI->SampleViewDirsW,  LInfo->SamplingDirection == E3dSAMPLE_VIEWDIRS,  FALSE);
					XeToggleButton_SetState(LResourcesGUI->SampleLightDirsW, LInfo->SamplingDirection == E3dSAMPLE_LIGHTDIRS, FALSE);
				}
				else
				{
					Widget            LTopMatrixW = LDrawer->Drawer.DrawerW, LW;
					Widget            LIndentWs[7];int	LIndentN;      // For indenting the Scale Widgets
					Widget            LTraverseWs[16];int	LTraverseN;  // For TextField traversal
					EArg              Args[256];int  ArgCnt;
					E3dResourcesGUI*  LResourcesGUI;ENew(LResourcesGUI);

					LDrawer->ResourcesGUI = LResourcesGUI;

					LTraverseN = 0;
					LIndentN = 0;

					{
						EXeStart;
						EXeSetArg(XeNmarginTop, 5);
						EXeSetArg(XeNradioBehavior, TRUE);
						EXeSetArg(XeNorientation, XeHORIZONTAL);
						EXeSetArg(XeNspacingX, 6);
						EXeSetArg(XeNmarginSizeY, 0);
						EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
						EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
						EXeSetArg(XeNheightAdjustable, FALSE);
						Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

						EXeStart;
						EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
						XeWidget_ManagedNew("Sample", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->SamplingDirection == E3dSAMPLE_VIEWDIRS);
						LResourcesGUI->SampleViewDirsW = LW = XeWidget_ManagedNew("View dirs", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SampleViewDirs, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->SamplingDirection == E3dSAMPLE_LIGHTDIRS);
						LResourcesGUI->SampleLightDirsW = LW = XeWidget_ManagedNew("Light dirs", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SampleLightDirs, (EPointer)LPanel);
					}


					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 2);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->BlackRadiusW = LW = XeScaleNew_Coord(LTopMatrixW, "Black radius",  Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->BlackRadius,  0.0, 10.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_BlackRadius, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 2);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->ScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Scale",  Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->Scale,  0.0, 100.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Scale, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 5);
					EXeSetArg(XeNminimum, 16);EXeSetArg(XeNmaximum, 4096);EXeSetArg(XeNvalue, LInfo->SamplesSqrt);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SamplesSqrtW = LW = XeScale_Newi32(LTopMatrixW, "Samples sqrt",  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SamplesSqrt, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 2);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SplatRadiusW = LW = XeScale_Newf32(LTopMatrixW, "Splat radius",  Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->SplatRadius,  0.005, 0.5,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SplatRadius, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNmarginTop, 0);EXeSetArg(XeNmarginBottom, 4);
					EXeSetArg(XeNset, LInfo->OrientSphere);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LResourcesGUI->OrientSphereW = LW = XeWidget_ManagedNew("Orient", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_OrientSphere, (EPointer)LPanel);

					Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);

// Set up TextField traversal
//
					Xe_TraversalRing(LTraverseWs, LTraverseN);
				}
			}
		break;

		case E_FREE:
		break;
	}
}


//========================================
// Scene changed callback
//========================================
static void _CB_Scene(ECoreObject* PScn, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = (E3dScene*)PScn;

	switch(((E3dCallbackStruct*)PCallData)->Reason)
	{
		case E3dCR_NODE_TRANSLATE:
			{
				E3dSceneNodesCallbackStruct*  LCBS = (E3dSceneNodesCallbackStruct*)PCallData;
				E3dModel*  LFirstVisibleLightModel = _SceneGetFirstVisibleLightModel(LScene);if(LFirstVisibleLightModel == NULL) return;

				EIndex  LC = 0;const EIndex  LNumOfNodes = LCBS->NumOfNodes;E3dNode**  LNodes = LCBS->Nodes;for(;LC < LNumOfNodes;LC++)
				{
					if(LNodes[LC] == (E3dNode*)LFirstVisibleLightModel)
					{
						EIndex  LRC = 0;const EIndex  LRN = LScene->RootNodes_Count;for(;LRC < LRN;LRC++)
						{
							E3dNode*  LNode = LScene->RootNodes[LRC];for(;LNode;LNode = LNode->Next)
							{
								E3dInfo*  LInfo = (E3dInfo*)E3dNodeComponent_ByClass(LNode, _NodeComponentClass);if(LInfo)
								{
									_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES|E3dChgLIGHT, FALSE);
								}
							}
						}
						break;
					}
				}
			}
		break;

		case E3dCR_MATERIAL_CHANGED:
			{
				E3dSceneMaterialsCallbackStruct*  LCBS = (E3dSceneMaterialsCallbackStruct*)PCallData;
				EIndex  LRC = 0;const EIndex  LRN = LScene->RootNodes_Count;for(;LRC < LRN;LRC++)
				{
					E3dNode*  LNode = LScene->RootNodes[LRC];for(;LNode;LNode = LNode->Next)
					{
						E3dInfo*  LInfo = (E3dInfo*)E3dNodeComponent_ByClass(LNode, _NodeComponentClass);if(LInfo)
						{
							E3dMaterial*  LMaterial = LInfo->Material;if(LMaterial)
							{
								EIndex  LC = 0;const EIndex  LN = LCBS->NumOfMaterials;for(;LC < LN;LC++)
								{
									if(LCBS->Materials[LC] == LMaterial) _Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES|E3dChgMATERIALTYPE, FALSE);
								}
							}
						}
					}
				}
			}
		break;
	}
}


static E3dMesh* _VisGeoNew(E3dVisGeo* PVisGeo, const char* PName, E3dObjectComponent* PSphereInfo, E3dMaterial* PMaterial, E3dModel* PModel)
{
	E3dMesh*  LMesh = (E3dMesh*)E3dObject_CustomNewInit(NULL, PSphereInfo, E3dUF_ALL, FALSE);EObject_SetName(LMesh, EStrDup(PName), FALSE);
	E3dModel_AppendObject(PModel, (E3dObject*)LMesh);
	E_AssignToMember(PMaterial, LMesh->PolyGroups[0],Material);
	return(LMesh);
}


//================================================
// ToolBar icon callback handler
//================================================
static void _CB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	E3dItem       LItem;
	EChangeFlags  LChanged = 0;

	if(E3dScene_GetSingleSelection(E3d_Scene, &LItem, E3dItemMESH|E3dItemFACE|E3dItemPOLYGROUP, 0) != E3dItemNONE)
	{
		E3dMaterial*  LMaterial = NULL;
		E3dObjItem*   LObjItem = &LItem.ObjItem;

		switch(LItem.Item.Type)
		{
			case E3dItemMESH:
				{
//printf("%s() %s:%d Mesh\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
					E3dMesh*  LMesh = (E3dMesh*)LObjItem->Object;
					if(LMesh->PolyGroups_Count > 0)
					{
						LMaterial = LMesh->PolyGroups[0]->Material;
					}
				}
			break;

			case E3dItemFACE:
				{
//printf("%s() %s:%d Face\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
					E3dFace*  LFace = (E3dFace*)LObjItem->Object;
					LMaterial = LFace->PolyGroup.Material;
				}
			break;

			case E3dItemPOLYGROUP:
				{
//printf("%s() %s:%d PolyGroup\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
					E3dPolyItem*  LPolyItem = &LItem.PolyItem;
					LMaterial = LPolyItem->PolyGroup->Material;
				}
			break;

			case E3dItemMODEL:
				{
//printf("%s() %s:%d Model\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
					E3dModel*  LModel = (E3dModel*)LObjItem->Node;
					if(LModel->Objects_Count > 0)
					{
						E3dObject*  LObject = LModel->Objects[0];
						E3dFace*    LFace = E_DynamicCast(LObject, E3dFace);if(LFace)
						{
							LMaterial = LFace->PolyGroup.Material;
							break;
						}
						E3dMesh*  LMesh = E_DynamicCast(LObject, E3dMesh);if(LMesh)
						{
							if(LMesh->PolyGroups_Count > 0)
							{
								LMaterial = LMesh->PolyGroups[0]->Material;
							}
							break;
						}
					}
				}
			break;

			default:
				{
					char  s[256];printf("%s() %s:%d  |%s|\n", __FUNCTION__, __FILE__, __LINE__, E3d_ItemFlagsToString(LItem.Item.Type, s));fflush(stdout);
				}
			break;
		}

		if(LMaterial == NULL) goto NoMaterial;

		E3dModel*     LModel = NULL;
		E3dMaterial*  LSMaterial = NULL;

		E3dObjectComponentClass*  LSphereObjectComponentClass = E3dObjectComponent_ClassFindByName("Sphere");if(LSphereObjectComponentClass)
		{
			E_New(E3dModel, LModel);EObject_SetName(LModel, EStrDup("BRDFVis"), FALSE);

			E3dMaterial_New(E3dLambertShader_class, LSMaterial);if(LSMaterial)
			{
				E3dLambertShader*  LShader = (E3dLambertShader*)(LSMaterial->Shader);

				EObject_SetName(LSMaterial, EStrDup("BRDF"), FALSE);
				EcColor_InitRGB(&LShader->Ambient,  0.1, 0.1, 0.1);
				EcColor_InitRGB(&LShader->Diffuse,  0.8, 0.8, 0.8);
				E3dMaterial_UpdateForDisplay(LSMaterial, NULL, E3dChgMATERIALTYPE, E3dTXTUF_ALL);

			}

			if(LModel)
			{
				E3dInfo*  LInfo = (E3dInfo*)E3dNodeComponent_Add((E3dNode*)LModel, _NodeComponentClass);

				E_AssignToMember(LMaterial, LInfo,Material);

				{
					E3dObjectComponent*  LSphereInfo = E3dObject_ComponentNew(LSphereObjectComponentClass);if(LSphereInfo)
					{
//						ERes_NamedResourceSet1i(LSphereInfo, "Type",        2);ERes_NamedResourceSet1i(LSphereInfo, "Iterations",  4);    // Geodesic sphere. It has a much more uniform distribution of Vertices
						ERes_NamedResourceSet1f(LSphereInfo, "Radius",    1.0);
						ERes_NamedResourceSet1i(LSphereInfo, "USections", 120);
						ERes_NamedResourceSet1i(LSphereInfo, "VSections",  90);
//						ERes_NamedResourceSet1f(LSphereInfo, "SouthMostLatitude", 0.0);

						{ E3dMesh*  LMesh = _VisGeoNew(&LInfo->ShadeSurface,     "ShadeSurface()",     LSphereInfo, LSMaterial, LModel);E_AssignToMember(LMesh, LInfo,ShadeSurface.Mesh); }
						{ E3dMesh*  LMesh = _VisGeoNew(&LInfo->SampleDirection,  "SampleDirection()",  LSphereInfo, LSMaterial, LModel);E_AssignToMember(LMesh, LInfo,SampleDirection.Mesh); LMesh->Visible = FALSE; }
						{ E3dMesh*  LMesh = _VisGeoNew(&LInfo->SampleDirections, "SampleDirections()", LSphereInfo, LSMaterial, LModel);E_AssignToMember(LMesh, LInfo,SampleDirections.Mesh); }

						E_Delete(LSphereInfo);  // Free this, as we were just using it as a template for E3dObject_CustomNewInit
					}
				}

				E3dScene_NodeHrcAdd(E3d_Scene, (E3dNode*)LModel, E3dUF_CALL_SCENE_CALLBACKS);
				E3dScene_ObjectsUpdateModelList(LScene);
				E3dScene_LayoutOnSchematics(LScene, TRUE);
				LChanged |= E3dModel_GetChangeFlagsByObjects(LModel)|E3dChgHIERARCHY;
				_Update(NULL, (E3dNode*)LModel, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);
			}
		}

		if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) E3dNode_Edit((E3dNode*)LModel, _NodeComponentClass);

		if(LChanged) E3dWindows_RedrawByChangeFlags(LChanged);
	}
	else
	{
NoMaterial:
		E3dp_PrintMessage(ELogMESSAGE, 10000, "Select something with a Material");
	}
}


static const char  _Tooltip[] = "Visualizes the BRDF on the selected object.\n\
You can Compare a Shader's ShadeSurface, SampleDirection and SampleDirection functions.\n\
It's great for slide presentations, or for debugging your shader.\n\
The light direction will be specified by the Scene's first point light's direction from the origin. Try moving the light around!";

//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	{
		E3dNodeComponentClass  LComponentClass=
		{
			.Name=        "BRDF visualizer",
			.StructSize=  sizeof(E3dInfo),
			.Resources=   _Resources,
			.Init=        _Init,
			.Free=        _Free,
			.Copy=        _Copy,
			.Update=      _Update,
			.GUI=         _GUI
		};
		_NodeComponentClass = E3dNodeComponent_ClassRegister(&LComponentClass);
	}

	if(_NodeComponentClass)
	{
		E3dScene_CallbackAdd(E3d_Scene, E3dChgTRANSFORM|E3dChgMATERIALTYPE|E3dChgMATERIALPARAM, _CB_Scene, (EPointer)0);
// Use default plugin icon
//
		Xe_PushButtonAddImg3(_MenuButtons[_NumOfMenuButtons++],       Plugin, "Menu->Material", "BRDFVis", '\0', NULL, NULL, FALSE, _Tooltip, _CB, (EPointer)0);
		EpM_CreateIconStates3_sp(BRDFVis);Xe_PushButtonAddImg3(_ToolbarButtons[_NumOfToolbarButtons++], BRDFVis, "Tool->Light",    "BRDFVis", '\0', NULL, NULL, FALSE, _Tooltip, _CB, (EPointer)0);
	}

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		{ int  LC = 0;for(;LC < _NumOfMenuButtons;LC++)    { XeItemFree(_MenuButtons[LC]); }    _NumOfMenuButtons = 0; }
		{ int  LC = 0;for(;LC < _NumOfToolbarButtons;LC++) { XeItemFree(_ToolbarButtons[LC]); } _NumOfToolbarButtons = 0; }
	}

	E3dScene_CallbackRemove(E3d_Scene, _CB_Scene, (EPointer)0);

	if(_NodeComponentClass)
	{
		E3dNodeComponent_ClassDeactivate(_NodeComponentClass);_NodeComponentClass = NULL;
	}

	EpIconFree3(BRDFVis);

	return(E_SUCCESS);
}
