//================================================================================================================================
// Plugin for EQUINOX-3D
//
// This sample code shows how to:
//  - Create a simple dialog box with a button. When the button is pressed:
//  - Look up runtime Object classes by name, in this case, the primitive shapes, "Box", "Sphere" and "Cylinder".
//  - Create Objects of those classes
//  - Build a simple Model hierarchy, using those primitive shapes
//  - Set transform values
//  - Add the hierarchy to the Scene, and redraw the 3DWindows
//
// A note on transforms:
// libE3D / EQUINOX-3D have the most powerful transform system that we are aware of:
// in addition to traditional scale, rotate and translate, it also allows for an arbitrary list of transforms on each Model.
// For example, a single Model can have 4 rotations, 2 scale operations, 5 translates, then 2 more scales, etc.
// Transforms can have a name and a semantic string.
// So, unlike any other 3D software, libE3D / EQUINOX-3D can accurately import / export transforms to / from any other system.
// For example, Maya and 3DS Max have very different transform system, and there's no way to accurately exchange hierarchies
// between those.
// However, libE3D / EQUINOX-3D can import and export accurate transform setups with either of those applications.
// This example shows how to use the traditional transform system.
//
// To avoid cluttering the UI, example plugins are not loaded by default, when EQUINOX-3D is started.
// You can manually load them from the Plugin manager, or you can edit your Plugins/Plugins.list file,
// and comment out (e.g. with a trailing "#") or remove the "Load = FALSE;" line for a plugin.
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-01 19:00:03
//
// Copyright (C) 2014-2021 by Gabor Nagy. All rights reserved.
//================================================================================================================================
#include <assert.h>
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/Scene.h>

#include <E3D/UI/Panel.h>

#include <Xe/PushButton.h>
#include <Xe/Shell.h>
#include <Xe/Dialog.h>


char*                E_PluginInfo = "Sample code: build hierarchy, with runtime-defined primitive shape classes.";

static EPlugin*      _Plugin = NULL;

static Widget        _DialogW = NULL;
static EPluginLock*  _DialogLock = NULL;

static EguiItem      _ToolBarButtons[16];static int  _ToolBarButtons_Count = 0;



enum
{
	CR_TEST = XeCR_CUSTOM0,
};


//========================================
// Dialog box callback
//========================================
static void _DCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;

// Create customized primitives (Box, Cylinder, Torus).
//
	{
// Look up the classes. Bail, if one or more of the classes don't exist (e.g. the appropriate plugin is not loaded)
// We only need to do this once, so these could even be global variables in your plugin.
//
		E3dObjectComponentClass*  LBoxClass      = E3dObjectComponent_ClassFindByName("Box");if(LBoxClass == NULL) return;
		E3dObjectComponentClass*  LCylinderClass = E3dObjectComponent_ClassFindByName("Cylinder");if(LCylinderClass == NULL) return;
		E3dObjectComponentClass*  LSphereClass   = E3dObjectComponent_ClassFindByName("Sphere");if(LSphereClass == NULL) return;
		E3dObjectComponentClass*  LTorusClass    = E3dObjectComponent_ClassFindByName("Torus");if(LTorusClass == NULL) return;

		{
// Create ObjectComponent templates. These are used for setting user-adjustable parameters before creating an Object.
// These can be reused.
//
			E3dObjectComponent*  LBoxInfo      = E3dObject_ComponentNew(LBoxClass);
			E3dObjectComponent*  LCylinderInfo = E3dObject_ComponentNew(LCylinderClass);
			E3dObjectComponent*  LSphereInfo   = E3dObject_ComponentNew(LSphereClass);
			E3dObjectComponent*  LTorusInfo    = E3dObject_ComponentNew(LTorusClass);


			{
// Set Resources by name.
// We use the Resource system, as we don't know the memory layout of runtime-typed objects, outside the plugin that implements them.
//
// Alternatively, if you set a Resource frequently (thousands of times), you could do this:
//
// Look up Resource by name once and get handle:
//  EResource*     LRadiusResource = ERes_ObjectResourceByName(LInfo, "Radius");
//
// Set value using the handle:
//  ERes_ResourceSet1f(LInfo, LRadiusResource, 2.0);
// This avoids the name-look-ups, so it's faster.
//


// Create an empty Model for the root of our hierarchy. This is how you create any object in the "E_" object-oriented system, from EpImage, to E3dModel.
//
				E3dModel*  LRootModel;E_New_n(E3dModel, LRootModel,  "Bicycle");
				LRootModel->TransformSet.EulerRotation.Angles.Y = -50.0;


// Create a torus for the wheels. We'll instantiate the same torus Mesh for both wheels.
				ERes_NamedResourceSet1f(LTorusInfo, "MainRadius", 0.5);
				ERes_NamedResourceSet1f(LTorusInfo, "CircleRadius", 0.04);
				ERes_NamedResourceSet1i(LTorusInfo, "MainSections", 30);     // Set the resolution along the main circle
				ERes_NamedResourceSet1i(LTorusInfo, "CircleSections", 16);   // Set the resolution along the small circle
				E3dObject*  LWheel = E3dObject_CustomNewInit(LScene, LTorusInfo, E3dUF_ALL, TRUE);

// Create a Model for the rear wheel, add our wheel Mesh to it, then add this Model to the root Model.
// This shows that, instead of using E_New_n(), you can set the object name separately, with EObject_SetName().
// This is how you change the name of any existing Object.
//
				E3dModel*  LModel;E_New(E3dModel, LModel);EObject_SetName(LModel, EStrDup("RearWheel"), FALSE);
				LModel->TransformSet.Translation.Y =  0.32;
				LModel->TransformSet.Translation.Z = -0.8;
				LModel->TransformSet.EulerRotation.Angles.Z = 90.0;
				E3dModel_AppendObject(LModel, LWheel);
				E3dNode_AddChild((E3dNode*)LRootModel, (E3dNode*)LModel);



				{
// Create a cylinder for parts of the bycicle frame. We'll just reuse this, by scaling it.
					ERes_NamedResourceSet1f(LCylinderInfo, "Radius", 0.03);
					ERes_NamedResourceSet1f(LCylinderInfo, "Height", 1.0);
					ERes_NamedResourceSet1i(LCylinderInfo, "OrigAtBottom", TRUE);
					E3dObject*  LFramePart = E3dObject_CustomNewInit(LScene, LCylinderInfo, E3dUF_ALL, TRUE);

// Create Models for the frame
					E_New_n(E3dModel, LModel,  "Top tube");E3dModel_AppendObject(LModel, LFramePart);
					E_SetXYZ_s(LModel->TransformSet.Scaling,      1.0, 1.06, 1.0);
					LModel->TransformSet.EulerRotation.Angles.X = 90.0;
					E_SetXYZ_s(LModel->TransformSet.Translation,  0.0, 1.0, -0.4);  // This is a macro for setting the X,Y and Z coordinates of a vector
					E3dNode_AddChild((E3dNode*)LRootModel, (E3dNode*)LModel);

					E_New_n(E3dModel, LModel,  "Seat tube");E3dModel_AppendObject(LModel, LFramePart);
					LModel->TransformSet.EulerRotation.Angles.X = -30.0;
					E_SetXYZ_s(LModel->TransformSet.Translation,  0.0, 0.3, 0.0);
					E3dNode_AddChild((E3dNode*)LRootModel, (E3dNode*)LModel);

					E_New_n(E3dModel, LModel,  "Down tube");E3dModel_AppendObject(LModel, LFramePart);
					E_SetXYZ_s(LModel->TransformSet.Scaling,      1.0, 0.86, 1.0);
					E_SetXYZ_s(LModel->TransformSet.Translation,  0.0, 0.32, 0.0);
					LModel->TransformSet.EulerRotation.Angles.X = 38.0;
					E3dNode_AddChild((E3dNode*)LRootModel, (E3dNode*)LModel);

					E_New_n(E3dModel, LModel,  "Seatstays");E3dModel_AppendObject(LModel, LFramePart);
					E_SetXYZ_s(LModel->TransformSet.Scaling,      1.0, 0.8, 1.0);
					LModel->TransformSet.EulerRotation.Angles.X = 32.0;
					E_SetXYZ_s(LModel->TransformSet.Translation,  0.0, 0.32, -0.82);
					E3dNode_AddChild((E3dNode*)LRootModel, (E3dNode*)LModel);

					E_New_n(E3dModel, LModel,  "Chainstays");E3dModel_AppendObject(LModel, LFramePart);
					E_SetXYZ_s(LModel->TransformSet.Scaling,      1.0, 0.86, 1.2);
					LModel->TransformSet.EulerRotation.Angles.X = -90.0;
					E_SetXYZ_s(LModel->TransformSet.Translation,  0.0, 0.32, 0.02);
					E3dNode_AddChild((E3dNode*)LRootModel, (E3dNode*)LModel);

// Steerable parts
					{
						E3dModel*  LSteeringModel;E_New_n(E3dModel, LSteeringModel,  "Steering");E3dModel_AppendObject(LSteeringModel, LFramePart);
						E_SetXYZ_s(LSteeringModel->TransformSet.EulerRotation.Angles,  -12.0, 20.0, 0.0);    // Turn the front wheel
						LSteeringModel->TransformSet.EulerRotation.Order = E3dXYZ;                           // We need to overwrite the default euler rotation order (ZYX), to XYZ, so the steering works correctly
						E_SetXYZ_s(LSteeringModel->TransformSet.Translation,  0.0, 0.32, 0.8);
						E3dNode_AddChild((E3dNode*)LRootModel, (E3dNode*)LSteeringModel);

// Another cylinder for the handle bar.
						ERes_NamedResourceSet1f(LCylinderInfo, "Radius", 0.03);
						ERes_NamedResourceSet1f(LCylinderInfo, "Height", 0.9);
						ERes_NamedResourceSet1i(LCylinderInfo, "OrigAtBottom", FALSE);  // We want this cylinder to be centered along its height
						LModel = E3d_ModelCustomObjectNewInit(LScene, LCylinderInfo, NULL, E3dUF_ALL, "HandleBar", FALSE);
						E_SetXYZ_s(LModel->TransformSet.Translation,  0.0, 1.0, 0.04);
						E_SetXYZ_s(LModel->TransformSet.EulerRotation.Angles,  90.0, 90.0, 0.0);
						E3dNode_AddChild((E3dNode*)LSteeringModel, (E3dNode*)LModel);

// Create a Model for the front wheel
						E_New_n(E3dModel, LModel, "FrontWheel");
						LModel->TransformSet.EulerRotation.Angles.Z = 90.0;
						E3dModel_AppendObject(LModel, LWheel);
						E3dNode_AddChild((E3dNode*)LSteeringModel, (E3dNode*)LModel);
					}
				}


// Create a box for the seat
				ERes_NamedResourceSet1f(LBoxInfo, "EdgeLengthX", 0.2);
				ERes_NamedResourceSet1f(LBoxInfo, "EdgeLengthY", 0.1);
				ERes_NamedResourceSet1f(LBoxInfo, "EdgeLengthZ", 0.36);
				LModel = E3d_ModelCustomObjectNewInit(LScene, LBoxInfo, NULL, E3dUF_ALL, "Seat", FALSE);  // This function creates a custom Geometry, and a Model for it
				E_SetXYZ_s(LModel->TransformSet.Translation,  0.0, 1.2, -0.48);
				E3dNode_AddChild((E3dNode*)LRootModel, (E3dNode*)LModel);


// Free our Component templates
				E3dObjectComponent_Free(LBoxInfo);
				E3dObjectComponent_Free(LCylinderInfo);
				E3dObjectComponent_Free(LSphereInfo);
				E3dObjectComponent_Free(LTorusInfo);

				E3dNodeHrc_UpdateConnections((E3dNode*)LRootModel);     // Update all internal connections in our hierarchy, at once
				E3dNodeHrc_UpdateMatrices((E3dNode*)LRootModel, TRUE);  // Update the Matrices of our new hierarchy, to make the hierarcy ready for rendering

// Add our hierarcy to the main Scene
				E3dScene_NodeHrcAdd(LScene, (E3dNode*)LRootModel, E3dUF_CALL_SCENE_CALLBACKS);

// These 2 calls are needed by EQUINOX-3D, so our new hierarchy will work correctly
				E3dScene_ObjectsUpdateModelList(LScene);
				E3dScene_LayoutOnSchematics(LScene, TRUE);


// This is needed, so the intelligent update system in libE3D / EQUINOX-3D knows what changed. We will be OR-ing flags to this, to indicate what we are changing in the Scene.
// In this case, we will be changing Meshes, but if we were just changing Splines, we wouldn't want all the ray-traced views to start redrawing.
				EChangeFlags  LChanged = E3dNodeHrc_GetChangeFlagsByObjects((E3dNode*)LRootModel) | E3dChgHIERARCHY;

				E3dWindows_RedrawByChangeFlags(LChanged);
			}
		}
	}
}


//========================================
// Dialog box callback
//========================================
static void _DCB_Close(Widget PW, EPointer PClientData, EPointer PCallData)
{
// Un-display dialog box
//
	Xe_WidgetUnmap(_DialogW);

// Delete plugin lock
//
	E_PluginLockDel(_Plugin, &_DialogLock, FALSE);
}


static int _DialogLockDel(EPlugin* PPlugin, EPluginLock* LLock)
{
	Xe_WidgetUnmap(_DialogW);

	return(E_SUCCESS);
}


//================================================
// ToolBar icon callback handler
//================================================
static void _CB(Widget PW, EPointer PClientData, EPointer PCallData)
{
// We'll be displaying a dialog box, so we'll lock the plugin to prevent unloading
// it (and pulling the rag out from beneath our global variables).
// When the application exits, or the user forces the plugin to unload, _DialogLockDel will get called before E_PluginExit.
// This two-pass exit (unlock & unload) of plugins is necessary to avoid certain inter-dependencies from breaking.
//
	if(_DialogLock == NULL) _DialogLock = E_PluginLockAdd(_Plugin, _DialogLockDel);

	if(_DialogW == NULL)
	{
		Widget   LW;
		EArg     Args[32];int  ArgCnt;
		EXeStart;
		EXeSetArg(XeNallowShellResize, TRUE);
		EXeSetArg(XeNbuttonsStretch, FALSE);
		EXeSetArg(XeNworkAreaType, XeMATRIX);
		EXeSetArg(XeNwmDecorations, XeWMDecorCLOSE|XeWMDecorBORDER|XeWMDecorRESIZEH|XeWMDecorTITLE);
		EXeSetArg(XeNtitle, "Hierarchy");EXeSetArg(XeNiconName, "Hierarchy");
		Widget  LDialogW = XeDialog_New(XeDIALOG_TEMPLATE, "Hierarchy", E3dp_TopShellW,  Args, ArgCnt);if(LDialogW == NULL) return;
		XeAddCallback(LDialogW, XeNwmCallback, _DCB_Close, (EPointer)NULL);

		_DialogW = LDialogW;

		Widget  LTopMatrixW = XeDialog_GetChild(LDialogW, XeDIALOG_WORK_AREA);

		EXeStart;
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNmarginSizeX, 8);EXeSetArg(XeNmarginSizeY, 8);
		EXeSetArg(XeNwidthAdjustable, TRUE);
		Xe_SetValues(LTopMatrixW, Args, ArgCnt);

		EXeStart;
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignCENTER);
		EXeSetArg(XeNheightAdjustable, FALSE);
		LW = XeWidget_ManagedNew("Create primitive hierarchy", XePushButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNactivateCallback, _DCB, (EPointer)NULL);


		XeShell_Raise(_DialogW);
		XeShell_CenterToWidget(_DialogW, E3dp_TopShellW);
	}
	else XeShell_Raise(_DialogW);
}


static const char*  _ToolTip = "This plugin demonstrates how to build a hierarchy, with primitive shapes, and instancing.";

//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

// Use default plugin icon
//
	Xe_PushButtonAddImg3(_ToolBarButtons[_ToolBarButtons_Count++], Plugin, "Tool->Create", "TestRuntimeClass", '\0', NULL, NULL, FALSE, _ToolTip, _CB, (EPointer)0);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC = 0;for(;LC < _ToolBarButtons_Count;LC++) { XeItemFree(_ToolBarButtons[LC]); } _ToolBarButtons_Count = 0;
	}

	return(E_SUCCESS);
}
