//================================================================================================
// Plugin for EQUINOX-3D
//
// Example to show how to create custom Node.
// This plugin demonstrates Snell's law of refraction.
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 03:54:05
//
// Copyright (C) 2014-2021 by Gabor Nagy. All rights reserved.
//================================================================================================
#include <assert.h>
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/Scene.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>

#include <Xe/PushButton.h>
#include <Xe/Scale.h>


char*            E_PluginInfo = "Demonstrates Snell's law of refraction";

static EPlugin*  _Plugin = NULL;

static EguiItem  _ToolBarButtons[16];static int  _ToolBarButtons_Count = 0;






typedef struct
{
	E3dNodeComponentCore;

// E3dRefractionInfo part
//
	E3dCoordinate  RefractiveIndexTop, RefractiveIndexBottom;
	E3dCoordinate  AngleTop, AngleBottom;

	EBool          SplinesCreated;
} E3dRefractionInfo;


#define ROffset(field)  (void*)(EPtrOffset(E3dRefractionInfo, field))
static EResource  _Resources[]=
{
	{ "RefractiveIndexTop",      E3dCOORDINATE_TYPE,     ROffset(RefractiveIndexTop) },
	{ "AngleTop",                E3dCOORDINATE_TYPE,     ROffset(AngleTop) },

	{ "RefractiveIndexBottom",   E3dCOORDINATE_TYPE,     ROffset(RefractiveIndexBottom) },
	{ "AngleBottom",             E3dCOORDINATE_TYPE,     ROffset(AngleBottom) },

	{ }
};
#undef ROffset


typedef struct
{
	Widget  RefractiveIndexTopW, RefractiveIndexBottomW, AngleTopW, AngleBottomW;
} E3dResourcesGUI;


static E3dNodeComponentClass*	_ComponentClass = NULL;



static EChangeFlags _Update(E3dNode* PChangedBranchRootNode, E3dNode* PNode, E3dNodeComponent* PInfo, const EChangeFlags PChanged, const EBool PCheckOnly)
{
	if(PCheckOnly) return(0);

	E3dRefractionInfo*  LInfo = (E3dRefractionInfo*)PInfo;
	E3dScene*           LScene = E3d_Scene;
	E3dModel*           LModel = (E3dModel*)PNode;
	E3dSpline*          LSpline0 = NULL;
	E3dSpline*          LSpline1 = NULL;
	E3dSpline*          LSpline2 = NULL;
	E3dSpline*          LSpline3 = NULL;
	EChangeFlags        LTriggeredChanges = 0;


	if(LInfo == NULL)
	{
		LInfo = (E3dRefractionInfo*)E3dNodeComponent_Add((E3dNode*)LModel, _ComponentClass);
	}

	if(!LInfo->SplinesCreated)
	{
		LSpline0 = (E3dSpline*)E3dModel_AddObject(LModel, E3dSpline_class);
		EObject_SetName(LSpline0, E3dScene_GetUniqueNodeName(LScene, "top"), FALSE);

		LSpline1 = (E3dSpline*)E3dModel_AddObject(LModel, E3dSpline_class);
		EObject_SetName(LSpline1, E3dScene_GetUniqueNodeName(LScene, "bottom"), FALSE);

		LSpline2 = (E3dSpline*)E3dModel_AddObject(LModel, E3dSpline_class);
		EObject_SetName(LSpline2, E3dScene_GetUniqueNodeName(LScene, "bottomV"), FALSE);

		LSpline3 = (E3dSpline*)E3dModel_AddObject(LModel, E3dSpline_class);
		EObject_SetName(LSpline3, E3dScene_GetUniqueNodeName(LScene, "topV"), FALSE);
		LInfo->SplinesCreated = TRUE;
	}

	if(PChanged == 0)
	{
// Just check consistency
//
		return(0);
	}
	else
	{
		if(LModel->Objects_Count == 4)
		{
			LSpline0 = E_DynamicCast(LModel->Objects[0], E3dSpline);
			LSpline1 = E_DynamicCast(LModel->Objects[1], E3dSpline);
			LSpline2 = E_DynamicCast(LModel->Objects[2], E3dSpline);
			LSpline3 = E_DynamicCast(LModel->Objects[3], E3dSpline);
		}
	}


	if(PChanged & E3dChgATTRIBUTES && LSpline0 && LSpline1 && LSpline2 && LSpline3)
	{
		if(LSpline0->CVs_Count != 2) EResizeNC(E3dSplineCV, 2,  LSpline0->CVs, LSpline0->CVs_Count);
		if(LSpline1->CVs_Count != 2) EResizeNC(E3dSplineCV, 2,  LSpline1->CVs, LSpline1->CVs_Count);
		if(LSpline2->CVs_Count != 2) EResizeNC(E3dSplineCV, 2,  LSpline2->CVs, LSpline2->CVs_Count);
		if(LSpline3->CVs_Count != 2) EResizeNC(E3dSplineCV, 2,  LSpline3->CVs, LSpline3->CVs_Count);

		LTriggeredChanges |= E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY;

		const E3dCoordinate  LLen = 10.0;

		E3dSplineCV*	LCV;
		LCV = LSpline0->CVs;
		E3dSplineCV_Init(LCV, E3dSplineLINEAR);E_SetXYZ_s(LCV->Position,  0.0, 0.0, 0.0);LCV++;

		const E3dCoordinate  LTopAngle = 90.0 - LInfo->AngleTop;
		E3dCoordinate  LX = E3d_Cos(LTopAngle * E3dDEGREES_TO_RADIANS), LY = sin(LTopAngle * E3dDEGREES_TO_RADIANS);
		{
			E3dSplineCV_Init(LCV, E3dSplineLINEAR);E_SetXYZ_s(LCV->Position,  LX * LLen, LY * LLen, 0.0);LCV++;
		}

		LCV = LSpline1->CVs;
		E3dSplineCV_Init(LCV, E3dSplineLINEAR);E_SetXYZ_s(LCV->Position,  0.0, 0.0, 0.0);LCV++;


		const E3dCoordinate  LEta12 = LInfo->RefractiveIndexTop / LInfo->RefractiveIndexBottom;
		LInfo->AngleBottom = asin(LX * LEta12) * E3dRADIANS_TO_DEGREES;


		E3dRay  LNewRay;

		{
			E3dRay           LTopRay;E_SetXYZ_s(LTopRay.Direction,  -LX, -LY, 0.0);
			const E3dCoord3  LNormal = { 0.0, 1.0, 0.0 };
			if(E3d_RayRefract(&LTopRay, &LNewRay, &LNormal, LEta12))
			{
				E3dSplineCV_Init(LCV, E3dSplineLINEAR);E_SetXYZ_s(LCV->Position,  LNewRay.Direction.X * LLen, LNewRay.Direction.Y * LLen, 0.0);LCV++;
			}
			else
			{
// Total internal reflection - indicate it with a 0 length vector
//
				LNewRay.Direction.X = LNewRay.Direction.Y = LNewRay.Direction.Z = 0.0;
			}
		}

		const E3dCoordinate  LVOffset = 1.0;


		LCV = LSpline2->CVs;
		E3dSplineCV_Init(LCV, E3dSplineLINEAR);E_SetXYZ_s(LCV->Position,  LVOffset, 0.0, 0.0);LCV++;
		E3dSplineCV_Init(LCV, E3dSplineLINEAR);E_SetXYZ_s(LCV->Position,  LVOffset + LNewRay.Direction.X * LLen, LNewRay.Direction.Y * LLen, 0.0);LCV++;



		E3dRay  LNewRayV;

		LCV = LSpline3->CVs;
		E3dSplineCV_Init(LCV, E3dSplineLINEAR);E_SetXYZ_s(LCV->Position,  LVOffset, 0.0, 0.0);LCV++;

		LNewRay.Direction.X *= -1.0;LNewRay.Direction.Y *= -1.0;
		const E3dCoord3  LNormal = { 0.0, -1.0, 0.0 };
		if(E3d_RayRefract(&LNewRay, &LNewRayV, &LNormal, 1.0 / LEta12))
		{
			E3dSplineCV_Init(LCV, E3dSplineLINEAR);E_SetXYZ_s(LCV->Position,  LVOffset + LNewRayV.Direction.X * LLen, LNewRayV.Direction.Y * LLen, 0.0);LCV++;
		}
		else
		{
			E3dSplineCV_Init(LCV, E3dSplineLINEAR);E_SetXYZ_s(LCV->Position,  LVOffset, 0.0, 0.0);LCV++;
		}


		E3dNode_UpdateForDisplay(PNode, E3duf_SHAPE);
	}



	return(LTriggeredChanges);
}


//================================================
// ToolBar icon callback handler
//================================================
static void _CB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dModel*  LModel = E3dModel_CustomNew(E3d_Scene, _ComponentClass, FALSE);if(LModel)
	{
		E3dScene*  LScene = E3d_Scene;
		E3dScene_NodeHrcAdd(E3d_Scene, (E3dNode*)LModel, E3dUF_CALL_SCENE_CALLBACKS);
		E3dScene_ObjectsUpdateModelList(LScene);
		E3dScene_LayoutOnSchematics(LScene, TRUE);
	}

	if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) E3dNode_Edit((E3dNode*)LModel, _ComponentClass);

	E3dWindows_RedrawByChangeFlags(E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY|E3dChgHIERARCHY);
}


//========================================
// Constructor
//========================================
static int _Init(E3dNode* PNode, E3dNodeComponent* PInfo)
{
	E3dRefractionInfo*  LThis = (E3dRefractionInfo*)PInfo;

	LThis->RefractiveIndexTop = 1.0; LThis->RefractiveIndexBottom = 1.0;
	LThis->AngleTop = 20.0, LThis->AngleBottom = 20.0;;
	LThis->SplinesCreated = FALSE;

	return(E_SUCCESS);
}


#define _M_Copy(mField)  LDInfo->mField = LSInfo->mField

//========================================
// Copy-constructor
//========================================
static int _Copy(const E3dNode* PSrcNode, const E3dNodeComponent* PSrcComponent,  E3dNode* PDstNode, E3dNodeComponent* PDstComponent)
{
	E3dRefractionInfo*  LSInfo = (E3dRefractionInfo*)PSrcComponent;
	E3dRefractionInfo*  LDInfo = (E3dRefractionInfo*)PDstComponent;

	_M_Copy(RefractiveIndexTop);
	_M_Copy(RefractiveIndexBottom);
	_M_Copy(AngleTop);
	_M_Copy(AngleBottom);

	return(E_SUCCESS);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_RefractiveIndexTop(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*       LPanel = (E3dNodePanel*)PClientData;
	E3dNode*            LNode = LPanel->Node;
	E3dRefractionInfo*  LInfo = (E3dRefractionInfo*)E3dNodeComponent_ByClass(LNode, _ComponentClass);


	LInfo->RefractiveIndexTop = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);

	E3dWindows_RedrawByChangeFlags(E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_AngleTop(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*       LPanel = (E3dNodePanel*)PClientData;
	E3dNode*            LNode = LPanel->Node;
	E3dRefractionInfo*  LInfo = (E3dRefractionInfo*)E3dNodeComponent_ByClass(LNode, _ComponentClass);

	LInfo->AngleTop = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);

	E3dNodeComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, LClass);if(LDrawer)
		{
			E3dResourcesGUI*  LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

			XeScale_SetValue_Coord(LResourcesGUI->AngleBottomW, LInfo->AngleBottom, FALSE);
		}
	}

	E3dWindows_RedrawByChangeFlags(E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_RefractiveIndexBottom(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*       LPanel = (E3dNodePanel*)PClientData;
	E3dNode*            LNode = LPanel->Node;
	E3dRefractionInfo*  LInfo = (E3dRefractionInfo*)E3dNodeComponent_ByClass(LNode, _ComponentClass);

	LInfo->RefractiveIndexBottom = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);

	E3dWindows_RedrawByChangeFlags(E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_AngleBottom(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*       LPanel = (E3dNodePanel*)PClientData;
	E3dNode*            LNode = LPanel->Node;
	E3dRefractionInfo*  LInfo = (E3dRefractionInfo*)E3dNodeComponent_ByClass(LNode, _ComponentClass);


	LInfo->AngleBottom = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_Update(NULL, LNode, (E3dNodeComponent*)LInfo, E3dChgATTRIBUTES, FALSE);

	E3dWindows_RedrawByChangeFlags(E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY);
}


static void _GUI(E3dNode* PNode, E3dNodeComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dNodePanel*            LPanel = (E3dNodePanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dRefractionInfo*       LInfo = (E3dRefractionInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LInfo)
			{
				if(LDrawer->ResourcesGUI)
				{
					E3dResourcesGUI*	LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

					XeScale_SetValue_Coord(LResourcesGUI->AngleTopW,              LInfo->AngleTop,              FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->RefractiveIndexTopW,    LInfo->RefractiveIndexTop,    FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->AngleBottomW,           LInfo->AngleBottom,           FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->RefractiveIndexBottomW, LInfo->RefractiveIndexBottom, FALSE);
				}
				else
				{
					Widget            LTopMatrixW = LDrawer->Drawer.DrawerW, LW;
					Widget            LIndentWs[7];int	LIndentN;      // For indenting the Scale Widgets
					Widget            LTraverseWs[16];int	LTraverseN;  // For TextField traversal
					EArg              Args[256];int  ArgCnt;
					E3dResourcesGUI*  LResourcesGUI;ENew(LResourcesGUI);

					LDrawer->ResourcesGUI = LResourcesGUI;

					EXeStart;
					EXeSetArg(XeNlabelAlignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNfontList, Xe_DropDownPanelFontList);
					EXeSetArg(XeNmarginTop, 5);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeWidget_ManagedNew("Top side", XeLabelWidget_class, LTopMatrixW, Args, ArgCnt);

					LTraverseN = 0;
					LIndentN = 0;
					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 120);
					EXeSetArg(XeNtextFieldDigits, 6);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					EXeSetArg(XeNuserData, LInfo->Class);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->RefractiveIndexTopW = LW = XeScaleNew_Coord(LTopMatrixW, "Refractive index", Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->RefractiveIndexTop,  0.001, 10.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_RefractiveIndexTop, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 120);
					EXeSetArg(XeNtextFieldDigits, 6);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					EXeSetArg(XeNuserData, LInfo->Class);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->AngleTopW = LW = XeScaleNew_Coord(LTopMatrixW, "Angle", Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->AngleTop,  0.0, 90.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_AngleTop, (EPointer)LPanel);


					EXeStart;
					EXeSetArg(XeNlabelAlignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNfontList, Xe_DropDownPanelFontList);
					EXeSetArg(XeNmarginTop, 5);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeWidget_ManagedNew("Bottom side", XeLabelWidget_class, LTopMatrixW, Args, ArgCnt);


					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 120);
					EXeSetArg(XeNtextFieldDigits, 6);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					EXeSetArg(XeNuserData, LInfo->Class);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->RefractiveIndexBottomW = LW = XeScaleNew_Coord(LTopMatrixW, "Refractive index", Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->RefractiveIndexBottom,  0.001, 10.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_RefractiveIndexBottom, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 120);
					EXeSetArg(XeNtextFieldDigits, 6);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					EXeSetArg(XeNuserData, LInfo->Class);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->AngleBottomW = LW = XeScaleNew_Coord(LTopMatrixW, "Angle", Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->AngleBottom,  0.0, 90.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_AngleBottom, (EPointer)LPanel);

					Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);

// Set up TextField traversal
//
					Xe_TraversalRing(LTraverseWs, LTraverseN);
				}
			}
		break;

		case E_FREE:
		break;
	}
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	{
		E3dNodeComponentClass  LComponentClass =
		{
			.Name=        "Refraction",
			.StructSize=  sizeof(E3dRefractionInfo),
			.Resources=   _Resources,
			.Init=        _Init,
			.Copy=        _Copy,
			.Update=      _Update,
			.GUI=         _GUI
		};
		_ComponentClass = E3dNodeComponent_ClassRegister(&LComponentClass);
	}

	_Plugin = PPlugin;

	if(_ComponentClass)
	{
// Use default plugin icon
//
		Xe_PushButtonAddImg3(_ToolBarButtons[_ToolBarButtons_Count++], Plugin, "Tool->Create", "Refraction", '\0', NULL, NULL, FALSE, "This tool demonstrates Snell's law of refraction", _CB, (EPointer)0);
	}

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC = 0;for(;LC < _ToolBarButtons_Count;LC++) { XeItemFree(_ToolBarButtons[LC]); }
		_ToolBarButtons_Count = 0;
	}

	if(_ComponentClass) { E3dNodeComponent_ClassDeactivate(_ComponentClass);_ComponentClass = NULL; }

	return(E_SUCCESS);
}
