//================================================================
// EQUINOX-3D plugin
//
// - Invert Geometry
// - Add back-faces
// - Invert Polygons
// - Separate (un-weld) selected Vertices of a Mesh
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-01 11:16:41
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <assert.h>
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Matrix.h>
#include <Xe/ToggleButton.h>

#include <E3D/Matrix.h>
#include <E3D/Scene.h>
#include <E3D/Select.h>

#include <E3D/PolygonEditing.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>

#include <Xe/PushButton.h>

char*  E_PluginInfo = "Various Geometry tools, including:\n- Invert Geometry\n- Add back faces\n- Separate (un-weld) selected Vertices of a Mesh";


static EPlugin*  _Plugin = NULL;

static EguiItem  _MenuButtons[16];static int  _MenuButtons_Count = 0;

//static const EBool  _SeparateSelectedObjects_AllInstances = TRUE;

static const char  _NewObjectAddingMode = E3dNEW_INSTANCE;
//static const char  _NewObjectAddingMode = E3dNEW_MODEL;


typedef struct
{
	E3dObjectComponentCore;

	E3dGeometry*   InputGeometry;
	EInterface     InputInterface;         // This is added to the profile Geometries as an output node

// Parameters
//
	EBool          BackSeparatePolyGroups;  // Make the back-side separate PolyGroup(s) than the front and the sides
} E3dInfo;


#define ROffsetP(field)  (void*)(EPtrOffset(E3dInfo, field))
#define ROffsetI(field)  EPtrOffset(E3dInfo, field)
static EResource  _Resources[]=
{
	{ "InputGeometry",  EResOBJECT,   ROffsetP(InputGeometry) },
	{ "Separate back",  EResBOOLEAN,  ROffsetP(BackSeparatePolyGroups) },
	{ }
};
#undef ROffsetP
#undef ROffsetI

static E3dObjectComponentClass*  _ObjectComponent_class = NULL;

typedef struct
{
	Widget            BackSeparatePolyGroupsW;
} E3dResourcesGUI;


//================================================================
// Callback for Profile-Spline changes
// This function should Not call any 3DWindow redrawing function,
// just return with TRUE if a redraw is needed!
//================================================================
static EChangeFlags _InputChanged(EInterface* PInterface, EInterfaceCallbackStruct* PCBS)
{
	E3dObject*    LObject  = (E3dObject*)PInterface->Object;
	E3dInfo*      LInfo    = (E3dInfo*)PInterface->ObjectComponent;
	EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass(LObject);

// Only respond if the Object is instantiated (e.g. not in the undo stack after being deleted)
//
	if(LObject->Models_Count)
	{
		E3dObject_CustomUpdate(LObject, LInfo, PCBS->CallOutputs ? E3dUF_ALL : E3dUF_ALL - E3dUF_CALL_OUTPUTS, &LChanged);
	}

	return(LChanged);
}


//================================================
// Add back-face Constructor
//================================================
static int _Init(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dInfo*  LThis = (E3dInfo*)PInfo;

	LThis->InputGeometry = NULL;
	LThis->BackSeparatePolyGroups = TRUE;

	{
		EInterface*  LInput = &LThis->InputInterface;E_InterfaceInit(LInput);
		LInput->InputChanged    = _InputChanged;
		LInput->Object          = (ECoreObject*)PObject;
		LInput->ObjectComponent = (ECoreObject*)PInfo;
	}

	return(E_SUCCESS);
}


//========================================
// Add back-face Copy-constructor
//========================================
static void _Copy(const E3dObject* PSObject, const E3dObjectComponent* PSInfo,  E3dObject* PDObject, E3dObjectComponent* PDInfo)
{
	const E3dInfo*  LSInfo = (E3dInfo*)PSInfo;
	E3dInfo*        LDInfo = (E3dInfo*)PDInfo;

	_Init(PDObject, PDInfo);

#define _CopyField(f)  LDInfo->f = LSInfo->f
	_CopyField(BackSeparatePolyGroups);
#undef _CopyField

	E_AssignToMember(LSInfo->InputGeometry, LDInfo,InputGeometry);
	if(LSInfo->InputGeometry)
	{
		E3dBaseObject_AddOutput((E3dBaseObject*)LDInfo->InputGeometry, &LDInfo->InputInterface);
	}
}

//================================================
// Add back-face Destructor
//================================================
static int _Free(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dInfo*  LInfo = (E3dInfo*)PInfo;

	E3dBaseObject_RemoveOutput((E3dBaseObject*)LInfo->InputGeometry, &LInfo->InputInterface);

	E_DeleteChk(LInfo->InputGeometry);

	E_InterfaceClear(&LInfo->InputInterface);

	return(E_SUCCESS);
}


//========================================
// Add back-face Update method
//========================================
static E3dObject* _Update(E3dObject* PObject, E3dObjectComponent* PInfoV, const EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	E3dScene*      LScene = E3d_Scene;
	E3dMesh*       LMesh;
	E3dInfo*       LInfo;

	if(PObject == NULL)
	{
		E_New(E3dMesh, LMesh);
		PObject = (E3dObject*)LMesh;

		LInfo = (E3dInfo*)E3dObjectComponent_Add(PObject, _ObjectComponent_class);
		if(PInfoV && PInfoV->Class == _ObjectComponent_class)
		{
			_Copy(PObject, PInfoV, PObject, (E3dObjectComponent*)LInfo);
		}

		E3dBaseObject_AddOutput((E3dBaseObject*)LInfo->InputGeometry, &LInfo->InputInterface);
	}
	else
	{
		if(!EObject_IsSubClass(PObject, E3dMesh)) return(NULL);

		LMesh = (E3dMesh*)PObject;
		LInfo = (E3dInfo*)E3dObjectComponent_ByClass(PObject, _ObjectComponent_class);

// When a Scene is loaded, Update methods are called to create this link.
// The Geometry will be up-to-date, so if the link wasn't there yet, just exit
// We should probably add a 3rd argument to this function ("Action=LINK"), so
// even input-less ComponentClasses don't rebuild their Geometry.
//
		if(PCheckOnly && LInfo->InputGeometry)
		{
			if(E3dBaseObject_AddOutput((E3dBaseObject*)LInfo->InputGeometry, &LInfo->InputInterface) == E_SUCCESS)
			{
				return(PObject);
			}
		}
	}

	if(LMesh)
	{
		E3dMesh_FreeCaches(LMesh);
		E3dMesh_EdgesFree(LMesh);
		E3dMesh_PolyGroupsFree(LMesh);

		const E3dMesh*  LInputMesh = (E3dMesh*)LInfo->InputGeometry;if(LInputMesh)
		{
			const EIndex  LVN = LInputMesh->Vertices_Count;if(LVN > 0)
			{
				if(LMesh->Vertices_Count != LVN)
				{
					E3dMesh_VerticesFree(LMesh);
					E3dMesh_VerticesAdd(LMesh, LVN);
				}
				E3dVertices_Copy(LInputMesh->Vertices, LInputMesh->VertexSize, &LInputMesh->VertexAttributes,  LMesh->Vertices, LMesh->VertexSize, &LMesh->VertexAttributes,  LVN);
			}
			else E3dMesh_VerticesFree(LMesh);
		}

		if(LInputMesh->Edges_Count > 0)
		{
			const EIndex  LEN = LInputMesh->Edges_Count;
			ECopyN(LInputMesh->Edges, LEN, LMesh->Edges);LMesh->Edges_Count = LMesh->Edges_CountAllocated = LEN;
		}

		LMesh->SaveEdges = LInputMesh->SaveEdges;

		const EBool  LBackSeparatePolyGroups = LInfo->BackSeparatePolyGroups;

		const EIndex  LPGN = LInputMesh->PolyGroups_Count;if(LPGN)
		{
			E3dPolyGroup**  LPolyGroups = LInputMesh->PolyGroups;
			E3dMatrix       LMatrix;E3dMatrix_LoadIdentity(LMatrix);

			EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
			{
				E3dPolyGroup*            LSrcPolyGroup = LPolyGroups[LPGC];
				const E3dVertexNodeSize  LVertexNodeSize = LSrcPolyGroup->VertexNodeSize;

				if(LBackSeparatePolyGroups)
				{
					{
//					E3dPolyGroup*  LNewPolyGroup = E3dPolyGroup_Copy(LSrcPolyGroup, NULL,  NULL,  ECopyNAME|E3dCopyPOLYGONS|E3dCopyMATERIALS);
						E3dPolyGroup*  LNewPolyGroup = E3dPolyGroup_Copy(LSrcPolyGroup, NULL,  NULL,  ECopyNAME|E3dCopyPOLYGONS);
						E3dMesh_AppendPolyGroup(LMesh, LNewPolyGroup);E3dM_ObjectUpdateDrawingMaterial(PObject, LNewPolyGroup->Material, LNewPolyGroup,DrawingMaterial);

						if(LSrcPolyGroup->Edges_Count > 0)
						{
							const EIndex  LEN = LSrcPolyGroup->Edges_Count;
							ECopyN(LSrcPolyGroup->Edges, LEN, LNewPolyGroup->Edges);LNewPolyGroup->Edges_Count = LEN;
						}
					}

					{
						E3dPolyGroup*  LNewPolyGroup = E3dPolyGroup_Copy(LSrcPolyGroup, NULL,  NULL,  E3dCopyPOLYGONS);
						const char*    LPGName = LSrcPolyGroup->Name;
						const int      LNL = (LPGName ? strlen(LPGName) : 0) + 6;char*  LName = EMalloc(LNL);sprintf(LName, "%s-back", LPGName ? LPGName : "");
						EObject_SetName(LNewPolyGroup, LName, FALSE);
						E3dMesh_AppendPolyGroup(LMesh, LNewPolyGroup);E3dM_ObjectUpdateDrawingMaterial(PObject, LNewPolyGroup->Material, LNewPolyGroup,DrawingMaterial);

						const int  LNormalOffset = E3dVertexNodeAttribute_Offset(&LNewPolyGroup->VertexAttributes,  E3dVtxAttribute_Normal, 0);
						{ EIndex  LPC = 0;const EIndex  LPN = LNewPolyGroup->Polygons_Count;E3dPolygon*  LPolygon = LNewPolyGroup->Polygons;for(;LPC < LPN;LPC++, LPolygon++) E3dPolygon_Invert(LPolygon, LVertexNodeSize, LNormalOffset); }

						if(LSrcPolyGroup->Edges_Count > 0)
						{
							const EIndex  LEN = LSrcPolyGroup->Edges_Count;
							ECopyN(LSrcPolyGroup->Edges, LEN, LNewPolyGroup->Edges);LNewPolyGroup->Edges_Count = LEN;
						}
					}
				}
				else
				{
					E3dPolyGroup*  LNewPolyGroup = E3dPolyGroup_Copy(LSrcPolyGroup, NULL,  NULL,  ECopyNAME|E3dCopyPOLYGONS);
					E3dMesh_AppendPolyGroup(LMesh, LNewPolyGroup);E3dM_ObjectUpdateDrawingMaterial(PObject, LNewPolyGroup->Material, LNewPolyGroup,DrawingMaterial);

					if(LSrcPolyGroup->Edges_Count > 0)
					{
						const EIndex  LEN = LSrcPolyGroup->Edges_Count;
						ECopyN(LSrcPolyGroup->Edges, LEN, LNewPolyGroup->Edges);LNewPolyGroup->Edges_Count = LEN;
					}

					const EIndex  LPN = LNewPolyGroup->Polygons_Count;if(LPN > 0)
					{
						const int  LNormalOffset = E3dVertexNodeAttribute_Offset(&LNewPolyGroup->VertexAttributes,  E3dVtxAttribute_Normal, 0);
						EResizeNT(E3dPolygon, LPN * 2, LNewPolyGroup->Polygons);
						EIndex  LPC = 0;const E3dPolygon*  LSrcPolygon = LNewPolyGroup->Polygons;E3dPolygon*  LDstPolygon = LNewPolyGroup->Polygons + LPN;for(;LPC < LPN;LPC++, LSrcPolygon++, LDstPolygon++)
						{
							E3d_PolygonCopy(LSrcPolygon, LVertexNodeSize, LDstPolygon);
							E3dPolygon_Invert(LDstPolygon, LVertexNodeSize, LNormalOffset);
						}
						LNewPolyGroup->Polygons_Count = LPN * 2;
					}
				}
			}
		}


		E3dMesh_EdgeFlagsUpdate(LMesh);

		EChangeFlags  LChanged;
//		if(LCreateEdges) LChanged = E3dObject_UpdateForDisplay(PObject, PUpdateFlags & ~E3dUF_EDGES, LScene);
//		else
LChanged = E3dObject_UpdateForDisplay(PObject, PUpdateFlags & ~E3dUF_EDGES, LScene);
//LChanged = E3dObject_UpdateForDisplay(PObject, E3duf_FULLGEO, LScene);

		if(PTriggeredChangesP) *PTriggeredChangesP |= LChanged;
	}

	return(PObject);
}


static void _DCB_ProfileShow(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);
	EChangeFlags     LChanged = 0;

	if(LInfo->InputGeometry)
	{
		if(E3dSel_ObjectSelect(E3d_Scene, (E3dObject*)LInfo->InputGeometry, FALSE, FALSE)) LChanged |= E3dChgSELECTION;

		E3dWindows_RedrawByChangeFlags(LChanged);
		E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
		E3dp_PrintMessage(ELogMESSAGE, 3000, "Revolve: selected profile Geometry <%s>", LInfo->InputGeometry->Name);
	}
	else E3dp_PrintMessage(ELogMESSAGE, 3000, "Revolve: no profile Geometry!");
}


static E3dInfo*         _TmpInfo = NULL;
static E3dObjectPanel*  _TmpPanel = NULL;
static const char*      E3dpLng_Pick_a_Mesh_as_the_new_input = "Pick a Mesh as the new input";

#define _PickInputGeometryTypeWarningStr  E3dpLng_Pick_a_Mesh_as_the_new_input
#define _PickInputGeometryObjectTypes     E3dItemMESH

#include <E3D/PickInput.ci>

static void _RequestProfilePick(void)
{
	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color = Xe_ForegroundRGBA8pColor;

	E3dp_PickRequest(
	 XeString_NewVa(TRUE,
	 XeFONT, XeFontDEFAULT,  XeCOLOR_RGBA, LLabelRGBAp8Color,  XeTEXT, _ObjectComponent_class->Name,  XeTEXT, ":    ",
	 E3dp_LMouseButtonS,     XeTEXT, ": pick a Mesh as the new input  ",
	 E3dp_RMouseButtonEscS,  XeTEXT, ": cancel",
	 0),
	 E3dItemMESH|E3dItemMODEL,  _PickCB_InputGeometry, NULL,  0, FALSE);
}


static void _DCB_ProfilePick(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	_TmpInfo = LInfo;
	_TmpPanel = LPanel;
	_RequestProfilePick();
}


static void _DCB_BackSeparatePolyGroups(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->BackSeparatePolyGroups = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass(LObject);E3dObject_CustomUpdate(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// GUI
//========================================
static void _GUI(E3dObject* PObject, E3dObjectComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dResourcesGUI*         LResourcesGUI;
	E3dInfo*                 LInfo = (E3dInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

				XeToggleButton_SetState(LResourcesGUI->BackSeparatePolyGroupsW, LInfo->BackSeparatePolyGroups, FALSE);
			}
			else
			{
				Widget  LTopMatrixW = LDrawer->Drawer.DrawerW, LW;
				Widget  LTraverseWs[7];int  LTraverseN;  // For TextField traversal
				EArg    Args[256];int  ArgCnt;

				ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

// Input Geometry
//
				{
					EXeStart;
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginSizeY, 5);
					EXeSetArg(XeNspacingX, 0);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
					EXeSetArg(XeNfontList, Xe_LabelFontList);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNmarginSizeX, 2);EXeSetArg(XeNmarginSizeY, 3);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeWidget_ManagedNew("Profile", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNspacingLeft, 6);
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);EXeSetArg(XeNactiveColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNlabelType, XeIMAGE);EXeSetArg(XeNimageSizeYPtr, &Xe_IconSize);EXeSetImages2(_Xe_Theme->Image_Find);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Select the input Mesh (so you can find it).");
					LW = XeWidget_ManagedNew("Show", XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_ProfileShow, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);EXeSetArg(XeNactiveColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNlabelType, XeIMAGE);EXeSetArg(XeNimageSizeYPtr, &Xe_IconSize);EXeSetImages2(_Xe_Theme->Image_Replace);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Pick a new input Mesh.");
					LW = XeWidget_ManagedNew("Pick", XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_ProfilePick, (EPointer)LPanel);
				}


/*
				EXeStart;
				EXeSetArg(XeNmarginSizeY, 0);
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNspacingX, 12);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				EXeSetArg(XeNheightAdjustable, FALSE);
				Widget  LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);
*/

				EXeStart;
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNset, LInfo->BackSeparatePolyGroups);
				XeToolTip8("Make the back-side out of separate PolyGroups than the front and the side.\nThis allows assigning a diferent Material to the inside of objects.");
				LResourcesGUI->BackSeparatePolyGroupsW = LW = XeWidget_ManagedNew("Separate back", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_BackSeparatePolyGroups, (EPointer)LPanel);

				LTraverseN = 0;


// Set up TextField traversal
//
				Xe_TraversalRing(LTraverseWs, LTraverseN);
			}
		break;
	}
}






//========================================
// Menu callback
//========================================
static void _MCB_Invert(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	int           LDid = 0;
	EChangeFlags  LChanged = 0;
	E3dObject**   LObjects;const EIndex	 LNumOfObjects = E3dScene_GetObjects(LScene, &LObjects, E3dItemGEOMETRY, E3dSELECTED, FALSE);if(LNumOfObjects)
	{
		EIndex  LC = 0;for(;LC < LNumOfObjects; LC++)
		{
			E3dObject*  LObject = LObjects[LC];

			if(E3dGeometry_Invert((E3dGeometry*)LObject, &LChanged) == E_SUCCESS)
			{
				LDid++;
				if(EObject_IsSubClass(LObject, E3dSpline) || EObject_IsSubClass(LObject, E3dFace))
				{
					E3dObject_UpdateForDisplay(LObject, E3dUF_ALL, LScene);
				}
			}
		}

		EFree(LObjects);
	}

	if(LDid == 0) E3dp_PrintMessage(ELogWARNING, 5000, "Select Mesh, PolyGroup, Spline or Face to invert");

	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Menu callback
//========================================
static void _MCB_AddBackFaces(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*   LScene = E3d_Scene;
	E3dObjItem  LItem;

	if(E3dScene_GetSingleSelection(LScene, &LItem, E3dItemMESH, E3dVISIBLE) != E3dItemNONE)
	{
		E3dObject*  LInputObject = LItem.Object;
		EBool       LSuccess = FALSE;

		E3dInfo  LTInfo;E3dObjectComponent_InitStatic(&LTInfo, _ObjectComponent_class);
		E_Assign(LInputObject, LTInfo.InputGeometry);

		E3dModel*  LOrigModel = E3dObject_GetSelectingModel(LInputObject, FALSE);
		if(LOrigModel == NULL) LOrigModel = LInputObject->Models[0];

		E3dObject*  LObject = E3dObject_CustomNew(LScene, NULL, LOrigModel, NULL, (E3dObjectComponent*)(&LTInfo), E3dUF_ALL, _NewObjectAddingMode, (E3dp_Prefs.Duplicate_Flags & E3dCopyMODEL_UNIQUE_NAME) != 0,  NULL);if(LObject)
		{
			((E3dGeometry*)LObject)->AutoLOD = FALSE;

			LSuccess = TRUE;
		  E3dScene_ObjectsUpdateModelList(LScene);
			E3dScene_LayoutOnSchematics(LScene, TRUE);
		}

		E_Delete(LTInfo.InputGeometry);

		if(LSuccess)
		{
			if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) E3dObject_Edit(LObject, _ObjectComponent_class);

			E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE|E3dChgHIERARCHY);
		}
	}
	else E3dp_PrintMessage(ELogWARNING, 5000, "Select a Mesh");
}


//========================================
// Menu callback
//========================================
static void _MCB_InvertPolygons(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	E3dObject**   LObjects;
	const EIndex  LNumOfObjects = E3dScene_GetObjects(LScene, &LObjects, E3dItemGEOMETRY, E3dSELECTED, FALSE);
	int           LDid = 0;


	if(LNumOfObjects)
	{
		EIndex  LC = 0;for(;LC < LNumOfObjects; LC++)
		{
			E3dObject*  LObject = LObjects[LC];


			if(EObject_IsSubClass(LObject, E3dMesh))
			{
				E3dMesh*  LMesh = (E3dMesh*)LObject;


				if(LMesh->PolyGroups_Count)
				{
					E3dPolyGroup**  LPolyGroups = LMesh->PolyGroups;
					EBool           LChanged = FALSE;


					EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
					{
						E3dPolyGroup*            LPolyGroup = LPolyGroups[LPGC];
						const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

						if((LPolyGroup->Selected) || (LMesh->Selection != E3dObjectSelPOLYGROUP))
						{
							E3dPolygon*  LPolygon = LPolyGroup->Polygons;if(LPolygon)
							{
								const int  LNormalOffset = E3dVertexNodeAttribute_Offset(&LPolyGroup->VertexAttributes,  E3dVtxAttribute_Normal, 0);
								EIndex  LPC = 0;const EIndex  LPN = LPolyGroup->Polygons_Count;for(;LPC < LPN;LPC++, LPolygon++) if(LPolygon->Flags & E3dPolygonSELECTED)
								{
									E3dPolygon_Invert(LPolygon, LVertexNodeSize, LNormalOffset);LChanged = TRUE;
								}
							}
						}
					}
					if(LChanged)
					{
						LDid++;
						E3dMesh_CreateRenderablePolygons(LMesh, TRUE);
						E3dObject_UpdateForDisplay((E3dObject*)LMesh, E3duf_SHAPE, LScene);
					}
				}
			}
		}
		EFree(LObjects);
	}

	if(LDid == 0) E3dp_PrintMessage(ELogWARNING, 5000, "Select Polygons to invert");
	else E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
}


//========================================================
// Split Mesh Vertices if their corresponding Polygon
// vertex-normals, texture coordinates etc. are different
//========================================================
static EIndex _MeshUnweldSelectedPoints(E3dMesh* PMesh)
{
	if(PMesh->Vertices_Count)
	{
		const E3dVertexUsage*  LVertexUsageList = E3dMesh_GetVertexUsageList(PMesh, TRUE, FALSE);
		const EIndex           LVN = PMesh->Vertices_Count;
		EIndex*                LNewToOldTable;ENewN(LVN, LNewToOldTable);
		EIndex                 LNewToOldTableAllocated = LVN;
		EIndex                 LNumOfVerticesAdded = 0, LNewNumOfVertices = LVN;

// Fill out first PMesh->Vertices_Count elements of LNewToOldTable
//
		{
			EIndex  LVC = 0;for(;LVC < LVN;LVC++)
			{
				LNewToOldTable[LVC] = LVC;
			}
		}

		EIndex  LVC = 0;const E3dVertex*  LVertex = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;const E3dVertexUsage*  LVertexUsageEntry = LVertexUsageList;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LVertexSize), LVertexUsageEntry++)
		{
			EIndex  LVNC = 1;const EIndex  LVNN = LVertexUsageEntry->VertexNodes_Count;for(;LVNC < LVNN;LVNC++)
			{
				E3dVertexNode*  LVertexNode = LVertexUsageEntry->VertexNodes[LVNC];

				if(LVertex->Flags & E3dVtxSELECTED)
				{
					if(LNewNumOfVertices >= LNewToOldTableAllocated) EResizeNC(EIndex, LNewToOldTableAllocated + 16, LNewToOldTable, LNewToOldTableAllocated);
					LNewToOldTable[LNewNumOfVertices] = LVC;

					LVertexNode->VertexID = LNewNumOfVertices;

					LNewNumOfVertices++;LNumOfVerticesAdded++;
				}
//printf("%s() unwelded vertex %d\n", __FUNCTION__, LVC);fflush(stdout);
			}
		}

		E3dMesh_VertexUsageListFree(PMesh);

		E3dMesh_VerticesRemap(PMesh, LNewToOldTable, LNewNumOfVertices);

		EFree(LNewToOldTable);

		return(LNumOfVerticesAdded);
	}

	return(0);
}


//========================================
// Menu callback
//========================================
static void _MCB_SeparateSelectedPoints(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	EIndex        LTotalNumOfPointsAdded = 0;
	E3dObject**   LSelectedObjects;const EIndex	LNumOfSelectedGeometries = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemMESH, E3dSELECTED, FALSE);
	EChangeFlags  LChanged = 0;

	if(LSelectedObjects)
	{
		EIndex  LC = 0;for(;LC < LNumOfSelectedGeometries;LC++)
		{
			E3dObject*  LObject = LSelectedObjects[LC];
			EIndex      LNumOfPointsAdded = _MeshUnweldSelectedPoints((E3dMesh*)LObject);

			if(LNumOfPointsAdded)
			{
				LTotalNumOfPointsAdded += LNumOfPointsAdded;

				LChanged |= E3dObject_GetChangeFlagsByClass(LObject) | E3dChgHIERARCHY;

				E3dObject_UpdateForDisplay(LObject, ~E3duf_NORMALS, LScene);
			}
		}

		EFree(LSelectedObjects);
	}

	if(LChanged)
	{
		if(LTotalNumOfPointsAdded > 1) E3dp_PrintMessage(ELogMESSAGE, 7000, "Added %d points", LTotalNumOfPointsAdded);
		else E3dp_PrintMessage(ELogMESSAGE, 7000, "Added %d point", LTotalNumOfPointsAdded);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
	else E3dp_PrintMessage(ELogWARNING, 7000, "Select Vertices to separate");
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	{
		E3dObjectComponentClass  LClass =
		{
			.Name=        "Add back-faces",
			.StructSize=  sizeof(E3dInfo),
			.Resources=   _Resources,

			.Init=        _Init,
			.Copy=        _Copy,
			.Free=        _Free,
			.Update=      _Update,
			.GUI=         _GUI
		};
		_ObjectComponent_class = E3dObjectComponent_ClassRegister(&LClass);
	}

	_Plugin = PPlugin;

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Geometry", "Invert", '\0', NULL, NULL, FALSE, NULL, _MCB_Invert, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Geometry", "Add back-faces", '\0', NULL, NULL, FALSE, NULL, _MCB_AddBackFaces, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Geometry", "Invert Polygons", '\0', NULL, NULL, FALSE, NULL, _MCB_InvertPolygons, (EPointer)0);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Geometry", "Separate Points",  '\0', NULL, NULL, FALSE, NULL, _MCB_SeparateSelectedPoints,  (EPointer)0);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC = 0;for(;LC < _MenuButtons_Count;LC++) XeItemFree(_MenuButtons[LC]);
		_MenuButtons_Count = 0;
	}

	return(E_SUCCESS);
}
