//================================================================
// EQUINOX-3D(TM) plugin
//
// Hierarchies
// - Connect / disconnect / reparent Nodes and Branches
// - Separate selected Objects (move them under a new Model)
// - Add Object to same Nodes as...
// - Collapse branches
// - Move BranchMoves / Objects within hierarchies
// - Group objects
// - Expand selected branches
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-01 04:50:44
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <assert.h>
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>


#ifndef _E3D_Matrix_h
#include <E3D/Matrix.h>
#endif

#include <E3D/Scene.h>
#include <E3D/Select.h>

#include <E3D/E3D.h>  // For undoable operators

#include <Xe/PushButton.h>
#include <Xe/ToggleButton.h>
#include <Xe/Shell.h>
#include <Xe/Dialog.h>

#include <E3D/UI/ModeStack.h>    // E3dFIXME : For E3dp_ResetWorkModes, temporarily
#include <E3D/UI/Panel.h>
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage



typedef struct
{
	E3dObject*  Object;
	EList(, EObjectAndIndex, AddToModels);
	EList(, EObjectAndIndex, DelFrModels);
} E3dObjectMove;


typedef struct
{
	E3dNode*  Node;
	E3dNode*  OldParent;EIndex  OldIndex;
	E3dNode*  NewParent;
} E3dBranchMove;


// Undoable Operator for changing hierarchy connections
//
typedef struct
{
	EOperatorCore;

	EList(, E3dBranchMove, BranchMoves);
	EList(, E3dObjectMove, ObjectMoves);
} E3dOperatorEditHierarchy;


EOperatorClass*  E3dOperatorEditHierarchy_class = NULL;


char*                E_PluginInfo = "Hierarchy manipulation:\n - Collapse branch\n - Connect / disconnect Nodes and Objects";
static EPlugin*      _Plugin = NULL;
static EPluginLock*  _ConnectPickLock = NULL;
static EPluginLock*  _DisconnectPickLock = NULL;

static const EBool  _SeparateSelectedObjects_NewModelUnderOldParent = TRUE;


EpIconDef3(Connect)
EpIconDef3(Disconnect)
EpIconDef3(Reparent)


static const char*  _Msg_SelectModelOrObjectToReparent     = "Select Models or Objects to reparent";
static const char*  _Msg_SelectOnlyModelOrObjectToReparent = "Select either only Models or only Objects to reparent";

static const char*  _Msg_SelectObjectToReInstantiate       = "Select Objects to re-instantiate";

static const char*  _Msg_SelectModelOrObjectToConnect  = "Select Models or Objects to connect";
static const char*  _Msg_SelectOnlyModelOrObject       = "Select either only Models or only Objects";


static const char*  _Msg_ConnectStr        = "Connect";
static const char*  _Msg_ConnectModelStr   = "Connect Model";
static const char*  _Msg_ConnectModelsStr  = "Connect Models";
static const char*  _Msg_ConnectObjectStr  = "Connect Object";
static const char*  _Msg_ConnectObjectsStr = "Connect Objects";

static const char*  _Msg_ReparentModelStr   = "Reparent Model";
static const char*  _Msg_ReparentModelsStr  = "Reparent Models";
static const char*  _Msg_ReparentObjectStr  = "Reparent Object";
static const char*  _Msg_ReparentObjectsStr = "Reparent Objects";

static const char*  _Msg_ReparentStr      = "Reparent";
static const char*  _Msg_InstantiateAsStr = "Connect Objects to the same Model as";


static Widget        _DialogW = NULL;
static EPluginLock*  _DialogLock = NULL;

static EguiItem  _MenuButtons[16];static int  _MenuButtons_Count = 0;
static EguiItem  _ToolBarButtons[16];static int  _ToolBarButtons_Count = 0;


static EBool    _TransformCompensation = TRUE;


#define E3dConnectCONNECT        0x01
#define E3dConnectREPARENT       0x02
#define E3dConnectINSTANTIATEAS  0x04
static uint8_t  _ConnectMode = 0;


enum
{
	CR_TOGGLE_HRC_COMPENSATION = XeCR_CUSTOM0,
};


enum
{
	E3dBA_PICK_CHILD_NODE_FOR_NODE = 1,
	E3dBA_PICK_PARENT_NODE_FOR_NODES,
	E3dBA_PICK_PARENT_MODEL_FOR_OBJECTS,
	E3dBA_PICK_CHILD_OBJECT_FOR_MODELS,
	E3dBA_PICK_NODE_TO_DISCONNECT_OBJECTS_FROM,
	E3dBA_PICK_OBJECT_FOR_ADDOBJECTTO_MODELSOF
};


static const char* _GetToolName(const int PConnectMode, const EIndex PSelectedNodes_Count, const EIndex PSelectedObjects_Count,  const char** PToolBaseNameP, EBool* PAcceptModelsP)
{
	switch(PConnectMode)
	{
		case E3dConnectCONNECT:
			*PToolBaseNameP = _Msg_ConnectStr;
			if(PAcceptModelsP) *PAcceptModelsP = TRUE;
			if(PSelectedNodes_Count == 0)
			{
				if(PSelectedObjects_Count == 1) return(_Msg_ConnectObjectStr);
				else if(PSelectedObjects_Count > 1) return(_Msg_ConnectObjectsStr);
			}
			else if(PSelectedNodes_Count == 1) return(_Msg_ConnectModelStr);
			else if(PSelectedNodes_Count > 1) return(_Msg_ConnectModelsStr);
		return(NULL);

		case E3dConnectREPARENT:
			*PToolBaseNameP = _Msg_ReparentStr;
			if(PAcceptModelsP) *PAcceptModelsP = TRUE;
			if(PSelectedNodes_Count == 0)
			{
				if(PSelectedObjects_Count == 1) return(_Msg_ReparentObjectStr);
				else if(PSelectedObjects_Count > 1) return(_Msg_ReparentObjectsStr);
			}
			else if(PSelectedNodes_Count == 1) return(_Msg_ReparentModelStr);
			else if(PSelectedNodes_Count > 1) return(_Msg_ReparentModelsStr);
		return(_Msg_ReparentStr);

		case E3dConnectINSTANTIATEAS:
			*PToolBaseNameP = _Msg_InstantiateAsStr;
			if(PAcceptModelsP) *PAcceptModelsP = FALSE;
		return(_Msg_InstantiateAsStr);

		default:
			*PToolBaseNameP = NULL;
			if(PAcceptModelsP) *PAcceptModelsP = TRUE;
		return(NULL);
	}
}



static XeString _ModelPickParent(const char* PToolName, int* PButtonActions)
{
	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
	const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

	 XeString  LXeString = XeString_NewVa(TRUE,
		XeFONT, XeFontDEFAULT,  XeCOLOR_RGBA, LLabelRGBAp8Color,  XeTEXT, PToolName,  XeTEXT, " - ",
		E3dp_MMouseButtonS,  XeTEXT, ": pick parent Model  ",
		XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
		XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
		0);
	PButtonActions[1] = E3dBA_PICK_PARENT_NODE_FOR_NODES;
	return(LXeString);
}


static XeString _ModelPickChild(const char* PToolName, int* PButtonActions)
{
	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
	const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

	XeString  LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, PToolName,  XeTEXT, " - ",
												 E3dp_LMouseButtonS,  XeTEXT, ": pick child Model  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
	PButtonActions[0] = E3dBA_PICK_CHILD_NODE_FOR_NODE;
	return(LXeString);
}


static XeString _ModelPickChildOrParent(const char* PToolName, int* PButtonActions)
{
	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
	const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

	XeString  LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, PToolName,  XeTEXT, " - ",
												 E3dp_LMouseButtonS,  XeTEXT, ": pick child Model  ",
												 E3dp_MMouseButtonS,  XeTEXT, ": pick parent Model  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
	PButtonActions[0] = E3dBA_PICK_CHILD_NODE_FOR_NODE;
	PButtonActions[1] = E3dBA_PICK_PARENT_NODE_FOR_NODES;
	return(LXeString);
}


static XeString _InstantiateAsObjectPicked(const char* PToolName, const E3dItem* PPickedItems,  E3dObject** PSelectedObjects, const EIndex PSelectedObjects_Count,  int* PButtonActions)
{
	const E3dObjItem*  LObjItem = (const E3dObjItem*)PPickedItems;
	XeString           LXeString = NULL;
	E3dObject*         LObject = LObjItem->Object;
	const E3dObject*   LSelectedObject = PSelectedObjects[0];

	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
	const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;
	const uint32_t    LObjectNameRGBAp8Color = LPrefs->ObjectNameRGBAp8Color;

	switch(PSelectedObjects_Count)
	{
		case 0:
			LXeString = XeString_NewVa(TRUE,
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, PToolName,  XeTEXT, " - Move pointer over an Object ",
			 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
			 0);
		break;

		case 1:
			{
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, PToolName, XeTEXT, " - ",
				 E3dp_LMouseButtonS, XeTEXT, " Add selected Object <",  XeCOLOR_RGBA, LObjectNameRGBAp8Color, XeTEXT, LSelectedObject->Name ? LSelectedObject->Name : "",  XeCOLOR_RGBA, LLabelRGBAp8Color, XeTEXT, "> to all Models of <",  XeCOLOR_RGBA, LObjectNameRGBAp8Color, XeTEXT, LObject->Name ? LObject->Name : "",  XeCOLOR_RGBA, LLabelRGBAp8Color,  XeTEXT, ">  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
				PButtonActions[0] = E3dBA_PICK_OBJECT_FOR_ADDOBJECTTO_MODELSOF;
			}
		break;

		default:
			{
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, PToolName, XeTEXT, " - ",
				 E3dp_LMouseButtonS,  XeTEXT, " Add selected Objects to all Models of <",  XeCOLOR_RGBA, LObjectNameRGBAp8Color, XeTEXT, LObject->Name ? LObject->Name : "",  XeCOLOR_RGBA, LLabelRGBAp8Color, XeTEXT, ">  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
				PButtonActions[0] = E3dBA_PICK_OBJECT_FOR_ADDOBJECTTO_MODELSOF;
			}
		break;
	}

	return(LXeString);
}


//======================================================================================
// Get actions and status-panel strint for SCHEMATICS view
//======================================================================================
static EChangeFlags _ConnectGetButtonActionsSchematicsView(const E3dItem* PPickedItems,  const int PNumOfPickedItems, E3dNode** PSelectedNodes, const int PSelectedNodes_Count,  E3dObject** PSelectedObjects, const int PSelectedObjects_Count,  int* PButtonActions, XeString* PXeStringRet)
{
	const E3dObjItem*   LObjItem = (const E3dObjItem*)PPickedItems;
	XeString            LXeString = NULL;
	EChangeFlags        LChanged = 0;
	EBool               LAcceptModels = FALSE;
	const char*         LToolBaseName;const char*  LToolName = _GetToolName(_ConnectMode, PSelectedNodes_Count, PSelectedObjects_Count,  &LToolBaseName, &LAcceptModels);

	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color = Xe_ForegroundRGBA8pColor,
	                  LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

	if(PNumOfPickedItems)
	{
		switch(LObjItem->Type)
		{
			case E3dItemMODEL:
				if(LAcceptModels)
				{
					E3dNode*   LPickedNode = LObjItem->Node;
					E3dModel*  LPickedModel = E_DynamicCast(LPickedNode, E3dModel);if(LPickedModel)
					{
						switch(PSelectedNodes_Count)
						{
							case 0:
								if(PSelectedObjects_Count == 1)
								{
									const E3dObject*  LSelectedObject = PSelectedObjects[0];
									if(E3dModel_GetIndexOfObject(LPickedModel, PSelectedObjects[0]) > -1)
									{
										LXeString = XeString_NewVa(TRUE,
										 XeFONT, XeFontDEFAULT,                                              XeTEXT, LToolName,  XeTEXT, " - The selected Object ",
										 E3dp_ObjectS(LSelectedObject->Name ? LSelectedObject->Name : ""), " is already on this Model ",
										 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,     XeTEXT, "Esc",
										 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,          XeTEXT, ": cancel",
										 0);
									}
									else
									{
										LXeString = XeString_NewVa(TRUE,
										 XeFONT, XeFontDEFAULT,                                            XeTEXT, LToolName,  XeTEXT, " - ",
										 E3dp_LMouseButtonS,  XeTEXT, ": add selected Object ",  E3dp_ObjectS(LSelectedObject->Name ? LSelectedObject->Name : ""),
										                                                                   XeTEXT, " to this Model  ",
										 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
										 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
										 0);
										PButtonActions[0] = E3dBA_PICK_PARENT_MODEL_FOR_OBJECTS;
									}
								}
								else
								{
									LXeString = XeString_NewVa(TRUE,
									 XeFONT, XeFontDEFAULT,                                            XeTEXT, LToolName,  XeTEXT, " - ",
									 E3dp_LMouseButtonS,  XeFTEXT, ": add %lld selected Objects to this Model  ", (long long)PSelectedObjects_Count,
									 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
									 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
									 0);
									PButtonActions[0] = E3dBA_PICK_PARENT_MODEL_FOR_OBJECTS;
								}
							break;

							case 1:  // One Model is selected
								{
									E3dNode*  LSelectedNode = PSelectedNodes[0];
									if(LSelectedNode == LPickedNode)
									{
										LXeString = XeString_NewVa(TRUE,
										 XeFONT, XeFontDEFAULT,                                           XeTEXT, LToolName,  XeTEXT, " - Pick a different Model   ",
										 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
										 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
										 0);
									}
									else
									{
										if(LPickedNode == LSelectedNode->Parent)
										{
											LXeString = XeString_NewVa(TRUE,
											 XeFONT, XeFontDEFAULT,                                           XeTEXT, LToolName,  XeTEXT, " - This Model is already the parent of the selected Model ",
											                           E3dp_NodeS(LSelectedNode->Name ? LSelectedNode->Name : ""), XeTEXT, "  ",
											 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
											 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
											 0);
										}
										else
										{
											const EIndex  LConnectionLevel1 = E3dNode_HrcIsDescendantOf(LPickedNode, LSelectedNode);
											if(LConnectionLevel1 != E_NOT_CONNECTED)
											{
												if(LPickedNode->Parent == LSelectedNode)
												{
													LXeString = XeString_NewVa(TRUE,
													 XeFONT, XeFontDEFAULT,                                            XeTEXT, LToolName,  XeTEXT, " - This Model is already the child of the selected Model ",
													                           E3dp_NodeS(LSelectedNode->Name ? LSelectedNode->Name : ""), XeTEXT, "  ",
													 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,   XeTEXT, "Esc",
													 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,        XeTEXT, ": cancel",
													 0);
												}
												else
												{
													LXeString = _ModelPickChild(LToolName, PButtonActions);
												}
											}
											else
											{
												const EIndex  LConnectionLevel2 = E3dNode_HrcIsDescendantOf(LSelectedNode, LPickedNode);
												if(LConnectionLevel2 >= 2)
												{
													LXeString = _ModelPickParent(LToolName, PButtonActions);
												}
												else
												{
													if(LPickedNode->Parent == NULL && LSelectedNode->Parent == NULL)
													{
														LXeString = _ModelPickChildOrParent(LToolName, PButtonActions);
													}
													else if(LSelectedNode->Parent == NULL)
													{
														LXeString = _ModelPickChildOrParent(LToolName, PButtonActions);
													}
													else if(LPickedNode->Parent == NULL)
													{
														LXeString = _ModelPickChildOrParent(LToolName, PButtonActions);
													}
													else
													{
														LXeString = _ModelPickChildOrParent(LToolName, PButtonActions);
													}
												}
											}
										}
									}
								}
							break;

							default:  // Multiple Models are selected
								{
									EListI(, E3dNode*, LSelectedBranches);LSelectedBranches_Count = E3dNodeHrc_CollectIndependentBranches(PSelectedNodes, PSelectedNodes_Count,  &LSelectedBranches);if(LSelectedBranches)
									{
										if(LSelectedBranches_Count == 1)
										{
											const E3dNode*  LSelectedBranch = LSelectedBranches[0];
											if(E3dNode_HrcIsDescendantOf(LPickedNode, LSelectedBranch) == E_NOT_CONNECTED)
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,                                           XeTEXT, LToolBaseName,  XeTEXT, " - ",
												 E3dp_MMouseButtonS,  XeTEXT, ": make this Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),
												                           XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, " the new parent of the selected branch ",
												                           E3dp_NodeS(LSelectedBranch->Name ? LSelectedBranch->Name : ""),  XeTEXT, "  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
												PButtonActions[1] = E3dBA_PICK_PARENT_NODE_FOR_NODES;
											}
											else
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,                                            XeTEXT, LToolName,  XeTEXT, " - ",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "This Model is in the selected branch  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
											}
										}
										else if(LSelectedBranches_Count > 1)
										{
											EBool  LPickedNodeValid = TRUE;

											{
												EIndex  LBC = 0;for(;LBC < LSelectedBranches_Count;LBC++)
												{
													const E3dNode*  LSelectedBranch = LSelectedBranches[LBC];const EIndex  LDescendantLevel = E3dNode_HrcIsDescendantOf(LPickedNode, LSelectedBranch);if(LDescendantLevel != E_NOT_CONNECTED)
													{
														if(LDescendantLevel == 0)
														{
															LXeString = XeString_NewVa(TRUE,
															 XeFONT, XeFontDEFAULT,                                            XeTEXT, LToolName,  XeTEXT, " - ",
															 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,        XeTEXT, "This Model ", 
															                           E3dp_NodeS(LSelectedBranch->Name ? LSelectedBranch->Name : ""), 
															                           XeCOLOR_RGBA, LLabelRGBAp8Color,        XeTEXT, " is one of the selected Models  ",
															 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,   XeTEXT, "Esc",
															 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,        XeTEXT, ": cancel",
															 0);
														}
														else
														{
															LXeString = XeString_NewVa(TRUE,
															 XeFONT, XeFontDEFAULT,                                           XeTEXT, LToolName,  XeTEXT, " - ",
															 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "This Model is inside a selected branch ",
															                           E3dp_NodeS(LSelectedBranch->Name ? LSelectedBranch->Name : ""),  XeTEXT, "  ",
															 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
															 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
															 0);
														}
														LPickedNodeValid = FALSE;
														break;
													}
												}
											}

											if(LPickedNodeValid)
											{
												EBool  LPassFound = FALSE;EIndex  LBC = 0;for(;LBC < LSelectedBranches_Count;LBC++)
												{
													const E3dNode*  LSelectedBranch = LSelectedBranches[LBC];const EIndex  LDescendantLevel = E3dNode_HrcIsDescendantOf(LSelectedBranch, LPickedNode);if(LDescendantLevel == E_NOT_CONNECTED || LDescendantLevel > 1)
													{
														LPassFound = TRUE;
														break;
													}
												}
												if(!LPassFound) LPickedNodeValid = FALSE;
											}

											if(LPickedNodeValid)
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,                                           XeTEXT, LToolName,  XeTEXT, " - ",
												 E3dp_MMouseButtonS,  XeTEXT, ": pick new parent Model for selected Models  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
												PButtonActions[1] = E3dBA_PICK_PARENT_NODE_FOR_NODES;
											}
										}

										EFree(LSelectedBranches);
									}
								}
							break;
						}
						if(PButtonActions[0] || PButtonActions[1] || PButtonActions[2]) { LPickedModel->Highlighted = TRUE;LChanged |= E3dChgSELECTION; }
					}
				}
			break;

			default:
				{
					E3dObject*  LObject = LObjItem->Object;

					if(_ConnectMode == E3dConnectINSTANTIATEAS) LXeString = _InstantiateAsObjectPicked(LToolName, PPickedItems,  PSelectedObjects, PSelectedObjects_Count, PButtonActions);
					else
					{
						if(PSelectedObjects_Count) LXeString = _InstantiateAsObjectPicked(LToolName, PPickedItems,  PSelectedObjects, PSelectedObjects_Count, PButtonActions);
						else
						{
							switch(PSelectedNodes_Count)
							{
								case 0:
									LXeString = XeString_NewVa(TRUE,
									 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - Move pointer over a Model ",
									 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
									 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
									 0);
								break;

								case 1:
									{
										E3dNode*   LNode = PSelectedNodes[0];
										E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
										{
											if(E3dModel_GetIndexOfObject(LModel, LObjItem->Object) > -1)
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - This Object is already on the selected Model ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
											}
											else
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
												 E3dp_LMouseButtonS,  XeTEXT, ": add this Object to the selected Model  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
												PButtonActions[0] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
											}
										}
									}
								break;

								default:
									LXeString = XeString_NewVa(TRUE,
									 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
									 E3dp_LMouseButtonS,  XeTEXT, ": add this Object to the selected Models  ",
									 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
									 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
									 0);
									PButtonActions[0] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
								break;
							}
						}
					}
					if(PButtonActions[0] || PButtonActions[1] || PButtonActions[2]) { LObject->Highlighted = TRUE;LChanged |= E3dChgSELECTION; }
				break;
			}
		}
	}
	else
	{
		switch(PSelectedNodes_Count)
		{
			case 0:
				if(PSelectedObjects_Count == 1)
				{
					LXeString = XeString_NewVa(TRUE,
					 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
					 E3dp_LMouseButtonS,  XeTEXT, ": pick parent Model for selected Object  ",
					 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
					 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
					 0);
				}
				else if(PSelectedObjects_Count > 1)
				{
					LXeString = XeString_NewVa(TRUE,
					 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
					 E3dp_LMouseButtonS,  XeTEXT, ": pick parent Model for selected Objects  ",
					 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
					 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
					 0);
				}
			break;

			default:
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - Move pointer over a Model  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
			break;
		}
	}


	*PXeStringRet = LXeString;

	return(LChanged);
}


//======================================================================================
// Get actions and status-panel strint for PERSPECTIVE, TOP, FRONT and RIGHT views
//======================================================================================
static EChangeFlags _ConnectGetButtonActions3DView(E3dItem* PPickedItems, int PNumOfPickedItems, E3dNode** PSelectedNodes, const int PSelectedNodes_Count, E3dObject** PSelectedObjects, const int PSelectedObjects_Count, int* PButtonActions, XeString* PXeStringRet)
{
	E3dObjItem*   LObjItem = (E3dObjItem*)PPickedItems;
	XeString      LXeString = NULL;
	EChangeFlags  LChanged = 0;
	EBool         LAcceptModels = FALSE;
	const char*   LToolBaseName;const char*   LToolName = _GetToolName(_ConnectMode, PSelectedNodes_Count, PSelectedObjects_Count,  &LToolBaseName, &LAcceptModels);

	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
	const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

// On these views, we'll always get a detailed pick (e.g. Model->Mesh->PolyGroup)
//
	if(PNumOfPickedItems)
	{
		switch(LObjItem->Type)
		{
			case E3dItemMODEL:
			case E3dItemPOLYGON:
			case E3dItemSPLINE:
			case E3dItemFACE:
				{
					E3dNode*   LPickedNode = LObjItem->Node;
					E3dModel*  LPickedModel = E_DynamicCast(LPickedNode, E3dModel);if(LPickedModel)
					{
						switch(PSelectedNodes_Count)
						{
							case 0:
								if(PSelectedObjects_Count == 1)
								{
									if(E3dModel_GetIndexOfObject(LPickedModel, PSelectedObjects[0]) > -1)
									{
										LXeString = XeString_NewVa(TRUE,
										 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - The selected Object is already on this Model  ",
										 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
										 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
										 0);
									}
									else
									{
										LXeString = XeString_NewVa(TRUE,
										 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
										 E3dp_LMouseButtonS,  XeTEXT, ": pick parent Model for Object  ",
										 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
										 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
										 0);
										PButtonActions[0] = E3dBA_PICK_PARENT_MODEL_FOR_OBJECTS;
									}
								}
								else
								{
									LXeString = XeString_NewVa(TRUE,
									 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
									 E3dp_LMouseButtonS,  XeTEXT, ": pick parent Model for Objects  ",
									 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
									 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
									 0);
									PButtonActions[0] = E3dBA_PICK_PARENT_MODEL_FOR_OBJECTS;
								}
							break;

							case 1:
								{
									E3dNode*    LSelectedNode = PSelectedNodes[0];E3dModel*  LSelectedModel = E_DynamicCast(LSelectedNode, E3dModel);
									E3dObject*  LPickedObject = LObjItem->Object;
									if(LSelectedNode == LPickedNode)
									{
										LXeString = XeString_NewVa(TRUE,
										 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - Pick a different Model or Object   ",
										 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
										 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
										 0);
									}
									else if(LPickedNode == LSelectedNode->Parent)
									{
										if(E3dModel_GetIndexOfObject(LSelectedModel, LPickedObject) > -1)
										{
											LXeString = XeString_NewVa(TRUE,
											 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - This Model ",
											  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),
											                                                                   XeTEXT, " is already the parent of selected Model ",
											  E3dp_NodeS(LSelectedNode->Name ? LSelectedNode->Name : ""),
											                                                                   XeTEXT, " and this Object ",
											  E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""),
											                                                                   XeTEXT, " is on the selected Model  ",
											 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
											 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
											 0);
										}
										else
										{
											LXeString = XeString_NewVa(TRUE,
											 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,         XeTEXT, LToolName,  XeTEXT, " - ",
											 E3dp_LMouseButtonS,  XeTEXT, ": add Object ",  E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""),  XeTEXT, " to selected Model ",
											                   E3dp_NodeS(LSelectedNode->Name ? LSelectedNode->Name : ""),
											 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,   XeTEXT, "   Esc",
											 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,        XeTEXT, ": cancel",
											 0);
											PButtonActions[0] = E3dBA_PICK_PARENT_MODEL_FOR_OBJECTS;
										}
									}
									else if(LSelectedNode == LPickedNode->Parent)
									{
										if(E3dModel_GetIndexOfObject(LSelectedModel, LPickedObject) > -1)
										{
											LXeString = XeString_NewVa(TRUE,
											 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,
											                                                                   XeTEXT, " - This Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),
											                                                                   XeTEXT, " is a child of selected Model ",  E3dp_NodeS(LSelectedNode->Name ? LSelectedNode->Name : ""),
											                                                                   XeTEXT, " and this Object ",E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""),
											                                                                   XeTEXT, " is on the selected Model  ",
											 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
											 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
											 0);
										}
										else
										{
											LXeString = XeString_NewVa(TRUE,
											 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
											 E3dp_LMouseButtonS,  XeTEXT, ": add this Object ",  E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""),  XeTEXT, " to the selected Model ",  E3dp_NodeS(LSelectedNode->Name ? LSelectedNode->Name : ""),
											 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "  Esc",
											 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
											 0);
											PButtonActions[0] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
										}
									}
									else
									{
										const EIndex  LConnectionLevel1 = E3dNode_HrcIsDescendantOf(LPickedNode, LSelectedNode);

										if(LConnectionLevel1 > 1)
										{
											if(E3dModel_GetIndexOfObject(LSelectedModel, LPickedObject) > -1)
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,
												                                                                   XeTEXT, " - This Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),
												                                                                   XeTEXT, " is a descendant of selected Model ",  E3dp_NodeS(LSelectedNode->Name ? LSelectedNode->Name : ""),
												                                                                   XeTEXT, " and this Object ",  E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""),
												                                                                   XeTEXT, " is on the selected Model",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "  Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
											}
											else
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,        XeTEXT, LToolName,
												                                                                   XeTEXT, " - This Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),
												                                                                   XeTEXT, " is a descendant of selected Model ",  E3dp_NodeS(LSelectedNode->Name ? LSelectedNode->Name : ""),
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "  Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
											}
										}
										else
										{
											if(LPickedNode->Parent == NULL && LSelectedNode->Parent == NULL)
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
												 E3dp_LMouseButtonS,  XeTEXT, ": pick child Model  ",
												 E3dp_MMouseButtonS,  XeTEXT, ": pick parent Model  ",
												 E3dp_RMouseButtonS,  XeTEXT, ": pick child Object  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
												PButtonActions[0] = E3dBA_PICK_CHILD_NODE_FOR_NODE;
												PButtonActions[1] = E3dBA_PICK_PARENT_NODE_FOR_NODES;
												PButtonActions[2] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
											}
											else if(LSelectedNode->Parent == NULL)
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
												 E3dp_MMouseButtonS,  XeTEXT, ": pick parent Model  ",
												 E3dp_RMouseButtonS,  XeTEXT, ": pick child Object  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
												PButtonActions[1] = E3dBA_PICK_PARENT_NODE_FOR_NODES;
												PButtonActions[2] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
											}
											else if(LPickedNode->Parent == NULL)
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
												 E3dp_LMouseButtonS,  XeTEXT, ": pick child Model  ",
												 E3dp_RMouseButtonS,  XeTEXT, ": pick child Object  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
												PButtonActions[0] = E3dBA_PICK_CHILD_NODE_FOR_NODE;
												PButtonActions[2] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
											}
											else
											{
												if(E3dModel_GetIndexOfObject(LSelectedModel, LObjItem->Object) > -1)
												{
													LXeString = XeString_NewVa(TRUE,
													 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - Selected Model ",  E3dp_NodeS(LSelectedModel->Name ? LSelectedModel->Name : ""),  XeTEXT, " and this Model are already in hierarchies ",
													 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
													 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
													 0);
												 }
												 else
												 {
printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
													 LXeString = XeString_NewVa(TRUE,
														XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
														E3dp_RMouseButtonS,  XeTEXT, ": pick child Object  ",
														XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
														XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
														0);
													PButtonActions[2] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
												}
											}
										}
									}
								}
							break;

							default:  // Multiple Models are selected
								{
									E3dObject*  LPickedObject = LObjItem->Object;

									EListI(, E3dNode*, LSelectedBranches);LSelectedBranches_Count = E3dNodeHrc_CollectIndependentBranches(PSelectedNodes, PSelectedNodes_Count,  &LSelectedBranches);if(LSelectedBranches)
									{
										if(LSelectedBranches_Count == 1)
										{
// A single branch is selected
//
											const E3dNode*  LSelectedBranch = LSelectedBranches[0];
											if(E3dNode_HrcIsDescendantOf(LPickedNode, LSelectedBranch) == E_NOT_CONNECTED)
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolBaseName,  XeTEXT, " - ",
												 E3dp_MMouseButtonS,  XeTEXT, ": make this Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),   XeTEXT, " the parent of the selected branch ",  E3dp_NodeS(LSelectedBranch->Name ? LSelectedBranch->Name : ""),  XeCOLOR_RGBA, LLabelRGBAp8Color, XeTEXT, "  ",
												 E3dp_RMouseButtonS,  XeTEXT, ": add Object ",  E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""),  XeTEXT, " to the seleceted Models  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
												PButtonActions[1] = E3dBA_PICK_PARENT_NODE_FOR_NODES;
												PButtonActions[2] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
											}
											else
											{
												LXeString = XeString_NewVa(TRUE,
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "This Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),  XeTEXT, " is in the selected branch ",  E3dp_NodeS(LSelectedBranch->Name ? LSelectedBranch->Name : ""),  XeTEXT, "  ",
												 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
												 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
												 0);
											}
										}
										else
										{
// Multiple branches are selected
//
											EBool  LPickedNodeValid = TRUE;

											{
												EIndex  LBC = 0;for(;LBC < LSelectedBranches_Count;LBC++)
												{
													const E3dNode*  LSelectedBranch = LSelectedBranches[LBC];const EIndex  LDescendantLevel = E3dNode_HrcIsDescendantOf(LPickedNode, LSelectedBranch);if(LDescendantLevel != E_NOT_CONNECTED)
													{
														if(LDescendantLevel == 0)
														{
															LXeString = XeString_NewVa(TRUE,
															 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
															 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "This Model ",  E3dp_NodeS(LSelectedBranch->Name ? LSelectedBranch->Name : ""),
															                                                                   XeTEXT, " is one of the selected Models",
															 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "  Esc",
															 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
															 0);
														}
														else
														{
															LXeString = XeString_NewVa(TRUE,
															 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
															 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "This Model is inside selected branch ",  E3dp_NodeS(LSelectedBranch->Name ? LSelectedBranch->Name : ""),
															 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "  Esc",
															 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
															 0);
														}
														LPickedNodeValid = FALSE;
														break;
													}
												}
											}

											if(LPickedNodeValid)
											{
												EBool  LPassFound = FALSE;EIndex  LBC = 0;for(;LBC < LSelectedBranches_Count;LBC++)
												{
													const E3dNode*  LSelectedBranch = LSelectedBranches[LBC];const EIndex  LDescendantLevel = E3dNode_HrcIsDescendantOf(LSelectedBranch, LPickedNode);if(LDescendantLevel == E_NOT_CONNECTED || LDescendantLevel > 1)
													{
														LPassFound = TRUE;
														break;
													}
												}
												if(!LPassFound) LPickedNodeValid = FALSE;
											}

											EBool  LPickedObjectValid = FALSE;

											{
												EIndex  LBC = 0;for(;LBC < LSelectedBranches_Count;LBC++)
												{
													const E3dNode*  LSelectedBranch = LSelectedBranches[LBC];const E3dNode*  LNodeAfterBranch = E3dNode_BranchGetNodeAfter(LSelectedBranch);const E3dNode*  LNode = LSelectedBranch;for(;LNode != LNodeAfterBranch;LNode = LNode->Next)
													{
														if(LNode->Selection)
														{
															E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
															{
																if(EList_PointerGetIndex(LModel->Objects, LPickedObject) == E_NOT_FOUND)
																{
																	LPickedObjectValid = TRUE;
																	break;
																}
															}
														}
													}
												}
											}

											if(LPickedNodeValid || LPickedObjectValid)
											{
												if(LPickedNodeValid && LPickedObjectValid)
												{
													LXeString = XeString_NewVa(TRUE,
													 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
													 E3dp_MMouseButtonS, XeTEXT, ": make Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),  XeTEXT, " the new parent of the selected Models  ",
													 E3dp_MMouseButtonS, XeTEXT, ": add Object ",  E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""), XeTEXT, " to the selected Models",
													 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "  Esc",
													 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
													 0);
													PButtonActions[1] = E3dBA_PICK_PARENT_NODE_FOR_NODES;
													PButtonActions[2] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
												}
												else if(LPickedNodeValid)
												{
													LXeString = XeString_NewVa(TRUE,
													 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
													 E3dp_MMouseButtonS,  XeTEXT, ": make Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),  XeTEXT, " the new parent of the selected Models  ",
													 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
													 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
													 0);
													PButtonActions[1] = E3dBA_PICK_PARENT_NODE_FOR_NODES;
												}
												else if(LPickedObjectValid)
												{
													LXeString = XeString_NewVa(TRUE,
													 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
													 E3dp_MMouseButtonS,  XeTEXT, ": add Object ",  E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""),  XeTEXT, " to the selected Models",
													 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "  Esc",
													 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
													 0);
													PButtonActions[2] = E3dBA_PICK_CHILD_OBJECT_FOR_MODELS;
												}
											}
										}

										EFree(LSelectedBranches);
									}
								}
							break;
						}
						if(PButtonActions[0] || PButtonActions[1] || PButtonActions[2]) { LPickedModel->Highlighted = TRUE;LChanged |= E3dChgSELECTION; }
					}
				break;
			}
		}
	}
	else
	{
		switch(PSelectedNodes_Count)
		{
			case 0:
				if(PSelectedObjects_Count == 1)
				{
					LXeString = XeString_NewVa(TRUE,
					 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
					 E3dp_LMouseButtonS,  XeTEXT, ": pick parent Model for selected Object  ",
					 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
					 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
					 0);
				}
				else if(PSelectedObjects_Count > 1)
				{
					LXeString = XeString_NewVa(TRUE,
					 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - ",
					 E3dp_LMouseButtonS,  XeTEXT, ": pick parent Model for selected Objects  ",
					 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
					 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
					 0);
				}
			break;

			default:
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - Move pointer over a Model  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
			break;
		}
	}


	*PXeStringRet = LXeString;

	return(LChanged);
}


//================================================================
// Add a child Node to a given Node
//================================================================
static void _NodeHrcConnectChild(E3dNode* PParent, E3dNode* PChild, const EIndex PWhere, const EBool PCompensateTransform, const EBool PUpdate)
{
	if(PChild->Parent == NULL)
	{
		E3dScene*  LScene = E3d_Scene;

// Remove this Node / Branch from the Scene as an individual hierarchy, but don't free its RenderData etc.
// That is done in E3dScene_NodeHrcRemove()
//
		E_Object(E3dNode, PChild, LTNode);  // Make sure ObjectRemoveA doesn't free LChild
		ELst_ObjectRemoveA_om(LScene,RootNodes, PChild);

		E3dNode_ConnectChild(PParent, PChild, PWhere, PCompensateTransform, PUpdate);
		E_DeleteChk(LTNode);
	}
}


//========================================
// Initialize EditHierarchy Operator
//========================================
static void _EditHierarchy_Init(EOperator* POperator)
{
	E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)POperator;

	EList_Init(LOperator->ObjectMoves);
	EList_Init(LOperator->BranchMoves);
}


//========================================
// Free ObjectsEdit Operator
//========================================
static int _EditHierarchy_Free(EOperator* POperator)
{
	E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)POperator;

	if(LOperator->ObjectMoves)
	{
		EIndex  LC = 0;const EIndex  LN = LOperator->ObjectMoves_Count;E3dObjectMove*  LObjectMove = LOperator->ObjectMoves;for(;LC < LN;LC++, LObjectMove++)
		{
			if(LObjectMove->AddToModels)
			{
				EIndex  LMC = 0;const EIndex  LMN = LObjectMove->AddToModels_Count;EObjectAndIndex*  LObjecAndIndex = LObjectMove->AddToModels;for(;LMC < LMN;LMC++, LObjecAndIndex++)
				{
					E_DeleteChk(LObjecAndIndex->Object);
				}

				EList_Free(LObjectMove->AddToModels);
			}

			if(LObjectMove->DelFrModels)
			{
				EIndex  LMC = 0;const EIndex  LMN = LObjectMove->DelFrModels_Count;EObjectAndIndex*  LObjecAndIndex = LObjectMove->DelFrModels;for(;LMC < LMN;LMC++, LObjecAndIndex++)
				{
					E_DeleteChk(LObjecAndIndex->Object);
				}

				EList_Free(LObjectMove->DelFrModels);
			}
			E_DeleteChk(LObjectMove->Object);
		}

		EList_FreeChk(LOperator->ObjectMoves);
	}

	if(LOperator->BranchMoves)
	{
		EIndex  LC = 0;const EIndex  LN = LOperator->BranchMoves_Count;E3dBranchMove*  LBranchMove = LOperator->BranchMoves;for(;LC < LN;LC++, LBranchMove++)
		{
			E_DeleteChk(LBranchMove->Node);
			E_DeleteChk(LBranchMove->OldParent);
			E_DeleteChk(LBranchMove->NewParent);
		}

		EList_Free(LOperator->BranchMoves);
	}


	return(E_SUCCESS);
}


//================================================
// EditObjects Operator proc
//================================================
static EChangeFlags _EditHierarchy_Proc(EOps* POps, EOperator* POperator, const int POperation)
{
	E3dScene*                  LScene = E3d_Scene;
	E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)POperator;
	EChangeFlags               LChanged = 0;
	EListAI(, E3dNode*, LMaterialInheritanceChangedBranches);
	EListAI(, E3dNode*, LHrcChangedBranches);

	switch(POperation)
	{
		case EOpDO:
			{
				if(LOperator->ObjectMoves)
				{
					EIndex  LC = 0;const EIndex  LN = LOperator->ObjectMoves_Count;E3dObjectMove*  LObjectMove = LOperator->ObjectMoves;for(;LC < LN;LC++, LObjectMove++)
					{
						E3dObject*  LObject = LObjectMove->Object;

//printf("%s() %s:%d <%s>\n", __FUNCTION__, __FILE__, __LINE__, LObject->Name);fflush(stdout);
						{
							EIndex  LMC = 0;const EIndex  LMN = LObjectMove->AddToModels_Count;EObjectAndIndex*  LAddToModel = LObjectMove->AddToModels;for(;LMC < LMN;LMC++, LAddToModel++)
							{
//printf("%s() %s:%d  Add [%s]\n", __FUNCTION__, __FILE__, __LINE__, LAddToModel->Object->Name);fflush(stdout);
								E3dModel*  LModel = (E3dModel*)LAddToModel->Object;
								if(E3dModel_AppendObject(LModel, LObject)) LChanged |= (E3dChgHIERARCHY|E3dObject_GetChangeFlagsByClass(LObject));

								E3dNode*  LRootNode = (E3dNode*)LModel;while(LRootNode->Parent) LRootNode = LRootNode->Parent;
								EList_PointerAddAChk(LMaterialInheritanceChangedBranches, 8, LRootNode);
							}
						}

						{
//printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
							EIndex  LMC = 0;const EIndex  LMN = LObjectMove->DelFrModels_Count;EObjectAndIndex*  LDelFrModel = LObjectMove->DelFrModels;for(;LMC < LMN;LMC++, LDelFrModel++)
							{
//printf("%s() %s:%d  Del [%s]\n", __FUNCTION__, __FILE__, __LINE__, LDelFrModel->Object->Name);fflush(stdout);
								E3dModel*  LModel = (E3dModel*)LDelFrModel->Object;
								E3dModel_RemoveObject(LModel, LObject);
								LChanged |= (E3dChgHIERARCHY|E3dObject_GetChangeFlagsByClass(LObject));

								E3dNode*  LRootNode = (E3dNode*)LModel;while(LRootNode->Parent) LRootNode = LRootNode->Parent;
								EList_PointerAddAChk(LMaterialInheritanceChangedBranches, 8, LRootNode);
							}
						}
					}
				}

				if(LOperator->BranchMoves)
				{
					E3dNode**  LRoots;ENewNClr(LOperator->BranchMoves_Count, LRoots);

					{
						EIndex  LBC = 0;E3dBranchMove*  LBranchMove = LOperator->BranchMoves;for(;LBC < LOperator->BranchMoves_Count;LBC++, LBranchMove++)
						{
							if(LBranchMove->NewParent == NULL)
							{
								E3dNode*  LRoot = LBranchMove->Node;while(LRoot->Parent) LRoot = LRoot->Parent;
								LRoots[LBC] = LRoot;
							}
						}
					}

					{
						EIndex  LBC = 0;E3dBranchMove*  LBranchMove = LOperator->BranchMoves;E3dNode*  LPrevRoot = NULL;EIndex  LSceneRootOffset = 0;for(;LBC < LOperator->BranchMoves_Count;LBC++, LBranchMove++)
						{
							E3dNode*  LNode = LBranchMove->Node;
							E3dNode*  LTNode = NULL;E_Assign(LNode, LTNode);  // Temporary assignment, so LSelectedBranch doesn't get freed by E3dNode_DisconnectFromParent() (we don't add LSelectedBranch to the Scene...) or E3dScene_NodeHrcRemove()
							if(LNode->Parent)
							{
								E3dNode*  LRootNode = LNode->Parent;while(LRootNode->Parent) LRootNode = LRootNode->Parent;

								E3dNode_DisconnectFromParent(LNode, _TransformCompensation);

								E3dNodeHrc_UpdateConnections(LRootNode);
								E3dNodeHrc_UpdateMatrices(LRootNode, FALSE);
							}
							else
							{
								E3dScene_NodeHrcRemove(LScene, LNode, NULL, FALSE);
							}

							LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNode) | E3dChgHIERARCHY;

							if(LBranchMove->NewParent)
							{
								_NodeHrcConnectChild(LBranchMove->NewParent, LNode, ELstEND, _TransformCompensation, FALSE);
								E3dNode*  LRootNode = LBranchMove->NewParent;while(LRootNode->Parent) LRootNode = LRootNode->Parent;

								E3dNodeHrc_UpdateConnections(LRootNode);
								E3dNodeHrc_UpdateMatrices(LRootNode, FALSE);
								EList_PointerAddAChk(LHrcChangedBranches, 8, LRootNode);
							}
							else
							{
								E3dNode*  LRoot = LRoots[LBC];const EIndex  LRootIndex = EList_PointerGetIndex(LScene->RootNodes, LRoot) + 1;
								if(LRoot == LPrevRoot) LSceneRootOffset++;
								else LSceneRootOffset = 0;

								LPrevRoot = LRoot;

								E3dNodeHrc_UpdateConnections(LNode);
								E3dNodeHrc_UpdateMatrices(LNode, FALSE);
								EList_PointerAddAChk(LHrcChangedBranches, 8, LNode);
								E3dScene_NodeHrcInsert(LScene, LNode, LRootIndex + LSceneRootOffset, FALSE);
							}

							E_DeleteChk(LTNode);
						}
					}

					EFreeChk(LRoots);
				}
			}
		break;

		case EOpUNDO:
			{
				if(LOperator->ObjectMoves)
				{
					EIndex  LC = 0;const EIndex  LN = LOperator->ObjectMoves_Count;E3dObjectMove*  LObjectMove = LOperator->ObjectMoves;for(;LC < LN;LC++, LObjectMove++)
					{
						E3dObject*  LObject = LObjectMove->Object;

						{
							EIndex  LMC = 0;const EIndex  LMN = LObjectMove->DelFrModels_Count;EObjectAndIndex*  LDelFrModel = LObjectMove->DelFrModels;for(;LMC < LMN;LMC++, LDelFrModel++)
							{
								E3dModel*  LModel = (E3dModel*)LDelFrModel->Object;
								if(E3dModel_AppendObject(LModel, LObject)) LChanged |= (E3dChgHIERARCHY|E3dObject_GetChangeFlagsByClass(LObject));

								E3dNode*  LRootNode = (E3dNode*)LModel;while(LRootNode->Parent) LRootNode = LRootNode->Parent;
								EList_PointerAddAChk(LMaterialInheritanceChangedBranches, 8, LRootNode);
							}
						}

						{
							EIndex  LMC = 0;const EIndex  LMN = LObjectMove->AddToModels_Count;EObjectAndIndex*  LAddToModel = LObjectMove->AddToModels;for(;LMC < LMN;LMC++, LAddToModel++)
							{
								E3dModel*  LModel = (E3dModel*)LAddToModel->Object;
								E3dModel_RemoveObject(LModel, LObject);
								LChanged |= (E3dChgHIERARCHY|E3dObject_GetChangeFlagsByClass(LObject));

								E3dNode*  LRootNode = (E3dNode*)LModel;while(LRootNode->Parent) LRootNode = LRootNode->Parent;
								EList_PointerAddAChk(LMaterialInheritanceChangedBranches, 8, LRootNode);
							}
						}
					}
				}

				if(LOperator->BranchMoves)
				{
					EIndex  LBC = 0;E3dBranchMove*  LBranchMove = LOperator->BranchMoves;for(;LBC < LOperator->BranchMoves_Count;LBC++, LBranchMove++)
					{
						E3dNode*  LNode = LBranchMove->Node;
						E3dNode*  LTNode = NULL;E_Assign(LNode, LTNode);  // Temporary assignment, so LSelectedBranch doesn't get freed by E3dNode_DisconnectFromParent() (we don't add LSelectedBranch to the Scene...) or E3dScene_NodeHrcRemove()
						if(LNode->Parent)
						{
							E3dNode_DisconnectFromParent(LNode, _TransformCompensation);
						}
						else E3dScene_NodeHrcRemove(LScene, LNode, NULL, FALSE);

						LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNode) | E3dChgHIERARCHY;

						if(LBranchMove->OldParent)
						{
							_NodeHrcConnectChild(LBranchMove->OldParent, LNode, LBranchMove->OldIndex, _TransformCompensation, FALSE);
						}
						else
						{
							E3dScene_NodeHrcInsert(LScene, LNode, LBranchMove->OldIndex, FALSE);
						}

						E_DeleteChk(LTNode);

						E3dNode*  LRootNode = LNode;while(LRootNode->Parent) LRootNode = LRootNode->Parent;
						E3dNodeHrc_UpdateConnections(LRootNode);
						E3dNodeHrc_UpdateMatrices(LRootNode, FALSE);
						EList_PointerAddAChk(LHrcChangedBranches, 8, LRootNode);
					}
				}
			}
		break;
	}


	if(LMaterialInheritanceChangedBranches)
	{
		EIndex  LBC = 0;for(;LBC < LMaterialInheritanceChangedBranches_Count;LBC++) E3dNodeHrc_UpdateMaterialInheritance(LMaterialInheritanceChangedBranches[LBC]);
	}

	if(LHrcChangedBranches)
	{
		EIndex  LBC = 0;for(;LBC < LHrcChangedBranches_Count;LBC++) E3dNodeHrc_CallNodeInfoUpdateProcs(LHrcChangedBranches[LBC], E3dChgHIERARCHY, FALSE);
	}

	if(LChanged)
	{
//								E3dNodeHrc_UpdateMaterialInheritance(LPickedNode);
	}

	EList_FreeAChk(LMaterialInheritanceChangedBranches);
	EList_FreeAChk(LHrcChangedBranches);

	return(LChanged);
}


//================================================================
// Connect pick callback
//================================================================
static void _PickCB_Connect(E3dPickCallbackStruct* PPickStruct, EPointer PClientData,  const EIndex PNumOfPickedItems, E3dItem* PPickedItems)
{
	E3dScene*     LScene = E3d_Scene;
	E3dpWindow*   L3DWindow = PPickStruct->E3DWindow;
	XeString      LXeString = NULL;
	E3dNode**     LSelectedNodes = NULL;  const EIndex  LSelectedNodes_Count   = E3dScene_GetSelectedNodes(LScene,  &LSelectedNodes, NULL,  E3dNodeSelNODE|E3dNodeSelBRANCH_ROOT,   FALSE);
	E3dObject**   LSelectedObjects = NULL;const EIndex  LSelectedObjects_Count = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemALL, E3dSELECTED|E3dEXCLUDE_MODEL_SELECTION, FALSE);
	int           LButtonActions[3] = { 0, 0, 0 };
	EChangeFlags  LChanged = 0;


	if(E3dScene_UnhighlightAll(LScene, E3dSF_ALL)) LChanged = E3dChgSELECTION;

	if(PPickStruct->Reason == E3dCR_ABORT)
	{
		_ConnectMode = 0;

		E_PluginLockDel(_Plugin, &_ConnectPickLock, FALSE);
		return;
	}

	if(LSelectedNodes_Count == 0 && LSelectedObjects_Count == 0) { E3dp_PrintMessage(ELogWARNING, 0, _Msg_SelectModelOrObjectToConnect);return; }

	if(LSelectedNodes_Count > 0 && LSelectedObjects_Count > 0)
	{
		E3dp_PrintMessage(ELogWARNING, 0, _Msg_SelectOnlyModelOrObject);

		EFreeChk(LSelectedNodes);
		EFreeChk(LSelectedObjects);
		return;
	}

	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
	const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

	const E3dViewClass*  LViewClass = L3DWindow->ViewMode->Class;if(LViewClass == E3dSchematicsView_class)
	{
		LChanged |= _ConnectGetButtonActionsSchematicsView(PPickedItems, PNumOfPickedItems,  LSelectedNodes, LSelectedNodes_Count,  LSelectedObjects, LSelectedObjects_Count,  LButtonActions, &LXeString);
//printf("%s() %d  %08x\n", __FUNCTION__, PNumOfPickedItems, LChanged);fflush(stdout);
	}
	else if(LViewClass == E3dTopView_class || LViewClass == E3dFrontView_class || LViewClass == E3dRightView_class || LViewClass == E3dPerspectiveView_class)
	{
		LChanged |= _ConnectGetButtonActions3DView(PPickedItems, PNumOfPickedItems, LSelectedNodes,  LSelectedNodes_Count, LSelectedObjects,  LSelectedObjects_Count, LButtonActions,  &LXeString);
//printf("%s() %d  %08x\n", __FUNCTION__, PNumOfPickedItems, LChanged);fflush(stdout);
	}
	else
	{
		LXeString = XeString_NewVa(TRUE,
		 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Connect - Move pointer to a 3D or a schematics view  ",
		 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
		 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
		 0);
	}

	if(LXeString)
	{
		E3dp_PrintCompoundMessage(ELogMESSAGE, 0, LXeString);
		XeString_Free(LXeString);
	}

	if(LChanged == 0)
	{
		EFreeChk(LSelectedNodes);
		EFreeChk(LSelectedObjects);
		return;
	}


	if(PNumOfPickedItems)
	{
		E3dObjItem*  LObjItem = (E3dObjItem*)PPickedItems;
		int          LAction = 0;

		switch(PPickStruct->Button)
		{
			case Button1:  LAction = LButtonActions[0];break;
			case Button2:  LAction = LButtonActions[1];break;
			case Button3:  LAction = LButtonActions[2];break;
		}

		switch(LAction)
		{
			case E3dBA_PICK_CHILD_NODE_FOR_NODE:
				{
					E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)EOp_OperatorNew(E3dOperatorEditHierarchy_class);

					E3dNode*  LNewParent = LSelectedNodes[0];
					E3dNode*  LChildNode = LObjItem->Node;

					E3dp_PrintMessage(ELogMESSAGE, 0, "Picked child: [%s]", LChildNode->Name);

					ENewNClr(1, LOperator->BranchMoves);LOperator->BranchMoves_Count = 1;E3dBranchMove*  LBranchMove = LOperator->BranchMoves;
					E_AssignToMember(LChildNode, LBranchMove,Node);
					E_AssignToMember(LChildNode->Parent, LBranchMove,OldParent);
					if(LChildNode->Parent) LBranchMove->OldIndex = E3dNode_GetChildIndex(LChildNode);
					else LBranchMove->OldIndex = EList_PointerGetIndex(LScene->RootNodes, LChildNode);
					E_AssignToMember(LNewParent, LBranchMove,NewParent);

					E3dp_UndoStackAddOperator((EOperator*)LOperator, NULL);
					LChanged |= _EditHierarchy_Proc(NULL, (EOperator*)LOperator, EOpDO);

					_ConnectMode = 0;
					E3dp_PickEnd();

					E3dSel_NodeSelectBranch(LScene, LNewParent, FALSE, TRUE);
				}
			break;

			case E3dBA_PICK_PARENT_NODE_FOR_NODES:
				{
					EListI(, E3dNode*, LSelectedBranches);LSelectedBranches_Count = E3dNodeHrc_CollectIndependentBranches(LSelectedNodes, LSelectedNodes_Count,  &LSelectedBranches);if(LSelectedBranches)
					{
						E3dNode*  LPickedNode = LObjItem->Node;

						E3dp_PrintMessage(ELogMESSAGE, 3000, "Picked parent: [%s]", LPickedNode->Name);

						EIndex  LBranchesN = 0;

						{
							EIndex  LC = 0;for(;LC < LSelectedBranches_Count;LC++)
							{
								E3dNode*  LSelectedBranch = LSelectedBranches[LC];
								const EIndex  LConnectionLevel = E3dNode_HrcIsDescendantOf(LSelectedBranch, LPickedNode);
								if(LConnectionLevel == E_NOT_CONNECTED || LConnectionLevel > 1)
								{
									if(E3dNode_HrcIsDescendantOf(LPickedNode, LSelectedBranch) == E_NOT_CONNECTED) LBranchesN++;
								}
							}
						}

						if(LBranchesN)
						{
							E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)EOp_OperatorNew(E3dOperatorEditHierarchy_class);

							{
								ENewNClr(LBranchesN, LOperator->BranchMoves);LOperator->BranchMoves_Count = LBranchesN;

								EIndex  LC = 0;E3dBranchMove*  LBranchMove = LOperator->BranchMoves;for(;LC < LSelectedBranches_Count;LC++)
								{
									E3dNode*  LSelectedBranch = LSelectedBranches[LC];
									const EIndex  LConnectionLevel = E3dNode_HrcIsDescendantOf(LSelectedBranch, LPickedNode);
									if(LConnectionLevel == E_NOT_CONNECTED || LConnectionLevel > 1)
									{
										if(E3dNode_HrcIsDescendantOf(LPickedNode, LSelectedBranch) == E_NOT_CONNECTED)
										{
											E_AssignToMember(LSelectedBranch, LBranchMove,Node);
											E_AssignToMember(LSelectedBranch->Parent, LBranchMove,OldParent);
											if(LSelectedBranch->Parent) LBranchMove->OldIndex = E3dNode_GetChildIndex(LSelectedBranch);
											else LBranchMove->OldIndex = EList_PointerGetIndex(LScene->RootNodes, LSelectedBranch);
											E_AssignToMember(LPickedNode, LBranchMove,NewParent);

											LBranchMove++;
										}
									}
								}
							}

							E3dp_UndoStackAddOperator((EOperator*)LOperator, NULL);
							LChanged |= _EditHierarchy_Proc(NULL, (EOperator*)LOperator, EOpDO);

							_ConnectMode = 0;
							E3dp_PickEnd();
						}
					}
				}
			break;

			case E3dBA_PICK_OBJECT_FOR_ADDOBJECTTO_MODELSOF:
				if(LObjItem->Object)
				{
					E3dObject*  LPickedObject = LObjItem->Object;
					E3dModel**  LPickedObjectModels = LPickedObject->Models;const EIndex  LPickedObjectModels_Count = LPickedObject->Models_Count;

					EIndex  LObjectsN = 0;

					{
						EIndex  LObjC = 0;for(;LObjC < LSelectedObjects_Count;LObjC++)
						{
							E3dObject*  LSelectedObject = LSelectedObjects[LObjC];
							if(LSelectedObject != LPickedObject)
							{
								EIndex  LIC = 0;for(;LIC < LPickedObjectModels_Count;LIC++)
								{
									E3dModel*  LModel = LPickedObjectModels[LIC];
									if(EList_PointerGetIndex(LModel->Objects, LSelectedObject) == E_NOT_FOUND) { LObjectsN++;break; }
								}
							}
						}
					}

					if(LObjectsN)
					{
						E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)EOp_OperatorNew(E3dOperatorEditHierarchy_class);

						E3dp_PrintCompoundMessageVa(ELogMESSAGE, 5000,  XeFONT, XeFontDEFAULT, XeTEXT, "Picked Object ",  E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""),  0);

						ENewNClr(LObjectsN, LOperator->ObjectMoves);LOperator->ObjectMoves_Count = LObjectsN;E3dObjectMove*  LObjectMove = LOperator->ObjectMoves;

						EIndex  LObjC = 0;for(;LObjC < LSelectedObjects_Count;LObjC++)
						{
							E3dObject*  LSelectedObject = LSelectedObjects[LObjC];

							if(LSelectedObject != LPickedObject)
							{
								EBool  LObjAdd = FALSE;

								EIndex  LIC = 0;for(;LIC < LPickedObjectModels_Count;LIC++)
								{
									E3dModel*  LModel = LPickedObjectModels[LIC];
									if(EList_PointerGetIndex(LModel->Objects, LSelectedObject) == E_NOT_FOUND)
									{
										EObjectAndIndex*  LAddToModel = EList_StructAddClr(LObjectMove->AddToModels);
										E_AssignToMember(LModel, LAddToModel,Object);
										LAddToModel->Index = LModel->Objects_Count;
										LObjAdd = TRUE;
									}
								}

								if(LObjAdd)
								{
									E_AssignToMember(LSelectedObject, LObjectMove,Object);
									LObjectMove++;
								}
							}
						}

						E3dp_UndoStackAddOperator((EOperator*)LOperator, NULL);
						LChanged |= _EditHierarchy_Proc(NULL, (EOperator*)LOperator, EOpDO);

						_ConnectMode = 0;
						E3dp_PickEnd();
					}
				}
			break;

			case E3dBA_PICK_CHILD_OBJECT_FOR_MODELS:
				switch(LObjItem->Type)
				{
					default:
						if(LObjItem->Object && LSelectedNodes_Count)
						{
							E3dObject*  LPickedObject = LObjItem->Object;

							EIndex  LAddToModelsN = 0;

							{
								EIndex  LC = 0;for(;LC < LSelectedNodes_Count;LC++)
								{
									E3dNode*   LSelectedNode = LSelectedNodes[LC];
									E3dModel*  LSelectedModel = E_DynamicCast(LSelectedNode, E3dModel);if(LSelectedModel)
									{
										const EIndex  LIndex = EList_PointerGetIndex(LSelectedModel->Objects, LPickedObject);if(LIndex == E_NOT_FOUND) LAddToModelsN++;
									}
								}
							}

							if(LAddToModelsN)
							{
								E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)EOp_OperatorNew(E3dOperatorEditHierarchy_class);

								E3dp_PrintCompoundMessageVa(ELogMESSAGE, 5000,  XeFONT, XeFontDEFAULT, XeTEXT, "Picked Object ",  E3dp_ObjectS(LPickedObject->Name ? LPickedObject->Name : ""),  0);

								ENewNClr(1, LOperator->ObjectMoves);LOperator->ObjectMoves_Count = 1;
								E3dObjectMove*  LObjectMove = LOperator->ObjectMoves;
								E_AssignToMember(LPickedObject, LObjectMove,Object);


								EIndex  LC = 0;for(;LC < LSelectedNodes_Count;LC++)
								{
									E3dNode*   LSelectedNode = LSelectedNodes[LC];
									E3dModel*  LSelectedModel = E_DynamicCast(LSelectedNode, E3dModel);if(LSelectedModel)
									{
										const EIndex  LIndex = EList_PointerGetIndex(LSelectedModel->Objects, LPickedObject);if(LIndex == E_NOT_FOUND)
										{
											EObjectAndIndex*  LAddToModel = EList_StructAddClr(LObjectMove->AddToModels);
											E_AssignToMember(LSelectedModel, LAddToModel,Object);
											LAddToModel->Index = LSelectedModel->Objects_Count;
										}
									}
								}

								if(_ConnectMode == E3dConnectREPARENT)
								{
									EIndex  LMC = 0;for(;LMC < LPickedObject->Models_Count;LMC++)
									{
										E3dModel*  LModel = LPickedObject->Models[LMC];if(EList_PointerGetIndex(LSelectedNodes, LModel) == E_NOT_FOUND)
										{
											EObjectAndIndex*  LDelFrModel = EList_StructAddClr(LObjectMove->DelFrModels);
											E_AssignToMember(LModel, LDelFrModel,Object);
											LDelFrModel->Index = EList_PointerGetIndex(LModel->Objects, LPickedObject);
										}
									}
								}

								E3dp_UndoStackAddOperator((EOperator*)LOperator, NULL);
								LChanged |= _EditHierarchy_Proc(NULL, (EOperator*)LOperator, EOpDO);
							}
						}
					break;
				}
			break;

			case E3dBA_PICK_PARENT_MODEL_FOR_OBJECTS:
printf("%s() %s:%d  %sGREEN%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout);
				if(LSelectedObjects_Count)
				{
					E3dNode*   LPickedNode = LObjItem->Node;
					E3dModel*  LPickedModel = E_DynamicCast(LPickedNode, E3dModel);if(LPickedModel)
					{
						EIndex  LObjectsN = 0;

						{
							EIndex  LC = 0;for(;LC < LSelectedObjects_Count;LC++)
							{
								E3dObject*  LObject = LSelectedObjects[LC];
								EBool       LObjectAdd = FALSE;
								EBool       LObjectDel = FALSE;

								const EIndex  LIndex = EList_PointerGetIndex(LPickedModel->Objects, LObject);if(LIndex == E_NOT_FOUND)
								{
									LObjectAdd = TRUE;
								}

								if(_ConnectMode == E3dConnectREPARENT)
								{
									EIndex  LMC = 0;for(;LMC < LObject->Models_Count;LMC++)
									{
										E3dModel*  LModel = LObject->Models[LMC];
										if(LModel != LPickedModel) { LObjectDel = TRUE;break; }
									}
								}

								if(LObjectAdd || LObjectDel) LObjectsN++;
							}
						}


						if(LObjectsN)
						{
							E3dp_PrintCompoundMessageVa(ELogMESSAGE, 5000,  XeFONT, XeFontDEFAULT, XeTEXT, "Picked Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),  0);

							E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)EOp_OperatorNew(E3dOperatorEditHierarchy_class);

							ENewNClr(LObjectsN, LOperator->ObjectMoves);LOperator->ObjectMoves_Count = LObjectsN;
							if(LObjectsN > 1) LOperator->Name = EStrDup("Add Objects to Model");
							else LOperator->Name = EStrDup("Add Object to Model");

							EIndex  LC = 0;E3dObjectMove*  LObjectMove = LOperator->ObjectMoves;for(;LC < LSelectedObjects_Count;)
							{
								E3dObject*  LObject = LSelectedObjects[LC];
								EBool       LObjectAdd = FALSE;
								EBool       LObjectDel = FALSE;

								const EIndex  LIndex = EList_PointerGetIndex(LPickedModel->Objects, LObject);if(LIndex == E_NOT_FOUND)
								{
									EObjectAndIndex*  LAddModel = EList_StructAddClr(LObjectMove->AddToModels);
									E_AssignToMember(LPickedModel, LAddModel,Object);
//									LAddModel->Index = LPickedModel->Objects_Count;  // E3dFIXME Add to the end...
									LObjectAdd = TRUE;
								}

								if(_ConnectMode == E3dConnectREPARENT)
								{
									EIndex  LMC = 0;for(;LMC < LObject->Models_Count;LMC++)
									{
										E3dModel*  LModel = LObject->Models[LMC];if(LModel != LPickedModel)
										{
											EObjectAndIndex*  LDelFrModel = EList_StructAddClr(LObjectMove->DelFrModels);
											E_AssignToMember(LModel, LDelFrModel,Object);
											LDelFrModel->Index = EList_PointerGetIndex(LModel->Objects, LObject);
											LObjectDel = TRUE;
										}
									}
								}

								if(LObjectAdd || LObjectDel)
								{
									E_AssignToMember(LObject, LObjectMove,Object);
									LC++;LObjectMove++;
								}
							}

							E3dp_UndoStackAddOperator((EOperator*)LOperator, NULL);
							LChanged |= _EditHierarchy_Proc(NULL, (EOperator*)LOperator, EOpDO);
						}
						else
						{
							E3dp_PrintCompoundMessageVa(ELogMESSAGE, 5000,
							 XeFONT, XeFontDEFAULT, XeTEXT, "This Model ",  E3dp_NodeS(LPickedNode->Name ? LPickedNode->Name : ""),  XeTEXT, " already has the selected Object(s)",  0);
						}
					}
					_ConnectMode = 0;
					E3dp_PickEnd();
				}
			break;
		}
	}

	EFreeChk(LSelectedNodes);
	EFreeChk(LSelectedObjects);

	if(LChanged & E3dChgHIERARCHY)
	{
		E3dScene_ObjectsUpdateModelList(LScene);
		E3dScene_LayoutOnSchematics(LScene, FALSE);
	}

	if(LChanged)
	{
		E3dWindows_RedrawByChangeFlags(LChanged);
		E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	}
}


//======================================================================================
// Get actions and status-panel strint for PERSPECTIVE, TOP, FRONT and RIGHT views
//======================================================================================
static EChangeFlags _DisconnectGetButtonActions(E3dItem* PPickedItems, const int PNumOfPickedItems, E3dNode** PSelectedNodes, const int PNumOfSelectedNodes, E3dObject** PSelectedObjects, const int PNumOfSelectedObjects, int* PButtonActions, XeString* PXeStringRet)
{
	E3dObjItem*   LObjItem = (E3dObjItem*)PPickedItems;
	XeString      LXeString = NULL;
	EChangeFlags  LChanged = 0;

	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
	const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

	switch(LObjItem->Type)
	{
		case E3dItemMODEL:
		case E3dItemPOLYGON:
		case E3dItemSPLINE:
		case E3dItemFACE:
			{
				E3dNode*   LNode = LObjItem->Node;
				E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
				{
					switch(PNumOfSelectedObjects)
					{
						case 0:
							LXeString = XeString_NewVa(TRUE,
							 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Objects -  Select Objects  ",
							 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
							 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
							 0);
						break;

						case 1:
							if(E3dModel_GetIndexOfObject(LModel, PSelectedObjects[0]) > -1)
							{
								LXeString = XeString_NewVa(TRUE,
								 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Object - ",
								 E3dp_LMouseButtonS, XeTEXT, ": Disconnect selected Object from this Model  ",
								 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
								 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
								 0);
								PButtonActions[0] = E3dBA_PICK_NODE_TO_DISCONNECT_OBJECTS_FROM;
							}
							else
							{
								LXeString = XeString_NewVa(TRUE,
								 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Object - Selected Object is not on this Model  ",
								 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
								 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
								 0);
							}
						break;

						default:
							{
								EBool  LModelContainsObject = FALSE;

								EIndex  LC = 0;for(;LC < PNumOfSelectedObjects;LC++)
								{
									if(E3dModel_GetIndexOfObject(LModel, PSelectedObjects[LC]) > -1)
									{
										LModelContainsObject = TRUE;
										break;
									}
								}

								if(LModelContainsObject)
								{
									LXeString = XeString_NewVa(TRUE,
									 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect - ",
									 E3dp_LMouseButtonS,  XeTEXT, ": Disconnect selected Objects from this Model  ",
									 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
									 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
									 0);
									PButtonActions[0] = E3dBA_PICK_NODE_TO_DISCONNECT_OBJECTS_FROM;
								}
								else
								{
									LXeString = XeString_NewVa(TRUE,
									 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Object -  None of the selected Objects are on this Model  ",
									 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
									 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
									0);
								}
							}
						break;
					}
				}
				if(PButtonActions[0] || PButtonActions[1] || PButtonActions[2]) { LModel->Highlighted = TRUE;LChanged |= E3dChgSELECTION; }
			}
		break;
	}

	*PXeStringRet = LXeString;

	return(LChanged);
}


//================================================================
// Connect pick callback
//================================================================
static void _PickCB_DisconnectObject(E3dPickCallbackStruct* PPickStruct, EPointer PClientData,  const EIndex PNumOfPickedItems, E3dItem* PPickedItems)
{
	E3dScene*     LScene = E3d_Scene;
	E3dpWindow*   L3DWindow = PPickStruct->E3DWindow;
	XeString      LXeString = NULL;
	E3dNode**     LSelectedNodes = NULL;  const EIndex  LSelectedNodes_Count   = E3dScene_GetSelectedNodes(LScene,  &LSelectedNodes, NULL,  E3dNodeSelNODE|E3dNodeSelBRANCH_ROOT,   FALSE);
	E3dObject**   LSelectedObjects = NULL;const EIndex  LSelectedObjects_Count = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemALL, E3dSELECTED|E3dEXCLUDE_MODEL_SELECTION, FALSE);
	int           LButtonActions[3] = { 0, 0, 0 };
	EChangeFlags  LChanged = 0;


	if(E3dScene_UnhighlightAll(LScene, E3dSF_ALL)) LChanged = E3dChgSELECTION;

	if(PPickStruct->Reason == E3dCR_ABORT)
	{
		E_PluginLockDel(_Plugin, &_DisconnectPickLock, FALSE);
		return;
	}

	if(PPickStruct->Button == Button3)
	{
		E3dp_PickEnd();
		E_PluginLockDel(_Plugin, &_DisconnectPickLock, FALSE);
		return;
	}

	if(LSelectedObjects_Count == 0)
	{
		E3dp_PrintMessage(ELogMESSAGE, 0, "Disconnect Objects -  Select Objects");

		if(LSelectedNodes) EFree(LSelectedNodes);
		if(LSelectedObjects) EFree(LSelectedObjects);
		return;
	}

	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
	const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

	const E3dViewClass*  LViewClass = L3DWindow->ViewMode->Class;

	if(PNumOfPickedItems)
	{
		if(LViewClass == E3dSchematicsView_class || LViewClass == E3dTopView_class || LViewClass == E3dFrontView_class || LViewClass == E3dRightView_class || LViewClass == E3dPerspectiveView_class)
		{
			LChanged |= _DisconnectGetButtonActions(PPickedItems, PNumOfPickedItems, LSelectedNodes, LSelectedNodes_Count, LSelectedObjects, LSelectedObjects_Count, LButtonActions, &LXeString);
		}
		else
		{
			if(LSelectedObjects_Count > 1)
			{
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Objects - Move pointer to a 3D or a schematics view  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
			}
			else
			{
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Object - Move pointer to a 3D or a schematics view  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
			}
		}
	}
	else
	{
		if(LViewClass == E3dSchematicsView_class || LViewClass == E3dTopView_class || LViewClass == E3dFrontView_class || LViewClass == E3dRightView_class || LViewClass == E3dPerspectiveView_class)
		{
			if(LSelectedObjects_Count > 1)
			{
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Objects - Move pointer over a Model  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
			}
			else
			{
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Object - Move pointer over a Model  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
			}
		}
		else
		{
			if(LSelectedObjects_Count > 1)
			{
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Objects - Move pointer to a 3D or a schematics view  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
			}
			else
			{
				LXeString = XeString_NewVa(TRUE,
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Disconnect Object - Move pointer to a 3D or a schematics view  ",
				 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
				 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
				 0);
			}
		}
	}

	if(LXeString)
	{
		E3dp_PrintCompoundMessage(ELogMESSAGE, 0, LXeString);
		XeString_Free(LXeString);
	}

	if(LChanged == 0)
	{
		if(LSelectedNodes) EFree(LSelectedNodes);
		if(LSelectedObjects) EFree(LSelectedObjects);
		return;
	}


	if(PNumOfPickedItems)
	{
		int  LAction = 0;

		switch(PPickStruct->Button)
		{
			case Button1:  LAction = LButtonActions[0];break;
			case Button2:  LAction = LButtonActions[1];break;
			case Button3:  LAction = LButtonActions[2];break;
		}

		switch(LAction)
		{
			case E3dBA_PICK_NODE_TO_DISCONNECT_OBJECTS_FROM:
				{
					E3dObjItem*  LObjItem = (E3dObjItem*)PPickedItems;
					E3dNode*     LNode = LObjItem->Node;
					E3dModel*    LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
					{
						E3dObject**  LObjects = NULL;const EIndex  LNumOfObjects = E3dScene_GetObjects(LScene, &LObjects, E3dItemANYOBJECT, E3dSELECTED, FALSE);if(LNumOfObjects)
						{
							EIndex  LObjectsN = 0;

							{
								EIndex  LC = 0;for(;LC < LModel->Objects_Count;LC++)
								{
									E3dObject*  LObject = LModel->Objects[LC];

									if(LObject->Selection && LObject->Models_Count > 1) LObjectsN++;
								}
							}

							if(LObjectsN)
							{
								E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)EOp_OperatorNew(E3dOperatorEditHierarchy_class);

								ENewNClr(LObjectsN, LOperator->ObjectMoves);LOperator->ObjectMoves_Count = LObjectsN;E3dObjectMove*  LObjectMove = LOperator->ObjectMoves;

								{
									EIndex  LC = 0;for(;LC < LModel->Objects_Count;LC++)
									{
										E3dObject*  LObject = LModel->Objects[LC];

										if(LObject->Selection && LObject->Models_Count > 1)
										{
											char*  LObjectName = LObject->Name ? LObject->Name : "";
											char*  LNodeName   = LNode->Name ? LNode->Name : "";

											E_AssignToMember(LObject, LObjectMove,Object);
											EObjectAndIndex*  LDelFrModel = EList_StructAddClr(LObjectMove->DelFrModels);
											E_AssignToMember(LModel, LDelFrModel,Object);
											LDelFrModel->Index = EList_PointerGetIndex(LModel->Objects, LObject);
											LObjectMove++;

											E3dp_PrintCompoundMessageVa(ELogMESSAGE, 5000,
											 XeFONT, XeFontDEFAULT,
											 XeCOLOR_RGBA, LLabelRGBAp8Color,  XeTEXT, "Object ",                   E3dp_ObjectS(LObjectName),
											                                   XeTEXT, " disconnected from node ",  E3dp_NodeS(LNodeName),
											 0);
										}
									}
								}

								E3dp_UndoStackAddOperator((EOperator*)LOperator, NULL);
								LChanged |= _EditHierarchy_Proc(NULL, (EOperator*)LOperator, EOpDO);

								LChanged |= E3dModel_GetTransformChangeFlagsByObjects(LModel) | E3dChgHIERARCHY;
							}

							EFree(LObjects);
						}
					}
				}
			break;
		}
	}

	if(LChanged & E3dChgHIERARCHY)
	{
		E3dScene_ObjectsUpdateModelList(LScene);
		E3dScene_LayoutOnSchematics(LScene, FALSE);
	}

	if(LChanged)
	{
		E3dWindows_RedrawByChangeFlags(LChanged);
		E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	}


// If none of the remaining selected Objects are
// instanced by multiple Models, end mode
//
	if(LSelectedObjects_Count)
	{
		EBool  LFoundMultiInstance = FALSE;

		EIndex  LC = 0;for(;LC < LSelectedObjects_Count;LC++)
		{
		  if(LSelectedObjects[LC]->Models_Count > 1) { LFoundMultiInstance = TRUE;break; }
		}

		if(!LFoundMultiInstance)
		{
			E3dp_PickEnd();
			E_PluginLockDel(_Plugin, &_DisconnectPickLock, FALSE);
		}
	}
}


//========================================
// Dialog box callback
//========================================
static void _DCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	switch((EPointerInt)PClientData)
	{
		case CR_TOGGLE_HRC_COMPENSATION:
			if(((XeToggleButtonCallbackStruct*)PCallData)->Set) _TransformCompensation = TRUE;
			else _TransformCompensation = FALSE;
		break;
	}
}


//========================================
// Dialog box callback
//========================================
static void _DCB_Close(Widget PW, EPointer PClientData, EPointer PCallData)
{
	Xe_WidgetUnmap(_DialogW);

	E_PluginLockDel(_Plugin, &_DialogLock, FALSE);
}


static int _DialogLockDel(EPlugin* PPlugin, EPluginLock* PLock)
{
	Xe_WidgetUnmap(_DialogW);

	return(E_SUCCESS);
}


//========================================================
// Pop up connect/disconnect settings dialog box
//========================================================
static void _PopupDialog(void)
{
	if(_DialogLock == NULL) _DialogLock = E_PluginLockAdd(_Plugin, _DialogLockDel);

	if(_DialogW == NULL)
	{
		Widget  LW;
		EArg    Args[256];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNallowShellResize, TRUE);
		EXeSetArg(XeNcreateButtons, FALSE);
		EXeSetArg(XeNworkAreaType, XeMATRIX);
		EXeSetArg(XeNwmDecorations, XeWMDecorCLOSE|XeWMDecorBORDER|XeWMDecorRESIZEH|XeWMDecorTITLE);
		EXeSetArg(XeNtitle, "Hierarchy connect/disconnect settings");EXeSetArg(XeNiconName, "Hierarchy");
		Widget  LDialogW = XeDialog_New(XeDIALOG_TEMPLATE, "Hierarchy", E3dp_TopShellW,  Args, ArgCnt);if(LDialogW == NULL) return;

		XeAddCallback(LDialogW, XeNwmCallback, _DCB_Close, (EPointer)XeCR_CANCEL);

		_DialogW = LDialogW;

		Widget  LTopMatrixW = XeDialog_GetChild(LDialogW, XeDIALOG_WORK_AREA);
		EXeStart;
		EXeSetArg(XeNmarginSizeX, 6);EXeSetArg(XeNmarginSizeY, 4);
		Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

		EXeStart;
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNset, _TransformCompensation);
		LW = XeWidget_ManagedNew("Transform compensation", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB, (EPointer)CR_TOGGLE_HRC_COMPENSATION);

		XeShell_Raise(_DialogW);
		XeShell_CenterToWidget(_DialogW, E3dp_TopShellW);
	}
	else XeShell_Raise(_DialogW);
}


static int _PickLockDel(EPlugin* PPlugin, EPluginLock* PLock)
{
	E3dp_PickEnd();

	return(E_SUCCESS);
}


//================================================================
// Connect
//================================================================
static void _CB_Connect(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dp_ResetWorkModes();
	if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) { _PopupDialog();return; }

	if(_ConnectMode == 0)
	{
		E3dScene*     LScene = E3d_Scene;
		XeString      LXeString = NULL;
		E3dNode**     LSelectedNodes = NULL;  const EIndex  LSelectedNodes_Count   = E3dScene_GetSelectedNodes(LScene,  &LSelectedNodes, NULL,  E3dNodeSelNODE|E3dNodeSelBRANCH_ROOT, FALSE);
		E3dObject**   LSelectedObjects = NULL;const EIndex  LSelectedObjects_Count = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemALL, E3dSELECTED|E3dEXCLUDE_MODEL_SELECTION, FALSE);

		if(LSelectedNodes_Count == 0 && LSelectedObjects_Count == 0) { E3dp_PrintMessage(ELogWARNING, 0, _Msg_SelectModelOrObjectToConnect);return; }
		if(LSelectedNodes_Count > 0 && LSelectedObjects_Count > 0) E3dp_PrintMessage(ELogWARNING, 0, _Msg_SelectOnlyModelOrObject);

		const char*   LToolBaseName;const char*  LToolName = _GetToolName(E3dConnectCONNECT, LSelectedNodes_Count, LSelectedObjects_Count,  &LToolBaseName, NULL);

		const E3dpPrefs*  LPrefs = &E3dp_Prefs;
		const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
		const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

		if(LSelectedObjects_Count)
		{
			LXeString = XeString_NewVa(TRUE,
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - Move pointer over a Model  ",
			 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
			 0);
		}
		else if(LSelectedNodes_Count)
		{
			LXeString = XeString_NewVa(TRUE,
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, LToolName,  XeTEXT, " - Move pointer over a Model or an Object  ",
			 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
			 0);
		}

		if(E3dp_PickRequest(LXeString, E3dItemMODEL|E3dItemANYOBJECT|E3dItemWORLD_POSITION,  _PickCB_Connect, NULL,  0, TRUE) == E_SUCCESS)
		{
			_ConnectMode = E3dConnectCONNECT;

			if(_ConnectPickLock == NULL) _ConnectPickLock = E_PluginLockAdd(_Plugin, _PickLockDel);
		}

		EFreeChk(LSelectedNodes);
		EFreeChk(LSelectedObjects);
	}
}


//================================================================
// Disconnect
//================================================================
static void _CB_Disconnect(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dp_ResetWorkModes();
	if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) { _PopupDialog();return; }

	{
		E3dScene*     LScene = E3d_Scene;
		EChangeFlags  LChanged = 0;

		EIndex  LNodesToDisconnectN = 0;

		{
			EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
			{
				E3dNode*  LRootNode = LScene->RootNodes[LC];
				E3dNode*  LNode = LRootNode;for(;LNode;LNode = LNode->Next)
				{
					switch(LNode->Selection)
					{
						case E3dNodeSelNODE:
						case E3dNodeSelBRANCH_ROOT:
							if(LNode->Parent) LNodesToDisconnectN++;
						break;
					}
				}
			}
		}


		if(LNodesToDisconnectN)
		{
			E3dOperatorEditHierarchy*  LOperator = (E3dOperatorEditHierarchy*)EOp_OperatorNew(E3dOperatorEditHierarchy_class);

			ENewNClr(LNodesToDisconnectN, LOperator->BranchMoves);LOperator->BranchMoves_Count = LNodesToDisconnectN;E3dBranchMove*  LBranchMove = LOperator->BranchMoves;

			EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
			{
				E3dNode*  LRootNode = LScene->RootNodes[LC];
				E3dNode*  LNode = LRootNode;for(;LNode;LNode = LNode->Next)
				{
					switch(LNode->Selection)
					{
						case E3dNodeSelNODE:
						case E3dNodeSelBRANCH_ROOT:
							if(LNode->Parent)
							{
								E_AssignToMember(LNode, LBranchMove,Node);
								E_AssignToMember(LNode->Parent, LBranchMove,OldParent);
								LBranchMove->OldIndex = E3dNode_GetChildIndex(LNode);
								LBranchMove++;
							}
						break;
					}
				}
			}

			E3dp_UndoStackAddOperator((EOperator*)LOperator, NULL);
			LChanged |= _EditHierarchy_Proc(NULL, (EOperator*)LOperator, EOpDO);

			if(LChanged & E3dChgHIERARCHY)
			{
				E3dScene_ObjectsUpdateModelList(LScene);
				E3dScene_LayoutOnSchematics(LScene, FALSE);
				E3dWindows_RedrawByChangeFlags(LChanged);
				E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
			}

			return;
		}

		const E3dpPrefs*  LPrefs = &E3dp_Prefs;
		const uint32_t    LLabelRGBAp8Color = Xe_ForegroundRGBA8pColor;

		EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
		{
			E3dNode*  LRootNode = LScene->RootNodes[LC];
			E3dNode*  LNode = LRootNode;for(;LNode;LNode = LNode->Next)
			{
				switch(LNode->Selection)
				{
					case E3dNodeSelOBJECT:
						{
							E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
							{
								EIndex  LOC = 0;const EIndex  LON = LModel->Objects_Count;for(;LOC < LON;LOC++)
								{
									E3dObject*  LObject = LModel->Objects[LOC];
									EBool       LValidDisconnectableObject = FALSE;

									if(LObject->Selection)
									{
										E3dModel*  LSingleInstanceModel = NULL;

										if(LObject->Models_Count > 1)
										{
											XeString	LXeString = XeString_NewVa(TRUE,
											 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,  XeTEXT, "Disconnect - ",
											 E3dp_LMouseButtonS,     XeTEXT, ": pick Model from which to disconnect the selected Object(s)  ",
											 E3dp_RMouseButtonEscS,  XeTEXT, ": cancel",
											 0);

											E3dp_PickRequest(LXeString, E3dItemMODEL,  _PickCB_DisconnectObject, NULL,  0, TRUE);
											if(_DisconnectPickLock == NULL) _DisconnectPickLock = E_PluginLockAdd(_Plugin, _PickLockDel);

											LValidDisconnectableObject = TRUE;
											break;
										}
										else LSingleInstanceModel = LObject->Models[0];

										if(LSingleInstanceModel && !LValidDisconnectableObject) E3dp_PrintCompoundMessageVa(ELogWARNING, 5000, XeFONT, XeFontDEFAULT,  XeTEXT, "Disconnect - Object ",  E3dp_ObjectS(LObject->Name ? LObject->Name : ""),  XeTEXT, " is only connected to 1 Model: ",  E3dp_NodeS(LSingleInstanceModel->Name ? LSingleInstanceModel->Name : ""),  0);
									}
								}
								return;
							}
						}
					break;
				}
			}
		}
	}
}


//================================================================
// Reparent
//================================================================
static void _CB_Reparent(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dp_ResetWorkModes();
	if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) { _PopupDialog();return; }

	if(_ConnectMode == 0)
	{
		E3dScene*     LScene = E3d_Scene;
		XeString      LXeString = NULL;
		E3dNode**     LSelectedNodes = NULL;const EIndex  LSelectedNodes_Count   = E3dScene_GetSelectedNodes(LScene,  &LSelectedNodes, NULL,  E3dNodeSelNODE|E3dNodeSelBRANCH_ROOT, FALSE);
		E3dObject**   LSelectedObjects;const EIndex  LSelectedObjects_Count = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemALL, E3dSELECTED|E3dEXCLUDE_MODEL_SELECTION, FALSE);

		if(LSelectedNodes_Count == 0 && LSelectedObjects_Count == 0) { E3dp_PrintMessage(ELogWARNING, 0, _Msg_SelectModelOrObjectToReparent);return; }

		if(LSelectedNodes_Count > 0 && LSelectedObjects_Count > 0) E3dp_PrintMessage(ELogWARNING, 0, _Msg_SelectOnlyModelOrObjectToReparent);

		const E3dpPrefs*  LPrefs = &E3dp_Prefs;
		const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
		const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

		if(LSelectedObjects_Count)
		{
			LXeString = XeString_NewVa(TRUE,
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Reparent Object(s):  Move pointer over a Model  ",
			 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
			 0);
		}
		else if(LSelectedNodes_Count)
		{
			LXeString = XeString_NewVa(TRUE,
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Reparent Model(s):  Move pointer over a Model or an Object  ",
			 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
			 0);
		}

		if(E3dp_PickRequest(LXeString, E3dItemMODEL|E3dItemANYOBJECT|E3dItemWORLD_POSITION,  _PickCB_Connect, NULL,  0, TRUE) == E_SUCCESS)
		{
			_ConnectMode = E3dConnectREPARENT;

			if(_ConnectPickLock == NULL) _ConnectPickLock = E_PluginLockAdd(_Plugin, _PickLockDel);
		}

		EFreeChk(LSelectedNodes);
		EFreeChk(LSelectedObjects);
	}
}


//========================================
// Menu callback
//========================================
static void _MCB_SeparateSelectedObjects(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	E3dNode**     LOldRootNodes;ENewN(LScene->RootNodes_Count, LOldRootNodes);
	EChangeFlags  LChanged = 0;

	EListAI(, E3dNode*, LNodesToSkip);

	{
		const EIndex	LRN = LScene->RootNodes_Count;memcpy(LOldRootNodes, LScene->RootNodes, sizeof(E3dNode*) * LRN);
		EIndex  LRC = 0;for(;LRC < LRN;LRC++)
		{
			EBool     LUpdateHrc = FALSE;
			E3dNode*  LRootNode = LOldRootNodes[LRC];
			EIndex    LNextRootIndex = EList_PointerGetIndex(LScene->RootNodes, LRootNode);  // We might add new Hierarchies, so we need to get this index here (can't just use LRC)
			E3dNode*  LNode = LRootNode;for(;LNode;LNode = LNode->Next)
			{
				const EIndex  LNodeIndex = EList_PointerGetIndex(LNodesToSkip, LNode);  // Skip Nodes that we just added...

				if(LNodeIndex == E_NOT_FOUND && LNode->Selection == E3dNodeSelOBJECT)
				{
					E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel && LModel->Objects_Count > 1)
					{
//printf("%s() %s:%d [%s]\n", __FUNCTION__, __FILE__, __LINE__, LNode->Name);fflush(stdout);
						E3dObject**  LObjects = LModel->Objects;

						EIndex  LObjC = 0;for(;LObjC < LModel->Objects_Count;)
						{
							E3dModel*		LNewModel = NULL;
							E3dObject*	LObject = LObjects[LObjC];

//printf("%s() %s:%d  <%s>\n", __FUNCTION__, __FILE__, __LINE__, LObject->Name);fflush(stdout);
							if(LObject->Selection == E3dObjectSelOBJECT)
							{
//printf("%s() %s:%d  <%s>\n", __FUNCTION__, __FILE__, __LINE__, LObject->Name);fflush(stdout);
								if(LNewModel == NULL)
								{
									E_New(E3dModel, LNewModel);

									EObject_SetName(LNewModel, E3dScene_GetUniqueNodeName(LScene, LModel->Name), FALSE);

									E3dTransformSet_Copy(&LModel->TransformSet, &LNewModel->TransformSet);
								}
								E3dModel_AppendObject(LNewModel, LObject);
								E3dModel_RemoveObject(LModel, LObject);
								LChanged |= E3dObject_GetTransformChangeFlagsByClass(LObject) | E3dChgHIERARCHY;
							}
							else LObjC++;

							if(LNewModel)
							{
								E3dNode*  LParentNode = LModel->Parent;if(LParentNode && _SeparateSelectedObjects_NewModelUnderOldParent)
								{
									const EIndex  LIndex = E3dNode_GetChildIndex((E3dNode*)LModel);
									E3dNode_ConnectChild(LModel->Parent, (E3dNode*)LNewModel, LIndex + 1, FALSE, FALSE);
									E3dNodeHrc_UpdateConnections(LRootNode);
									LUpdateHrc = TRUE;

									EList_PointerAddAChk(LNodesToSkip, 16, LNewModel);
//printf("%s() %s:%d  Added [%s]\n", __FUNCTION__, __FILE__, __LINE__, LNewModel->Name);fflush(stdout);
								}
								else
								{
									E3dScene_NodeHrcInsert(LScene, (E3dNode*)LNewModel, LNextRootIndex++, TRUE);
									if(LParentNode) E3dMatrix_DecomposeToScaleRotateTranslate(LModel->LocalToWorldMatrix, NULL, LNewModel->TransformSet.EulerRotation.Order,  &LNewModel->TransformSet);
									E3dNodeHrc_UpdateMatrices((E3dNode*)LNewModel, FALSE);
								}
							}
						}
					}
				}
			}

			if(LUpdateHrc)
			{
				E3dNodeHrc_UpdateMatrices(LRootNode, TRUE);
				E3dNodeHrc_UpdateMaterialInheritance(LRootNode);
				E3dNodeHrc_CallNodeInfoUpdateProcs(LRootNode, E3dChgHIERARCHY, FALSE);
			}

			LNodesToSkip_Count = 0;
		}
	}

	EFree(LOldRootNodes);

	EList_FreeAChk(LNodesToSkip);

	if(LChanged)
	{
		E3dScene_ObjectsUpdateModelList(LScene);
		E3dScene_LayoutOnSchematics(LScene, TRUE);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
	else E3dp_PrintMessage(ELogWARNING, 7000, "Select Objects that are not the only Objects under their Models");
}


//================================================================
// Add selected Objects to the same Nodes as the one picked
//================================================================
static void _CB_Reinstantiate(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dp_ResetWorkModes();
	if(_ConnectMode == 0)
	{
		E3dScene*     LScene = E3d_Scene;
		XeString      LXeString = NULL;
		E3dObject**   LSelectedObjects;
		const EIndex  LSelectedObjects_Count = E3dScene_GetObjects(LScene, &LSelectedObjects, E3dItemALL, E3dSELECTED|E3dEXCLUDE_MODEL_SELECTION, FALSE);

		if(LSelectedObjects_Count == 0) { E3dp_PrintMessage(ELogWARNING, 0, _Msg_SelectObjectToReInstantiate);return; }

		if(LSelectedObjects_Count)
		{
			const E3dpPrefs*  LPrefs = &E3dp_Prefs;
			const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
			const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

			LXeString = XeString_NewVa(TRUE,
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, "Add Object(s) to the same Nodes as:  Move pointer over an Object  ",
			 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": cancel",
			 0);
		}

		if(E3dp_PickRequest(LXeString, E3dItemMODEL|E3dItemANYOBJECT|E3dItemWORLD_POSITION,  _PickCB_Connect, NULL,  0, TRUE) == E_SUCCESS)
		{
			_ConnectMode = E3dConnectINSTANTIATEAS;

			if(_ConnectPickLock == NULL) _ConnectPickLock = E_PluginLockAdd(_Plugin, _PickLockDel);
		}

		EFreeChk(LSelectedObjects);
	}
}


//================================================================================
// Check if any of the Objects in a hierarchy are instantiated more than once
//================================================================================
static EBool _NodeHrcHasInstancing(E3dNode* PNode)
{
	for(;PNode;PNode = PNode->Next)
	{
		E3dModel*  LModel = E_DynamicCast(PNode, E3dModel);

		if(LModel)
		{
			EIndex  LOC = 0;const EIndex  LON = LModel->Objects_Count;for(;LOC < LON;LOC++)
			{
				E3dObject*  LObject = LModel->Objects[LOC];

				if(LObject->Models_Count > 1) return(TRUE);
			}
		}
	}

	return(FALSE);
}


//================================================================================
// Collapse branch
//================================================================================
static E3dNode* _BranchCollapse(E3dNode* PBranchRootNode)
{
	E3dNode*   LNodeNextAfterBranch = E3dNode_BranchGetNodeAfter(PBranchRootNode);
	E3dModel*  LBranchRootModel     = E_DynamicCast(PBranchRootNode, E3dModel);


	if(LBranchRootModel)
	{
		E3dNode**  LNodesToRemove = NULL;EIndex  LNumOfNodesToRemove = 0;
		E3dMatrix  LBranchRootWorldToLocalMatrix;

		E3dMatrix_Invert3x4(LBranchRootModel->LocalToWorldMatrix, LBranchRootWorldToLocalMatrix);


// Disconnect children of BranchRoot from the hierarchy
//
		{
			E3dNode*  LNode = PBranchRootNode->Child;do
			{
				ELst_ObjectAdd(&LNodesToRemove, &LNumOfNodesToRemove, LNode);

				LNode = LNode->Next;
			} while(LNode != LNodeNextAfterBranch);
		}

		{
			EIndex  LC = 0;for(;LC < LNumOfNodesToRemove;LC++)
			{
				E3dNode*  LNode = LNodesToRemove[LC];
				E3dNode_DisconnectFromParent(LNode, FALSE);
			}
		}

		EIndex  LC = 0;for(LC = 0;LC < LNumOfNodesToRemove;LC++)
		{
			E3dNode*   LNode = LNodesToRemove[LC];
			E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
			{
				const EIndex  LObjN = LModel->Objects_Count;if(LObjN)
				{
					E3dObject**  LObjects;ENewN(LObjN, LObjects);
					E3dMatrix    LMatrix;

					memcpy(LObjects, LModel->Objects, sizeof(E3dObject*) * LObjN);

					E3dMatrix_Mult4x4_3(LModel->LocalToWorldMatrix, LBranchRootWorldToLocalMatrix, LMatrix);
					const E3dCoordinate  LMatrix3x3Determinant = E3dMatrix_Determinant3x3(LMatrix);

					EIndex  LObjC = 0;for(;LObjC < LObjN;LObjC++)
					{
						E3dObject*  LObject = LObjects[LObjC];

// If LObject is multi-instantiated, copy it before baking in the transforms
//
						if(LObject->Models_Count > 1)
						{
							E3dObject*  LOldObject = LObject;

							if(EObject_IsSubClass(LObject, E3dGeometry))
							{
								LObject = E3dObject_Copy(LObject, E3dCopyOBJECT_UNIQUE_NAME|E3dCopyEDGES|E3dCopyCOMPONENTS, E3d_Scene);

								LObject->Selection = 0;

								E3dGeometry*  LGeometry = (E3dGeometry*)LObject;
								E3dGeometry_Transform(LGeometry, LMatrix);
								if(LMatrix3x3Determinant < 0.0)
								{
									E3dGeometry_Invert(LGeometry, NULL);
								}
							}

							E3dModel_AppendObject(LBranchRootModel, LObject);
							E3dModel_RemoveObject(LModel, LOldObject);
						}
						else
						{
							if(EObject_IsSubClass(LObject, E3dGeometry))
							{
								E3dGeometry*  LGeometry = (E3dGeometry*)LObject;

								E3dGeometry_Transform(LGeometry, LMatrix);
								if(LMatrix3x3Determinant < 0.0)
								{
									E3dGeometry_Invert(LGeometry, NULL);
								}
							}

							E3dModel_AppendObject(LBranchRootModel, LObject);
							E3dModel_RemoveObject(LModel, LObject);
						}

						if(LObject->Name == NULL) EObject_SetName(LObject, EStrDup(LModel->Name), FALSE);
					}
					EFree(LObjects);
				}
			}
		}

		ELst_ObjectsFree(&LNodesToRemove, &LNumOfNodesToRemove);
	}

	E3dNode_UpdateForDisplay(PBranchRootNode, E3dUF_ALL - (E3dUF_EDGES|E3dUF_VERTEX_NORMAL|E3dUF_POLYGON_NORMAL));

	return(LNodeNextAfterBranch);
}


//================================================================
// Collapse selected branches into one Node per branch
//================================================================
static void _CB_CollapseBranch(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	E3dNode**  LRootNodes = LScene->RootNodes;
	EIndex     LBranchesCollapsed = 0;


	EIndex  LRC = 0;const EIndex  LRN = LScene->RootNodes_Count;for(;LRC < LRN;LRC++)
	{
		E3dNode*  LRootNode = LRootNodes[LRC];


		if(_NodeHrcHasInstancing(LRootNode))
		{
			E3dp_PrintMessage(ELogMESSAGE, 5000, "Branch %s has multi-instantiated Objects. These will be duplicated.", LRootNode->Name);
		}

		E3dNode*  LNode = LRootNode;do
		{
			if((LNode->Selection == E3dNodeSelBRANCH_ROOT) && LNode->Child)
			{
				LNode = _BranchCollapse(LNode);
				LBranchesCollapsed++;
			}
			else LNode = LNode->Next;
		} while(LNode);

		E3dNodeHrc_UpdateConnections(LRootNode);
		E3dNodeHrc_UpdateMatrices(LRootNode, TRUE);
		E3dNodeHrc_CallNodeInfoUpdateProcs(LRootNode, E3dChgHIERARCHY, FALSE);
		E3dNodeHrc_UpdateMaterialInheritance(LRootNode);
	}

	if(LBranchesCollapsed == 0) E3dp_PrintMessage(ELogWARNING, 5000, "No branches selected");
	else
	{
		if(LBranchesCollapsed == 1) E3dp_PrintMessage(ELogMESSAGE, 5000, "Collapsed 1 branch");
		else E3dp_PrintMessage(ELogMESSAGE, 5000, "Collapsed %d branches", LBranchesCollapsed);

		E3dScene_ObjectsUpdateModelList(LScene);
		E3dScene_LayoutOnSchematics(LScene, FALSE);
		E3dWindows_RedrawByChangeFlags(E3dChgHIERARCHY);
		E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	}
}



typedef struct
{
	E3dNode*  Parent;
	EList(, E3dNode*, Nodes);
	EIndex    FirstNodeChildIndex;
} E3dPNode;


static E3dPNode* _FindNode(E3dPNode* PPNode, const EIndex PPNodes_Count, E3dNode* PNode)
{
	E3dNode*  LParent = PNode->Parent;

	EIndex  LC = 0;for(;LC < PPNodes_Count;LC++, PPNode++) if(PPNode->Parent == LParent) return(PPNode);

	return(NULL);
}


//================================================================
// Insert Nodes between the selected Nodes and their parents
//================================================================
static void _CB_AddParent(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	E3dNode**  LRootNodes = LScene->RootNodes;
	E3dPNode*  LPNodes = NULL;EIndex  LPNodes_Count = 0;
	E3dNode**  LAffectedRoots = NULL;EIndex  LNumOfAffectedRoots = 0;


// Collect candidate Nodes for grouping
//
	EIndex  LRC = 0;const EIndex  LRN = LScene->RootNodes_Count;for(;LRC < LRN;LRC++)
	{
		E3dNode*  LRootNode = LRootNodes[LRC];

		E3dNode*  LNode = LRootNode;for(;LNode;LNode = LNode->Next)
		{
			if(LNode->Selection == E3dNodeSelNODE || LNode->Selection == E3dNodeSelBRANCH_ROOT)
			{
				E3dPNode*  LPNode = _FindNode(LPNodes, LPNodes_Count, LNode);if(LPNode == NULL)
				{
					LPNode = EList_StructAdd(LPNodes);
					EList_Init(LPNode->Nodes);
					E_AssignToMemberNoChkNoFree(LNode->Parent, LPNode,Parent);


					if(LNode->Parent)
					{
						E3dNode*  LChild = LNode->Parent->Child;
						EIndex    LIndex = 0;while(LChild != LNode && LChild->NextSibling) { LChild = LChild->NextSibling;LIndex++; }
						LPNode->FirstNodeChildIndex = LIndex;
					}
					else LPNode->FirstNodeChildIndex = EList_PointerGetIndex(LScene->RootNodes, LNode);

					if(LNode->Parent)
					{
						E3dNode*  LRoot = LNode->Parent;while(LRoot->Parent) LRoot = LRoot->Parent;
						ELst_PointerAddChk(&LAffectedRoots, &LNumOfAffectedRoots, LRoot);
					}
				}
				EList_ObjectAdd(LPNode->Nodes, LNode);
			}
		}
	}

	if(LPNodes)
	{
		E3dPNode*     LPNode = LPNodes;
		EChangeFlags  LChanged = 0;

		EIndex  LPC = 0;for(;LPC < LPNodes_Count;LPC++, LPNode++)
		{
			E3dNode*      LParent = LPNode->Parent;
			const EIndex  LN = LPNode->Nodes_Count;

			if(LParent)
			{
				EIndex  LC = 0;for(;LC < LN;LC++)
				{
					E3dNode*  LNode = LPNode->Nodes[LC];

					E3dNode_DisconnectFromParent(LNode, FALSE);
				}
			}

			E3dModel*  LNewParent;E_New(E3dModel, LNewParent);EObject_SetName(LNewParent, E3dScene_GetUniqueNodeName(LScene, "Group"), FALSE);
			if(LParent)
			{
				E3dNode_ConnectChild(LParent, (E3dNode*)LNewParent, LPNode->FirstNodeChildIndex, FALSE, FALSE);
			}
			else
			{
				ELst_PointerAddChk(&LAffectedRoots, &LNumOfAffectedRoots, LNewParent);
				E3dScene_NodeHrcInsert(LScene, (E3dNode*)LNewParent, LPNode->FirstNodeChildIndex, FALSE);
			}


			EIndex  LC = 0;for(;LC < LN;LC++)
			{
				E3dNode*  LNode = LPNode->Nodes[LC];

				E3dScene_NodeHrcRemove(LScene, LNode, &LChanged, FALSE);
				E3dNode_ConnectChild((E3dNode*)LNewParent, LNode, ELstEND, FALSE, FALSE);

				E_Delete(LPNode->Nodes[LC]);
			}
			E_DeleteChk(LPNode->Parent);
		}
		EFree(LPNodes);


		if(LAffectedRoots)
		{
			EIndex  LC = 0;for(;LC < LNumOfAffectedRoots;LC++)
			{
				E3dNodeHrc_UpdateConnections(LAffectedRoots[LC]);
				E3dNodeHrc_UpdateMatrices(LAffectedRoots[LC], TRUE);
			}
			EFree(LAffectedRoots);
		}

		E3dScene_LayoutOnSchematics(LScene, FALSE);
		E3dWindows_RedrawByChangeFlags(E3dChgHIERARCHY);
	}
	else E3dp_PrintMessage(ELogWARNING, 5000, "Select at least one Model.");
}


//================================================================
// Expand selected Nodes in the Schematic view
//================================================================
static void _CB_ExpandAll(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	E3dNode**  LRootNodes = LScene->RootNodes;
	EIndex     LChanged = 0;

// Collect candidate Nodes for grouping
//
	EIndex  LRC = 0;const EIndex  LRN = LScene->RootNodes_Count;for(;LRC < LRN;LRC++)
	{
		E3dNode*  LRootNode = LRootNodes[LRC];

		E3dNode*  LNode = LRootNode;for(;LNode;LNode = LNode->Next)
		{
			if(LNode->Selection)
			{
				if(!LNode->SchemObject.BranchExpanded) { LNode->SchemObject.BranchExpanded = TRUE;LChanged++; }
			}
		}
	}

	if(LChanged)
	{
		E3dScene_LayoutOnSchematics(LScene, FALSE);
		E3dWindows_RedrawByChangeFlags(E3dChgSCHEMATICS);
	}
}


static void _KCB_AddParent(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	_CB_AddParent(NULL, NULL, NULL);
}


//================================================================
// Move Node or Object to the 'left' within the hierarchy
//================================================================
static void _CB_MoveLeft(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	E3dItem    LSelectedItem;

	switch(E3dScene_GetSingleSelection(LScene, &LSelectedItem, E3dItemMODEL|E3dItemANYOBJECT, 0))
	{
		case E3dItemMODEL:
			{
				E3dNode*  LNode = LSelectedItem.ObjItem.Node;

				if(LNode->Parent)
				{
					E3dNode*  LParent = LNode->Parent;

					if(LNode->PrevSibling)
					{
						E_Object(E3dNode, LNode->NextSibling, LNextSibling);  // Temporary references, so these won't get freed
						E_Object(E3dNode, LNode->PrevSibling, LPrevSibling);

						E_AssignToMember(LPrevSibling, LNode,NextSibling);

						if(LPrevSibling->PrevSibling)
						{
							E_AssignToMember(LNode, LPrevSibling->PrevSibling,NextSibling);
						}
						else
						{
							E_AssignToMember(LNode, LParent,Child);
						}

						E_AssignToMember(LNextSibling, LPrevSibling,NextSibling);

						E_DeleteChk(LNextSibling);
						E_DeleteChk(LPrevSibling);

						while(LNode->Parent) LNode = LNode->Parent;
						E3dNodeHrc_UpdateConnections(LNode);

						E3dScene_ObjectsUpdateModelList(LScene);
						E3dScene_LayoutOnSchematics(LScene, FALSE);

						E3dWindows_RedrawByChangeFlags(E3dChgHIERARCHY);
					}
					else E3dp_PrintMessage(ELogWARNING, 5000, "This Model is already the first child of its parent");
				}
				else
				{
// It's a root Node
//
					EIndex  LIndex = EList_PointerGetIndex(LScene->RootNodes, LNode);if(LIndex > 0)
					{
						E_Object(E3dNode, LScene->RootNodes[LIndex], LTNode);  // Temporary reference, so LScene->RootNodes[LIndex] doesn't get freed

						E_AssignToMember(LScene->RootNodes[LIndex - 1], LScene,RootNodes[LIndex]);
						E_AssignToMember(LTNode, LScene,RootNodes[LIndex - 1]);
						E_Delete(LTNode);

						E3dScene_ObjectsUpdateModelList(LScene);
						E3dScene_LayoutOnSchematics(LScene, FALSE);

						E3dWindows_RedrawByChangeFlags(E3dChgHIERARCHY);
					}
					else E3dp_PrintMessage(ELogWARNING, 5000, "This Model is already the first root in the Scene");
				}
			}
		break;

		case E3dItemOBJECT:
			{
				E3dObject*  LObject = LSelectedItem.ObjItem.Object;
				E3dModel*   LModel = LObject->Models[0];
				EIndex      LIndex = EList_PointerGetIndex(LModel->Objects, LObject);if(LIndex > 0)
				{
					E_Object(E3dObject, LModel->Objects[LIndex], LTObject);  // Temporary reference, so LModel->Objects[LIndex] doesn't get freed

					E_AssignToMember(LModel->Objects[LIndex - 1], LModel,Objects[LIndex]);
					E_AssignToMember(LTObject, LModel,Objects[LIndex - 1]);
					E_Delete(LTObject);

					E3dScene_ObjectsUpdateModelList(LScene);
					E3dScene_LayoutOnSchematics(LScene, FALSE);

					E3dWindows_RedrawByChangeFlags(E3dChgHIERARCHY);
				}
				else E3dp_PrintMessage(ELogWARNING, 5000, "This Object is already the first one under its (first) Model");
			}
		break;

		default:
		 E3dp_PrintMessage(ELogWARNING, 5000, "Select a Model or an Object");
		break;
	}
}


//================================================================
// Move Node or Object to the 'right' within the hierarchy
//================================================================
static void _CB_MoveRight(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	E3dItem    LSelectedItem;

	switch(E3dScene_GetSingleSelection(LScene, &LSelectedItem, E3dItemMODEL|E3dItemANYOBJECT, 0))
	{
		case E3dItemMODEL:
			{
				E3dNode*  LNode = LSelectedItem.ObjItem.Node;

				if(LNode->Parent)
				{
					E3dNode*	LParent = LNode->Parent;

					if(LNode->NextSibling)
					{
						E_Object(E3dNode, LNode->NextSibling, LNextSibling);	// Temporary references, so these won't get freed
						E_Object(E3dNode, LNode->PrevSibling, LPrevSibling);

						E_AssignToMember(LNextSibling->NextSibling, LNode,NextSibling);
						E_AssignToMember(LNode, LNextSibling,NextSibling);

						if(LPrevSibling)
						{
							E_AssignToMember(LNextSibling, LPrevSibling,NextSibling);
						}
						else
						{
							E_AssignToMember(LNextSibling, LParent,Child);
						}

						E_DeleteChk(LNextSibling);
						E_DeleteChk(LPrevSibling);

						while(LNode->Parent) LNode = LNode->Parent;
						E3dNodeHrc_UpdateConnections(LNode);

						E3dScene_ObjectsUpdateModelList(LScene);
						E3dScene_LayoutOnSchematics(LScene, FALSE);

						E3dWindows_RedrawByChangeFlags(E3dChgHIERARCHY);
					}
					else E3dp_PrintMessage(ELogWARNING, 5000, "This Node is already the last child of its parent");
				}
				else
				{
// It's a root Node
//
					EIndex  LIndex = EList_PointerGetIndex(LScene->RootNodes, LNode);

					if(LIndex < LScene->RootNodes_Count - 1)
					{
						E_Object(E3dNode, LScene->RootNodes[LIndex], LTNode);	// Temporary reference, so LScene->RootNodes[LIndex] doesn't get freed

						E_AssignToMember(LScene->RootNodes[LIndex + 1], LScene,RootNodes[LIndex]);
						E_AssignToMember(LTNode, LScene,RootNodes[LIndex + 1]);
						E_Delete(LTNode);

						E3dScene_ObjectsUpdateModelList(LScene);
						E3dScene_LayoutOnSchematics(LScene, FALSE);

						E3dWindows_RedrawByChangeFlags(E3dChgHIERARCHY);
					}
					else E3dp_PrintMessage(ELogWARNING, 5000, "This Model is already the last root in the Scene");
				}
			}
		break;

		case E3dItemOBJECT:
			{
				E3dObject*  LObject = LSelectedItem.ObjItem.Object;
				E3dModel*   LModel = LObject->Models[0];
				EIndex      LIndex = EList_PointerGetIndex(LModel->Objects, LObject);

				if(LIndex < LModel->Objects_Count - 1)
				{
					E_Object(E3dObject, LModel->Objects[LIndex], LTObject);  // Temporary reference, so LModel->Objects[LIndex] doesn't get freed

					E_AssignToMember(LModel->Objects[LIndex + 1], LModel,Objects[LIndex]);
					E_AssignToMember(LTObject, LModel,Objects[LIndex + 1]);
					E_Delete(LTObject);

					E3dScene_ObjectsUpdateModelList(LScene);
					E3dScene_LayoutOnSchematics(LScene, FALSE);

					E3dWindows_RedrawByChangeFlags(E3dChgHIERARCHY);
				}
				else E3dp_PrintMessage(ELogWARNING, 5000, "This Object is already the last one under its (first) Model");
			}
		break;

		default:
			E3dp_PrintMessage(ELogWARNING, 5000, "Select a Model or an Object");
		break;
	}
}


static void _KCB_MoveLeft(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	_CB_MoveLeft(NULL, NULL, NULL);
}


static void _KCB_MoveRight(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	_CB_MoveRight(NULL, NULL, NULL);
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	static EOperatorClass  LOperatorEditHierarchy_classDesc[] =
	{
		{
			.Name=        "Edit hierarchies",
			.StructSize=  sizeof(E3dOperatorEditHierarchy),
			.Init=        _EditHierarchy_Init,
			.Free=        _EditHierarchy_Free,
			.Proc=        _EditHierarchy_Proc
		}
	};
	E3dOperatorEditHierarchy_class  = EOp_OperatorClassRegister(LOperatorEditHierarchy_classDesc);


	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Connect Model, branch or Object", '\0', NULL, NULL, TRUE, NULL, _CB_Connect, (EPointer)NULL);
	Xe_PushButtonAddImg3_sp(_ToolBarButtons[_ToolBarButtons_Count++], Connect, "Tool->Hierarchy", "Connect", '\0', NULL, NULL, FALSE, "Connect selected Model(s) to a hierarchy or selected Objects (Geometry, Light etc.) to a Model\nRight mouse button: settings for this tool", _CB_Connect, (EPointer)NULL);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Disconnect Model, branch or Object", '\0', NULL, NULL, TRUE, NULL, _CB_Disconnect, (EPointer)NULL);
	Xe_PushButtonAddImg3_sp(_ToolBarButtons[_ToolBarButtons_Count++], Disconnect, "Tool->Hierarchy", "Disconnect", '\0', NULL, NULL, FALSE, "Disconnect selected Model(s) from their parent(s)\nRight mouse button: settings for this tool", _CB_Disconnect, (EPointer)NULL);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Reparent Object", '\0', NULL, NULL, TRUE, NULL, _CB_Reparent, (EPointer)NULL);
	Xe_PushButtonAddImg3_sp(_ToolBarButtons[_ToolBarButtons_Count++], Reparent, "Tool->Hierarchy", "Reparent", '\0', NULL, NULL, FALSE, "Reparent: Disconnect selected Models or Objects from their parent and move them under a different Model\nRight mouse button: settings for this tool", _CB_Reparent, (EPointer)NULL);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Separate Objects", '\0', NULL, NULL, FALSE, NULL, _MCB_SeparateSelectedObjects, (EPointer)0);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Add Object to same Nodes as...", '\0', NULL, NULL, FALSE, NULL, _CB_Reinstantiate, (EPointer)NULL);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Collapse branch", '\0', NULL, NULL, FALSE, NULL, _CB_CollapseBranch, (EPointer)0);

	const EKeyModifiers  LCtrlMask = Xe_CommandKeyForCtrl ? E_CommandModifierMask : EControlMask;const char*  LCtrlName = Xe_CommandKeyForCtrl ? "Cmd" : "Ctrl";char                 LAccText[256];

	sprintf(LAccText, "%s+left", LCtrlName);_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Move Model / Object 'left'", '\0', NULL, LAccText, FALSE, "Move Model or Object to the 'left' within the hierarchy", _CB_MoveLeft, (EPointer)NULL);
	E_KeySetCallback("Move Model or Object to the 'left' within the hierarchy",  EKeyLEFT,  LCtrlMask, EKeyPRESS, _KCB_MoveLeft, (EPointer)NULL);

	sprintf(LAccText, "%s+right", LCtrlName);_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Move Model / Object 'right'", '\0', NULL, LAccText, FALSE, "Move Model or Object to the 'right' within the hierarchy", _CB_MoveRight, (EPointer)NULL);
	E_KeySetCallback("Move Model or Object to the 'right' within the hierarchy", EKeyRIGHT, LCtrlMask, EKeyPRESS, _KCB_MoveRight, (EPointer)NULL);

	sprintf(LAccText, "%s+G", LCtrlName);_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Group objects", '\0', NULL, LAccText, FALSE, "Add new parent Model to the selected Model.", _CB_AddParent, (EPointer)0);
	E_KeySetCallback("Group objects", 'g',                                                 LCtrlMask, EKeyPRESS, _KCB_AddParent, (EPointer)NULL);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Hierarchy", "Expand selected", '\0', NULL, NULL, FALSE, "Expand all selected Nodes", _CB_ExpandAll, (EPointer)NULL);
//	E_KeySetCallback("Expand selected", EKeyRIGHT, EControlMask, EKeyPRESS, _KCB_ExpandAll, (EPointer)NULL);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC;

		for(LC = 0;LC < _MenuButtons_Count;LC++) { XeItemFree(_MenuButtons[LC]); }
		_MenuButtons_Count = 0;

		for(LC = 0;LC < _ToolBarButtons_Count;LC++) { XeItemFree(_ToolBarButtons[LC]); }
		_ToolBarButtons_Count = 0;
	}

	EpIconFree3(Connect);
	EpIconFree3(Disconnect);
	EpIconFree3(Reparent);

	const EKeyModifiers  LCtrlMask = Xe_CommandKeyForCtrl ? E_CommandModifierMask : EControlMask;
	E_KeyRemoveCallback(EKeyLEFT,  LCtrlMask, EKeyPRESS, _KCB_MoveLeft, (EPointer)NULL);
	E_KeyRemoveCallback(EKeyRIGHT, LCtrlMask, EKeyPRESS, _KCB_MoveRight, (EPointer)NULL);
	E_KeyRemoveCallback('g',       LCtrlMask, EKeyPRESS, _KCB_AddParent, (EPointer)NULL);

	if(E3dOperatorEditHierarchy_class) { EOp_OperatorClassDeactivate(E3dOperatorEditHierarchy_class);E3dOperatorEditHierarchy_class = NULL; }

	return(E_SUCCESS);
}
