//================================================================
// EQUINOX-3D plugin
//
// Information about the selected Models and Objects.
// This plugin is great for debugging things in EQUINOX-3D, as well
// as for learning how objects are stored inside EQUINOX-3D.
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 11:15:03
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/PushButton.h>

#include <E3D/Matrix.h>
#include <E3D/Scene.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/StatusPanel.h>


static EguiItem  _MenuButtons[16];static int     _MenuButtons_Count = 0;
static EguiItem  _ToolBarButtons[16];static int  _ToolBarButtons_Count = 0;


static EBool     _E3dp_ShowMaterials = TRUE;

enum
{
	CR_INFO_ON_SELECTION,
	CR_INFO_LIST_ROOTS,
	CR_INFO_LIST_NODES,
	CR_INFO_LIST_MATERIALS,
	CR_INFO_LIST_ANIMATIONS
};


//========================================
// Print information about a Spline
//========================================
static void _SplinePrintInfo(E3dSpline* PSpline)
{
	const EIndex  LN = PSpline->CVs_Count;

	printf(" %sSpline%s: CVs: %lld  Type: ", ECyanCode, ENormalCode, (long long)LN);

	switch(PSpline->SplineType)
	{
		case E3dSplineBEZIER:
			printf("BEZIER\n");fflush(stdout);
		break;

		case E3dSplineBSPLINE:
			printf("B-SPLINE\n");fflush(stdout);
		break;

		case E3dSplineCARDINAL:
			printf("CARDINAL\n");fflush(stdout);
		break;

		case E3dSplineLINEAR:
			printf("LINEAR\n");fflush(stdout);
		break;
	}

	switch(PSpline->SplineType)
	{
		case E3dSplineLINEAR:
			{
				E3dSplineCV*  LSplineCV = (E3dSplineCV*)(PSpline->CVs);

				EIndex  LC = 0;for(;LC < LN;LC++, LSplineCV++)
				{
					printf(" Position: %f,%f,%f\n", LSplineCV->Position.X, LSplineCV->Position.Y, LSplineCV->Position.Z);
				}
			}
		break;
	}
	fflush(stdout);
}


//================================================
// Accumulate the number of Polygons of a Mesh
//================================================
static void _MeshAddPolyCounts(E3dMesh* PMesh, EIndex* PTotalPolyNumP, EIndex* PSelectedPolyNumP)
{
	EIndex  LPGC = 0;for(;LPGC < PMesh->PolyGroups_Count;LPGC++)
	{
		E3dPolyGroup*  LPolyGroup = PMesh->PolyGroups[LPGC];

		if((PMesh->Selection == E3dObjectSelPOLYGROUP && LPolyGroup->Selected) || PMesh->Selection != E3dObjectSelPOLYGROUP)
		{
			(*PTotalPolyNumP) += LPolyGroup->Polygons_Count;

			const EIndex  LPN = LPolyGroup->Polygons_Count;if(LPN)
			{
				EIndex  LPC = 0;const E3dPolygon*  LPolygon = LPolyGroup->Polygons;for(;LPC < LPN;LPC++, LPolygon++)
				{
					if(LPolygon->VertexNodes)
					{
						if(LPolygon->Flags & E3dPolygonSELECTED)
						{
							(*PSelectedPolyNumP)++;
						}
					}
				}
			}
		}
	}
}


//========================================
// Print information about a Mesh
//========================================
static void _MeshPrintInfo(E3dMesh* PMesh, EIndex* PTotalPolyNumP, EIndex* PSelectedPolyNumP)
{
	printf(" %sMesh%s: vertices: %lld, PolyGroup(s): %lld\n {\n", ECyanCode, ENormalCode, (long long)PMesh->Vertices_Count, (long long)(PMesh->PolyGroups_Count));

// VertexAttributes
//
	{
		printf("  VertexAttributes\n");fflush(stdout);
		EIndex  LC = 0;const EIndex  LN = PMesh->VertexAttributes.Count;const EResource*  LAttr = PMesh->VertexAttributes.Items;for(;LC < LN;LC++, LAttr++)
		{
			printf("   Name %-8s  Semantic  %-8s\n", LAttr->Name, (char*)LAttr->Semantic);fflush(stdout);
		}
	}

	EIndex  LPGC = 0;for(;LPGC < PMesh->PolyGroups_Count;LPGC++)
	{
		E3dPolyGroup*  LPolyGroup = PMesh->PolyGroups[LPGC];

		if(LPolyGroup->Selected || PMesh->Selection == E3dObjectSelOBJECT)
		{
			*PTotalPolyNumP += LPolyGroup->Polygons_Count;

			printf("  PolyGroup #%lld <<%s%s%s>>:  Selected %d  Polygons: %lld  TriangleStrips: %lld  Edges: %lld\n  {\n", (long long)LPGC,  EGreenCode, LPolyGroup->Name ? LPolyGroup->Name : "", ENormalCode,  LPolyGroup->Selected,  (long long)(LPolyGroup->Polygons_Count), (long long)(LPolyGroup->TriangleStrips_Count), (long long)(LPolyGroup->Edges_Count));
			printf("   TxMappers %d  MapperIndices %d\n", (int)LPolyGroup->TxMappers_Count, (int)LPolyGroup->TextureMapperIndices_Count);

			printf("   Use vertex-normals: ");
			if(LPolyGroup->Flags & E3dUSE_VERTEX_NORMALS) printf("YES\n");
			else printf("NO\n");
			fflush(stdout);

			printf("   Vertex-normal type: ");
			switch(LPolyGroup->VertexNormalType)
			{
				case 0:
					printf("NONE");
				break;

				case E3dNormalPROCESS:
					printf("FROM PROCESS");
				break;

				case E3dNormalFROM_SMOOTHING_ANGLE:
					printf("FROM DISCONTINUITY ANGLE: %f degrees\n", LPolyGroup->SmoothingAngle);
				break;
			}
			printf("\n");


// VertexNodeAttributes
//
			{
				printf("   VertexNodeAttributes\n");fflush(stdout);
				EIndex  LC = 0;const EIndex  LN = LPolyGroup->VertexAttributes.Count;const EResource*  LAttr = LPolyGroup->VertexAttributes.Items;for(;LC < LN;LC++, LAttr++)
				{
					printf("    Name %-8s  Semantic  %-8s\n", LAttr->Name, (char*)LAttr->Semantic);fflush(stdout);
				}
			}


			if(_E3dp_ShowMaterials)
			{
				E3dMaterial*  LMaterial = LPolyGroup->Material;
				if(LMaterial == NULL)
				{
					LMaterial = LPolyGroup->DrawingMaterial;
					if(LMaterial == E3d_DefaultMaterial) printf("\n   No Material (using default)\n");
					else
					{
						printf("\n   Inherited Material\n");fflush(stdout);
						E3dMaterial_Print(LMaterial, 4);
					}
				}
				else
				{
					printf("\n   PolyGroupMaterial:\n");E3dMaterial_Print(LMaterial, 4);
				}
			}
			fflush(stdout);


			const EIndex  LPN = LPolyGroup->Polygons_Count;if(LPN)
			{
				const E3dVertex*         LVertices = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;
				const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

				EIndex  LPC = 0;const E3dPolygon*  LPolygon = LPolyGroup->Polygons;for(;LPC < LPN;LPC++, LPolygon++)
				{
					EIndex  LNonTris = 0;
					if(LPolygon->VertexNodes)
					{
						if(LPolygon->Flags & E3dPolygonSELECTED)
						{
							printf("   Polygon % 3d:\n", (int)LPC);fflush(stdout);
							EIndex  LVC = 0;const E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;for(;LVC < LPolygon->NumOfVertexNodes;LVC++)
							{
								const E3dVertexIndex  LVID = LVertexNode->VertexID;if(LVID >= 0)
								{
									const E3dVertex*  LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVID);
									printf("    %s% 3lld%s : % f, % f, % f\n", ECyanCode, (long long)LVID, ENormalCode, XYZ_s(LVertex->Position));
								}
								else
								{
									printf("    Hole with %lld vertices:\n", -(long long)LVID);fflush(stdout);
								}

								EStruct_Inc(LVertexNode, LVertexNodeSize);
							}

							(*PSelectedPolyNumP)++;
						}

						if(LPolygon->NumOfVertices == 3)
						{
						}
						else LNonTris++;
					}
					else
					{
						printf("%s() %s:%d  %sNO VERTEXNODES%s\n", __FUNCTION__, __FILE__, __LINE__, ERedCode, ENormalCode);fflush(stdout);
					}
				}
			}

/*
// Pring GLArray
//
			if(LPolyGroup->GLArrayValid)
			{
				float*  LValues = (float*)(LPolyGroup->GLArray);

				EIndex  LTC = 0;const EIndex  LVN = LPolyGroup->GLArraySize / LPolyGroup->GLArrayVertexNodeSize;for(;LTC < LTN;LTC++)
				{
					EIndex  LVC = 0;const EInde  LTN = LVN / 3;for(;LVC < 3;LVC++)
					{
						EIndex  LFC = 0;const EIndex  LFN = LPolyGroup->GLArrayVertexNodeSize / sizeof(float);for(;LFC < LFN;LFC++)
						{
							printf(" %f", *(LValues++));
						}
						printf("\n");fflush(stdout);
					}
					printf("\n");fflush(stdout);
				}
			}
*/

			printf("  }\n");  // Close PolyGroup
		}
	}


// Show info about selected Edges
//
	if(E3dMesh_NumOfSelectedEdges(PMesh) > 0)
	{
		EBool  LWingedEdgesCreated = FALSE;

		if(PMesh->WingedEdges == NULL) { E3dMesh_CreateWingedEdgesFromEdges(PMesh, FALSE, FALSE);LWingedEdgesCreated = TRUE; }

		if(PMesh->WingedEdges)
		{
			EIndex  LEC = 0;const EIndex  LEN = PMesh->Edges_Count;const E3dEdge*  LEdge = PMesh->Edges;const E3dWingedEdge*  LWingedEdge = PMesh->WingedEdges;for(;LEC < LEN;LEC++, LEdge++, LWingedEdge++)
			{
				if(LEdge->Flags & E3dEdgeSELECTED)
				{
					printf("  Edge  % 5lld  Polygons: %d", (long long)LEC, LWingedEdge->Polygons_Count);
/*
					int32_t  LPC = 0;const int32_t  LPN = LWingedEdge->Polygons_Count;const E3dEdgePolygonInfo*  LEdgePolygonInfo = LWingedEdge->Polygons;for(;LPC < LPN;LPC++, LEdgePolygonInfo++)
					{
						printf(" ", LEdgePolygonInfo);fflush(stdout);
					}
*/
					printf("\n");fflush(stdout);
				}
			}
		}

		if(LWingedEdgesCreated) E3dMesh_WingedEdgesFree(PMesh);

		printf("\n");fflush(stdout);
	}


	if(E3dMesh_Check(PMesh, 2) == E_SUCCESS) { printf("  E3dMesh_Check() %sOK%s\n", EGreenCode, ENormalCode);fflush(stdout); }

	printf(" }\n");	// Close Mesh
}


//================================================================
// Print information about a specified link member of an E3dNode
//================================================================
#define _M_NodeLinkPrint(mLink)\
{\
	if(LNode->mLink)\
	{\
		printf(" %s: [%s]", #mLink, LNode->mLink->Name);fflush(stdout);\
	}\
}


//================================================================
// Print information about an E3dObject
//================================================================
static void _ObjectInfo(const E3dObject* PObject, EIndex* PTotalPolyNumP, EIndex* PSelectedPolyNumP)
{
	if(EObject_IsSubClass(PObject, E3dGeometry))
	{
		E3dGeometry*  LGeometry = (E3dGeometry*)PObject;

		printf(" NumOfModels: %lld", (long long)(PObject->Models_Count));
		printf(" Selection: %08x", PObject->Selection);fflush(stdout);
		switch(PObject->Selection)
		{
			case E3dObjectSelOBJECT:          printf(" (OBJECT)");break;
			case E3dObjectSelPOLYGROUP:       printf(" (POLYGROUP)");break;
			case E3dObjectSelSPLINE_SEGMENT:  printf(" (SPLINE_SEGMENT)");break;
			case E3dObjectSelFACE_CONTOUR:    printf(" (FACE_CONTOUR)");break;
			default:  break;
		}
		printf("\n");fflush(stdout);


		{
			printf(" Inputs:  %2lld:", (long long)PObject->Inputs_Count);
			EIndex  LC = 0;for(;LC < PObject->Inputs_Count;LC++)
			{
				const EInterface*  LInterface = PObject->Inputs[LC];
				printf("  %p", LInterface);
				if(LInterface->Object) printf(" <%s>", LInterface->Object->Name ? LInterface->Object->Name : "-");
				if(LInterface->Internal) printf(" (Internal)");
			}
			printf("\n");fflush(stdout);
		}


		{
			printf(" Outputs: %2lld:", (long long)PObject->Outputs_Count);
			EIndex  LC = 0;for(;LC < PObject->Outputs_Count;LC++)
			{
				const EInterface*  LInterface = PObject->Outputs[LC];
				printf("  %p", LInterface);
				if(LInterface->Object) printf(" <%s>", LInterface->Object->Name ? LInterface->Object->Name : "-");
				if(LInterface->Internal) printf(" (Internal)");
			}
			printf("\n");fflush(stdout);
		}


		if(EObject_IsSubClass(PObject, E3dMesh))
		{
			E3dMesh*  LMesh = (E3dMesh*)LGeometry;

			_MeshPrintInfo(LMesh, PTotalPolyNumP, PSelectedPolyNumP);
		}
		else if(EObject_IsSubClass(PObject, E3dSpline))
		{
			_SplinePrintInfo((E3dSpline*)PObject);
		}
	}

	if(PObject->Components_Count)
	{
		const EIndex  LIN = PObject->Components_Count;

		printf(" Components: %lld\n {\n", (long long)LIN);fflush(stdout);

		EIndex  LIC = 0;for(;LIC < LIN;LIC++)
		{
			E3dObjectComponent*  LInfo = PObject->Components[LIC];

			printf("  %s%p%s  Class {%s%s%s}\n", ECyanCode, LInfo, ENormalCode, EGreenCode, LInfo->Class->Name, ENormalCode);fflush(stdout);

			if(LInfo->Class->Resources)
			{
				printf("  {\n");fflush(stdout);
				ERes_ObjectPrintResources(LInfo, LInfo->Class->Resources, 3);
				printf("  }\n");fflush(stdout);
			}
		}
		printf(" }\n");fflush(stdout);
	}
}


//========================================
// Info about the selection
//========================================
static void _SelInfo(E3dScene* PScene)
{
	EIndex  LTotalPolyNum = 0, LSelectedPolyNum = 0;


// Collect selected Objects
// Note that E3dScene_GetObjects() collects a list with non-repeating entries (even if an Object is instantiated from multiple Models)
//
	{
		E3dObject**  LObjects = NULL;const EIndex	 LNumOfObjects = E3dScene_GetObjects(PScene, &LObjects, E3dItemANYOBJECT, E3dSELECTED, FALSE);if(LNumOfObjects)
		{
			EIndex  LObjC = 0;for(;LObjC < LNumOfObjects;LObjC++)
			{
				E3dObject*  LObject = LObjects[LObjC];

// Check if the Object directly, or because one of its instantiating Models are selected.
// Only print info if it's a direct selection, because we list selected Models separately, later.
//
//				if(LObject->Selection == E3dObjectSelOBJECT)
				if(LObject->Selection)
				{
					printf(" Object #%2lld <%s%p%s:%s::%s%s%s>\n", (long long)LObjC, ECyanCode, LObject, ENormalCode, LObject->Class->Name, EGreenCode, LObject->Name, ENormalCode);fflush(stdout);
					_ObjectInfo(LObject, &LTotalPolyNum, &LSelectedPolyNum);
				}
/*
				else if(LObject->Selection == E3dObjectSelPOLYGROUP)
				{
printf("%s() %s:%d  %sIMPLEMENT ME%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
				}
*/
			}

// We passed FALSE for PAddReferences, to E3dScene_GetObjects(), so the Objects in the LObjects array are not reference-assigned.
// This means, we can just shallow-free the LObjects.
// Otherwise, we would have to use EList_ObjectsFree().
// You should do the latter method, if you expect that other functions might release/free one or more of those Objects (via E_Delete), while you have a pointer to them.
//
			EFree(LObjects);
		}
	}

// Walk through the Node hierarchies of PScene
//
	EIndex  LRC = 0;const EIndex  LRN = PScene->RootNodes_Count;for(;LRC < LRN;LRC++)
	{
		E3dNode*  LNode = PScene->RootNodes[LRC];for(;LNode;LNode = LNode->Next)
		{
			const EBool  LModelSelected = E3dNode_IsSelected(LNode);if(LModelSelected)
			{
				printf("{%s} [%s%p%s:%s%s%s]  RefCnt: %d\n", LNode->Class->Name, ECyanCode, LNode, ENormalCode, EGreenCode, LNode->Name, ENormalCode, LNode->RefCnt);
				printf(" Selection: %08x", LNode->Selection);fflush(stdout);
				switch(LNode->Selection)
				{
					case E3dNodeSelOBJECT:       printf(" (OBJECT)");break;
					case E3dNodeSelNODE:         printf(" (NODE)");break;
					case E3dNodeSelBRANCH:       printf(" (BRANCH)");break;
					case E3dNodeSelBRANCH_ROOT:  printf(" (BRANCH_ROOT)");break;
					default:  break;
				}
				printf("\n");fflush(stdout);


// Print hierarchy link members of LNode
//
				_M_NodeLinkPrint(Parent);
				_M_NodeLinkPrint(Child);
				_M_NodeLinkPrint(Prev);
				_M_NodeLinkPrint(Next);
				_M_NodeLinkPrint(PrevSibling);
				_M_NodeLinkPrint(NextSibling);

// Print info about the Component objects of LNode
//
				if(LNode->Components_Count)
				{
					const EIndex  LIN = LNode->Components_Count;

					printf(" Components: %d\n", (int)LIN);fflush(stdout);
					EIndex  LIC = 0;for(;LIC < LIN;LIC++)
					{
						E3dNodeComponent*  LInfo = LNode->Components[LIC];

						const E3dNodeComponentClass*  LClasses[EMAX_CLASS_DEPTH];
						EIndex                        LCC = 0;

// Collect superclasses
//
						{
							const E3dNodeComponentClass*  LClass = LInfo->Class;
							while(LClass)
							{
								LClasses[LCC] = LClass;
								LClass = LClass->SuperClass;
								LCC++;
							}
						}

						printf(" {\n");fflush(stdout);

						do
						{
							const E3dNodeComponentClass*  LClass = LClasses[LCC - 1];

							printf("  %s%p%s Class {%s%s%s}\n", ECyanCode, LInfo, ENormalCode, EGreenCode, LClass->Name, ENormalCode);fflush(stdout);
							if(LClass->Resources)
							{
								printf("  {\n");fflush(stdout);
								ERes_ObjectPrintResources(LInfo, LClass->Resources, 3);
								printf("  }\n");fflush(stdout);
							}
						} while(--LCC);
						printf(" }\n");fflush(stdout);
					}
				}

/*
				if(_E3dp_ShowMaterials)
				{
					LMaterial = E3dNode_InheritMaterial(LNode);

					if(LModel->Material) { printf("   ModelMaterial:\n");E3dMaterial_Print(LMaterial, 4); }
					else
					{
						if(LMaterial == E3d_DefaultMaterial) printf("   No Material (using default)\n");
						else { printf("   Inherited ModelMaterial:\n");E3dMaterial_Print(LMaterial, 4); }
					}
				}
*/


// Check if LNode is a Model - note the dynamic cast
//
				E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel && LModel->Objects_Count)
				{
					E3dObject**  LObjects = LModel->Objects;

					printf("Objects: %lld\n{\n", (long long)(LModel->Objects_Count));

					EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
					{
						E3dObject*  LObject = LObjects[LObjC];

						printf(" Object #%2lld <%s%p%s:%s::%s%s%s>\n", (long long)LObjC, ECyanCode, LObject, ENormalCode, LObject->Class->Name, EGreenCode, LObject->Name, ENormalCode);fflush(stdout);

						if(EObject_IsSubClass(LObject, E3dMesh))
						{
							E3dMesh*  LMesh = (E3dMesh*)LObject;

							_MeshAddPolyCounts(LMesh, &LTotalPolyNum, &LSelectedPolyNum);
						}
					}

					printf("}");  // Close Object list
				}

				printf("\n");  // Close Model
				if(LModel->Next) printf("\n");
			}
		}
	}
	printf("Total # of Polygons: %lld  selected: %lld\n", (long long)LTotalPolyNum, (long long)LSelectedPolyNum);fflush(stdout);
}


//========================================
// List RootNodes
//========================================
static void _RootsInfo(const E3dScene* PScene)
{
printf("RootNodes:\n");fflush(stdout);
	EIndex  LC = 0;const EIndex  LN = PScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*  LNode = PScene->RootNodes[LC];
printf("[%s] %p RefCnt: %d\n", LNode->Name, LNode, LNode->RefCnt);fflush(stdout);
	}
}


//================================================================================
// Get the next Node of a Node, based on depth-first walk of the hierarchy
//================================================================================
E3dNode* E3dNodeHrc_NextNodeDepthFirstD(E3dNode* PNode, int* PDepth)
{
	if(PNode->Child) { PNode = PNode->Child;*PDepth += 1; }  // Does the current Node have a child ? (1)
	                                                         // (1Y) Yes, let the child be the next...
	else                                                     // (1N) No, does it have a sibling ? (2)
	{
		if(PNode->NextSibling) PNode = PNode->NextSibling;     // (2Y) Yes, let the sibling be the next...
		else                                                   // (2N) No, go back upward in this branch until
		{                                                      // we find a Node that has a sibling
			while((PNode->NextSibling == NULL) && PNode->Parent)
			{
				PNode = PNode->Parent;*PDepth -= 1;
			}
			if(PNode) PNode = PNode->NextSibling;
		}
	}
	return(PNode);
}


//========================================
// List all Nodes
//========================================
static void _NodesInfo(const E3dScene* PScene)
{
	printf("Nodes:\n");fflush(stdout);
	EIndex  LC = 0;const EIndex  LN = PScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*  LNode = PScene->RootNodes[LC];

		int  PDepth = 0;

		if(LC > 0) { printf("\n");fflush(stdout); }

		while(LNode)
		{
			char  LIndent[256];memset(LIndent, ' ', PDepth);LIndent[PDepth] = 0;

			printf("%s[%p:%s%s%s]\n", LIndent, LNode, ECyanCode, LNode->Name, ENormalCode);fflush(stdout);

//			const int  LOldDepth = PDepth;
			LNode = E3dNodeHrc_NextNodeDepthFirstD(LNode, &PDepth);
		}
	}
}


//========================================
// List Materials
//========================================
static void _MaterialsInfo(const E3dScene* PScene)
{
	printf("Materials:");fflush(stdout);

	EIndex  LC = 0;const EIndex  LN = PScene->Materials_Count;for(;LC < LN;LC++)
	{
		E3dMaterial*  LMaterial = PScene->Materials[LC];

		printf("\n");fflush(stdout);
		E3dMaterial_Print(LMaterial, 1);
	}

	printf("PScene->Materials_Count: %lld\n", (long long)(PScene->Materials_Count));fflush(stdout);
}


//========================================
// List Animations
//========================================
static void _AnimationsInfo(const E3dScene* PScene)
{
	printf("Animations:");fflush(stdout);

	EIndex  LASC = 0;const EIndex  LASN = PScene->AnimStacks_Count;for(;LASC < LASN;LASC++)
	{
		E3dAnimStack*  LAnimStack = PScene->AnimStacks[LASC];
		printf("Stack [%s] animations: %lld\n", LAnimStack->Name, (long long)(LAnimStack->Animations_Count));fflush(stdout);
		EIndex  LC = 0;const EIndex  LN = LAnimStack->Animations_Count;for(;LC < LN;LC++)
		{
			E3dAnimation*  LAnimation = LAnimStack->Animations[LC];E3dKeyFrameAnimation*  LKeyFrameAnimation = E_DynamicCast(LAnimation, E3dKeyFrameAnimation);
			printf(" |%s| {%s} Channels_Count %lld\n", LAnimation->Name, LAnimation->Class->Name, (long long)LAnimation->Channels_Count);fflush(stdout);

			EIndex  LCC = 0;const EIndex  LCN = LAnimation->Channels_Count;for(;LCC < LCN;LCC++)
			{
				E3dAnimationChannel*  LChannel = LAnimation->Channels[LCC];
				printf("%s() %s:%d  %3lld %s", __FUNCTION__, __FILE__, __LINE__, (long long)LCC, LChannel->Name);fflush(stdout);
				if(LKeyFrameAnimation)
				{
					printf(" Curve %p\n", ((E3dKeyFrameAnimationChannel*)LChannel)->Curve);fflush(stdout);
				}
				printf("\n");fflush(stdout);
			}
		}
	}
}


//========================================
// Append a number string to POutStr
//========================================
static void _AddCountString(const char* PName, const int PCount, char* POutStr)
{
	if(PCount)
	{
		char  LTmpStr[1024];

		if(PCount > 1) sprintf(LTmpStr, " %d %ss", PCount, PName);
		else sprintf(LTmpStr, " 1 %s", PName);
		strcat(POutStr, LTmpStr);
	}
}


//========================================
// Menu callback
//========================================
static void _MCB_Info(Widget PW, EPointer PClientData, EPointer PCallData)
{
	switch((EPointerInt)PClientData)
	{
		case CR_INFO_ON_SELECTION:
			_SelInfo(E3d_Scene);
		break;

		case CR_INFO_LIST_ROOTS:
			_RootsInfo(E3d_Scene);
		break;

		case CR_INFO_LIST_NODES:
			_NodesInfo(E3d_Scene);
		break;

		case CR_INFO_LIST_MATERIALS:
			_MaterialsInfo(E3d_Scene);
		break;

		case CR_INFO_LIST_ANIMATIONS:
			_AnimationsInfo(E3d_Scene);
		break;
	}
}


//========================================
// Key callback
//========================================
static void _KCB_Info(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	_MCB_Info(NULL, PClientData, NULL);
}



/*
//========================================
// Menu callback handler
//========================================
static void _MCB_Ruler(Widget PW, EPointer PClientData, EPointer PCallData)
{
// Just create a Spline with a "Ruler" ObjectComponent and its custom GLDraw function
// then interactively print the Spline's length on the StatusPanel
}
*/



//========================================================================================
// Remove NodeInfo from selected Nodes and GeoInfo from selected Geometries
//========================================================================================
static void _RemoveNodeAndObjectComponents(E3dScene* PScene)
{
	E3dNode**  LRootNodes = PScene->RootNodes;
	EIndex     LAttDelNodeN = 0, LAttDelObjectN = 0;

	EIndex  LC = 0;const EIndex  LN = PScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*  LNode = LRootNodes[LC];for(;LNode;LNode = LNode->Next)
		{
			if(LNode->Selection != E3dNodeSelNONE)
			{
				const EBool  LNodeSelected = E3dNode_IsSelected(LNode);if(LNodeSelected)
				{
					if(LNode->Components)
					{
						EIndex  LIC = 0;const EIndex  LIN = LNode->Components_Count;for(;LIC < LIN;LIC++)
						{
							E3dNodeComponentClass*  LClass = LNode->Components[LIC]->Class;
							if(LClass->Free) LClass->Free(LNode, LNode->Components[LIC]);
							EFree(LNode->Components[LIC]);
						}

						EList_Free(LNode->Components);

						LAttDelNodeN++;
					}
				}

				{
					E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
					{
						EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;E3dObject**  LObjects = LModel->Objects;for(;LObjC < LObjN;LObjC++)
						{
							E3dObject*  LObject = LObjects[LObjC];

							if(LNodeSelected || LObject->Selection != E3dObjectSelNONE)
							{
								if(LObject->Components)
								{
									EIndex  LIC = 0;EIndex  LIN = LObject->Components_Count;for(;LIC < LIN;LIC++)
									{
										E3dObjectComponentClass*  LClass = LObject->Components[LIC]->Class;
										if(LClass->Free) LClass->Free(LObject, LObject->Components[LIC]);
										EFree(LObject->Components[LIC]);
									}

									EList_Free(LObject->Components);

									LAttDelObjectN++;
								}
							}
						}
					}
				}
			}
		}
	}

	if(LAttDelObjectN > 0 || LAttDelNodeN > 0)
	{
		char  LTmpStr[1024] = "Removed attributes from";

		_AddCountString("Node", LAttDelNodeN, LTmpStr);
		_AddCountString("Object", LAttDelObjectN, LTmpStr);

		E3dp_PrintMessage(ELogMESSAGE, 5000, LTmpStr);
		E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
	}
	else E3dp_PrintMessage(ELogWARNING, 5000, "No Node or Object with attributes was selected");
}


//========================================
// Menu callback
//========================================
static void _MCB_RemoveAttributes(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_RemoveNodeAndObjectComponents(E3d_Scene);
}


//========================================
// Key callback
//========================================
static void _KCB_RemoveAttributes(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	_MCB_RemoveAttributes(NULL, PClientData, NULL);
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* LPlugin)
{
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Debug", "Selected objects", '\0', NULL, NULL, FALSE, NULL, _MCB_Info, (EPointer)CR_INFO_ON_SELECTION);
//	E_KeySetCallback("Info about selected objects", 'I', EControlMask, EKeyPRESS, _KCB_Info, (EPointer)CR_INFO_ON_SELECTION);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Debug", "Roots", '\0', "~Shift ~Ctrl <Key>R", NULL, FALSE, NULL, _MCB_Info, (EPointer)CR_INFO_LIST_ROOTS);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Debug", "Nodes", '\0', "~Shift ~Ctrl <Key>N", NULL, FALSE, NULL, _MCB_Info, (EPointer)CR_INFO_LIST_NODES);
//	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Debug", "List Scene Materials", '\0', "~Shift ~Ctrl <Key>R", NULL, FALSE, NULL, _MCB_Info, (EPointer)CR_INFO_LIST_MATERIALS);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Debug", "List Scene Animations", '\0', NULL, NULL, FALSE, NULL, _MCB_Info, (EPointer)CR_INFO_LIST_ANIMATIONS);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Delete", "Attributes", '\0', "Alt <Key>Delete", "Alt+Del", FALSE, "Remove special attributes such as construction history,\nfrom selected Nodes and Objects", _MCB_RemoveAttributes, (EPointer)NULL);
	E_KeySetCallback("Delete attributes", EKeyDELETE, EAltMask, EKeyPRESS, _KCB_RemoveAttributes, (EPointer)NULL);

// _MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Info", "Ruler", '\0', NULL, NULL, FALSE, NULL, _MCB_Ruler, (EPointer)NULL);



// _ToolBarButtons[_ToolBarButtons_Count++] = XePushButton_Add("Tool->Transform", "Anims", '\0', NULL, NULL, FALSE, NULL, _MCB_Info, (EPointer)CR_INFO_LIST_ANIMATIONS);
	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC;

		for(LC = 0;LC < _MenuButtons_Count;LC++) { XeItemFree(_MenuButtons[LC]); } _MenuButtons_Count = 0;
		for(LC = 0;LC < _ToolBarButtons_Count;LC++) { XeItemFree(_ToolBarButtons[LC]); } _ToolBarButtons_Count = 0;
	}

	E_KeyRemoveCallback('I',        EControlMask, EKeyPRESS, _KCB_Info, (EPointer)CR_INFO_ON_SELECTION);
	E_KeyRemoveCallback(EKeyDELETE, EAltMask,     EKeyPRESS, _KCB_RemoveAttributes, (EPointer)NULL);
	return(E_SUCCESS);
}
