//================================================================
// EQUINOX-3D plugin
//
// Create a box and change it interactively
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 19:33:44
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <math.h>
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Label.h>
#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/Separator.h>
#include <Xe/ToggleButton.h>


#include <E3D/UI/CreateMacros.h>
#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>

// Import our icons
// In EQUINOX-3D, icon images are compiled inlined in the main and the plugin binaries (rather than loading hundreds of .png / .jpg etc. files on startup).
// This is one of the reasons EQX starts orders of magnitude faster than other large 3D applications.
//
EpIconDef3(Box)

char*             E_PluginInfo = "Box primitive";  // This shows up in the information dialog box of the plugin manager (System->Plugins)

static EPlugin*  _Plugin = NULL;

typedef struct
{
	Widget  CenterAtBottomBW, UVsBW,
	        MinMaxBW, SizeBW,
	        EdgeLengthXScaleW, EdgeLengthYScaleW, EdgeLengthZScaleW,

	        MinXScaleW, MinYScaleW, MinZScaleW,
	        MaxXScaleW, MaxYScaleW, MaxZScaleW,

	        SubdivisionsXScaleW, SubdivisionsYScaleW, SubdivisionsZScaleW;
} E3dResourcesGUI;


static EguiItem  _MenuButton = NULL, _ToolBarButton = NULL;


enum
{
	ModeSIZE = 1,
	ModeMIN_MAX
};


typedef struct
{
	E3dObjectComponentCore;  // E3dInfo is the sub-class of the E3dObjectComponent class. This is how class inheritance works in EQX

// E3dInfo part
//
	E3dCoord3       EdgeLength;

	E3dCoord3       Min, Max;

	EInt32_3        Subdivisions;
	unsigned short  CoordinateMode;
	EBool           CenterAtBottom, Triangles, UVs;
} E3dInfo;


// Resources for the "Box" ObjectComponent class
// These are saved / restored when you save a 3D file that supports Resources (e.g. .dae)
//
#define ROffset(field)  (void*)(EPtrOffset(E3dInfo, field))
static EResource  _Resources[]=
{
	{ "EdgeLengthX",     E3dCOORDINATE_TYPE,  ROffset(EdgeLength.X) },
	{ "EdgeLengthY",     E3dCOORDINATE_TYPE,  ROffset(EdgeLength.Y) },
	{ "EdgeLengthZ",     E3dCOORDINATE_TYPE,  ROffset(EdgeLength.Z) },

	{ "Min",             E3dCOORD3_TYPE,      ROffset(Min) },
	{ "Max",             E3dCOORD3_TYPE,      ROffset(Max) },

	{ "SubdivisionsX",   EResINT32,           ROffset(Subdivisions.X) },
	{ "SubdivisionsY",   EResINT32,           ROffset(Subdivisions.Y) },
	{ "SubdivisionsZ",   EResINT32,           ROffset(Subdivisions.Z) },

	{ "CoordinateMode",  EResUINT16,          ROffset(CoordinateMode) },

	{ "CenterAtBottom",  EResBOOLEAN,         ROffset(CenterAtBottom) },
	{ "Triangles",       EResBOOLEAN,         ROffset(Triangles) },
	{ "UVs",             EResBOOLEAN,         ROffset(UVs) },
	{ }
};
#undef ROffset


static E3dObjectComponentClass*  _ComponentClass = NULL;

//========================================
// Set vertex coordinates
//========================================
static void _SetVertices(E3dVertex* PVertices, const E3dVertexSize PVertexSize,  E3dInfo* PInfo)
{
	E3dCoordinate  LMinX, LMinY, LMinZ,
	               LMaxX, LMaxY, LMaxZ,
	               LX, LY, LZ;
	const EIndex   LXSubdiv = PInfo->Subdivisions.X, LYSubdiv = PInfo->Subdivisions.Y, LZSubdiv = PInfo->Subdivisions.Z;

	switch(PInfo->CoordinateMode)
	{
		case ModeSIZE:
			LMinX = -PInfo->EdgeLength.X * 0.5;LMaxX = PInfo->EdgeLength.X * 0.5;
			LMinY = -PInfo->EdgeLength.Y * 0.5;
			LMinZ = -PInfo->EdgeLength.Z * 0.5;LMaxZ = PInfo->EdgeLength.Z * 0.5;

			if(PInfo->CenterAtBottom) { LMinY = 0.0;LMaxY = PInfo->EdgeLength.Y; }
			else LMaxY = PInfo->EdgeLength.Y * 0.5;
		break;

		case ModeMIN_MAX:
		default:
			LMinX = PInfo->Min.X;LMaxX = PInfo->Max.X;
			LMinY = PInfo->Min.Y;LMaxY = PInfo->Max.Y;
			LMinZ = PInfo->Min.Z;LMaxZ = PInfo->Max.Z;
		break;
	}

//printf("%s() %s:%d  %f %f  %f\n", __FUNCTION__, __FILE__, __LINE__, PInfo->Min.X, PInfo->Max.X, PInfo->EdgeLength.X);fflush(stdout);

	const E3dCoordinate  LXStep = PInfo->EdgeLength.X / (E3dCoordinate)LXSubdiv;
	const E3dCoordinate  LYStep = PInfo->EdgeLength.Y / (E3dCoordinate)LYSubdiv;
	const E3dCoordinate  LZStep = PInfo->EdgeLength.Z / (E3dCoordinate)LZSubdiv;


	E3dVertex*  LVertex = PVertices;

// Front side
//
	EIndex  LYCnt = 0;for(LY = LMaxY;LYCnt <= LYSubdiv;LYCnt++, LY -= LYStep)
	{
		EIndex  LXCnt = 0;for(LX = LMinX;LXCnt <= LXSubdiv;LXCnt++, LX += LXStep)
		{
			E3dVertex_Init(LVertex, PVertexSize,  LX, LY, LMaxZ);EStruct_Inc(LVertex, PVertexSize);
		}
	}


// Sides
//
	EIndex  LZCnt = 1;for(LZ = LMaxZ - LZStep;LZCnt < LZSubdiv;LZCnt++, LZ -= LZStep)
	{
		LY = LMaxY;
// Top edge
//
		EIndex  LXCnt = 0;for(LX = LMinX;LXCnt <= LXSubdiv;LXCnt++, LX += LXStep)
		{
			E3dVertex_Init(LVertex, PVertexSize,  LX, LY, LZ);EStruct_Inc(LVertex, PVertexSize);
		}

		LY = LMaxY - LYStep;
// Side points
//
		for(LYCnt = 1;LYCnt < LYSubdiv;LYCnt++, LY -= LYStep)
		{
			E3dVertex_Init(LVertex, PVertexSize,  LMinX, LY, LZ);EStruct_Inc(LVertex, PVertexSize);
			E3dVertex_Init(LVertex, PVertexSize,  LMaxX, LY, LZ);EStruct_Inc(LVertex, PVertexSize);
		}

// Bottom edge
//
		LY = LMinY;
		for(LXCnt = 0, LX = LMinX;LXCnt <= LXSubdiv;LXCnt++, LX += LXStep)
		{
		 E3dVertex_Init(LVertex, PVertexSize,  LX, LY, LZ);EStruct_Inc(LVertex, PVertexSize);
		}
	}

// Back side
//
	for(LYCnt = 0, LY = LMaxY;LYCnt <= LYSubdiv;LYCnt++, LY -= LYStep)
	{
		EIndex  LXCnt = 0;for(LX = LMinX;LXCnt <= LXSubdiv;LXCnt++, LX += LXStep)
		{
			E3dVertex_Init(LVertex, PVertexSize,  LX, LY, LMinZ);EStruct_Inc(LVertex, PVertexSize);
		}
	}
}


#define _M_AddQuad(pPolygon,  pV0, pV1, pV2, pV3) do\
{\
	E3dPolygon_InitAsQuad(pPolygon, LVertexNodeSize,  pV0, pV1, pV2, pV3);\
	ENewN(2, pPolygon->ITriangles);\
	E3dITriangle*  lITriangle = pPolygon->ITriangles;\
	pPolygon->ITriangles_Count = 2;\
	lITriangle->VertexNodes[0] = 0;lITriangle->VertexNodes[1] = 1;lITriangle->VertexNodes[2] = 2;lITriangle++;\
	lITriangle->VertexNodes[0] = 2;lITriangle->VertexNodes[1] = 3;lITriangle->VertexNodes[2] = 0;\
	pPolygon++;\
} while(0)


#define _M_AddQuadTris(pPolygon,  pV0, pV1, pV2, pV3) do\
{\
	pPolygon->ITriangles = NULL;\
	E3dPolygon_SetAsTriangle(pPolygon, LVertexNodeSize,  pV0, pV1, pV2);\
	pPolygon++;\
\
	pPolygon->ITriangles = NULL;\
	E3dPolygon_SetAsTriangle(pPolygon, LVertexNodeSize,  pV2, pV3, pV0);\
	pPolygon++;\
} while(0)


#define _M_AddQuadUV(pPolygon,  pV0, pV1, pV2, pV3) do\
{\
	E3dVertexNode*  lVertexNode;\
	E3dITriangle*   lITriangle;\
	E3dCoord2*      lUV;\
\
	E3dPolygon_SetVa(pPolygon, LVertexNodeSize, 4,  pV0, pV1, pV2, pV3);\
	ENewN(2, pPolygon->ITriangles);\
	lITriangle = pPolygon->ITriangles;\
	pPolygon->ITriangles_Count = 2;\
	lITriangle->VertexNodes[0] = 0;lITriangle->VertexNodes[1] = 1;lITriangle->VertexNodes[2] = 2;lITriangle++;\
	lITriangle->VertexNodes[0] = 2;lITriangle->VertexNodes[1] = 3;lITriangle->VertexNodes[2] = 0;\
	lVertexNode = pPolygon->VertexNodes;\
\
	lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
\
	lUV->X = LU;lUV->Y = LV;EStruct_Inc(lUV, LVertexNodeSize);\
\
	lUV->X = LU;lUV->Y = LV + LVStep;EStruct_Inc(lUV, LVertexNodeSize);\
\
	lUV->X = LU + LUStep;lUV->Y = LV + LVStep;EStruct_Inc(lUV, LVertexNodeSize);\
\
	lUV->X = LU + LUStep;lUV->Y = LV;\
	pPolygon++;\
} while(0)


#define _M_AddQuadUVTris(pPolygon,  pV0, pV1, pV2, pV3) do\
{\
	E3dVertexNode*  lVertexNode;\
	E3dCoord2*      lUV;\
\
	pPolygon->ITriangles = NULL;\
	E3dPolygon_SetAsTriangle(pPolygon, LVertexNodeSize, pV0, pV1, pV2);\
	lVertexNode = pPolygon->VertexNodes;\
	lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
\
	lUV->X = LU;lUV->Y = LV;EStruct_Inc(lUV, LVertexNodeSize);\
\
	lUV->X = LU;lUV->Y = LV + LVStep;EStruct_Inc(lUV, LVertexNodeSize);\
\
	lUV->X = LU + LUStep;lUV->Y = LV + LVStep;\
	pPolygon++;\
\
\
	pPolygon->ITriangles = NULL;\
	E3dPolygon_SetAsTriangle(pPolygon, LVertexNodeSize, pV2, pV3, pV0);\
	lVertexNode = pPolygon->VertexNodes;\
	lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
\
	lUV->X = LU + LUStep;lUV->Y = LV + LVStep;EStruct_Inc(lUV, LVertexNodeSize);\
\
	lUV->X = LU + LUStep;lUV->Y = LV;EStruct_Inc(lUV, LVertexNodeSize);\
\
	lUV->X = LU;lUV->Y = LV;\
	pPolygon++;\
} while(0)


//========================================================
// Create a grid of rectangles for a side of a box
//========================================================
static E3dPolygon* _SideUV(E3dPolygon* PPolygon, const E3dVertexNodeSize LVertexNodeSize, const int LUVOffsetInVtxNode,
 const int PUSubdiv, const E3dCoordinate PU, const E3dCoordinate LUStep,
 const int PVSubdiv, E3dCoordinate LV, const E3dCoordinate LVStep,
 int PRowA,
 const int LRowAStep, const int LLastRowAStepCorrection)
{
	E3dVertexIndex  LNextRowA = PRowA + LRowAStep;
	E3dVertexIndex  LV00, LV01, LV10, LV11;


	EIndex  LVC = 1;for(;LVC < PVSubdiv;LVC++, LV += LVStep)
	{
		E3dCoordinate  LU = PU;

		LV00 = PRowA;
		LV10 = PRowA + 1;
		LV01 = LNextRowA;
		LV11 = LNextRowA + 1;
		EIndex  LUC = 0;for(;LUC < PUSubdiv;LUC++, LU += LUStep)
		{
			_M_AddQuadUV(PPolygon,  LV00, LV01, LV11, LV10);
			LV00 = LV10;
			LV01 = LV11;
			LV10 += 1;
			LV11 += 1;
		}
		PRowA = LNextRowA;
		LNextRowA += LRowAStep;
	}

	LNextRowA += LLastRowAStepCorrection;

	{
		E3dCoordinate  LU = PU;
		LV00 = PRowA;
		LV10 = PRowA + 1;
		LV01 = LNextRowA;
		LV11 = LNextRowA + 1;
		EIndex  LUC = 0;for(;LUC < PUSubdiv;LUC++, LU += LUStep)
		{
			_M_AddQuadUV(PPolygon,  LV00, LV01, LV11, LV10);
			LV00 = LV10;
			LV01 = LV11;
			LV10 += 1;
			LV11 += 1;
		}
		PRowA = LNextRowA;
		LNextRowA += LRowAStep;
	}

	return(PPolygon);
}


//========================================================
// Create a grid of rectangles for a side of a box
//========================================================
static E3dPolygon* _SideUVr(E3dPolygon* PPolygon, const E3dVertexNodeSize LVertexNodeSize, const int LUVOffsetInVtxNode,
 const int PUSubdiv, const E3dCoordinate PU, const E3dCoordinate LUStep,
 const int PVSubdiv, E3dCoordinate LV, const E3dCoordinate LVStep,
 const int LWSubdiv,
 int LRow,
 const int LColStepMatrix[3][3][2], const int LOneVDivStep,
 const int LRowStep00,
 const int LRowStep01,
 const int LRowStep02,
const EBool PFullEnable)
{
	E3dVertexIndex  LNextRow = LRow + PVSubdiv > 1 ? LRow + LRowStep00 : LRow + LRowStep02;
	E3dVertexIndex  LV00, LV01, LV10, LV11;


	EIndex  LVC = 0;for(;LVC < PVSubdiv;LVC++, LV += LVStep)
	{
		E3dCoordinate  LU = PU;
		EIndex         LUCode, LVCode,
		               LNextRowVCode,
		               LThisRowColStep, LNextRowColStep;


		if(PVSubdiv > 1)
		{
			if(LVC == 0) LVCode = 0;
			else if(LVC < PVSubdiv) LVCode = 1;
			else LVCode = 2;

			if(LVC < (PVSubdiv - 1)) LNextRowVCode = 1;
			else LNextRowVCode = 2;
		}
		else { LVCode = 0;LNextRowVCode = 2; }


		LV00 = LRow;
		LV01 = LNextRow;

		if(PUSubdiv > 1)
		{
			LUCode = 0;
			LThisRowColStep = LColStepMatrix[LVCode][LUCode][0] + LColStepMatrix[LVCode][LUCode][1] * (LVC - 1) * (LWSubdiv - 1);
			LNextRowColStep = LColStepMatrix[LNextRowVCode][LUCode][0] + LColStepMatrix[LNextRowVCode][LUCode][1] * LVC * (LWSubdiv - 1);
		}
		else
		{
			LThisRowColStep = LNextRowColStep = LOneVDivStep * (LWSubdiv + 1) * (PVSubdiv + 1);
			LUCode = 2;
		}


		LV10 = LRow + LThisRowColStep;
		LV11 = LNextRow + LNextRowColStep;

/*
		{
			static EBool	f = TRUE;

			if(f)
			{
				printf("%s() %d  %d %d\n", __FUNCTION__, LColStepMatrix[0][0], LColStepMatrix[0][1], LColStepMatrix[0][2]);fflush(stdout);
				printf("  %d %d  LNextRowVCode %d\n", LThisRowColStep, LNextRowColStep, LNextRowVCode);fflush(stdout);
				f = FALSE;
			}
		}
*/

		EIndex  LUC = 0;for(;LUC < PUSubdiv;LUC++, LU += LUStep)
		{
			if(PFullEnable || (LVC == 0))
			{
				_M_AddQuadUV(PPolygon,  LV00, LV01, LV11, LV10);
//printf("%s() %d %d %d %d", __FUNCTION__, LRow, LRow + LThisRowColStep, LNextRow, LNextRow + LNextRowColStep);fflush(stdout);
			}
			else { _M_AddQuadUV(PPolygon,  0, 0, 0, 0); }

			LV00 = LV10;
			LV01 = LV11;

			if(LUC < (PUSubdiv - 2)) LUCode = 1;
			else LUCode = 2;

			LThisRowColStep = LColStepMatrix[LVCode][LUCode][0] + LColStepMatrix[LVCode][LUCode][1] * (LVC - 1) * (LWSubdiv - 1);
			LNextRowColStep = LColStepMatrix[LNextRowVCode][LUCode][0] + LColStepMatrix[LNextRowVCode][LUCode][1] * LVC * (LWSubdiv - 1);

			LV10 += LThisRowColStep;
			LV11 += LNextRowColStep;
		}

		LRow = LNextRow;
		if(LVC < (PVSubdiv - 2))
		{
			LNextRow += LRowStep01;
		}
		else LNextRow += LRowStep02;
	}

	return(PPolygon);
}


static void _CheckConsistency(E3dInfo* PInfo)
{
	switch(PInfo->CoordinateMode)
	{
		case ModeSIZE:
			PInfo->Min.X = -PInfo->EdgeLength.X * 0.5;
			PInfo->Max.X =  PInfo->EdgeLength.X * 0.5;

			PInfo->Min.Y = -PInfo->EdgeLength.Y * 0.5;
			PInfo->Max.Y =  PInfo->EdgeLength.Y * 0.5;

			PInfo->Min.Z = -PInfo->EdgeLength.Z * 0.5;
			PInfo->Max.Z =  PInfo->EdgeLength.Z * 0.5;
		break;

		case ModeMIN_MAX:
			E3d_Sub3_vvv(PInfo->Max, PInfo->Min, PInfo->EdgeLength);
		break;
	}
}


//========================================
// Update() (create/change) method
//========================================
static E3dObject* _Update(E3dObject* PObject, E3dObjectComponent* PInfo, const EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	E3dScene*      LScene = E3d_Scene;
	E3dInfo*       LInfo;
	E3dPolyGroup*  LPolyGroup;

	if(PCheckOnly)
	{
		if(PObject)
		{
			LInfo = (E3dInfo*)E3dObjectComponent_ByClass(PObject, _ComponentClass);

			_CheckConsistency(LInfo);
		}
		return(NULL);
	}

	E3dMesh*  LMesh;
	if(PObject == NULL)
	{
		E_New(E3dMesh, LMesh);
		PObject = (E3dObject*)LMesh;

		LInfo = (E3dInfo*)E3dObjectComponent_Add((E3dObject*)LMesh, _ComponentClass);
		if(PInfo && PInfo->Class == _ComponentClass) memcpy(LInfo, PInfo, PInfo->Class->StructSize);

		 _CheckConsistency(LInfo);

		if((LPolyGroup = E3dMesh_PolyGroupAdd(LMesh)) == NULL) { E_Delete(LMesh);return(NULL); }
	}
	else
	{
		if(!EObject_IsSubClass(PObject, E3dMesh)) return(NULL);

		LMesh = (E3dMesh*)PObject;
		LInfo = (E3dInfo*)E3dObjectComponent_ByClass(PObject, _ComponentClass);
		LPolyGroup = LMesh->PolyGroups[0];
	}

	if(LMesh)
	{
		E3dMesh_FreeCaches(LMesh);
		E3dMesh_EdgesFree(LMesh);

		LPolyGroup->VertexNormalType |= E3dNormalFROM_SMOOTHING_ANGLE;
		LPolyGroup->SmoothingAngle = 0.0;
		LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;

		const EIndex  LXSubdiv = LInfo->Subdivisions.X;
		const EIndex  LYSubdiv = LInfo->Subdivisions.Y;
		const EIndex  LZSubdiv = LInfo->Subdivisions.Z;

		const EIndex  LVNum = (LXSubdiv + 1) * (LYSubdiv + 1) * 2 + (LZSubdiv - 1) * (LXSubdiv + LYSubdiv) * 2;
		const EIndex  LPNum = LXSubdiv * LYSubdiv * 2 + LXSubdiv * LZSubdiv * 2 + LZSubdiv * LYSubdiv * 2;


// Allocate Vertices for the Mesh
//
		if(LMesh->Vertices_Count != LVNum)
		{
			E3dMesh_EdgesFree(LMesh);
			E3dMesh_VerticesFree(LMesh);
			E3dMesh_VerticesAdd(LMesh, LVNum);
		}

// Allocate Polygons for the Mesh
//
		E3dPolygon*  LPolygons;
		if(LPNum == LPolyGroup->Polygons_Count) LPolygons = LPolyGroup->Polygons;
		else
		{
			if(LPolyGroup->Polygons) E3dPolyGroup_FreePolygons(LPolyGroup);

			LPolygons = E3dPolygons_New(LPNum);if(LPolygons)
			{
				LPolyGroup->Polygons = LPolygons;LPolyGroup->Polygons_Count = LPNum;
			}
		}

		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);


// Build the Polygons
//
		E3dPolygon*  LPolygon = LPolyGroup->Polygons;

		if(LInfo->UVs) E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_UV, 0,  NULL);

		const int                LUVOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(LPolyGroup, E3dVtxAttribute_UV, 0);
		const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

// Left side (-X)
//
		if(LInfo->UVs)
		{
			const int  LColStep = -((LXSubdiv + 1) + (LYSubdiv - 1)) * 2, LColStepL = LColStep - (LXSubdiv - 1) * (LYSubdiv - 1);
			const int  LColStepMatrix[3][3][2] = 
			{
				{ { LColStep,  0 },        { LColStep, 0 },  { LColStepL, 0 }  },
				{ { LColStep, -1 },        { LColStep, 0 },  { LColStep - (LXSubdiv - 1) * (LYSubdiv - 1), 1 }  },
				{ { LColStepL, 0 },        { LColStep, 0 },  { LColStep,  0 }  }
			};

			LPolygon = _SideUVr(LPolygon, LVertexNodeSize, LUVOffsetInVtxNode,
			                    LZSubdiv, 0.0, 1.0 / (E3dCoordinate)LZSubdiv / 3.0,
			                    LYSubdiv, 0.0, 1.0 / (E3dCoordinate)LYSubdiv / 2.0,
			                    LXSubdiv,
			                    (LXSubdiv + 1) * (LYSubdiv + 1) + ((LXSubdiv + 1) + (LYSubdiv - 1)) * 2 * (LZSubdiv - 1),

			                    LColStepMatrix, -1,

			                    (LXSubdiv + 1),
			                    (LXSubdiv + 1),
			                    (LXSubdiv + 1), TRUE);
		}


// Right side (+X)
//
		if(LInfo->UVs)
		{
			const int  LColStepF = (LXSubdiv + 1) * (LYSubdiv + 1);
			const int  LColStep = ((LXSubdiv + 1) + (LYSubdiv - 1)) * 2, LColStepL = LColStep + (LXSubdiv - 1) * (LYSubdiv - 1);
			const int  LColStepMatrix[3][3][2] = 
			{
				{ { LColStepL, 0 },  { LColStep, 0 },  { LColStep, 0 }  },
				{ { LColStep + (LXSubdiv - 1) * (LYSubdiv - 2), -1 },  { LColStep, 0 },  { LColStep + LXSubdiv - 1, 1 }  },
				{ { LColStep, 0 },  { LColStep, 0 },  { LColStepF,  0 }  }
			};

			LPolygon = _SideUVr(LPolygon, LVertexNodeSize, LUVOffsetInVtxNode,
			                    LZSubdiv, 0.0, 1.0 / (E3dCoordinate)LZSubdiv / 3.0,
			                    LYSubdiv, 0.5, 1.0 / (E3dCoordinate)LYSubdiv / 2.0,
			                    LXSubdiv,
			                    LXSubdiv,

			                    LColStepMatrix, 1,

			                    (LXSubdiv + 1),
			                    (LXSubdiv + 1),
			                    (LXSubdiv + 1), TRUE);
		}



// Bottom side (-Y)
//
		if(LInfo->UVs)
		{
			LPolygon = _SideUV(LPolygon, LVertexNodeSize, LUVOffsetInVtxNode,
			                   LXSubdiv, 1.0 / 3.0, 1.0 / (E3dCoordinate)LXSubdiv / 3.0,  LZSubdiv, 0.0, 1.0 / (E3dCoordinate)LZSubdiv / 2.0,
			                   (LXSubdiv + 1) * LYSubdiv,
			                   ((LXSubdiv + 1) + (LYSubdiv - 1)) * 2,  LYSubdiv > 1 ? (LXSubdiv - 1) * (LYSubdiv - 1) : 0);
		}


// Top side (+Y)
//
		if(LInfo->UVs)
		{
			LPolygon = _SideUV(LPolygon, LVertexNodeSize, LUVOffsetInVtxNode,
			                   LXSubdiv, 1.0 / 3.0, 1.0 / (E3dCoordinate)LXSubdiv / 3.0,  LZSubdiv, 0.5, 1.0 / (E3dCoordinate)LZSubdiv / 2.0,
			                   (LXSubdiv + 1) * (LYSubdiv + 1) + (LXSubdiv + LYSubdiv) * 2 * (LZSubdiv - 1),
			                   -((LXSubdiv + 1) + (LYSubdiv - 1)) * 2,  LYSubdiv > 1 ? -(LXSubdiv - 1) * (LYSubdiv - 1) : 0);
		}


// Rear side (-Z)
//
		EIndex  LRowA = (LXSubdiv + 1) * (LYSubdiv + 1) + (LZSubdiv - 1) * (LXSubdiv + LYSubdiv) * 2;
		EIndex  LNextRowA = LRowA + LXSubdiv + 1;
		if(LInfo->UVs)
		{
			if(LInfo->Triangles)
			{
				E3dCoordinate  LV = 0.0, LVStep = 1.0 / (E3dCoordinate)LYSubdiv / 2.0;
				EIndex  LYCnt = 0;for(;LYCnt < LYSubdiv;LYCnt++, LRowA += LXSubdiv + 1, LNextRowA += LXSubdiv + 1, LV += LVStep)
				{
					const E3dCoordinate  LUStep = 1.0 / (E3dCoordinate)LXSubdiv / 3.0;E3dCoordinate	LU = 2.0 / 3.0 - LUStep;
					EIndex  LXCnt = 0;for(;LXCnt < LXSubdiv;LXCnt++, LU -= LUStep)
					{
						_M_AddQuadUVTris(LPolygon, LRowA + LXCnt + 1, LNextRowA + LXCnt + 1, LNextRowA + LXCnt, LRowA + LXCnt);
					}
				}
			}
			else
			{
				E3dCoordinate  LV = 0.0, LVStep = 1.0 / (E3dCoordinate)LYSubdiv / 2.0;
				EIndex  LYCnt = 0;for(;LYCnt < LYSubdiv;LYCnt++, LRowA += LXSubdiv + 1, LNextRowA += LXSubdiv + 1, LV += LVStep)
				{
					const E3dCoordinate  LUStep = 1.0 / (E3dCoordinate)LXSubdiv / 3.0;E3dCoordinate	LU = 2.0 / 3.0;
					EIndex  LXCnt = LXSubdiv - 1;for(;LXCnt >= 0;LXCnt--, LU += LUStep)
					{
						_M_AddQuadUV(LPolygon,  LRowA + LXCnt + 1, LNextRowA + LXCnt + 1, LNextRowA + LXCnt, LRowA + LXCnt);
					}
				}
			}
		}
		else
		{
			if(LInfo->Triangles)
			{
				EIndex  LYCnt = 0;for(;LYCnt < LYSubdiv;LYCnt++, LRowA += LXSubdiv+1, LNextRowA += LXSubdiv + 1)
				{
					EIndex  LXCnt = LXSubdiv - 1;for(;LXCnt >= 0;LXCnt--)
					{
						_M_AddQuadTris(LPolygon,  LRowA + LXCnt + 1, LNextRowA + LXCnt + 1, LNextRowA + LXCnt, LRowA + LXCnt);
					}
				}
			}
			else
			{
				EIndex  LYCnt = 0;for(;LYCnt < LYSubdiv;LYCnt++, LRowA += LXSubdiv + 1, LNextRowA += LXSubdiv + 1)
				{
					EIndex  LXCnt = LXSubdiv - 1;for(;LXCnt >= 0;LXCnt--)
					{
						_M_AddQuad(LPolygon,  LRowA + LXCnt + 1, LNextRowA + LXCnt + 1, LNextRowA + LXCnt, LRowA + LXCnt);
					}
				}
			}
		}


// Front side (+Z)
//
		LRowA = 0;LNextRowA = LXSubdiv + 1;
		if(LInfo->UVs)
		{
			if(LInfo->Triangles)
			{
				E3dCoordinate  LV = 0.5, LVStep = 1.0 / (E3dCoordinate)LYSubdiv / 2.0;
				EIndex  LYCnt = 0;for(;LYCnt < LYSubdiv;LYCnt++, LRowA += LXSubdiv + 1, LNextRowA += LXSubdiv + 1, LV += LVStep)
				{
					E3dCoordinate  LU = 2.0 / 3.0, LUStep = 1.0 / (E3dCoordinate)LXSubdiv / 3.0;
					EIndex  LXCnt = 0;for(;LXCnt < LXSubdiv;LXCnt++, LU += LUStep)
					{
						_M_AddQuadUVTris(LPolygon,  LRowA + LXCnt, LNextRowA + LXCnt, LNextRowA + LXCnt + 1, LRowA + LXCnt + 1);
					}
				}
			}
			else
			{
				E3dCoordinate  LV = 0.5, LVStep = 1.0 / (E3dCoordinate)LYSubdiv / 2.0;
				EIndex  LYCnt = 0;for(;LYCnt < LYSubdiv;LYCnt++, LRowA += LXSubdiv + 1, LNextRowA += LXSubdiv + 1, LV += LVStep)
				{
					E3dCoordinate  LU = 2.0 / 3.0, LUStep = 1.0 / (E3dCoordinate)LXSubdiv / 3.0;
					EIndex  LXCnt = 0;for(;LXCnt < LXSubdiv;LXCnt++, LU += LUStep)
					{
						_M_AddQuadUV(LPolygon,  LRowA + LXCnt, LNextRowA + LXCnt, LNextRowA + LXCnt + 1, LRowA + LXCnt + 1);
					}
				}
			}
		}
		else
		{
			if(LInfo->Triangles)
			{
				EIndex  LYCnt = 0;for(;LYCnt < LYSubdiv;LYCnt++, LRowA += LXSubdiv + 1, LNextRowA += LXSubdiv + 1)
				{
					EIndex  LXCnt = 0;for(;LXCnt < LXSubdiv;LXCnt++)
					{
						_M_AddQuadTris(LPolygon,  LRowA + LXCnt, LNextRowA + LXCnt, LNextRowA + LXCnt + 1, LRowA + LXCnt + 1);
					}
				}
			}
			else
			{
				EIndex  LYCnt = 0;for(;LYCnt < LYSubdiv;LYCnt++, LRowA += LXSubdiv + 1, LNextRowA += LXSubdiv + 1)
				{
					EIndex  LXCnt = 0;for(;LXCnt < LXSubdiv;LXCnt++)
					{
						_M_AddQuad(LPolygon,  LRowA + LXCnt, LNextRowA + LXCnt, LNextRowA + LXCnt + 1, LRowA + LXCnt + 1);
					}
				}
			}
		}

		EChangeFlags  LUpdateFlags = PUpdateFlags;
		if(LInfo->UVs) LUpdateFlags |= E3dUF_VERTEX_ST;
		const EChangeFlags	LChanged = E3dObject_UpdateForDisplay(PObject, LUpdateFlags, LScene);
		if(PTriggeredChangesP) *PTriggeredChangesP |= LChanged;
	}

	return(PObject);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_CenterAtBottom(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->CenterAtBottom = (EBool)(((XeToggleButtonCallbackStruct*)PCallData)->Set);

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


/*
//========================================
// Dialog box callback handler
//========================================
static void _DCB_UVs(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);

	LInfo->UVs = (EBool)(((XeToggleButtonCallbackStruct*)PCallData)->Set);

	EChangeFlags  LChanged = 0;

	E3dObject_CustomUpdate(LObject, LInfo, E3duf_SHAPE, &LChanged);

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}
*/


//================================================================
// Enable / disable GUI widgets, based on the coordinate mode
//================================================================
static void _SetGUIByMode(E3dObjectPanel* PPanel)
{
	E3dObject*               LObject = PPanel->Object;
	XeChangeManage           LWidgetsToManageUnmanage[10];
	XeChangeManage*          LChild = LWidgetsToManageUnmanage;
	EBool                    LSizeMode, LMinMaxMode;
	E3dInfo*                 LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)PPanel, _ComponentClass);
	E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	switch(LInfo->CoordinateMode)
	{
		case ModeSIZE:  LSizeMode = TRUE;LMinMaxMode = FALSE;break;
		default:        LSizeMode = FALSE;LMinMaxMode = TRUE;break;
	}


	LChild = LWidgetsToManageUnmanage;
	LChild->W = LResourcesGUI->CenterAtBottomBW;LChild->Manage = LSizeMode;LChild++;
	XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);


	LChild = LWidgetsToManageUnmanage;
	LChild->W = LResourcesGUI->EdgeLengthXScaleW;LChild->Manage = LSizeMode;LChild++;
	LChild->W = LResourcesGUI->EdgeLengthYScaleW;LChild->Manage = LSizeMode;LChild++;
	LChild->W = LResourcesGUI->EdgeLengthZScaleW;LChild->Manage = LSizeMode;LChild++;

	LChild->W = LResourcesGUI->MinXScaleW;LChild->Manage = LMinMaxMode;LChild++;
	LChild->W = LResourcesGUI->MinYScaleW;LChild->Manage = LMinMaxMode;LChild++;
	LChild->W = LResourcesGUI->MinZScaleW;LChild->Manage = LMinMaxMode;LChild++;

	LChild->W = LResourcesGUI->MaxXScaleW;LChild->Manage = LMinMaxMode;LChild++;
	LChild->W = LResourcesGUI->MaxYScaleW;LChild->Manage = LMinMaxMode;LChild++;
	LChild->W = LResourcesGUI->MaxZScaleW;LChild->Manage = LMinMaxMode;LChild++;

	XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);


	XeScale_SetValue_Coord(LResourcesGUI->EdgeLengthXScaleW, LInfo->EdgeLength.X, FALSE);
	XeScale_SetValue_Coord(LResourcesGUI->EdgeLengthYScaleW, LInfo->EdgeLength.Y, FALSE);
	XeScale_SetValue_Coord(LResourcesGUI->EdgeLengthZScaleW, LInfo->EdgeLength.Z, FALSE);


	XeScale_SetValue_Coord(LResourcesGUI->MinXScaleW, LInfo->Min.X, FALSE);
	XeScale_SetValue_Coord(LResourcesGUI->MinYScaleW, LInfo->Min.Y, FALSE);
	XeScale_SetValue_Coord(LResourcesGUI->MinZScaleW, LInfo->Min.Z, FALSE);

	XeScale_SetValue_Coord(LResourcesGUI->MaxXScaleW, LInfo->Max.X, FALSE);
	XeScale_SetValue_Coord(LResourcesGUI->MaxYScaleW, LInfo->Max.Y, FALSE);
	XeScale_SetValue_Coord(LResourcesGUI->MaxZScaleW, LInfo->Max.Z, FALSE);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_Size(Widget PW, EPointer PClientData, EPointer PCallData)
{
	if(((XeToggleButtonCallbackStruct*)PCallData)->Set)
	{
		E3dScene*        LScene = E3d_Scene;
		E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
		E3dObject*       LObject = LPanel->Object;
		E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
		EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->CoordinateMode = ModeSIZE;

		{
			E3dMesh*  LMesh = (E3dMesh*)LObject;
			_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
		}

		LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
		E3dWindows_RedrawByChangeFlags(LChanged);

		_SetGUIByMode(LPanel);
	}
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_MinMax(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);


	if(((XeToggleButtonCallbackStruct*)PCallData)->Set)
	{
		LInfo->CoordinateMode = ModeMIN_MAX;

		{
			E3dMesh*  LMesh = (E3dMesh*)LObject;
			_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
		}

		LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
		E3dWindows_RedrawByChangeFlags(LChanged);

		_SetGUIByMode(LPanel);
	}
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_EdgeLengthX(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);


	LInfo->EdgeLength.X = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	LInfo->Min.X = -LInfo->EdgeLength.X * 0.5;
	LInfo->Max.X =  LInfo->EdgeLength.X * 0.5;

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_EdgeLengthY(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->EdgeLength.Y = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	LInfo->Min.Y = -LInfo->EdgeLength.Y * 0.5;
	LInfo->Max.Y =  LInfo->EdgeLength.Y * 0.5;

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_EdgeLengthZ(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->EdgeLength.Z = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	LInfo->Min.Z = -LInfo->EdgeLength.Z * 0.5;
	LInfo->Max.Z =  LInfo->EdgeLength.Z * 0.5;

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_MinX(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);


	LInfo->Min.X = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	if(LInfo->Max.X < LInfo->Min.X)
	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
		E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

		LInfo->Max.X = LInfo->Min.X;

		XeScale_SetValue_Coord(LResourcesGUI->MaxXScaleW, LInfo->Max.X, FALSE);
	}

	LInfo->EdgeLength.X = LInfo->Max.X - LInfo->Min.X;

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_MinY(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);


	LInfo->Min.Y = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	if(LInfo->Max.Y < LInfo->Min.Y)
	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
		E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

		LInfo->Max.Y = LInfo->Min.Y;

		XeScale_SetValue_Coord(LResourcesGUI->MaxYScaleW, LInfo->Max.Y, FALSE);
	}

	LInfo->EdgeLength.Y = LInfo->Max.Y - LInfo->Min.Y;

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_MinZ(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);


	LInfo->Min.Z = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	if(LInfo->Max.Z < LInfo->Min.Z)
	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
		E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

		LInfo->Max.Z = LInfo->Min.Z;

		XeScale_SetValue_Coord(LResourcesGUI->MaxZScaleW, LInfo->Max.Z, FALSE);
	}

	LInfo->EdgeLength.Z = LInfo->Max.Z - LInfo->Min.Z;

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_MaxX(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->Max.X = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	if(LInfo->Min.X > LInfo->Max.X)
	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
		E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

		LInfo->Min.X = LInfo->Max.X;

		XeScale_SetValue_Coord(LResourcesGUI->MinXScaleW, LInfo->Min.X, FALSE);
	}

	LInfo->EdgeLength.X = LInfo->Max.X - LInfo->Min.X;

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_MaxY(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->Max.Y = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	if(LInfo->Min.Y > LInfo->Max.Y)
	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
		E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

		LInfo->Min.Y = LInfo->Max.Y;

		XeScale_SetValue_Coord(LResourcesGUI->MinYScaleW, LInfo->Min.Y, FALSE);
	}

	LInfo->EdgeLength.Y = LInfo->Max.Y - LInfo->Min.Y;

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_MaxZ(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->Max.Z = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	if(LInfo->Min.Z > LInfo->Max.Z)
	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
		E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

		LInfo->Min.Z = LInfo->Max.Z;

		XeScale_SetValue_Coord(LResourcesGUI->MinZScaleW, LInfo->Min.Z, FALSE);
	}

	LInfo->EdgeLength.Z = LInfo->Max.Z - LInfo->Min.Z;

	{
		E3dMesh*  LMesh = (E3dMesh*)LObject;
		_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	}

	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_XSubdivisions(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);

	LInfo->Subdivisions.X = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	EChangeFlags  LChanged = 0;E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_YSubdivisions(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);

	LInfo->Subdivisions.Y = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	EChangeFlags  LChanged = 0;E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handler
//========================================
static void _DCB_ZSubdivisions(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);

	LInfo->Subdivisions.Z = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	EChangeFlags  LChanged = 0;E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//================================================
// GUI method for the "Box" ObjectComponent class
//================================================
static void _GUI(E3dObject* PObject, E3dObjectComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dInfo*                 LInfo = (E3dInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				E3dResourcesGUI*  LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

				_SetGUIByMode(LPanel);

				XeToggleButton_SetState(LResourcesGUI->CenterAtBottomBW, LInfo->CenterAtBottom, FALSE);
//				XeToggleButton_SetState(LResourcesGUI->TrianglesBW, LInfo->Triangles, FALSE);
//				XeToggleButton_SetState(LResourcesGUI->UVsBW, LInfo->UVs, FALSE);


				switch(LInfo->CoordinateMode)
				{
					case ModeSIZE:
						XeToggleButton_SetState(LResourcesGUI->SizeBW, TRUE, FALSE);
						XeToggleButton_SetState(LResourcesGUI->MinMaxBW, FALSE, FALSE);
					break;

					case ModeMIN_MAX:
						XeToggleButton_SetState(LResourcesGUI->SizeBW, FALSE, FALSE);
						XeToggleButton_SetState(LResourcesGUI->MinMaxBW, TRUE, FALSE);
					break;
				}


				XeScale_SetValue_Coord(LResourcesGUI->EdgeLengthXScaleW, LInfo->EdgeLength.X, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->EdgeLengthYScaleW, LInfo->EdgeLength.Y, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->EdgeLengthZScaleW, LInfo->EdgeLength.Z, FALSE);


				XeScale_SetValue_Coord(LResourcesGUI->MinXScaleW, LInfo->Min.X, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->MaxXScaleW, LInfo->Max.X, FALSE);

				XeScale_SetValue_Coord(LResourcesGUI->MinYScaleW, LInfo->Min.Y, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->MaxYScaleW, LInfo->Max.Y, FALSE);

				XeScale_SetValue_Coord(LResourcesGUI->MinZScaleW, LInfo->Min.Z, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->MaxZScaleW, LInfo->Max.Z, FALSE);

				XeScale_SetValue(LResourcesGUI->SubdivisionsXScaleW, LInfo->Subdivisions.X, FALSE);
				XeScale_SetValue(LResourcesGUI->SubdivisionsYScaleW, LInfo->Subdivisions.Y, FALSE);
				XeScale_SetValue(LResourcesGUI->SubdivisionsZScaleW, LInfo->Subdivisions.Z, FALSE);
			}
			else
			{
				Widget            LTopMatrixW = LDrawer->Drawer.DrawerW, LTypeMatrixW, LMatrixW, LW;
				Widget            LIndentWs[7];int  LIndentN;       // For indenting the Scale Widgets
				Widget            LTraverseWs[16];int  LTraverseN;  // For TextField traversal
				EArg              Args[256];int  ArgCnt;
				E3dResourcesGUI*  LResourcesGUI;ENew(LResourcesGUI);


				LDrawer->ResourcesGUI = LResourcesGUI;

				EXeStart;
				EXeSetArg(XeNspacingLeft, Xe_DrawerIndent);
				Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

				{
					EXeStart;
					EXeSetArg(XeNradioBehavior, TRUE);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNspacingX, 6);
					EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LTypeMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					XeWidget_ManagedNew("Shape by:", XeLabelWidget_class, LTypeMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
					EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
					EXeSetArg(XeNset, (EBool)(LInfo->CoordinateMode == ModeSIZE));
					LResourcesGUI->SizeBW = LW = XeWidget_ManagedNew("Size", XeToggleButton_class, LTypeMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Size, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
					EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
					EXeSetArg(XeNset, (EBool)(LInfo->CoordinateMode == ModeMIN_MAX));
					LResourcesGUI->MinMaxBW = LW = XeWidget_ManagedNew("Min-Max", XeToggleButton_class, LTypeMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MinMax, (EPointer)LPanel);
				}




				EXeStart;
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNspacingX, 16);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

				EXeStart;
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
				EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNmarginTop, 0);EXeSetArg(XeNmarginBottom, 4);
				EXeSetArg(XeNset, LInfo->CenterAtBottom);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LResourcesGUI->CenterAtBottomBW = LW = XeWidget_ManagedNew("Center at bottom", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CenterAtBottom, (EPointer)LPanel);





/*
				EXeStart;
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNselectColor, &Xe_HighlightRGBAiColor);
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNmarginTop, 0);EXeSetArg(XeNmarginBottom, 4);
				EXeSetArg(XeNset, LInfo->UVs);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LResourcesGUI->UVsBW = LW = XeWidget_ManagedNew("UVs", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_UVs, (EPointer)LPanel);
*/

				LTraverseN = 0;
				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNscaleLengthInPixels, 120);
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->EdgeLengthXScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Width (X)",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->EdgeLength.X,  0.0001, 200000.0,  1.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_EdgeLengthX, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNscaleLengthInPixels, 120);
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->EdgeLengthYScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Height (Y)",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->EdgeLength.Y,  0.0001, 200000.0,  1.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_EdgeLengthY, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNscaleLengthInPixels, 120);
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->EdgeLengthZScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Depth (Z)",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->EdgeLength.Z,  0.0001, 200000.0,  1.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_EdgeLengthZ, (EPointer)LPanel);

				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);




				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNscaleLengthInPixels, 120);
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->MinXScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Min X",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Min.X,  -100000.0, 100000.0,  -100.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MinX, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNscaleLengthInPixels, 120);
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->MaxXScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Max X",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Max.X,  -100000.0, 100000.0,  -100.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MaxX, (EPointer)LPanel);


				EXeStart;
				EXeSetArg(XeNscaleLengthInPixels, 120);
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->MinYScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Min Y",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Min.Y,  -100000.0, 100000.0,  -100.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MinY, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNscaleLengthInPixels, 120);
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->MaxYScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Max Y",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Max.Y,  -100000.0, 100000.0,  -100.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MaxY, (EPointer)LPanel);


				EXeStart;
				EXeSetArg(XeNscaleLengthInPixels, 120);
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->MinZScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Min Z",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Min.Z,  -100000.0, 100000.0,  -100.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MinZ, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNscaleLengthInPixels, 120);
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->MaxZScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Max Z",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Max.X,  -100000.0, 100000.0,  -100.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MaxZ, (EPointer)LPanel);



				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


				_SetGUIByMode(LPanel);


				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 3);
				EXeSetArg(XeNminimum, 1);EXeSetArg(XeNmaximum, 100);
				EXeSetArg(XeNscaleMaximum, 20);
				EXeSetArg(XeNvalue, LInfo->Subdivisions.X);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SubdivisionsXScaleW = LW = XeWidget_ManagedNew("X subdivisions", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_XSubdivisions, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 3);
				EXeSetArg(XeNminimum, 1);EXeSetArg(XeNmaximum, 100);
				EXeSetArg(XeNscaleMaximum, 20);
				EXeSetArg(XeNvalue, LInfo->Subdivisions.Y);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SubdivisionsYScaleW = LW = XeWidget_ManagedNew("Y subdivisions", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_YSubdivisions, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 3);
				EXeSetArg(XeNminimum, 1);EXeSetArg(XeNmaximum, 100);
				EXeSetArg(XeNscaleMaximum, 20);
				EXeSetArg(XeNvalue, LInfo->Subdivisions.Z);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SubdivisionsZScaleW = LW = XeWidget_ManagedNew("Z subdivisions", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_ZSubdivisions, (EPointer)LPanel);

// Indent the Scale Widgets, so they line up nicely
//
				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 0);

// Set up TextField traversal between the TextFields of the Scale Widgets
//
				Xe_TraversalRing(LTraverseWs, LTraverseN);
			}
		break;
	}
}


//========================================
// Main callback (create new box)
//========================================
static void _MCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dTpl_ObjectComponentActivateCB_Geometry(E3dBox, _, _Plugin);
}


//========================================================
// Constructor of the "Box" ObjectComponent class
//========================================================
static int _Init(E3dObject* LObject, E3dObjectComponent* PInfo)
{
	E3dInfo*  LThis = (E3dInfo*)PInfo;

	E_SetXYZ_s(LThis->Min,  -0.5, -0.5, -0.5);
	E_SetXYZ_s(LThis->Max,   0.5,  0.5,  0.5);

	E_SetXYZ_s(LThis->EdgeLength,   1.0, 1.0, 1.0);

	E_SetXYZ_s(LThis->Subdivisions,   1, 1, 1);

	LThis->CenterAtBottom = FALSE;
	LThis->Triangles = FALSE;
	LThis->UVs = TRUE;

	LThis->CoordinateMode = ModeSIZE;

	return(E_SUCCESS);
}


#define _M_Copy(pField) LDInfo->pField = LSInfo->pField

//========================================================
// Copy-constructor of the "Box" ObjectComponent class
//========================================================
static void _Copy(const E3dObject* PSObject, const E3dObjectComponent* PSInfo, E3dObject* LDObject, E3dObjectComponent* PDInfo)
{
	const E3dInfo*  LSInfo = (E3dInfo*)PSInfo;
	E3dInfo*        LDInfo = (E3dInfo*)PDInfo;

	_M_Copy(Min);
	_M_Copy(Max);
	_M_Copy(EdgeLength);
	_M_Copy(Subdivisions);
	_M_Copy(CenterAtBottom);
	_M_Copy(Triangles);
	_M_Copy(UVs);
	_M_Copy(CoordinateMode);
}


//========================================================================
// No destructor for the "Box" ObjectComponent class, as we don't store
// any dynamically allocatied data in the ObjectComponent instances.
//========================================================================


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	E3dObjectComponentClass  LComponentClass=
	{
		.Name=        "Box",
		.StructSize=  sizeof(E3dInfo),
		.Resources=   _Resources,

		.Init=        _Init,
		.Copy=        _Copy,
		.Update=      _Update,
		.GUI=         _GUI
	};

	EpM_CreateIconStates3_sp(Box);

// This is another way of adding information to the plugin
// You can also just define the symbol:
// char*	E_PluginInfo ...
//
	EFreeChkClr(PPlugin->Info);                // First, you have to free any existing text
	PPlugin->Info = EStrDup("Box primitive");  // This will be freed when EQUINOX-3D unloads the plugin (at exit, or when the user unloads it in the plugin manager)

	_ComponentClass = E3dObjectComponent_ClassRegister(&LComponentClass);if(_ComponentClass)
	{
		E_Assign(_Image_Box, _ComponentClass->Icon);
		Xe_PushButtonAddImg3(_ToolBarButton, Box, "Tool->Create", "Box", '\0', NULL, NULL, TRUE, "Left button: Create cube or box with optional subdivisions\nRight button: dialog box for this tool", _MCB, (EPointer)0);
		_MenuButton = XePushButton_Add("Menu->Create", "Cube / Box", '\0', NULL, NULL, TRUE, NULL, _MCB, (EPointer)0);
	}

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
// If the plugin is being unloaded while EQX is still running, destroy the buttons we created
// (when EQX is exiting, everything will be freed automatically)
//
		XeItemFree(_MenuButton);
		XeItemFree(_ToolBarButton);
	}

// Deactivate the ObjectComponent class we registered
// We deactivate classes, rather than destroying them, because this way the user can unload / recompile and reload plugins,
// while retaining Objects in the Scene that use classes defined by that plugin!
// When the plugin is reloaded, E3dObjectComponent_ClassRegister will find and reactivate the same class.
//
	if(_ComponentClass) { E3dObjectComponent_ClassDeactivate(_ComponentClass);_ComponentClass = NULL; }

	EpIconFree3(Box);

	return(E_SUCCESS);
}
