//================================================================
// EQUINOX-3D(TM) plugin
//
// Create a cylinder or a capsule and change it interactively
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 17:25:03
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <float.h>
#include <math.h>
#include <stdio.h>
#include <assert.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/Separator.h>
#include <Xe/ToggleButton.h>

#include <E3D/PolygonEditing.h>

#include <E3D/UI/CreateMacros.h>
#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>



static EPlugin*  _Plugin = NULL;

static EBool     _AddPieSides = FALSE;

typedef struct
{
	Widget  SmoothW,
          SideW, TopW, BottomW, OriginAtBottomW, PieW, RadiusScaleW, HeightScaleW,
          StartAngleScaleW, EndAngleScaleW,
          SidesScaleW, RadialSectionsScaleW, HeightSectionsScaleW;
} E3dResourcesGUI;


char*            E_PluginInfo = "Cylinder and capsule primitives";

static EguiItem  _MenuButtons[16];static int  _MenuButtons_Count = 0;
static EguiItem  _ToolBarButtons[2];static int  _ToolBarButtons_Count = 0;

enum
{
	E3dCylinderPOLYGONS = 0,
	E3dCylinderNURBS,
	E3dCylinderBEZIER,
};

#define E3dRPI  (1.0 / 180.0 * E3dPI)


EpIconDef3(Cylinder)
EpIconDef3(Capsule)


static const EBool  _UseRevolve = TRUE;


typedef struct
{
	E3dObjectComponentCore;

	E3dCoordinate  Radius, Height, TopDomeHeight, BottomDomeHeight;
	E3dCoordinate  StartAngle, EndAngle;

	unsigned int   Sides, RadialSections, HeightSections;
	unsigned int   Type;

	EBool          Full, Smooth, Side, TopCap, BottomCap, Pie, OriginAtBottom;
} E3dCylinderInfo;


#define ROffset(field) (void*)(EPtrOffset(E3dCylinderInfo, field))
static EResource _Resources[]=
{
	{ "Radius",          E3dCOORDINATE_TYPE,  ROffset(Radius) },
	{ "Height",          E3dCOORDINATE_TYPE,  ROffset(Height) },
	{ "StartAngle",      E3dCOORDINATE_TYPE,  ROffset(StartAngle) },
	{ "EndAngle",        E3dCOORDINATE_TYPE,  ROffset(EndAngle) },
	{ "Sides",           EResINT32,           ROffset(Sides) },
	{ "RadialSections",  EResINT32,           ROffset(RadialSections) },
	{ "HeightSections",  EResINT32,           ROffset(HeightSections) },
	{ "Type",            EResINT32,           ROffset(Type) },
	{ "Smooth",          EResBOOLEAN,         ROffset(Smooth) },
	{ "OrigAtBottom",    EResBOOLEAN,         ROffset(OriginAtBottom) },
	{ "Side",            EResBOOLEAN,         ROffset(Side) },
	{ "TopCap",          EResBOOLEAN,         ROffset(TopCap) },
	{ "BottomCap",       EResBOOLEAN,         ROffset(BottomCap) },
	{ }
};
#undef ROffset


static E3dObjectComponentClass*  _CylinderComponentClass = NULL;
static E3dObjectComponentClass*  _CapsuleComponentClass = NULL;

static EIndex  _VNum;

//================================================
// Compute vertex coordinates of the cylinder
//================================================
static void _SetVertices(E3dMesh* PMesh, E3dCylinderInfo* PInfo, E3dCoord2* PUnitArcPoints)
{
	E3dCoord2*  LUnitArcPoints;

	if(PUnitArcPoints) LUnitArcPoints = PUnitArcPoints;
	else LUnitArcPoints = E3d_CreateUnitArc(PInfo->Sides, PInfo->StartAngle, PInfo->EndAngle,  NULL);


	const E3dCoordinate  LRadius = PInfo->Radius;
	const E3dCoordinate  LHeight = PInfo->Height;
	const E3dCoordinate  LTopDomeHeight = PInfo->TopDomeHeight;
	const E3dCoordinate  LBottomDomeHeight = PInfo->BottomDomeHeight;


	const EBool    LFull = PInfo->Full;
	EIndex         LSides = PInfo->Sides;
	EIndex         LRadialSections = PInfo->RadialSections;
	const EIndex   LHeightSections = PInfo->HeightSections;

	const E3dCoordinate  LUStep = 1.0 / (E3dCoordinate)LSides;
	const E3dCoordinate  LVStep = 1.0 / (E3dCoordinate)LHeightSections;

	const E3dCoordinate  LRadStep = LRadius / (E3dCoordinate)LRadialSections;

	EBool  LPie;
	if((!LFull) && PInfo->Pie) LPie = TRUE;
	else LPie = FALSE;

	E3dCoordinate  LTopY, LBottomY;
	E3dCoordinate  LY;

	if(PInfo->OriginAtBottom)
	{
		LY = LHeight;LTopY = LHeight;LBottomY = 0.0;
	}
	else
	{
		LY = LHeight * 0.5;LTopY = LHeight * 0.5;LBottomY = -LHeight * 0.5;
	}


	E3dVertex*  LVertices = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;
	E3dVertex*  LVertex = LVertices;

	switch(PInfo->Type)
	{
		case E3dCylinderPOLYGONS:
			{
				const E3dCoordinate  LHStep = LHeight / (double)LHeightSections;
// Sides
//
				if(LPie)
				{
					if(LRadialSections < 1) LRadialSections = 1;
					LSides++;
				}
				else
				{
					if(!LFull) LSides++;
				}

				{
					EIndex  LC = 0;E3dCoordinate  LVi = 0.0;for(;LC <= LHeightSections;LC++, LVi += LVStep)
					{
						int  LC1 = 0;E3dCoordinate  LU = 0.0;const E3dCoord2*  LPoint = LUnitArcPoints;for(;LC1 < LSides;LC1++, LU += LUStep)
						{
							LVertex->Position.X = LRadius * LPoint->X;LVertex->Position.Y = LY;LVertex->Position.Z = LRadius * LPoint->Y;

							LPoint++;
							EStruct_Inc(LVertex, LVertexSize);
						}
						LY -= LHStep;
					}
				}

				LY = LTopY;
				const EIndex  LRN = LRadialSections - 1;

				if(LRadialSections > 0)
				{
					if(PInfo->TopCap)
					{
// Middle Vertex
//
						LVertex->Position.X = 0.0;LVertex->Position.Y = LY + LTopDomeHeight * LRadius;LVertex->Position.Z = 0.0;EStruct_Inc(LVertex, LVertexSize);

						if(LRadialSections > 1 && E3d_Abs(LTopDomeHeight) > 0.0)
						{
							E3dCoordinate  LAngle, LAngleStep, LHt;

							LAngleStep = E3dPI * 0.5 / ((E3dCoordinate)LRadialSections);
							LAngle = E3dPI * 0.5 - LAngleStep;

							EIndex  LC = 0;for(;LC < LRN;LC++)
							{
								E3dCoordinate  LRad = E3d_Cos(LAngle) * LRadius;LHt = LRadius * E3d_Sin(LAngle);
								EIndex  LC1 = 0;const E3dCoord2*  LPoint = LUnitArcPoints;for(;LC1 < LSides;LC1++)
								{
									LVertex->Position.X = LRad * LPoint->X;LVertex->Position.Y = LY + LTopDomeHeight * LHt;LVertex->Position.Z = LRad * LPoint->Y;
									LPoint++;
									EStruct_Inc(LVertex, LVertexSize);
								}
								LAngle -= LAngleStep;
							}
						}
						else
						{
							EIndex  LC = 0;E3dCoordinate  LRad = LRadStep;for(;LC < LRN;LC++)
							{
								EIndex  LC1 = 0;const E3dCoord2*  LPoint = LUnitArcPoints;for(;LC1 < LSides;LC1++)
								{
									LVertex->Position.X = LRad * LPoint->X;LVertex->Position.Y = LY;LVertex->Position.Z = LRad * LPoint->Y;
									LPoint++;
									EStruct_Inc(LVertex, LVertexSize);
								}
								LRad += LRadStep;
							}
						}
					}


					LY = LBottomY;

// Radial segments in bottom cap
//
					if(PInfo->BottomCap)
					{
// Middle Vertex
//
						LVertex->Position.X = 0.0;LVertex->Position.Y = LY - LBottomDomeHeight*LRadius;LVertex->Position.Z = 0.0;EStruct_Inc(LVertex, LVertexSize);

						if((LRadialSections > 1) && (E3d_Abs(LTopDomeHeight) > 0.0))
						{
							E3dCoordinate  LAngle, LAngleStep, LHt;

							LAngleStep = E3dPI * 0.5 / ((E3dCoordinate)LRadialSections);
							LAngle = E3dPI * 0.5 - LAngleStep;

							EIndex  LC = 0;for(;LC < LRN;LC++)
							{
								E3dCoordinate  LRad = E3d_Cos(LAngle) * LRadius;LHt = LRadius * E3d_Sin(LAngle);
								EIndex  LC1 = 0;const E3dCoord2*  LPoint = LUnitArcPoints;for(;LC1 < LSides;LC1++)
								{
									LVertex->Position.X = LRad * LPoint->X;LVertex->Position.Y = LY - LTopDomeHeight * LHt;LVertex->Position.Z = LRad * LPoint->Y;
									LPoint++;
									EStruct_Inc(LVertex, LVertexSize);
								}
								LAngle -= LAngleStep;
							}
						}
						else
						{
							EIndex  LC = 0;E3dCoordinate  LRad = LRadStep;for(;LC < LRN;LC++)
							{
								EIndex  LC1 = 0;const E3dCoord2*  LPoint = LUnitArcPoints;for(;LC1 < LSides;LC1++)
								{
									LVertex->Position.X = LRad * LPoint->X;LVertex->Position.Y = LY;LVertex->Position.Z = LRad * LPoint->Y;
									LPoint++;
									EStruct_Inc(LVertex, LVertexSize);
								}
								LRad += LRadStep;
							}
						}
					}
				}

// Inner Vertices on the two "pie cut" sides
//
				if(LPie && LHeightSections > 1 && _AddPieSides)
				{
					{
						LY = LTopY - LHStep;
						EIndex  LC = 1;const E3dCoord2*  LPoint = LUnitArcPoints;E3dCoordinate  LX = LPoint->X, LZ = LPoint->Y;for(;LC < LHeightSections;LC++)
						{
							EIndex  LC1 = 0;E3dCoordinate  LRad = 0.0;for(;LC1 < LRadialSections;LC1++)
							{
								LVertex->Position.X = LRad * LX;LVertex->Position.Y = LY;LVertex->Position.Z = LRad * LZ;
								LRad += LRadStep;
								EStruct_Inc(LVertex, LVertexSize);
							}
							LY -= LHStep;
						}
					}

					if(LRadialSections > 1)
					{
						const EIndex  LN1 = LRadialSections - 1;

						LY = LTopY - LHStep;

						EIndex  LC = 1;const E3dCoord2*  LPoint = LUnitArcPoints + LSides - 1;E3dCoordinate  LX = LPoint->X, LZ = LPoint->Y;for(;LC < LHeightSections;LC++)
						{
							int  LC1 = 0;E3dCoordinate  LRad = LRadStep;for(;LC1 < LN1;LC1++)
							{
								LVertex->Position.X = LRad * LX;LVertex->Position.Y = LY;LVertex->Position.Z = LRad * LZ;
								LRad += LRadStep;
								EStruct_Inc(LVertex, LVertexSize);
							}
							LY -= LHStep;
						}
					}
				}

// This is for debugging only
// If an existing cylinder is edited, _VNum mignt not be initialized, this is normal!!!
//
/*
				const EIndex  LVIndex = EStruct_GetIndex(PMesh->Vertices, LVertexSize, LVertex);if(LVIndex > _VNum)
				{
					printf("%s() %d vs %d\n", __FUNCTION__, LVIndex, _VNum);fflush(stdout);
					assert(0);
				}
*/
			}
		break;
	}

	if(LUnitArcPoints && LUnitArcPoints == NULL) EFree(LUnitArcPoints);
}



void _CreatePolygonsForCylinder(E3dCylinderInfo* PInfo, E3dMesh* PMesh, E3dPolyGroup* PPolyGroup, const EIndex PNSideVertices, E3dCoord2* PUnitArcPoints)
{
	E3dPolyGroup_VertexAttributeAdd(PPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);

// Make sure the VertexNodes have UVs
//
	E3dVertexNodeSize  LUVOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(PPolyGroup,  E3dVtxAttribute_UV, 0);
	if(LUVOffsetInVtxNode < 0)
	{
		E3dPolyGroup_VertexAttributeAdd(PPolyGroup,  NULL, E3dVtxAttribute_UV, 0,  NULL);
		LUVOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(PPolyGroup,  E3dVtxAttribute_UV, 0);
	}

	const EIndex    LHeightSections = PInfo->HeightSections;
	const EIndex    LSides = PNSideVertices;

	EIndex          LNumOfVertexNodes = 0;
	if(PInfo->Side) LNumOfVertexNodes += LHeightSections * LSides * 4;

	const EIndex    LRadialSections = PInfo->RadialSections;

	EIndex  LNCapVertexNodes;
	if(LRadialSections == 0) LNCapVertexNodes = LSides;
	else
	{
		LNCapVertexNodes = LSides * 3 + (LRadialSections - 1) * LSides * 4;
	}
	if(PInfo->TopCap) LNumOfVertexNodes += LNCapVertexNodes;
	if(PInfo->BottomCap) LNumOfVertexNodes += LNCapVertexNodes;


	const E3dCoordinate  LInvR = 1.0 / PInfo->Radius;
	EIndex               LVi = 0;

// Side
//
	E3dCoordinate  LUStep = -1.0 / (E3dCoordinate)LSides;
	E3dCoordinate  LVStep =  1.0 / (E3dCoordinate)LHeightSections;

	const E3dVertexNodeSize  LVertexNodeSize = PPolyGroup->VertexNodeSize;

	E3dCoordinate   LSideHeightH, LDomeHeight,
	                LU, LV;

	E3dPolygon*     LPolygon = PPolyGroup->Polygons;

	E3dVertex*      LVertices = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;

	const int       LNormalOffset = E3dVertexNodeAttribute_Offset(&PPolyGroup->VertexAttributes,  E3dVtxAttribute_Normal, 0);

	if(PInfo->Side)
	{
//	PMesh->PolyGroups[0]->Polygons_Count += PNSideVertices * LHeightSections;

		LV = 0.0;
		EIndex  LC = 0;for(;LC < LHeightSections;LC++, LVi += PNSideVertices, LV += LVStep)
		{
			LU = 1.0;

// Use last vtx on this section
//
			E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4,  PNSideVertices - 1 + LVi, LVi, LVi + PNSideVertices, LVi + PNSideVertices - 1 + PNSideVertices);

			{
				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

				const E3dVertex*  LVertex0 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[PNSideVertices - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[PNSideVertices - 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = -LUStep;LUV->Y = LV;
				EStruct_Inc(LVertexNode, LVertexNodeSize);EStruct_Inc(LUV, LVertexNodeSize);

				const E3dVertex*  LVertex1 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[0].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[0].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = 0.0;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[0].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[0].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = 0.0;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[PNSideVertices - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[PNSideVertices - 1].Y; }
				LUV->X = -LUStep;LUV->Y = LV + LVStep;

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);

				LPolygon++;
			}

			LU += LUStep;

			EIndex  LH = 1;for(;LH < LSides;LH++, LU += LUStep)
			{
				E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4,  LH - 1 + LVi, LH + LVi, LVi + LH + PNSideVertices, LVi + LH - 1 + PNSideVertices);
				E3dVertexNode*  LVertexNode   = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				E3dCoord2*      LUV           = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

				const E3dVertex*  LVertex0 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH - 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU - LUStep;LUV->Y = LV;
				EStruct_Inc(LVertexNode, LVertexNodeSize);EStruct_Inc(LUV, LVertexNodeSize);

				const E3dVertex*  LVertex1 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH].X;    LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH].X;    LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH - 1].Y; }
				LUV->X = LU-LUStep;LUV->Y = LV + LVStep;

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);
				LPolygon++;
			}
		}
	}


	LVStep = 1.0 / (E3dCoordinate)LRadialSections;

	if(PInfo->TopCap)
	{
		E3dCoord2*  LUnitCirclePoint;

		if(LRadialSections == 0)
		{
			E3dVertexNode*  LVertexNode = E3d_PolygonVertexNodesNew(LPolygon, LVertexNodeSize, PNSideVertices);if(LVertexNode)
			{
				E3dCoord3*  LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				E3dCoord2*  LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

				LUnitCirclePoint = PUnitArcPoints + PNSideVertices - 1;
				EIndex  LC = 0, LH = PNSideVertices - 1;for(;LC < PNSideVertices;LC++, --LUnitCirclePoint)
				{
					LVertexNode->VertexID = LH;
					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LUnitCirclePoint->X * 0.5 + 0.5;LUV->Y = LUnitCirclePoint->Y * 0.5 + 0.5;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);
					LH--;
				}
				LPolygon->NumOfVertices = LPolygon->NumOfExteriorVertices = PNSideVertices;
				LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;
				LPolygon++;
			}
		}
		else
		{
			EIndex  LCenterV = PNSideVertices * (LHeightSections + 1);

			EIndex  LNextCircleV;
			if(LRadialSections == 1) LNextCircleV = 0;
			else LNextCircleV = LCenterV+1;

			LVi = LNextCircleV;
			LV = 0.0;
			LU = 1.0;
			if((LDomeHeight = PInfo->TopDomeHeight) == 0.0)
			{
// Flat top
				EIndex  LVC = 0;for(;LVC < PNSideVertices;LVC++, LVi++, LU += LUStep)
				{
					if(LVC < PNSideVertices - 1) E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LVi + 1, LVi, LCenterV);
					else E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LNextCircleV, LVi, LCenterV);
					LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
					E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0; }
					LUV->X = LU + LUStep * 0.5;LUV->Y = LV;

					LPolygon++;
				}
			}
			else
			{
// Rounded top
				if(PInfo->OriginAtBottom) LSideHeightH = PInfo->Height;
				else LSideHeightH = PInfo->Height * 0.5;

				EIndex  LVC = 0;for(;LVC < PNSideVertices;LVC++, LVi++, LU += LUStep)
				{
					if(LVC < PNSideVertices - 1) E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LVi + 1, LVi, LCenterV);
					else E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LNextCircleV, LVi, LCenterV);

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
					E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

					const E3dVertex*  LVertex;
					const E3dVertex*  LVertex0 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					const E3dVertex*  LVertex1 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					const E3dVertex*  LVertex2 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR; }
					LUV->X = LU + LUStep * 0.5;LUV->Y = LV;

					E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);

					LPolygon++;
				}
			}

			LV += LVStep;

			if(LRadialSections > 1)
			{
				const EIndex  LRN = LRadialSections - 1;

				if(LDomeHeight == 0.0)
				{
// Flat top
					EIndex  LRC = 0;for(;LRC < LRN;LRC++, LV += LVStep)
					{
						LVi = LNextCircleV;
						if(LRC < LRN - 1) LNextCircleV = LVi + PNSideVertices;
						else LNextCircleV = 0;    // Top outer side vertices start at index 0

						LU = 1.0;
						EIndex  LVC = 0;for(;LVC < PNSideVertices;LVC++, LVi++, LU += LUStep)
						{
							if(LVC < PNSideVertices - 1) E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi + 1, LVC + LNextCircleV + 1, LVC + LNextCircleV, LVi);
							else E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi - PNSideVertices + 1, LVC + LNextCircleV - PNSideVertices + 1, LVC + LNextCircleV, LVi);

							LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;

							E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
							E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
							E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

							if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU + LUStep;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0; }
							LUV->X = LU;LUV->Y = LV;

							LPolygon++;
						}
					}
				}
				else
				{
// Rounded top
					if(PInfo->OriginAtBottom) LSideHeightH = PInfo->Height;
					else LSideHeightH = PInfo->Height * 0.5;

					EIndex  LRC = 0;for(;LRC < LRN;LRC++, LV += LVStep)
					{
						LVi = LNextCircleV;
						if(LRC < LRN - 1) LNextCircleV = LVi + PNSideVertices;
						else LNextCircleV = 0;    // Top outer side vertices start at index 0

						LU = 1.0;
						EIndex  LVC = 0;for(;LVC < PNSideVertices;LVC++, LVi++, LU += LUStep)
						{
							if(LVC < PNSideVertices - 1) E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi + 1, LVC + LNextCircleV + 1, LVC + LNextCircleV, LVi);
							else E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi - PNSideVertices + 1, LVC + LNextCircleV - PNSideVertices + 1, LVC + LNextCircleV, LVi);

							E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
							E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
							E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

							const E3dVertex*  LVertex;

							const E3dVertex*  LVertex0 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
							if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU + LUStep;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							const E3dVertex*  LVertex1 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
							if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							const E3dVertex*  LVertex2 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
							if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
							if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU;LUV->Y = LV;

							E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);

							LPolygon++;
						}
					}
				}

			}

		}
	}

	LVStep = -1.0 / (E3dCoordinate)LRadialSections;

	if(PInfo->BottomCap)
	{
		E3dCoord2*	LUnitCirclePoint;

		if(LRadialSections == 0)
		{
			E3dVertexNode*  LVertexNode = E3d_PolygonVertexNodesNew(LPolygon, LVertexNodeSize, PNSideVertices);if(LVertexNode)
			{
				E3dCoord3*  LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				E3dCoord2*  LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);
				LVi = PNSideVertices * LHeightSections;
				LUnitCirclePoint = PUnitArcPoints;
				EIndex  LVC = 0;for(;LVC < PNSideVertices;LVC++, LVi++, LUnitCirclePoint++)
				{
					LVertexNode->VertexID = LVi;
					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LUnitCirclePoint->X * 0.5 + 0.5;LUV->Y = LUnitCirclePoint->Y * 0.5 + 0.5;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);
				}
				LPolygon->NumOfVertices = LPolygon->NumOfExteriorVertices = PNSideVertices;
				LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = -1.0;LPolygon->Normal.Z = 0.0;
				LPolygon++;
			}
		}
		else
		{
			EIndex	LCenterV, LNextCircleV;

			if(PInfo->TopCap) LCenterV = PNSideVertices * (LHeightSections + 1 + LRadialSections - 1) + 1;
			else LCenterV = PNSideVertices * (LHeightSections + 1);

			if(LRadialSections == 1) LNextCircleV = PNSideVertices * LHeightSections;
			else LNextCircleV = LCenterV + 1;

			LVi = LNextCircleV;
			LV = 1.0;
			LU = 1.0;
			if((LDomeHeight = PInfo->TopDomeHeight) == 0.0)
			{
// Flat bottom
				EIndex  LVC = 0;for(;LVC < PNSideVertices;LVC++, LVi++, LU += LUStep)
				{
					if(LVC < PNSideVertices - 1) E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LCenterV, LVi, LVi + 1);
					else E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LCenterV, LVi, LNextCircleV);
					LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = -1.0;LPolygon->Normal.Z = 0.0;

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
					E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep * 0.5;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0; }
					LUV->X = LU + LUStep;LUV->Y = LV + LVStep;

					LPolygon++;
				}
			}
			else
			{
// Rounded bottom
				if(PInfo->OriginAtBottom) LSideHeightH = 0.0;
				else LSideHeightH = -PInfo->Height * 0.5;

				EIndex  LVC = 0;for(;LVC < PNSideVertices;LVC++, LVi++, LU += LUStep)
				{
					if(LVC < PNSideVertices - 1) E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LCenterV, LVi, LVi + 1);
					else E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LCenterV, LVi, LNextCircleV);

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
					E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

					const E3dVertex*  LVertex;

					const E3dVertex*  LVertex0 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep * 0.5;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					const E3dVertex*  LVertex1 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					const E3dVertex*  LVertex2 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR; }
					LUV->X = LU + LUStep;LUV->Y = LV + LVStep;

					E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);

					LPolygon++;
				}
			}

			LV += LVStep;

			if(LRadialSections > 1)
			{
				const EIndex  LRN = LRadialSections - 1;

				if(LDomeHeight == 0.0)
				{
// Flat bottom
					EIndex  LRC = 0;for(;LRC < LRN;LRC++, LV += LVStep)
					{
						LVi = LNextCircleV;
						if(LRC < LRN - 1) LNextCircleV = LVi + PNSideVertices;
						else LNextCircleV = PNSideVertices * LHeightSections;    // Bottom-outer side vertices start at index PNSideVertices * LHeightSections

						LU = 1.0;
						EIndex  LVC = 0;for(;LVC < PNSideVertices;LVC++, LVi++, LU += LUStep)
						{
							if(LVC < PNSideVertices - 1) E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi, LVC + LNextCircleV, LVC + LNextCircleV + 1, LVi + 1);
							else E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi, LVC + LNextCircleV, LVC + LNextCircleV - PNSideVertices + 1, LVi - PNSideVertices + 1);

							LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = -1.0;LPolygon->Normal.Z = 0.0;

							E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
							E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
							E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

							if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU + LUStep;LUV->Y = LV+LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0; }
							LUV->X = LU + LUStep;LUV->Y = LV;

							LPolygon++;
						}
					}
				}
				else
				{
// Rounded bottom
					if(PInfo->OriginAtBottom) LSideHeightH = 0.0;
					else LSideHeightH = -PInfo->Height * 0.5;

					EIndex  LRC = 0;for(;LRC < LRN;LRC++, LV += LVStep)
					{
						LVi = LNextCircleV;
						if(LRC < LRN - 1) LNextCircleV = LVi + PNSideVertices;
						else LNextCircleV = PNSideVertices * LHeightSections;    // Bottom-outer side vertices start at index PNSideVertices * LHeightSections

						LU = 1.0;
						EIndex  LVC = 0;for(;LVC < PNSideVertices;LVC++, LVi++, LU += LUStep)
						{
							if(LVC < PNSideVertices - 1) E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi, LVC + LNextCircleV, LVC + LNextCircleV + 1, LVi + 1);
							else E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi, LVC + LNextCircleV, LVC + LNextCircleV - PNSideVertices + 1, LVi - PNSideVertices + 1);

							E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
							E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
							E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

							const E3dVertex*  LVertex;

							const E3dVertex*  LVertex0 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
							if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							const E3dVertex*  LVertex1 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
							if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							const E3dVertex*  LVertex2 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
							if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
							LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							EStruct_Inc(LVertexNode, LVertexNodeSize);

							LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
							if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR; }
							LUV->X = LU + LUStep;LUV->Y = LV;

							E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);

							LPolygon++;
						}
					}
				}
			}
		}
	}
}


static void _CreatePolygonsForPie(const E3dCylinderInfo* PInfo, E3dMesh* PMesh, E3dPolyGroup* PPolyGroup, const EIndex PSides, const EIndex PNSideVertices, const E3dCoord2* PUnitArcPoints)
{
	E3dCoordinate   LSideHeightH, LDomeHeight;
	EIndex          LNumOfVertexNodes, LNCapVertexNodes;
	EIndex          LRadialSections = PInfo->RadialSections;
	const EIndex    LHeightSections = PInfo->HeightSections;

	LNumOfVertexNodes = 0;

	E3dPolyGroup_VertexAttributeAdd(PPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);

// Make sure the VertexNodes have UVs
//
	E3dVertexNodeSize  LUVOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_UV, 0);
	if(LUVOffsetInVtxNode < 0)
	{
		E3dPolyGroup_VertexAttributeAdd(PPolyGroup,  NULL, E3dVtxAttribute_UV, 0,  NULL);
		LUVOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, E3dVtxAttribute_UV, 0);
	}

	if(PInfo->Side) LNumOfVertexNodes += LHeightSections * PSides * 4;

	if(LRadialSections == 0) LNCapVertexNodes = PSides;
	else
	{
		LNCapVertexNodes = PSides * 3 + (LRadialSections - 1) * PSides * 4;
	}
	if(PInfo->TopCap) LNumOfVertexNodes += LNCapVertexNodes;
	if(PInfo->BottomCap) LNumOfVertexNodes += LNCapVertexNodes;


	if(LRadialSections < 1) LRadialSections = 1;

	const E3dCoordinate  LUStep = -1.0 / (E3dCoordinate)PSides;
	E3dCoordinate        LVStep = 1.0 / (E3dCoordinate)LHeightSections;

	const E3dCoordinate  LInvR = 1.0/PInfo->Radius;
	EIndex               LVi = 0;

	const int            LNormalOffset = E3dVertexNodeAttribute_Offset(&PPolyGroup->VertexAttributes,  E3dVtxAttribute_Normal, 0);

	E3dVertex*       LVertices = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;

// Side
//
	const E3dVertexNodeSize  LVertexNodeSize = PPolyGroup->VertexNodeSize;
	E3dPolygon*  LPolygon = PPolyGroup->Polygons;
	if(PInfo->Side)
	{
		EIndex  LC = 0;E3dCoordinate  LV = 0.0;for(;LC < LHeightSections;LC++, LVi += PNSideVertices, LV += LVStep)
		{
			EIndex  LH = 0;E3dCoordinate  LU = 1.0;for(;LH < PSides;LH++, LU += LUStep)
			{
				E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LH + LVi, LH + 1 + LVi, LVi + LH + 1 + PNSideVertices, LVi + LH + PNSideVertices);

				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

				const E3dVertex*  LVertex0 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH].X;    LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex1 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH + 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH + 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU + LUStep;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH + 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH + 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH].X;    LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH].Y; }
				LUV->X = LU;LUV->Y = LV + LVStep;

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);
				LPolygon++;
			}
		}
	}

	LVStep = 1.0 / (E3dCoordinate)LRadialSections;

	if(PInfo->TopCap)
	{
		EIndex  LCenterV = PNSideVertices * (LHeightSections + 1);
		EIndex  LNextCircleV;

		if(LRadialSections == 1) LNextCircleV = 0;
		else LNextCircleV = LCenterV + 1;

// Triangles in the middle
//
		LVi = LNextCircleV;
		E3dCoordinate  LV = 0.0;
		E3dCoordinate  LU = 1.0;
		if((LDomeHeight = PInfo->TopDomeHeight) == 0.0)
		{
// Flat top
			EIndex  LVC = 0;for(;LVC < PSides;LVC++, LVi++, LU += LUStep)
			{
				E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LVi + 1, LVi, LCenterV);
				LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;

				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

				if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0; }
				LUV->X = LU + LUStep * 0.5;LUV->Y = LV;

				LPolygon++;
			}
		}
		else
		{
// Rounded top
			if(PInfo->OriginAtBottom) LSideHeightH = PInfo->Height;
			else LSideHeightH = PInfo->Height * 0.5;

			EIndex  LVC = 0;for(;LVC < PSides;LVC++, LVi++, LU += LUStep)
			{
				E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LVi + 1, LVi, LCenterV);

				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

				const E3dVertex*  LVertex;

				const E3dVertex*  LVertex0 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex1 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR; }
				LUV->X = LU + LUStep * 0.5;LUV->Y = LV;

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);

				LPolygon++;
			}
		}

		LV += LVStep;
		const EIndex  LRN = LRadialSections - 1;

		if(LDomeHeight == 0.0)
		{
// Flat top
			EIndex  LRC = 0;for(;LRC < LRN;LRC++, LV += LVStep)
			{
				LVi = LNextCircleV;
				if(LRC < LRN - 1) LNextCircleV = LVi + PNSideVertices;
				else LNextCircleV = 0;			// Top outer side vertices start at index 0

				LU = 1.0;
				EIndex  LVC = 0;for(;LVC < PSides;LVC++, LVi++, LU += LUStep)
				{
					E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi + 1, LVC + LNextCircleV + 1, LVC + LNextCircleV, LVi);

					LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
					E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0; }
					LUV->X = LU;LUV->Y = LV;

					LPolygon++;
				}
			}
		}
		else
		{
// Rounded top
			if(PInfo->OriginAtBottom) LSideHeightH = PInfo->Height;
			else LSideHeightH = PInfo->Height * 0.5;

			EIndex  LRC = 0;for(;LRC < LRN;LRC++, LV += LVStep)
			{
				LVi = LNextCircleV;
				if(LRC < LRN - 1) LNextCircleV = LVi + PNSideVertices;
				else LNextCircleV = 0;  // Top outer side vertices start at index 0

				EIndex  LVC = 0;for(;LVC < PSides;LVC++, LVi++, LU += LUStep)
				{
					E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi + 1, LVC + LNextCircleV + 1, LVC + LNextCircleV, LVi);
					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
					E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

					const E3dVertex*  LVertex;

					const E3dVertex*  LVertex0 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					const E3dVertex*  LVertex1 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					const E3dVertex*  LVertex2 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR; }
					LUV->X = LU;LUV->Y = LV;

					E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);

					LPolygon++;
				}
			}

		}
	}

	LVStep = -1.0 / (E3dCoordinate)LRadialSections;

	if(PInfo->BottomCap)
	{
		EIndex  LCenterV;
		if(PInfo->TopCap) LCenterV = PNSideVertices*(LHeightSections + 1 + LRadialSections - 1) + 1;
		else LCenterV = PNSideVertices * (LHeightSections + 1);

		EIndex  LNextCircleV;
		if(LRadialSections == 1) LNextCircleV = PNSideVertices * LHeightSections;
		else LNextCircleV = LCenterV + 1;

		LVi = LNextCircleV;
		E3dCoordinate  LV = 1.0;
		E3dCoordinate  LU = 1.0;
		if((LDomeHeight = PInfo->TopDomeHeight) == 0.0)
		{
// Flat bottom
			EIndex  LVC = 0;for(;LVC < PSides;LVC++, LVi++, LU += LUStep)
			{
				E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 3, LCenterV, LVi, LVi + 1);
				LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = -1.0;LPolygon->Normal.Z = 0.0;

				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

				if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU + LUStep * 0.5;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0; }
				LUV->X = LU + LUStep;LUV->Y = LV + LVStep;

				LPolygon++;
			}
		}
		else
		{
// Rounded bottom
			if(PInfo->OriginAtBottom) LSideHeightH = 0.0;
			else LSideHeightH = -PInfo->Height * 0.5;

			EIndex  LVC = 0;for(;LVC < PSides;LVC++, LVi++, LU += LUStep)
			{
				E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LCenterV, LVi, LVi + 1);

				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

				const E3dVertex*  LVertex;

				const E3dVertex*  LVertex0 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU+LUStep*0.5;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex1 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR; }
				LUV->X = LU + LUStep;LUV->Y = LV + LVStep;

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);

				LPolygon++;
			}
		}

		LV += LVStep;
		const EIndex  LRN = LRadialSections - 1;

		if(LDomeHeight == 0.0)
		{
// Flat bottom
			EIndex  LRC = 0;for(;LRC < LRN;LRC++, LV += LVStep)
			{
				LVi = LNextCircleV;
				if(LRC < LRN - 1) LNextCircleV = LVi + PNSideVertices;
				else LNextCircleV = PNSideVertices * LHeightSections;  // Bottom-outer side vertices start at index PNSideVertices * LHeightSections

				LU = 1.0;
				EIndex  LVC = 0;for(;LVC < PSides;LVC++, LVi++, LU += LUStep)
				{
					E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi, LVC + LNextCircleV, LVC + LNextCircleV + 1, LVi + 1);

					LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = -1.0;LPolygon->Normal.Z = 0.0;

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
					E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0; }
					LUV->X = LU + LUStep;LUV->Y = LV;

					LPolygon++;
				}
			}
		}
		else
		{
// Rounded bottom
			if(PInfo->OriginAtBottom) LSideHeightH = 0.0;
			else LSideHeightH = -PInfo->Height * 0.5;

			EIndex  LRC = 0;for(;LRC < LRN;LRC++, LV += LVStep)
			{
				LVi = LNextCircleV;
				if(LRC < LRN - 1) LNextCircleV = LVi + PNSideVertices;
				else LNextCircleV = PNSideVertices * LHeightSections;	// Bottom-outer side vertices start at index PNSideVertices * LHeightSections

				LU = 1.0;
				EIndex  LVC = 0;for(;LVC<PSides;LVC++, LVi++, LU += LUStep)
				{
					E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi, LVC + LNextCircleV, LVC + LNextCircleV + 1, LVi + 1);

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
					E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

					const E3dVertex*  LVertex;

					const E3dVertex*  LVertex0 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					const E3dVertex*  LVertex1 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					const E3dVertex*  LVertex2 = LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
					if(LVertexNormal) { LVertexNormal->X = LVertex->Position.X * LInvR;LVertexNormal->Y = (LVertex->Position.Y - LSideHeightH) * LInvR;LVertexNormal->Z = LVertex->Position.Z * LInvR; }
					LUV->X = LU + LUStep;LUV->Y = LV;

					E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);

					LPolygon++;
				}
			}
		}
	}


// Extra sides
//
	{
//		EIndex  LNextCircleV = PNSideVertices;
		EIndex  LRC = 0;const EIndex  LRN = LRadialSections - 1;for(;LRC < LRN;LRC++)
		{
			EIndex  LVC = 0;for(;LVC < LHeightSections;LVC++)
			{
//				E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LVi, LVC + LNextCircleV, LVC + LNextCircleV + 1, LVi + 1);
			}
		}
	}

	PPolyGroup->Polygons_Count = LPolygon - PPolyGroup->Polygons;
}


//========================================
// Update() (create/change) method
//========================================
static E3dObject* _Update(E3dObject* PObject, E3dObjectComponent* PInfo, E3dObjectComponentClass* PClass, const EChangeFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP)
{
	E3dScene*         LScene = E3d_Scene;
	E3dMesh*          LMesh;
	E3dPolyGroup*     LPolyGroup;
	E3dCylinderInfo*  LInfo;

	if(PObject == NULL)
	{
		E_New(E3dMesh, LMesh);

		LInfo = (E3dCylinderInfo*)E3dObjectComponent_Add((E3dObject*)LMesh, PClass);
		if(PInfo) memcpy(LInfo, PInfo, PClass->StructSize);
		LPolyGroup = E3dMesh_PolyGroupAdd(LMesh);if(LPolyGroup == NULL)
		{
			E_Delete(LMesh);return(NULL);
		}
		LPolyGroup->VertexNormalType = E3dNormalPROCESS;
	}
	else
	{
		if(!EObject_IsSubClass(PObject, E3dMesh)) return(NULL);
		LMesh = (E3dMesh*)PObject;
		LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(PObject, PClass);
		LPolyGroup = LMesh->PolyGroups[0];
	}


/*
	if(LRadialSections < 1)
	{
		E3dObjectComponentClass*  LExtrudeClass = E3dObjectComponent_ClassFindByName("Extrude");if(LExtrudeClass == NULL) return(NULL);
	}
*/

// Create a profile Spline and use the Revolve tool to create the cylinder
//
	if(_UseRevolve)
	{
		E3dObjectComponentClass*  LRevolveClass = E3dObjectComponent_ClassFindByName("Revolve");if(LRevolveClass)
		{
			const EIndex   LRadialSections = LInfo->RadialSections >= 1 ? LInfo->RadialSections : 1;
			const EIndex   LSides = LInfo->Sides, LHeightSections = LInfo->HeightSections;
			E3dCoordinate  LStartAngle = LInfo->StartAngle, LEndAngle = LInfo->EndAngle;

			E3dSpline*  LSpline = NULL;E3dObjectComponent*  LRevolveInfo = NULL;
			if(PObject == NULL)
			{
				LRevolveInfo = E3dObjectComponent_Add((E3dObject*)LMesh, LRevolveClass);

// Revolve is an internal representation of the Cylinder, so hide its ObjectComponent in the GUI
//
				LRevolveInfo->InternalFlags &= ~(EComponentVISIBLE|EComponentREMOVABLE);
				LRevolveInfo->InternalFlags = 0;

				E_New_n(E3dSpline, LSpline, "profile");if(LSpline)
				{
					LSpline->SplineType = E3dSplineBEZIER;  // Make the Spline Bezier, for flexibility. But, we'll just create linear segments here.

					ERes_NamedResourceSetObject(LRevolveInfo, "ProfileGeometry", LSpline);  // This is how you pass an Object as a Resource. In this case, the profile Spline to be revolved
				}
			}
			else
			{
				LRevolveInfo = E3dObjectComponent_ByClass((E3dObject*)LMesh, LRevolveClass);if(LRevolveInfo)
				{
					ERes_NamedResourceGetObject(LRevolveInfo, "ProfileGeometry", LSpline);  // This is how you get an Object as a Resource. In this case, the profile Spline to be revolved
				}
			}

// Create Spline CVs in counter-clockwise order, so the revolved surface will be the correct side out.
//
			if(LRevolveInfo && LSpline)
			{
				const E3dCoordinate  LRadius = LInfo->Radius, LHeight = LInfo->Height;
				const E3dCoordinate  LRadialStep = LRadius / (E3dCoordinate)LRadialSections;
				const E3dCoordinate  LHeightStep = LHeight / (E3dCoordinate)LHeightSections;

				EIndex  LNCVs = 0;

				LNCVs += LRadialSections;
				LNCVs += LHeightSections;
				LNCVs += LRadialSections;
				LNCVs++;

//printf("%s() %s:%d  %sNCVs%s %lld\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode, (long long)LNCVs);fflush(stdout);

				E3dSplineCV*  LCV = E3dSplineCVs_New(E3dSplineLINEAR, LNCVs);E3dSpline_SetCVs(LSpline, LCV, LNCVs);

				{
					E3dCoordinate  LY = LInfo->OriginAtBottom ? 0.0 : -LHeight / 2.0;
					{
						const uint16_t  LSegmentFlags = LInfo->BottomCap ? E3dSegmentDEFAULT : E3dSegmentDEFAULT & ~E3dSegmentVISIBLE;EIndex  LC = 0;E3dCoordinate  LX = 0.0;for(;LC < LRadialSections;LC++, LX += LRadialStep) { E_SetXYZ_s(LCV->Position,  LX, LY, 0.0);LCV->SegmentFlags = LSegmentFlags;LCV++; }
					}
					{
						const uint16_t  LSegmentFlags = LInfo->Side ? E3dSegmentDEFAULT : E3dSegmentDEFAULT & ~E3dSegmentVISIBLE;EIndex  LC = 0;E3dCoordinate  LX = LRadius;for(;LC < LHeightSections;LC++, LY += LHeightStep) { E_SetXYZ_s(LCV->Position,  LX, LY, 0.0);LCV->SegmentFlags = LSegmentFlags;LCV++; }
					}
				}

				{
					E3dCoordinate  LY = LInfo->OriginAtBottom ? LHeight : LHeight / 2.0;
					{
						const uint16_t  LSegmentFlags = LInfo->TopCap ? E3dSegmentDEFAULT : E3dSegmentDEFAULT & ~E3dSegmentVISIBLE;EIndex  LC = 0;E3dCoordinate  LX = LRadius;for(;LC < LRadialSections;LC++, LX -= LRadialStep) { E_SetXYZ_s(LCV->Position,  LX, LY, 0.0);LCV->SegmentFlags = LSegmentFlags;LCV++; }
					}
					E_SetXYZ_s(LCV->Position,  0.0, LY, 0.0);LCV++;
				}

				E3dSpline_CreateLinearSegments(LSpline);

				ERes_NamedResourceSet1i(LRevolveInfo, "Sides",          LSides);        // Steps for the revolve
				ERes_NamedResourceSet1f(LRevolveInfo, "SmoothingAngle", 30.0);          // Normal smoothing angle for the Revolve
				ERes_NamedResourceSet1f(LRevolveInfo, "StartAngle",     LStartAngle);
				ERes_NamedResourceSet1f(LRevolveInfo, "EndAngle",       LEndAngle);
				ERes_NamedResourceSet1b(LRevolveInfo, "UseProfileSegmentVisibility", TRUE);

				LRevolveClass->Update((E3dObject*)LMesh, NULL,  E3dUF_ALL, NULL, FALSE);

				return((E3dObject*)LMesh);
			}
		}
	}


	const EIndex   LSides = LInfo->Sides, LHeightSections = LInfo->HeightSections;
	E3dCoordinate  LStartAngle = LInfo->StartAngle, LEndAngle = LInfo->EndAngle;


	EBool         LFull, LPie;
	if(LEndAngle - LStartAngle == 360.0) LFull = TRUE;else LFull = FALSE;

	if(LFull || (!(LInfo->Pie))) LPie = FALSE;else LPie = TRUE;

	LInfo->Full = LFull;

	E3dCoord2*    LUnitArcPoints = NULL;
	EIndex        LSidesN = 0;
	EIndex        LVNum = 0, LPNum = 0;

	EIndex  LRadialSections = LInfo->RadialSections;

	switch(LInfo->Type)
	{
		case E3dCylinderPOLYGONS:
			LVNum = LPNum = 0;

// Create number of Vertices and Polygons and initialize the unit cross-section
//
			if(LFull || (!(LInfo->Pie)))
			{
				if(LFull) LSidesN = LSides;
				else LSidesN = LSides + 1;

				LVNum += LSidesN * (LHeightSections + 1);
				if(LInfo->Side)
				{
					LPNum += LSidesN * LHeightSections;
				}

				if(LRadialSections < 1)
				{
					if(LInfo->TopCap) LPNum++;
					if(LInfo->BottomCap) LPNum++;
				}
				else
				{
					if(LInfo->TopCap)
					{
						LVNum += ((LRadialSections - 1) * LSidesN + 1);    // "+1" middle vertex
						LPNum += LRadialSections * LSidesN;
					}

					if(LInfo->BottomCap)
					{
						LVNum += ((LRadialSections - 1) * LSidesN + 1);    // "+1" middle vertex
						LPNum += LRadialSections * LSidesN;
					}
				}
			}
			else
			{
// "Pie"
//
				LPNum = LSides * LHeightSections;
				LSidesN = LSides + 1;

				if(LRadialSections < 1) LRadialSections = 1;
				LVNum = (LSides + 1) * (LHeightSections + 1);

				if(LInfo->TopCap)
				{
					LVNum += (LSides + 1) * (LRadialSections - 1) + 1;LPNum += LRadialSections * LSides;
				}
				else if(_AddPieSides) LVNum++;

				if(LInfo->BottomCap)
				{
					LVNum += (LSides + 1) * (LRadialSections - 1) + 1;LPNum += LRadialSections * LSides;
				}
				else if(_AddPieSides) LVNum++;
			}

			LUnitArcPoints = E3d_CreateUnitArc(LSides, LStartAngle, LEndAngle,  NULL);

			_VNum = LVNum;
		break;
	}

	E3dMesh_FreeCaches(LMesh);
	E3dMesh_EdgesFree(LMesh);

	if(LVNum != LMesh->Vertices_Count)
	{
		E3dMesh_VerticesFree(LMesh);
		E3dMesh_VerticesAdd(LMesh, LVNum);
	}

	if(LMesh->Vertices)
	{
		switch(LInfo->Type)
		{
			case E3dCylinderPOLYGONS:
				{
					E3dPolygon*  LPolygon;

					if(LPNum == LPolyGroup->Polygons_Count) LPolygon = LPolyGroup->Polygons;
					else
					{
						if(LPolyGroup->Polygons) E3dPolyGroup_FreePolygons(LPolyGroup);
						LPolygon = LPolyGroup->Polygons = E3dPolygons_New(LPolyGroup->Polygons_Count = LPNum);
					}

					_SetVertices(LMesh, LInfo, LUnitArcPoints);

					if(LPolygon)
					{
						if(LPie) _CreatePolygonsForPie(LInfo, LMesh, LPolyGroup, LSides, LSidesN, LUnitArcPoints);
						else _CreatePolygonsForCylinder(LInfo, LMesh, LPolyGroup, LSidesN, LUnitArcPoints);

						if(LInfo->Smooth) LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;
						else LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;
						LPolyGroup->SmoothingAngle = 30.0;

						EChangeFlags  LTriggeredChanges = E3dObject_UpdateForDisplay((E3dObject*)LMesh, PUpdateFlags, LScene);
						if(PTriggeredChangesP) *PTriggeredChangesP |= LTriggeredChanges;
					}
				}
			break;
		}
	}

	EFreeChk(LUnitArcPoints);

	return((E3dObject*)LMesh);
}


static E3dObject* _CylinderUpdate(E3dObject* PObject, E3dObjectComponent* PInfo, const EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	if(PCheckOnly) return(NULL);
	else return(_Update(PObject, PInfo, _CylinderComponentClass, PUpdateFlags, PTriggeredChangesP));
}


static E3dObject* _CapsuleUpdate(E3dObject* PObject, E3dObjectComponent* PInfo, const EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	if(PCheckOnly) return(NULL);
	else return(_Update(PObject, PInfo, _CapsuleComponentClass, PUpdateFlags, PTriggeredChangesP));
}


//========================================
// Set GUI for cylinder type
//========================================
static void _SetGUIToType(E3dCylinderInfo* PInfo, E3dResourcesGUI* PResourcesGUI)
{
	Widget  LIndentWs[7];int  LIndentN;      // For indenting the ScaleWidgets
	Widget  LTraverseWs[7];int  LTraverseN;  // For TextField traversal

	switch(PInfo->Type)
	{
		case E3dCylinderNURBS:
		case E3dCylinderPOLYGONS:
			LIndentN = LTraverseN = 0;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->RadiusScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->HeightScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->StartAngleScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->EndAngleScaleW;
			Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);

			LIndentN = 0;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->SidesScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->RadialSectionsScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->HeightSectionsScaleW;
			Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);

			Xe_TraversalRing(LTraverseWs, LTraverseN);
		break;
	}
}


/*
//========================================
// Set Cylinder type
//========================================
static void _CBSetType(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dCylinderInfo*  PInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass((E3dObject*)_Geometry, _ComponentClass);


	if(LInfo->Type != (int)PClientData)
	{
		EFreeChkClr(_UnitCirclePoints);
		LInfo->Type = (int)PClientData;
		_SetGUIToType(LInfo);

		_Create(_Model, _Geometry, LInfo, E3dUF_ALL);

		E3dWindows_RedrawByChangeFlags(E3dChgSHAPE);
	}
}
*/


//========================================
// Dialog-box Xt callbacks
//========================================
static void _DCB_Smooth(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dMesh*                  LMesh = (E3dMesh*)LObject;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LObject, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->Smooth = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

		if(LInfo->Smooth) LMesh->PolyGroups[0]->Flags |= E3dUSE_VERTEX_NORMALS;
		else LMesh->PolyGroups[0]->Flags &= ~E3dUSE_VERTEX_NORMALS;

		E3dMesh_RecomputeNormals(LMesh, FALSE);
		LChanged |= E3dChgVERTEX_ATTRIBUTE;

		E3dObject_UpdateForDisplay((E3dObject*)LMesh, E3duf_NORMALS|E3duf_CALLBACKS, E3d_Scene);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_Side(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LObject, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->Side = (((XeToggleButtonCallbackStruct*)PCallData)->Set);

		E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		E3dWindows_RedrawByChangeFlags(LChanged);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_CapTop(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LObject, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->TopCap = (((XeToggleButtonCallbackStruct*)PCallData)->Set);

		E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_CapBottom(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LPanel->Object, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->BottomCap = (((XeToggleButtonCallbackStruct*)PCallData)->Set);

		E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_OriginAtBottom(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*                 LScene = E3d_Scene;
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LObject, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->OriginAtBottom = (((XeToggleButtonCallbackStruct*)PCallData)->Set);

		if(_UseRevolve) E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		else _SetVertices((E3dMesh*)LObject, LInfo, NULL);

		LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_Pie(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LPanel->Object, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->Pie = (((XeToggleButtonCallbackStruct*)PCallData)->Set);

		E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_Radius(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*                 LScene = E3d_Scene;
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LPanel->Object, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->Radius = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

		if(_UseRevolve) E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		else _SetVertices((E3dMesh*)LObject, LInfo, NULL);

		LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_Height(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*                 LScene = E3d_Scene;
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LPanel->Object, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->Height = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

		if(_UseRevolve) E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		else _SetVertices((E3dMesh*)LObject, LInfo, NULL);

		LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_StartAngle(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*         LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LObject, LClass);
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, LClass);if(LDrawer)
		{
			E3dResourcesGUI*  LResourcesGUI = (E3dResourcesGUI*)LDrawer->ResourcesGUI;
			EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

			LInfo->StartAngle = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

			if(LInfo->StartAngle > LInfo->EndAngle)
			{
				LInfo->EndAngle = LInfo->StartAngle;
				XeScale_SetValue_Coord(LResourcesGUI->EndAngleScaleW, (double)LInfo->EndAngle, FALSE);
			}
			if(LInfo->StartAngle < LInfo->EndAngle - 360.0)
			{
				LInfo->EndAngle = LInfo->StartAngle + 360.0;
				XeScale_SetValue_Coord(LResourcesGUI->EndAngleScaleW, (double)LInfo->EndAngle, FALSE);
			}

//printf("%s() %f-%f  span: %f\n", __FUNCTION__, LInfo->StartAngle, LInfo->EndAngle, LInfo->EndAngle-LInfo->StartAngle);fflush(stdout);
			E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
			E3dWindows_RedrawByChangeFlags(LChanged);
		}
	}
}


static void _DCB_EndAngle(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*         LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LObject, LClass);
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, LClass);if(LDrawer)
		{
			E3dResourcesGUI*  LResourcesGUI = (E3dResourcesGUI*)LDrawer->ResourcesGUI;
			EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);


//			EBool  LWasFull;if((LInfo->EndAngle - LInfo->StartAngle) == 360.0) LWasFull = TRUE;else LWasFull = FALSE;

			LInfo->EndAngle = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

			if(LInfo->StartAngle > LInfo->EndAngle)
			{
				LInfo->StartAngle = LInfo->EndAngle;
				XeScale_SetValue_Coord(LResourcesGUI->StartAngleScaleW, (double)(LInfo->StartAngle), FALSE);
			}
			if(LInfo->StartAngle < LInfo->EndAngle - 360.0)
			{
				LInfo->StartAngle = LInfo->EndAngle - 360.0;
				XeScale_SetValue_Coord(LResourcesGUI->StartAngleScaleW, (double)(LInfo->StartAngle), FALSE);
			}

//			EBool  LFull;if((LInfo->EndAngle - LInfo->StartAngle) == 360.0) LFull = TRUE;else LFull = FALSE;

			E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
			E3dWindows_RedrawByChangeFlags(LChanged);
		}
	}
}


static void _DCB_Sides(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LObject, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);


		LInfo->Sides = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

		E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_RadialSections(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LObject, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->RadialSections = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

		E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_HeightSections(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*           LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*                LObject = LPanel->Object;
	E3dObjectComponentClass*  LClass = NULL;Xe_GetValuesVa(PW,  XeNuserData, &LClass,  NULL);

	{
		E3dCylinderInfo*  LInfo = (E3dCylinderInfo*)E3dObjectComponent_ByClass(LObject, LClass);
		EChangeFlags      LChanged = E3dObject_GetChangeFlagsByClass(LObject);

		LInfo->HeightSections = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

		E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


//========================================
// GUI
//========================================
static void _GUI(E3dObject* PObject, E3dObjectComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dResourcesGUI*         LResourcesGUI;
	E3dCylinderInfo*         LInfo = (E3dCylinderInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LInfo)
			{
				if(LDrawer->ResourcesGUI)
				{
					LResourcesGUI = (E3dResourcesGUI*)LDrawer->ResourcesGUI;

					XeToggleButton_SetState(LResourcesGUI->SmoothW, LInfo->Smooth, FALSE);

/*
					if(LInfo->TopDomeHeight == 0.0) { XeWidget_Manage(LResourcesGUI->CylinderBigW); XeWidget_Unmanage(LResourcesGUI->CapsuleBigW); }
					else { XeWidget_Unmanage(LResourcesGUI->CylinderBigW); XeWidget_Manage(LResourcesGUI->CapsuleBigW); }
*/

					XeToggleButton_SetState(LResourcesGUI->SideW, LInfo->Side, FALSE);
					XeToggleButton_SetState(LResourcesGUI->TopW, LInfo->TopCap, FALSE);
					XeToggleButton_SetState(LResourcesGUI->BottomW, LInfo->BottomCap, FALSE);
					XeToggleButton_SetState(LResourcesGUI->OriginAtBottomW, LInfo->OriginAtBottom, FALSE);
					XeToggleButton_SetState(LResourcesGUI->PieW, LInfo->Pie, FALSE);

					XeScale_SetValue_Coord(LResourcesGUI->RadiusScaleW, LInfo->Radius, FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->HeightScaleW, LInfo->Height, FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->StartAngleScaleW, LInfo->StartAngle, FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->EndAngleScaleW, LInfo->EndAngle, FALSE);
					XeScale_SetValue(LResourcesGUI->SidesScaleW, LInfo->Sides, FALSE);
					XeScale_SetValue(LResourcesGUI->RadialSectionsScaleW, LInfo->RadialSections, FALSE);
					XeScale_SetValue(LResourcesGUI->HeightSectionsScaleW, LInfo->HeightSections, FALSE);

					_SetGUIToType(LInfo, LResourcesGUI);
				}
				else
				{
					Widget  LTopMatrixW = LDrawer->Drawer.DrawerW, LW;
					Widget  LIndentWs[7];int  LIndentN;      // For indenting the ScaleWidgets
					Widget  LTraverseWs[7];int  LTraverseN;  // For TextField traversal
					EArg    Args[256];int  ArgCnt;

					ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

					{
						EXeStart;
						EXeSetArg(XeNorientation, XeHORIZONTAL);
						EXeSetArg(XeNspacingX, 8);
						EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
						EXeSetArg(XeNheightAdjustable, FALSE);
						Widget  LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->Smooth);
						EXeSetArg(XeNheightAdjustable, FALSE);
						EXeSetArg(XeNuserData, LInfo->Class);
						LResourcesGUI->SmoothW = LW = XeWidget_ManagedNew("Smooth", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Smooth, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->OriginAtBottom);
						EXeSetArg(XeNuserData, LInfo->Class);
						LResourcesGUI->OriginAtBottomW = LW = XeWidget_ManagedNew("Origin at bottom", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_OriginAtBottom, (EPointer)LPanel);
					}


					{
						EXeStart;
						EXeSetArg(XeNorientation, XeHORIZONTAL);
						EXeSetArg(XeNspacingX, 8);
						EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
						EXeSetArg(XeNheightAdjustable, FALSE);
						Widget  LMatrixW = XeWidget_ManagedNew("LMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->Side);
						EXeSetArg(XeNuserData, LInfo->Class);
						LResourcesGUI->SideW = LW = XeWidget_ManagedNew("Side", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Side, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->TopCap);
						EXeSetArg(XeNuserData, LInfo->Class);
						LResourcesGUI->TopW = LW = XeWidget_ManagedNew("Cap top", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CapTop, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->BottomCap);
						EXeSetArg(XeNuserData, LInfo->Class);
						LResourcesGUI->BottomW = LW = XeWidget_ManagedNew("Cap bottom", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CapBottom, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->Pie);
						EXeSetArg(XeNuserData, LInfo->Class);
						LResourcesGUI->PieW = LW = XeWidget_ManagedNew("Pie", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Pie, (EPointer)LPanel);
					}

					EXeStart;
					EXeSetArg(XeNshadowThickness, 2);
					EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeWidget_ManagedNew("", XeSeparatorWidget_class, LTopMatrixW,  Args, ArgCnt);

					LIndentN = 0;
					LTraverseN = 0;
					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNuserData, LInfo->Class);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->RadiusScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Radius",  Args, ArgCnt);
					XeScale_SetValueAndRanges_Coord(LW,  LInfo->Radius,  0.0001, 1000000.0,  0.0001, 100.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Radius, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNuserData, LInfo->Class);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->HeightScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Height",  Args, ArgCnt);
					XeScale_SetValueAndRanges_Coord(LW,  LInfo->Height,  0.0001, 1000000.0,  0.0001, 100.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Height, (EPointer)LPanel);

					Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


					LIndentN = 0;
					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 2);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					EXeSetArg(XeNuserData, LInfo->Class);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->StartAngleScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Start angle",  Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->StartAngle,  -360.0, 360.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_StartAngle, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 2);
					EXeSetArg(XeNincrement, 100);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNuserData, LInfo->Class);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->EndAngleScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "End angle",  Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->EndAngle,  -360.0, 360.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_EndAngle, (EPointer)LPanel);

					Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


					LIndentN = 0;
					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 4);
					EXeSetArg(XeNminimum, 3);EXeSetArg(XeNmaximum, 360);EXeSetArg(XeNvalue, LInfo->Sides);
					EXeSetArg(XeNscaleMaximum, 60);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNuserData, LInfo->Class);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SidesScaleW = LW = XeWidget_ManagedNew("Sides", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Sides, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 4);
					EXeSetArg(XeNminimum, 0);EXeSetArg(XeNmaximum, 100);EXeSetArg(XeNvalue, LInfo->RadialSections);
					EXeSetArg(XeNscaleMinimum, 0);EXeSetArg(XeNscaleMaximum, 24);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNuserData, LInfo->Class);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->RadialSectionsScaleW = LW = XeWidget_ManagedNew("Radial segments", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_RadialSections, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 4);
					EXeSetArg(XeNminimum, 1);EXeSetArg(XeNmaximum, 100);EXeSetArg(XeNvalue, LInfo->HeightSections);
					EXeSetArg(XeNscaleMinimum, 1);EXeSetArg(XeNscaleMaximum, 20);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNuserData, LInfo->Class);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->HeightSectionsScaleW = LW = XeWidget_ManagedNew("Height sections", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_HeightSections, (EPointer)LPanel);

					Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


// Set up TextField traversal
//
					Xe_TraversalRing(LTraverseWs, LTraverseN);


					_SetGUIToType(LInfo, LResourcesGUI);
				}
			}
		break;
	}
}


//================================================
// Menu callback for Cylinder
//================================================
static void _MCB_Cylinder(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dTpl_ObjectComponentActivateCB_Geometry(E3dCylinder, _Cylinder, _Plugin);
}


//================================================
// Menu callback for Capsule
//================================================
static void _MCB_Capsule(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dTpl_ObjectComponentActivateCB_Geometry(E3dCapsule, _Capsule, _Plugin);
}


//================================================
//================================================
static int _CylinderInit(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dCylinderInfo*	LThis = (E3dCylinderInfo*)PInfo;

	LThis->TopDomeHeight = 0.0;LThis->BottomDomeHeight = 0.0;
	LThis->RadialSections = 0;
	LThis->Radius = 1.0;LThis->Height = 2.0;

	LThis->Sides          = 18;
	LThis->HeightSections = 1;

	LThis->Type = E3dCylinderPOLYGONS;

	LThis->StartAngle = 0.0;LThis->EndAngle = 360.0;LThis->Full = TRUE;
	LThis->Smooth = TRUE;
	LThis->Side = TRUE;LThis->TopCap = TRUE;LThis->BottomCap = TRUE;
	LThis->Pie = TRUE;
	LThis->OriginAtBottom = FALSE;

	return(E_SUCCESS);
}


//================================================
//================================================
static int _CapsuleInit(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dCylinderInfo*	LThis = (E3dCylinderInfo*)PInfo;

	LThis->TopDomeHeight = 1.0;LThis->BottomDomeHeight = 1.0;
	LThis->RadialSections = 8;
	LThis->Radius = 1.0;LThis->Height = 2.0;

	LThis->Sides          = 18;
	LThis->HeightSections = 1;

	LThis->Type = E3dCylinderPOLYGONS;

	LThis->StartAngle = 0.0;LThis->EndAngle = 360.0;LThis->Full = TRUE;
	LThis->Smooth = TRUE;
	LThis->Side = TRUE;LThis->TopCap = TRUE;LThis->BottomCap = TRUE;
	LThis->Pie = TRUE;
	LThis->OriginAtBottom = FALSE;

	return(E_SUCCESS);
}


//========================================
// Copy-constructor
//========================================
static void _Copy(const E3dObject* LSObject, const E3dObjectComponent* LSInfoIn, E3dObject* LDObject, E3dObjectComponent* LDInfoIn)
{
	const E3dCylinderInfo*  LSInfo = (E3dCylinderInfo*)LSInfoIn;
	E3dCylinderInfo*        LDInfo = (E3dCylinderInfo*)LDInfoIn;

	LDInfo->TopDomeHeight = LSInfo->TopDomeHeight;
	LDInfo->RadialSections = LSInfo->RadialSections;
	LDInfo->Radius = LSInfo->Radius;
	LDInfo->Height = LSInfo->Height;
	LDInfo->Sides = LSInfo->Sides;
	LDInfo->HeightSections = LSInfo->HeightSections;
	LDInfo->Type = LSInfo->Type;
	LDInfo->StartAngle = LSInfo->StartAngle;
	LDInfo->EndAngle = LSInfo->EndAngle;
	LDInfo->Full = LSInfo->Full;
	LDInfo->Smooth = LSInfo->Smooth;
	LDInfo->Side = LSInfo->Side;
	LDInfo->TopCap = LSInfo->TopCap;
	LDInfo->BottomCap = LSInfo->BottomCap;
	LDInfo->Pie = LSInfo->Pie;
	LDInfo->OriginAtBottom = LSInfo->OriginAtBottom;
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	E3dObjectComponentClass	LCylinderClass =
	{
		.Name=        "Cylinder",
		.StructSize=  sizeof(E3dCylinderInfo),
		.Resources=   _Resources,

		.Init=        _CylinderInit,
		.Copy=        _Copy,
		.Update=      _CylinderUpdate,
		.GUI=         _GUI,
	};

	E3dObjectComponentClass	LCapsuleClass =
	{
		.Name=        "Capsule",
		.StructSize=  sizeof(E3dCylinderInfo),
		.Resources=   _Resources,

		.Init=        _CapsuleInit,
		.Copy=        _Copy,
		.Update=      _CapsuleUpdate,
		.GUI=         _GUI,
	};

	if((_CylinderComponentClass = E3dObjectComponent_ClassRegister(&LCylinderClass)) != NULL)
	{
		EpM_CreateIconStates3_sp(Cylinder);E_Assign(_Image_Cylinder, _CylinderComponentClass->Icon);
		Xe_PushButtonAddImg3(_ToolBarButtons[_ToolBarButtons_Count++], Cylinder, "Tool->Create", "Cylinder", '\0', NULL, NULL, FALSE, "Create cylinder\nRight button:  dialog box for this tool", _MCB_Cylinder, (EPointer)0);
		_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Create", "Cylinder", '\0', NULL, NULL, TRUE, NULL, _MCB_Cylinder, (EPointer)0);
	}

	if((_CapsuleComponentClass = E3dObjectComponent_ClassRegister(&LCapsuleClass)) != NULL)
	{
		EpM_CreateIconStates3_sp(Capsule);E_Assign(_Image_Capsule, _CapsuleComponentClass->Icon);
		Xe_PushButtonAddImg3(_ToolBarButtons[_ToolBarButtons_Count++], Capsule, "Tool->Create", "Capsule", '\0', NULL, NULL, FALSE, "Create capsule\nRight button:  dialog box for this tool", _MCB_Capsule, (EPointer)0);
		_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Create", "Capsule", '\0', NULL, NULL, TRUE, NULL, _MCB_Capsule, (EPointer)0);
	}

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC;

		for(LC = 0;LC < _MenuButtons_Count;LC++) { XeItemFree(_MenuButtons[LC]); } _MenuButtons_Count = 0;
		for(LC = 0;LC < _ToolBarButtons_Count;LC++) { XeItemFree(_ToolBarButtons[LC]); } _ToolBarButtons_Count = 0;
	}

	if(_CylinderComponentClass) { E3dObjectComponent_ClassDeactivate(_CylinderComponentClass);_CylinderComponentClass = NULL; }
	if(_CapsuleComponentClass) { E3dObjectComponent_ClassDeactivate(_CapsuleComponentClass);_CapsuleComponentClass = NULL; }

	EpIconFree3(Cylinder);
	EpIconFree3(Capsule);

	return(E_SUCCESS);
}
