//================================================================
// EQUINOX-3D plugin
//
// Create a grid and change it interactively
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 12:02:33
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <math.h>
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/Separator.h>
#include <Xe/ToggleButton.h>

#include <E3D/UI/CreateMacros.h>
#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>


static EPlugin*  _Plugin = NULL;

enum
{
	E3dPO_XY = 0,
	E3dPO_XZ
};


typedef struct
{
	Widget  HorizontalBW, CenterAtMinBW, SmoothBW,
	        TrianglesBW, BackSideBW, SidesBW,
	        UEdgeLengthScaleW, VEdgeLengthScaleW, ThicknessScaleW,
	        USubdivisionsScaleW, VSubdivisionsScaleW,
	        UCurvatureScaleW, VCurvatureScaleW;
} E3dResourcesGUI;


static EguiItem  _MenuButton = NULL, _ToolBarButton = NULL;

EpIconDef3(Grid)


typedef struct
{
	E3dObjectComponentCore;

// E3dGridInfo part
//
	E3dCoordinate  UCurvature, VCurvature;
	E3dCoordinate  UEdgeLength, VEdgeLength, Thickness;
	int            USubdivisions, VSubdivisions, Orientation;
	char           UCenterAt, VCenterAt;
	EBool          Smooth, BackSide, Sides, Triangles;
} E3dGridInfo;


#define ROffset(field)  (void*)(EPtrOffset(E3dGridInfo, field))
static EResource  _Resources[]=
{
	{ "UCurvature",     E3dCOORDINATE_TYPE,  ROffset(UCurvature) },
	{ "VCurvature",     E3dCOORDINATE_TYPE,  ROffset(VCurvature) },
	{ "UEdgeLength",    E3dCOORDINATE_TYPE,  ROffset(UEdgeLength) },
	{ "VEdgeLength",    E3dCOORDINATE_TYPE,  ROffset(VEdgeLength) },
	{ "Thickness",      E3dCOORDINATE_TYPE,  ROffset(Thickness) },

	{ "USubdivisions",  EResINT32,           ROffset(USubdivisions) },
	{ "VSubdivisions",  EResINT32,           ROffset(VSubdivisions) },
	{ "Orientation",    EResINT32,           ROffset(Orientation) },

	{ "UCenterAt",      EResINT8,            ROffset(UCenterAt) },
	{ "VCenterAt",      EResINT8,            ROffset(VCenterAt) },
	{ "Smooth",         EResBOOLEAN,         ROffset(Smooth) },
	{ "BackSide",       EResBOOLEAN,         ROffset(BackSide) },
	{ "Sides",          EResBOOLEAN,         ROffset(Sides) },
	{ "Triangles",      EResBOOLEAN,         ROffset(Triangles) },
	{ }
};
#undef ROffset

static E3dObjectComponentClass*  _ComponentClass = NULL;

//========================================
// Set Vertex coordinates
//========================================
static void _SetVertices(E3dVertex* PVertices, const E3dVertexSize PVertexSize,  E3dGridInfo* PInfo)
{
	const int  LUSubdiv = PInfo->USubdivisions;
	const int  LVSubdiv = PInfo->VSubdivisions;

	const E3dCoordinate  LUEdgeLength = PInfo->UEdgeLength;
	const E3dCoordinate  LVEdgeLength = PInfo->VEdgeLength;

	const E3dCoordinate  LULenH = LUEdgeLength * 0.5;
	const E3dCoordinate  LVLenH = LVEdgeLength * 0.5;
	const E3dCoordinate  LThicknessH = PInfo->Thickness * 0.5;

	const E3dCoordinate  LUStep = LUEdgeLength / (E3dCoordinate)LUSubdiv;
	const E3dCoordinate  LVStep = LVEdgeLength / (E3dCoordinate)LVSubdiv;

	const E3dCoordinate  LUCurvature = PInfo->UCurvature;
	const E3dCoordinate  LVCurvature = PInfo->VCurvature;


	EIndex         LUCnt, LVCnt;
	E3dCoordinate  LX, LV;

	E3dVertex*  LVertex = PVertices;
	switch(PInfo->Orientation)
	{
		case E3dPO_XY:
			{
				E3dCoordinate  LVStart;

				switch(PInfo->VCenterAt)
				{
					case E3dCENTER:  LVStart = LVLenH;break;
					case E3dMIN:     LVStart = LVEdgeLength;break;
					default:         LVStart = 0;break;
				}

// Front side
//
				for(LVCnt = 0, LV = LVStart;LVCnt <= LVSubdiv;LVCnt++, LV -= LVStep)
				{
					for(LUCnt = 0, LX = -LULenH;LUCnt <= LUSubdiv;LUCnt++, LX += LUStep)
					{
						E3dVertex_Init(LVertex, PVertexSize,  LX, LV, LThicknessH + LUCurvature * LUEdgeLength * E3d_Cos(LX / LUEdgeLength * E3dPI) + LVCurvature * LVEdgeLength * E3d_Cos(LV / LVEdgeLength * E3dPI));
						EStruct_Inc(LVertex, PVertexSize);
					}
				}


				if(PInfo->BackSide)
				{
// Back side
//
					for(LVCnt = 0, LV = LVStart;LVCnt <= LVSubdiv;LVCnt++, LV -= LVStep)
					{
						for(LUCnt = 0, LX = -LULenH;LUCnt <= LUSubdiv;LUCnt++, LX += LUStep)
						{
							E3dVertex_Init(LVertex, PVertexSize,  LX, LV, -LThicknessH + LUCurvature * LUEdgeLength * E3d_Cos(LX / LUEdgeLength * E3dPI) + LVCurvature * LVEdgeLength * E3d_Cos(LV / LVEdgeLength * E3dPI));
							EStruct_Inc(LVertex, PVertexSize);
						}
					}
				}
			}
		break;

		case E3dPO_XZ:
			{
				E3dCoordinate  LVStart;

				switch(PInfo->VCenterAt)
				{
					case E3dCENTER:  LVStart = -LVLenH;break;
					case E3dMIN:     LVStart = 0.0;break;
					default:         LVStart = LVEdgeLength;break;
				}

// Front side
//
				for(LVCnt = 0, LV = LVStart;LVCnt <= LVSubdiv;LVCnt++, LV += LVStep)
				{
					for(LUCnt = 0, LX = -LULenH;LUCnt <= LUSubdiv;LUCnt++, LX += LUStep)
					{
						E3dVertex_Init(LVertex, PVertexSize,  LX, -(-LThicknessH + LUCurvature * LUEdgeLength * E3d_Cos(LX / LUEdgeLength * E3dPI)+LVCurvature*LVEdgeLength*E3d_Cos(LV/LVEdgeLength*3.1415)), LV);
						EStruct_Inc(LVertex, PVertexSize);
					}
				}


				if(PInfo->BackSide)
				{
// Back side
//
					for(LVCnt = 0, LV = LVStart;LVCnt <= LVSubdiv;LVCnt++, LV += LVStep)
					{
						for(LUCnt = 0, LX = -LULenH;LUCnt <= LUSubdiv;LUCnt++, LX += LUStep)
						{
							E3dVertex_Init(LVertex, PVertexSize, LX, -(LThicknessH + LUCurvature * LUEdgeLength * E3d_Cos(LX / LUEdgeLength * E3dPI) + LVCurvature * LVEdgeLength * E3d_Cos(LV / LVEdgeLength * E3dPI)), LV);
							EStruct_Inc(LVertex, PVertexSize);
						}
					}
				}
			}
		break;
	}
}


//========================================
// Update() (create/change) method
//========================================
static E3dObject* _Update(E3dObject* PObject, E3dObjectComponent* PInfo, const EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	if(PCheckOnly) return(NULL);

	E3dScene*      LScene = E3d_Scene;
	E3dGridInfo*   LInfo;
	E3dMesh*       LMesh;
	E3dPolyGroup*  LPolyGroup;

	if(PObject == NULL)
	{
		E_New(E3dMesh, LMesh);

		LInfo = (E3dGridInfo*)E3dObjectComponent_Add((E3dObject*)LMesh, _ComponentClass);
		if(PInfo) memcpy(LInfo, PInfo, _ComponentClass->StructSize);
		LPolyGroup = E3dMesh_PolyGroupAdd(LMesh);if(LPolyGroup) E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);
		else
		{
			E_Delete(LMesh);return(NULL);
		}

		PObject = (E3dObject*)LMesh;
	}
	else
	{
		if(!EObject_IsSubClass(PObject, E3dMesh)) return(NULL);

		LMesh = (E3dMesh*)PObject;
		LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass((E3dObject*)LMesh, _ComponentClass);
		LPolyGroup = LMesh->PolyGroups[0];
	}


	if(LMesh)
	{
//		const E3dCoordinate  LUEdgeLength = LInfo->UEdgeLength, LVEdgeLength = LInfo->VEdgeLength;
		int  LUSubdiv = LInfo->USubdivisions;
		int  LVSubdiv = LInfo->VSubdivisions;

		EIndex  LVNum = (LUSubdiv + 1) * (LVSubdiv + 1);
		if(LInfo->BackSide) LVNum *= 2;

		EIndex  LPNum = LUSubdiv * LVSubdiv;

		E3dMesh_FreeCaches(LMesh);
		E3dMesh_EdgesFree(LMesh);

// Allocate vertices for the Mesh
//
		const E3dVertexSize  LVertexSize = LMesh->VertexSize;
		if(LVNum != LMesh->Vertices_Count)
		{
			E3dMesh_VerticesFree(LMesh);
			E3dMesh_VerticesAdd(LMesh, LVNum);
		}

// Allocate Polygons for the Mesh
//
		if(LInfo->BackSide) LPNum *= 2;
		if(LInfo->Sides) LPNum += LUSubdiv * 2 + LVSubdiv * 2;
		if(LInfo->Triangles) LPNum *= 2;

		E3dPolygon*  LPolygons;
		if(LPNum == LPolyGroup->Polygons_Count) LPolygons = LPolyGroup->Polygons;
		else
		{
			if(LPolyGroup->Polygons) E3dPolyGroup_FreePolygons(LPolyGroup);
			if((LPolygons = E3dPolygons_New(LPNum)) != NULL)
			{
				LPolyGroup->Polygons = LPolygons;LPolyGroup->Polygons_Count = LPNum;
			}
		}

// Set up the Mesh vertices
//
		_SetVertices(LMesh->Vertices, LVertexSize,  LInfo);


// Build the Polygons
//
// Front side
//
		const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;
//printf("%s() %s:%d LVertexNodeSize %d\n", __FUNCTION__, __FILE__, __LINE__, LVertexNodeSize);fflush(stdout);
		E3dPolygon*  LPolygon = E3d_GridBuildPolygons(LPolyGroup, LPolyGroup->Polygons, LUSubdiv, LVSubdiv, 0, NULL, NULL, LInfo->Triangles, FALSE);

//		E3d_GridBuildPolygons(LPolyGroup->Polygons, LVertexNodeSize, LUSubdiv, LVSubdiv, (LUSubdiv + 1) * (LVSubdiv + 1), LInfo->Triangles, FALSE);


// Rear side
//
		if(LInfo->BackSide)
		{
			EIndex  LRowA = (LUSubdiv + 1) * (LVSubdiv + 1), LNextRowA = LRowA + LUSubdiv + 1;

			if(LInfo->Triangles)
			{
				EIndex  LVCnt = 0;for(;LVCnt < LVSubdiv;LVCnt++, LRowA += LUSubdiv + 1, LNextRowA += LUSubdiv + 1)
				{
					EIndex  LUCnt = 0;for(;LUCnt < LUSubdiv;LUCnt++)
					{
						E3dPolygon_SetAsTriangle(LPolygon++, LVertexNodeSize, LNextRowA + LUCnt, LRowA + LUCnt, LRowA + LUCnt + 1);
						E3dPolygon_SetAsTriangle(LPolygon++, LVertexNodeSize, LRowA + LUCnt + 1, LNextRowA + LUCnt + 1, LNextRowA + LUCnt);
					}
				}
			}
			else
			{
				EIndex  LVCnt = 0;for(;LVCnt < LVSubdiv;LVCnt++, LRowA += LUSubdiv + 1, LNextRowA += LUSubdiv + 1)
				{
					EIndex  LUCnt = 0;for(;LUCnt < LUSubdiv;LUCnt++)
					{
						E3dPolygon_SetVa(LPolygon++, LVertexNodeSize, 4, LRowA + LUCnt + 1, LNextRowA + LUCnt + 1, LNextRowA + LUCnt, LRowA + LUCnt);
					}
				}
			}

		}


		if(LInfo->Smooth)
		{
			LPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;
			LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;
		}
		else
		{
			LPolyGroup->VertexNormalType = E3dNONE;
			LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;
		}

		EChangeFlags	LChanged = E3dObject_UpdateForDisplay(PObject, PUpdateFlags, LScene);
		if(PTriggeredChangesP) *PTriggeredChangesP |= LChanged;
	}

	return(PObject);
}


//========================================
// Dialog-box callback handlers
//========================================
static void _DCB_Horizontal(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dMesh*         LMesh = (E3dMesh*)LObject;
	E3dPolyGroup*    LPolyGroup = LMesh->PolyGroups[0];
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);


	if(((XeToggleButtonCallbackStruct*)PCallData)->Set) LInfo->Orientation = E3dPO_XZ;
	else LInfo->Orientation = E3dPO_XY;

	_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	E3dMesh_RecomputePolygonNormals(LMesh);
	E3dMeshPolyGroup_RefreshNormals(LMesh, LPolyGroup, LPolyGroup->VertexNormalType, LPolyGroup->SmoothingAngle);
	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);

	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_Smooth(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dMesh*         LMesh = (E3dMesh*)LObject;
	E3dPolyGroup*    LPolyGroup = LMesh->PolyGroups[0];
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);


	LInfo->Smooth = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
	if(LInfo->Smooth)
	{
		LPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;
		E3dMesh_RecomputePolygonNormals(LMesh);
		E3dMeshPolyGroup_RefreshNormals(LMesh, LPolyGroup, LPolyGroup->VertexNormalType, LPolyGroup->SmoothingAngle);
		LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;
	}
	else
	{
		LPolyGroup->VertexNormalType = E3dNONE;
		LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;
	}

	E3dWindows_RedrawByChangeFlags(E3dChgVERTEX_ATTRIBUTE);
}


static void _DCB_Triangles(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->Triangles = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_Back(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->BackSide = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_CenterAtMin(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);


	if(((XeToggleButtonCallbackStruct*)PCallData)->Set)
	{
		LInfo->UCenterAt = E3dCENTER;
		LInfo->VCenterAt = E3dMIN;
	}
	else
	{
		LInfo->UCenterAt = E3dCENTER;
		LInfo->VCenterAt = E3dCENTER;
	}

	E3dObject_CustomUpdate(LObject, LInfo, E3duf_SHAPE|E3duf_CALLS, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_UEdgeLength(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dMesh*         LMesh = (E3dMesh*)LObject;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->UEdgeLength = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_VEdgeLength(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dMesh*         LMesh = (E3dMesh*)LObject;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->VEdgeLength = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_Thickness(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dMesh*         LMesh = (E3dMesh*)LObject;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->Thickness = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_USubdivisions(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->USubdivisions = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_VSubdivisions(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->VSubdivisions = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_UCurvature(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dMesh*         LMesh = (E3dMesh*)LObject;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->UCurvature = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
	_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_VCurvature(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dMesh*         LMesh = (E3dMesh*)LObject;
	E3dGridInfo*     LInfo = (E3dGridInfo*)E3dObjectComponent_ByClass(LPanel->Object, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->VCurvature = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
	_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo);
	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE|E3duf_CALLS, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Pop up the dialog box
//========================================
static void _GUI(E3dObject* PObject, E3dObjectComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dResourcesGUI*         LResourcesGUI;
	E3dGridInfo*             LInfo = (E3dGridInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LInfo)
			{
				if(LDrawer->ResourcesGUI)
				{
					LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

					XeToggleButton_SetState(LResourcesGUI->HorizontalBW, (EBool)(LInfo->Orientation == E3dPO_XZ), FALSE);
					XeToggleButton_SetState(LResourcesGUI->SmoothBW, LInfo->Smooth, FALSE);
					XeToggleButton_SetState(LResourcesGUI->TrianglesBW, LInfo->Triangles, FALSE);
					XeToggleButton_SetState(LResourcesGUI->BackSideBW, LInfo->BackSide, FALSE);
					XeToggleButton_SetState(LResourcesGUI->CenterAtMinBW, LInfo->UCenterAt == E3dCENTER && LInfo->VCenterAt == E3dMIN, FALSE);

					XeScale_SetValue_Coord(LResourcesGUI->UEdgeLengthScaleW, LInfo->UEdgeLength, FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->VEdgeLengthScaleW, LInfo->VEdgeLength, FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->ThicknessScaleW, LInfo->Thickness, FALSE);
					XeScale_SetValue(LResourcesGUI->USubdivisionsScaleW, LInfo->USubdivisions, FALSE);
					XeScale_SetValue(LResourcesGUI->VSubdivisionsScaleW, LInfo->VSubdivisions, FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->UCurvatureScaleW, LInfo->UCurvature, FALSE);
					XeScale_SetValue_Coord(LResourcesGUI->VCurvatureScaleW, LInfo->VCurvature, FALSE);
				}
				else
				{
					Widget	LTopMatrixW = LDrawer->Drawer.DrawerW, LMatrixW, LW;
					Widget	LIndentWs[7];int  LTraverseN;
					Widget	LTraverseWs[8];int  LIndentN;
					EArg    Args[256];int  ArgCnt;

					ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

					{
						EXeStart;
						EXeSetArg(XeNorientation, XeHORIZONTAL);
						EXeSetArg(XeNspacingX, 8);
						EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
						EXeSetArg(XeNheightAdjustable, FALSE);
						LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->Orientation == E3dPO_XZ);
						LResourcesGUI->HorizontalBW = LW = XeWidget_ManagedNew("Horizontal", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Horizontal, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->Smooth);
						LResourcesGUI->SmoothBW = LW = XeWidget_ManagedNew("Smooth", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Smooth, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNfontList, Xe_LabelFontList);
						LResourcesGUI->TrianglesBW = LW = XeWidget_ManagedNew("Triangles", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Triangles, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->BackSide);
						LResourcesGUI->BackSideBW = LW = XeWidget_ManagedNew("Back", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Back, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->UCenterAt == E3dCENTER && LInfo->VCenterAt == E3dMIN);
						LResourcesGUI->CenterAtMinBW = LW = XeWidget_ManagedNew("Center at bottom", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CenterAtMin, (EPointer)LPanel);
					}

/*
					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->Sides);
					LResourcesGUI->SidesBW = LW = XeWidget_ManagedNew("Sides", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Sides, (EPointer)LPanel);
*/

					EXeStart;
					EXeSetArg(XeNshadowThickness, 2);
					EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeWidget_ManagedNew("", XeSeparatorWidget_class, LTopMatrixW,  Args, ArgCnt);


					LTraverseN = 0;LIndentN = 0;
					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 200);
					EXeSetArg(XeNtextFieldDigits, 7);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->UEdgeLengthScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "U Edge length",  Args, ArgCnt);
					XeScale_SetValueAndRanges_Coord(LW,  LInfo->UEdgeLength,  0.0001, 1000000000.0,  0.0001, 1000.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_UEdgeLength, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 200);
					EXeSetArg(XeNtextFieldDigits, 7);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->VEdgeLengthScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "V Edge length",  Args, ArgCnt);
					XeScale_SetValueAndRanges_Coord(LW,  LInfo->VEdgeLength,  0.0001, 1000000000.0,  0.0001, 1000.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_VEdgeLength, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 200);
					EXeSetArg(XeNtextFieldDigits, 7);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->ThicknessScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Thickness",  Args, ArgCnt);
					XeScale_SetValueAndRanges_Coord(LW,  LInfo->Thickness,  0.0, 1000000000.0,  0.0, 1000.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Thickness, (EPointer)LPanel);

					Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 0);


					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 200);
					EXeSetArg(XeNtextFieldDigits, 5);
					EXeSetArg(XeNminimum, 1);EXeSetArg(XeNmaximum, 10000);EXeSetArg(XeNvalue, LInfo->USubdivisions);
					EXeSetArg(XeNscaleMinimum, 1);EXeSetArg(XeNscaleMaximum, 100);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->USubdivisionsScaleW = LW = XeWidget_ManagedNew("U Subdivisions", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_USubdivisions, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 200);
					EXeSetArg(XeNtextFieldDigits, 5);
					EXeSetArg(XeNminimum, 1);EXeSetArg(XeNmaximum, 10000);EXeSetArg(XeNvalue, LInfo->VSubdivisions);
					EXeSetArg(XeNscaleMinimum, 1);EXeSetArg(XeNscaleMaximum, 100);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->VSubdivisionsScaleW = LW = XeWidget_ManagedNew("V Subdivisions", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_VSubdivisions, (EPointer)LPanel);

					Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


					EXeStart;
					EXeSetArg(XeNshadowThickness, 2);
					EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeWidget_ManagedNew("", XeSeparatorWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 200);
					EXeSetArg(XeNtextFieldDigits, 7);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->UCurvatureScaleW = LW = XeWidget_ManagedNew("U Curvature", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->UCurvature,  -1.0, 1.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_UCurvature, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNscaleLengthInPixels, 200);
					EXeSetArg(XeNtextFieldDigits, 7);EXeSetArg(XeNdecimalPoints, 4);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->VCurvatureScaleW = LW = XeWidget_ManagedNew("V Curvature", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
					XeScale_SetValueAndRange_Coord(LW,  LInfo->VCurvature,  -1.0, 1.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_VCurvature, (EPointer)LPanel);

					Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


// Set up TextField traversal
//
					Xe_TraversalRing(LTraverseWs, LTraverseN);
				}
			}
		break;
	}
}


//================================================
// Menu callback
//================================================
static void _MCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dTpl_ObjectComponentActivateCB_Geometry(E3dGrid, _, _Plugin);
}


//================================================
//================================================
static int _Init(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dGridInfo*  LThis = (E3dGridInfo*)PInfo;

	LThis->USubdivisions = 1;LThis->VSubdivisions = 1;
	LThis->Orientation = E3dPO_XZ;
	LThis->UCurvature = 0.0;LThis->VCurvature = 0.0;
	LThis->UEdgeLength = 12.0;
	LThis->VEdgeLength = 12.0;
	LThis->Thickness = 0.0;

	LThis->Smooth = TRUE;
	LThis->Triangles = FALSE;
	LThis->BackSide = FALSE;
	LThis->Sides = FALSE;

	LThis->UCenterAt = E3dCENTER;
	LThis->VCenterAt = E3dCENTER;

	return(E_SUCCESS);
}


#define _M_Copy(mField) LDstInfo->mField = LSrcInfo->mField

//========================================
// Copy-constructor
//========================================
static void _Copy(const E3dObject* PSObject, const E3dObjectComponent* PSrcInfo,  E3dObject* PDObject, E3dObjectComponent* PDstInfo)
{
	const E3dGridInfo*  LSrcInfo = (E3dGridInfo*)PSrcInfo;
	E3dGridInfo*        LDstInfo = (E3dGridInfo*)PDstInfo;

	_M_Copy(USubdivisions);
	_M_Copy(VSubdivisions);
	_M_Copy(Orientation);
	_M_Copy(UCurvature);
	_M_Copy(VCurvature);
	_M_Copy(UEdgeLength);
	_M_Copy(VEdgeLength);
	_M_Copy(Thickness);
	_M_Copy(Smooth);
	_M_Copy(Triangles);
	_M_Copy(BackSide);
	_M_Copy(Sides);

	_M_Copy(UCenterAt);
	_M_Copy(VCenterAt);
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	{
		E3dObjectComponentClass  LObjectComponentClass=
		{
			.Name=        "Grid",
			.StructSize=  sizeof(E3dGridInfo),
			.Resources=   _Resources,
			.Init=        _Init,
			.Copy=        _Copy,
			.Update=      _Update,
			.GUI=         _GUI
		};
		_ComponentClass = E3dObjectComponent_ClassRegister(&LObjectComponentClass);
	}

	if(_ComponentClass)
	{
		EpM_CreateIconStates3_sp(Grid);E_Assign(_Image_Grid, _ComponentClass->Icon);
		Xe_PushButtonAddImg3(_ToolBarButton, Grid, "Tool->Create", "Grid", '\0', NULL, NULL, TRUE, "Create plane or grid.\nRight button:  dialog box for this tool", _MCB, (EPointer)0);
		_MenuButton = XePushButton_Add("Menu->Create", "Grid", '\0', NULL, NULL, TRUE, NULL, _MCB, (EPointer)0);
	}

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		XeItemFree(_MenuButton);
		XeItemFree(_ToolBarButton);
	}

	if(_ComponentClass) E3dObjectComponent_ClassDeactivate(_ComponentClass);

	EpIconFree3(Grid);

	return(E_SUCCESS);
}
