//================================================================
// Add a torus to the scene and change it interactively
//
// Plugin for EQUINOX-3D(TM)
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 19:55:11
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <math.h>
#include <stdio.h>


#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/Separator.h>
#include <Xe/ToggleButton.h>

#include <E3D/E3D.h>

#include <E3D/UI/CreateMacros.h>
#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>



static EPlugin*  _Plugin = NULL;


typedef struct
{
	Widget  SmoothUW, SmoothVW, TrianglesW,
          MainRadiusScaleW, CircleRadiusScaleW, MainSectionsScaleW, CircleSectionsScaleW;
} E3dResourcesGUI;


static EguiItem  _MenuButton = NULL, _ToolBarButton = NULL;

EpIconDef3(Torus)


typedef struct
{
	E3dObjectComponentCore;

// E3dTorusInfo part
//
	E3dCoordinate  MainRadius, CircleRadius;

	unsigned int  MainSections, CircleSections;
	EBool         SmoothU, SmoothV, Triangles;
} E3dTorusInfo;


#define ROffset(field)  (void*)(EPtrOffset(E3dTorusInfo, field))
static EResource  _Resources[]=
{
	{ "MainRadius",     E3dCOORDINATE_TYPE,  ROffset(MainRadius) },
	{ "CircleRadius",   E3dCOORDINATE_TYPE,  ROffset(CircleRadius) },
	{ "MainSections",   EResINT32,           ROffset(MainSections) },
	{ "CircleSections", EResINT32,           ROffset(CircleSections) },

	{ "SmoothU",        EResBOOLEAN,         ROffset(SmoothU) },
	{ "SmoothV",        EResBOOLEAN,         ROffset(SmoothV) },
	{ "Triangles",      EResBOOLEAN,         ROffset(Triangles) },
	{ }
};
#undef ROffset


static E3dObjectComponentClass*  _ComponentClass = NULL;


//========================================
// Rebuild() (create/change) method
//========================================
static E3dObject* _Update(E3dObject* PObject, E3dObjectComponent* PInfo, const EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	if(PCheckOnly) return(NULL);

	E3dScene*      LScene = E3d_Scene;
	E3dMesh*       LMesh;
	E3dPolyGroup*  LPolyGroup = NULL;
	E3dTorusInfo*  LInfo;


	if(PObject == NULL)
	{
		E_New(E3dMesh, LMesh);
		PObject = (E3dObject*)LMesh;

		LInfo = (E3dTorusInfo*)E3dObjectComponent_Add((E3dObject*)LMesh, _ComponentClass);
		if(PInfo && PInfo->Class == _ComponentClass) memcpy(LInfo, PInfo, PInfo->Class->StructSize);
		LPolyGroup = E3dMesh_PolyGroupAdd(LMesh);if(LPolyGroup == NULL)
		{
			E_Delete(LMesh);return(NULL);
		}
		LPolyGroup->VertexNormalType = E3dNormalPROCESS;
	}
	else
	{
		if(!EObject_IsSubClass(PObject, E3dMesh)) return(NULL);

		LMesh = (E3dMesh*)PObject;
		LInfo = (E3dTorusInfo*)E3dObjectComponent_ByClass((E3dObject*)LMesh, _ComponentClass);
		LPolyGroup = LMesh->PolyGroups[0];
	}

	E3dPolygon*          LPolygon;
	EBool                LRefreshTopology = TRUE;
	const EBool          LSmoothU = LInfo->SmoothU, LSmoothV = LInfo->SmoothV;

	const E3dCoordinate  LMainRadius = LInfo->MainRadius;
	const E3dCoordinate  LCircleRadius = LInfo->CircleRadius;
	const EIndex         LMainSections = LInfo->MainSections;
	const EIndex         LCircleSections = LInfo->CircleSections;

//	EIndex  LNumOfVertexNodes;if(LInfo->Triangles) LNumOfVertexNodes = LMainSections * LCircleSections * 2 * 3;else LNumOfVertexNodes = LMainSections * LCircleSections * 4;

	E3dMesh_FreeCaches(LMesh);
	E3dMesh_EdgesFree(LMesh);

	{
		const EIndex  LVN = LCircleSections * LMainSections;
		if(LVN != LMesh->Vertices_Count)
		{
			E3dMesh_VerticesFree(LMesh);
			E3dMesh_VerticesAdd(LMesh, LVN);
		}
	}

	if(LMesh->Vertices)
	{
		E3dVertex*  LVertex = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;

		{
			EIndex  LH = 0;for(;LH < LMainSections;LH++)
			{
				const E3dCoordinate  LTheta = E3dPI * 2.0 * LH / LMainSections;
				const E3dCoordinate  cx = E3d_Cos(LTheta) * LMainRadius;
				const E3dCoordinate  cy = E3d_Sin(LTheta) * LMainRadius;
				EIndex  LC = 0;for(;LC < LCircleSections;LC++)
				{
					const  E3dCoordinate  LPhi = E3dPI * 2.0 * LC / LCircleSections;
					const  E3dCoordinate  dx = E3d_Cos(LTheta) * E3d_Cos(LPhi) * LCircleRadius;
					const  E3dCoordinate  dy = E3d_Sin(LTheta) * E3d_Cos(LPhi) * LCircleRadius;
					const  E3dCoordinate  dz = E3d_Sin(LPhi) * LCircleRadius;

					LVertex->Position.X = cx + dx;LVertex->Position.Y = dz;LVertex->Position.Z = cy + dy;
					EStruct_Inc(LVertex, LVertexSize);
				}
			}
		}

		if(LRefreshTopology)
		{
			EIndex  LN = LMainSections * LCircleSections;
			if(LInfo->Triangles) LN *= 2;
			if(LN == LPolyGroup->Polygons_Count) LPolygon = LPolyGroup->Polygons;
			else
			{
				if(LPolyGroup->Polygons) E3dPolyGroup_FreePolygons(LPolyGroup);
				LPolygon = LPolyGroup->Polygons = E3dPolygons_New(LPolyGroup->Polygons_Count = LN);
			}

			if(LPolygon)
			{
// Make sure the VertexNodes have normals and UVs
//
				E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);
				E3dVertexNodeSize  LUVOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(LPolyGroup, E3dVtxAttribute_UV, 0);
				if(LUVOffsetInVtxNode < 0)
				{
					E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_UV, 0,  NULL);
					LUVOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(LPolyGroup, E3dVtxAttribute_UV, 0);
				}
				const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

				const E3dCoordinate  LUStep = 1.0 / (E3dCoordinate)LMainSections;
				const E3dCoordinate  LVStep = 1.0 / (E3dCoordinate)LCircleSections;

#define left(row, col)	((row) * LCircleSections + (col))
#define right(row, col)	((row) * LCircleSections + ((col) + 1) % LCircleSections)
				if(LInfo->Triangles)
				{
					EIndex  LH = 0;E3dCoordinate  LU = 0.0;for(;LH < LMainSections;LH++, LU += LUStep)
					{
						EIndex  LC = 0;E3dCoordinate  LV = 0.0;for(;LC < LCircleSections;LC++, LV += LVStep)
						{
							E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, left(LH, LC), right(LH, LC), left((LH + 1) % LMainSections, LC));
							E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
							E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);
							LUV->X = LU;LUV->Y = LV;         EStruct_Inc(LUV, LVertexNodeSize);
							LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							LUV->X = LU+LUStep;LUV->Y = LV;
							LPolygon++;

							E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, right(LH, LC), right((LH+1)%LMainSections, LC), left((LH+1)%LMainSections, LC));
							LVertexNode = LPolygon->VertexNodes;
							LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);
							LUV->X = LU;LUV->Y = LV + LVStep;       EStruct_Inc(LUV, LVertexNodeSize);
							LUV->X = LU+LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							LUV->X = LU+LUStep;LUV->Y = LV;
							LPolygon++;
						}
					}
				}
				else
				{
					EIndex  LH = 0;E3dCoordinate  LU = 0.0;for(;LH < LMainSections;LH++, LU += LUStep)
					{
						EIndex  LC = 0;E3dCoordinate  LV = 0.0;for(;LC < LCircleSections;LC++, LV += LVStep)
						{
							E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, left(LH, LC), right(LH, LC), right((LH+1)%LMainSections, LC), left((LH+1)%LMainSections, LC));

							E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
							E3dCoord2*      LUV = (E3dCoord2*)EOffsetToPtr(LVertexNode, LUVOffsetInVtxNode);

							LUV->X = LU;LUV->Y = LV;         EStruct_Inc(LUV, LVertexNodeSize);
							LUV->X = LU;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							LUV->X = LU + LUStep;LUV->Y = LV + LVStep;EStruct_Inc(LUV, LVertexNodeSize);
							LUV->X = LU + LUStep;LUV->Y = LV;

							LPolygon++;
						}
					}
				}

			}
			E3dMesh_CreateEdgesFromPolygons(LMesh, NULL, FALSE, FALSE);
		}

		if(LSmoothU && LSmoothV)
		{
			LPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;
			LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;
		}
		else
		{
			LPolyGroup->VertexNormalType = 0;
			LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;
		}
		LPolyGroup->SmoothingAngle = 48.0;

		EChangeFlags	LChanged = E3dObject_UpdateForDisplay((E3dObject*)LMesh, PUpdateFlags, LScene);
		if(PTriggeredChangesP) *PTriggeredChangesP |= LChanged;
	}

	return(PObject);
}


//========================================
// Dialog box callback handlers
//========================================
static void _DCB_Smooth(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTorusInfo*    LInfo = (E3dTorusInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);


	{
		E3dMesh*       LMesh = (E3dMesh*)LObject;
		E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[0];

		LInfo->SmoothU = ((XeToggleButtonCallbackStruct*)PCallData)->Set;
		if(LInfo->SmoothU)
		{
			LPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;
			LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;
			E3dMesh_RecomputeNormals(LMesh, FALSE);
		}
		else
		{
			LPolyGroup->VertexNormalType = 0;
			LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;
		}

		LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_NORMALS, LScene);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_Triangles(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTorusInfo*    LInfo = (E3dTorusInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->Triangles = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_MainRadius(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTorusInfo*    LInfo = (E3dTorusInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->MainRadius = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_CircleRadius(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTorusInfo*    LInfo = (E3dTorusInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->CircleRadius = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_MainSections(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTorusInfo*    LInfo = (E3dTorusInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->MainSections = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_CircleSections(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTorusInfo*    LInfo = (E3dTorusInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->CircleSections = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// GUI
//========================================
static void _GUI(E3dObject* PObject, E3dObjectComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dResourcesGUI*         LResourcesGUI;
	E3dTorusInfo*            LInfo = (E3dTorusInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

				XeToggleButton_SetState(LResourcesGUI->SmoothUW, LInfo->SmoothU, FALSE);
//				XeToggleButton_SetState(LResourcesGUI->SmoothVW, LInfo->SmoothV, FALSE);
				XeToggleButton_SetState(LResourcesGUI->TrianglesW, LInfo->Triangles, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->MainRadiusScaleW, LInfo->MainRadius, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->CircleRadiusScaleW, LInfo->CircleRadius, FALSE);
				XeScale_SetValue(LResourcesGUI->MainSectionsScaleW, LInfo->MainSections, FALSE);
				XeScale_SetValue(LResourcesGUI->CircleSectionsScaleW, LInfo->CircleSections, FALSE);
			}
			else
			{
				Widget  LTopMatrixW = LDrawer->Drawer.DrawerW, LMatrixW, LW;
				Widget  LIndentWs[7];int  LIndentN;      // For indenting the ScaleWidgets
				Widget  LTraverseWs[7];int  LTraverseN;  // For TextField traversal
				EArg    Args[256];int  ArgCnt;


				ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

				EXeStart;
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNspacingX, 12);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LMatrixW = XeWidget_ManagedNew("LMatrixW", XeMatrixWidget_class, LTopMatrixW, Args, ArgCnt);

				EXeStart;
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
				EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNfontList, Xe_LabelFontList);
				EXeSetArg(XeNset, LInfo->SmoothU);
				LResourcesGUI->SmoothUW = LW = XeWidget_ManagedNew("Smooth", XeToggleButton_class, LMatrixW, Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Smooth, (EPointer)LPanel);

/*
				EXeStart;
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNset, LInfo->SmoothV);
				LResourcesGUI->SmoothVW = LW = XeWidget_ManagedNew("Smooth V", XeToggleButton_class, LMatrixW, Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SmoothV, (EPointer)LPanel);
*/

				EXeStart;
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNset, LInfo->Triangles);
				LResourcesGUI->TrianglesW = LW = XeWidget_ManagedNew("Triangles", XeToggleButton_class, LMatrixW, Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Triangles, (EPointer)LPanel);


				LTraverseN = 0;
				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->MainRadiusScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Main radius", Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->MainRadius,  0.0, 100000.0,  0.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MainRadius, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 8);
				EXeSetArg(XeNminimum, 3);EXeSetArg(XeNmaximum, 100);EXeSetArg(XeNvalue, LInfo->MainSections);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->MainSectionsScaleW = LW = XeWidget_ManagedNew("Main sections", XeScaleWidget_class, LTopMatrixW, Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_MainSections, (EPointer)LPanel);


				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNincrement, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->CircleRadiusScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Circle radius", Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->CircleRadius,  0.0, 100000.0,  0.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CircleRadius, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 8);
				EXeSetArg(XeNminimum, 3);EXeSetArg(XeNmaximum, 100);EXeSetArg(XeNvalue, LInfo->CircleSections);
				EXeSetArg(XeNincrement, 1);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->CircleSectionsScaleW = LW = XeWidget_ManagedNew("Circle sections", XeScaleWidget_class, LTopMatrixW, Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CircleSections, (EPointer)LPanel);

				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


// Set up TextField traversal
//
				Xe_TraversalRing(LTraverseWs, LTraverseN);
			}
		break;
	}
}


//========================================
// Main callback (create new torus)
//========================================
static void _MCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dTpl_ObjectComponentActivateCB_Geometry(E3dTorus, _, _Plugin);
}


//================================================
//================================================
static int _Init(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dTorusInfo*  PThis = (E3dTorusInfo*)PInfo;

	PThis->MainRadius = 1.0;PThis->CircleRadius = 0.35;
	PThis->MainSections = 24;PThis->CircleSections = 12;
	PThis->SmoothU = TRUE;PThis->SmoothV = TRUE;PThis->Triangles = FALSE;

	return(E_SUCCESS);
}


//========================================
// Copy-constructor
//========================================
static void _Copy(const E3dObject* PSrcObject, const E3dObjectComponent* PSrcInfo,  E3dObject* PDstObject, E3dObjectComponent* PDstInfo)
{
	const E3dTorusInfo*  LSrcInfo = (E3dTorusInfo*)PSrcInfo;
	E3dTorusInfo*        LDstInfo = (E3dTorusInfo*)PDstInfo;

#define _Copy(m)  LDstInfo->m = LSrcInfo->m
	_Copy(MainRadius);
	_Copy(CircleRadius);
	_Copy(MainSections);
	_Copy(CircleSections);
	_Copy(SmoothU);
	_Copy(SmoothV);
	_Copy(Triangles);
#undef _Copy
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	E3dObjectComponentClass  LComponentClass=
	{
		.Name=        "Torus",
		.StructSize=  sizeof(E3dTorusInfo),
		.Resources=   _Resources,

		.Init=        _Init,
		.Copy=        _Copy,
		.Update=      _Update,
		.GUI=         _GUI
	};

	if((_ComponentClass = E3dObjectComponent_ClassRegister(&LComponentClass)) != NULL)
	{
		EpM_CreateIconStates3_sp(Torus);E_Assign(_Image_Torus, _ComponentClass->Icon);
		Xe_PushButtonAddImg3(_ToolBarButton, Torus, "Tool->Create", "Torus", '\0', NULL, NULL, TRUE, "Create torus\nRight button:  dialog box for this tool", _MCB, (EPointer)0);

		_MenuButton = XePushButton_Add("Menu->Create", "Torus", '\0', NULL, NULL, TRUE, NULL, _MCB, (EPointer)0);
	}
	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		XeItemFree(_MenuButton);
		XeItemFree(_ToolBarButton);
	}

	if(_ComponentClass)
	{
		E3dObjectComponent_ClassDeactivate(_ComponentClass);_ComponentClass = NULL;
	}

	EpIconFree3(Torus);

	return(E_SUCCESS);
}
