//================================================================
// EQUINOX-3D(TM) plugin
//
// Create a tube and change it interactively
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 00:50:07
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <math.h>
#include <stdio.h>
#include <assert.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/Separator.h>
#include <Xe/ToggleButton.h>

#include <E3D/PolygonEditing.h>

#include <E3D/UI/CreateMacros.h>
#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>


static EPlugin*  _Plugin = NULL;


enum
{
	E3dTubePOLYGONS = 0,
	E3dTubeBEZIER,
	E3dTubeNURBS
};


typedef struct
{
	Widget  SmoothW,
	        CapTopW, CapBottomW, OriginAtBottomW,
	        InnerRadiusScaleW, OuterRadiusScaleW, HeightScaleW,
	        StartAngleScaleW, EndAngleScaleW, SidesScaleW, RadialSectionsScaleW, HeightSectionsScaleW;
} E3dResourcesGUI;

static EguiItem  _MenuButton = NULL, _ToolBarButton = NULL;

EpIconDef3(Tube)


typedef struct
{
	E3dObjectComponentCore;


	E3dSpline*    ProfileSpline;  // "Tube" is built on top of "Revolve"

// E3dTubeInfo part
//
	float         InnerRadius, OuterRadius, Height;
	float         StartAngle, EndAngle;

	unsigned int  Sides, RadialSections, HeightSections;
	unsigned int  Type;

	EBool         Full, Smooth, CapTop, CapBottom, AllowHoledPolys, OriginAtBottom;
} E3dTubeInfo;


#define ROffset(field)  (void*)(EPtrOffset(E3dTubeInfo, field))
static EResource  _Resources[]=
{
	{ "InnerRadius",      E3dCOORDINATE_TYPE,  ROffset(InnerRadius) },
	{ "OuterRadius",      E3dCOORDINATE_TYPE,  ROffset(OuterRadius) },
	{ "Height",           E3dCOORDINATE_TYPE,  ROffset(Height) },
	{ "StartAngle",       E3dCOORDINATE_TYPE,  ROffset(StartAngle) },
	{ "EndAngle",         E3dCOORDINATE_TYPE,  ROffset(EndAngle) },

	{ "Sides",            EResINT32,           ROffset(Sides) },
	{ "RadialSections",   EResINT32,           ROffset(RadialSections) },
	{ "HeightSections",   EResINT32,           ROffset(HeightSections) },

	{ "Type",             EResINT32,           ROffset(Type) },

	{ "Smooth",           EResBOOLEAN,         ROffset(Smooth) },
	{ "CapTop",           EResBOOLEAN,         ROffset(CapTop) },
	{ "CapBottom",        EResBOOLEAN,         ROffset(CapBottom) },
	{ "AllowHoledPolys",  EResBOOLEAN,         ROffset(AllowHoledPolys) },
	{ "OriginAtBottom",   EResBOOLEAN,         ROffset(OriginAtBottom) },
	{ }
};
#undef ROffset


static E3dObjectComponentClass*  _ComponentClass = NULL;


static int  _VNum;

//================================================
// Compute vertex coordinates of the tube
//================================================
static void _SetVertices(E3dVertex* PVertices, const E3dVertexSize PVertexSize,  E3dTubeInfo* PInfo, E3dCoord2* PUnitArcPoints)
{
	E3dCoord2*           LUnitArcPoints = NULL;
	const EBool          LFull = PInfo->Full;
	const E3dCoordinate  LOuterRadius = PInfo->OuterRadius;
	const E3dCoordinate  LInnerRadius = PInfo->InnerRadius;
	const E3dCoordinate  LHeight = PInfo->Height;
	EIndex               LSides = PInfo->Sides;
	const EIndex         LRadialSections = PInfo->RadialSections;
	const EIndex         LHeightSections = PInfo->HeightSections;

	E3dCoordinate        LTopY, LBottomY;
	if(PInfo->OriginAtBottom)
	{
		LTopY = LHeight;LBottomY = 0.0;
	}
	else
	{
		LTopY = LHeight * 0.5;LBottomY = -LHeight * 0.5;
	}

	E3dVertex*  LVertex = PVertices;

	switch(PInfo->Type)
	{
		case E3dTubePOLYGONS:
			if(PUnitArcPoints) LUnitArcPoints = PUnitArcPoints;
			else LUnitArcPoints = E3d_CreateUnitArc(PInfo->Sides, PInfo->StartAngle, PInfo->EndAngle,  NULL);

			if(!LFull) LSides++;

			const E3dCoordinate  LHSize = LHeight / (double)LHeightSections;

			const E3dCoordinate  LRadStep = (LOuterRadius - LInnerRadius) / (E3dCoordinate)LRadialSections;


// Sides
//
			E3dCoordinate  LY = LTopY;
			EIndex  LC = 0;for(;LC <= LHeightSections;LC++)
			{
				EIndex  LC1 = 0;const E3dCoord2*  LPoint = LUnitArcPoints;for(;LC1 < LSides;LC1++)
				{
					LVertex->Position.X = LInnerRadius * LPoint->X;LVertex->Position.Y = LY;LVertex->Position.Z = LInnerRadius * LPoint->Y;
					EStruct_Inc(LVertex, PVertexSize);
					LPoint++;
				}
				LY -= LHSize;
			}


			LY = LTopY;
			for(LC = 0;LC <= LHeightSections;LC++)
			{
				EIndex  LC1 = 0;const E3dCoord2*  LPoint = LUnitArcPoints;for(;LC1 < LSides;LC1++)
				{
					LVertex->Position.X = LOuterRadius * LPoint->X;LVertex->Position.Y = LY;LVertex->Position.Z = LOuterRadius * LPoint->Y;
					EStruct_Inc(LVertex, PVertexSize);
					LPoint++;
				}
				LY -= LHSize;
			}


			if(LRadialSections > 1)
			{
// Radial section dividers in top cap
//
				if(PInfo->CapTop)
				{
					unsigned int	LRN = LRadialSections-1;

					LY = LTopY;
					E3dCoordinate  LRad = LInnerRadius + LRadStep;for(LC = 0;LC < LRN;LC++)
					{
						EIndex  LC1 = 0;const E3dCoord2*  LPoint = LUnitArcPoints;for(;LC1 < LSides;LC1++)
						{
							LVertex->Position.X = LRad * LPoint->X;LVertex->Position.Y = LY;LVertex->Position.Z = LRad * LPoint->Y;
							EStruct_Inc(LVertex, PVertexSize);
							LPoint++;
						}
						LRad += LRadStep;
					}
				}
			}

// Radial section dividers in bottom cap
//
			if(LRadialSections > 1)
			{
				if(PInfo->CapBottom)
				{
					LY = LBottomY;
					E3dCoordinate  LRad = LInnerRadius+LRadStep;
					const EIndex	LRN = LRadialSections - 1;for(LC = 0;LC < LRN;LC++)
					{
						EIndex  LC1 = 0;const E3dCoord2*  LPoint = LUnitArcPoints;for(;LC1 < LSides;LC1++)
						{
							LVertex->Position.X = LRad * LPoint->X;LVertex->Position.Y = LY;LVertex->Position.Z = LRad * LPoint->Y;
							EStruct_Inc(LVertex, PVertexSize);
							LPoint++;
						}
						LRad += LRadStep;
					}
				}
			}
		break;
	}
	if(LUnitArcPoints && (PUnitArcPoints == NULL)) EFree(LUnitArcPoints);
}




static void _CreatePolygonsForFullTube(E3dTubeInfo* PInfo, E3dMesh* PMesh, E3dPolyGroup* PPolyGroup, E3dPolygon* PPolygon, const EIndex PSides, E3dCoord2* PUnitArcPoints)
{
	const EIndex      LRadialSections = PInfo->RadialSections;
	const EIndex      LHeightSections = PInfo->HeightSections;
	E3dPolyGroup_VertexAttributeAdd(PPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);
	const EIndex      LVertexNodeSize = PPolyGroup->VertexNodeSize;
	const int         LNormalOffset = E3dVertexNodeAttribute_Offset(&PPolyGroup->VertexAttributes,  E3dVtxAttribute_Normal, 0);
	E3dPolygon*       LPolygon = PPolygon;


	const E3dVertex*  LVertices = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;

// Inner sides
//
	{
		EIndex  LC = 0, LV = 0;for(;LC < LHeightSections;LC++, LV += PSides)
		{
// Use last vtx on this section
//
			E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV + PSides - 1 + PSides, LV + PSides, LV, LV + PSides - 1);

			{
				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

				const E3dVertex*  LVertex0 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[PSides - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[PSides - 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex1 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[0].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[0].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[0].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[0].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[PSides - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[PSides - 1].Y; }

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);
				LPolygon++;
			}

			EIndex  LH = 1;const EIndex  LN = PSides;for(;LH < LN;LH++)
			{
				E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV + LH - 1 + PSides, LV + LH + PSides, LV + LH, LV + LH - 1);
				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

				const E3dVertex*  LVertex0 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[LH - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[LH - 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex1 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[LH].X;    LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[LH].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[LH].X;    LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[LH].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[LH - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[LH - 1].Y; }

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);
				LPolygon++;
			}

		}
	}

// Outer sides
//
	{
		EIndex  LC = 0, LV = PSides * (LHeightSections + 1);;for(;LC < LHeightSections;LC++, LV += PSides)
		{
// Use last vtx on this section
//
			{
				E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV + PSides - 1, LV, LV + PSides, LV + PSides - 1 + PSides);
				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

				const E3dVertex*  LVertex0 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[PSides - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[PSides - 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex1 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[0].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[0].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[0].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[0].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[PSides - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[PSides - 1].Y; }

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);
				LPolygon++;
			}

			EIndex  LH = 1;const EIndex  LN = PSides;for(;LH < LN;LH++)
			{
				E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV + LH - 1, LV + LH, LV + LH + PSides, LV + LH - 1 + PSides);
				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

				const E3dVertex*  LVertex0 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH - 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex1 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH - 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH - 1].Y; }
				LVertexNode = LPolygon->VertexNodes;

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);
				LPolygon++;
			}
		}
	}

	if(PInfo->CapTop)
	{
		EIndex  LV = 0;
		if(LRadialSections < 2 && PInfo->AllowHoledPolys)
		{
			E3dVertexNode*  LVertexNode = E3d_PolygonVertexNodesNew(LPolygon, LVertexNodeSize, PSides * 2 + 1);if(LVertexNode)
			{
				E3dCoord3*  LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				EIndex  LC = 0, LH = PSides - 1;for(;LC < PSides;LH--, LC++)
				{
					LVertexNode->VertexID = LH;
					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);
				}
				LPolygon->NumOfVertices = LPolygon->NumOfExteriorVertices = PSides;
				LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;
				LPolygon++;
			}
		}
		else
		{
			EIndex  LRC = 0;for(;LRC < LRadialSections;LRC++)
			{
				EIndex  LNextCircleV;
				if(LRC < LRadialSections - 1)
				{
					if(LRC == 0 && LRadialSections > 1) LNextCircleV = PSides * (LHeightSections + 1) * 2;
					else LNextCircleV = LV + PSides;
				}
				else LNextCircleV = PSides * (LHeightSections + 1);

				EIndex  LVC = 0;for(;LVC < PSides;LVC++, LV++)
				{
					if(LVC < PSides - 1) E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV + 1, LVC + LNextCircleV + 1, LVC + LNextCircleV, LV);
					else E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV - PSides + 1, LVC + LNextCircleV - PSides + 1, LVC + LNextCircleV, LV);
					LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0; }
					LPolygon++;
				}
				LV = LNextCircleV;
			}

		}
	}

//goto Done;

	if(PInfo->CapBottom)
	{
		if(LRadialSections < 2 && PInfo->AllowHoledPolys)
		{
			E3dVertexNode*  LVertexNode = E3d_PolygonVertexNodesNew(LPolygon, LVertexNodeSize, PSides * 2 + 1);if(LVertexNode)
			{
				E3dCoord3*  LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				EIndex  LC = 0, LH = PSides - 1;for(;LC < PSides;LH--, LC++)
				{
					LVertexNode->VertexID = LH;
					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);
				}
				LPolygon->NumOfVertices = LPolygon->NumOfExteriorVertices = PSides;
				LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;
				LPolygon++;
			}
		}
		else
		{
			EIndex  LV = PSides * LHeightSections;
			EIndex  LRC = 0;for(;LRC < LRadialSections;LRC++)
			{
				EIndex  LNextCircleV;
				if(LRC < LRadialSections - 1)
				{
					if(LRC == 0 && LRadialSections > 1) LNextCircleV = PSides * ((LHeightSections + 1) * 2 + LRadialSections - 1);
					else LNextCircleV = LV + PSides;
				}
				else LNextCircleV = PSides * (LHeightSections * 2 + 1);

printf("%s() %s:%d  %d %d\n", __FUNCTION__, __FILE__, __LINE__, (int)LV, (int)LNextCircleV);fflush(stdout);

				EIndex  LVC = 0;for(;LVC < PSides;LVC++, LV++)
				{
					if(LVC < PSides - 1) E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV, LVC + LNextCircleV, LVC + LNextCircleV + 1, LV + 1);
					else E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV, LVC + LNextCircleV, LVC + LNextCircleV - PSides + 1, LV - PSides + 1);
					LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = -1.0;LPolygon->Normal.Z = 0.0;

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0; }
					LPolygon++;
				}
				LV = LNextCircleV;
			}

		}
	}

#ifndef ERELEASE_BUILD
Done:
#endif

printf("%s() %s:%d  %d %d\n", __FUNCTION__, __FILE__, __LINE__, (int)PPolyGroup->Polygons_Count, (int)(LPolygon - PPolyGroup->Polygons));fflush(stdout);

	PPolyGroup->Polygons_Count = LPolygon - PPolyGroup->Polygons;
}


static void _CreatePolygonsForPie(E3dTubeInfo* PInfo, E3dMesh* PMesh, E3dPolyGroup* PPolyGroup, E3dPolygon* PPolygon, EIndex PSides, EIndex PNSideVertices, E3dCoord2* PUnitArcPoints)
{
	const EIndex             LRadialSections = PInfo->RadialSections,
	                         LHeightSections = PInfo->HeightSections;
	E3dPolyGroup_VertexAttributeAdd(PPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);
	const E3dVertexNodeSize  LVertexNodeSize = PPolyGroup->VertexNodeSize;
	const int                LNormalOffset = E3dVertexNodeAttribute_Offset(&PPolyGroup->VertexAttributes,  E3dVtxAttribute_Normal, 0);
	E3dPolygon*              LPolygon = PPolygon;

	const E3dVertex*         LVertices = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;

// Inner sides
//
	{
		EIndex  LV = 0;
		EIndex  LC = 0;for(;LC < LHeightSections;LC++, LV += PNSideVertices)
		{
			EIndex  LH = 0;for(;LH < PSides;LH++)
			{
				E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV + LH + PNSideVertices, LV + LH + PNSideVertices + 1, LV + LH + 1, LV + LH);
				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

				const E3dVertex*  LVertex0 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[LH].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[LH].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex1 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[LH + 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[LH + 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[LH + 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[LH + 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = -PUnitArcPoints[LH].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = -PUnitArcPoints[LH].Y; }
				LVertexNode = LPolygon->VertexNodes;

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);
				LPolygon++;
			}
		}
	}

// Outer sides
//
	{
		EIndex  LV = PNSideVertices * (LHeightSections + 1);
		EIndex  LC = 0;for(;LC < LHeightSections;LC++, LV += PNSideVertices)
		{
			EIndex  LH = 0;for(;LH < PSides;LH++)
			{
				E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV + LH, LV + LH + 1, LV + LH + PNSideVertices + 1, LV + LH + PNSideVertices);
				E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

				const E3dVertex*  LVertex0 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex1 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH + 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH + 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				const E3dVertex*  LVertex2 = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH + 1].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH + 1].Y;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
				EStruct_Inc(LVertexNode, LVertexNodeSize);

				if(LVertexNormal) { LVertexNormal->X = PUnitArcPoints[LH].X;LVertexNormal->Y = 0.0;LVertexNormal->Z = PUnitArcPoints[LH].Y; }
				LVertexNode = LPolygon->VertexNodes;

				E3d_TriangleNormal(&LVertex0->Position, &LVertex1->Position, &LVertex2->Position,  &LPolygon->Normal);
				LPolygon++;
			}
		}
	}

	if(PInfo->CapTop)
	{
		EIndex  LV = 0;
		if(LRadialSections < 2 && PInfo->AllowHoledPolys)
		{
			E3dVertexNode*  LVertexNode = E3d_PolygonVertexNodesNew(LPolygon, LVertexNodeSize, PSides * 2 + 1);if(LVertexNode)
			{
				E3dCoord3*  LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				EIndex  LC = 0,  LH = PSides - 1;for(;LC < PSides;LH--, LC++)
				{
					LVertexNode->VertexID = LH;
					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);
				}
				LPolygon->NumOfVertices = LPolygon->NumOfExteriorVertices = PSides;
				LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;
				LPolygon++;
			}
		}
		else
		{
			EIndex  LNextCircleV;

			EIndex  LRC = 0;for(;LRC < LRadialSections;LRC++)
			{
				if(LRC < (LRadialSections - 1))
				{
					if((LRC == 0) && (LRadialSections > 1)) LNextCircleV = PNSideVertices * (LHeightSections + 1) * 2;
					else LNextCircleV = LV + PNSideVertices;
				}
				else LNextCircleV = PNSideVertices*(LHeightSections+1);

				EIndex  LVC = 0;for(;LVC < PSides;LVC++, LV++)
				{
					E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV + 1, LVC + LNextCircleV + 1, LVC + LNextCircleV, LV);
					LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0; }
					LPolygon++;
				}
				LV = LNextCircleV;
			}

		}
	}


	if(PInfo->CapBottom)
	{
		EIndex  LV = PNSideVertices*LHeightSections;
		if(LRadialSections < 2 && PInfo->AllowHoledPolys)
		{
			E3dVertexNode*  LVertexNode = E3d_PolygonVertexNodesNew(LPolygon, PPolyGroup->VertexNodeSize, PSides * 2 + 1);if(LVertexNode)
			{
				E3dCoord3*  LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;
				EIndex  LC = 0, LH = PSides - 1;for(;LC < PSides;LH--, LC++)
				{
					LVertexNode->VertexID = LH;
					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = 1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);
				}
				LPolygon->NumOfVertices = LPolygon->NumOfExteriorVertices = PSides;
				LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = 1.0;LPolygon->Normal.Z = 0.0;
				LPolygon++;
			}
		}
		else
		{
			EIndex  LNextCircleV;

			EIndex  LRC = 0;for(;LRC < LRadialSections;LRC++)
			{
				if(LRC < LRadialSections - 1)
				{
					if(LRC == 0 && LRadialSections > 1) LNextCircleV = PNSideVertices * ((LHeightSections + 1) * 2 + LRadialSections - 1);
					else LNextCircleV = LV + PNSideVertices;
				}
				else LNextCircleV = PNSideVertices * (LHeightSections * 2 + 1);

				EIndex  LVC = 0;for(;LVC < PSides;LVC++, LV++)
				{
					E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, LV, LVC + LNextCircleV, LVC + LNextCircleV + 1, LV + 1);
					LPolygon->Normal.X = 0.0;LPolygon->Normal.Y = -1.0;LPolygon->Normal.Z = 0.0;

					E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					E3dCoord3*      LVertexNormal = LNormalOffset >= 0 ? (E3dCoord3*)EOffsetToPtr(LVertexNode, LNormalOffset) : NULL;

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0;EStruct_Inc(LVertexNormal, LVertexNodeSize); }
					EStruct_Inc(LVertexNode, LVertexNodeSize);

					if(LVertexNormal) { LVertexNormal->X = 0.0;LVertexNormal->Y = -1.0;LVertexNormal->Z = 0.0; }
					LPolygon++;
				}
				LV = LNextCircleV;
			}

		}
	}
}


//========================================
// Rebuild() (create/change) method
//========================================
static E3dObject* _Update(E3dObject* PObject, E3dObjectComponent* PInfo, const EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	E3dScene*      LScene = E3d_Scene;

	E3dMesh*       LMesh;
	E3dCoord2*     LUnitArcPoints = NULL;
	E3dTubeInfo*   LInfo;
	E3dPolyGroup*  LPolyGroup;


	if(PObject == NULL)
	{
		E_New(E3dMesh, LMesh);
		PObject = (E3dObject*)LMesh;

		LInfo = (E3dTubeInfo*)E3dObjectComponent_Add((E3dObject*)LMesh, _ComponentClass);
		if(PInfo && PInfo->Class == _ComponentClass) memcpy(LInfo, PInfo, PInfo->Class->StructSize);
		if((LPolyGroup = E3dMesh_PolyGroupAdd(LMesh)) == NULL)
		{
			E_Delete(LMesh);return(NULL);
		}
		LPolyGroup->VertexNormalType = E3dNormalPROCESS;
	}
	else
	{
		if(!EObject_IsSubClass(PObject, E3dMesh)) return(NULL);

		LMesh = (E3dMesh*)PObject;
		if(LMesh->PolyGroups == NULL) return(NULL);
		LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass((E3dObject*)LMesh, _ComponentClass);
		LPolyGroup = LMesh->PolyGroups[0];
	}

	const E3dCoordinate  LStartAngle = LInfo->StartAngle, LEndAngle = LInfo->EndAngle;

	EBool          LFull;
	if((LEndAngle - LStartAngle) == 360.0) LFull = TRUE;
	else LFull = FALSE;
	LInfo->Full = LFull;

	if(PCheckOnly) return(NULL);

	E3dPolygon*    LPolygon;

//	const E3dCoordinate  LRadius = LInfo->OuterRadius;
//	const E3dCoordinate  LHeight = LInfo->Height;
	const EIndex   LSides = LInfo->Sides;
	const EIndex   LRadialSections = LInfo->RadialSections;
	const EIndex   LHeightSections = LInfo->HeightSections;

	EIndex         LSidesN = 0, LVNum = 0, LPNum = 0;

	switch(LInfo->Type)
	{
		case E3dTubePOLYGONS:
			LUnitArcPoints = E3d_CreateUnitArc(LInfo->Sides, LInfo->StartAngle, LInfo->EndAngle,  NULL);

// Get number of Vertices and Polygons and initialize the unit cross-section
//
			if(LFull)
			{
				LSidesN = LSides;

				LVNum = LSides * (LHeightSections + 1) * 2;
				LPNum = LSides * LHeightSections * 2;

				if((LRadialSections < 2) && (LInfo->AllowHoledPolys))
				{
					if(LInfo->CapTop) LPNum++;
					if(LInfo->CapBottom) LPNum++;
				}
				else
				{
					if(LInfo->CapTop)
					{
						LVNum += ((LRadialSections - 1) * LSides);
						LPNum += LRadialSections*LSides;
					}

					if(LInfo->CapBottom)
					{
						LVNum += ((LRadialSections-1)*LSides);
						LPNum += LRadialSections*LSides;
					}
				}
			}
			else
			{
				LSidesN = LSides + 1;

				LVNum = LSidesN * (LHeightSections + 1) * 2;
				LPNum = LSides * LHeightSections * 2;

				if((LRadialSections < 2) && (LInfo->AllowHoledPolys))
				{
					if(LInfo->CapTop) LPNum++;
					if(LInfo->CapBottom) LPNum++;
				}
				else
				{
					if(LInfo->CapTop)
					{
						LVNum += ((LRadialSections - 1) * LSidesN);
						LPNum += LRadialSections * LSides;
					}

					if(LInfo->CapBottom)
					{
						LVNum += ((LRadialSections - 1) * LSidesN);
						LPNum += LRadialSections * LSides;
					}
				}
			}


			_VNum = LVNum;
//printf("%d %d LVNum %d  LPNum %d\n", LSides, LRadialSections, LVNum, LPNum);fflush(stdout);
		break;
	}

	E3dMesh_FreeCaches(LMesh);

	if(LVNum != LMesh->Vertices_Count)
	{
		E3dMesh_EdgesFree(LMesh);
		E3dMesh_VerticesFree(LMesh);
		E3dMesh_VerticesAdd(LMesh, LVNum);
	}

	if(LMesh->Vertices)
	{
		switch(LInfo->Type)
		{
			case E3dTubePOLYGONS:
				if(LPNum == LPolyGroup->Polygons_Count) LPolygon = LPolyGroup->Polygons;
				else
				{
					if(LPolyGroup->Polygons) E3dPolyGroup_FreePolygons(LPolyGroup);
					LPolygon = LPolyGroup->Polygons = E3dPolygons_New(LPolyGroup->Polygons_Count = LPNum);
				}

				_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo, LUnitArcPoints);

				if(LPolygon)
				{
					if(LFull) _CreatePolygonsForFullTube(LInfo, LMesh, LPolyGroup, LPolygon, LSides, LUnitArcPoints);
					else _CreatePolygonsForPie(LInfo, LMesh, LPolyGroup, LPolygon, LSides, LSidesN, LUnitArcPoints);

					if(LInfo->Smooth) LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;
					else LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;

					LPolyGroup->SmoothingAngle = 60.0;

					EChangeFlags  LChanged = E3dObject_UpdateForDisplay((E3dObject*)LMesh, PUpdateFlags, LScene);
					if(PTriggeredChangesP) *PTriggeredChangesP |= LChanged;
				}
			break;
		}
	}

	EFreeChk(LUnitArcPoints);

	return(PObject);
}


//========================================
// Set GUI for tube type
//========================================
static void _SetGUIToType(E3dTubeInfo* PInfo, E3dResourcesGUI* PResourcesGUI)
{
	Widget  LIndentWs[8];int  LIndentN;      // For indenting the ScaleWidgets
	Widget  LTraverseWs[8];int  LTraverseN;  // For TextField traversal


	switch(PInfo->Type)
	{
		case E3dTubeNURBS:
		case E3dTubePOLYGONS:
			XeScale_SetValueAndRange(PResourcesGUI->SidesScaleW,  PInfo->Sides,  3, 360,  FALSE);

			LIndentN = LTraverseN = 0;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->OuterRadiusScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->InnerRadiusScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->HeightScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->StartAngleScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->EndAngleScaleW;
			Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);

			LIndentN = 0;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->SidesScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->RadialSectionsScaleW;
			LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PResourcesGUI->HeightSectionsScaleW;
			Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);
		break;
	}
}


static void _CustomObjectChanged(E3dObject* PObject, E3dObjectComponent* PInfo, const EChangeFlags PUpdateFlags, E3dObjectPanel* PPanel)
{
	EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass(PObject);

	E3dObject_CustomUpdate(PObject, PInfo, PUpdateFlags, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handlers
//========================================
static void _DCB_Smooth(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dMesh*         LMesh = (E3dMesh*)LObject;
	E3dTubeInfo*     LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);

	LInfo->Smooth = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	if(LInfo->Smooth) LMesh->PolyGroups[0]->Flags |= E3dUSE_VERTEX_NORMALS;
	else LMesh->PolyGroups[0]->Flags &= ~E3dUSE_VERTEX_NORMALS;

	E3dMesh_RecomputeNormals(LMesh, FALSE);

	EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass(LObject);
	LChanged |= E3dChgVERTEX_ATTRIBUTE;

	E3dObject_UpdateForDisplay((E3dObject*)LMesh, E3duf_NORMALS|E3duf_CALLBACKS, E3d_Scene);
	E3dWindows_RedrawByChangeFlags(LChanged);

//	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3duf_NORMALS, LPanel);
}


static void _DCB_OriginAtBottom(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTubeInfo*     LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);

	LInfo->OriginAtBottom = (((XeToggleButtonCallbackStruct*)PCallData)->Set);

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3duf_SHAPE, LPanel);
}


static void _DCB_CapTop(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTubeInfo*     LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);

	LInfo->CapTop = (((XeToggleButtonCallbackStruct*)PCallData)->Set);

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_CapBottom(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTubeInfo*     LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);

	LInfo->CapBottom = (((XeToggleButtonCallbackStruct*)PCallData)->Set);

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_OuterRadius(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*                LScene = E3d_Scene;
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dTubeInfo*             LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags             LChanged = E3dObject_GetChangeFlagsByClass(LObject);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
	E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	LInfo->OuterRadius = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
	if(LInfo->OuterRadius < LInfo->InnerRadius)
	{
		LInfo->InnerRadius = LInfo->OuterRadius - 0.01;
		XeScale_SetValue_Coord(LResourcesGUI->InnerRadiusScaleW, LInfo->InnerRadius, FALSE);
	}

	{ E3dMesh*  LMesh = (E3dMesh*)LObject;_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo, NULL); }
	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_InnerRadius(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*                LScene = E3d_Scene;
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dTubeInfo*             LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags             LChanged = E3dObject_GetChangeFlagsByClass(LObject);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
	E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	LInfo->InnerRadius = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
	if(LInfo->InnerRadius > LInfo->OuterRadius)
	{
		LInfo->OuterRadius = LInfo->InnerRadius + 0.01;
		XeScale_SetValue_Coord(LResourcesGUI->OuterRadiusScaleW, LInfo->OuterRadius, FALSE);
	}

	{ E3dMesh*  LMesh = (E3dMesh*)LObject;_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo, NULL); }
	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_Height(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTubeInfo*     LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->Height = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	{ E3dMesh*  LMesh = (E3dMesh*)LObject;_SetVertices(LMesh->Vertices, LMesh->VertexSize,  LInfo, NULL); }
	LChanged |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_StartAngle(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dTubeInfo*             LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
	E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	LInfo->StartAngle = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
	if(LInfo->StartAngle > LInfo->EndAngle)
	{
		LInfo->EndAngle = LInfo->StartAngle;
		XeScale_SetValue_Coord(LResourcesGUI->EndAngleScaleW, LInfo->EndAngle, FALSE);
	}
	if(LInfo->StartAngle < (LInfo->EndAngle - 360.0))
	{
		LInfo->EndAngle = LInfo->StartAngle + 360.0;
		XeScale_SetValue_Coord(LResourcesGUI->EndAngleScaleW, LInfo->EndAngle, FALSE);
	}

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_EndAngle(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dTubeInfo*             LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
	E3dResourcesGUI*          LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	LInfo->EndAngle = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	if(LInfo->StartAngle > LInfo->EndAngle)
	{
		LInfo->StartAngle = LInfo->EndAngle;
		XeScale_SetValue_Coord(LResourcesGUI->StartAngleScaleW, LInfo->StartAngle, FALSE);
	}
	if(LInfo->StartAngle < (LInfo->EndAngle - 360.0))
	{
		LInfo->StartAngle = LInfo->EndAngle - 360.0;
		XeScale_SetValue_Coord(LResourcesGUI->StartAngleScaleW, LInfo->StartAngle, FALSE);
	}

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_Sides(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTubeInfo*     LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);


	LInfo->Sides = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_RadialSections(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTubeInfo*     LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);


	LInfo->RadialSections = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_HeightSections(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dTubeInfo*     LInfo = (E3dTubeInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);


	LInfo->HeightSections = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


//========================================
// GUI
//========================================
static void _GUI(E3dObject* PObject, E3dObjectComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dResourcesGUI*         LResourcesGUI;
	E3dTubeInfo*             LInfo = (E3dTubeInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

				XeToggleButton_SetState(LResourcesGUI->SmoothW, LInfo->Smooth, FALSE);
				XeToggleButton_SetState(LResourcesGUI->OriginAtBottomW, LInfo->OriginAtBottom, FALSE);

//				XeOptionMenu_SetOptionByNumber(LResourcesGUI->TypesOptionMenuW, LInfo->Type);

				XeToggleButton_SetState(LResourcesGUI->CapTopW, LInfo->CapTop, FALSE);
				XeToggleButton_SetState(LResourcesGUI->CapBottomW, LInfo->CapBottom, FALSE);

				XeScale_SetValue_Coord(LResourcesGUI->OuterRadiusScaleW, LInfo->OuterRadius, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->InnerRadiusScaleW, LInfo->InnerRadius, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->HeightScaleW, LInfo->Height, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->StartAngleScaleW, LInfo->StartAngle, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->EndAngleScaleW, LInfo->EndAngle, FALSE);
				XeScale_SetValue(LResourcesGUI->SidesScaleW, LInfo->Sides, FALSE);
				XeScale_SetValue(LResourcesGUI->RadialSectionsScaleW, LInfo->RadialSections, FALSE);
				XeScale_SetValue(LResourcesGUI->HeightSectionsScaleW, LInfo->HeightSections, FALSE);

				_SetGUIToType(LInfo, LResourcesGUI);
			}
			else
			{
//assert(0);
				Widget  LTopMatrixW = LDrawer->Drawer.DrawerW, LW;
				Widget  LIndentWs[8];int  LIndentN;      // For indenting the ScaleWidgets
				Widget  LTraverseWs[8];int  LTraverseN;  // For TextField traversal
				EArg    Args[256];int  ArgCnt;

// New, more correct cast-shadow blending
//				Xe_SetValuesVa(LDrawer->Folder.LabelW,  XeNimageBlendMode, EpSRC_plus_DSTxALPHA,  NULL);

				ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

				{
					EXeStart;
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNspacingX, 8);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("LMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->Smooth);
					LResourcesGUI->SmoothW = LW = XeWidget_ManagedNew("Smooth", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Smooth, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->OriginAtBottom);
					LResourcesGUI->OriginAtBottomW = LW = XeWidget_ManagedNew("Origin at bottom", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_OriginAtBottom, (EPointer)LPanel);
				}

				{
					EXeStart;
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNspacingX, 8);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("LMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->CapTop);
					LResourcesGUI->CapTopW = LW = XeWidget_ManagedNew("Cap top", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CapTop, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->CapBottom);
					LResourcesGUI->CapBottomW = LW = XeWidget_ManagedNew("Cap bottom", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CapBottom, (EPointer)LPanel);
				}

				EXeStart;
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNshadowThickness, 2);
				EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeWidget_ManagedNew("", XeSeparatorWidget_class, LTopMatrixW,  Args, ArgCnt);

				LTraverseN = 0;
				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->OuterRadiusScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Outer radius",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->OuterRadius,  0.0, 100000.0,  0.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_OuterRadius, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->InnerRadiusScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Inner radius",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->InnerRadius,  0.0, 100000.0,  0.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_InnerRadius, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->HeightScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Height",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Height,  0.0, 100000.0,  0.0, 100.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Height, (EPointer)LPanel);

				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 2);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->StartAngleScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Start angle",  Args, ArgCnt);
				XeScale_SetValueAndRange_Coord(LW,  LInfo->StartAngle,  -360.0, 360.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_StartAngle, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 2);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->EndAngleScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "End angle",  Args, ArgCnt);
				XeScale_SetValueAndRange_Coord(LW,  LInfo->EndAngle,  -360.0, 360.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_EndAngle, (EPointer)LPanel);

				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 4);
				EXeSetArg(XeNminimum, 3);EXeSetArg(XeNmaximum, 500);EXeSetArg(XeNvalue, LInfo->Sides);
				EXeSetArg(XeNscaleMinimum, 3);EXeSetArg(XeNscaleMaximum, 50);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SidesScaleW = LW = XeWidget_ManagedNew("Sides", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Sides, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 4);
				EXeSetArg(XeNminimum, 1);EXeSetArg(XeNmaximum, 100);EXeSetArg(XeNvalue, LInfo->RadialSections);
				EXeSetArg(XeNscaleMinimum, 1);EXeSetArg(XeNscaleMaximum, 10);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->RadialSectionsScaleW = LW = XeWidget_ManagedNew("Radial segments", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_RadialSections, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 4);
				EXeSetArg(XeNminimum, 1);EXeSetArg(XeNmaximum, 100);EXeSetArg(XeNvalue, LInfo->HeightSections);
				EXeSetArg(XeNscaleMinimum, 1);EXeSetArg(XeNscaleMaximum, 20);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->HeightSectionsScaleW = LW = XeWidget_ManagedNew("Height sections", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_HeightSections, (EPointer)LPanel);

				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


// Set up TextField traversal
//
				Xe_TraversalRing(LTraverseWs, LTraverseN);


				_SetGUIToType(LInfo, LResourcesGUI);
			}
		break;
	}
}


//================================================
// Menu callback
//================================================
static void _MCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dTpl_ObjectComponentActivateCB_Geometry(E3dTube, _, _Plugin);
}


//================================================
//================================================
static int _Init(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dTubeInfo*  LThis = (E3dTubeInfo*)PInfo;

	LThis->InnerRadius = 0.7;LThis->OuterRadius = 1.0;LThis->Height = 2.0;
	LThis->Sides          = 18;
	LThis->RadialSections = 1;
	LThis->HeightSections = 1;

	LThis->Type = E3dTubePOLYGONS;

	LThis->StartAngle = 0.0;LThis->EndAngle = 360.0;
	LThis->Smooth = TRUE;
	LThis->CapTop = TRUE;LThis->CapBottom = TRUE;
	LThis->OriginAtBottom = FALSE;
	LThis->AllowHoledPolys = FALSE;

	LThis->ProfileSpline = NULL;

	return(E_SUCCESS);
}

static int _Free(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dTubeInfo*  LThis = (E3dTubeInfo*)PInfo;

	E_DeleteChk(LThis->ProfileSpline);

	return(E_SUCCESS);
}


#define _M_Copy(mField)  LDInfo->mField = LSInfo->mField

//========================================
// Copy-constructor
//========================================
static void _Copy(const E3dObject* PSrcObject, const E3dObjectComponent* PSrcInfo,  E3dObject* PDstObject, E3dObjectComponent* PDstInfo)
{
	const E3dTubeInfo*  LSInfo = (E3dTubeInfo*)PSrcInfo;
	E3dTubeInfo*        LDInfo = (E3dTubeInfo*)PDstInfo;

	LDInfo->ProfileSpline = NULL;

	_M_Copy(InnerRadius);
	_M_Copy(OuterRadius);
	_M_Copy(Height);
	_M_Copy(Sides);
	_M_Copy(RadialSections);
	_M_Copy(HeightSections);
	_M_Copy(Type);
	_M_Copy(StartAngle);
	_M_Copy(EndAngle);
	_M_Copy(Smooth);
	_M_Copy(CapTop);
	_M_Copy(CapBottom);
	_M_Copy(OriginAtBottom);
	_M_Copy(AllowHoledPolys);
}

//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	E3dObjectComponentClass  LComponentClass=
	{
		.Name=        "Tube",
		.StructSize=  sizeof(E3dTubeInfo),
		.Resources=   _Resources,

		.Init=        _Init,
		.Free=        _Free,
		.Copy=        _Copy,
		.Update=      _Update,
		.GUI=         _GUI
	};

	if((_ComponentClass = E3dObjectComponent_ClassRegister(&LComponentClass)) != NULL)
	{
		EpM_CreateIconStates3_sp(Tube);E_Assign(_Image_Tube, _ComponentClass->Icon);
		Xe_PushButtonAddImg3(_ToolBarButton, Tube, "Tool->Create", "Tube", '\0', NULL, NULL, TRUE, "Create tube\nRight button:  dialog box for this tool", _MCB, (EPointer)0);

		_MenuButton = XePushButton_Add("Menu->Create", "Tube", '\0', NULL, NULL, TRUE, NULL, _MCB, (EPointer)0);
	}

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		XeItemFree(_ToolBarButton);
		XeItemFree(_MenuButton);
	}

	if(_ComponentClass)
	{
		E3dObjectComponent_ClassDeactivate(_ComponentClass);_ComponentClass = NULL;
	}

	EpIconFree3(Tube);

	return(E_SUCCESS);
}
