//================================================================
// Create a vector (represented by a linear Spline)
//
// Plugin for EQUINOX-3D
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 04:12:33
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <math.h>
#include <stdio.h>


#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Label.h>
#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/ToggleButton.h>

#include <E3D/E3D.h>

#include <E3D/UI/CreateMacros.h>
#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>



extern EpCImage  Image_VectorBig;

static EPlugin*  _Plugin = NULL;


typedef struct
{
	Widget  ArrowTBW,
	        TypeMatrixW, P1P2BW, P1LengthBW, P1DirLengthBW,
	        VX0ScaleW, VY0ScaleW, VZ0ScaleW, VX1ScaleW, VY1ScaleW, VZ1ScaleW, LengthW;
} E3dResourcesGUI;

static EguiItem		_MenuButton = NULL;


static E3dObjectComponentClass*  _ComponentClass = NULL;

//========================================
// Update() (create/change) method
//========================================
static E3dObject* _Update(E3dObject* PObject, E3dObjectComponent* PInfoV, EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	if(PCheckOnly) return(NULL);

	E3dSpline*      LSpline;
	E3dVectorInfo*  LInfo;

	if(PObject == NULL)
	{
		E_New(E3dSpline, LSpline);
		PObject = (E3dObject*)LSpline;

		LInfo = (E3dVectorInfo*)E3dObjectComponent_Add((E3dObject*)LSpline, _ComponentClass);
		if(PInfoV && PInfoV->Class == _ComponentClass) memcpy(LInfo, PInfoV, PInfoV->Class->StructSize);
	}
	else
	{
		if(!EObject_IsSubClass(PObject, E3dSpline)) return(NULL);

		LSpline = (E3dSpline*)PObject;
		LInfo = (E3dVectorInfo*)E3dObjectComponent_ByClass((E3dObject*)LSpline, _ComponentClass);
	}

	{
		E3dSplineCV*  LCV;
		EIndex        LNV;

		if(LInfo->ArrowOn) LNV = 10;
		else LNV = 2;

		if(LSpline->CVs_Count != LNV)
		{
			E3dSpline_FreeCVs(LSpline);
			LSpline->CVs = E3dSplineCVs_New(E3dSplineLINEAR, LNV);
			LSpline->CVs_Count = LNV;
		}

		LCV = LSpline->CVs;
		LCV->Position = LInfo->P0;LCV++;

		if(LInfo->ArrowOn) E3d_CreateVectorArrow(LSpline->CVs,  LInfo->P1.X, LInfo->P1.Y, LInfo->P1.Z);
		else
		{
			LCV->Position = LInfo->P1;
		}
		E3dSpline_CreateLinearSegments(LSpline);
	}

	return(PObject);
}


//========================================
// Dialog box callback handlers
//========================================
static void _DCB_Arrow(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dVectorInfo*   LInfo = (E3dVectorInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags     LChanged = E3dObject_GetChangeFlagsByClass(LObject);

	LInfo->ArrowOn = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_Values(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dVectorInfo*           LInfo = (E3dVectorInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags             LChanged = E3dObject_GetChangeFlagsByClass(LObject);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
	E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	LInfo->P0.X = Xe_ScaleGetValue_Coord(LResourcesGUI->VX0ScaleW);
	LInfo->P0.Y = Xe_ScaleGetValue_Coord(LResourcesGUI->VY0ScaleW);
	LInfo->P0.Z = Xe_ScaleGetValue_Coord(LResourcesGUI->VZ0ScaleW);

	switch(LInfo->CoordinateMode)
	{
		case E3dVTM_P1_P2:
			LInfo->P1.X = Xe_ScaleGetValue_Coord(LResourcesGUI->VX1ScaleW);
			LInfo->P1.Y = Xe_ScaleGetValue_Coord(LResourcesGUI->VY1ScaleW);
			LInfo->P1.Z = Xe_ScaleGetValue_Coord(LResourcesGUI->VZ1ScaleW);
		break;

		case E3dVTM_P1_DIR:
			LInfo->P1.X = Xe_ScaleGetValue_Coord(LResourcesGUI->VX1ScaleW) + LInfo->P0.X;
			LInfo->P1.Y = Xe_ScaleGetValue_Coord(LResourcesGUI->VY1ScaleW) + LInfo->P0.Y;
			LInfo->P1.Z = Xe_ScaleGetValue_Coord(LResourcesGUI->VZ1ScaleW) + LInfo->P0.Z;
		break;
	}

	const E3dCoord3  LP0ToP1 = { LInfo->P1.X - LInfo->P0.X,  LInfo->P1.Y - LInfo->P0.Y, LInfo->P1.Z - LInfo->P0.Z };
	E3dCoordinate    LLength;E3d_Length3_v(LP0ToP1, LLength);

	XeScale_SetValue_Coord(LResourcesGUI->LengthW, LLength, FALSE);

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_Length(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dVectorInfo*           LInfo = (E3dVectorInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags             LChanged = E3dObject_GetChangeFlagsByClass(LObject);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
	E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);
	E3dCoord3                LP0ToP1 = { LInfo->P1.X - LInfo->P0.X,  LInfo->P1.Y - LInfo->P0.Y, LInfo->P1.Z - LInfo->P0.Z };
	const E3dCoordinate      LLength = Xe_ScaleGetValue_Coord(LResourcesGUI->LengthW);


	E3d_Normalize3_v(LP0ToP1);
	LInfo->P1.X = LInfo->P0.X + LP0ToP1.X * LLength;
	LInfo->P1.Y = LInfo->P0.Y + LP0ToP1.Y * LLength;
	LInfo->P1.Z = LInfo->P0.Z + LP0ToP1.Z * LLength;

	switch(LInfo->CoordinateMode)
	{
		case E3dVTM_P1_P2:
			XeScale_SetValue_Coord(LResourcesGUI->VX1ScaleW, LInfo->P1.X, FALSE);
			XeScale_SetValue_Coord(LResourcesGUI->VY1ScaleW, LInfo->P1.Y, FALSE);
			XeScale_SetValue_Coord(LResourcesGUI->VZ1ScaleW, LInfo->P1.Z, FALSE);
		break;

		case E3dVTM_P1_DIR:
			XeScale_SetValue_Coord(LResourcesGUI->VX1ScaleW, LInfo->P1.X - LInfo->P0.X, FALSE);
			XeScale_SetValue_Coord(LResourcesGUI->VY1ScaleW, LInfo->P1.Y - LInfo->P0.Y, FALSE);
			XeScale_SetValue_Coord(LResourcesGUI->VZ1ScaleW, LInfo->P1.Z - LInfo->P0.Z, FALSE);
		break;
	}

	E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_P1P2(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dVectorInfo*           LInfo = (E3dVectorInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags             LChanged = E3dObject_GetChangeFlagsByClass(LObject);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
	E3dResourcesGUI*          LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	if(((XeToggleButtonCallbackStruct*)PCallData)->Set)
	{
		LInfo->CoordinateMode = E3dVTM_P1_P2;

		LInfo->P1.X = Xe_ScaleGetValue_Coord(LResourcesGUI->VX1ScaleW);
		LInfo->P1.Y = Xe_ScaleGetValue_Coord(LResourcesGUI->VY1ScaleW);
		LInfo->P1.Z = Xe_ScaleGetValue_Coord(LResourcesGUI->VZ1ScaleW);

		E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_P1Dir(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dVectorInfo*           LInfo = (E3dVectorInfo*)E3dObjectComponent_ByClass(LObject, _ComponentClass);
	EChangeFlags             LChanged = E3dObject_GetChangeFlagsByClass(LObject);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ComponentClass);
	E3dResourcesGUI*          LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	if(((XeToggleButtonCallbackStruct*)PCallData)->Set)
	{
		LInfo->CoordinateMode = E3dVTM_P1_DIR;

		LInfo->P1.X = Xe_ScaleGetValue_Coord(LResourcesGUI->VX1ScaleW) + LInfo->P0.X;
		LInfo->P1.Y = Xe_ScaleGetValue_Coord(LResourcesGUI->VY1ScaleW) + LInfo->P0.Y;
		LInfo->P1.Z = Xe_ScaleGetValue_Coord(LResourcesGUI->VZ1ScaleW) + LInfo->P0.Z;

		E3dObject_CustomUpdate(LObject, LInfo, E3dUF_ALL, &LChanged);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _UpdateDialog(E3dVectorInfo* PInfo, E3dResourcesGUI* PResourcesGUI)
{
	XeToggleButton_SetState(PResourcesGUI->ArrowTBW, PInfo->ArrowOn, FALSE);

	XeScale_SetValue_Coord(PResourcesGUI->VX0ScaleW, PInfo->P0.X, FALSE);
	XeScale_SetValue_Coord(PResourcesGUI->VY0ScaleW, PInfo->P0.Y, FALSE);
	XeScale_SetValue_Coord(PResourcesGUI->VZ0ScaleW, PInfo->P0.Z, FALSE);

	switch(PInfo->CoordinateMode)
	{
		case E3dVTM_P1_P2:
			XeToggleButton_SetState(PResourcesGUI->P1P2BW, TRUE, FALSE);
			XeToggleButton_SetState(PResourcesGUI->P1LengthBW, FALSE, FALSE);

			XeScale_SetValue_Coord(PResourcesGUI->VX1ScaleW, PInfo->P1.X, FALSE);
			XeScale_SetValue_Coord(PResourcesGUI->VY1ScaleW, PInfo->P1.Y, FALSE);
			XeScale_SetValue_Coord(PResourcesGUI->VZ1ScaleW, PInfo->P1.Z, FALSE);

			XeScale_SetValue_Coord(PResourcesGUI->LengthW, PInfo->P1.Z, FALSE);
		break;

		case E3dVTM_P1_DIR:
			XeToggleButton_SetState(PResourcesGUI->P1P2BW, FALSE, FALSE);
			XeToggleButton_SetState(PResourcesGUI->P1LengthBW, TRUE, FALSE);

			XeScale_SetValue_Coord(PResourcesGUI->VX1ScaleW, PInfo->P1.X - PInfo->P0.X, FALSE);
			XeScale_SetValue_Coord(PResourcesGUI->VY1ScaleW, PInfo->P1.Y - PInfo->P0.Y, FALSE);
			XeScale_SetValue_Coord(PResourcesGUI->VZ1ScaleW, PInfo->P1.Z - PInfo->P0.Z, FALSE);
		break;
	}

	const E3dCoord3  LP0ToP1 = { PInfo->P1.X - PInfo->P0.X,  PInfo->P1.Y - PInfo->P0.Y, PInfo->P1.Z - PInfo->P0.Z };
	E3dCoordinate    LLength;E3d_Length3_v(LP0ToP1, LLength);

	XeScale_SetValue_Coord(PResourcesGUI->LengthW, LLength, FALSE);
}


//========================================
// GUI
//========================================
static void _GUI(E3dObject* PObject, E3dObjectComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dResourcesGUI*         LResourcesGUI;
	E3dVectorInfo*           LInfo = (E3dVectorInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

				_UpdateDialog(LInfo, LResourcesGUI);
			}
			else
			{
				Widget  LTopMatrixW = LDrawer->Drawer.DrawerW, LMatrixW, LTypeMatrixW, LW;
				Widget  LIndentWs[7];int  LIndentN;      // For indenting the Scale Widgets
				Widget  LTraverseWs[7];int  LTraverseN;  // For TextField traversal
				EArg    Args[256];int  ArgCnt;


				ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

				EXeStart;
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
				EXeSetArg(XeNfontList, Xe_LabelFontList);
				EXeSetArg(XeNset, LInfo->ArrowOn);
				LResourcesGUI->ArrowTBW = LW = XeWidget_ManagedNew("Arrow", XeToggleButton_class, LTopMatrixW, Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Arrow, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNmarginTop, 5);
				EXeSetArg(XeNradioBehavior, TRUE);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNspacingX, 6);
				EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				LTypeMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW, Args, ArgCnt);

				EXeStart;
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
				EXeSetArg(XeNfontList, Xe_LabelFontList);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				XeWidget_ManagedNew("Type:", XeLabelWidget_class, LTypeMatrixW, Args, ArgCnt);

				EXeStart;
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
				EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
				EXeSetArg(XeNfontList, Xe_LabelFontList);
				EXeSetArg(XeNset, (EBool)(LInfo->CoordinateMode == E3dVTM_P1_P2));
				LResourcesGUI->P1P2BW = LW = XeWidget_ManagedNew("P1-P2", XeToggleButton_class, LTypeMatrixW, Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_P1P2, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
				EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
				EXeSetArg(XeNfontList, Xe_LabelFontList);
				EXeSetArg(XeNset, (EBool)(LInfo->CoordinateMode == E3dVTM_P1_DIR));
				LResourcesGUI->P1LengthBW = LW = XeWidget_ManagedNew("P1 & direction", XeToggleButton_class, LTypeMatrixW, Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_P1Dir, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNmarginSizeX, 2);EXeSetArg(XeNmarginSizeY, 2);
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNcolumns, 3);
				EXeSetArg(XeNspacingX, 2);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW, Args, ArgCnt);

				LTraverseN = 0;
				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 5);
				EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
				EXeSetArg(XeNorientation, XeHORIZONTAL);EXeSetArg(XeNcreateScale, FALSE);
				LTraverseWs[LTraverseN++] = LIndentWs[LIndentN++] = LResourcesGUI->VX0ScaleW = LW = XeScaleNew_Coord(LMatrixW, "XYZ:", Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  0.0,  -1000000.0,  1000000,  -10.0, 10,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Values, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 5);
				EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
				EXeSetArg(XeNorientation, XeHORIZONTAL);EXeSetArg(XeNcreateScale, FALSE);
				LTraverseWs[LTraverseN++] = LResourcesGUI->VY0ScaleW = LW = XeScaleNew_Coord(LMatrixW, "", Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  0.0,  -1000000.0,  1000000,  -10.0, 10,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Values, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 5);
				EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
				EXeSetArg(XeNorientation, XeHORIZONTAL);EXeSetArg(XeNcreateScale, FALSE);
				LTraverseWs[LTraverseN++] = LResourcesGUI->VZ0ScaleW = LW = XeScaleNew_Coord(LMatrixW, "", Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  0.0,  -1000000.0,  1000000,  -10.0, 10,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Values, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 5);
				EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
				EXeSetArg(XeNorientation, XeHORIZONTAL);EXeSetArg(XeNcreateScale, FALSE);
				LTraverseWs[LTraverseN++] = LIndentWs[LIndentN++] = LResourcesGUI->VX1ScaleW = LW = XeScaleNew_Coord(LMatrixW, "XYZ:", Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  0.0,  -1000000.0,  1000000,  -10.0, 10,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Values, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 5);
				EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
				EXeSetArg(XeNorientation, XeHORIZONTAL);EXeSetArg(XeNcreateScale, FALSE);
				LTraverseWs[LTraverseN++] = LResourcesGUI->VY1ScaleW = LW = XeScaleNew_Coord(LMatrixW, "", Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  0.0,  -1000000.0,  1000000,  -10.0, 10,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Values, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 5);
				EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
				EXeSetArg(XeNorientation, XeHORIZONTAL);EXeSetArg(XeNcreateScale, FALSE);
				LTraverseWs[LTraverseN++] = LResourcesGUI->VZ1ScaleW = LW = XeScaleNew_Coord(LMatrixW, "", Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  0.0,  -1000000.0,  1000000,  -10.0, 10,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Values, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 9);EXeSetArg(XeNdecimalPoints, 5);
				EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
				EXeSetArg(XeNorientation, XeHORIZONTAL);EXeSetArg(XeNcreateScale, FALSE);
				LTraverseWs[LTraverseN++] = LIndentWs[LIndentN++] = LResourcesGUI->LengthW = LW = XeScaleNew_Coord(LTopMatrixW, "Length", Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  0.0,  -1000000.0,  1000000,  -10.0, 10,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Length, (EPointer)LPanel);

				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 0);

				_UpdateDialog(LInfo, LResourcesGUI);

// Set up TextField traversal
//
				Xe_TraversalRing(LTraverseWs, LTraverseN);
			}
		break;
	}
}


//========================================
// Menu callback
//========================================
static void _MCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dTpl_ObjectComponentActivateCB_Geometry(E3dVector, _, _Plugin);
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	{
		_ComponentClass = E3d_VectorObjectComponentClass;
		_ComponentClass->Update = _Update;
		_ComponentClass->GUI = _GUI;
	}

	_MenuButton = XePushButton_Add("Menu->Create", "Vector", '\0', NULL, NULL, TRUE, NULL, _MCB, (EPointer)0);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit) XeItemFree(_MenuButton);

// Check if ObjectComponentClasses have been already removed
//
	if(E3dObjectComponent_Classes)
	{
		_ComponentClass->Update = NULL;
		_ComponentClass->GUI = NULL;
	}

	return(E_SUCCESS);
}
