//================================================================
// EQUINOX-3D plugin
//
// Cg renderer
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 21:23:46
//
// Copyright (C) 2014-2018 By Gabor Nagy. All rights reserved.
//================================================================
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>

// cgc -profile fp30 ...

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/E3D.h>
#include <E3D/Matrix.h>
#include <E3D/Material.h>
#include <E3D/SL.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage

#include <E3D/GL/Draw.h>

#include <E3D/SkinMesh.h>

#ifndef _E3D_PointSetDraw_h
#include <E3D/GL/PointSetDraw.h>
#endif

#include <E3D/GL/MeshDraw.h>

static EPlugin*  _Plugin = NULL;

#include "../../HW/Shadow.h"
#include "../GPURendererGUI.h"

#include "Cg.h"

#include <Cg/cgGL.h>

char*             E_PluginInfo = "GPU renderer with advanced shaders and full-screen\neffects, using the Cg shading language.";

static const EBool    _PrintShaderSource = FALSE;
//static const uint8_t  _SaveShaderSourceOn = E3dSL_ALWAYS;
static const uint8_t  _SaveShaderSourceOn = E3dSL_COMPILE_ERROR;
//static const uint8_t  _SaveShaderSourceOn = E3dSL_COMPILE_ERROR|E3dSL_SYMBOL_RESOLVE_ERROR;


#define TIMING  0

#if TIMING
#define TIMER_ENABLE          1
#define TIMER_USE_AUTO_SCOPE  0
#include <ETiming.h>

extern ETmTimerStack  E3dp_RendererTimerStack;
#else
#define TIMER_PUSH(m0, m1)
#define TIMER_PUSH_WITH_OPTIONS(mTimerStack, mName, mBracket)  do { } while(0)
#define TIMER_POP(m0)
#endif

static ETime  _BindMaterial_Microseconds  = 0;
static ETime  _HierarchyDraw_Microseconds = 0;

static EBool        _ProfileInitialized = FALSE;
static const int    _FBOSharingMode = E3dFBOSharingFULL;

CGcontext	E3dCg_ShaderContext = 0;

CGprofile	E3dCg_VertexProfile = 0;
CGprofile	E3dCg_FragmentProfile = 0;


static char     _StateCompilingFragmentProgram[] = "Compiling fragment program";
static char     _StateLoadingFragmentProgram[]   = "Loading fragment program";

static E3dShaderClass*		_ConstantShaderClass = NULL;
static E3dShaderClass*		_LambertShaderClass = NULL;
static E3dShaderClass*		_PhongShaderClass = NULL;


#define E3dSL_Uniform1i(mParameter, m0)              cgGLSetParameter1f(mParameter, m0)   // Cg doesn't have a way to set integer uniforms, so we have to use the float version
#define E3dSL_Uniform1f(mParameter, m0)              cgGLSetParameter1f(mParameter, m0)
#define E3dSL_Uniform2f(mParameter, m0, m1)          cgGLSetParameter2f(mParameter, m0, m1)
#define E3dSL_Uniform3f(mParameter, m0, m1, m2)      cgGLSetParameter3f(mParameter, m0, m1, m2)
#define E3dSL_Uniform4f(mParameter, m0, m1, m2, m3)  cgGLSetParameter4f(mParameter, m0, m1, m2, m3)

#define E3dSL_Uniform1fv(mParameter, mPtr)           cgGLSetParameter1fv(mParameter, mPtr)
#define E3dSL_Uniform2fv(mParameter, mPtr)           cgGLSetParameter2fv(mParameter, mPtr)
#define E3dSL_Uniform3fv(mParameter, mPtr)           cgGLSetParameter3fv(mParameter, mPtr)
#define E3dSL_Uniform4fv(mParameter, mPtr)           cgGLSetParameter4fv(mParameter, mPtr)

#define E3dSL_GetUniformfv(mShaderVariant, mParameter, mPtr)

#define E3dSLParameter	        CGparameter
#define E3dSLParameterValid(m)	(m != NULL)
#define E3dSLParameterNONE      NULL
#define E3dThisRenderer	        E3dCgRenderer


static E3dFeature*  _FeatureCg  = NULL;
static E3dFeature*  _FeatureFBO = NULL;
static E3dFeature*  _FeatureHDR = NULL;


static EArg	 EmptyUI[]=
{
	{ NULL, 0 }
};

static EArg	 DrawZOnlyFirstUI[]=
{
	{ XeNtipString8, (EArgVal)("Draw Z-buffer only first. This might produce a better\nframe-rate if there are complex shaders in the scene.") },
	{ NULL, 0 }
};


#define ROffset(field)  (void*)(EPtrOffset(E3dThisRenderer, field))
static EResource  _Resources[]=
{
	{ "Shadows",           EResBOOLEAN,  ROffset(ShadowsOn),       .UI= EmptyUI },
	{ "Reflections",       EResBOOLEAN,  ROffset(ReflectionsOn),   .UI= EmptyUI },
	{ "Transparency",      EResBOOLEAN,  ROffset(TransparencyOn),  .UI= EmptyUI },
	{ "Z-first",           EResBOOLEAN,  ROffset(DrawZOnlyFirst),  .UI= DrawZOnlyFirstUI },

	{ "Bloom radius",       EResFLOAT32,  ROffset(Bloom.Radius),      .UI= E3dBloomUI_Radius },
	{ "Bloom strength",     EResFLOAT32,  ROffset(Bloom.Strength),    .UI= E3dBloomUI_Strength },
	{ "Bloom power",        EResFLOAT32,  ROffset(Bloom.Power),       .UI= E3dBloomUI_Power },
	{ }
};
#undef ROffset


static E3dRendererClass*  _RendererClass = NULL;


#define E3dSL_SetParameterPointer(mParam, mFSize, mType, mStride, mPointer)	cgGLSetParameterPointer(mParam, mFSize, mType, mStride, mPointer)

// For example:
//
//	cgGLSetParameterPointer(LPositionParam, 3, GL_FLOAT, LPolyGroup->GLArrayVertexNodeSize_SL, LPtr);LPtr += sizeof(float) * 3;


static inline void E3dSL_EnableClientState(const E3dSLParameter PParam)
{
	cgGLEnableClientState(PParam);
}


static inline void E3dSL_DisableClientState(const E3dSLParameter PParam)
{
	cgGLDisableClientState(PParam);
}


static void _ErrorCallback(void)
{
	const CGerror  LastError = cgGetError();
	
	if(LastError)
	{
//	const char*  Listing = cgGetLastListing(E3dCg_ShaderContext);
		char*        LStr = NULL;
		char*        LPtr = NULL;
		int          LStrLen = 0;


		if(E3dSL_DebugRec.Function)
		{
			LStr = ERealloc(LStr, LStrLen + 256);LPtr = LStr + LStrLen;

			if(E3dSL_DebugRec.Line) sprintf(LPtr, "%s() %s:%d", E3dSL_DebugRec.Function, E3dSL_DebugRec.File, E3dSL_DebugRec.Line);
			else sprintf(LPtr, "%s()", E3dSL_DebugRec.Function);

			LStrLen += strlen(LPtr);LPtr = LStr + LStrLen;
		}


		if(E3dSL_DebugRec.Renderer)
		{
			LStr = ERealloc(LStr, LStrLen + 16);LPtr = LStr + LStrLen;

			sprintf(LPtr, " Frame: %lld", (long long)(E3dSL_DebugRec.Renderer->FrameCnt));

			LStrLen += strlen(LPtr);LPtr = LStr + LStrLen;
		}


		if(E3dSL_DebugRec.StateString)
		{
			const int    LLen = strlen(E3dSL_DebugRec.StateString);
			const char*  LLStr = "  State: ";const int LLLen = strlen(LLStr);

			LStr = ERealloc(LStr, LStrLen + LLLen + LLen + 1);LPtr = LStr + LStrLen;
			memcpy(LPtr, LLStr, LLLen);LPtr += LLLen;

			memcpy(LPtr, E3dSL_DebugRec.StateString, LLen + 1);LPtr += LLen;
			LStrLen = LPtr - LStr;
		}


		if(E3dSL_DebugRec.Material && E3dSL_DebugRec.Material->Name)
		{
			const int    LLen = strlen(E3dSL_DebugRec.Material->Name);
			const char*  LLStr = "  Material: ";const int LLLen = strlen(LLStr);

			LStr = ERealloc(LStr, LStrLen + LLLen + LLen + 1);LPtr = LStr + LStrLen;
			memcpy(LPtr, LLStr, LLLen);LPtr += LLLen;

			memcpy(LPtr, E3dSL_DebugRec.Material->Name, LLen + 1);LPtr += LLen;
			LStrLen = LPtr - LStr;
		}


		if(E3dSL_DebugRec.ShaderVariant && E3dSL_DebugRec.ShaderVariant->Signature)
		{
			const int    LLen = strlen(E3dSL_DebugRec.ShaderVariant->Signature);
			const char*  LLStr = "  CgID: ";const int LLLen = strlen(LLStr);

			LStr = ERealloc(LStr, LStrLen + LLLen + LLen + 1);LPtr = LStr + LStrLen;
			memcpy(LPtr, LLStr, LLLen);LPtr += LLLen;

			memcpy(LPtr, E3dSL_DebugRec.ShaderVariant->Signature, LLen + 1);LPtr += LLen;
			LStrLen = LPtr - LStr;
		}


		if(E3dSL_DebugRec.ParameterName)
		{
			const int    LLen = (int)strlen(E3dSL_DebugRec.ParameterName);
			const char*  LLStr = "  Param: ";const int  LLLen = (int)strlen(LLStr);

			LStr = ERealloc(LStr, LStrLen + LLLen + LLen + 1);LPtr = LStr + LStrLen;
			memcpy(LPtr, LLStr, LLLen);LPtr += LLLen;

			memcpy(LPtr, E3dSL_DebugRec.ParameterName, LLen + 1);LPtr += LLen;
			LStrLen = LPtr - LStr;
//assert(0);
		}


//  printf("%s :: %s\n", cgGetErrorString(LastError), Listing);fflush(stdout);

// On old nVidia drivers, this causes a lock-up with the FP30 profile
//
		E3d_Log(ELogWARNING, "Cg", "%s  %s", LStr, cgGetErrorString(LastError));

//assert(0);

		if(LStr) EFree(LStr);
	}
}


static int E3dSL_ProfileInit(void)
{
	if(!_ProfileInitialized)
	{
	 if(cgGLIsProfileSupported(CG_PROFILE_VP40)) E3dCg_VertexProfile = CG_PROFILE_VP40;
	 else if(cgGLIsProfileSupported(CG_PROFILE_VP30)) E3dCg_VertexProfile = CG_PROFILE_VP30;
	 else if(cgGLIsProfileSupported(CG_PROFILE_ARBVP1)) E3dCg_VertexProfile = CG_PROFILE_ARBVP1;
	 else
	 {
		 E3d_Log(ELogWARNING, "Cg", "Vertex programming extensions (GL_ARB_vertex_program or " "GL_NV_vertex_program) not supported");
		 return(-1);
	 }

	 if(cgGLIsProfileSupported(CG_PROFILE_FP40)) E3dCg_FragmentProfile = CG_PROFILE_FP40;
	 else if(cgGLIsProfileSupported(CG_PROFILE_ARBFP1)) E3dCg_FragmentProfile = CG_PROFILE_ARBFP1;
	 else
	 {
		 E3d_Log(ELogWARNING, "Cg", "Fragment programming extensions (GL_ARB_fragment_program or " "GL_NV_fragment_program) not supported");
		 return(-1);
	 }

	 cgSetErrorCallback(_ErrorCallback);
	 E3dCg_ShaderContext = cgCreateContext();

	 _ProfileInitialized = TRUE;
	}

	return(E_SUCCESS);
}


static void  E3dSL_CodeGeneratorInit(E3dRendererClass* PRendererClass);
static void* _BindMaterial(E3dGPUContext* PGPUContext,  E3dMaterial* PMaterial, E3dScene* PScene, E3dRenderer* PRenderer,  E3dModel* PModel, E3dGeometry* PGeometry, E3dShadeable* PShadeable,  const E3dLightInstance* PLightInstances, const int PNLightInstances);
static void  _UnbindMaterial(E3dGPUContext* PGPUContext,  E3dMaterial* PMaterial, void* PHWShader, E3dRenderer* PRenderer);

static inline void E3dSL_SetTextureParameter(const E3dTextureMapper* PMapper, const E3dSLParameter PParam, const GLint PGLIndex, const int PTextureEngineIndex)
{
//printf("%s() %s:%d  TxEngine %d  GLID %d  GLTarget %d\n", __FUNCTION__, __FILE__, __LINE__, PTextureEngineIndex, PGLIndex, LGLTarget);fflush(stdout);

	cgGLSetTextureParameter(PParam, PGLIndex);
	cgGLEnableTextureParameter(PParam);
}

static void _SLVertexAttrPointer(const E3dSLParameter PSLID, const GLint PSLSize, const GLenum PType, const GLsizei PStride, const GLvoid* PPointer)
{
	cgGLSetParameterPointer(PSLID,  PSLSize,  PType,  PStride, PPointer);
	cgGLEnableClientState(PSLID);
}

static void _SLDisableVertexAttrArray(const E3dSLParameter PSLID)
{
	cgGLDisableClientState(PSLID);
}

#include "../SLCommon.ci"


static void* _BindMaterial(E3dGPUContext* PGPUContext,  E3dMaterial* PMaterial, E3dScene* PScene, E3dRenderer* PRenderer,  E3dModel* PModel, E3dGeometry* PGeometry, E3dShadeable* PShadeable,  const E3dLightInstance* PLightInstances, const int PNLightInstances)
{
TIMER_PUSH(E3dp_RendererTimerStack, "Cg BindMaterial");

//const ETime  LStartMicroseconds = E_GetMicroseconds();
	E3dTxMapperIndex*         LTextureMapperIndices;E3dTxMapperIndex  LNTxMappers;
	E3dTextureMapper* const*  LTxMappers;
	EBool                     LShadeableReceivesShadows = TRUE;

	if(EObject_IsSubClass(PShadeable, E3dPolyGroup))
	{
		const E3dPolyGroup*  LPolyGroup = (E3dPolyGroup*)PShadeable;
		LTextureMapperIndices = LPolyGroup->TextureMapperIndices;
		LTxMappers            = LPolyGroup->TxMappers;LNTxMappers = LPolyGroup->TxMappers_Count;
		LShadeableReceivesShadows = LPolyGroup->ReceivesShadows;
	}
	else return(NULL);

	E3dThisRenderer*                LRenderer = (E3dThisRenderer*)PRenderer;
	E3dRendererClass*               LRendererClass = LRenderer->Class;
	E3dShader*                      LShader = PMaterial->Shader;
	E3dShaderClass*                 LShaderClass = LShader->Class;
	E3dShaderVariant*               LShaderVariant = NULL;
	E3dShaderClassRendererBinding*  LShaderClassBinding = (E3dShaderClassRendererBinding*)E3dShader_ClassGetRendererBinding(LShader->Class, LRendererClass);if(LShaderClassBinding)
	{
#ifndef EGLES
		E3glDisable(GL_LIGHTING);
#endif
		_SL_BindMaterial0();

//printf("%s() %s:%d LShaderVariant [\n%s\n]\n", __FUNCTION__, __FILE__, __LINE__, LShaderVariant->VertexProgram->Source);fflush(stdout);
//printf("%s() %s:%d LShaderVariant [\n%s\n]\n", __FUNCTION__, __FILE__, __LINE__, LShaderVariant->FragmentProgram->Source);fflush(stdout);

// For _ErrorCallback
//
		E3dSL_DebugRec.Material = PMaterial;
		E3dSL_DebugRec.Shader = LShader;
		E3dSL_DebugRec.ShaderVariant = LShaderVariant;
		E3dSL_DebugRec.ParameterName = NULL;

		if(LShaderVariant->VertexProgram == NULL) { LShaderVariant->CreationResult = E_NO_PROGRAM;return(NULL); }

		E3dThisVertexProgram*  LThisVertexProgram = (E3dThisVertexProgram*)(LShaderVariant->VertexProgram);
		E3dCgFragmentProgram*  LFragmentProgram = (E3dCgFragmentProgram*)(LShaderVariant->FragmentProgram);
		E3dThisShaderVariant*  LThisShaderVariant = (E3dThisShaderVariant*)LShaderVariant;

		if(LShaderVariant->CreationResult == E_NOT_SET)  // Only try to create this GLSL shader, if we haven't tried yet (if it failed, there's no point spitting out the same errors int every frame)
		{
			EBool  LSaveShaderSource = (_SaveShaderSourceOn & E3dSL_ALWAYS) != 0;

// Compile the vertex program
//
			if(LThisVertexProgram->CgProgram == NULL)
			{
TIMER_PUSH(E3dp_RendererTimerStack, "VertexProgram");
				E3dSL_DebugRec.StateString = "Compiling vertex-program";E3dSL_DebugRec.Material = PMaterial;E3dSL_DebugRec.ShaderVariant = NULL;E3dSL_DebugRec.ParameterName = NULL;
				LThisVertexProgram->CgProgram = cgCreateProgram(E3dCg_ShaderContext, CG_SOURCE, LThisVertexProgram->Source, E3dCg_VertexProfile, "main", NULL);
//printf("%s() %s:%d [%d:%d]\n", __FUNCTION__, __FILE__, __LINE__, CG_SOURCE, E3dCg_VertexProfile);fflush(stdout);
				E3dSL_ClearDebugRec();

				if(LThisVertexProgram->CgProgram)
				{
					cgGLLoadProgram(LThisVertexProgram->CgProgram);

					_E3dSL_VertexProgramResolveSymbols(PRenderer, LShaderClassBinding, LThisShaderVariant, PMaterial,  LTxMappers, LNTxMappers,  LTextureMapperIndices, LNumOfTextures,  "vtx.cg");
				}
				else
				{
					E3d_Log(ELogWARNING, PRenderer->Class->Name, "Unable to compile vertex program for '%s'", LShaderVariant->Signature);
					LSaveShaderSource = TRUE;
					LShaderVariant->CreationResult = E_COMPILE_ERROR;
				}

				if((LSaveShaderSource || (_SaveShaderSourceOn & E3dSL_ALWAYS)) && LShaderVariant->Signature)
				{
					char*  LFileName = EMalloc(strlen(PMaterial->Name) + strlen(LShaderVariant->Signature) + 1 + 16);

					sprintf(LFileName, "%s%s-v.cg", PMaterial->Name, LShaderVariant->Signature);

					if(LThisVertexProgram->Source)
					{
						FILE*  LFile = fopen(LFileName, "w");if(LFile)
						{
							fwrite(LThisVertexProgram->Source, strlen(LThisVertexProgram->Source), 1, LFile);
							fclose(LFile);
						}
					}
				}
TIMER_POP(E3dp_RendererTimerStack);
			}


if(_PrintShaderSource)
{
	printf("%s() %s:%d  |%s| %sVertexProgram%s\n", __FUNCTION__, __FILE__, __LINE__, PMaterial->Name, EGreenCode, ENormalCode);fflush(stdout);
	EStr_PrintWithLineNumbers(LShaderVariant->VertexProgram->Source, ENormalCode);
	printf("%s() %s:%d  |%s| %sFragmentProgram%s [%s]\n", __FUNCTION__, __FILE__, __LINE__, PMaterial->Name, EGreenCode, ENormalCode, LShaderVariant->Signature);fflush(stdout);
	EStr_PrintWithLineNumbers(LShaderVariant->FragmentProgram->Source, ENormalCode);
}


// Compile the fragment program and resolve symbols in it
//
			const E3dSLCodeGenParams*  LCodeGenParams = &LShaderVariant->CodeGenParams;

			if(LFragmentProgram->Source && LFragmentProgram->CgProgram == NULL)
			{
TIMER_PUSH(E3dp_RendererTimerStack, "Fragment program");
				E3dSL_Location();E3dSL_DebugRec.StateString = _StateCompilingFragmentProgram;E3dSL_DebugRec.Material = PMaterial;E3dSL_DebugRec.ShaderVariant = LShaderVariant;
				LFragmentProgram->CgProgram = cgCreateProgram(E3dCg_ShaderContext, CG_SOURCE, LFragmentProgram->Source, E3dCg_FragmentProfile, "main", NULL);
				E3dSL_ClearDebugRec();
//printf("%s() %s:%d [%d:%d]\n", __FUNCTION__, __FILE__, __LINE__, CG_SOURCE, E3dCg_FragmentProfile);fflush(stdout);

				if(LFragmentProgram->CgProgram)
				{
					E3dSL_DebugRec.StateString = _StateLoadingFragmentProgram;
					cgGLLoadProgram(LFragmentProgram->CgProgram);

//E3dSLTEXTURE
E3dSL_Location();
					E3dSLM_ShaderResolveUniforms(PMaterial,  LTxMappers, LTextureMapperIndices);
E3dSL_ClearDebugRec();

/*
					if(LShaderVariant->TextureBlenderLists)
					{
						for(LC = 0;LC < LShaderVariant->Textures_Count;LC++)
						{
							EFree(LShaderVariant->TextureBlenderLists[LC]);
						}
						EFree(LShaderVariant->TextureBlenderLists);LShaderVariant->TextureBlenderLists = NULL;
					}
*/
				}
				else
				{
					LSaveShaderSource = TRUE;

					E3d_Log(ELogMESSAGE, "Cg", "%s() Error calling '%s'. Saving as '%s.cg' for debugging", __FUNCTION__, PMaterial->Name, LShaderVariant->Signature);fflush(stdout);
					E3d_Log(ELogWARNING, "Cg", "Shader-class %s: Unable to load fragment program: %s()", LShaderClass->Name, LShaderVariant->Signature);
					LShaderVariant->CreationResult = E_NO_PROGRAM;
				}


				if(LSaveShaderSource && LShaderVariant->Signature) E3dSL_FrgSaveSource(PMaterial, LShaderVariant, ".cg");
TIMER_POP(E3dp_RendererTimerStack);
			}

			if(LShaderVariant->CreationResult != E_NOT_SET && LShaderVariant->CreationResult != E_SUCCESS) return(NULL);

			LShaderVariant->CreationResult = E_SUCCESS;
		}


		if(LShaderVariant->CreationResult != E_SUCCESS) return(NULL);


E3dSL_ClearDebugRec();
TIMER_PUSH(E3dp_RendererTimerStack, "Bind programs");
E3dSL_Location();
		if(LThisVertexProgram->CgProgram)
		{
			cgGLBindProgram(LThisVertexProgram->CgProgram);
			cgGLEnableProfile(E3dCg_VertexProfile);
		}
		if(LFragmentProgram->CgProgram)
		{
			cgGLBindProgram(LFragmentProgram->CgProgram);
			cgGLEnableProfile(E3dCg_FragmentProfile);
		}
TIMER_POP(E3dp_RendererTimerStack);



E3dSL_ClearDebugRec();
TIMER_PUSH(E3dp_RendererTimerStack, "Vertex program inputs");
// Set vertex program inputs
//
//  cgGLSetStateMatrixParameter(LThisVertexProgram->ModelViewMatrixParam, CG_GL_MODELVIEW_MATRIX, CG_GL_MATRIX_IDENTITY);
E3dSL_Location();
		cgGLSetStateMatrixParameter(LThisVertexProgram->Param_VtU_ModelViewProjectionMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);
E3dSL_ClearDebugRec();


// Fragment program uniforms
//
TIMER_PUSH(E3dp_RendererTimerStack, "Vertex program uniforms");
// Set vertex-program uniform parameters
//
		if(((E3dCgVertexProgram*)(LShaderVariant->VertexProgram))->ResourceParameters)
		{
			if(LThisVertexProgram->Param_EyeLocalPosition) E3dSL_Uniform3f(LThisVertexProgram->Param_EyeLocalPosition,  LEyePos.X, LEyePos.Y, LEyePos.Z);

			E3dCgVertexProgram*  LSLVertexProgram = (E3dCgVertexProgram*)(LThisShaderVariant->VertexProgram);
			E3dSL_ShaderSetUniforms(PScene,  LLightInstancesForRenderPass, LNLightsForPass,  LRenderer, PMaterial, LThisShaderVariant, LShaderClassBinding, E3dVERTEX_PROGRAM, LSLVertexProgram->ResourceParameters,  LWorldToLocalMatrix,  LWorldToLocalUnitVectorMatrix,  &LCameraPosition, &LEyePos,  LRenderer->AmbientMult, &E3dSL_DebugRec);
		}
/*
		else
		{
printf("%s() '%s' No LThisShaderVariant->VertexProgram->ResourceParameters\n", __FUNCTION__, PMaterial->Name);fflush(stdout);
		}
*/
TIMER_POP(E3dp_RendererTimerStack);


// Set fragment-program uniform parameters
//
		if(LFragmentProgram->CgProgram)
		{
TIMER_PUSH(E3dp_RendererTimerStack, "Fragment program uniforms");

	TIMER_PUSH(E3dp_RendererTimerStack, "E3dSL_ShaderSetUniforms");
			E3dSL_ShaderSetUniforms(PScene,  LLightInstancesForRenderPass, LNLightsForPass,  LRenderer, PMaterial, LThisShaderVariant, LShaderClassBinding, E3dFRAGMENT_PROGRAM, LThisShaderVariant->FragmentUniformParams,  LWorldToLocalMatrix, LWorldToLocalUnitVectorMatrix,  &LCameraPosition, &LEyePos,  LRenderer->AmbientMult, &E3dSL_DebugRec);
	TIMER_POP(E3dp_RendererTimerStack);

	TIMER_PUSH(E3dp_RendererTimerStack, "cgGLEnableTextureParameter calls");
			{
				int  LC = 0;E3dSLLight*  LSHLight = LThisShaderVariant->Lights;for(;LC < LNLightsForPass;LC++, LSHLight++)
				{
					if(E3dSLParameterValid(LSHLight->Param_ShadowMap)) { E3dSL_DebugRec.ParameterName = "ShadowMap";cgGLEnableTextureParameter(LSHLight->Param_ShadowMap); }
				}
			}
	TIMER_POP(E3dp_RendererTimerStack);

	TIMER_PUSH(E3dp_RendererTimerStack, "E3dSL_SetTextureParams");
			E3dSL_SetTextureParams(PScene, LRenderer, PModel, PMaterial, LThisShaderVariant,  LNumOfTextures, LTxMappers, LTextureMapperIndices,  LLightInstancesForRenderPass, LNLightsForPass);
	TIMER_POP(E3dp_RendererTimerStack);


TIMER_POP(E3dp_RendererTimerStack);
		}
	}

//E_TimeStamp(PMaterial->Name);

//if(TIMING) E_NamedTime(_BindMaterial);
TIMER_POP(E3dp_RendererTimerStack);

	return(LShaderVariant);
}


static void _UnbindMaterial(E3dGPUContext* PGPUContext,  E3dMaterial* PMaterial, void* PHWShader, E3dRenderer* PRenderer)
{
	E3dThisRenderer*       LThisRenderer = (E3dThisRenderer*)PRenderer;
	E3dThisShaderVariant*  LThisShaderVariant = (E3dThisShaderVariant*)PHWShader;

	E3dSL_Location();

	if(LThisShaderVariant)
	{
		{
			EIndex  LC = 0;const int  LN = LThisRenderer->MaxNLightsPerPass;E3dSLLight*  LSHLight = LThisShaderVariant->Lights;for(;LC < LN;LC++, LSHLight++)
			{
				if(E3dSLParameterValid(LSHLight->Param_ShadowMap)) cgGLDisableTextureParameter(LSHLight->Param_ShadowMap);
			}
		}

		{
			EIndex  LC = 0;for(;LC < PMaterial->Textures_Count;LC++)
			{
				if(E3dSLParameterValid(LThisShaderVariant->TextureParam[LC])) cgGLDisableTextureParameter(LThisShaderVariant->TextureParam[LC]);
				if(E3dSLParameterValid(LThisShaderVariant->NormalMapParam[LC])) cgGLDisableTextureParameter(LThisShaderVariant->NormalMapParam[LC]);
			}
		}
	}

	cgGLDisableProfile(E3dCg_FragmentProfile);
	cgGLDisableProfile(E3dCg_VertexProfile);

/*
	{
//printf("%s() %s:%d  [%s] TextureEnginesUsed %d\n", __FUNCTION__, __FILE__, __LINE__, PMaterial->Name, LShaderVariant->TextureEnginesUsed);fflush(stdout);
		int  LC = LShaderVariant->TextureEnginesUsed - 1;for(;LC > 0;LC--)
		{
			E3glActiveTexture(GL_TEXTURE0 + LC);
			E3dGL_TexturesDisable();
		}

		E3glActiveTexture(GL_TEXTURE0);
	}
*/

	E3dSL_ClearDebugRec();
}


static void _InfoParse(E3dThisRenderer* PRenderer, E3dGeometry* PGeometry, E3dShadeable* PShadeable,  va_list PVArgs, const EBool PGenerateShader, EResource*** PVtxProgramResourcesPtrP, EResource*** PFrgProgramResourcesPtrP)
{
	E3dMaterial*        LMaterial;
//E3dTextureMapper**  LTextureMappers;

	if(EObject_IsSubClass(PShadeable, E3dPolyGroup))
	{
		E3dPolyGroup*  LPolyGroup = (E3dPolyGroup*)PShadeable;
		LMaterial = LPolyGroup->DrawingMaterial;
//	LTextureMappers = LPolyGroup->TextureMappers;
	}
	else return;

	E3dThisShaderVariant*  LThisShaderVariant = NULL;
	EIndex                 LFragmentProgramNumOfResources = 0;
	EIndex                 LVertexProgramNumOfResources = 0;
	short                  LArgType;

	while((LArgType = (short)va_arg(PVArgs, int)) != 0)  // Get the argument type, is it a terminating 0 ?
	{
		switch(LArgType)
		{
			case E3dSHADER_ID:
				{
					char**  LIDPtr = (char**)va_arg(PVArgs, void*);

					if(PGenerateShader)
					{
						if(LIDPtr) *LIDPtr = NULL;
						if(LMaterial->Shader && LMaterial->Shader->Class)
						{
							E3dShaderClassRendererBinding*	LShaderClassBinding = (E3dShaderClassRendererBinding*)E3dShader_ClassGetRendererBinding(LMaterial->Shader->Class, PRenderer->Class);

							if(LShaderClassBinding && LThisShaderVariant == NULL)
							{
//printf("%s() %s:%d E3dSHADER_ID\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
								LThisShaderVariant = (E3dThisShaderVariant*)E3dShader_GenerateCode(PGeometry, PShadeable, (E3dRenderer*)PRenderer, LShaderClassBinding,  E3dShaderPassBASE|E3dShaderPassLIGHTADD,  NULL, 1, NULL,  *PFrgProgramResourcesPtrP, &LFragmentProgramNumOfResources, *PVtxProgramResourcesPtrP, &LVertexProgramNumOfResources,  _PrintShaderSource || (_SaveShaderSourceOn & E3dSL_ALWAYS));
							}

							if(LThisShaderVariant)
							{
								if(LIDPtr) *LIDPtr = LThisShaderVariant->Signature;
							}
						}
					}
				}
			break;

			case E3dVERTEX_PROGRAM_RESOURCES:
				{
					*PVtxProgramResourcesPtrP = (EResource**)va_arg(PVArgs, void*);

					if(PGenerateShader)
					{
						if(LMaterial->Shader && LMaterial->Shader->Class)
						{
							E3dShaderClassRendererBinding*  LShaderClassBinding = (E3dShaderClassRendererBinding*)E3dShader_ClassGetRendererBinding(LMaterial->Shader->Class, PRenderer->Class);

							if(LShaderClassBinding && LThisShaderVariant == NULL)
							{
//printf("%s() E3dVERTEX_PROGRAM_RESOURCES %s:%d \n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
								LThisShaderVariant = (E3dThisShaderVariant*)E3dShader_GenerateCode(PGeometry, PShadeable, (E3dRenderer*)PRenderer, LShaderClassBinding,  E3dShaderPassBASE|E3dShaderPassLIGHTADD,  NULL, 1, NULL,  *PFrgProgramResourcesPtrP, &LFragmentProgramNumOfResources, *PVtxProgramResourcesPtrP, &LVertexProgramNumOfResources,  _PrintShaderSource || (_SaveShaderSourceOn & E3dSL_ALWAYS));
							}
						}

						if(LThisShaderVariant && LThisShaderVariant->VertexProgram)
						{
							if(*PVtxProgramResourcesPtrP)
							{
								if(**PVtxProgramResourcesPtrP)
								{
// Terminate resource-list
//
									EResource*  LResource = ELst_StructAdd(*PVtxProgramResourcesPtrP, &LVertexProgramNumOfResources, sizeof(EResource));

									memset(LResource, 0, sizeof(EResource));
								}
							}
						}
					}
					else if(*PVtxProgramResourcesPtrP) **PVtxProgramResourcesPtrP = NULL;
				}
			break;

			case E3dVERTEX_PROGRAM_SOURCE:
				{
					int     LIndex = (int)va_arg(PVArgs, int);
					char**  LSourcePtr = (char**)va_arg(PVArgs, void*);
					char**  LSignaturePtr = (char**)va_arg(PVArgs, void*);


					if(LSourcePtr) *LSourcePtr = NULL;
					if(LSignaturePtr) *LSignaturePtr = NULL;

					if(PGenerateShader)
					{
						if(LMaterial->Shader && LMaterial->Shader->Class)
						{
							E3dShaderClassRendererBinding*	LShaderClassBinding = (E3dShaderClassRendererBinding*)E3dShader_ClassGetRendererBinding(LMaterial->Shader->Class, PRenderer->Class);

							if(LShaderClassBinding)
							{
// Hand-written vertex program(s)?
//
								if(LShaderClassBinding->VertexProgramVariants_Count)
								{
									if(LIndex < LShaderClassBinding->VertexProgramVariants_Count)
									{
										E3dStageProgram*	LVertexProgramVariant = LShaderClassBinding->VertexProgramVariants[LIndex];

										if(LSourcePtr) *LSourcePtr = LVertexProgramVariant->Source;
										if(LSignaturePtr) *LSignaturePtr = LVertexProgramVariant->Signature;
									}
								}
								else if(LThisShaderVariant == NULL)
								{
//printf("%s() E3dVERTEX_PROGRAM_SOURCE %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
									LThisShaderVariant = (E3dThisShaderVariant*)E3dShader_GenerateCode(PGeometry, PShadeable, (E3dRenderer*)PRenderer, LShaderClassBinding,  E3dShaderPassBASE|E3dShaderPassLIGHTADD,  NULL, 1, NULL,  *PFrgProgramResourcesPtrP, &LFragmentProgramNumOfResources, *PVtxProgramResourcesPtrP, &LVertexProgramNumOfResources,  _PrintShaderSource || (_SaveShaderSourceOn & E3dSL_ALWAYS));

// Only LIndex == 0 is valid for generated vertex programs
//
									if(LIndex == 0 && LThisShaderVariant && LThisShaderVariant->VertexProgram)
									{
										if(LSourcePtr) *LSourcePtr = LThisShaderVariant->VertexProgram->Source;
									}
								}
							}
						}
					}
				}
			break;

			case E3dFRAGMENT_PROGRAM_RESOURCES:
				{
					*PFrgProgramResourcesPtrP = (EResource**)va_arg(PVArgs, void*);

					if(PGenerateShader)
					{
						if(LMaterial->Shader && LMaterial->Shader->Class)
						{
							E3dShaderClassRendererBinding*	LShaderClassBinding = (E3dShaderClassRendererBinding*)E3dShader_ClassGetRendererBinding(LMaterial->Shader->Class, PRenderer->Class);

							if(LShaderClassBinding && LThisShaderVariant == NULL)
							{
//printf("%s() E3dFRAGMENT_PROGRAM_RESOURCES %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
								LThisShaderVariant = (E3dThisShaderVariant*)E3dShader_GenerateCode(PGeometry, PShadeable, (E3dRenderer*)PRenderer, LShaderClassBinding,  E3dShaderPassBASE|E3dShaderPassLIGHTADD,  NULL, 1, NULL,  *PFrgProgramResourcesPtrP, &LFragmentProgramNumOfResources, *PVtxProgramResourcesPtrP, &LVertexProgramNumOfResources,  _PrintShaderSource || (_SaveShaderSourceOn & E3dSL_ALWAYS));
							}
						}

						if(LThisShaderVariant)
						{
							if(*PFrgProgramResourcesPtrP)
							{
// Does the resource list have any entries?
								if(**PFrgProgramResourcesPtrP)
								{
// Terminate resource-list
//
									EResource*  LResource = ELst_StructAdd(*PFrgProgramResourcesPtrP, &LFragmentProgramNumOfResources, sizeof(EResource));

									memset(LResource, 0, sizeof(EResource));
								}
							}
						}
					}
					else if(*PFrgProgramResourcesPtrP) **PFrgProgramResourcesPtrP = NULL;
				}
			break;

			case E3dFRAGMENT_PROGRAM_SOURCE:
				{
					char**  LSourcePtr = (char**)va_arg(PVArgs, void*);

					if(LSourcePtr) *LSourcePtr = NULL;

					if(PGenerateShader)
					{
						if(LMaterial->Shader && LMaterial->Shader->Class)
						{
							E3dShaderClassRendererBinding*	LShaderClassBinding = (E3dShaderClassRendererBinding*)E3dShader_ClassGetRendererBinding(LMaterial->Shader->Class, PRenderer->Class);

							if(LShaderClassBinding && LThisShaderVariant == NULL)
							{
//printf("%s() %s:%d E3dFRAGMENT_PROGRAM_SOURCE\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
								LThisShaderVariant = (E3dThisShaderVariant*)E3dShader_GenerateCode(PGeometry, PShadeable, (E3dRenderer*)PRenderer, LShaderClassBinding,  E3dShaderPassBASE|E3dShaderPassLIGHTADD,  NULL, 1, NULL,  *PFrgProgramResourcesPtrP, &LFragmentProgramNumOfResources, *PVtxProgramResourcesPtrP, &LVertexProgramNumOfResources,  _PrintShaderSource || (_SaveShaderSourceOn & E3dSL_ALWAYS));
							}
						}

						if(LThisShaderVariant)
						{
							E3dCgFragmentProgram*  LFragmentProgram = (E3dCgFragmentProgram*)(LThisShaderVariant->FragmentProgram);

							if(LSourcePtr)
							{
								if(LFragmentProgram && LFragmentProgram->Source) *LSourcePtr = LFragmentProgram->Source;
								else *LSourcePtr = NULL;
							}
						}
					}
				}
			break;
		}
	}
}


//================================================
// Info member function of the RendererClass
//================================================
static int _Info(E3dRenderer* PRenderer, E3dGeometry* PGeometry, E3dShadeable* PShadeable,  const int PAction, ...)
{
	switch(PAction)
	{
		case E3dGET:
			{
				E3dThisRenderer*  LRenderer = (E3dThisRenderer*)PRenderer;
				EResource**       LFrgProgramResourcesPtr = NULL;
				EResource**       LVtxProgramResourcesPtr = NULL;

				va_list  LVArgs;

// Parse arglist to see if we need FrgProgramResources and/or VtxProgramResources
//
				va_start(LVArgs, PAction);        // PAction is the last argument before the varargs
					_InfoParse(LRenderer, PGeometry, PShadeable,  LVArgs, FALSE,  &LVtxProgramResourcesPtr, &LFrgProgramResourcesPtr);
				va_end(LVArgs);

				va_start(LVArgs, PAction);        // PAction is the last argument before the varargs
					_InfoParse(LRenderer, PGeometry, PShadeable,  LVArgs, TRUE,   &LVtxProgramResourcesPtr, &LFrgProgramResourcesPtr);
				va_end(LVArgs);
			}
		break;
	}

	return(E_SUCCESS);
}


//================================================
// Free Cg bindings of a ShaderClass
//================================================
static void _ShaderClassFree(E3dShaderClass* PShaderClass, E3dShaderClassRendererBinding* PShaderClassRendererBinding)
{
	E3dShaderClassRendererBinding*  LShaderClassRendererBinding = (E3dShaderClassRendererBinding*)PShaderClassRendererBinding;

//printf("%s() {%s} NumOfHwShaders %d\n", __FUNCTION__, PShaderClass->Name, LShaderClassRendererBinding->NumOfHwShaders);fflush(stdout);
	if(LShaderClassRendererBinding->ShaderVariants)
	{
		E3dRendererClass*       LRendererClass = (E3dRendererClass*)(LShaderClassRendererBinding->RendererClass);
		const E3dShaderDevice*  LShaderDevice = LRendererClass->ShaderDevice;

		EIndex  LSVC = 0;const EIndex  LSVN = LShaderClassRendererBinding->ShaderVariants_Count;for(;LSVC < LSVN;LSVC++)
		{
			E3dThisShaderVariant*  LThisShaderVariant = (E3dThisShaderVariant*)LShaderClassRendererBinding->ShaderVariants[LSVC];
			EFreeChk(LThisShaderVariant->Signature);

			if(LThisShaderVariant->FragmentProgram) E3dCg_FragmentProgramFree((E3dCgFragmentProgram*)(LThisShaderVariant->FragmentProgram));

			if(LThisShaderVariant->TextureBlendingParameters)
			{
				int  LTxC = 0;const int  LTxN = LThisShaderVariant->NumOfTextures;for(;LTxC < LTxN;LTxC++)
				{
					EFreeChk(LThisShaderVariant->TextureBlendingParameters[LTxC]);
				}

				EFree(LThisShaderVariant->TextureBlendingParameters);
			}

			if(LThisShaderVariant->VertexProgram) LShaderDevice->VertexProgramFree(LThisShaderVariant->VertexProgram);


			if(LThisShaderVariant->TextureBlenderLists)
			{
				int  LC = 0;for(;LC < LThisShaderVariant->NumOfTextures;LC++)
				{
					EFree(LThisShaderVariant->TextureBlenderLists[LC]);
				}
				EFree(LThisShaderVariant->TextureBlenderLists);LThisShaderVariant->TextureBlenderLists = NULL;
			}
			EFree(LThisShaderVariant);
		}

		if(LShaderClassRendererBinding->VertexProgramVariants)
		{
			int  LVC = 0;const int  LVN = LShaderClassRendererBinding->VertexProgramVariants_Count;for(;LVC < LVN;LVC++)
			{
				LShaderDevice->VertexProgramFree(LShaderClassRendererBinding->VertexProgramVariants[LVC]);
			}

			EList_Free(LShaderClassRendererBinding->VertexProgramVariants);
		}

		EList_FreeA(LShaderClassRendererBinding->ShaderVariants);
	}
}



//
// Shader code generator setup
//

// Resource-translator for sampler2D
//
static const char  _sampler2D[] = "sampler2D";
static const char  _tex2D[] = "tex2D";

static int _RT_sampler2D(EResContext* PContext, const void* PDStructAddress, const void* PSPtr, void* PDPtr, int PAction)
{
	switch(PAction)
	{
		case EResGET_TYPE_NAME:
			*((const char**)PDPtr) = _sampler2D;
		return(E_SUCCESS);

		case EResGET_FUNC_NAME:
			*((const char**)PDPtr) = _tex2D;
		return(E_SUCCESS);
	}
	return(E_NOTHING_TO_DO);
}


// Resource-translator for sampler3D
//
static const char  _sampler3D[] = "sampler3D";

static int _RT_sampler3D(EResContext* PContext, const void* PDStructAddress, const void* PSPtr, void* PDPtr, const int PAction)
{
	switch(PAction)
	{
		case EResGET_TYPE_NAME:
			*((const char**)PDPtr) = _sampler3D;
		return(E_SUCCESS);

		case EResSET_VALUE:
			{
				cgGLSetTextureParameter((CGparameter)PSPtr, *((GLuint*)PDPtr));
				cgGLEnableTextureParameter((CGparameter)PSPtr);
			}
		return(E_SUCCESS);
	}
	return(E_NOTHING_TO_DO);
}


// Resource-translator for samplerCUBE
//
static const char  _samplerCUBE[] = "samplerCUBE";
static const char  _texCUBE[] = "texCUBE";

static int _RT_samplerCUBE(EResContext* PContext, const void* PDStructAddress, const void* PSPtr, void* PDPtr, const int PAction)
{
	switch(PAction)
	{
		case EResGET_TYPE_NAME:
			*((const char**)PDPtr) = _samplerCUBE;
		return(E_SUCCESS);

		case EResGET_FUNC_NAME:
			*((const char**)PDPtr) = _texCUBE;
		return(E_SUCCESS);
	}
	return(E_NOTHING_TO_DO);
}


// Resource-translator for float3x3
//
static const char  _float3x3[] = "float3x3";
static const char  _float3x3_Default[] = "\n1 0 0\n0 1 0\n0 0 1";

static int _RT_float3x3(EResContext* PContext, const void* PDStructAddress, const void* PSPtr, void* PDPtr, const int PAction)
{
	switch(PAction)
	{
		case EResGET_TYPE_NAME:
			*((const char**)PDPtr) = _float3x3;
		return(E_SUCCESS);

		case EResGET_DEFAULT_VALUE_STRING:
			*((const char**)PDPtr) = _float3x3_Default;
		return(E_SUCCESS);
	}
	return(E_NOTHING_TO_DO);
}


// Resource-translator for float3x3
//
static const char  _float3x4[] = "float3x4";
static const char  _float3x4_Default[] = "\n1 0 0\n0 1 0\n0 0 1\n0 0 0";

static int _RT_float3x4(EResContext* PContext, const void* PDStructAddress, const void* PSPtr, void* PDPtr, const int PAction)
{
	switch(PAction)
	{
		case EResGET_TYPE_NAME:
			*((const char**)PDPtr) = _float3x4;
		return(E_SUCCESS);

		case EResGET_DEFAULT_VALUE_STRING:
			*((const char**)PDPtr) = _float3x4_Default;
		return(E_SUCCESS);
	}
	return(E_NOTHING_TO_DO);
}


void E3dCg_FragmentProgramFree(E3dCgFragmentProgram* LProgram)
{
	if(LProgram->RefCnt > 0) LProgram->RefCnt -= 1;
	if(LProgram->RefCnt > 0) return;

	if(LProgram->CgProgram) cgDestroyProgram(LProgram->CgProgram);
	if(LProgram->Source && !LProgram->SourceStatic) EFree(LProgram->Source);

	EFree(LProgram);
}


static int _VertexProgramInit(E3dStageProgram* PVertexProgram, E3dMaterial* PMaterial, EResource* PResources,  E3dTextureMapper* const* PTxMappers, const E3dTxMapperIndex* PTextureMapperIndices, const int PNumOfTextures)
{
//printf("%s() %s:%d  %sUNUSED CODE%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
	return(E_SUCCESS);
}


static void _VertexProgramFree(E3dStageProgram* PPrg)
{
	E3dCgVertexProgram*  LProgram = (E3dCgVertexProgram*)PPrg;

	if(LProgram->RefCnt > 0) LProgram->RefCnt -= 1;
	if(LProgram->RefCnt > 0) return;

	if(LProgram->Source && !LProgram->SourceStatic) EFree(LProgram->Source);
	if(LProgram->ResourceParameters && !LProgram->ResourceParametersStatic) EFree(LProgram->ResourceParameters);

	if(LProgram->CgProgram) cgDestroyProgram(LProgram->CgProgram);

	EFree(LProgram);
}

static const char*  _VertexHeader=
"\
#define E3dFloat2              float2\n\
#define E3dFloat3              float3\n\
#define E3dFloat3_cast         float3\n\
#define E3dFloat4              float4\n\
#define E3dFloat3x3            float3x3\n\
#define E3dFloat4x4            float4x4\n\
#define E3dMatrix_Mul(mM, mV)  mul(mM, mV)\n\
";

static const char*  _BuiltIn_VertexNormalInParamName   = "NORMAL";
static const char*  _BuiltIn_VertexTexCoordInParamName = "TEXCOORD";

static const char*  _Fragment_OutColorName = "OutColor";

static const char*  _VUParamQualifier = "uniform";
static const char*  _VUParamPrefix = "P";
static const char*  _FUParamQualifier = "uniform";
static const char*  _FUParamPrefix = "P";
static const char*  _Param_DefFormatString = "uniform %s %s";
static const char*  _Sampler2D_DefFormatString = "uniform sampler2D %s";
static const char*  _FragmentHeader=
"\
#define E3dFloat                 float\n\
#define E3dFloat_cast            float\n\
#define E3dFloat2                float2\n\
#define E3dFloat2_cast           float2\n\
#define E3dFloat3                float3\n\
#define E3dFloat3_cast           float3\n\
#define E3dFloat4                float4\n\
#define E3dFloat4_cast           float4\n\
#define E3dFloat3x3              float3x3\n\
#define E3dSamplerCube           samplerCUBE\n\
#define E3dTextureCube           texCUBE\n\
#define E3dHalf                  half\n\
#define E3dHalf2                 half2\n\
#define E3dHalf3                 half3\n\
#define E3dHalf4                 half4\n\
#define Pconst                   const\n\
#define Lconst                   const\n\
#define E3dMatrix_Mul(mM, mV)    mul(mM, mV)\n\
#define LightingFloat3_t         float3\n\
#define LightingColor_t          half4\n\
#define TextureColor_t           half4\n\
#define OutColor_t               half4\n\
#define OutColor_t_cast          half4\n\
#define TextureBlendingFactor_t  half\n\
#define SL_MIN(m0, m1)           min(m0, m1)\n\
#include <cstdlib>\n\n\
";

static const char*  _FragmentFunctionPrefix = "void";
static const char*  _FragmentEntryPointName = "main";


//========================================================================================
// Configure EQUINOX-3D's universal shader code-generator to generate Cg shaders
//========================================================================================
static void E3dSL_CodeGeneratorInit(E3dRendererClass* PRendererClass)
{
	E3dShaderDevice*  LShaderDevice = PRendererClass->ShaderDevice;if(LShaderDevice->CodeGeneratorInitialized) return;

	LShaderDevice->ShadowMappingAutoColor = TRUE;

	LShaderDevice->LanguageVersionString = EStrDup("Cg 3.1");

	LShaderDevice->ParametersGlobal = FALSE;  // Cg can use structs as function args for shader parameters
	LShaderDevice->SupportsConstForLocalVariables = TRUE;

	LShaderDevice->GetShaderSignature = E3dShaderVariant_GetSignature;
	LShaderDevice->MaxNumOfTextures = E3d_Min(E3dSLMAX_NUM_OF_2DTEXTURES, E3dGL_MAX_TEXTURE_IMAGE_UNITS);

	LShaderDevice->VertexHeader = _VertexHeader;

	LShaderDevice->FragmentHeader = _FragmentHeader;
	LShaderDevice->FragmentFunctionPrefix = _FragmentFunctionPrefix;
	LShaderDevice->FragmentEntryPointName = _FragmentEntryPointName;

	LShaderDevice->BuiltIn_VertexNormalInParamName   = _BuiltIn_VertexNormalInParamName;
	LShaderDevice->BuiltIn_VertexTexCoordInParamName = _BuiltIn_VertexTexCoordInParamName;

	LShaderDevice->BuiltIn_FragmentOutColorName = _Fragment_OutColorName;

	LShaderDevice->RT_sampler2D       = _RT_sampler2D;
	LShaderDevice->RT_sampler2DShadow = _RT_sampler2D;
	LShaderDevice->RT_sampler3D   = _RT_sampler3D;
	LShaderDevice->RT_samplerCUBE = _RT_samplerCUBE;
	LShaderDevice->RT_float3x3    = _RT_float3x3;
	LShaderDevice->RT_float3x4    = _RT_float3x4;
	EpPixelFormat_Set(&LShaderDevice->NormalMapPixelFormat,  EpRGB, 3, EResUINT8);

	LShaderDevice->HasSwizzle = TRUE;
	LShaderDevice->XName = 'x';LShaderDevice->YName = 'y';LShaderDevice->ZName = 'z';LShaderDevice->WName = 'w';
	LShaderDevice->RName = 'r';LShaderDevice->GName = 'g';LShaderDevice->BName = 'b';LShaderDevice->AName = 'a';

	LShaderDevice->VUParamQualifier = _VUParamQualifier;
	LShaderDevice->VUParamPrefix = _VUParamPrefix;

	LShaderDevice->FUParamQualifier = _FUParamQualifier;
	LShaderDevice->FUParamPrefix = _FUParamPrefix;

	LShaderDevice->Param_DefFormatString = _Param_DefFormatString;
	LShaderDevice->Sampler2D_DefFormatString = _Sampler2D_DefFormatString;
	LShaderDevice->Sampler2DShadow_DefFormatString = _Sampler2D_DefFormatString;
	LShaderDevice->Sampler_PassCoordinatesAsVector = TRUE;
	LShaderDevice->VectorOperators = TRUE;

	LShaderDevice->AccumulateLightPerShadeCall = TRUE;

	LShaderDevice->CodeGeneratorInitialized = TRUE;
	PRendererClass->ShaderDevice = LShaderDevice;
}


E3dRendererClass*  E3dPGPURenderer_class = NULL;

//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
// Register common superclass for programmable-GPU renderers.
// This is a "blank" class. It's only used to identify PGPU renderers.
//
	{
		E3dRendererClassDescriptor  LClassDescriptor =
		{
			.SuperClass=       E3dGPURenderer_class,
			.Name=             "PGPURenderer",
			.StructSize=       sizeof(E3dPGPURenderer),
		};

		E3dPGPURenderer_class = E3dRenderer_ClassRegister(&LClassDescriptor);
	}

	E3dRendererClassDescriptor  LRendererClassDescriptor =
	{
		.SuperClass=       E3dPGPURenderer_class,
		.Name=             "Cg",
		.StructSize=       sizeof(E3dThisRenderer),
		.Resources=        _Resources,

		.Init=             _Init,
		.Free=             _Free,
		.Update=           _Update,
		.Render=           _Render,
		.BindMaterial=     _BindMaterial,
		.UnbindMaterial=   _UnbindMaterial,
		.ShaderClassFree=  _ShaderClassFree,

		.GUI=              E3dPGPURenderer_GUI,
		.Flags=            E3dHDR_IMAGE
	};


	_Plugin = PPlugin;

	E3dSL_ClearDebugRec();

	E3dRendererClass*  LRendererClass = E3dRenderer_ClassRegister(&LRendererClassDescriptor);if(LRendererClass)
	{
		_RendererClass = LRendererClass;

		LRendererClass->ShaderClassRendererBindingSize = sizeof(E3dShaderClassRendererBinding);
		LRendererClass->ShaderVariantSize = sizeof(E3dThisShaderVariant);
		LRendererClass->Info = _Info;

		E3dShaderDevice*  LShaderDevice;ENewClr(LShaderDevice);

		LRendererClass->ShaderDevice = LShaderDevice;

		LShaderDevice->VertexProgramInit = _VertexProgramInit;
		LShaderDevice->VertexProgramFree = _VertexProgramFree;
		LShaderDevice->VertexProgramSize = sizeof(E3dCgVertexProgram);

		LShaderDevice->FragmentProgramSize    = sizeof(E3dCgFragmentProgram);
		LShaderDevice->ShaderCheckNeededTerms = E3dSL_ShaderCheckNeededTerms;


		_ConstantShaderClass = E3dShader_ClassFindByName("Constant");if(_ConstantShaderClass)
		{
			E3dShaderClassRendererBinding  LDescriptor=
			{
				.ShaderBindingStructSize=  sizeof(E3dShaderCgBinding),
				.Resources=                E3dSL_ConstantResources,
			};
			E3dShaderClassRendererBinding*  LShaderClassBinding = E3dShader_ClassAddRenderer(_ConstantShaderClass, LRendererClass, &LDescriptor);

			if(LShaderClassBinding) LShaderClassBinding->SourceCore = (char*)Material_Constant_core_sl;
		}

		_LambertShaderClass = E3dShader_ClassFindByName("Lambert");if(_LambertShaderClass)
		{
			E3dShaderClassRendererBinding  LDescriptor=
			{
				.ShaderBindingStructSize=  sizeof(E3dShaderCgBinding),
				.Resources=                E3dSL_LambertResources,
			};
			E3dShaderClassRendererBinding*  LShaderClassBinding = E3dShader_ClassAddRenderer(_LambertShaderClass, LRendererClass, &LDescriptor);

			if(LShaderClassBinding) LShaderClassBinding->SourceCore = (char*)Material_Lambert_core_sl;
		}

		_PhongShaderClass = E3dShader_ClassFindByName("Phong");if(_PhongShaderClass)
		{
			E3dShaderClassRendererBinding  LDescriptor=
			{
				.ShaderBindingStructSize=  sizeof(E3dShaderCgBinding),
				.Resources=                E3dSL_PhongResources,
			};
			E3dShaderClassRendererBinding*  LShaderClassBinding = E3dShader_ClassAddRenderer(_PhongShaderClass, LRendererClass, &LDescriptor);if(LShaderClassBinding)
			{
				LShaderClassBinding->SourceCore = (char*)Material_Phong_core_sl;

				_FeatureCg = E3d_FeatureRegister(E3dFEATURE_HARDWARE_RENDERER, "Cg shaders");
			}
		}
	}

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(_RendererClass) { E3dRenderer_ClassDeactivate(_RendererClass);_RendererClass = NULL; }
	if(E3dPGPURenderer_class) { E3dRenderer_ClassDeactivate(E3dPGPURenderer_class);E3dPGPURenderer_class = NULL; }

//printf("%s:%d  Cg Renderers_Count %d\n", __FILE__, __LINE__, _RendererClass->Renderers_Count);fflush(stdout);

	if(E3dCg_ShaderContext) { cgDestroyContext(E3dCg_ShaderContext);E3dCg_ShaderContext = 0; }

	if(_FeatureCg) { E3d_FeatureRemove(_FeatureCg);_FeatureCg = NULL; }
	if(_FeatureFBO) { E3d_FeatureRemove(_FeatureFBO);_FeatureFBO = NULL; }
	if(_FeatureHDR) { E3d_FeatureRemove(_FeatureHDR);_FeatureHDR = NULL; }

	return(E_SUCCESS);
}
