//================================================================
// EQUINOX-3D(TM) plugin
//
// - Hierarchy walking
// - Select Objects, parts of Objects/Geometries (PolyGroup, Polygon, Edge, Vertex etc.)
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 11:18:28
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <E3D/E3D.h>
#include <E3D/Select.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage

#include <Xe/PushButton.h>
#include <Xe/Shell.h>
#include <Xe/Label.h>
#include <Xe/Dialog.h>
#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/ToggleButton.h>

char*	E_PluginInfo = "Advanced selection";
static EPlugin*      _Plugin = NULL;

static EguiItem          _MenuButtons[32];static int  _MenuButtons_Count = 0;
static EScriptFunction*  _ScriptFunctions[32];static int  _ScriptFunctions_Count = 0;

static Widget        _SelectEdgesAtNormalCreasesDialogW = NULL;
static EPluginLock*  _SelectEdgesAtNormalCreasesDialogLock = NULL;

// Operations
//
enum
{
	E3dCrALL_HIERARCHIES = 1,
	E3dCrALL_NORMAL_MODELS,
	E3dCrEMPTY_MODELS,
	E3dCrHIDDEN_GEOMETRIES,
	E3dCrUNSELECT_EVERYTHING,
	E3dCrUNSELECT_ALL_MODELS,
	E3dCrSELECT_BY_NAME,
	E3dCrSELECT_EDGES,
	E3dCrSELECT_HULLS,
	E3dCrSELECT_TRIS,
	E3dCrSELECT_QUADS,

	E3dCrSELECT_EDGES_WITH_SELECTED_VERTICES,
	E3dCrSELECT_EDGES_OF_SELECTED_POLYGONS,
	E3dCrSELECT_EDGES_AT_NORMAL_CREASES,
	E3dCrUNSELECT_EDGES,
	E3dCrSELECT_BOUNDARY_EDGES,

	E3dCrPOLYGONS_WITH_ALL_VERTICES_SELECTED,
	E3dCrPOLYGONS_WITH_A_SELECTED_EDGE,
	E3dCrUNSELECT_ALL_POLYS,
	E3dCrINVERT_POLYGON_SELECTION,
	E3dCrINVERT_POLYGROUP_SELECTION,
	E3dCrGO_UP,
	E3dCrGO_DOWN,
	E3dCrGO_LEFT,
	E3dCrGO_RIGHT
};


enum
{
	E3dSELECT_POINTS_OF_SELECTION = 1,
	E3dUNSELECT_POINTS_OF_SELECTION,
	E3dSELECT_UNUSED_POINTS,
};


//========================================
// Node selection
//========================================
static void _NodeSelection(const int PAction)
{
	E3dScene*  LScene = E3d_Scene;
	E3dNode**  LRootNodes = LScene->RootNodes;

	EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*  LNode = LRootNodes[LC];for(;LNode;LNode = LNode->Next)
		{
			switch(PAction)
			{
				case E3dCrALL_NORMAL_MODELS:
					LNode->Selection = E3dNodeSelNODE;
				break;

				case E3dCrEMPTY_MODELS:
					{
						E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
						{
							if(LModel->Objects_Count == 0) LNode->Selection = E3dNodeSelNODE;
						}
					}
				break;

				case E3dCrALL_HIERARCHIES:
					if(LNode->Parent == NULL) LNode->Selection = E3dNodeSelBRANCH_ROOT;
					else LNode->Selection = E3dNodeSelBRANCH;
				break;

				case E3dCrUNSELECT_EVERYTHING:
					LNode->Selection = E3dNodeSelNONE;
					if(EObject_IsSubClass(LNode, E3dModel)) E3dModel_UnselectObjects(LScene, (E3dModel*)LNode,  NULL, 0,  E3dSF_ALL, TRUE);
				break;

				case E3dCrUNSELECT_ALL_MODELS:
					LNode->Selection = E3dNodeSelNONE;
					if(EObject_IsSubClass(LNode, E3dModel)) E3dModel_UnselectObjects(LScene, (E3dModel*)LNode,  NULL, 0,  E3dSF_MODELS, TRUE);
				break;
			}
		}
	}
}


//========================================
// Node/Geometry
//========================================
static void _SelectHidden(const uint16_t PWhatToInclude)
{
	E3dScene*  LScene = E3d_Scene;

	E3dSel_SceneUnselectAll(LScene);

	if(PWhatToInclude & E3dItemGEOMETRY)
	{
		E3dObject**  LObjects;EIndex  LObjects_Count = E3dScene_GetObjects(LScene, &LObjects, E3dItemGEOMETRY, E3dVISIBLE|E3dHIDDEN, FALSE);if(LObjects_Count)
		{
			EIndex  LObjC = 0;for(;LObjC < LObjects_Count;LObjC++)
			{
				E3dObject*  LObject = LObjects[LObjC];if(LObject->Visible)
				{
//printf("%s() %s:%d <%s>\n", __FUNCTION__, __FILE__, __LINE__, LObject->Name);fflush(stdout);
					EBool  LObjectHasVisibleInstance = FALSE;
					EIndex  LMC = 0;for(;LMC < LObject->Models_Count;LMC++)
					{
						if(LObject->Models[LMC]->Visible) { LObjectHasVisibleInstance = TRUE;break; }
					}
					if(!LObjectHasVisibleInstance) E3dSel_ObjectSelect(LScene, LObjects[LObjC], FALSE, FALSE);
				}
				else E3dSel_ObjectSelect(LScene, LObjects[LObjC], FALSE, FALSE);
			}

			if(LObjects_Count == 1) E3dp_PrintMessage(ELogMESSAGE, 5000, "Selected <%s>", LObjects[0]->Name ? LObjects[0]->Name : "");
			else E3dp_PrintMessage(ELogMESSAGE, 5000, "Selected %lld Objects", (long long)LObjects_Count);

			EFree(LObjects);
		}
	}


	if(PWhatToInclude & E3dItemNODE)
	{
		E3dNode**  LRootNodes = E3d_Scene->RootNodes;
		EIndex     LC = 0;const EIndex  LN = E3d_Scene->RootNodes_Count;for(;LC < LN;LC++)
		{
			E3dNode*  LNode = LRootNodes[LC];for(;LNode;LNode = LNode->Next)
			{
				if(!LNode->Visible) E3dSel_NodeSelect(LScene, LNode, FALSE, FALSE);
			}
		}
	}
}


static EIndex _TagVerticesOfSelectedPolygons(E3dPolyGroup* PPolyGroup, const EBool PSelectAllForPolyGroupsWithNoSelectedPolygons, E3dVertex* PVertices, const E3dVertexSize PVertexSize)
{
	EIndex  LPointSelectionChanged = 0;

	const E3dVertexNodeSize  LVertexNodeSize = PPolyGroup->VertexNodeSize;

	if(E3dPolyGroup_HasSelectedPolygons(PPolyGroup))
	{
		EIndex  LPC = 0;const EIndex  LPN = PPolyGroup->Polygons_Count;E3dPolygon*  LPolygon = PPolyGroup->Polygons;for(;LPC < LPN;LPC++, LPolygon++)
		{
			if(LPolygon->VertexNodes)
			{
				if(LPolygon->Flags & E3dPolygonSELECTED)
				{
					const E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertexNodes;for(;LVC < LVN;LVC++)
					{
						const EIndex  LVIndex = LVertexNode->VertexID;if(LVIndex > -1)
						{
							E3dVertex*  LVertex = EStruct_ByIndex(PVertices, PVertexSize, LVIndex);
							if((LVertex->Flags & E3dVtxSELECTED) == 0) { LVertex->Flags |= E3dVtxSELECTED;LPointSelectionChanged++; }
						}
						EStruct_Inc(LVertexNode, LVertexNodeSize);
					}
				}
			}
		}
	}
	else if(PSelectAllForPolyGroupsWithNoSelectedPolygons)
	{
		EIndex  LPC = 0;const EIndex  LPN = PPolyGroup->Polygons_Count;E3dPolygon*  LPolygon = PPolyGroup->Polygons;for(;LPC < LPN;LPC++, LPolygon++)
		{
			if(LPolygon->VertexNodes)
			{
				const E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertexNodes;for(;LVC < LVN;LVC++)
				{
					const EIndex  LVIndex = LVertexNode->VertexID;if(LVIndex > -1)
					{
						E3dVertex*  LVertex = EStruct_ByIndex(PVertices, PVertexSize, LVIndex);
						if((LVertex->Flags & E3dVtxSELECTED) == 0) { LVertex->Flags |= E3dVtxSELECTED;LPointSelectionChanged++; }
					}
					EStruct_Inc(LVertexNode, LVertexNodeSize);
				}
			}
		}
	}

	return(LPointSelectionChanged);
}


static EIndex _UnTagVerticesOfSelectedPolygons(E3dPolyGroup* PPolyGroup, const EBool PSelectAllForPolyGroupsWithNoSelectedPolygons, E3dVertex* PVertices, const E3dVertexSize PVertexSize)
{
	EIndex  LPointSelectionChanged = 0;

	const E3dVertexNodeSize  LVertexNodeSize = PPolyGroup->VertexNodeSize;

	if(E3dPolyGroup_HasSelectedPolygons(PPolyGroup))
	{
		EIndex  LPC = 0;const EIndex  LPN = PPolyGroup->Polygons_Count;E3dPolygon*  LPolygon = PPolyGroup->Polygons;for(;LPC < LPN;LPC++, LPolygon++)
		{
			if(LPolygon->VertexNodes)
			{
				if(LPolygon->Flags & E3dPolygonSELECTED)
				{
					const E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
					EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertexNodes;for(;LVC < LVN;LVC++)
					{
						const EIndex  LVIndex = LVertexNode->VertexID;if(LVIndex > -1)
						{
							E3dVertex*  LVertex = EStruct_ByIndex(PVertices, PVertexSize, LVIndex);
							if(LVertex->Flags & E3dVtxSELECTED) { LVertex->Flags &= ~E3dVtxSELECTED;LPointSelectionChanged++; }
						}
						EStruct_Inc(LVertexNode, LVertexNodeSize);
					}
				}
			}
		}
	}
	else if(PSelectAllForPolyGroupsWithNoSelectedPolygons)
	{
		EIndex  LPC = 0;const EIndex  LPN = PPolyGroup->Polygons_Count;E3dPolygon*  LPolygon = PPolyGroup->Polygons;for(;LPC < LPN;LPC++, LPolygon++)
		{
			if(LPolygon->VertexNodes)
			{
				const E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
				EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertexNodes;for(;LVC < LVN;LVC++)
				{
					const EIndex  LVIndex = LVertexNode->VertexID;if(LVIndex > -1)
					{
						E3dVertex*  LVertex = EStruct_ByIndex(PVertices, PVertexSize, LVIndex);
						if(LVertex->Flags & E3dVtxSELECTED) { LVertex->Flags &= ~E3dVtxSELECTED;LPointSelectionChanged++; }
					}
					EStruct_Inc(LVertexNode, LVertexNodeSize);
				}
			}
		}
	}

	return(LPointSelectionChanged);
}


static EIndex _SplineTagCVs(E3dSpline* PSpline)
{
	EIndex  LPointSelectionChanged = 0;

	EIndex  LC = 0;const EIndex  LN = PSpline->CVs_Count;E3dSplineCV*  LCV = PSpline->CVs;for(;LC < LN;LC++, LCV++)
	{
		if((LCV->CVFlags & E3dCV_TAGGED) == 0)
		{
			LCV->CVFlags |= E3dCV_TAGGED;
			LPointSelectionChanged++;
		}
	}
	return(LPointSelectionChanged);
}


static EIndex _SplineUntagCVs(E3dSpline* PSpline)
{
	EIndex  LPointSelectionChanged = 0;

	EIndex  LC = 0;const EIndex  LN = PSpline->CVs_Count;E3dSplineCV*  LCV = PSpline->CVs;for(;LC < LN;LC++, LCV++)
	{
		if(LCV->CVFlags & E3dCV_TAGGED)
		{
			LCV->CVFlags &= ~E3dCV_TAGGED;
			LPointSelectionChanged++;
		}
	}
	return(LPointSelectionChanged);
}


static EIndex _MeshSelectIntermediatePoints(E3dMesh* PMesh, const EBool PDoBoundaryEdges)
{
	E3dMesh_GetVertexUsageList(PMesh, FALSE, TRUE);
	E3dMesh_CreateWingedEdgesFromEdges(PMesh, TRUE, FALSE);
	EIndex  LNVerticesSelected = 0;

	const E3dVertexUsage*  LVertexUsageList = PMesh->VertexUsageList;
	const E3dWingedEdge*   LMeshWingedEdges = PMesh->WingedEdges;

	EIndex  LMVC = 0;const EIndex  LMVN = PMesh->Vertices_Count;E3dVertex*  LVertex = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;const E3dVertexUsage*  LVertexUsage = LVertexUsageList;for(;LMVC < LMVN;LMVC++, EStruct_Inc(LVertex, LVertexSize), LVertexUsage++)
	{
		const E3dValenceCount  LVN = LVertexUsage->EdgeEnds_Count;if(LVN == 2)
		{
			E3dValenceCount  LVC = 0;const E3dEdgeEnd*  LEdgeEnd = LVertexUsage->EdgeEnds;for(;LVC < LVN;LVC++, LEdgeEnd++)
			{
				if(!PDoBoundaryEdges)
				{
					const EIndex          LEIndex = LEdgeEnd->EdgeIndex;
					const E3dWingedEdge*  LTWingedEdge = LMeshWingedEdges + LEIndex;
					if(LTWingedEdge->Polygons_Count < 2) goto NextVertex;
				}
			}

			LVertex->Flags |= E3dVtxSELECTED;
			LNVerticesSelected++;
NextVertex:;
		}
	}
printf("%s() %s:%d %lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LNVerticesSelected);fflush(stdout);
	return(LNVerticesSelected);
}


static EChangeFlags _Select_IntermediatePoints(E3dScene* PScene)
{
	EIndex        LNTotalSelectedUnselected = 0;
	EChangeFlags  LSceneChangedFlags = 0;

	E3dObject**  LObjects;EIndex  LObjects_Count = E3dScene_GetObjects(PScene, &LObjects, E3dItemALL, E3dSELECTED, FALSE);if(LObjects_Count)
	{
		EIndex  LObjC = 0;for(;LObjC < LObjects_Count;LObjC++)
		{
			E3dObject*  LObject = LObjects[LObjC];
			EIndex      LPointSelectedUnselected = 0;

			if(EObject_IsSubClass(LObject, E3dMesh))
			{
				E3dMesh*    LMesh = (E3dMesh*)LObject;

// Select intermediate points on selected Objects
// - On Meshes: select Vertices that subdivide Edges (valence 2, non-boundary Vertices)
//
				LPointSelectedUnselected += _MeshSelectIntermediatePoints(LMesh, TRUE);
//printf("%s() %s:%d %d\n", __FUNCTION__, __FILE__, __LINE__, LPointSelectedUnselected);fflush(stdout);
			}
			else if(EObject_IsSubClass(LObject, E3dSpline))
			{
printf("%s() %s:%d  %sIMPLEMENT ME%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
			}

			LNTotalSelectedUnselected += LPointSelectedUnselected;
		}

		EFree(LObjects);
	}


	if(LNTotalSelectedUnselected) LSceneChangedFlags |= E3dChgSELECTION;
	else E3dp_PrintMessage(ELogMESSAGE, 5000, "No points were selected");

//{ char  s[256];printf("%s() %s:%d  %d %s\n", __FUNCTION__, __FILE__, __LINE__, LNTotalSelectedUnselected, E3d_SceneChangeFlagsToString(LSceneChangedFlags, s));fflush(stdout); }

	return(LSceneChangedFlags);
}


static EChangeFlags _Select_PolygonsWithASelectedEdge(E3dScene* PScene)
{
	EIndex        LNTotalSelectedUnselected = 0;
	EChangeFlags  LSceneChangedFlags = 0;

	E3dObject**  LObjects;EIndex  LObjects_Count = E3dScene_GetObjects(PScene, &LObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LObjects_Count)
	{
		EIndex  LObjC = 0;for(;LObjC < LObjects_Count;LObjC++)
		{
			E3dObject*  LObject = LObjects[LObjC];

			if(EObject_IsSubClass(LObject, E3dMesh))
			{
				E3dMesh*  LMesh = (E3dMesh*)LObject;
				E3dMesh_CreateWingedEdgesFromEdges(LMesh, FALSE, FALSE);
				EIndex  LEC = 0;const EIndex  LEN = LMesh->Edges_Count;const E3dEdge*  LEdge = LMesh->Edges;const E3dWingedEdge*  LWingedEdge = LMesh->WingedEdges;for(;LEC < LEN;LEC++, LEdge++, LWingedEdge++)
				{
					if(LEdge->Flags & E3dEdgeSELECTED)
					{
						EIndex  LPC = 0;const EIndex  LPN = LWingedEdge->Polygons_Count;E3dEdgePolygonInfo*  LPolygonInfo = LWingedEdge->Polygons;for(;LPC < LPN;LPC++, LPolygonInfo++)
						{
							E3dPolygon*  LPolygon = LPolygonInfo->Polygon;
							if((LPolygon->Flags & E3dPolygonSELECTED) == 0)
							{
								LPolygon->Flags |= E3dPolygonSELECTED;
								LNTotalSelectedUnselected++;
							}
						}
					}
				}
			}
		}

		EFree(LObjects);
	}

	if(LNTotalSelectedUnselected > 0) LSceneChangedFlags |= E3dChgSELECTION;

	return(LSceneChangedFlags);
}


//========================================
// Polygon selecting/unselecting
//========================================
static EChangeFlags _PointSelection(E3dScene* PScene, const int PAction)
{
	EIndex       LNTotalSelectedUnselected = 0;
	E3dObject**  LObjects;EIndex  LObjects_Count = E3dScene_GetObjects(PScene, &LObjects, E3dItemALL, E3dSELECTED, FALSE);if(LObjects_Count)
	{
		EIndex  LObjC = 0;for(;LObjC < LObjects_Count;LObjC++)
		{
			E3dObject*  LObject = LObjects[LObjC];
			EIndex      LPointSelectedUnselected = 0;

			if(EObject_IsSubClass(LObject, E3dMesh))
			{
				E3dMesh*    LMesh = (E3dMesh*)LObject;
				E3dVertex*  LVertices = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;

				switch(PAction)
				{
					case E3dSELECT_POINTS_OF_SELECTION:
					case E3dUNSELECT_POINTS_OF_SELECTION:
						{
							E3dPolyGroup**  LPolyGroups = LMesh->PolyGroups;
							const EIndex    LPGN = LMesh->PolyGroups_Count;
							const EBool     LSelect = PAction == E3dSELECT_POINTS_OF_SELECTION;

// If the whole Mesh is selected, check for selected Polygons
//
							if(LObject->Selection != E3dObjectSelPOLYGROUP)
							{
								EBool  LHaveSelectedPolygons = FALSE;
								{
									EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
									{
//printf("%s() %s:%d <<%s>>\n", __FUNCTION__, __FILE__, __LINE__, LPolyGroups[LPGC]->Name);fflush(stdout);
									 if(E3dPolyGroup_HasSelectedPolygons(LPolyGroups[LPGC])) { LHaveSelectedPolygons = TRUE;break; }
									}
								}

								if(LHaveSelectedPolygons)
								{
//printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
									EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
									{
										if(LSelect) LPointSelectedUnselected += _TagVerticesOfSelectedPolygons(LPolyGroups[LPGC], FALSE, LVertices, LVertexSize);
										else LPointSelectedUnselected += _UnTagVerticesOfSelectedPolygons(LPolyGroups[LPGC], FALSE, LVertices, LVertexSize);
									}
								}
								else
								{
// No selected Polygons, select all Vertices of the Mesh
//
									if(LSelect)
									{
										EIndex  LVC = 0;const EIndex  LVN = LMesh->Vertices_Count;E3dVertex*  LVertex = LVertices;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LVertexSize))
										{
											if((LVertex->Flags & E3dVtxSELECTED) == 0) { LVertex->Flags |= E3dVtxSELECTED;LPointSelectedUnselected++; }
										}
									}
									else
									{
										EIndex  LVC = 0;const EIndex  LVN = LMesh->Vertices_Count;E3dVertex*  LVertex = LVertices;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LVertexSize))
										{
										 if(LVertex->Flags & E3dVtxSELECTED) { LVertex->Flags &= ~E3dVtxSELECTED;LPointSelectedUnselected++; }
										}
									}
								}
							}
							else
							{
								EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
								{
									E3dPolyGroup*  LPolyGroup = LPolyGroups[LPGC];if(LPolyGroup->Selected)
									{
										if(LSelect) LPointSelectedUnselected += _TagVerticesOfSelectedPolygons(LPolyGroups[LPGC], TRUE, LVertices, LVertexSize);
										else LPointSelectedUnselected += _UnTagVerticesOfSelectedPolygons(LPolyGroups[LPGC], TRUE, LVertices, LVertexSize);
									}
								}
							}
						}
					break;

					case E3dSELECT_UNUSED_POINTS:
						{
							E3dMesh_GetVertexUsageList(LMesh, FALSE, TRUE);

							EIndex  LVC = 0;const EIndex  LVN = LMesh->Vertices_Count;E3dVertex*  LVertex = LVertices;const E3dVertexUsage*  LVertexUsage = LMesh->VertexUsageList;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LVertexSize), LVertexUsage++)
							{
								if(LVertexUsage->EdgeEnds == 0 && (LVertex->Flags & E3dVtxSELECTED) == 0)
								{
									LVertex->Flags |= E3dVtxSELECTED;
									LNTotalSelectedUnselected++;
								}
							}
						}
					break;
				}
			}
			else if(EObject_IsSubClass(LObject, E3dFace))
			{
				E3dFace*  LFace = (E3dFace*)LObject;

				switch(PAction)
				{
					case E3dSELECT_POINTS_OF_SELECTION:
					case E3dUNSELECT_POINTS_OF_SELECTION:
						{
							const EBool  LSelect = PAction == E3dSELECT_POINTS_OF_SELECTION;

							if(LSelect) LPointSelectedUnselected += _SplineTagCVs(LFace->Exterior);
							else LPointSelectedUnselected += _SplineUntagCVs(LFace->Exterior);
							EIndex  LHC = 0;const EIndex  LHN = LFace->Holes_Count;for(;LHC < LHN;LHC++)
							{
								if(LSelect) LPointSelectedUnselected += _SplineTagCVs(LFace->Holes[LHC]);
								else LPointSelectedUnselected += _SplineUntagCVs(LFace->Holes[LHC]);
							}
						}
					break;
				}
			}
			else if(EObject_IsSubClass(LObject, E3dSpline))
			{
				switch(PAction)
				{
					case E3dSELECT_POINTS_OF_SELECTION:
					case E3dUNSELECT_POINTS_OF_SELECTION:
						{
							const EBool  LSelect = PAction == E3dSELECT_POINTS_OF_SELECTION;
							if(LSelect) LPointSelectedUnselected += _SplineTagCVs((E3dSpline*)LObject);
							else LPointSelectedUnselected += _SplineUntagCVs((E3dSpline*)LObject);
						}
					break;
				}
			}
			else if(EObject_IsSubClass(LObject, E3dPointSet))
			{
printf("%s() %s:%d %sIMPLEMENT ME%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
			}

			if(LPointSelectedUnselected)
			{
				E3dSceneObjectCallbackStruct  LCBS;

				LCBS.Reason = E3dCR_POINTS_SELECTION;
				LCBS.Object = LObject;
				LCBS.SubObject = NULL;
				E3dScene_CallCallbacks((ECoreObject*)PScene,  PScene->Callbacks, PScene->Callbacks_Count,  &LCBS);
			}

			LNTotalSelectedUnselected += LPointSelectedUnselected;
		}

		EFree(LObjects);
	}

	if(LNTotalSelectedUnselected) return(E3dChgSELECTION);
	else return(0);
}


static EChangeFlags _Unselect_Points(E3dScene* PScene, EPointer POptions)
{
	E3dScene*           LScene = E3d_Scene;
	const EChangeFlags  LSceneChangeFlags = _PointSelection(LScene, E3dUNSELECT_POINTS_OF_SELECTION);

	if(LSceneChangeFlags) return(LSceneChangeFlags);
	else E3dp_PrintMessage(ELogMESSAGE, 5000, "No points were selected or unselected");
	return(0);
}


//========================================
//========================================
static void _MCB_Select_IntermediatePoints(Widget PW, EPointer PClientData, EPointer PCallData)
{
	const EChangeFlags  LSceneChangedFlags = _Select_IntermediatePoints(E3d_Scene);
	if(LSceneChangedFlags) E3dWindows_RedrawByChangeFlags(LSceneChangedFlags);
}


//========================================
// Polygon selecting/unselecting
//========================================
static void _MCB_PointSelection(Widget PW, EPointer PClientData, EPointer PCallData)
{
	const int           LAction = (EPointerInt)PClientData;
	E3dScene*           LScene = E3d_Scene;
	const EChangeFlags  LSceneChangeFlags = _PointSelection(LScene, LAction);

	if(LSceneChangeFlags) E3dWindows_RedrawByChangeFlags(LSceneChangeFlags);
	else E3dp_PrintMessage(ELogMESSAGE, 5000, "No points were selected or unselected");
}


static const uint8_t* _GetVertexAttribute(const E3dPolyGroup* PPolyGroup, const E3dPolygon* PPolygon, const char* PSemantic, const EIndex PMeshVertexIndex)
{
	const int  LVAOffset = E3dPolyGroup_VertexAttributeOffset(PPolyGroup, PSemantic, 0);if(LVAOffset < 0) return(NULL);
	const E3dVertexNodeSize  LVertexNodeSize = PPolyGroup->VertexNodeSize;
	EIndex  LVC = 0;const EIndex  LVN = PPolygon->NumOfVertexNodes;const E3dVertexNode*  LVertexNode = PPolygon->VertexNodes;for(;LVC < LVN;LVC++, EStruct_Inc(LVertexNode, LVertexNodeSize))
	{
		if(LVertexNode->VertexID == PMeshVertexIndex)
		{
			return(((const uint8_t*)LVertexNode + LVAOffset));
		}
	}
	return(NULL);
}


//========================================
// Dialog box callback
//========================================
static void _DCB_Close(Widget PW, EPointer PClientData, EPointer PCallData)
{
	Xe_WidgetUnmap(_SelectEdgesAtNormalCreasesDialogW);

	E_PluginLockDel(_Plugin, &_SelectEdgesAtNormalCreasesDialogLock, FALSE);
}


static int _DialogLockDel(EPlugin* PPlugin, EPluginLock* PLock)
{
	Xe_WidgetUnmap(_SelectEdgesAtNormalCreasesDialogW);

	return(E_SUCCESS);
}


enum
{
	E3dSelectEdgesAtNormalCreasesVERTEX_NORMAL = 1,
	E3dSelectEdgesAtNormalCreasesPOLYGON_NORMAL,
};

static E3dCoordinate  _CreaseAngle = 10.0;
static int            _SelectEdgesAtNormalCreases_NormalMode = E3dSelectEdgesAtNormalCreasesVERTEX_NORMAL;


static void _SelectEdgesAtNormalCreases(const int PMode, const float PCreaseAngle)
{
	E3dScene*    LScene = E3d_Scene;
	E3dObject**  LObjects;EIndex  LObjects_Count = E3dScene_GetObjects(LScene, &LObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LObjects_Count)
	{
		EIndex  LAllEdgeSelectionChanged = 0;

		EIndex  LObjC = 0;for(;LObjC < LObjects_Count;LObjC++)
		{
			E3dObject*  LObject = LObjects[LObjC];

			if(EObject_IsSubClass(LObject, E3dMesh))
			{
				EIndex        LThisMeshEdgeSelectionChanged = 0;
				E3dMesh*      LMesh = (E3dMesh*)LObject;E3dEdge*  LMeshEdges = LMesh->Edges;
				const EIndex  LEN = LMesh->Edges_Count;

				if(LMesh->WingedEdges == NULL) E3dMesh_CreateWingedEdgesFromEdges(LMesh, TRUE, FALSE);
				if(LMesh->WingedEdges)
				{
					const float  LCreaseAngleCos = E3d_Cos(PCreaseAngle * E3dDEGREES_TO_RADIANS);

// Using the temporary flag E3dEdgeMARK0 allows us to track if we selected any Edges, while keeping the selection non-additive
//
					{ EIndex  LEC = 0;E3dEdge*  LEdge = LMeshEdges;for(;LEC < LEN;LEC++, LEdge++) LEdge->Flags &= ~E3dEdgeMARK0; }

					{
						EIndex  LEC = 0;E3dEdge*  LEdge = LMeshEdges;const E3dWingedEdge*  LWingedEdge = LMesh->WingedEdges;for(;LEC < LEN;LEC++, LEdge++, LWingedEdge++)
						{
							if(LWingedEdge->Polygons_Count > 0)
							{
								if(LWingedEdge->Polygons_Count > 1)
								{
									const E3dPolygon*  LPolygon0 = LWingedEdge->Polygons[0].Polygon;const E3dPolyGroup*  LPolyGroup0 = LMesh->PolyGroups[LWingedEdge->Polygons[0].PolyGroupIndex];
									const E3dPolygon*  LPolygon1 = LWingedEdge->Polygons[1].Polygon;const E3dPolyGroup*  LPolyGroup1 = LMesh->PolyGroups[LWingedEdge->Polygons[1].PolyGroupIndex];

									switch(_SelectEdgesAtNormalCreases_NormalMode)
									{
										case E3dSelectEdgesAtNormalCreasesPOLYGON_NORMAL:
											{
												if(E3d_Abs(E3d_Dot3_vv(LPolygon0->Normal, LPolygon1->Normal)) < LCreaseAngleCos)
												{
													if((LEdge->Flags & E3dEdgeMARK0) == 0)
													{
														LEdge->Flags |= E3dEdgeMARK0;
														LThisMeshEdgeSelectionChanged++;
													}
												}
											}
										break;

										case E3dSelectEdgesAtNormalCreasesVERTEX_NORMAL:
											{
												const EIndex       LEdgeStartVertexIndex = LEdge->Start, LEdgeEndVertexIndex = LEdge->End;
												const E3dCoord3*  L0SNormal = (const E3dCoord3*)_GetVertexAttribute(LPolyGroup0, LPolygon0, E3dVtxAttribute_Normal, LEdgeStartVertexIndex);
												const E3dCoord3*  L0ENormal = (const E3dCoord3*)_GetVertexAttribute(LPolyGroup0, LPolygon0, E3dVtxAttribute_Normal, LEdgeEndVertexIndex);
												const E3dCoord3*  L1SNormal = (const E3dCoord3*)_GetVertexAttribute(LPolyGroup1, LPolygon1, E3dVtxAttribute_Normal, LEdgeStartVertexIndex);
												const E3dCoord3*  L1ENormal = (const E3dCoord3*)_GetVertexAttribute(LPolyGroup1, LPolygon1, E3dVtxAttribute_Normal, LEdgeEndVertexIndex);

												if(L0SNormal && L0ENormal && L1SNormal && L1ENormal)
												{
													if(E3d_Abs(E3d_Dot3_pp(L0SNormal, L1SNormal)) < LCreaseAngleCos || E3d_Abs(E3d_Dot3_pp(L0ENormal, L1ENormal)) < LCreaseAngleCos)
													{
														if((LEdge->Flags & E3dEdgeMARK0) == 0)
														{
															LEdge->Flags |= E3dEdgeMARK0;
															LThisMeshEdgeSelectionChanged++;
														}
													}
												}
												else  // One or both Polygons are in a PolyGroup with no vertex normals
												{
													if((LEdge->Flags & E3dEdgeMARK0) == 0)
													{
														LEdge->Flags |= E3dEdgeMARK0;
														LThisMeshEdgeSelectionChanged++;
													}
												}
											}
										break;
									}
								}
								else  // Boundary Edge
								{
									if((LEdge->Flags & E3dEdgeMARK0) == 0)
									{
										LEdge->Flags |= E3dEdgeMARK0;
										LThisMeshEdgeSelectionChanged++;
									}
								}
							}
							else  // Bad WingedEdge. No Polygons.
							{
							}
						}
					}

					{
						EIndex  LEC = 0;E3dEdge*  LEdge = LMeshEdges;for(;LEC < LEN;LEC++, LEdge++)
						{
							if(LEdge->Flags & E3dEdgeMARK0)
							{
								if((LEdge->Flags & E3dEdgeSELECTED) == 0) { LEdge->Flags |= E3dEdgeSELECTED;LThisMeshEdgeSelectionChanged++; }
							}
							else
							{
								if(LEdge->Flags & E3dEdgeSELECTED) { LEdge->Flags &= ~E3dEdgeSELECTED;LThisMeshEdgeSelectionChanged++; }
							}
						}
					}
				}

				if(LThisMeshEdgeSelectionChanged)
				{
//printf("%s() %s:%d <%s> %d\n", __FUNCTION__, __FILE__, __LINE__, LObject->Name, (int)LThisMeshEdgeSelectionChanged);fflush(stdout);
					LMesh->EdgeFlags |= E3dSELECTED_EDGES;
					E3dMesh_EdgeFlagsUpdate(LMesh);
					E3dObject_UpdateForDisplay((E3dObject*)LMesh, ~(E3dUF_EDGES|E3dUF_REMOVE_STRIPS|E3dUF_TRIANGULATE|E3dUF_CALL_OUTPUTS), LScene);
					LAllEdgeSelectionChanged++;
				}
			}
		}

		EFree(LObjects);

		if(LAllEdgeSelectionChanged)
		{
			E3dCallbackStruct  LCBS;
			LCBS.Reason = E3dCR_SELECTION_CHANGED;
			E3dScene_CallCallbacks((ECoreObject*)LScene,  LScene->Callbacks, LScene->Callbacks_Count,  &LCBS);

			E3dWindows_RedrawByChangeFlags(E3dChgSELECTION);
		}
	}
}


static void _DCB_SelectEdgesAtNormalCreases_Angle(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_CreaseAngle = ((XeScaleCallbackStruct*)PCallData)->Value.f32;

	_SelectEdgesAtNormalCreases(_SelectEdgesAtNormalCreases_NormalMode, _CreaseAngle);
}


static void _DCB_SelectEdgesAtNormalCreases_Activate(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_SelectEdgesAtNormalCreases(_SelectEdgesAtNormalCreases_NormalMode, _CreaseAngle);
}


static void _DCB_SelectEdgesAtNormalCreases_NormalMode(Widget PW, EPointer PClientData, EPointer PCallData)
{
	const int  LMode = (EPointerInt)PClientData;

	if(_SelectEdgesAtNormalCreases_NormalMode != LMode)
	{
		_SelectEdgesAtNormalCreases_NormalMode = LMode;
		_SelectEdgesAtNormalCreases(_SelectEdgesAtNormalCreases_NormalMode, _CreaseAngle);
	}
}


//========================================================
// Pop up connect/disconnect settings dialog box
//========================================================
static void _SelectEdgesAtNormalCreasesPopupDialog(void)
{
	if(_SelectEdgesAtNormalCreasesDialogLock == NULL) _SelectEdgesAtNormalCreasesDialogLock = E_PluginLockAdd(_Plugin, _DialogLockDel);

	if(_SelectEdgesAtNormalCreasesDialogW == NULL)
	{
		Widget  LW;
		EArg    Args[256];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNallowShellResize, TRUE);
		EXeSetArg(XeNcreateButtons, FALSE);
		EXeSetArg(XeNworkAreaType, XeMATRIX);
		EXeSetArg(XeNwmDecorations, XeWMDecorCLOSE|XeWMDecorBORDER|XeWMDecorRESIZEH|XeWMDecorTITLE);
		EXeSetArg(XeNtitle, "Select Mesh Edges at creases");EXeSetArg(XeNiconName, "Edge selection");
		Widget  LDialogW = XeDialog_New(XeDIALOG_TEMPLATE, "Edge selection", E3dp_TopShellW,  Args, ArgCnt);if(LDialogW == NULL) return;

		XeAddCallback(LDialogW, XeNwmCallback, _DCB_Close, (EPointer)XeCR_CANCEL);

		_SelectEdgesAtNormalCreasesDialogW = LDialogW;

		Widget  LTopMatrixW = XeDialog_GetChild(LDialogW, XeDIALOG_WORK_AREA);
		EXeStart;
		EXeSetArg(XeNmarginSizeX, 6);EXeSetArg(XeNmarginSizeY, 4);
		Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

		EXeStart;
		EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 4);
		EXeSetArg(XeNorientation, XeHORIZONTAL);
		LW = XeScaleNew_Coord(LTopMatrixW, "Crease angle",  Args, ArgCnt);
		XeScale_SetValueAndRanges_f32(LW,  _CreaseAngle,  1.0, 90.0,  1.0, 90.0,  FALSE);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SelectEdgesAtNormalCreases_Angle,    (EPointer)NULL);
		XeAddCallback(LW, XeNactivateCallback,     _DCB_SelectEdgesAtNormalCreases_Activate, (EPointer)NULL);

		{
			EXeStart;
			EXeSetArg(XeNradioBehavior, TRUE);
			Widget LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

			EXeStart;
			XeWidget_ManagedNew("Normal source", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

			EXeStart;
			EXeSetArg(XeNmarginLeft, 8);
			EXeSetArg(XeNset, _SelectEdgesAtNormalCreases_NormalMode == E3dSelectEdgesAtNormalCreasesVERTEX_NORMAL);
			LW = XeWidget_ManagedNew("Vertex normals", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SelectEdgesAtNormalCreases_NormalMode, (EPointer)E3dSelectEdgesAtNormalCreasesVERTEX_NORMAL);

			EXeStart;
			EXeSetArg(XeNmarginLeft, 4);
			EXeSetArg(XeNset, _SelectEdgesAtNormalCreases_NormalMode == E3dSelectEdgesAtNormalCreasesPOLYGON_NORMAL);
			LW = XeWidget_ManagedNew("Polygon normals", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
			XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SelectEdgesAtNormalCreases_NormalMode, (EPointer)E3dSelectEdgesAtNormalCreasesPOLYGON_NORMAL);
		}

		XeShell_Raise(_SelectEdgesAtNormalCreasesDialogW);
		XeShell_CenterToWidget(_SelectEdgesAtNormalCreasesDialogW, E3dp_TopShellW);
	}
	else XeShell_Raise(_SelectEdgesAtNormalCreasesDialogW);

	_SelectEdgesAtNormalCreases(_SelectEdgesAtNormalCreases_NormalMode, _CreaseAngle);
}


//========================================
// Polygon selecting/unselecting
//========================================
static EBool _EdgeSelection(const int PAction)
{
	E3dScene*    LScene = E3d_Scene;
	E3dObject**  LObjects;EIndex  LObjects_Count = E3dScene_GetObjects(LScene, &LObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LObjects_Count)
	{
		EIndex  LAllEdgeSelectionChanged = 0;

		EIndex  LObjC = 0;for(;LObjC < LObjects_Count;LObjC++)
		{
			E3dObject*  LObject = LObjects[LObjC];

			if(EObject_IsSubClass(LObject, E3dMesh))
			{
				EIndex        LThisMeshEdgeSelectionChanged = 0;
				E3dMesh*      LMesh = (E3dMesh*)LObject;E3dEdge*  LMeshEdges = LMesh->Edges;
				const EIndex  LEN = LMesh->Edges_Count;

				switch(PAction)
				{
					case E3dCrSELECT_EDGES_WITH_SELECTED_VERTICES:
						{
							const E3dVertexSize  LVertexSize = LMesh->VertexSize;
							EIndex  LEC = 0;const E3dVertex*  LVertices = LMesh->Vertices;E3dEdge*  LEdge = LMeshEdges;for(;LEC < LEN;LEC++, LEdge++)
							{
								const E3dVertex*  LStartVertex = EStruct_ByIndex(LVertices, LVertexSize, LEdge->Start);
								const E3dVertex*  LEndVertex   = EStruct_ByIndex(LVertices, LVertexSize, LEdge->End);
								if(LStartVertex->Flags & E3dVtxSELECTED && LEndVertex->Flags & E3dVtxSELECTED)
								{
									if((LEdge->Flags & E3dEdgeSELECTED) == 0)
									{
										LEdge->Flags |= E3dEdgeSELECTED;
										LThisMeshEdgeSelectionChanged++;
									}
								}
							}
						}
					break;

					case E3dCrSELECT_EDGES_OF_SELECTED_POLYGONS:
						{
							const EIndex  LPGN = LMesh->PolyGroups_Count;
							EBool         LRebuildWingedEdges = FALSE;

							{
								EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
								{
									E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[LPGC];
									if(E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_EdgeIndex, 0,  NULL) != E_NOTHING_TO_DO) LRebuildWingedEdges = TRUE;
								}
							}

							if(LRebuildWingedEdges) E3dMesh_CreateWingedEdgesFromEdges(LMesh, TRUE, FALSE);
							if(LMesh->WingedEdges)
							{
								EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
								{
									const E3dPolyGroup*      LPolyGroup = LMesh->PolyGroups[LPGC];
									const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;
									const E3dVertexNodeSize  LEdgeIndexOffset = E3dPolyGroup_VertexAttributeOffset(LPolyGroup, E3dVtxAttribute_EdgeIndex, 0);

									if(LEdgeIndexOffset >= 0)
									{
										EIndex  LPC = 0;const EIndex  LPN = LPolyGroup->Polygons_Count;const E3dPolygon*  LPolygon = LPolyGroup->Polygons;for(;LPC < LPN;LPC++, LPolygon++)
										{
											if(LPolygon->Flags & E3dPolygonSELECTED)
											{
												EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertexNodes;const E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;for(;LVC < LVN;LVC++, EStruct_Inc(LVertexNode, LVertexNodeSize))
												{
													if(LVertexNode->VertexID >= 0)
													{
														const E3dEdgeIndex  LEdgeIndexV = *((E3dEdgeIndex*)((const uint8_t*)LVertexNode + LEdgeIndexOffset)), LEdgeIndex = LEdgeIndexV & E3dVtxNodeFlagEDGE_INDEX_MASK;
														E3dEdge*            LEdge = LMeshEdges + LEdgeIndex;

														if((LEdge->Flags & E3dEdgeSELECTED) == 0)
														{
															LEdge->Flags |= E3dEdgeSELECTED;
															LThisMeshEdgeSelectionChanged++;
														}
													}
												}
											}
										}
									}
								}
							}
						}
					break;

					case E3dCrUNSELECT_EDGES:
						{
							EIndex  LEC = 0;E3dEdge*  LEdge = LMeshEdges;for(;LEC < LEN;LEC++, LEdge++)
							{
								if(LEdge->Flags & E3dEdgeSELECTED)
								{
									LEdge->Flags &= ~E3dEdgeSELECTED;
									LThisMeshEdgeSelectionChanged++;
								}
							}
						}
					break;

					case E3dCrSELECT_EDGES_AT_NORMAL_CREASES:
						_SelectEdgesAtNormalCreasesPopupDialog();

/*
						{
							if(LMesh->WingedEdges == NULL) E3dMesh_CreateWingedEdgesFromEdges(LMesh, TRUE, FALSE);if(LMesh->WingedEdges)
							{
								const float  LCreaseAngle = 10.0, LCreaseAngleCos = E3d_Cos(LCreaseAngle * E3dDEGREES_TO_RADIANS);
								EIndex  LEC = 0;E3dEdge*  LEdge = LMeshEdges;const E3dWingedEdge*  LWingedEdge = LMesh->WingedEdges;for(;LEC < LEN;LEC++, LEdge++, LWingedEdge++)
								{
									if(LWingedEdge->Polygons_Count > 0)
									{
										if(LWingedEdge->Polygons_Count > 1)
										{
											const EIndex       LEdgeStartVertexIndex = LEdge->Start, LEdgeEndVertexIndex = LEdge->End;
											const E3dPolygon*  LPolygon0 = LWingedEdge->Polygons[0].Polygon;const E3dPolyGroup*  LPolyGroup0 = LMesh->PolyGroups[LWingedEdge->Polygons[0].PolyGroupIndex];
											const E3dPolygon*  LPolygon1 = LWingedEdge->Polygons[1].Polygon;const E3dPolyGroup*  LPolyGroup1 = LMesh->PolyGroups[LWingedEdge->Polygons[1].PolyGroupIndex];

											const E3dCoord3*  L0SNormal = (const E3dCoord3*)_GetVertexAttribute(LPolyGroup0, LPolygon0, E3dVtxAttribute_Normal, LEdgeStartVertexIndex);
											const E3dCoord3*  L0ENormal = (const E3dCoord3*)_GetVertexAttribute(LPolyGroup0, LPolygon0, E3dVtxAttribute_Normal, LEdgeEndVertexIndex);
											const E3dCoord3*  L1SNormal = (const E3dCoord3*)_GetVertexAttribute(LPolyGroup1, LPolygon1, E3dVtxAttribute_Normal, LEdgeStartVertexIndex);
											const E3dCoord3*  L1ENormal = (const E3dCoord3*)_GetVertexAttribute(LPolyGroup1, LPolygon1, E3dVtxAttribute_Normal, LEdgeEndVertexIndex);
											if(L0SNormal && L0ENormal && L1SNormal && L1ENormal)
											{
												if(E3d_Abs(E3d_Dot3_pp(L0SNormal, L1SNormal)) < LCreaseAngleCos || E3d_Abs(E3d_Dot3_pp(L0ENormal, L1ENormal)) < LCreaseAngleCos)
												{
													if((LEdge->Flags & E3dEdgeSELECTED) == 0)
													{
														LEdge->Flags |= E3dEdgeSELECTED;
														LThisMeshEdgeSelectionChanged++;
													}
												}
											}
											else  // One or both Polygons are in a PolyGroup with no normals
											{
												if((LEdge->Flags & E3dEdgeSELECTED) == 0)
												{
													LEdge->Flags |= E3dEdgeSELECTED;
													LThisMeshEdgeSelectionChanged++;
												}
											}
										}
										else  // Boundary Edge
										{
											if((LEdge->Flags & E3dEdgeSELECTED) == 0)
											{
												LEdge->Flags |= E3dEdgeSELECTED;
												LThisMeshEdgeSelectionChanged++;
											}
										}
									}
									else  // Bad WingedEdge. No Polygons.
									{
									}
								}
							}
						}
*/
					break;

					case E3dCrSELECT_BOUNDARY_EDGES:
printf("%s() %s:%d  %sIMPLEMENT ME%s select boundary edges of selected Polygons, or the selected Mesh or PolyGroup\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
						{
							if(LMesh->WingedEdges == NULL) E3dMesh_CreateWingedEdgesFromEdges(LMesh, TRUE, FALSE);
							if(LMesh->WingedEdges)
							{
								EIndex  LEC = 0;E3dEdge*  LEdge = LMeshEdges;const E3dWingedEdge*  LWingedEdge = LMesh->WingedEdges;for(;LEC < LEN;LEC++, LEdge++, LWingedEdge++)
								{
									if(LWingedEdge->Polygons_Count > 0)
									{
										if(LWingedEdge->Polygons_Count > 1)
										{
											const E3dPolygon*  LPolygon0 = LWingedEdge->Polygons[0].Polygon;const E3dPolyGroup*  LPolyGroup0 = LMesh->PolyGroups[LWingedEdge->Polygons[0].PolyGroupIndex];
											const E3dPolygon*  LPolygon1 = LWingedEdge->Polygons[1].Polygon;const E3dPolyGroup*  LPolyGroup1 = LMesh->PolyGroups[LWingedEdge->Polygons[1].PolyGroupIndex];

											if((LPolygon0->Flags & E3dPolygonSELECTED) != (LPolygon1->Flags & E3dPolygonSELECTED) || LPolyGroup0 != LPolyGroup1)
											{
												if((LEdge->Flags & E3dEdgeSELECTED) == 0)
												{
													LEdge->Flags |= E3dEdgeSELECTED;
													LThisMeshEdgeSelectionChanged++;
												}
											}
										}
										else  // Boundary Edge
										{
											if((LEdge->Flags & E3dEdgeSELECTED) == 0)
											{
												LEdge->Flags |= E3dEdgeSELECTED;
												LThisMeshEdgeSelectionChanged++;
											}
										}
									}
									else  // Bad WingedEdge. No Polygons.
									{
									}
								}
							}
						}
					break;
				}

				if(LThisMeshEdgeSelectionChanged)
				{
					LMesh->EdgeFlags |= E3dSELECTED_EDGES;
					E3dMesh_EdgeFlagsUpdate(LMesh);
					E3dObject_UpdateForDisplay((E3dObject*)LMesh, ~(E3dUF_EDGES|E3dUF_REMOVE_STRIPS|E3dUF_TRIANGULATE|E3dUF_CALL_OUTPUTS), LScene);
				}

printf("%s() %s:%d <%s> %d\n", __FUNCTION__, __FILE__, __LINE__, LObject->Name, (int)LThisMeshEdgeSelectionChanged);fflush(stdout);
				if(LThisMeshEdgeSelectionChanged) LAllEdgeSelectionChanged += LThisMeshEdgeSelectionChanged;
			}
		}

		EFree(LObjects);

		if(LAllEdgeSelectionChanged)
		{
			E3dCallbackStruct  LCBS;
			LCBS.Reason = E3dCR_SELECTION_CHANGED;
			E3dScene_CallCallbacks((ECoreObject*)LScene,  LScene->Callbacks, LScene->Callbacks_Count,  &LCBS);

			return(TRUE);
		}
	}
	else
	{
		E3dp_PrintMessage(ELogMESSAGE, 5000, "Select a Mesh");return(FALSE);
	}
	return(FALSE);
}


//========================================
// Polygon selecting/unselecting
//========================================
static void _PolySelection(const int PAction)
{
	E3dScene*    LScene = E3d_Scene;
	E3dObject**  LObjects;EIndex  LObjects_Count = E3dScene_GetObjects(LScene, &LObjects, E3dItemALL, E3dSELECTED, FALSE);

	if(LObjects_Count)
	{
		EIndex  LTotalPolyNum = 0;

		EIndex  LObjC = 0;for(;LObjC < LObjects_Count;LObjC++)
		{
			E3dObject*  LObject = LObjects[LObjC];
			EBool       LPolygonSelectionChanged = FALSE, LPolyGroupSelectionChanged = FALSE, LPointSelectionChanged = FALSE;

			if(EObject_IsSubClass(LObject, E3dMesh))
			{
				E3dMesh*        LMesh = (E3dMesh*)LObject;

				E3dPolyGroup**  LPolyGroups = LMesh->PolyGroups;
				const EIndex    LPGN = LMesh->PolyGroups_Count;

				{
					E3dVertex*  LVertices = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;

					EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
					{
						E3dPolyGroup*  LPolyGroup = LPolyGroups[LPGC];

						if(PAction == E3dCrINVERT_POLYGROUP_SELECTION)
						{
							if(LPolyGroup->Selected)
							{
								LPolyGroup->Selected = FALSE;
							}
							else LPolyGroup->Selected = TRUE;
							LPolyGroupSelectionChanged = TRUE;
						}
						else if(LPolyGroup->Selected || LObject->Selection != E3dObjectSelPOLYGROUP)
						{
							EBool  LPGPolygonSelectionChanged = FALSE;

							LTotalPolyNum += LPolyGroup->Polygons_Count;
							const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;
							EIndex  LPolyCnt;

							switch(PAction)
							{
								case E3dCrPOLYGONS_WITH_ALL_VERTICES_SELECTED:
									if((LPolyCnt = LPolyGroup->Polygons_Count) > 0)
									{
										E3dPolygon*  LPolygon = LPolyGroup->Polygons;
										do
										{
											E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;if(LVertexNode)
											{
												EIndex  LNSelVs = 0;
												EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertexNodes;for(;LVC < LVN;LVC++)
												{
													if(LVertexNode->VertexID > -1)
													{
														const E3dVertex*  LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);
														if(LVertex->Flags & E3dVtxSELECTED) LNSelVs++;
													}
													EStruct_Inc(LVertexNode, LVertexNodeSize);
												}
												if(LNSelVs == LPolygon->NumOfVertices && (LPolygon->Flags & E3dPolygonSELECTED) == 0)
												{
													LPolygon->Flags |= E3dPolygonSELECTED;
													LPGPolygonSelectionChanged = TRUE;
												}
											}
											LPolygon++;
										} while(--LPolyCnt);
										if(LPGPolygonSelectionChanged) LPolygonSelectionChanged = TRUE;
									}
								break;

								case E3dCrPOLYGONS_WITH_A_SELECTED_EDGE:
									if(_Select_PolygonsWithASelectedEdge(LScene)) LPGPolygonSelectionChanged = TRUE;
								break;

								case E3dCrSELECT_TRIS:
									if((LPolyCnt = LPolyGroup->Polygons_Count) > 0)
									{
										E3dPolygon*  LPolygon = LPolyGroup->Polygons;
										do
										{
											if(LPolygon->NumOfVertices == 3 && (LPolygon->Flags & E3dPolygonSELECTED) == 0)
											{
												LPolygon->Flags |= E3dPolygonSELECTED;
												LPGPolygonSelectionChanged = TRUE;
											}
											LPolygon++;
										} while(--LPolyCnt);
										if(LPGPolygonSelectionChanged) LPolygonSelectionChanged = TRUE;
									}
								break;

								case E3dCrSELECT_QUADS:
									if((LPolyCnt = LPolyGroup->Polygons_Count) > 0)
									{
										E3dPolygon*  LPolygon = LPolyGroup->Polygons;
										do
										{
											if(LPolygon->NumOfVertices == 4 && (LPolygon->Flags & E3dPolygonSELECTED) == 0)
											{
												LPolygon->Flags |= E3dPolygonSELECTED;
												LPGPolygonSelectionChanged = TRUE;
											}
											LPolygon++;
										} while(--LPolyCnt);
										if(LPGPolygonSelectionChanged) LPolygonSelectionChanged = TRUE;
									}
								break;

								case E3dCrUNSELECT_ALL_POLYS:
									if((LPolyCnt = LPolyGroup->Polygons_Count) > 0)
									{
										E3dPolygon*  LPolygon = LPolyGroup->Polygons;
										do
										{
											if(LPolygon->Flags & E3dPolygonSELECTED)
											{
												LPolygon->Flags &= ~E3dPolygonSELECTED;
												LPGPolygonSelectionChanged = TRUE;
											}
											LPolygon++;
										} while(--LPolyCnt);
										if(LPGPolygonSelectionChanged) LPolygonSelectionChanged = TRUE;
									}
								break;

								case E3dCrINVERT_POLYGON_SELECTION:
									if((LPolyCnt = LPolyGroup->Polygons_Count) > 0)
									{
										E3dPolygon*  LPolygon = LPolyGroup->Polygons;
										do
										{
											LPolygon->Flags ^= E3dPolygonSELECTED;
											LPGPolygonSelectionChanged = TRUE;
											LPolygon++;
										} while(--LPolyCnt);
										LPolygonSelectionChanged = TRUE;
									}
								break;
							}

							if(LPGPolygonSelectionChanged)
							{
								E3dSceneObjectCallbackStruct  LCBS;
								LCBS.Reason = E3dCR_POLYGON_SELECTION;
								LCBS.Object = LObject;
								LCBS.SubObject = (EObject*)LPolyGroup;
								E3dScene_CallCallbacks((ECoreObject*)LScene,  LScene->Callbacks, LScene->Callbacks_Count,  &LCBS);
							}
						}
					}
				}

				if(LPolyGroupSelectionChanged)
				{
// Set the Mesh's Selection mode, based on whether any PolyGroups are selected
//
					EBool  LPolyGroupSelected = FALSE;
					EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
					{
						E3dPolyGroup*  LPolyGroup = LPolyGroups[LPGC];
						if(LPolyGroup->Selected) { LPolyGroupSelected = TRUE;break; }
					}
					if(LPolyGroupSelected)
					{
						LMesh->Selection = E3dObjectSelPOLYGROUP;
					}
					else LMesh->Selection = E3dObjectSelNONE;
				}
			}

			if(LPointSelectionChanged)
			{
				E3dSceneObjectCallbackStruct  LCBS;

				LCBS.Reason = E3dCR_POINTS_SELECTION;
				LCBS.Object = LObject;
				LCBS.SubObject = NULL;
				E3dScene_CallCallbacks((ECoreObject*)LScene,  LScene->Callbacks, LScene->Callbacks_Count,  &LCBS);
			}

			(void)LPolygonSelectionChanged;
		}

		EFree(LObjects);
	}
	else E3dp_PrintMessage(ELogMESSAGE, 5000, "Select something");
}


//========================================
// Menu callback
//========================================
static void _MCB_NodeSelection(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_NodeSelection((EPointerInt)PClientData);
	E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	E3dWindows_RedrawByChangeFlags(E3dChgSELECTION);
}


//========================================
// Menu callback
//========================================
static void _MCB_SelectHidden(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_SelectHidden((EPointerInt)PClientData);
	E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	E3dWindows_RedrawByChangeFlags(E3dChgSELECTION);
}


//========================================
// Menu callback
//========================================
static void _MCB_SelectObjects(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;

	E3dSel_SceneUnselectAll(LScene);

	E3dObject**  LObjects;EIndex  LObjects_Count = E3dScene_GetObjects(LScene, &LObjects, (EPointerInt)PClientData, 0, FALSE);if(LObjects_Count)
	{
		EIndex  LObjC = 0;for(;LObjC < LObjects_Count;LObjC++)
		{
			E3dSel_ObjectSelect(LScene, LObjects[LObjC], FALSE, FALSE);
		}

		if(LObjects_Count == 1) E3dp_PrintMessage(ELogMESSAGE, 5000, "Selected <%s>", LObjects[0]->Name ? LObjects[0]->Name : "");
		else E3dp_PrintMessage(ELogMESSAGE, 5000, "Selected %lld Objects", (long long)LObjects_Count);

		EFree(LObjects);
	}


	E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	E3dWindows_RedrawByChangeFlags(E3dChgSELECTION);
}


//========================================
// Menu callback
//========================================
static void _MCB_EdgeSelection(Widget PW, EPointer PClientData, EPointer PCallData)
{
	if(_EdgeSelection((EPointerInt)PClientData)) E3dWindows_RedrawByChangeFlags(E3dChgSELECTION);
}


//========================================
// Menu callback
//========================================
static void _MCB_PolySelection(Widget PW, EPointer PClientData, EPointer PCallData)
{
	_PolySelection((EPointerInt)PClientData);
	E3dWindows_RedrawByChangeFlags(E3dChgSELECTION);
}


//========================================
// Key callback	
//========================================
static void _KCB_Selection(EKeyCode LKey, unsigned int LModifiers, int LKeyEventType, EPointer PClientData)
{
	_NodeSelection((EPointerInt)PClientData);
	E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	E3dWindows_RedrawByChangeFlags(E3dChgSELECTION);
}


//========================================
// Key callback
//========================================
static void _KCB_HrcWalk(const EKeyCode PKey, const unsigned int PModifiers, const int PKeyEventType, EPointer PClientData)
{
	E3dScene*     LScene = E3d_Scene;
	E3dNode**     LRootNodes = LScene->RootNodes;
	EChangeFlags  LChanged = 0;

	EIndex  LRC = 0;const EIndex  LRN = LScene->RootNodes_Count;for(;LRC < LRN;LRC++)
	{
		E3dNode*  LRootNode = LRootNodes[LRC];
		E3dNode*  LNode = LRootNode;for(;LNode;LNode = LNode->Next)
		{
			if(LNode->Selection != E3dNodeSelNONE)
			{
				switch((EPointerInt)PClientData)
				{
					case E3dCrGO_UP:
						if(LNode->Selection == E3dNodeSelOBJECT)
						{
// If an Object is selected and it's only instantiated once, we can switch to Node selection mode.
//
printf("%s() %s:%d %sUP%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout);

							int  LNodesSelected = 0;

							E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
							{
								EIndex  LC = 0;const EIndex  LN = LModel->Objects_Count;E3dObject**  LObjects = LModel->Objects;for(;LC < LN;LC++)
								{
									const E3dObject*  LObject = LObjects[LC];if(LObject->Selection)
									{
										switch(LObject->Selection)
										{
											case E3dObjectSelOBJECT:
												if(LObject->Models_Count < 2)
												{
													E3dSel_NodeSelect(LScene, LNode, TRUE, TRUE);
													LChanged |= E3dChgSELECTION;
												}
												else
												{
// LObject instantiated by multiple Models. Select all of them.
//
													E3dSel_NodeSelect(LScene, LNode, TRUE, FALSE);
													EIndex  LIC = 0;for(;LIC < LObject->Models_Count;LIC++)
													{
														if(LObject->Models[LIC] != (E3dModel*)LNode) { E3dSel_NodeSelect(LScene, LNode, FALSE, FALSE);LChanged |= E3dChgSELECTION; }
													}
													E3dp_PrintMessage(ELogMESSAGE, 2000, "Object \"%s\" is instantiated by multiple Nodes (it's not clear which one to select).", LObject->Name ? LObject->Name : "");
												}
												LNodesSelected = LObject->Models_Count;
											break;

											case E3dObjectSelPOLYGROUP:
												{
													E3dMesh*  LMesh = E_DynamicCast(LObject, E3dMesh);if(LMesh)
													{
// If a PolyGroup is selected going up means selecting the Mesh
//
														LMesh->Selection = 0;  // We must set this to 0, so E3dSel_ObjectSelect will work properly

												    EIndex  LPGC = 0;const EIndex  LPGN = LMesh->PolyGroups_Count;for(;LPGC < LPGN;LPGC++)
												    {
												      E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[LPGC];
												      LPolyGroup->Selected = FALSE;
												    }

														E3dSel_ObjectSelect(LScene, LObjects[LC], TRUE, TRUE);
														LChanged |= E3dChgSELECTION;
													}
												}
											break;
										}
										break;
									}
								}
							}

							if(LNodesSelected)
							{
								if(LNodesSelected == 1) E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected Node: \"%s\"", LNode->Name ? LNode->Name : "");
								else E3dp_PrintMessage(ELogMESSAGE, 2000, "Multiple Nodes selected");
							}
						}
						else
						{
							E3dNode*  LTNode = LNode->Parent;if(LTNode)
							{
								int  LSelected = 0;
								switch(LNode->Selection)
								{
									case E3dNodeSelNODE:         E3dSel_NodeSelect(LScene, LTNode, TRUE, TRUE);LSelected = 1;break;
									case E3dNodeSelBRANCH_ROOT:  E3dSel_NodeSelectBranch(LScene, LTNode, TRUE, TRUE);LSelected = 2;break;
								}

								if(LSelected)
								{
									LChanged |= E3dChgSELECTION;
									if(LSelected == 1) E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected parent Node: \"%s\"", LTNode->Name ? LTNode->Name : "");
									else E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected parent Branch: \"%s\"", LTNode->Name ? LTNode->Name : "");
								}
							}
							else E3dp_PrintMessage(ELogWARNING, 2000, "Node \"%s\" has no parent.", LNode->Name ? LNode->Name : "");
						}
					break;

					case E3dCrGO_DOWN:
						if(LNode->Selection == E3dNodeSelOBJECT)
						{
							E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
							{
								EIndex  LC = 0;const EIndex  LN = LModel->Objects_Count;E3dObject**  LObjects = LModel->Objects;for(;LC < LN;LC++)
								{
									E3dObject*  LObject = LObjects[LC];if(LObject->Selection)
									{
										switch(LObject->Selection)
										{
											case E3dObjectSelOBJECT:
												{
													E3dMesh*  LMesh = E_DynamicCast(LObject, E3dMesh);if(LMesh)
													{
														if(LMesh->PolyGroups_Count)
														{
															E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[0];

															E3dSel_MeshPolyGroupSelect(LScene, LMesh, LPolyGroup, TRUE, TRUE);
															LChanged |= E3dChgSELECTION;
														}
													}
												}
											break;
										}
									}
								}
							}
						}
						else
						{
							E3dNode*  LTNode = LNode->Child;if(LTNode)
							{
								int  LSelected = 0;
								switch(LNode->Selection)
								{
									case E3dNodeSelNODE:         E3dSel_NodeSelect(LScene, LTNode, TRUE, TRUE);LSelected = 1;break;
									case E3dNodeSelBRANCH_ROOT:  E3dSel_NodeSelectBranch(LScene, LTNode, TRUE, TRUE);LSelected = 2;break;
								}

								if(LSelected)
								{
									LChanged |= E3dChgSELECTION;
									if(LSelected == 1) E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected child Node:  \"%s\"", LTNode->Name ? LTNode->Name : "");
									else E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected child Branch:  \"%s\"", LTNode->Name ? LTNode->Name : "");
								}
							}
							else
							{
								E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
								{
									if(LModel->Objects_Count)
									{
										E3dSel_ObjectSelect(LScene, LModel->Objects[0], TRUE, TRUE);
										LChanged |= E3dChgSELECTION;
										E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected Object:  \"%s\"", LModel->Objects[0]->Name ? LModel->Objects[0]->Name : "");
									}
								}
								else E3dp_PrintMessage(ELogWARNING, 2000, "Node \"%s\" has no child Node and no Objects.", LNode->Name ? LNode->Name : "");
							}
						}
					break;

					case E3dCrGO_LEFT:
						if(LNode->Selection == E3dNodeSelOBJECT)
						{
							E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
							{
								EIndex  LC = 0;const EIndex  LN = LModel->Objects_Count;E3dObject**  LObjects = LModel->Objects;for(;LC < LN;LC++)
								{
									E3dObject*  LObject = LObjects[LC];if(LObject->Selection)
									{
										switch(LObject->Selection)
										{
											case E3dObjectSelOBJECT:
												{
													if(LC > 0)
													{
														E3dSel_ObjectSelect(LScene, LObjects[LC - 1], TRUE, TRUE);
														LChanged |= E3dChgSELECTION;
													}
													else E3dp_PrintMessage(ELogWARNING, 2000, "Object \"%s\" has no next sibling.", LObject->Name ? LObject->Name : "");
												}
											break;

											case E3dObjectSelPOLYGROUP:
												{
													E3dMesh*  LMesh = E_DynamicCast(LObject, E3dMesh);if(LMesh)
													{
														const EIndex  LPGN = LMesh->PolyGroups_Count;if(LPGN > 1)
														{
															EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
															{
																if(LMesh->PolyGroups[LPGC]->Selected) break;
															}
															if(LPGC == LPGN);  // No PolyGroup selected
															else if(LPGC == 0)
															{
																E3dp_PrintMessage(ELogWARNING, 2000, "PolyGroup \"%s\" has no previous sibling.", LMesh->PolyGroups[LPGC]->Name ? LMesh->PolyGroups[LPGC]->Name : "");
															}
															else
															{
																E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[LPGC - 1];

																E3dSel_MeshPolyGroupSelect(LScene, LMesh, LPolyGroup, TRUE, TRUE);
																LChanged |= E3dChgSELECTION;
															}
														}
														else if(LPGN > 0) E3dp_PrintMessage(ELogWARNING, 2000, "PolyGroup \"%s\" has no next sibling.", LMesh->PolyGroups[0]->Name ? LMesh->PolyGroups[0]->Name : "");
														else E3dp_PrintMessage(ELogWARNING, 2000, "Mesh \"%s\" has no next PolyGroups.", LMesh->Name ? LMesh->Name : "");
													}
												}
											break;
										}
										break;
									}
								}
							}
						}
						else
						{
							E3dNode*  LTNode = LNode->PrevSibling;if(LTNode)
							{
								int  LSelected = 0;
								switch(LNode->Selection)
								{
									case E3dNodeSelNODE:         E3dSel_NodeSelect(LScene, LTNode, TRUE, TRUE);LSelected = 1;break;
									case E3dNodeSelBRANCH_ROOT:  E3dSel_NodeSelectBranch(LScene, LTNode, TRUE, TRUE);LSelected = 2;break;
								}

								if(LSelected)
								{
									LChanged |= E3dChgSELECTION;
									if(LSelected == 1) E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected prev sibling Node:  \"%s\"", LTNode->Name ? LTNode->Name : "");
									else E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected prev sibling Branch:  \"%s\"", LTNode->Name ? LTNode->Name : "");
								}
							}
							else
							{
								if(LNode == LRootNode)
								{
									if(LRC > 0)
									{
										int  LSelected = 0;
										LTNode = LScene->RootNodes[LRC - 1];
										switch(LNode->Selection)
										{
											case E3dNodeSelNODE:         E3dSel_NodeSelect(LScene, LTNode, TRUE, TRUE);LSelected = 1;break;
											case E3dNodeSelBRANCH_ROOT:  E3dSel_NodeSelectTree(LScene, LTNode, TRUE, TRUE);LSelected = 2;break;
										}

										if(LSelected)
										{
											LChanged |= E3dChgSELECTION;
											if(LSelected == 1) E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected previous hierarchy root Node:  \"%s\"", LTNode->Name ? LTNode->Name : "");
											else E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected previous hierarchy:  \"%s\"", LTNode->Name ? LTNode->Name : "");
										}
									}
								}
								else E3dp_PrintMessage(ELogWARNING, 2000, "Node \"%s\" has no previous hierarchy root.", LNode->Name ? LNode->Name : "");
							}
						}
					break;

					case E3dCrGO_RIGHT:
						if(LNode->Selection == E3dNodeSelOBJECT)
						{
							E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
							{
								EIndex  LC = 0;const EIndex  LN = LModel->Objects_Count;E3dObject**  LObjects = LModel->Objects;for(;LC < LN;LC++)
								{
									E3dObject*  LObject = LObjects[LC];if(LObject->Selection)
									{
										switch(LObject->Selection)
										{
											case E3dObjectSelOBJECT:
												if(LC < LN - 1)
												{
													E3dSel_ObjectSelect(LScene, LObjects[LC + 1], TRUE, TRUE);
													LChanged |= E3dChgSELECTION;
												}
												else E3dp_PrintMessage(ELogWARNING, 2000, "Object \"%s\" has no next sibling.", LObjects[LC]->Name);
											break;

											case E3dObjectSelPOLYGROUP:
												{
													E3dMesh*  LMesh = E_DynamicCast(LObject, E3dMesh);if(LMesh)
													{
														const EIndex  LPGN = LMesh->PolyGroups_Count;if(LPGN > 1)
														{
															EIndex  LPGC = 0;for(;LPGC < LPGN;LPGC++)
															{
																if(LMesh->PolyGroups[LPGC]->Selected) break;
															}
															if(LPGC == LPGN);  // No PolyGroup selected
															else if(LPGC == LPGN - 1)
															{
																E3dp_PrintMessage(ELogWARNING, 2000, "PolyGroup \"%s\" has no next sibling.", LMesh->PolyGroups[LPGC]->Name ? LMesh->PolyGroups[LPGC]->Name : "");
															}
															else
															{
																E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[LPGC + 1];

																E3dSel_MeshPolyGroupSelect(LScene, LMesh, LPolyGroup, TRUE, TRUE);
																LChanged |= E3dChgSELECTION;
															}
														}
														else if(LPGN > 0) E3dp_PrintMessage(ELogWARNING, 2000, "PolyGroup \"%s\" has no next sibling.", LMesh->PolyGroups[0]->Name ? LMesh->PolyGroups[0]->Name : "");
														else E3dp_PrintMessage(ELogWARNING, 2000, "Mesh \"%s\" has no next PolyGroups.", LMesh->Name ? LMesh->Name : "");
													}
												}
											break;
										}
										break;
									}
								}
							}
						}
						else
						{
							E3dNode*  LTNode = LNode->NextSibling;if(LTNode)
							{
								int  LSelected = 0;
								switch(LNode->Selection)
								{
									case E3dNodeSelNODE:         E3dSel_NodeSelect(LScene, LTNode, TRUE, TRUE);LSelected = 1;break;
									case E3dNodeSelBRANCH_ROOT:  E3dSel_NodeSelectBranch(LScene, LTNode, TRUE, TRUE);LSelected = 2;break;
								}

								if(LSelected)
								{
									LChanged |= E3dChgSELECTION;
									if(LSelected == 1) E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected next sibling Node:  \"%s\"", LTNode->Name ? LTNode->Name : "");
									else E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected next sibling Branch:  \"%s\"", LTNode->Name ? LTNode->Name : "");
								}
							}
							else
							{
								if(LNode == LRootNode)
								{
									if(LRC < LRN - 1)
									{
										int  LSelected = 0;
										LTNode = LScene->RootNodes[LRC + 1];

										switch(LNode->Selection)
										{
											case E3dNodeSelNODE:         E3dSel_NodeSelect(LScene, LTNode, TRUE, TRUE);LSelected = 1;break;
											case E3dNodeSelBRANCH_ROOT:  E3dSel_NodeSelectTree(LScene, LTNode, TRUE, TRUE);LSelected = 2;break;
										}

										if(LSelected)
										{
											LChanged |= E3dChgSELECTION;
											if(LSelected == 1) E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected next hierarchy root Node:  \"%s\"", LTNode->Name ? LTNode->Name : "");
											else E3dp_PrintMessage(ELogMESSAGE, 2000, "Selected next hierarchy:  \"%s\"", LTNode->Name);
										}
									}
									else E3dp_PrintMessage(ELogWARNING, 2000, "Node \"%s\" has no next sibling.", LNode->Name ? LNode->Name : "");
								}
								else E3dp_PrintMessage(ELogWARNING, 2000, "Node \"%s\" has no next sibling.", LNode->Name ? LNode->Name : "");
							}
						}
					break;
				}
				goto Done;
			}
		}
	}

Done:
	if(LChanged)
	{
		E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

//	E3dp_SelectByNameMenuButton = XePushButton_Add("Menu/Select", "By name", '\0', NULL, NULL, NULL, _MCB_NodeSelection, (EPointer)E3dCrSELECT_BY_NAME);

	const EKeyModifiers  LCtrlMask = Xe_CommandKeyForCtrl ? E_CommandModifierMask : EControlMask;const char*  LCtrlName = Xe_CommandKeyForCtrl ? "Cmd" : "Ctrl";char                 LAccText[256];

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "All Models",                    '\0', NULL,                NULL,           FALSE, NULL, _MCB_NodeSelection,         (EPointer)E3dCrALL_NORMAL_MODELS);
	sprintf(LAccText, "Shift+%s+A", LCtrlName);_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "All hierarchies",               '\0', "Shift Ctrl <Key>A", LAccText, FALSE, NULL, _MCB_NodeSelection,         (EPointer)E3dCrALL_HIERARCHIES);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "Empty Models",                  '\0', NULL,                NULL,           FALSE, NULL, _MCB_NodeSelection,         (EPointer)E3dCrEMPTY_MODELS);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "All Splines",                   '\0', NULL,                NULL,           FALSE, NULL, _MCB_SelectObjects,         (EPointer)E3dItemSPLINE);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "All Spline Faces",              '\0', NULL,                NULL,           FALSE, NULL, _MCB_SelectObjects,         (EPointer)E3dItemFACE);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "All Meshes",                    '\0', NULL,                NULL,           FALSE, NULL, _MCB_SelectObjects,         (EPointer)E3dItemMESH);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "All Geometries",                '\0', NULL,                NULL,           FALSE, NULL, _MCB_SelectObjects,         (EPointer)E3dItemGEOMETRY);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "Hidden Geometries",             '\0', NULL,                NULL,           FALSE, NULL, _MCB_SelectHidden,          (EPointer)E3dItemGEOMETRY);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "Hidden Nodes and Geometries",   '\0', NULL,                NULL,           FALSE, NULL, _MCB_SelectHidden,          (EPointer)(E3dItemNODE|E3dItemGEOMETRY));

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "Unselect everything",           '\0', NULL, NULL, FALSE, NULL, _MCB_NodeSelection, (EPointer)E3dCrUNSELECT_EVERYTHING);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "Unselect all Nodes",            '\0', NULL, NULL, FALSE, NULL, _MCB_NodeSelection, (EPointer)E3dCrUNSELECT_ALL_MODELS);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "Invert PolyGroup selection",    '\0', NULL, NULL, FALSE, NULL, _MCB_PolySelection, (EPointer)E3dCrINVERT_POLYGROUP_SELECTION);

//	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "Edges",      '\0', NULL, NULL, FALSE, NULL, _MCB_PolySelection, (EPointer)E3dCrSELECT_EDGES);
//	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select", "Hulls",      '\0', NULL, NULL, FALSE, NULL, _MCB_PolySelection, (EPointer)E3dCrSELECT_HULLS);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Points", "Select points of selected Objects",    '\0', NULL, NULL, FALSE, NULL, _MCB_PointSelection,           (EPointer)E3dSELECT_POINTS_OF_SELECTION);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Points", "Un-select points of selected Objects", '\0', NULL, NULL, FALSE, NULL, _MCB_PointSelection,           (EPointer)E3dUNSELECT_POINTS_OF_SELECTION);
	E3dScene_ScriptFunctionRegister("Unselect_Points", "Unselect points", "Unselect selected Vertices on selected Meshes, or selected CVs on selected Splines.", _Unselect_Points,  _ScriptFunctions + _ScriptFunctions_Count++);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Points", "Select unused points",                 '\0', NULL, NULL, FALSE, NULL, _MCB_PointSelection,           (EPointer)E3dSELECT_UNUSED_POINTS);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Points", "Select Edge splitter points",          '\0', NULL, NULL, FALSE, NULL, _MCB_Select_IntermediatePoints, (EPointer)0);
	E3dScene_ScriptFunctionRegister("Select_IntermediatePoints", "Select intermediate points", "Select Edge bisector Vertices on selected Meshes, or intermediate CVs on selected Splines.", _Select_IntermediatePoints,  _ScriptFunctions + _ScriptFunctions_Count++);


	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Edges", "Select Edges with selected Vertices",   '\0', NULL, NULL, FALSE, NULL, _MCB_EdgeSelection, (EPointer)E3dCrSELECT_EDGES_WITH_SELECTED_VERTICES);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Edges", "Select Edges of selected Polygons",     '\0', NULL, NULL, FALSE, NULL, _MCB_EdgeSelection, (EPointer)E3dCrSELECT_EDGES_OF_SELECTED_POLYGONS);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Edges", "Unselect Edges",                        '\0', NULL, NULL, FALSE, NULL, _MCB_EdgeSelection, (EPointer)E3dCrUNSELECT_EDGES);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Edges", "Select Edges at normal creases",        '\0', NULL, NULL, FALSE, NULL, _MCB_EdgeSelection, (EPointer)E3dCrSELECT_EDGES_AT_NORMAL_CREASES);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Edges", "Select boundary Edges",                 '\0', NULL, NULL, FALSE, "Select Edges at the boundaries between PolyGroups, between selected Polygons,\nand at the boundaries of Meshes (e.g. to find holes in a Mesh)", _MCB_EdgeSelection, (EPointer)E3dCrSELECT_BOUNDARY_EDGES);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Polygons", "Polygons with all Vertices selected",'\0', NULL, NULL, FALSE, NULL, _MCB_PolySelection, (EPointer)E3dCrPOLYGONS_WITH_ALL_VERTICES_SELECTED);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Polygons", "Polygons with a selected Edge",      '\0', NULL, NULL, FALSE, NULL, _MCB_PolySelection, (EPointer)E3dCrPOLYGONS_WITH_A_SELECTED_EDGE);
	E3dScene_ScriptFunctionRegister("Select_PolygonsWithASelectedEdge", "Select Polygons with a selected Edge", "Select Polygons that have at least one selected Edge.", _Select_PolygonsWithASelectedEdge,  _ScriptFunctions + _ScriptFunctions_Count++);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Polygons", "Triangles",                          '\0', NULL, NULL, FALSE, NULL, _MCB_PolySelection, (EPointer)E3dCrSELECT_TRIS);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Polygons", "Quads",                              '\0', NULL, NULL, FALSE, NULL, _MCB_PolySelection, (EPointer)E3dCrSELECT_QUADS);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Polygons", "Unselect all Polygons",              '\0', NULL, NULL, FALSE, NULL, _MCB_PolySelection, (EPointer)E3dCrUNSELECT_ALL_POLYS);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Select/Polygons", "Invert Polygon selection",           '\0', NULL, NULL, FALSE, NULL, _MCB_PolySelection, (EPointer)E3dCrINVERT_POLYGON_SELECTION);

//	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu/Delete", "Hidden Nodes and Objects",    '\0', NULL, NULL, FALSE, NULL, _MCB_DeleteHiddenNodesAndObjects, (EPointer)FALSE);

	E_KeySetCallback("Select all hierarchies", 'A', EShiftMask|LCtrlMask, EKeyDOWN, _KCB_Selection, (EPointer)E3dCrALL_HIERARCHIES);

	E_KeySetCallback("Go up",    EKeyUP,    0, EKeyPRESS, _KCB_HrcWalk, (EPointer)E3dCrGO_UP);
	E_KeySetCallback("Go down",  EKeyDOWN,  0, EKeyPRESS, _KCB_HrcWalk, (EPointer)E3dCrGO_DOWN);
	E_KeySetCallback("Go left",  EKeyLEFT,  0, EKeyPRESS, _KCB_HrcWalk, (EPointer)E3dCrGO_LEFT);
	E_KeySetCallback("Go right", EKeyRIGHT, 0, EKeyPRESS, _KCB_HrcWalk, (EPointer)E3dCrGO_RIGHT);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		{ int  LC = 0;for(;LC < _MenuButtons_Count;LC++) XeItemFree(_MenuButtons[LC]); _MenuButtons_Count = 0;  }
		{ int  LC = 0;for(;LC < _ScriptFunctions_Count;LC++) { E3dScene_ScriptFunctionRemove(_ScriptFunctions[LC]); } _ScriptFunctions_Count = 0; }

		E_KeyRemoveCallback(EKeyUP,    0, EKeyPRESS, _KCB_HrcWalk, (EPointer)E3dCrGO_UP);
		E_KeyRemoveCallback(EKeyDOWN,  0, EKeyPRESS, _KCB_HrcWalk, (EPointer)E3dCrGO_DOWN);
		E_KeyRemoveCallback(EKeyLEFT,  0, EKeyPRESS, _KCB_HrcWalk, (EPointer)E3dCrGO_LEFT);
		E_KeyRemoveCallback(EKeyRIGHT, 0, EKeyPRESS, _KCB_HrcWalk, (EPointer)E3dCrGO_RIGHT);
	}

	return(E_SUCCESS);
}
