//================================================================
// EQUINOX-3D(TM) plugin
//
// - Create and edit Splines
// - Edges to Splines tool
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 09:38:25
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <math.h>
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>


#include <E3D/E3D.h>
#include <E3D/Item.h>
#include <E3D/Matrix.h>
#include <E3D/Model.h>

#include <E3D/SplineOps.h>  // Undoable Spline operations

#include <E3D/UI/ModeStack.h>

#ifndef _E3D_Panel_h
#include <E3D/UI/Panel.h>
#endif

#ifndef _E3D_Panels_h
#include <E3D/UI/Panels.h>
#endif

#ifndef _E3D_StatusPanel_h
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage
#endif

#include <E3D/GL/Draw.h>          // For E3dGL_GetFloatv

#include <Color/Panel.h>

#include <Xe/ArrowButton.h>
#include <Xe/Label.h>
#include <Xe/Matrix.h>
#include <Xe/PushButton.h>
#include <Xe/Scale.h>
#include <Xe/Separator.h>
#include <Xe/ToggleButton.h>

// Icons
//
EpIconDef3(SplSegmentLin)
EpIconDef3(SplSegmentCrv)
EpIconDef3(SplSegmentArc)
EpIconDef3(TgC0)
EpIconDef3(TgC1)
EpIconDef3(TgG1)


//------------------------------------------------
// E3DP/TransformPanel.c
//------------------------------------------------
extern E3dMatrix     E3dp_WorldToLocalMatrix;
extern E3dGeometry*  E3dp_ActiveGeometry;
extern E3dSpline*    E3dp_ActiveSpline;
extern int           E3dp_ActiveSubCV;
extern E3dSplineCV*  E3dp_ActiveSplineCV;


static EguiItem   _MenuButtons[16];static int  _MenuButtons_Count = 0;

static int		    _SplineType = E3dSplineNONE;  // Type of Spline being drawn

static E3dModel*  _Model = NULL;

// The Geometry the currently edited Spline belongs to
// If this the Spline itself, _Spline will be the same.
// If this is a Face, _Spline will point to a Contour of that Face
//
static E3dGeometry*  _Geometry = NULL;
static E3dSpline*    _Spline = NULL;



static EChangeFlags _EdgesToSplines_InputChanged(EInterface* PInterface, EInterfaceCallbackStruct* PCBS);

#define E3dSnVERTEX  1
#define E3dSnEDGE    2

typedef struct
{
	E3dNodeComponentCore;

	EList(, E3dModel*,  SrcModels);            // ...Corresponding Models
	EList(, E3dObject*, SrcObjects);           // Source Objects and... (instances, really)
	EBool*  SrcActiveFlags;                    // ...Corresponding Active flags

	EBool    DoHiddenEdges, DoUnselectedEdges,
	         MergeContiguous;                  // Generate a single Spline for non-branching Edges
	uint8_t  SplineNormalsMode;                // 0: don't generate Spline normals, E3dSnVERTEX: from Vertex normals, E3dSnEDGE from Edge normals

// Internal variables
//
	EInterface          InputInterface;        // This is added to SrcObjects and SrcModels as an output node

	EListA(, E3dSpline*, Splines);             // We cache Splines here, so we don't have to create / destroy tons of Splines every time the source Meshes change
} E3dEdgesToSplines;

#define ROffsetP(field)  (void*)(EPtrOffset(E3dEdgesToSplines, field))
#define ROffsetI(field)  EPtrOffset(E3dEdgesToSplines, field)
static EResource  _EdgesToSplines_Resources[]=
{
	{ "SrcModels",          EResOBJECT|EResDYNAMIC,  ROffsetP(SrcModels),      .Ext.Count= ROffsetI(SrcModels_Count) },
	{ "SrcObjects",         EResOBJECT|EResDYNAMIC,  ROffsetP(SrcObjects),     .Ext.Count= ROffsetI(SrcObjects_Count) },
	{ "SrcActiveFlags",     EResBOOLEAN|EResDYNAMIC, ROffsetP(SrcActiveFlags), .Ext.Count= ROffsetI(SrcObjects_Count) },

	{ "DoHiddenEdges",      EResBOOLEAN,             ROffsetP(DoHiddenEdges) },
	{ "DoUnselectedEdges",  EResBOOLEAN,             ROffsetP(DoUnselectedEdges) },
	{ "MergeContiguous",    EResBOOLEAN,             ROffsetP(MergeContiguous) },
	{ "SplineNormalsMode",  EResUINT8,               ROffsetP(SplineNormalsMode) },
	{ }
};
#undef ROffsetP
#undef ROffsetI

typedef struct
{
	Widget  DoHiddenEdgesW, DoUnselectedEdgesW,
	        SplineNormalsModeNoneW, SplineNormalsModeVertexW, SplineNormalsModeEdgeW,
	        MergeContiguousW;
} E3dResourcesGUI;

static E3dNodeComponentClass*  _EdgesToSplines_class = NULL;


enum
{
	CR_SPLINE_OPEN,
	CR_SPLINE_CLOSE,
	CR_SPLINE_DRAW,
	CR_SPLINE_EDIT_POINTS,
	CR_DELETE_POINTS
};


enum
{
 EditMode_ADD_ADD_ADD = 1,
 EditMode_DELETE
};

static int  _CVEditMode = EditMode_ADD_ADD_ADD;

#define E3dTRANSFORM_EDGES   0
#define E3dUSE_EDGE_NORMALS  1

#define EPSILON	0.00001


//========================================================
// Create compound message string for the given mode
//========================================================
static XeString _GetMessage(const int PType, const int PIndex)
{
	XeString  LXeString;

	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color      = Xe_ForegroundRGBA8pColor;
	const uint32_t    LButtonNameRGBAp8Color = LPrefs->ButtonNameRGBAp8Color;

	switch(PType)
	{
		case EditMode_ADD_ADD_ADD:
			LXeString = XeString_NewVa(True, XeFONT, XeFontDEFAULT,  XeCOLOR_RGBA, LLabelRGBAp8Color,    XeTEXT,  "Edit Spline CVs   ",
			 E3dp_LMouseButtonS,  XeTEXT,  ": add CV after last    ",
			 E3dp_MMouseButtonS,  XeTEXT,  ": add CV between CVs    ",
			 E3dp_RMouseButtonS,  XeTEXT,  ": add CV before first    ",
			 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
			 XeFONT, XeFontDEFAULT,    XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT,  ": end mode",
			 0);
		return(LXeString);

		case EditMode_DELETE:
		 LXeString = XeString_NewVa(True, XeFONT, XeFontDEFAULT,  XeCOLOR_RGBA, LLabelRGBAp8Color,    XeTEXT,  "Edit Spline CVs   ",
			 E3dp_LMouseButtonS,
			 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeFTEXT, ": delete CV #%d  ", PIndex,
			 XeFONT, XeFontFIXEDWIDTH,  XeCOLOR_RGBA, LButtonNameRGBAp8Color,  XeTEXT, "Esc",
			 XeFONT, XeFontDEFAULT,     XeCOLOR_RGBA, LLabelRGBAp8Color,       XeTEXT, ": end mode",
			 0);
		return(LXeString);
	}
	return(NULL);
}


//========================================
// Open or close selected Splines
//========================================
static void _OpenCloseSplines(const int PAction)
{
	E3dScene*     LScene = E3d_Scene;
	E3dObject**   LObjects = NULL;const EIndex  LNumOfSelectedSplines = E3dScene_GetObjects(LScene, &LObjects, E3dItemSPLINE, E3dSELECTED, FALSE);
	EIndex        LPerformedCnt = 0;
	EChangeFlags  LChanged = 0;

	
	if(LNumOfSelectedSplines)
	{
		EListAI(, EOpNode, LOpNodes);

		EIndex  LC = 0;for(;LC < LNumOfSelectedSplines;LC++)
		{
			E3dSpline*  LSpline = (E3dSpline*)(LObjects[LC]);

			switch(PAction)
			{
				case CR_SPLINE_OPEN:
					if(LSpline->Closed)
					{
						LSpline->Closed = FALSE;
						LChanged |= E3dObject_UpdateForDisplay((E3dObject*)LSpline, E3dUF_ALL - E3dUF_CALL_OUTPUTS, LScene) | E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY;

						EOpNode*  LOpNode = EList_StructAddAClr(LOpNodes, 4);if(LOpNode)
						{
							LOpNode->Object  = (ECoreObject*)LSpline;
							LOpNode->Changed = E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY;
						}

						LPerformedCnt++;
					}
				break;

				case CR_SPLINE_CLOSE:
					if(!LSpline->Closed)
					{
						LSpline->Closed = TRUE;
						LChanged |= E3dObject_UpdateForDisplay((E3dObject*)LSpline, E3dUF_ALL - E3dUF_CALL_OUTPUTS, LScene) | E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY;

						EOpNode*  LOpNode = EList_StructAddAClr(LOpNodes, 4);if(LOpNode)
						{
							LOpNode->Object  = (ECoreObject*)LSpline;
							LOpNode->Changed = E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY;
						}

						LPerformedCnt++;
					}
				break;
			}
		}

		switch(PAction)
		{
		 case CR_SPLINE_CLOSE:
			 if(LPerformedCnt > 0)
			 {
				 E3dp_PrintMessage(ELogMESSAGE, 5000, "%d Spline(s) closed", LPerformedCnt);
			 }
			 else E3dp_PrintMessage(ELogWARNING, 5000, "No open Spline selected");
		 break;

		 case CR_SPLINE_OPEN:
			 if(LPerformedCnt > 0)
			 {
				 E3dp_PrintMessage(ELogMESSAGE, 5000, "%d Spline(s) opened", LPerformedCnt);
			 }
			 else E3dp_PrintMessage(ELogWARNING, 5000, "Select at least 1 closed Spline");
		 break;
		}

		EFree(LObjects);

		if(LOpNodes)
		{
			LChanged |= E3dObjects_CallOutputs(ECbINTERNAL|ECbUSER, LOpNodes, LOpNodes_Count);
			EFree(LOpNodes);
		}

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
	else E3dp_PrintMessage(ELogWARNING, 5000, "No Spline selected");
}


//================================================
// Start CV moving
//================================================
static void _CVStartMoving(E3dpWindow* P3DWindow,  const int PPtrX, const int PPtrY,  E3dModel* PModel, E3dSpline* PSpline, E3dSplineCV* PCV, const int PSubCV, E3dCoord3* PPoint)
{
	E3dp_PushPointerMode(E3dpPointerMode_MOVE_POINT, XeString_SimpleNew("Moving Spline CV"));

	E3dp_ActiveGeometry = _Geometry;
	E3dp_ActiveSpline   = PSpline;

	E3dp_ActiveSplineCV = PCV;
	E3dp_ActiveSubCV    = PSubCV;
	E3dp_OldPointFlags  = PCV->CVFlags;

	PCV->CVFlags |= E3dCV_ACTIVE;

	E3dMatrix_Invert3x4(PModel->LocalToWorldMatrix, E3dp_WorldToLocalMatrix);
	E3dMatrix_Invert3x4(PModel->LocalToWorldMatrix, E3dp_WorldToLocalMatrix);

	E3dp_PointMoveInit(P3DWindow,  PPtrX, PPtrY,  PPoint->X, PPoint->Y, PPoint->Z, FALSE);
}


//========================================================
// Add a CV to a Spline with and undo-able operator
//========================================================
static EChangeFlags _AddCV(E3dPickCallbackStruct* PPickStruct, E3dSpline* PSpline, const EIndex PCVIndex, const E3dSplineCV* PCV, const E3dSplineCV* PReferenceCV, const E3dCoord3* PLocalPosition, const E3dCoord3* PWorldPosition, const int PSubCV, E3dModel* PModel, EOps* POps)
{
	EChangeFlags  LChanged = 0;

	E3dp_PickStop();    // Must stop picking if we change a Geometry

// Initialize Operator, execute it and add it to the undo stack
//
	{
		E3dOperatorAddDelSplineCVs*  LOperator = (E3dOperatorAddDelSplineCVs*)EOp_OperatorNew(E3d_OperatorAddSplineCVs_class);
		E3dSplineCV*                 LCV;ENew(LCV);E3dSplineCV_Init(LCV, PSpline->SplineType);
//		const EBool                  LHasTangents = (EBool)(PSpline->SplineType == E3dSplineBEZIER || PSpline->SplineType == E3dSplineHERMITE);
		const EBool                  LHasTangents = TRUE;

		if(PCV) *LCV = *PCV;
		else
		{
			LCV->Position = *PLocalPosition;

//			if(LHasTangents)
			{
// With this part commented out, it's a neuron-drawer
// :)
				LCV->Previous = *PLocalPosition;

				LCV->Next = *PLocalPosition;
			}
		}

// Copy constraint, segment type and resolution from PReferenceCV, if it's given
//
		if(PReferenceCV)
		{
			LCV->Constraint           = PReferenceCV->Constraint;
			LCV->SegmentShape         = PReferenceCV->SegmentShape;
			LCV->SegmentNumOfLinSteps = PReferenceCV->SegmentNumOfLinSteps;
		}


		E_AssignToMember(_Geometry, LOperator,Object);
		E_AssignToMember(PSpline,   LOperator,Spline);
		ENew(LOperator->CVs_Indices);LOperator->CVs_Indices[0] = PCVIndex;LOperator->CVs = LCV;LOperator->CVs_Count = 1;

		EList_ObjectAdd(POps->Operators, LOperator);
		if(POps->Operators_Count)
		{
			LChanged |= EOp_OpsRun(POps, EOpDO);    // Run OperatorList
		}

		LCV = PSpline->CVs + PCVIndex;

		switch(PSubCV)
		{
			case E3dCV_POSITION:
				_CVStartMoving(PPickStruct->E3DWindow,  PPickStruct->PtrX, PPickStruct->PtrY,  PModel, PSpline, LCV, LHasTangents ? E3dCV_OUT_TANGENT : E3dCV_POSITION, &LCV->Position);

				E3dp_PrintMessage(ELogMESSAGE, 0, "Point #%d: % .4f,% .4f,% .4f (local: % .4f,% .4f,% .4f)", PCVIndex,  XYZ_p(PWorldPosition),  XYZ_s(LCV->Position));
			break;

			case E3dCV_IN_TANGENT:
				_CVStartMoving(PPickStruct->E3DWindow,  PPickStruct->PtrX, PPickStruct->PtrY,  PModel, PSpline, LCV, LHasTangents ? E3dCV_OUT_TANGENT : E3dCV_POSITION, &LCV->Previous);

				E3dp_PrintMessage(ELogMESSAGE, 0, "Point #%d (prev): % .4f,% .4f,% .4f (local: % .4f,% .4f,% .4f)", PCVIndex,  XYZ_p(PWorldPosition),  XYZ_s(LCV->Previous));
			break;

			case E3dCV_OUT_TANGENT:
				_CVStartMoving(PPickStruct->E3DWindow,  PPickStruct->PtrX, PPickStruct->PtrY,  PModel, PSpline, LCV, LHasTangents ? E3dCV_OUT_TANGENT : E3dCV_POSITION, &LCV->Next);

				E3dp_PrintMessage(ELogMESSAGE, 0, "Point #%d (next): % .4f,% .4f,% .4f (local: % .4f,% .4f,% .4f)", PCVIndex,  XYZ_p(PWorldPosition),  XYZ_s(LCV->Next));
			break;
		}
	}

	E3dp_PickResume();

	return(LChanged);
}


static void _PickPosition(E3dPickCallbackStruct* PPickStruct, const E3dCoord3* PPositionIn)
{
	E3dScene*     LScene = E3d_Scene;
	E3dModel*     LModel = NULL;
	E3dSpline*    LSpline = _Spline;
	E3dModel*     LCameraModel = PPickStruct->CameraModel;
	E3dMatrix     LViewerToWorldMatrix;E3dMatrix_Copy(LCameraModel->LocalToWorldMatrix, LViewerToWorldMatrix);
	E3dMatrix     LWorldToViewerMatrix;E3dMatrix_Invert3x4(LViewerToWorldMatrix, LWorldToViewerMatrix);
	E3dMatrix     LMatrix;
	E3dRay        LRay;
	E3dCoord3     LRO = PPickStruct->RayOrigin, LRD = PPickStruct->RayDirection;
	EChangeFlags	LChanged = 0;
	EBool         LNewModel = FALSE;

	E3dCoord3     LPosition;

	if(PPositionIn) LPosition = *PPositionIn;
	else
	{
		LRay.Origin    = PPickStruct->RayOrigin;
		LRay.Direction = PPickStruct->RayDirection;

// Intersect the ray cast by the pick algorithm with a plane that is perpendicular
// to the Camera's view direction and passes through a reference point
//
		{
			E3dCoord3      LDiff, LViewDirection;  // Camera Direction vector
			E3dCoordinate  LD;                     // t parameter
			E3dCoord3      LAimPoint, LAimDirection;E3dModel_GetAim(LCameraModel,  &LAimPoint, &LAimDirection, NULL, NULL, NULL);

			LViewDirection.X = -LAimDirection.X;
			LViewDirection.Y = -LAimDirection.Y;
			LViewDirection.Z = -LAimDirection.Z;

// Get D in the plane equation
//
			LModel = _Model;
			if(LModel && LSpline && LSpline->CVs)
			{
				const unsigned int	LCVSize = sizeof(E3dSplineCV);

				switch(PPickStruct->Button)
				{
					case Button1:	// Add after last CV, use last CV as the reference point
						{
							const E3dSplineCV*  LReferenceCV = (E3dSplineCV*)(((char*)(LSpline->CVs)) + LCVSize * (LSpline->CVs_Count - 1));
							E3dMatrix_Transform3x4_vv(LModel->LocalToWorldMatrix,  LReferenceCV->Position,  LDiff);
							LDiff.X *= LViewDirection.X;
							LDiff.Y *= LViewDirection.Y;
							LDiff.Z *= LViewDirection.Z;
						}
					break;

					case Button3:	// Add before first CV, use first CV as the reference point
						{
							const E3dSplineCV*  LReferenceCV = (E3dSplineCV*)(LSpline->CVs);
							E3dMatrix_Transform3x4_vv(LModel->LocalToWorldMatrix,  LReferenceCV->Position,  LDiff);
							LDiff.X *= LViewDirection.X;
							LDiff.Y *= LViewDirection.Y;
							LDiff.Z *= LViewDirection.Z;
						}
					break;

					default:
						LDiff.X = LViewDirection.X * LAimPoint.X;
						LDiff.Y = LViewDirection.Y * LAimPoint.Y;
						LDiff.Z = LViewDirection.Z * LAimPoint.Z;
					break;
				}
			}
			else
			{
				LDiff.X = LViewDirection.X * LAimPoint.X;
				LDiff.Y = LViewDirection.Y * LAimPoint.Y;
				LDiff.Z = LViewDirection.Z * LAimPoint.Z;
			}

			LD = LDiff.X + LDiff.Y + LDiff.Z;

// t is the parameter for the ray where it intersects the plane
//
			const E3dCoordinate	LT = (LD - LViewDirection.X * LRO.X - LViewDirection.Y * LRO.Y - LViewDirection.Z * LRO.Z) / (LViewDirection.X * LRD.X + LViewDirection.Y * LRD.Y + LViewDirection.Z * LRD.Z);

			LPosition.X = LRO.X + LT * LRD.X;
			LPosition.Y = LRO.Y + LT * LRD.Y;
			LPosition.Z = LRO.Z + LT * LRD.Z;

			if(PPickStruct->SnapToX) LPosition.X = E3d_SnapCoordinateTo(LPosition.X, PPickStruct->SnapToXV);
			if(PPickStruct->SnapToY) LPosition.Y = E3d_SnapCoordinateTo(LPosition.Y, PPickStruct->SnapToYV);
			if(PPickStruct->SnapToZ) LPosition.Z = E3d_SnapCoordinateTo(LPosition.Z, PPickStruct->SnapToZV);
		}
	}

	switch(PPickStruct->Button)
	{
		case Button1:
		case Button3:
			if(_Model == NULL)
			{
				char*  LModelName = NULL;
				char*  LGeometryName = NULL;

				switch(_SplineType)
				{
					case E3dSplineBEZIER:
						LModelName    = "BezSpline";
						LGeometryName = "bezSpline";
					break;

					case E3dSplineBSPLINE:
						LModelName    = "BSpline";
						LGeometryName = "bSpline";
					break;

					case E3dSplineCARDINAL:
						LModelName    = "CardSpline";
						LGeometryName = "cardSpline";
					break;

					case E3dSplineLINEAR:
						LModelName    = "LinSpline";
						LGeometryName = "linSpline";
					break;
				}

				E_New(E3dModel, LModel);if(LModel)
				{
					EObject_SetName(LModel, E3dScene_GetUniqueNodeName(LScene, LModelName), FALSE);

					E_New(E3dSpline, LSpline);if(LSpline)
					{
						EObject_SetName(LSpline, E3dScene_GetUniqueObjectName(LScene, LGeometryName), FALSE);
						LSpline->SplineType = _SplineType;
						E3dModel_AppendObject(LModel, (E3dObject*)LSpline);

						E3dScene_NodeHrcAdd(E3d_Scene, (E3dNode*)LModel, E3dUF_CALL_SCENE_CALLBACKS);
						E3dScene_ObjectsUpdateModelList(LScene);
						E3dScene_LayoutOnSchematics(LScene, TRUE);

						LModel->Selection = E3dNodeSelNODE;
						E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);

						_Model = LModel;
						_Geometry = (E3dGeometry*)LSpline;
						_Spline = LSpline;
						LChanged |= E3dChgHIERARCHY;
						LNewModel = TRUE;
					}
				}
			}
		break;
	}


	if(LSpline && E3d_OperatorAddSplineCVs_class)
	{
		EOps   LOps;EOp_OpsInit(&LOps);
		EBool  LHasTangents = (EBool)(LSpline->SplineType == E3dSplineBEZIER);

//printf("%s() %s:%d [%s]\n", __FUNCTION__, __FILE__, __LINE__, LModel ? LModel->Name : "");fflush(stdout);
// Find the Model that 'makes' this Spline selected and use the
// inverse of its LocalToWorldMatrix to go to local space...
//
		if(LModel == NULL) LModel = E3dObject_GetSelectingModel((E3dObject*)_Geometry, TRUE);

		E3dCoord3  LWorldPosition = LPosition;

		if(LModel)
		{
			E3dMatrix_Invert3x4(LModel->LocalToWorldMatrix, LMatrix);
			E3dMatrix_Transform3x4_v(LMatrix,  LPosition);    
		}
		else
		{
// If there's no LModel, return
// Maybe we should exit CV editing if the Spline gets unselected?
//
			return;
//			E3dMatrix_LoadIdentity(LMatrix);
		}


		switch(PPickStruct->Button)
		{
// Add CV after the last one
//
			case Button1:
				LChanged |= _AddCV(PPickStruct, LSpline, LSpline->CVs_Count, NULL, LSpline->CVs_Count ? LSpline->CVs + LSpline->CVs_Count - 1 : NULL, &LPosition, &LWorldPosition, LHasTangents ? E3dCV_OUT_TANGENT : E3dCV_POSITION, LModel, &LOps);
			break;

// Add CV between 2 existing ones
//
			case Button2:
				if(PPositionIn == NULL)
				{
					if(LSpline->CVs_Count >= 2)
					{
						E3dSplineCV  LTCV;EIndex  LCVIndex = E3dSpline_CVInsertBetween(LSpline, LMatrix,  &LPosition, &LRay, &LTCV);if(LCVIndex > -1)
						{
							LChanged |= _AddCV(PPickStruct, LSpline, LCVIndex,  &LTCV, NULL,  &LPosition, &LWorldPosition, (LHasTangents && LTCV.SegmentShape != E3dSplineLINEAR) ? E3dCV_IN_TANGENT : E3dCV_POSITION, LModel, &LOps);
						}
					}
					else
					{
						E3dp_PrintMessage(ELogWARNING, 4000, "Spline must have at least 2 CVs to add between");
						return;
					}
				}
			break;

// Add CV to the beginning
//
			case Button3:
				LChanged |= _AddCV(PPickStruct, LSpline, 0,  NULL, LSpline->CVs,  &LPosition, &LWorldPosition, LHasTangents ? E3dCV_IN_TANGENT : E3dCV_POSITION, LModel, &LOps);
			break;
		}

		if(LOps.Operators_Count)
		{
			if(!LNewModel) E3dp_UndoStackAddOps(&LOps, NULL);    // Add OperatorList to undo stack
			else EOp_OpsFreeContents(&LOps);
		}

// Note that _Geometry may not be the same as LSpline (e.g. if it's a Face)
//
		if(LChanged) LChanged |= E3dObject_UpdateForDisplay((E3dObject*)_Geometry, E3dUF_ALL, LScene);
	}

	E3dWindows_RedrawByChangeFlags(LChanged);
}


static EBool _PolygonIntersect(E3dGridRay* PRay, E3dModel* PModel, const E3dMesh* PMesh, const E3dPolyGroup* PPolyGroup, const E3dPolygon* PPolygon)
{
	const E3dVertex*         LVertices = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;
	const E3dVertexNodeSize  LVertexNodeSize = PPolyGroup->VertexNodeSize;
	const E3dCoordinate*     LVertexMatrix = PModel->LocalToWorldMatrix;

  E3dRTriangle  LRTriangle;

//printf("%s() %s:%d %d\n", __FUNCTION__, __FILE__, __LINE__, (int)(PPolygon->NumOfVertices));fflush(stdout);
	if(PPolygon->NumOfVertices == 3)  // Is this Polygon a triangle?
	{
		const E3dVertexNode*  LVertexNode = PPolygon->VertexNodes;
 
		LRTriangle.Normal = PPolygon->Normal;
		{ const E3dVertex*  LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);E3dMatrix_Transform3x4_vv(LVertexMatrix, LVertex->Position, LRTriangle.Vertex0);EStruct_Inc(LVertexNode, LVertexNodeSize); }
		{ const E3dVertex*  LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);E3dMatrix_Transform3x4_vv(LVertexMatrix, LVertex->Position, LRTriangle.Vertex1);EStruct_Inc(LVertexNode, LVertexNodeSize); }
		{ const E3dVertex*  LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);E3dMatrix_Transform3x4_vv(LVertexMatrix, LVertex->Position, LRTriangle.Vertex2); }

    if(E3dRTriangle_PrepForRayTracing(&LRTriangle) == E_SUCCESS)
    {
			if(E3d_RayIntersectTriangleFrontBack(PRay, (E3dRTriangleICore*)&LRTriangle)) return(TRUE);
		}
	}
	else
	{
		const E3dITriangle*  LITriangle = PPolygon->ITriangles;
 
		if(LITriangle)
		{
			const E3dVertexNode*  LVertexNodes = PPolygon->VertexNodes;
 			const EIndex          LTN = PPolygon->ITriangles_Count;
	
			EIndex  LTC = 0;for(;LTC < LTN;LTC++, LITriangle++)
			{
				const E3dVertexNodeIndex*  LVertexNodeIndices = LITriangle->VertexNodes;

//printf("%s() %s:%d %d\n", __FUNCTION__, __FILE__, __LINE__, (int)LTC);fflush(stdout);
				LRTriangle.Normal = PPolygon->Normal;
				{ const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndices[0]);const E3dVertex*  LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);E3dMatrix_Transform3x4_vv(LVertexMatrix, LVertex->Position, LRTriangle.Vertex0); }
				{ const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndices[1]);const E3dVertex*  LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);E3dMatrix_Transform3x4_vv(LVertexMatrix, LVertex->Position, LRTriangle.Vertex1); }
				{ const E3dVertexNode*  LVertexNode = EStruct_ByIndex(LVertexNodes, LVertexNodeSize, LVertexNodeIndices[2]);const E3dVertex*  LVertex = EStruct_ByIndex(LVertices, LVertexSize, LVertexNode->VertexID);E3dMatrix_Transform3x4_vv(LVertexMatrix, LVertex->Position, LRTriangle.Vertex2); }

		    if(E3dRTriangle_PrepForRayTracing(&LRTriangle) == E_SUCCESS)
    		{
					if(E3d_RayIntersectTriangleFrontBack(PRay, (E3dRTriangleICore*)&LRTriangle)) return(TRUE);
				}
				else return(FALSE);
			}
		}
	}

	return(FALSE);
}


//================================================================
// Add Spline CV pick-callback
//================================================================
static void _PickCB_EditCVs(E3dPickCallbackStruct* PPickStruct, EPointer PClientData,  const EIndex PNumOfItems, E3dItem* PPickedItems)
{
	if(PPickStruct->Reason == E3dCR_ABORT)
	{
		_SplineType = E3dSplineNONE;
		_Model    = NULL;
		_Geometry = NULL;
		_Spline   = NULL;
		return;
	}



/*
if(_Spline)
{
 E3dSpline*  LSpline = _Spline;
 EIndex  LC = 0;for(;LC < LSpline->CVs_Count;LC++)
{
printf(" %08x", LSpline->CVs[LC].Flags);fflush(stdout);
}
printf("\n");fflush(stdout);
}
*/



	if(PPickStruct->Button != 0)
	{
		if(PPickedItems)
		{
			E3dItem*  LItem = PPickedItems;

			switch(LItem->Item.Type)
			{
				case E3dItemMODEL:
				 {
					 E3dObjItem*  LTItem = (E3dObjItem*)PPickedItems;

// Model picked explicitly (e.g. on Schematics)
//
					 if(LTItem->Node && LTItem->Object == NULL)
					 {
						 E3dNode*   LNode = LTItem->Node;
						 E3dCoord3  LPos = { LNode->LocalToWorldMatrix[M30], LNode->LocalToWorldMatrix[M31], LNode->LocalToWorldMatrix[M32] };
						 _PickPosition(PPickStruct, &LPos);
					 }
				 }
				break;
			}
		}
	}


	const E3dViewClass*  LViewClass = PPickStruct->E3DWindow->ViewMode->Class;
	if(LViewClass == E3dSchematicsView_class) return;

// If PPickStruct->Button is 0, it's just pointer motion
//
//printf("%s() %s:%d  %d %d\n", __FUNCTION__, __FILE__, __LINE__, PPickStruct->Button, PPickStruct->HighlightedItemChanged);fflush(stdout);
	if(PPickStruct->Button == 0 && PPickStruct->HighlightedItemChanged)
	{
		if(PNumOfItems)
		{
			const E3dItem*  LCVItem = PPickedItems;

			if(LCVItem && LCVItem->Item.Type == E3dItemSPLINE_CV)
			{
				if(_CVEditMode != EditMode_DELETE)
				{
					const E3dSplineItem*  LSplineItem = (const E3dSplineItem*)LCVItem;
					E3dp_ChangeCompoundMessage(_GetMessage(EditMode_DELETE, LSplineItem->PointIndex));
					_CVEditMode = EditMode_DELETE;
				}
			}
			else goto NoCV;
		}
		else
		{
NoCV:
			if(_CVEditMode != EditMode_ADD_ADD_ADD)
			{
				E3dp_ChangeCompoundMessage(_GetMessage(EditMode_ADD_ADD_ADD, 0));
				_CVEditMode = EditMode_ADD_ADD_ADD;
			}
		}
		return;
	}

	if(PPickStruct->Button != 0)
	{
		if(PPickedItems)
		{
			const E3dItem*  LClosestItem = PPickedItems;

//printf("%s() %s:%d  CV %d  Surface %d\n", __FUNCTION__, __FILE__, __LINE__, LCVItem ? LCVItem->ObjItem.ZNear : 0, LSurfaceItem ? LSurfaceItem->ObjItem.ZNear : 0);fflush(stdout);

			if(LClosestItem)
			{
				switch(LClosestItem->Item.Type)
				{
// Surface was hit, place new CV on the surface
//
					case E3dItemPOLYGON:
						{
							E3dPolyItem*  LPolyItem = (E3dPolyItem*)LClosestItem;
							E3dGridRay    LRay;
							E3dModel*     LCameraModel;
							const int     LX = PPickStruct->PtrX, LY = PPickStruct->PtrY;
							E3dpWindow*   L3DWindow = PPickStruct->E3DWindow;
							E3dCamera*    LCamera = E3dWindow_CastPrimaryRay(L3DWindow,  LX, LY,  (E3dRay*)&LRay, &LCameraModel);if(LCamera)
							{
								E3dCoord3  LPosition = { 0.0, 0.0, 0.0 };
								const E3dPolyGroup*  LPolyGroup = LPolyItem->PolyGroup;

								if(0)
								{
// E3dIMPLEMENTME
// High-precision version
									if(LPolyItem->PolygonIndex >= 0)
									{
//printf("%s() %s:%d  <%s/%s/%s>\n", __FUNCTION__, __FILE__, __LINE__, LPolyItem->Node->Name, LPolyItem->Object->Name, LPolyGroup->Name);fflush(stdout);
										if(_PolygonIntersect(&LRay, (E3dModel*)LPolyItem->Node, (E3dMesh*)LPolyItem->Object, LPolyGroup, LPolyGroup->Polygons + LPolyItem->PolygonIndex))
										{
											const E3dCoordinate  LT = LRay.LastT;
											E3d_AddScaled3_vvv(LRay.Origin, LRay.Direction, LT, LPosition);
printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
											_PickPosition(PPickStruct, &LPosition);
										}
									}
								}
								else
								{
// Low-precision version, using OpenGL and the Z-buffer
//
									E3dMatrix      LViewerToWorldMatrix;E3dMatrix_Copy(LCameraModel->LocalToWorldMatrix, LViewerToWorldMatrix);
									E3dMatrix      LWorldToViewerMatrix;E3dMatrix_Invert3x4(LViewerToWorldMatrix, LWorldToViewerMatrix);
									E3dCoordinate  LZNear = (E3dCoordinate)(LPolyItem->ZNear) / 4294967295.0, LXF = (E3dCoordinate)LX, LYF = (E3dCoordinate)(L3DWindow->Size.Y - LY);
									E3dMatrix      LProjectionMatrix;E3dGL_GetFloatv(GL_PROJECTION_MATRIX, LProjectionMatrix);
									GLint          LViewPort[4];E3glGetIntegerv(GL_VIEWPORT, LViewPort);

									E3d_UnProject(LXF, LYF - (E3dCoordinate)(L3DWindow->Size.Y), LZNear,  LViewPort, LProjectionMatrix, LWorldToViewerMatrix,  &LPosition);
									_PickPosition(PPickStruct, &LPosition);
								}
							}
						}
					break;

					case E3dItemSURFACE:
printf("%s() %s:%d  %sIMPLEMENT ME%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
					break;

/*
					case E3dItemMODEL:
					 {
						 E3dObjItem*  LTItem = (E3dObjItem*)PPickedItems;

						 if(LTItem->Node && LTItem->Object == NULL)
						 {
printf("%s() %s:%d Node\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
						 }
						 else
						 {
							 _PickPosition(PPickStruct, PNumOfItems, PPickedItems);
printf("%s() %s:%d Node\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
							}
						}
					break;
*/

					case E3dItemVERTEX:
					break;

					case E3dItemSPLINE_CV:
					case E3dItemTANGENT_PREVIOUS:
					case E3dItemTANGENT_NEXT:
// Delete CV
//
						{
							E3dSplineItem*  LSplineItem = (E3dSplineItem*)LClosestItem;
							E3dSpline*      LSpline = LSplineItem->Spline;

							if(LSpline->CVs_Count > 1)
							{
								const EIndex  LCVIndex = LSplineItem->PointIndex;

								E3dp_PickStop();    // Must stop picking if we change a Geometry

//								EList_StructRemoveByIndex(LSpline->CVs, LCVIndex);

// Unhighlight CV (it was highlighted because the pointer was over it)
//
								LSpline->CVs[LCVIndex].CVFlags &= ~E3dCV_ACTIVE;
// Initialize Operator, execute it and add it to the undo stack
//
								if(E3d_OperatorDeleteSplineCVs_class)
								{
									EOps  LOps;EOp_OpsInit(&LOps);

									E3dOperatorAddDelSplineCVs*  LOperator = (E3dOperatorAddDelSplineCVs*)EOp_OperatorNew(E3d_OperatorDeleteSplineCVs_class);

									E_AssignToMember(LSplineItem->Object, LOperator,Object);
									E_AssignToMember(LSpline, LOperator,Spline);
									ENew(LOperator->CVs_Indices);LOperator->CVs_Indices[0] = LCVIndex;LOperator->CVs_Count = 1;

									EList_ObjectAdd(LOps.Operators, LOperator);
									E3dp_RunOperators(&LOps, EOpDO);    // Run Operators
									E3dp_UndoStackAddOps(&LOps, NULL);  // Add OperatorList to undo stack
								}

								E3dp_PickResume();

								E3dp_PrintMessage(ELogMESSAGE, 2500, "Deleted Spline CV #%d", LCVIndex);
							}
							else E3dp_PrintMessage(ELogMESSAGE, 2500, "Spline has only one CV. Can not delete it.");
						}
					break;
				}

				return;
			}
		}
		else
		{
			_PickPosition(PPickStruct, NULL);
		}
	}
}


//========================================
// Create Spline
//========================================
static void _MCB_SplineCreate(Widget PW, EPointer PClientData, EPointer PCallData)
{
	XeString  LXeString = _GetMessage(EditMode_ADD_ADD_ADD, 0);

	E3dp_ResetWorkModes();
// E3dp_PickRequest(LXeString, E3dItemWORLD_POSITION|E3dItemSPLINE_CV, _PickCB_EditCVs, 0, TRUE);
	E3dp_PickRequest(LXeString, E3dItemWORLD_POSITION|E3dItemSPLINE_CV|E3dItemMODEL|E3dItemSURFACE,  _PickCB_EditCVs, NULL,  0, TRUE);

	_SplineType = (EPointerInt)PClientData;
}


//========================================
// Menu callback
//========================================
static void _MCB_Splines(Widget PW, EPointer PClientData, EPointer PCallData)
{
	switch((EPointerInt)PClientData)
	{
		case CR_SPLINE_OPEN:
		case CR_SPLINE_CLOSE:
			_OpenCloseSplines((EPointerInt)PClientData);
		break;

		case CR_SPLINE_EDIT_POINTS:
			{
				E3dItem  LItem;

				E3dp_ResetWorkModes();
				if(E3dScene_GetSingleSelection(E3d_Scene, &LItem, E3dItemSPLINE|E3dItemFACE_CONTOUR, E3dVISIBLE) != E3dItemNONE)
				{
					E3dSplineItem*  LSplineItem = (E3dSplineItem*)(&LItem);

					_Geometry = (E3dGeometry*)(LSplineItem->Object);
					_Spline   = LSplineItem->Spline;
					_Model    = E3dObject_GetSelectingModel((E3dObject*)LSplineItem->Spline, TRUE);

					E3dp_PickRequest(_GetMessage(EditMode_ADD_ADD_ADD, 0), E3dItemWORLD_POSITION|E3dItemSPLINE_CV|E3dItemMODEL|E3dItemSURFACE,  _PickCB_EditCVs, NULL,  0, TRUE);
				}
				else E3dp_PrintMessage(ELogWARNING, 3000, "Select a visible Spline to edit");
			}
		break;
	}
}


typedef struct
{
	Widget         UniformStepsBW, StepsScaleW, ClosedBW,
								 CardinalTensionW,
								 SplineLinearW, SplineBezierW, SplineBSplineW, SplineCardinalW,
								 StatsValLW, CVScaleW, SegmentVisibleW, ArcLengthTesselateW,
								 SegmentDrawStepsMatrixW, SegmentDrawStepsW,
								 LinSegmentBW, BezierSegmentBW, ArcSegmentBW,
								 CVTangentConstraintMatrixW, C0TangentConstraintBW, C1TangentConstraintBW, G1TangentConstraintBW,
								 SegmentShapeMatrixW, SegmentSettingsMatrixW, SegTypeLabelW, SegmentSettingsSeparatorW;

	XeDrawer       ColorDrawer;
	Widget         UseCustomColorW, ColorSwatchW;
	EcColorPanel*  ColorPanel;

	Widget         VectorDrawScaleW;

	EIndex         CVIndex;
} E3dSplineResourcesGUI;


static void  _SplineUpdatePanel(E3dObjectPanel* PPanel);


//========================================
// Update Spline stats label
//========================================
static void _SplineUpdateStatsLabel(E3dObjectPanel* PPanel)
{
	E3dSpline*               LSpline = (E3dSpline*)(PPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)PPanel, E3dSpline_class);if(LClassDrawer)
	{
		XeString                LXeStr;
		char                    LTmpStr[256];
		E3dSplineResourcesGUI*  LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);

/*
		switch(LSpline->SplineType)
		{
			case E3dSplineLINEAR:
				if(LSpline->Closed) sprintf(LTmpStr, "LINEAR\n%6lld\n%6lld\n%f", (long long)(LSpline->CVs_Count), (long long)(LSpline->CVs_Count), LSpline->Length);
				else sprintf(LTmpStr, "LINEAR\n%6lld\n%6lld\n%f", (long long)(LSpline->CVs_Count), (long long)(LSpline->CVs_Count - 1), LSpline->Length);
			break;

			case E3dSplineBEZIER:
				if(LSpline->Closed) sprintf(LTmpStr, "BEZIER\n%6lld\n%6lld\n%f", (long long)(LSpline->CVs_Count), (long long)(LSpline->CVs_Count), LSpline->Length);
				else sprintf(LTmpStr, "BEZIER\n%6lld\n%6lld\n%f", (long long)(LSpline->CVs_Count), (long long)(LSpline->CVs_Count - 1), LSpline->Length);
			break;

			case E3dSplineBSPLINE:
				if(LSpline->Closed) sprintf(LTmpStr, "B-SPLINE\n%6lld\n%6lld\n%f", (long long)(LSpline->CVs_Count), (long long)(LSpline->CVs_Count), LSpline->Length);
				else sprintf(LTmpStr, "B-SPLINE\n%6lld\n%6lld\n%f", (long long)(LSpline->CVs_Count), (long long)(LSpline->CVs_Count - 3), LSpline->Length);
			break;

			case E3dSplineCARDINAL:
				if(LSpline->Closed) sprintf(LTmpStr, "CARDINAL\n%6lld\n%6lld\n%f", (long long)(LSpline->CVs_Count), (long long)(LSpline->CVs_Count), LSpline->Length);
				else sprintf(LTmpStr, "CARDINAL\n%6lld\n%6lld\n%f", (long long)(LSpline->CVs_Count), (long long)(LSpline->CVs_Count - 3), LSpline->Length);
			break;
		}
*/
		if(LSpline->Closed) sprintf(LTmpStr, "%lld / %lld / %f", (long long)(LSpline->CVs_Count), (long long)(LSpline->CVs_Count), LSpline->Length);
		else sprintf(LTmpStr, "%6lld / %lld / %f", (long long)(LSpline->CVs_Count), LSpline->CVs_Count > 0 ? (long long)(LSpline->CVs_Count - 1) : 0, LSpline->CVs_Count ? LSpline->Length : 0.0);


		LXeStr = XeString_NewVa(TRUE,
						XeFONT, XeFontDEFAULT,
						XeALIGNMENT, XeSTRING_ALIGN_RIGHT,
						XeTEXT, LTmpStr,
						0);
		if(LXeStr)
		{
			EArg  Args[2];int  ArgCnt;

			EXeStart;
			EXeSetArg(XeNlabelString, LXeStr);
			Xe_SetValues(LResourcesGUI->StatsValLW,  Args, ArgCnt);

			XeString_Free(LXeStr);
		}
	}
}


//========================================================
// Update Spline-segment stuff in the Spline dialog-box
//========================================================
static void _UpdateSplineSegmentDisplay(E3dObjectPanel* PPanel)
{
	E3dSpline*               LSpline = (E3dSpline*)(PPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)PPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI;


	if(LClassDrawer == NULL) return;
	LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);

	if(LResourcesGUI->CVIndex < 0) LResourcesGUI->CVIndex = 0;
	else if(LResourcesGUI->CVIndex >= LSpline->CVs_Count) LResourcesGUI->CVIndex = LSpline->CVs_Count - 1;

	if(LSpline->CVs)
	{
		E3dSplineCV*  LCVs = LSpline->CVs;
		E3dSplineCV*  LCV = LCVs + LResourcesGUI->CVIndex;

		XeToggleButton_SetState(LResourcesGUI->SegmentVisibleW, (LCV->SegmentFlags & E3dSegmentVISIBLE) != 0, FALSE);
		XeToggleButton_SetState(LResourcesGUI->ArcLengthTesselateW, (LCV->SegmentFlags & E3dSegmentARC_LENGTH_TESSELATE) != 0, FALSE);

		switch(LSpline->SplineType)
		{
			case E3dSplineBEZIER:
				{
					XeScale_SetValue(LResourcesGUI->SegmentDrawStepsW, LCV->SegmentNumOfLinSteps, FALSE);

					int  LC = 0;const int  LN = LSpline->CVs_Count;for(;LC < LN;LC++)
					{
						if((LCVs + LC) == LCV)
						{
							LCVs[LC].CVFlags |= E3dCV_ACTIVE;
							LCVs[LC].SegmentFlags |= E3dSegmentSELECTED;
						}
						else
						{
							LCVs[LC].CVFlags &= ~E3dCV_ACTIVE;
							LCVs[LC].SegmentFlags &= ~E3dSegmentSELECTED;
						}
					}


					switch(LCV->SegmentShape)
					{
					 case E3dSplineLINEAR:
						 XeToggleButton_SetState(LResourcesGUI->LinSegmentBW, TRUE, FALSE);
						 XeToggleButton_SetState(LResourcesGUI->BezierSegmentBW, FALSE, FALSE);
					 break;

					 case E3dSplineBEZIER:
						 XeToggleButton_SetState(LResourcesGUI->LinSegmentBW, FALSE, FALSE);
						 XeToggleButton_SetState(LResourcesGUI->BezierSegmentBW, TRUE, FALSE);
					 break;
					}

					if(LCV->SegmentShape == E3dSplineLINEAR) Xe_WidgetSetSensitive(LResourcesGUI->SegmentDrawStepsW, FALSE);  // Composite Widgets need this called (rather than just SetValues)
					else Xe_WidgetSetSensitive(LResourcesGUI->SegmentDrawStepsW, TRUE);  // Composite Widgets need this called (rather than just SetValues)

					switch(LCV->Constraint)
					{
					 case E3dBezierC0:
						 XeToggleButton_SetState(LResourcesGUI->C0TangentConstraintBW, TRUE, FALSE);
						 XeToggleButton_SetState(LResourcesGUI->C1TangentConstraintBW, FALSE, FALSE);
						 XeToggleButton_SetState(LResourcesGUI->G1TangentConstraintBW, FALSE, FALSE);
					 break;

					 case E3dBezierC1:
						 XeToggleButton_SetState(LResourcesGUI->C0TangentConstraintBW, FALSE, FALSE);
						 XeToggleButton_SetState(LResourcesGUI->C1TangentConstraintBW, TRUE, FALSE);
						 XeToggleButton_SetState(LResourcesGUI->G1TangentConstraintBW, FALSE, FALSE);
					 break;

					 case E3dBezierG1:
						 XeToggleButton_SetState(LResourcesGUI->C0TangentConstraintBW, FALSE, FALSE);
						 XeToggleButton_SetState(LResourcesGUI->C1TangentConstraintBW, FALSE, FALSE);
						 XeToggleButton_SetState(LResourcesGUI->G1TangentConstraintBW, TRUE, FALSE);
					 break;
					}
				}
			break;

			default:
				if(LSpline->CVs)
				{
					XeScale_SetValue(LResourcesGUI->SegmentDrawStepsW, LCV->SegmentNumOfLinSteps, FALSE);

					int  LC = 0;const int  LN = LSpline->CVs_Count;for(;LC < LN;LC++)
					{
						if((LCVs + LC) == LCV)
						{
							LCVs[LC].CVFlags |= E3dCV_ACTIVE;
							LCVs[LC].SegmentFlags |= E3dSegmentSELECTED;
						}
						else
						{
							LCVs[LC].CVFlags &= ~E3dCV_ACTIVE;
							LCVs[LC].SegmentFlags &= ~E3dSegmentSELECTED;
						}
					}
				}
			break;
		}
	}
	else XeScale_SetValue(LResourcesGUI->SegmentDrawStepsW, 0, FALSE);
}


static void _DCB_SplineToggleUniform(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)(LPanel->Object);

	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline)
	{
		LSpline->UniformLinSteps = (EBool)(((XeToggleButtonCallbackStruct*)PCallData)->Set);

		EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
		_SplineUpdateStatsLabel(LPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_SplineSteps(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LClassDrawer)
	{
		E3dSplineResourcesGUI*  LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);

		LSpline->UniformLinSteps = TRUE;
		XeToggleButton_SetState(LResourcesGUI->UniformStepsBW, TRUE, FALSE);

		LSpline->NumOfLinSteps = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

		EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
		_SplineUpdateStatsLabel(LPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_SplineTension(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LClassDrawer)
	{
		LSpline->Tension = ((XeScaleCallbackStruct*)PCallData)->Value.f32;

		EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
		_SplineUpdateStatsLabel(LPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_SplineClosed(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)(LPanel->Object);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline)
	{
		LSpline->Closed = (EBool)(((XeToggleButtonCallbackStruct*)PCallData)->Set);

		EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
		_SplineUpdateStatsLabel(LPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


//================================================
// Shift CVs backward
//================================================
static void _DCB_CVsBack(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)(LPanel->Object);

	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline->CVs_Count)
	{
		E3dSplineCV   LCV;memcpy(&LCV, LSpline->CVs, sizeof(E3dSplineCV));

		memmove(LSpline->CVs, LSpline->CVs + 1, sizeof(E3dSplineCV) * (LSpline->CVs_Count - 1));
		memcpy(LSpline->CVs + LSpline->CVs_Count - 1, &LCV, sizeof(E3dSplineCV));

		EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
		_SplineUpdatePanel(LPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


//================================================
// Shift CVs forward
//================================================
static void _DCB_CVsForward(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)(LPanel->Object);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline->CVs_Count)
	{
		E3dSplineCV   LCV;memcpy(&LCV, LSpline->CVs + LSpline->CVs_Count - 1, sizeof(E3dSplineCV));

		memmove(LSpline->CVs + 1, LSpline->CVs, sizeof(E3dSplineCV) * (LSpline->CVs_Count - 1));
		memcpy(LSpline->CVs, &LCV, sizeof(E3dSplineCV));

		EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
		_SplineUpdatePanel(LPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_SplineSelectCV(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline)
	{
		LResourcesGUI->CVIndex = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

		_UpdateSplineSegmentDisplay(LPanel);
		E3dWindows_RedrawByChangeFlags(E3dChgSELECTION);
	}
}


static void _DCB_SegmentVisible(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*  LCV = LSpline->CVs + LResourcesGUI->CVIndex;

		if(((XeToggleButtonCallbackStruct*)PCallData)->Set) LCV->SegmentFlags |= E3dSegmentVISIBLE;
		else LCV->SegmentFlags &= ~E3dSegmentVISIBLE;

		EChangeFlags  LChanged = E3dChgSELECTION;
		LChanged |= E3dSpline_CallOutputs(LSpline, E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY, FALSE);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_ArcLengthTesselate(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*  LCV = LSpline->CVs + LResourcesGUI->CVIndex;

		if(((XeToggleButtonCallbackStruct*)PCallData)->Set) LCV->SegmentFlags |= E3dSegmentARC_LENGTH_TESSELATE;
		else LCV->SegmentFlags &= ~E3dSegmentARC_LENGTH_TESSELATE;

		E3dSpline_CreateLinearSegments(LSpline);

		EChangeFlags  LChanged = E3dChgSELECTION;
		LChanged |= E3dSpline_CallOutputs(LSpline, E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY, FALSE);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_SplineChangeSegmentSteps(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LClassDrawer)
	{
		E3dSplineResourcesGUI*  LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);
		if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
		{
			E3dSplineCV*  LCV = LSpline->CVs + LResourcesGUI->CVIndex;

			LSpline->UniformLinSteps = FALSE;
			XeToggleButton_SetState(LResourcesGUI->UniformStepsBW, FALSE, FALSE);

			LCV->SegmentNumOfLinSteps = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

			EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
			_SplineUpdateStatsLabel(LPanel);

			E3dWindows_RedrawByChangeFlags(LChanged);
		}
	}
}


static void _SetBezierSegmentType(E3dObjectPanel* PPanel, E3dSpline* PSpline, E3dSplineCV* PCV, int PType)
{
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)PPanel, E3dSpline_class);if(LClassDrawer)
	{
		E3dSplineResourcesGUI*  LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);

		PCV->SegmentShape = PType;

		if(PCV->SegmentShape == E3dSplineLINEAR) Xe_WidgetSetSensitive(LResourcesGUI->SegmentDrawStepsW, FALSE);  // Composite Widgets need this called (rather than just SetValues)
		else Xe_WidgetSetSensitive(LResourcesGUI->SegmentDrawStepsW, TRUE);  // Composite Widgets need this called (rather than just SetValues)

		EChangeFlags  LChanged = E3dSpline_FullUpdate(PSpline);
		_SplineUpdateStatsLabel(PPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_SetBezierSegmentTypeLinear(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*  LCV = LSpline->CVs + LResourcesGUI->CVIndex;
		_SetBezierSegmentType(LPanel, LSpline, LCV, E3dSplineLINEAR);
	}
}


static void _DCB_SetBezierSegmentTypeBezier(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		if(LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
		{
			E3dSplineCV*  LCV = LSpline->CVs + LResourcesGUI->CVIndex;
			_SetBezierSegmentType(LPanel, LSpline, LCV, E3dSplineBEZIER);
		}
	}
}


static void _DCB_SetBezierSegmentTypeArc(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);

	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*	LCV = LSpline->CVs + LResourcesGUI->CVIndex;

		_SetBezierSegmentType(LPanel, LSpline, LCV, E3dSplineARC);
	}
}


static void _DCB_SetSegmentShapeForAll(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);

	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*	LCV = LSpline->CVs + LResourcesGUI->CVIndex;
		const int     LSegmentType = LCV->SegmentShape;

		EIndex  LC = 0;const EIndex  LN = LSpline->CVs_Count;LCV = LSpline->CVs;for(;LC < LN;LC++, LCV++) LCV->SegmentShape = LSegmentType;

		EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
		_SplineUpdateStatsLabel(LPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}



static void _SetTangentConstraintType(E3dObjectPanel* PPanel, E3dSpline* PSpline, E3dSplineCV* LCV, const int PType)
{
	LCV->Constraint = PType;

	EChangeFlags  LChanged = E3dSpline_FullUpdate(PSpline);
	_SplineUpdateStatsLabel(PPanel);

	E3dWindows_RedrawByChangeFlags(LChanged);
}


static void _DCB_SetTangentConstraintTypeC0(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*  LCV = LSpline->CVs + LResourcesGUI->CVIndex;

		_SetTangentConstraintType(LPanel, LSpline, LCV, E3dBezierC0);
	}
}


static void _DCB_SetTangentConstraintTypeC1(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*  LCV = LSpline->CVs + LResourcesGUI->CVIndex;

		_SetTangentConstraintType(LPanel, LSpline, LCV, E3dBezierC1);
	}
}


static void _DCB_SetTangentConstraintTypeG1(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);


	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*  LCV = LSpline->CVs + LResourcesGUI->CVIndex;

		_SetTangentConstraintType(LPanel, LSpline, LCV, E3dBezierG1);
	}
}


static void _DCB_SetConstraintTypeForAll(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);

	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*	LCV = LSpline->CVs + LResourcesGUI->CVIndex;
		const int     LConstraintType = LCV->Constraint;

		EIndex  LC = 0;const EIndex  LN = LSpline->CVs_Count;LCV = LSpline->CVs;for(;LC < LN;LC++, LCV++) LCV->Constraint = LConstraintType;

		EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
		_SplineUpdateStatsLabel(LPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


static void _DCB_SetDrawingStepsForAll(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);

	if(LSpline->LODStep)
	{
		if(EObject_IsSubClass(LSpline->LODs[LSpline->LODStep - 1], E3dSpline)) LSpline = (E3dSpline*)(LSpline->LODs[LSpline->LODStep - 1]);
		else LSpline = NULL;
	}

	if(LSpline && LSpline->CVs_Count > 0 && LResourcesGUI->CVIndex >= 0 && LResourcesGUI->CVIndex < LSpline->CVs_Count)
	{
		E3dSplineCV*	LCV = LSpline->CVs + LResourcesGUI->CVIndex;
		const int     LSegmentNumOfLinSteps = LCV->SegmentNumOfLinSteps;

		EIndex  LC = 0;const EIndex  LN = LSpline->CVs_Count;LCV = LSpline->CVs;for(;LC < LN;LC++, LCV++) LCV->SegmentNumOfLinSteps = LSegmentNumOfLinSteps;

		EChangeFlags  LChanged = E3dSpline_FullUpdate(LSpline);
		_SplineUpdateStatsLabel(LPanel);

		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


//========================================
// Update Spline dialog-box
//========================================
static void _SplineUpdatePanel(E3dObjectPanel* LPanel)
{
	E3dSpline*               LSpline = (E3dSpline*)(LPanel->Object);
	XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((const XeInfoPanel*)LPanel, E3dSpline_class);
	if(LClassDrawer == NULL) return;
	E3dSplineResourcesGUI*   LResourcesGUI = (E3dSplineResourcesGUI*)(LClassDrawer->ResourcesGUI);

	_SplineUpdateStatsLabel(LPanel);

	XeToggleButton_SetState(LResourcesGUI->UniformStepsBW, LSpline->UniformLinSteps, FALSE);
	XeToggleButton_SetState(LResourcesGUI->ClosedBW, LSpline->Closed, FALSE);
	XeScale_SetValue(LResourcesGUI->StepsScaleW, LSpline->NumOfLinSteps, FALSE);

	_UpdateSplineSegmentDisplay(LPanel);


	EBool  LSplineHasNormals = FALSE;


	{
		EIndex  LC = 0;const EIndex  LN = LSpline->CVs_Count;const E3dSplineCV*  LCV = LSpline->CVs;for(;LC < LN;LC++, LCV++)
		{
			if(LCV->Normal.X != 0.0 || LCV->Normal.Y != 0.0 || LCV->Normal.Z != 0.0) { LSplineHasNormals = TRUE;break; }
		}
	}

	{
		XeChangeManage   LWidgetsToManageUnmanage[16];
		XeChangeManage*  LChild = LWidgetsToManageUnmanage;

		if(LSplineHasNormals) XeWidget_Manage(LResourcesGUI->VectorDrawScaleW);else XeWidget_Unmanage(LResourcesGUI->VectorDrawScaleW);

		switch(LSpline->SplineType)
		{
			case E3dSplineLINEAR:
				LChild = LWidgetsToManageUnmanage;
				LChild->W = LResourcesGUI->UniformStepsBW;            LChild->Manage = FALSE;LChild++;
				LChild->W = LResourcesGUI->StepsScaleW;               LChild->Manage = FALSE;LChild++;
				LChild->W = LResourcesGUI->CardinalTensionW;          LChild->Manage = FALSE;LChild++;
				XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);

				LChild = LWidgetsToManageUnmanage;
				LChild->W = LResourcesGUI->SegmentSettingsSeparatorW; LChild->Manage = FALSE;LChild++;
				LChild->W = LResourcesGUI->SegmentDrawStepsMatrixW;   LChild->Manage = FALSE;LChild++;
				LChild->W = LResourcesGUI->SegmentShapeMatrixW;       LChild->Manage = FALSE;LChild++;
				LChild->W = LResourcesGUI->CVTangentConstraintMatrixW;LChild->Manage = FALSE;LChild++;
				XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);

				XeWidget_Unmanage(LResourcesGUI->ArcLengthTesselateW);
			break;

			case E3dSplineBEZIER:
				LChild = LWidgetsToManageUnmanage;
				LChild->W = LResourcesGUI->UniformStepsBW;            LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->StepsScaleW;               LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->CardinalTensionW;          LChild->Manage = FALSE;LChild++;
				XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);

				LChild = LWidgetsToManageUnmanage;
				LChild->W = LResourcesGUI->SegmentSettingsSeparatorW; LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->SegmentDrawStepsMatrixW;   LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->SegmentShapeMatrixW;       LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->CVTangentConstraintMatrixW;LChild->Manage = TRUE;LChild++;
				XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);

				XeWidget_Manage(LResourcesGUI->ArcLengthTesselateW);
			break;

			case E3dSplineCARDINAL:
				LChild = LWidgetsToManageUnmanage;
				LChild->W = LResourcesGUI->UniformStepsBW;            LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->StepsScaleW;               LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->CardinalTensionW;          LChild->Manage = TRUE;LChild++;
				XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);

				LChild = LWidgetsToManageUnmanage;
				LChild->W = LResourcesGUI->SegmentSettingsSeparatorW; LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->SegmentDrawStepsMatrixW;   LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->SegmentShapeMatrixW;       LChild->Manage = FALSE;LChild++;
				LChild->W = LResourcesGUI->CVTangentConstraintMatrixW;LChild->Manage = FALSE;LChild++;
				XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);

				XeWidget_Manage(LResourcesGUI->ArcLengthTesselateW);
			break;

			default:
				LChild = LWidgetsToManageUnmanage;
				LChild->W = LResourcesGUI->UniformStepsBW;            LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->StepsScaleW;               LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->CardinalTensionW;          LChild->Manage = FALSE;LChild++;
				XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);

				LChild = LWidgetsToManageUnmanage;
				LChild->W = LResourcesGUI->SegmentSettingsSeparatorW; LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->SegmentDrawStepsMatrixW;   LChild->Manage = TRUE;LChild++;
				LChild->W = LResourcesGUI->SegmentShapeMatrixW;       LChild->Manage = FALSE;LChild++;
				LChild->W = LResourcesGUI->CVTangentConstraintMatrixW;LChild->Manage = FALSE;LChild++;
				XeWidgets_ChangeManage(LWidgetsToManageUnmanage, LChild - LWidgetsToManageUnmanage);

				XeWidget_Manage(LResourcesGUI->ArcLengthTesselateW);
			break;
		}
	}
}


static void _CB_SplineLinear(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)LPanel->Object;

	if(LSpline->SplineType != E3dSplineLINEAR)
	{
		LSpline->SplineType = E3dSplineLINEAR;

		EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass((E3dObject*)LSpline);

		LChanged |= E3dObject_UpdateForDisplay((E3dObject*)LSpline, E3dUF_ALL, LScene);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}

static void _CB_SplineBezier(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)LPanel->Object;

	if(LSpline->SplineType != E3dSplineBEZIER)
	{
		LSpline->SplineType = E3dSplineBEZIER;

		EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass((E3dObject*)LSpline);

// The Prev and Next fields of CVs may have not been initialized, so set them to Position.
//
/*
		{
			EIndex  LC = 0;const EIndex  LN = LSpline->CVs_Count;E3dSplineCV*  LCV = LSpline->CVs;for(;LC < LN;LC++, LCV++)
			{
				LCV->Previous = LCV->Position;
				LCV->Next     = LCV->Position;
			}
		}
*/

		LChanged |= E3dObject_UpdateForDisplay((E3dObject*)LSpline, E3dUF_ALL, LScene);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}

static void _CB_SplineBSpline(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)LPanel->Object;

	if(LSpline->SplineType != E3dSplineBSPLINE)
	{
		LSpline->SplineType = E3dSplineBSPLINE;

		EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass((E3dObject*)LSpline);

		LChanged |= E3dObject_UpdateForDisplay((E3dObject*)LSpline, E3dUF_ALL, LScene);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}

static void _CB_SplineCardinal(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*        LScene = E3d_Scene;
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)LPanel->Object;

	if(LSpline->SplineType != E3dSplineCARDINAL)
	{
		LSpline->SplineType = E3dSplineCARDINAL;

		EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass((E3dObject*)LSpline);

		LChanged |= E3dObject_UpdateForDisplay((E3dObject*)LSpline, E3dUF_ALL, LScene);
		E3dWindows_RedrawByChangeFlags(LChanged);
	}
}


//========================================
// Color ColorPanel callback
//========================================
static void _CB_ColorChanged(EcColorPanel* PColorPanel, EPointer PClientData, EPointer PCallData)
{
	EcColorPanelCallbackStruct*  LCPCBR = (EcColorPanelCallbackStruct*)PCallData;
	int                          LColorIndex;

	if(LCPCBR->NumOfColors > 0) LColorIndex = LCPCBR->ColorIDs[0];
	else return;

	E3dObjectPanel*  LObjectPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LObjectPanel->Object;E3dSpline*  LSpline = E_DynamicCast(LObject, E3dSpline);if(LSpline)
	{
		XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LObjectPanel, E3dSpline_class);if(LDrawer)
		{
			E3dSplineResourcesGUI*  LResourcesGUI = (E3dSplineResourcesGUI*)(LDrawer->ResourcesGUI);

			const EBool  LColorWasInactive = LSpline->RGBAfColor.A < 0.0 ? TRUE : FALSE;

			LSpline->RGBAfColor = PColorPanel->RGBAfColors[LColorIndex];

// If the user changed the Spline's color for the first time (it was "inactive" before), set it to "active"
//
			if(LColorWasInactive)
			{
//printf("%s() %s:%d %sACTIVATE%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout);
				LSpline->RGBAfColor.A = 1.0;
				EcColorPanel_SetRGBAfColor(LResourcesGUI->ColorPanel, 0, &LSpline->RGBAfColor);
			}

//printf("%s() %s:%d  <%s>  RGBAfColor % 7.2f % 7.2f % 7.2f % 7.2f\n", __FUNCTION__, __FILE__, __LINE__, LSpline->Name, RGBA_s(LSpline->RGBAfColor));fflush(stdout);

			EcRGBAiColor  LRGBAiColor;Ec_RGBAf32ToRGBAi(&LSpline->RGBAfColor, &LRGBAiColor);
			LRGBAiColor.A = 0xFFFF;  // LSpline->RGBAfColor may have a 0 alpha, which would affect the drawing of this Widget
			Xe_SetValuesVa(LResourcesGUI->ColorSwatchW,  XeNbackgroundColor, &LRGBAiColor,  NULL);
		}

		E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
	}
}


static void _DCB_NormalDrawScale(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)(LPanel->Object);

	LSpline->VectorDrawScale = ((XeScaleCallbackStruct*)PCallData)->Value.f32;

	E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
}


//========================================
// Spline dialog callback handler
//========================================
static void _DCB_ColorEnable(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dSpline*       LSpline = (E3dSpline*)LPanel->Object;

	LSpline->UseCustomColor = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	E3dWindows_RedrawByChangeFlags(E3dChgWIREFRAME);
}


//================================================================
// GUIProc for E3dSpline
//================================================================
static void _SplineGUI(ECoreObject* PObject, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	E3dSpline*               LSpline = (E3dSpline*)PObject;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dSplineResourcesGUI*   LResourcesGUI;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dSplineResourcesGUI*)(LDrawer->ResourcesGUI);

				if(LPanel->PointIndex >= 0) LResourcesGUI->CVIndex = LPanel->PointIndex;
				else
				{
// LPanel->PointIndex == -1 means, "don't set it", so just do a range check here
//
					if(LResourcesGUI->CVIndex >= LSpline->CVs_Count) LResourcesGUI->CVIndex = LSpline->CVs_Count - 1;
				}

				XeScale_SetValue_f32(LResourcesGUI->CardinalTensionW, LSpline->Tension, FALSE);
				XeScale_SetValueAndRange(LResourcesGUI->CVScaleW,  LResourcesGUI->CVIndex,  0, LSpline->CVs_Count - 1,  FALSE);
				_SplineUpdatePanel(LPanel);

				XeToggleButton_SetState(LResourcesGUI->SplineLinearW,   LSpline->SplineType == E3dSplineLINEAR,   FALSE);
				XeToggleButton_SetState(LResourcesGUI->SplineBezierW,   LSpline->SplineType == E3dSplineBEZIER,   FALSE);
				XeToggleButton_SetState(LResourcesGUI->SplineCardinalW, LSpline->SplineType == E3dSplineCARDINAL, FALSE);
				XeToggleButton_SetState(LResourcesGUI->SplineBSplineW,  LSpline->SplineType == E3dSplineBSPLINE,  FALSE);

		    XeScale_SetValue_f32(LResourcesGUI->VectorDrawScaleW,  LSpline->VectorDrawScale,  FALSE);

				EcRGBAiColor  LRGBAiColor;Ec_RGBAf32ToRGBAi(&LSpline->RGBAfColor, &LRGBAiColor);
				LRGBAiColor.A = 0xFFFF;  // LSpline->RGBAfColor may have a 0 alpha, which would affect the drawing of this Widget
				Xe_SetValuesVa(LResourcesGUI->ColorSwatchW,  XeNbackgroundColor, &LRGBAiColor,  NULL);

				EcColorPanel_SetRGBAfColor(LResourcesGUI->ColorPanel, LResourcesGUI->ColorPanel->DrawColor0, &LSpline->RGBAfColor);
				XeToggleButton_SetState(LResourcesGUI->UseCustomColorW, LSpline->UseCustomColor, FALSE);
			}
			else
			{
				Widget        LTopMatrixW = LDrawer->Drawer.DrawerW, LW;
				Widget        LIndentWs[7];int  LIndentN;    // For indenting the ScaleWidgets
				Widget        LTraverseWs[7];int  LTraverseN;  // For TextField traversal
				EcRGBAiColor  LButtonRGBAiColor, LButtonActiveRGBAiColor, LButtonArmRGBAiColor;
				EArg          Args[256];int  ArgCnt;


				ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

				if(LPanel->PointIndex < 0) LResourcesGUI->CVIndex = 0;
				else LResourcesGUI->CVIndex = LPanel->PointIndex;

				LTraverseN = 0;

				{
// Spline types
//
					EXeStart;
					EXeSetArg(XeNradioBehavior, TRUE);
					EXeSetArg(XeNmarginRight, 8);
					EXeSetArg(XeNspacingX, 8);
					EXeSetArg(XeNadjustEntryWidths, FALSE);
					EXeSetArg(XeNadjustEntryHeights, FALSE);
					Widget  LRadioMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);


					EXeStart;
					EXeSetArg(XeNfontList, Xe_LabelFontList);
					XeWidget_ManagedNew("Type", XeLabelWidget_class, LRadioMatrixW,  Args, ArgCnt);


					EXeStart;
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LSpline->SplineType == E3dSplineLINEAR);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LResourcesGUI->SplineLinearW = LW = XeWidget_ManagedNew("Linear", XeToggleButton_class, LRadioMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _CB_SplineLinear, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LSpline->SplineType == E3dSplineBEZIER);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LResourcesGUI->SplineBezierW = LW = XeWidget_ManagedNew("Bezier", XeToggleButton_class, LRadioMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _CB_SplineBezier, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LSpline->SplineType == E3dSplineBSPLINE);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LResourcesGUI->SplineBSplineW = LW = XeWidget_ManagedNew("B-spline", XeToggleButton_class, LRadioMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _CB_SplineBSpline, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LSpline->SplineType == E3dSplineCARDINAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LResourcesGUI->SplineCardinalW = LW = XeWidget_ManagedNew("Cardinal", XeToggleButton_class, LRadioMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _CB_SplineCardinal, (EPointer)LPanel);
				}


				EXeStart;
				EXeSetArg(XeNshadowThickness, 0);
				EXeSetArg(XeNhighlightThickness, 0);
				EXeSetArg(XeNspacingTop, 4);EXeSetArg(XeNspacingBottom, 2);
				EXeSetArg(XeNset, LSpline->Closed);
				LResourcesGUI->ClosedBW = LW = XeWidget_ManagedNew("Closed", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SplineClosed, (EPointer)LPanel);


				{
					EXeStart;
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginSizeY, 3);EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNspacingX, 6);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;

					XeString  LXeStr = XeString_NewVa(TRUE,
					 XeFONT, XeFontDEFAULT,
					 XeALIGNMENT, XeSTRING_ALIGN_LEFT,
					 XeTEXT, "CVs / segments / length:  ",
					 0);

					if(LXeStr) { EXeSetArg(XeNlabelString, LXeStr); }

					XeWidget_ManagedNew("", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);
					if(LXeStr) XeString_Free(LXeStr);

					EXeStart;
					LResourcesGUI->StatsValLW = XeWidget_ManagedNew("", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);
				}

// Shift CVs
				{
					EXeStart;
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginTop, 5);EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNspacingX, 8);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNfontList, Xe_LabelFontList);
					XeWidget_ManagedNew("Shift CVs", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNsizeX, 24);EXeSetArg(XeNsizeY, 24);
					EXeSetArg(XeNarrowDirection, XeLEFT);
					EXeSetArg(XeNarrowBehavior, XeARROW_ALWAYS_IN);
					LW = XeWidget_ManagedNew(" ", XeArrowButtonWidget_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_CVsBack, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNsizeX, 24);EXeSetArg(XeNsizeY, 24);
					EXeSetArg(XeNarrowDirection, XeRIGHT);
					EXeSetArg(XeNarrowBehavior, XeARROW_ALWAYS_IN);
					LW = XeWidget_ManagedNew(" ", XeArrowButtonWidget_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_CVsForward, (EPointer)LPanel);
				}


				{
					EXeStart;
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginSizeY, 3);EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNspacingX, 4);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LSpline->UniformLinSteps);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					LResourcesGUI->UniformStepsBW = LW = XeWidget_ManagedNew("Uniform drawing steps", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SplineToggleUniform, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 5);
					EXeSetArg(XeNminimum, 2);EXeSetArg(XeNmaximum, 1000);EXeSetArg(XeNvalue, LSpline->NumOfLinSteps);
					EXeSetArg(XeNscaleMinimum, 2);EXeSetArg(XeNscaleMaximum, 32);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					LTraverseWs[LTraverseN++] = LResourcesGUI->StepsScaleW = LW = XeWidget_ManagedNew(NULL, XeScaleWidget_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SplineSteps, (EPointer)LPanel);
				}

				LIndentN = 0;

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->CardinalTensionW = LW = XeScale_Newf32(LTopMatrixW, "Tension",  Args, ArgCnt);
				XeScale_SetValueAndRange_f32(LW,  LSpline->Tension,  -2.0, 1.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SplineTension, (EPointer)LPanel);


//				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);


				EXeStart;
				EXeSetArg(XeNshadowThickness, 1);EXeSetArg(XeNshadowType, XeShadowIN);
				EXeSetArg(XeNorientation, XeVERTICAL);
				EXeSetArg(XeNspacingX, 4);EXeSetArg(XeNspacingY, 0);
				EXeSetArg(XeNmarginSizeX, 6);EXeSetArg(XeNmarginSizeY, 6);
				LResourcesGUI->SegmentSettingsMatrixW = XeWidget_ManagedNew("_SplineSegmentSettingsMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 4);
				EXeSetArg(XeNminimum, 0);EXeSetArg(XeNmaximum, LSpline->CVs_Count - 1);EXeSetArg(XeNvalue, LResourcesGUI->CVIndex);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				EXeSetArg(XeNcreateUpButton, TRUE);
				EXeSetArg(XeNcreateDownButton, TRUE);
				LResourcesGUI->CVScaleW = LTraverseWs[LTraverseN++] = LW = XeWidget_ManagedNew("CV/segment", XeScaleWidget_class, LResourcesGUI->SegmentSettingsMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SplineSelectCV, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNshadowThickness, 2);
				EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LResourcesGUI->SegmentSettingsSeparatorW = XeWidget_ManagedNew("", XeSeparatorWidget_class, LResourcesGUI->SegmentSettingsMatrixW,  Args, ArgCnt);


				{
					EXeStart;
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginSizeY, 3);EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNspacingX, 8);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LResourcesGUI->SegmentSettingsMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					XeToolTip8("Segment is visible / hidden. If a segment is not visible, tools like Extrude will not create a corresponding side.");
					LResourcesGUI->SegmentVisibleW = LW = XeWidget_ManagedNew("Segment visible", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SegmentVisible, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					XeToolTip8("Create equal-length linear segments to approximate this segment.");
					LResourcesGUI->ArcLengthTesselateW = LW = XeWidget_ManagedNew("Proportional", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_ArcLengthTesselate, (EPointer)LPanel);
				}

				Ec_GainRGBAi(&Xe_BackgroundRGBAiColor, 1.0, &LButtonRGBAiColor);
				Ec_GainRGBAi(&Xe_BackgroundRGBAiColor, XeActiveColorFactor, &LButtonActiveRGBAiColor);
				Ec_GainRGBAi(&Xe_BackgroundRGBAiColor, XeArmColorFactor, &LButtonArmRGBAiColor);


				{
					EXeStart;
					EXeSetArg(XeNradioBehavior, TRUE);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginTop, 6);
					EXeSetArg(XeNspacingX, 7);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LResourcesGUI->SegmentSettingsMatrixW,  Args, ArgCnt);
					LResourcesGUI->SegmentShapeMatrixW = LMatrixW;

					LIndentN = 0;
					EXeStart;
					EXeSetArg(XeNwidthAdjustable, FALSE);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					LResourcesGUI->SegTypeLabelW = LIndentWs[LIndentN++] = XeWidget_ManagedNew("Segment shape", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNindicatorOn, FALSE);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNset, FALSE);
					EXeSetArg(XeNwidthAdjustable, FALSE);
					EXeSetArg(XeNactivateOnArm, TRUE);
					EXeSetArg(XeNlabelType, XeIMAGE);
					EXeSetImages3(_Image_SplSegmentLin);
					XeToolTip8("Linear segment");
					LResourcesGUI->LinSegmentBW = LW = XeWidget_ManagedNew("", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SetBezierSegmentTypeLinear, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNindicatorOn, FALSE);
//					EXeSetArg(XeNindicatorSize, 18);
//					EXeSetArg(XeNbuttonType, XeButtonTypeLED);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNset, FALSE);
					EXeSetArg(XeNwidthAdjustable, FALSE);
					EXeSetArg(XeNactivateOnArm, TRUE);
					EXeSetArg(XeNlabelType, XeIMAGE);
					EXeSetImages3(_Image_SplSegmentCrv);
					XeToolTip8("Bezier-curve segment");
					LResourcesGUI->BezierSegmentBW = LW = XeWidget_ManagedNew("", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SetBezierSegmentTypeBezier, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNindicatorOn, FALSE);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNset, FALSE);
					EXeSetArg(XeNwidthAdjustable, FALSE);
					EXeSetArg(XeNactivateOnArm, TRUE);
					EXeSetArg(XeNlabelType, XeIMAGE);
					EXeSetImages3(_Image_SplSegmentArc);
					XeToolTip8("Circular arc segment");
					LResourcesGUI->ArcSegmentBW = LW = XeWidget_ManagedNew("", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SetBezierSegmentTypeArc, (EPointer)LPanel);



					EXeStart;
					EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Set this shape for all segments.");
					LW = XeWidget_ManagedNew("All", XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_SetSegmentShapeForAll, (EPointer)LPanel);
				}


				{
					EXeStart;
					EXeSetArg(XeNradioBehavior, TRUE);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNspacingX, 7);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LResourcesGUI->SegmentSettingsMatrixW,  Args, ArgCnt);
					LResourcesGUI->CVTangentConstraintMatrixW = LMatrixW;

					LIndentN = 0;
					EXeSetArg(XeNwidthAdjustable, FALSE);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					LIndentWs[LIndentN++] = XeWidget_ManagedNew("Continuity constraint", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNindicatorOn, FALSE);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNset, FALSE);
					EXeSetArg(XeNwidthAdjustable, FALSE);
					EXeSetArg(XeNactivateOnArm, TRUE);
					EXeSetArg(XeNlabelType, XeIMAGE);EXeSetImages3(_Image_TgC0);
					XeToolTip8("\"C0\" constraint:\nTangents may have different lengths and directions.");
					LResourcesGUI->C0TangentConstraintBW = LW = XeWidget_ManagedNew("", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SetTangentConstraintTypeC0, (EPointer)LPanel);


					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNindicatorOn, FALSE);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNset, FALSE);
					EXeSetArg(XeNwidthAdjustable, FALSE);
					EXeSetArg(XeNactivateOnArm, TRUE);
					EXeSetArg(XeNlabelType, XeIMAGE);
					EXeSetImages3(_Image_TgG1);
					XeToolTip8("\"G1\" constraint\nTangents may have different lengths, but the directions match.");
					LResourcesGUI->G1TangentConstraintBW = LW = XeWidget_ManagedNew("", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SetTangentConstraintTypeG1, (EPointer)LPanel);


					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNindicatorOn, FALSE);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNset, FALSE);
					EXeSetArg(XeNwidthAdjustable, FALSE);
					EXeSetArg(XeNactivateOnArm, TRUE);
					EXeSetArg(XeNlabelType, XeIMAGE);
					EXeSetImages3(_Image_TgC1);
					XeToolTip8("\"C1\" constraint\nTangent directions and lenghts match.");
					LResourcesGUI->C1TangentConstraintBW = LW = XeWidget_ManagedNew("", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SetTangentConstraintTypeC1, (EPointer)LPanel);

//				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 4);

					EXeStart;
					EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
					EXeSetArg(XeNbackgroundColorPtr, &Xe_ButtonRGBAiColor);
					EXeSetArg(XeNfontList, Xe_LabelFontList);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Set this for all CVs.");
					LW = XeWidget_ManagedNew("All", XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_SetConstraintTypeForAll, (EPointer)LPanel);
				}


				{
					EXeStart;
					EXeSetArg(XeNradioBehavior, TRUE);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNspacingX, 2);
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LResourcesGUI->SegmentSettingsMatrixW,  Args, ArgCnt);
					LResourcesGUI->SegmentDrawStepsMatrixW = LMatrixW;

					EXeStart;
					EXeSetArg(XeNtextFieldDigits, 4);
					EXeSetArg(XeNminimum, 2);EXeSetArg(XeNmaximum, 1000);EXeSetArg(XeNvalue, LSpline->NumOfLinSteps);
					EXeSetArg(XeNscaleMinimum, 2);EXeSetArg(XeNscaleMaximum, 32);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginBottom, 5);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SegmentDrawStepsW = LW = XeWidget_ManagedNew("Drawing steps", XeScaleWidget_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SplineChangeSegmentSteps, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
					EXeSetArg(XeNbackgroundColorPtr, &Xe_ButtonRGBAiColor);
					EXeSetArg(XeNfontList, Xe_LabelFontList);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Set this for all CVs / segments.");
					LW = XeWidget_ManagedNew("All", XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_SetDrawingStepsForAll, (EPointer)LPanel);
				}


// Color Drawer
//
				{
					EXeStart;
					EXeSetArg(XeNmarginTop, 4);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNspacingX, 4);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("ColorMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					XeDrawer*  LColorDrawer = &LResourcesGUI->ColorDrawer;XeDrawer_NewVa(LMatrixW, LColorDrawer, "Color", 0, 0);

					LResourcesGUI->ColorSwatchW = EcColorSwatch_NewRGBAf32(NULL, LMatrixW, NULL, &LSpline->RGBAfColor, XeCB_DrawerOpenClose, (EPointer)LColorDrawer);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNspacingLeft, 4);EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNset, LSpline->UseCustomColor);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Use the color specified here, for the drawing this Spline.\nWhen this is OFF, use the default wireframe color.");
					LResourcesGUI->UseCustomColorW = LW = XeWidget_ManagedNew("Enable", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_ColorEnable, (EPointer)PPanel);

					EXeStart;
					EXeSetArg(XeNspacingLeft, 8);
					EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 3);
					EXeSetArg(XeNincrement, 10);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNheightAdjustable, FALSE);
					LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->VectorDrawScaleW = LW = XeScaleNew_Coord(LMatrixW, "Normal scale:",  Args, ArgCnt);
					XeScale_SetValueAndRanges_Coord(LW,  LSpline->VectorDrawScale,  0.001, 1000.0,  0.001, 100.0,  FALSE);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_NormalDrawScale, (EPointer)LPanel);


					EcColorPanel*  LColorPanel;
					LResourcesGUI->ColorPanel = LColorPanel = EcColorPanel_NewVa(LTopMatrixW, "Color",  XeNcolorChangedCallback, _CB_ColorChanged,  XeNcolorChangedCallbackClientData, LPanel,
					 XeNalphaOn, TRUE,
					 XeNcolorComponentType, EcFLOAT_0_1,
					 XeNRGBAfColor, &LSpline->RGBAfColor,
					 XeNintensityOn, TRUE,  XeNcreateColorWheel, TRUE,  NULL);

					LColorDrawer->DrawerW = LColorPanel->TopMatrixW;
				}

				Xe_TraversalRing(LTraverseWs, LTraverseN);
			}

			_SplineUpdatePanel(LPanel);


// If it's a contour Spline of the Face that's attributes are being shown, don't display the "Geometry" subclass folder
//
			if(LResourcesGUI)
			{
				if(LSpline->Faces)
				{
					EBool  LHideGeometryPanel = FALSE;

					EIndex  LPC = 0;for(;LPC < E3dp_ObjectPanels_Count;LPC++)
					{
						E3dObjectPanel*  LTPanel = E3dp_ObjectPanels[LPC];
						EIndex  LC = 0;const EIndex  LN = LSpline->Faces_Count;for(;LC < LN;LC++)
						{
							E3dObject*  LFace = LSpline->Faces[LC];
							if(LTPanel->Object == LFace) { LHideGeometryPanel = TRUE;goto DoneChecking; }
						}
					}


DoneChecking:
					if(LHideGeometryPanel)
					{
						XeComponentClassDrawer*  LClassDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, E3dGeometry_class);if(LClassDrawer)
						{
							LClassDrawer->Show = FALSE;
						}
					}
				}
			}
		break;

		case E_FREE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dSplineResourcesGUI*)(LDrawer->ResourcesGUI);
				if(LResourcesGUI->ColorPanel) { EcColorPanel_Free(LResourcesGUI->ColorPanel);LResourcesGUI->ColorPanel = NULL; }
			}
		break;
	}
}


//================================================
// Constructor
//================================================
static int _EdgesToSplines_Init(E3dNode* PNode, E3dNodeComponent* PInfo)
{
	E3dEdgesToSplines*  LThis = (E3dEdgesToSplines*)PInfo;

	EList_Init(LThis->SrcObjects);
	EList_Init(LThis->SrcModels);
	LThis->SrcActiveFlags = NULL;

	LThis->DoHiddenEdges = FALSE;
	LThis->DoUnselectedEdges = TRUE;
	LThis->MergeContiguous = TRUE;
	LThis->SplineNormalsMode = 1;

	EList_InitA(LThis->Splines);

// Initialize InputInterface for _InputChanged()
//
	{
		EInterface*  LInput = &LThis->InputInterface;E_InterfaceInit(LInput);
		LInput->InputChanged    = _EdgesToSplines_InputChanged;
		LInput->Object          = (ECoreObject*)PNode;
		LInput->ObjectComponent = (ECoreObject*)PInfo;
	}

	return(E_SUCCESS);
}

//========================================
// Copy-constructor
//========================================
static int _EdgesToSplines_Copy(const E3dNode* PSrcNode, const E3dNodeComponent* PSrcComponent,  E3dNode* PDstNode, E3dNodeComponent* PDstComponent)
{
	E3dEdgesToSplines*  LSInfo = (E3dEdgesToSplines*)PSrcComponent;
	E3dEdgesToSplines*  LDInfo = (E3dEdgesToSplines*)PDstComponent;

	_EdgesToSplines_Init(PDstNode, PDstComponent);

	if(LSInfo->SrcObjects_Count)
	{
		ENewN(LSInfo->SrcObjects_Count, LDInfo->SrcActiveFlags);
		const EBool  LHasSrcActiveFlags = LSInfo->SrcActiveFlags != NULL;

		EIndex  LC = 0;const EIndex  LN = LSInfo->SrcObjects_Count;for(;LC < LN;LC++)
		{
			EList_ObjectAdd(LDInfo->SrcObjects, LSInfo->SrcObjects[LC]);
			if(LHasSrcActiveFlags) LDInfo->SrcActiveFlags[LC] = LSInfo->SrcActiveFlags[LC];
			else LDInfo->SrcActiveFlags[LC] = TRUE;
		}
	}

	{
		EIndex  LC = 0;const EIndex  LN = LSInfo->SrcModels_Count;for(;LC < LN;LC++)
		{
			EList_ObjectAdd(LDInfo->SrcModels, LSInfo->SrcModels[LC]);
		}
	}

#define _CopyField(mField)  LDInfo->mField = LSInfo->mField
	_CopyField(DoHiddenEdges);
	_CopyField(DoUnselectedEdges);
	_CopyField(SplineNormalsMode);
	_CopyField(MergeContiguous);
#undef _CopyField

	return(E_SUCCESS);
}

//================================================
// Destructor
//================================================
static int _EdgesToSplines_Free(E3dNode* PNode, E3dNodeComponent* PInfo)
{
	E3dEdgesToSplines*  LThis = (E3dEdgesToSplines*)PInfo;

	{
		EIndex  LC = 0;const EIndex  LN = LThis->SrcObjects_Count;for(;LC < LN;LC++)
		{
			if(LThis->SrcObjects[LC]) E3dBaseObject_RemoveOutput((E3dBaseObject*)LThis->SrcObjects[LC], &LThis->InputInterface);
		}
	}
	EList_ObjectsFree(LThis->SrcObjects);

	{
		EIndex  LC = 0;const EIndex  LN = LThis->SrcModels_Count;for(;LC < LN;LC++)
		{
			if(LThis->SrcModels[LC]) E3dBaseObject_RemoveOutput((E3dBaseObject*)LThis->SrcModels[LC], &LThis->InputInterface);
		}
	}
	EList_ObjectsFree(LThis->SrcModels);

	EFreeChkClr(LThis->SrcActiveFlags);

	EList_ObjectsFreeA(LThis->Splines);

	return(E_SUCCESS);
}

static EChangeFlags _EdgesToSplines_Update(E3dNode* PChangedBranchRootNode, E3dNode* PNode, E3dNodeComponent* PInfo, const EUpdateFlags PUpdateFlags, const EBool PCheckOnly)
{
	E3dEdgesToSplines*  LInfo = (E3dEdgesToSplines*)PInfo;

	if(PCheckOnly)
	{
		if(LInfo->SrcObjects_Count > 0)
		{
			ENewN(LInfo->SrcObjects_Count, LInfo->SrcActiveFlags);
			EIndex  LC = 0;const EIndex  LN = LInfo->SrcObjects_Count;for(;LC < LN;LC++)
			{
				LInfo->SrcActiveFlags[LC] = TRUE;

				E3dBaseObject_AddOutput((E3dBaseObject*)LInfo->SrcObjects[LC], &LInfo->InputInterface);
				E3dBaseObject_AddOutput((E3dBaseObject*)LInfo->SrcModels[LC], &LInfo->InputInterface);
			}
		}
		return(0);
	}

	EChangeFlags  LChangeFlags = 0;
	E3dModel*     LModel = (E3dModel*)PNode;

	{
		EIndex  LC = 0;const EIndex  LN = LInfo->Splines_Count;for(;LC < LN;LC++)
		{
			E3dSpline*  LSpline = LInfo->Splines[LC];
			E3dModel_RemoveObject(LModel, (E3dObject*)LSpline);
		}
	}

	const EBool  LDoHiddenEdges = LInfo->DoHiddenEdges, LDoUnselectedEdges = LInfo->DoUnselectedEdges, LMergeContiguous = LInfo->MergeContiguous;

	uint8_t  LSplineNormalsMode = LInfo->SplineNormalsMode;
	if(LMergeContiguous && LSplineNormalsMode == E3dSnEDGE) LSplineNormalsMode = E3dSnVERTEX;

	const EBool  LDoTransform = FALSE;

	{
		EIndex  LC = 0;const EIndex  LN = LInfo->SrcObjects_Count;for(;LC < LN;LC++)
		{
			const E3dObject*  LSrcObject = LInfo->SrcObjects[LC];E3dMesh*  LSrcMesh = E_DynamicCast(LSrcObject, E3dMesh);if(LSrcMesh)
			{
				const E3dModel*  LSrcModel = LInfo->SrcModels[LC];
				E3dMatrix        LLocalToWorldMatrix;E3dMatrix_Copy(LSrcModel->LocalToWorldMatrix, LLocalToWorldMatrix);
				const EIndex     LVN = LSrcMesh->Vertices_Count;
				E3dCoord3*       LDVertices = NULL;
				const float      LVectorDrawScale = LSrcMesh->PolyGroups_Count > 0 ? LSrcMesh->PolyGroups[0]->VectorDrawScale : 1.0;

				if(LVN)
				{
					ENewN(LVN, LDVertices);
					const E3dVertexSize  LVertexSize = LSrcMesh->VertexSize;
					if(LDoTransform)
					{
						EIndex  LVC = 0;const E3dVertex*  LSVertex = LSrcMesh->Vertices;E3dCoord3*  LDVertex = LDVertices;for(;LVC < LVN;LVC++, EStruct_Inc(LSVertex, LVertexSize), LDVertex++)
						{
							E3dMatrix_Transform3x4_vp(LLocalToWorldMatrix, LSVertex->Position, LDVertex);
						}
					}
					else
					{
						EIndex  LVC = 0;const E3dVertex*  LSVertex = LSrcMesh->Vertices;E3dCoord3*  LDVertex = LDVertices;for(;LVC < LVN;LVC++, EStruct_Inc(LSVertex, LVertexSize), LDVertex++) *LDVertex = LSVertex->Position;
					}
				}

// Check if there are any selected Edges on LMesh. If so, only create Splines for those
//
				EIndex  LNSelectedEdges = 0;
				{ EIndex  LEdgeC = 0;const EIndex  LEdgeN = LSrcMesh->Edges_Count;const E3dEdge*  LEdge = LSrcMesh->Edges;for(;LEdgeC < LEdgeN;LEdgeC++, LEdge++) if(LEdge->Flags & E3dEdgeSELECTED) LNSelectedEdges++; }


				EBool  LWeCreatedWingedEdges = FALSE, LWeCreatedVertexLinks = FALSE;
				if(LSrcMesh->WingedEdges == NULL) { E3dMesh_CreateWingedEdgesFromEdges(LSrcMesh, TRUE, FALSE);LWeCreatedWingedEdges = TRUE;LWeCreatedVertexLinks = TRUE; }
				else if(LSrcMesh->VertexLinks == NULL) { E3dMesh_VertexLinksBuild(LSrcMesh);LWeCreatedVertexLinks = TRUE; }

//if(LWeCreatedVertexLinks) { printf("%s() %s:%d %sLWeCreatedVertexLinks%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout); }
//if(LWeCreatedWingedEdges) { printf("%s() %s:%d %sLWeCreatedWingedEdges%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout); }


// Count Edges
//
				const EIndex  LEN = LSrcMesh->Edges_Count;
				EIndex  LNTotalSplinesNeeded = 0;

				{
					EIndex  LEC = 0;E3dEdge*  LEdge = LSrcMesh->Edges;for(;LEC < LEN;LEC++, LEdge++)
					{
						if((LDoHiddenEdges || (LEdge->Flags & E3dEdgeVISIBLE)) && (LDoUnselectedEdges || (LEdge->Flags & E3dEdgeSELECTED)))
						{
							LEdge->Flags |= E3dEdgePASS;   // We'll use this flag to identify which Edges should be processed
							LEdge->Flags |= E3dEdgeMARK0;  // We'll use this flag to identify which Edges have not been processed yet (cleared after an Edge is processed)
							LNTotalSplinesNeeded++;
						}
						else LEdge->Flags &= ~(E3dEdgePASS|E3dEdgeMARK0);
					}
				}

// Add Splines to LInfo->Splines, if needed
//
				if(LInfo->Splines_Count < LNTotalSplinesNeeded)
				{
					EIndex  LEC = LInfo->Splines_Count;for(;LEC < LNTotalSplinesNeeded;LEC++)
					{
						E3dSpline*  LSpline;E_New(E3dSpline, LSpline);LSpline->SplineType = E3dSplineLINEAR;
						EList_ObjectAddA(LInfo->Splines, 256, LSpline);
					}
				}

				E3dCoord3Sum*  LEdgeNormalSums = NULL;
				E3dCoord3Sum*  LVertexNormalSums = NULL;

				switch(LSplineNormalsMode)
				{
					case E3dSnVERTEX:  LVertexNormalSums = E3dMesh_GetVertexNormals(LSrcMesh, E3dMeshVertexNormalFROM_EDGES);break;
					case E3dSnEDGE:    LEdgeNormalSums = E3dMesh_GetEdgeNormals(LSrcMesh);break;
				}

				{
					E3dEdge*  LSrcMeshEdges = LSrcMesh->Edges;

					EIndex  LEdgeC = 0, LEAC = 0;E3dEdge*  LEdge = LSrcMesh->Edges;for(;LEdgeC < LEN;LEdgeC++, LEdge++)
					{
						if(LEdge->Flags & E3dEdgeMARK0)
						{
							LEdge->Flags &= ~E3dEdgeMARK0;

							E3dSpline*  LSpline = LInfo->Splines[LEAC];LSpline->SplineType = E3dSplineLINEAR;
							if(LSpline->CVs_Count != 2)
							{
								EList_Resize(LSpline->CVs, 2);
							}

							E3dSplineCVs_Init(LSpline->CVs, E3dSplineLINEAR, LSpline->CVs_Count);
							E3dSplineCV*  LCV = LSpline->CVs;


							{
								E3dCoord3  LEdgeNormal;
								if(LSplineNormalsMode == E3dSnEDGE)
								{
									if(LDoTransform) { const E3dCoord3  LNml = LEdgeNormalSums[LEdgeC].Vector;E3dMatrix_Transform3x3_vv(LLocalToWorldMatrix, LNml, LEdgeNormal); }
									else LEdgeNormal = LEdgeNormalSums[LEdgeC].Vector;
								}


								{
									const EIndex  LVertexIndex = LEdge->Start;

									LCV->Position = LDVertices[LVertexIndex];

									if(LDoTransform)
									{
										switch(LSplineNormalsMode)
										{
											case E3dSnVERTEX:  { const E3dCoord3  LNml = LVertexNormalSums[LVertexIndex].Vector;E3dMatrix_Transform3x3_vv(LLocalToWorldMatrix, LNml, LCV->Normal); } break;
											case E3dSnEDGE:    LCV->Normal = LEdgeNormal;break;
										}
									}
									else
									{
										switch(LSplineNormalsMode)
										{
											case E3dSnVERTEX:  LCV->Normal = LVertexNormalSums[LVertexIndex].Vector;break;
											case E3dSnEDGE:    LCV->Normal = LEdgeNormal;break;
										}
									}
								}

								LCV++;

								{
									const EIndex  LVertexIndex = LEdge->End;

									LCV->Position = LDVertices[LVertexIndex];

									if(LDoTransform)
									{
										switch(LSplineNormalsMode)
										{
											case E3dSnVERTEX:  { const E3dCoord3  LNml = LVertexNormalSums[LVertexIndex].Vector;E3dMatrix_Transform3x3_vv(LLocalToWorldMatrix, LNml, LCV->Normal); } break;
											case E3dSnEDGE:    LCV->Normal = LEdgeNormal;break;
										}
									}
									else
									{
										switch(LSplineNormalsMode)
										{
											case E3dSnVERTEX:  LCV->Normal = LVertexNormalSums[LVertexIndex].Vector;break;
											case E3dSnEDGE:    LCV->Normal = LEdgeNormal;break;
										}
									}
								}
							}

							LEdge->Flags &= ~E3dEdgeMARK0;


							if(LMergeContiguous)
							{
// 1  62 20 63  2

//printf("%s() %s:%d  %3lld -> %3lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LEdge->Start, (long long)LEdge->End);fflush(stdout);
// Walk from the Start and End points along the VertexLinks, until we hit Edges with valence > 2
// When evaluating the valence, we have to take Edge visibility, etc. into consideration.
// This simply means checking the E3dEdgeMARK0 flag on the Edge, as we set this flag above, for Edges that need to be considered
//
								{
									EIndex                LPrevVertexIndex = LEdge->Start;  // This is used for preventing back-tracking
									EIndex                LVertexIndex = LEdge->End;
									const E3dVertexLink*  LVertexLink = LSrcMesh->VertexLinks + LVertexIndex;

									for(;1;)
									{
// Find the first marked Edge, and check if this Vertex's valence is 2
//
										EIndex  LEVC = 0, LValence = 0, LNextVertexIndex = -1, LNextEdgeIndex = -1;const EIndex  LEVN = LVertexLink->EdgeEnds_Count;const E3dEdgeEnd*  LEE = LVertexLink->EdgeEnds;for(;LEVC < LEVN;LEVC++, LEE++)
										{
											const EIndex  LTNextVertexIndex = LEE->VertexIndex, LTNextEdgeIndex = LEE->EdgeIndex;
											if(LSrcMeshEdges[LTNextEdgeIndex].Flags & E3dEdgePASS)
											{
												LValence++;if(LValence > 2) break;
//printf("%s() %s:%d  LValence %3lld  LTNextVertexIndex %3lld  LTNextEdgeIndex %3lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LValence, (long long)LTNextVertexIndex, (long long)LTNextEdgeIndex);fflush(stdout);
												if(LTNextVertexIndex != LPrevVertexIndex) { LNextVertexIndex = LTNextVertexIndex;LNextEdgeIndex = LTNextEdgeIndex; }
											}
										}
//printf("%s() %s:%d  LVertexIndex %3lld  LValence %3lld  LNextVertexIndex %3lld  LNextEdgeIndex %3lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LVertexIndex, (long long)LValence, (long long)LNextVertexIndex, (long long)LNextEdgeIndex);fflush(stdout);
										if(LValence == 2)
										{
											E3dEdge*  LNEdge = LSrcMeshEdges + LNextEdgeIndex;if(LNEdge->Flags & E3dEdgeMARK0)
											{
												LNEdge->Flags &= ~E3dEdgeMARK0;
												E3dSplineCV*  LNCV = EList_StructAdd(LSpline->CVs);
												E3dSplineCV_Init(LNCV, E3dSplineLINEAR);
												LNCV->Position = LDVertices[LNextVertexIndex];
												if(LSplineNormalsMode == E3dSnVERTEX)
												{
													if(LDoTransform) { const E3dCoord3  LNml = LVertexNormalSums[LNextVertexIndex].Vector;E3dMatrix_Transform3x3_vv(LLocalToWorldMatrix, LNml, LNCV->Normal); }
													else LNCV->Normal = LVertexNormalSums[LNextVertexIndex].Vector;
												}

												LVertexLink = LSrcMesh->VertexLinks + LNextVertexIndex;
												LPrevVertexIndex = LVertexIndex;
												LVertexIndex = LNextVertexIndex;
											}
											else break;
										}
										else break;
									}
								}

								{
									EIndex                LPrevVertexIndex = LEdge->End;  // This is used for preventing back-tracking
									EIndex                LVertexIndex = LEdge->Start;
									const E3dVertexLink*  LVertexLink = LSrcMesh->VertexLinks + LVertexIndex;

									for(;1;)
									{
// Find the first marked Edge, and check if this Vertex's valence is 2
//
										EIndex  LEVC = 0, LValence = 0, LNextVertexIndex = -1, LNextEdgeIndex = -1;const EIndex  LEVN = LVertexLink->EdgeEnds_Count;const E3dEdgeEnd*  LEE = LVertexLink->EdgeEnds;for(;LEVC < LEVN;LEVC++, LEE++)
										{
											const EIndex  LTNextVertexIndex = LEE->VertexIndex, LTNextEdgeIndex = LEE->EdgeIndex;
											if(LSrcMeshEdges[LTNextEdgeIndex].Flags & E3dEdgePASS)
											{
												LValence++;if(LValence > 2) break;
												if(LTNextVertexIndex != LPrevVertexIndex) { LNextVertexIndex = LTNextVertexIndex;LNextEdgeIndex = LTNextEdgeIndex; }
											}
										}
//printf("%s() %s:%d  LVertexIndex %3lld  LValence %3lld  LNextVertexIndex %3lld  LNextEdgeIndex %3lld\n", __FUNCTION__, __FILE__, __LINE__, (long long)LVertexIndex, (long long)LValence, (long long)LNextVertexIndex, (long long)LNextEdgeIndex);fflush(stdout);
										if(LValence == 2)
										{
											E3dEdge*  LNEdge = LSrcMeshEdges + LNextEdgeIndex;if(LNEdge->Flags & E3dEdgeMARK0)
											{
												LNEdge->Flags &= ~E3dEdgeMARK0;
												E3dSplineCV*  LNCV = EList_StructInsert(LSpline->CVs, 0);
												E3dSplineCV_Init(LNCV, E3dSplineLINEAR);
												LNCV->Position = LDVertices[LNextVertexIndex];
												if(LSplineNormalsMode == E3dSnVERTEX)
												{
													if(LDoTransform) { const E3dCoord3  LNml = LVertexNormalSums[LNextVertexIndex].Vector;E3dMatrix_Transform3x3_vv(LLocalToWorldMatrix, LNml, LNCV->Normal); }
													else LNCV->Normal = LVertexNormalSums[LNextVertexIndex].Vector;
												}

												LVertexLink = LSrcMesh->VertexLinks + LNextVertexIndex;
												LPrevVertexIndex = LVertexIndex;
												LVertexIndex = LNextVertexIndex;
											}
											else break;
										}
										else break;
									}
								}
							}


							LSpline->VectorDrawScale = LVectorDrawScale;
							E3dSpline_CreateLinearSegments(LSpline);
							LEAC++;

							E3dModel_AppendObject(LModel, (E3dObject*)LSpline);
						}
					}
				}

				EFreeChkClr(LEdgeNormalSums);
				EFreeChkClr(LVertexNormalSums);

				if(E3dp_Prefs.OptimizeFor == EOptimizeMEMORY)
				{
					if(LWeCreatedWingedEdges) E3dMesh_WingedEdgesFree(LSrcMesh);
					if(LWeCreatedVertexLinks) E3dMesh_VertexLinksFree(LSrcMesh);
				}

//			E3dSpline*  LSpline = LInfo->Splines[LC];

				EFreeChkClr(LDVertices);
			}
		}

		LChangeFlags |= E3dChgSPLINESHAPE|E3dChgSPLINECONNECTIVITY|E3dChgHIERARCHY;
	}

	return(LChangeFlags);
}


//================================================================
// Callback for input changes
// This function should Not call any 3DWindow redrawing function,
// just return with TRUE if a redraw is needed!
//================================================================
static EChangeFlags _EdgesToSplines_InputChanged(EInterface* PInterface, EInterfaceCallbackStruct* PCBS)
{
	E3dNode*            LNode = (E3dNode*)PInterface->Object;
	E3dEdgesToSplines*  LThis = (E3dEdgesToSplines*)E3dNodeComponent_ByClass(LNode, _EdgesToSplines_class);
	EChangeFlags        LChangeFlags = _EdgesToSplines_Update(NULL, LNode, (E3dNodeComponent*)LThis, E3dUF_ALL, FALSE);
/*
	E3dScene_ObjectsUpdateModelList(LScene);
	E3dScene_LayoutOnSchematics(LScene, FALSE);
	if(LChangeFlags) E3dWindows_RedrawByChangeFlags(LChangeFlags);
	E3dNodePanel_UpdateValues(LPanel, NULL);
*/


/*
	E3dEdgesToSplines*  LInfo    = (E3dEdgesToSplines*)PInterface->ObjectComponent;
	EChangeFlags        LChanged = E3dObject_GetChangeFlagsByClass(LObject);

// Only respond if the Object is instantiated (e.g. not in the undo stack after being deleted)
//
	if(LObject->Models_Count)
	{
		E3dObject_CustomUpdate(LObject, LInfo, PCBS->CallOutputs ? E3dUF_ALL : E3dUF_ALL - E3dUF_CALL_OUTPUTS, &LChanged);
	}
*/

	return(LChangeFlags);
}


#define _CB(m)\
static void _EdgesToSplines_CB_##m(Widget PW, EPointer PClientData, EPointer PCallData)\
{\
	E3dNodePanel*       LPanel = (E3dNodePanel*)PClientData;\
	E3dNode*            LNode = LPanel->Node;\
	E3dEdgesToSplines*  LThis = (E3dEdgesToSplines*)E3dNodeComponent_ByClass(LNode, _EdgesToSplines_class);\
	E3dScene*           LScene = E3d_Scene;\
\
	LThis->m = ((XeToggleButtonCallbackStruct*)PCallData)->Set;\
\
	EChangeFlags  LChangeFlags = _EdgesToSplines_Update(NULL, LNode, (E3dNodeComponent*)LThis, E3dUF_ALL, FALSE);\
	E3dScene_ObjectsUpdateModelList(LScene);\
	E3dScene_LayoutOnSchematics(LScene, FALSE);\
	if(LChangeFlags) E3dWindows_RedrawByChangeFlags(LChangeFlags);\
	E3dNodePanel_UpdateValues(LPanel, NULL);\
}


_CB(DoHiddenEdges)
_CB(DoUnselectedEdges)
_CB(MergeContiguous)

static void _EdgesToSplines_CB_SplineNormalsNone(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*       LPanel = (E3dNodePanel*)PClientData;E3dNode*  LNode = LPanel->Node;
	E3dEdgesToSplines*  LThis = (E3dEdgesToSplines*)E3dNodeComponent_ByClass(LNode, _EdgesToSplines_class);
	E3dScene*           LScene = E3d_Scene;

	LThis->SplineNormalsMode = 0;

	EChangeFlags  LChangeFlags = _EdgesToSplines_Update(NULL, LNode, (E3dNodeComponent*)LThis, E3dUF_ALL, FALSE);
	E3dScene_ObjectsUpdateModelList(LScene);
	E3dScene_LayoutOnSchematics(LScene, FALSE);
	if(LChangeFlags) E3dWindows_RedrawByChangeFlags(LChangeFlags);
	E3dNodePanel_UpdateValues(LPanel, NULL);
}

static void _EdgesToSplines_CB_SplineNormalsVertex(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*       LPanel = (E3dNodePanel*)PClientData;E3dNode*  LNode = LPanel->Node;
	E3dEdgesToSplines*  LThis = (E3dEdgesToSplines*)E3dNodeComponent_ByClass(LNode, _EdgesToSplines_class);
	E3dScene*           LScene = E3d_Scene;

	LThis->SplineNormalsMode = E3dSnVERTEX;

	EChangeFlags  LChangeFlags = _EdgesToSplines_Update(NULL, LNode, (E3dNodeComponent*)LThis, E3dUF_ALL, FALSE);
	E3dScene_ObjectsUpdateModelList(LScene);
	E3dScene_LayoutOnSchematics(LScene, FALSE);
	if(LChangeFlags) E3dWindows_RedrawByChangeFlags(LChangeFlags);
	E3dNodePanel_UpdateValues(LPanel, NULL);
}

static void _EdgesToSplines_CB_SplineNormalsEdge(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dNodePanel*       LPanel = (E3dNodePanel*)PClientData;E3dNode*  LNode = LPanel->Node;
	E3dEdgesToSplines*  LThis = (E3dEdgesToSplines*)E3dNodeComponent_ByClass(LNode, _EdgesToSplines_class);
	E3dScene*           LScene = E3d_Scene;

	LThis->SplineNormalsMode = E3dSnEDGE;

	EChangeFlags  LChangeFlags = _EdgesToSplines_Update(NULL, LNode, (E3dNodeComponent*)LThis, E3dUF_ALL, FALSE);
	E3dScene_ObjectsUpdateModelList(LScene);
	E3dScene_LayoutOnSchematics(LScene, FALSE);
	if(LChangeFlags) E3dWindows_RedrawByChangeFlags(LChangeFlags);
	E3dNodePanel_UpdateValues(LPanel, NULL);
}

static void _EdgesToSplines_GUI(E3dNode* PNode, E3dNodeComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dNodePanel*            LPanel = (E3dNodePanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dEdgesToSplines*       LInfo = (E3dEdgesToSplines*)PInfo;
	E3dResourcesGUI*         LResourcesGUI;

	switch(PAction)
	{
		case E_UPDATE:
			if(LInfo)
			{
				if(LDrawer->ResourcesGUI)
				{
					LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

					XeToggleButton_SetState(LResourcesGUI->DoHiddenEdgesW, LInfo->DoHiddenEdges, FALSE);
					XeToggleButton_SetState(LResourcesGUI->DoUnselectedEdgesW, LInfo->DoUnselectedEdges, FALSE);
					XeToggleButton_SetState(LResourcesGUI->MergeContiguousW, LInfo->MergeContiguous, FALSE);
					XeToggleButton_SetState(LResourcesGUI->SplineNormalsModeNoneW, LInfo->SplineNormalsMode == 0, FALSE);
					XeToggleButton_SetState(LResourcesGUI->SplineNormalsModeVertexW, LInfo->SplineNormalsMode == E3dSnVERTEX, FALSE);
					XeToggleButton_SetState(LResourcesGUI->SplineNormalsModeEdgeW, LInfo->SplineNormalsMode == E3dSnEDGE, FALSE);
				}
				else
				{
					Widget  LTopMatrixW = LDrawer->Drawer.DrawerW, LW;
					EArg    Args[256];int  ArgCnt;

					ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

					EXeStart;
					EXeSetArg(XeNheightAdjustable, TRUE);
					EXeSetArg(XeNmarginSizeX, 2);
					EXeSetArg(XeNspacingY, 0);
					Xe_SetValues(LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->DoHiddenEdges);
					XeToolTip8("Use hidden Edges as well.");
					LResourcesGUI->DoHiddenEdgesW = LW = XeWidget_ManagedNew("Hidden Edges", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _EdgesToSplines_CB_DoHiddenEdges, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->DoUnselectedEdges);
					XeToolTip8("Use unselected Edges as well (if FALSE, only selected Edges are used).");
					LResourcesGUI->DoUnselectedEdgesW = LW = XeWidget_ManagedNew("Unselected Edges", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _EdgesToSplines_CB_DoUnselectedEdges, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->MergeContiguous);
					XeToolTip8("Generate a single Spline for non-branching Edge sequences.");
					LResourcesGUI->MergeContiguousW = LW = XeWidget_ManagedNew("Merge continuous", XeToggleButton_class, LTopMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _EdgesToSplines_CB_MergeContiguous, (EPointer)LPanel);

					{
						EXeStart;
						EXeSetArg(XeNorientation, XeHORIZONTAL);
						EXeSetArg(XeNspacingLeft, 4);EXeSetArg(XeNspacingRight, 4);
						EXeSetArg(XeNradioBehavior, TRUE);
						EXeSetArg(XeNspacingX, 4);
						EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
						Widget  LMatrixW = XeWidget_ManagedNew("SplineNormalModeMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

						EXeStart;
						XeWidget_ManagedNew("Normals from", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

						EXeStart;
						EXeSetArg(XeNspacingLeft, 8);
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->SplineNormalsMode == 0);
						XeToolTip8("Don't generate Spline normals.");
						LResourcesGUI->SplineNormalsModeNoneW = LW = XeWidget_ManagedNew("None", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _EdgesToSplines_CB_SplineNormalsNone, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->SplineNormalsMode == 0);
						XeToolTip8("Don't generate Spline normals.");
						LResourcesGUI->SplineNormalsModeVertexW = LW = XeWidget_ManagedNew("Vertex", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _EdgesToSplines_CB_SplineNormalsVertex, (EPointer)LPanel);

						EXeStart;
						EXeSetArg(XeNshadowThickness, 0);
						EXeSetArg(XeNhighlightThickness, 0);
						EXeSetArg(XeNset, LInfo->SplineNormalsMode == 0);
						XeToolTip8("Don't generate Spline normals.");
						LResourcesGUI->SplineNormalsModeEdgeW = LW = XeWidget_ManagedNew("Edge", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
						XeAddCallback(LW, XeNvalueChangedCallback, _EdgesToSplines_CB_SplineNormalsEdge, (EPointer)LPanel);
					}
				}
			}
		break;

		case E_FREE:
		break;
	}
}


//========================================================
// Create linear Splines for Edges of selected Meshes
//========================================================
static void _EdgesToSplines_MCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;

	E3dNodeComponentClass*  LClass = _EdgesToSplines_class;

	E3dObject**  LObjects = NULL;EIndex  LObjects_Count = E3dScene_GetObjects(LScene, &LObjects, E3dItemMESH, E3dSELECTED, FALSE);if(LObjects_Count)
	{
		E3dModel*  LModel;E_New(E3dModel, LModel);EObject_SetName(LModel, E3dScene_GetUniqueNodeName(LScene, LClass->Name), FALSE);

//	E3dModel*  LModel = E3dModel_CustomNew(LScene, LClass, E3dp_Prefs.Duplicate_Flags & E3dCopyMODEL_UNIQUE_NAME ? TRUE : FALSE);if(LModel)

		if(LModel)
		{
			E3dScene_NodeHrcAdd(LScene, (E3dNode*)LModel, E3dUF_CALL_SCENE_CALLBACKS);
			E3dScene_ObjectsUpdateModelList(LScene);
			E3dScene_LayoutOnSchematics(LScene, TRUE);
		}

		if(LModel)
		{
			E3dEdgesToSplines*  LInfo = (E3dEdgesToSplines*)E3dNodeComponent_Add((E3dNode*)LModel, LClass);

			LModel->SchemObject.BranchExpanded = FALSE;

//printf("%s() %s:%d  %s[%p]%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, LInfo, ENormalCode);fflush(stdout);

// Collect source Meshes
//
			{
				EIndex  LC = 0;for(;LC < LObjects_Count;LC++)
				{
					E3dObject*  LObject = LObjects[LC];
					EIndex  LMC = 0;for(;LMC < LObject->Models_Count;LMC++)
					{
						E3dModel*  LSrcModel = LObject->Models[LMC];
						if(LSrcModel->Selection == E3dNodeSelNODE || LSrcModel->Selection == E3dNodeSelBRANCH || LSrcModel->Selection == E3dNodeSelBRANCH_ROOT ||
						  (LSrcModel->Selection == E3dNodeSelOBJECT && LObject->Selection == E3dObjectSelOBJECT))
						{
							EList_ObjectAdd(LInfo->SrcObjects, LObject);
							EList_ObjectAdd(LInfo->SrcModels,  LSrcModel);
						}
					}
				}
			}

			if(LInfo->SrcObjects_Count > 0)
			{
				ENewN(LInfo->SrcObjects_Count, LInfo->SrcActiveFlags);
				EIndex  LC = 0;const EIndex  LN = LInfo->SrcObjects_Count;for(;LC < LN;LC++)
				{
					LInfo->SrcActiveFlags[LC] = TRUE;

					E3dBaseObject_AddOutput((E3dBaseObject*)LInfo->SrcObjects[LC], &LInfo->InputInterface);
					E3dBaseObject_AddOutput((E3dBaseObject*)LInfo->SrcModels[LC], &LInfo->InputInterface);
				}
			}

			LClass->Update(NULL, (E3dNode*)LModel, (E3dNodeComponent*)LInfo, E3dUF_ALL, FALSE);

			E3dScene_ObjectsUpdateModelList(LScene);
			E3dScene_LayoutOnSchematics(LScene, TRUE);

			E3dWindows_RedrawByChangeFlags(E3dChgFACE|E3dChgHIERARCHY);

			if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) E3dNode_Edit((E3dNode*)LModel, _EdgesToSplines_class);
		}
		else
		{
	printf("%s() %s:%d  %sCould not create Model%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
		}

		EFree(LObjects);
	}
	else E3dp_PrintMessage(ELogMESSAGE, 5000, "%s: Select a Mesh as a source", LClass->Name);
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	E3dSpline_class->GUI = _SplineGUI;

	EpM_CreateIconStates3_sp(SplSegmentLin);
	EpM_CreateIconStates3_sp(SplSegmentCrv);
	EpM_CreateIconStates3_sp(SplSegmentArc);
	EpM_CreateIconStates3_sp(TgC0);
	EpM_CreateIconStates3_sp(TgC1);
	EpM_CreateIconStates3_sp(TgG1);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Spline", "Close", '\0', NULL, NULL, FALSE, NULL, _MCB_Splines, (EPointer)CR_SPLINE_CLOSE);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Spline", "Open", '\0', NULL, NULL, FALSE, NULL, _MCB_Splines, (EPointer)CR_SPLINE_OPEN);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Spline", "Draw Linear", '\0', NULL, NULL, FALSE, NULL, _MCB_SplineCreate, (EPointer)E3dSplineLINEAR);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Spline", "Draw Bezier", '\0', NULL, NULL, FALSE, NULL, _MCB_SplineCreate, (EPointer)E3dSplineBEZIER);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Spline", "Draw BSpline", '\0', NULL, NULL, FALSE, NULL, _MCB_SplineCreate, (EPointer)E3dSplineBSPLINE);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Spline", "Draw Cardinal", '\0', NULL, NULL, FALSE, NULL, _MCB_SplineCreate, (EPointer)E3dSplineCARDINAL);
// _MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Spline", "Draw NURBS", '\0', NULL, NULL, FALSE, NULL, _MCB_Splines, (EPointer)E3dSplineNURBS);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Spline", "Edit points", '\0', NULL, NULL, FALSE, NULL, _MCB_Splines, (EPointer)CR_SPLINE_EDIT_POINTS);


	{
		E3dNodeComponentClass  LNodeComponentClass =
		{
			.Name=        "Edges to Splines",
			.StructSize=  sizeof(E3dEdgesToSplines),
			.Resources=   _EdgesToSplines_Resources,
			.Init=        _EdgesToSplines_Init,
			.Copy=        _EdgesToSplines_Copy,
			.Free=        _EdgesToSplines_Free,
			.Update=      _EdgesToSplines_Update,
			.GUI=         _EdgesToSplines_GUI
		};
		_EdgesToSplines_class = E3dNodeComponent_ClassRegister(&LNodeComponentClass);

		_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Spline", "Edges to Splines", '\0', NULL, NULL, TRUE, NULL, _EdgesToSplines_MCB, (EPointer)NULL);
	}


	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC = 0;for(;LC < _MenuButtons_Count;LC++) XeItemFree(_MenuButtons[LC]);
		_MenuButtons_Count = 0;
	}

	if(_EdgesToSplines_class) { E3dNodeComponent_ClassDeactivate(_EdgesToSplines_class);_EdgesToSplines_class = NULL; }

	EpIconFree3(SplSegmentLin);
	EpIconFree3(SplSegmentCrv);
	EpIconFree3(SplSegmentArc);
	EpIconFree3(TgC0);
	EpIconFree3(TgC1);
	EpIconFree3(TgG1);

	E3dSpline_class->GUI = NULL;

	return(E_SUCCESS);
}
