//================================================================
// EQUINOX-3D plugin
//
// Interactive Extrude tool
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 03:10:52
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <assert.h>
#include <math.h>
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Label.h>
#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/PushButton.h>
#include <Xe/Separator.h>
#include <Xe/ToggleButton.h>

#include <E3D/Math.h>
#include <E3D/Matrix.h>
#include <E3D/PolygonEditing.h>
#include <E3D/Select.h>
#include <E3D/Tool.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>
#include <E3D/UI/WidgetCombos.h>


static EPlugin*	_Plugin = NULL;

char*  E_PluginInfo = "Extrude a Spline or Face along a straight line, or along a Spline.";

static E3dObjectComponentClass*  _ObjectComponent_class = NULL;

EguiItem  _MenuButtons[2];   int  _MenuButtons_Count = 0;
EguiItem  _ToolBarButtons[2];int  _ToolBarButtons_Count = 0;


#include "ExtrudeSpline.h"

//----------------------------------------
// Extrude
//----------------------------------------
EpIconDef3(Extrude)

static const char  _NewObjectAddingMode = E3dNEW_INSTANCE;
//static const char  _NewObjectAddingMode = E3dNEW_MODEL;

typedef struct
{
	E3dExtrudeSplineInfoCore;

// Parameters
//
	E3dCoordinate  Depth, Offset;
	E3dCoordinate  Twist;     // Degrees
	E3dCoord2      Scale;     // Gradually scale size along extrusion
	E3dCoord3      Axis;
	short          ProfileFollowing;
	EBool          Profile3D, CenterAlongPath, TwistBevels, ScaleBevels;
} E3dInfo;



#define ROffset(field)   (void*)(EPtrOffset(E3dInfo, field))
#define ROffsetI(field)  EPtrOffset(E3dInfo, field)
static EResource  _ExtrudeResources[]=
{
	{ "ProfileGeometries",            EResOBJECT|EResDYNAMIC,  ROffset(ProfileGeometries),    .Ext.CountOffset= ROffsetI(ProfileGeometries_Count) },

	{ "ProfileFollowing",             EResINT16,               ROffset(ProfileFollowing) },
	{ "Profile3D",                    EResBOOLEAN,             ROffset(Profile3D) },
	{ "PathSteps",                    EResINT32,               ROffset(PathSteps) },

	{ "AxX",                          E3dCOORDINATE_TYPE,      ROffset(Axis.X) },
	{ "AxY",                          E3dCOORDINATE_TYPE,      ROffset(Axis.Y) },
	{ "AxZ",                          E3dCOORDINATE_TYPE,      ROffset(Axis.Z) },
	{ "Depth",                        E3dCOORDINATE_TYPE,      ROffset(Depth) },
	{ "Offset",                       E3dCOORDINATE_TYPE,      ROffset(Offset) },
	{ "Twist",                        E3dCOORDINATE_TYPE,      ROffset(Twist) },
	{ "Scale.X",                      E3dCOORDINATE_TYPE,      ROffset(Scale.X) },
	{ "Scale.Y",                      E3dCOORDINATE_TYPE,      ROffset(Scale.Y) },

	{ "StartBevel.Steps",             EResINT32,               ROffset(StartBevel.NumOfRoundingSteps) },
	{ "StartBevel.Radius",            E3dCOORDINATE_TYPE,      ROffset(StartBevel.Radius) },
	{ "StartBevel.Invert",            EResBOOLEAN,             ROffset(StartBevel.Invert) },
	{ "StartBevel.AddStartSmoother",  EResBOOLEAN,             ROffset(StartBevel.AddStartSmoother) },
	{ "StartBevel.AddEndSmoother",    EResBOOLEAN,             ROffset(StartBevel.AddEndSmoother) },

	{ "EndBevel.Steps",               EResINT32,               ROffset(EndBevel.NumOfRoundingSteps) },
	{ "EndBevel.Radius",              E3dCOORDINATE_TYPE,      ROffset(EndBevel.Radius) },
	{ "EndBevel.Invert",              EResBOOLEAN,             ROffset(EndBevel.Invert) },
	{ "EndBevel.AddStartSmoother",    EResBOOLEAN,             ROffset(EndBevel.AddStartSmoother) },
	{ "EndBevel.AddEndSmoother",      EResBOOLEAN,             ROffset(EndBevel.AddEndSmoother) },

	{ "SmoothingAngle",               EResFLOAT32,             ROffset(SmoothingAngle) },
	{ "Center_along_path",            EResBOOLEAN,             ROffset(CenterAlongPath) },
	{ "Triangles on sides",           EResBOOLEAN,             ROffset(TrianglesOnSides) },
	{ "Triangles on caps",            EResBOOLEAN,             ROffset(TrianglesOnCaps) },
	{ "Create_UVs",                   EResBOOLEAN,             ROffset(CreateUVs) },
	{ "Create_tangents",              EResBOOLEAN,             ROffset(CreateTangents) },
	{ "Cap_start",                    EResBOOLEAN,             ROffset(CloseStart) },
	{ "Cap_end",                      EResBOOLEAN,             ROffset(CloseEnd) },
	{ "Sides",                        EResBOOLEAN,             ROffset(CloseSides) },
	{ "Include bevels in length",     EResBOOLEAN,             ROffset(IncludeBevelsInLength) },
	{ "Twist bevels",                 EResBOOLEAN,             ROffset(TwistBevels) },
	{ "Scale bevels",                 EResBOOLEAN,             ROffset(ScaleBevels) },
	{ }
};
#undef ROffset
#undef ROffsetI


//----------------------------------------
// Extrude along Spline
//----------------------------------------


enum
{
	CR_SET_CLOSE_FRONT = XeCR_CUSTOM0,
	CR_SET_CLOSE_BACK,
	CR_TRIANGLES,
	CR_SET_DEPTH,
	CR_SET_PATH_STEPS,
	CR_CHANGE_SMOOTHING_ANGLE
};


static E3dCoordinate* _GetUs(const E3dExtrudeInfoCommon* PInfo, const E3dSpline* PPathSpline, const E3dCoordinate PPathLength, const int PPathSteps, EIndex* PNumOfUsPtr, E3dCoordinate* PTotalLengthRet)
{
	E3dCoordinate     LPathLength = PPathLength;
	int               LPathSteps  = PPathSteps;
	E3dCoordinate*    LUs;
	E3dCoordinate*    LEndBevelUs = NULL;
	const E3dSpline*  LStartBevelProfile = PInfo->StartBevel.Profile;
	const E3dSpline*  LEndBevelProfile = PInfo->EndBevel.Profile;
	E3dCoordinate     LTotalULength, LStartBevelProfileLength = 0.0, LEndBevelProfileLength = 0.0;
	EIndex            LNumOfUs = 0, LOpenPath = TRUE;


	if(PPathSpline && PPathSpline->SNumOfLinearSegments > 0)
	{
		if(PPathSpline->Closed)
		{
			LNumOfUs = PPathSpline->SNumOfLinearSegments + 1;  // E3dSpline_GetDrawingArcLengthParameters stores U for the end point as well
			LOpenPath = FALSE;
		}
		else
		{
			if(LStartBevelProfile && LStartBevelProfile->SNumOfLinearSegments) LNumOfUs += LStartBevelProfile->SNumOfLinearSegments;
			LNumOfUs += PPathSpline->SNumOfLinearSegments + 1;
			if(LEndBevelProfile && LEndBevelProfile->SNumOfLinearSegments) LNumOfUs += LEndBevelProfile->SNumOfLinearSegments;
			LPathSteps = PPathSpline->SNumOfLinearSegments;
		}
	}
	else
	{
		EIndex  LSideSteps = 0;

		if(LPathSteps) LSideSteps = LPathSteps + 1;
		else LSideSteps = 0;

		if(LStartBevelProfile && LStartBevelProfile->SNumOfLinearSegments) LNumOfUs += LStartBevelProfile->SNumOfLinearSegments;
		if(LSideSteps) LNumOfUs += LSideSteps;
		if(LEndBevelProfile && LEndBevelProfile->SNumOfLinearSegments) LNumOfUs += LEndBevelProfile->SNumOfLinearSegments;
	}


	ENewN(LNumOfUs, LUs);
	*PNumOfUsPtr = LNumOfUs;

// Use arc-length parameterization for the whole StartBevel-Sides-EndBevel curve, as if it was a single Spline
//
// First, arc-length parameterize the bevel Splines and get their
// tesseleted length (sum of the lengths of the linear segments)
//
	if(LStartBevelProfile && LStartBevelProfile->SNumOfLinearSegments && LOpenPath)
	{
		LStartBevelProfileLength = E3dSpline_GetDrawingArcLengthParameters(LStartBevelProfile, 0.0, FALSE,  LUs, NULL);
	}

	if(LEndBevelProfile && LEndBevelProfile->SNumOfLinearSegments && LOpenPath)
	{
		ENewN(LEndBevelProfile->SNumOfLinearSegments + 1, LEndBevelUs);E3dCoordinate*	LUPtr = LEndBevelUs;
		LEndBevelProfileLength = E3dSpline_GetDrawingArcLengthParameters(LEndBevelProfile, 0.0, TRUE,  LUPtr, NULL);
	}


// If a PPathSpline is given, do arc-length parameterization for U
//
	{
		E3dCoordinate*  LUPtr;


		if(LStartBevelProfile && LStartBevelProfile->SNumOfLinearSegments && LOpenPath) LUPtr = LUs + LStartBevelProfile->SNumOfLinearSegments;
		else LUPtr = LUs;

		if(PPathSpline)
		{
			LPathLength = E3dSpline_GetDrawingArcLengthParameters(PPathSpline, 0.0, FALSE,  LUPtr, NULL);
		}

		if(PPathSpline && LPathLength > 0.0)
		{
			E3dCoordinate  LUMult, LUOffset;
			const EIndex   LUN = PPathSpline->SNumOfLinearSegments + 1;


			LTotalULength = LStartBevelProfileLength + LPathLength + LEndBevelProfileLength;
			if(LTotalULength < 0.00001) LTotalULength = 0.00001;

			LUMult = LPathLength / LTotalULength;LUOffset = LStartBevelProfileLength / LTotalULength;
			EIndex  LUC = 0;for(;LUC < LUN;LUC++, LUPtr++)
			{
				*LUPtr = *LUPtr * LUMult + LUOffset;
			}
		}
		else
		{
			E3dCoordinate  LU, LUStep;
			const EIndex   LUN = LPathSteps + 1;


			LTotalULength = LStartBevelProfileLength + LPathLength + LEndBevelProfileLength;
			if(LTotalULength < 0.00001) LTotalULength = 0.00001;
			LUStep = LPathLength / LTotalULength / (E3dCoordinate)LPathSteps;

			EIndex  LUC = 0;for(LU = LStartBevelProfileLength / LTotalULength;LUC < LUN;LUC++, LU += LUStep, LUPtr++)
			{
				*LUPtr = LU;
			}
		}
	}


	{
		E3dCoordinate*  LUPtr = LUs;

		if(LStartBevelProfile && LStartBevelProfile->SNumOfLinearSegments && LOpenPath)
		{
		 E3dCoordinate  LUMult = LStartBevelProfileLength / LTotalULength;
		 EIndex  LUC = 0;const EIndex  LUN = LStartBevelProfile->SNumOfLinearSegments;for(;LUC < LUN;LUC++) { *LUPtr = *LUPtr * LUMult;LUPtr++; }
		}

		if(LEndBevelProfile && LEndBevelProfile->SNumOfLinearSegments && LOpenPath)
		{
			E3dCoordinate   LUMult = LEndBevelProfileLength / LTotalULength, LUOffset = (LStartBevelProfileLength + LPathLength) / LTotalULength;
			E3dCoordinate*  LSUPtr;
			EIndex          LUN = LEndBevelProfile->SNumOfLinearSegments + 1;

			LUPtr += LPathSteps;

			LSUPtr = LEndBevelUs + LEndBevelProfile->SNumOfLinearSegments;
			EIndex  LUC = 0;for(;LUC < LUN;LUC++, LSUPtr--, LUPtr++) *LUPtr = *LSUPtr * LUMult + LUOffset;
		}
	}

	EFreeChk(LEndBevelUs);

	if(PTotalLengthRet) *PTotalLengthRet = LTotalULength;

	return(LUs);
}


void E3d_BevelCreateSegments(E3dBevel* PBevel)
{
	E3dSplineCV*   LStartBevelV;
	E3dSplineCV*   LEndBevelV;
	E3dCoordinate  LAngle = 0.0;
	E3dCoordinate  LBevelStep = (E3dPI * 0.5) / (E3dCoordinate)(PBevel->NumOfRoundingSteps), LRadius;


	PBevel->Type = E3dBevelARC;
	if(PBevel->Profile == NULL)
	{
		E_NewAssign(E3dSpline, PBevel->Profile);
		EObject_SetName(PBevel->Profile, EStrDup("BevelProfile"), FALSE);
	}

	EIndex  LNumOfProfileSteps = PBevel->NumOfRoundingSteps;
	if(PBevel->AddStartSmoother) LNumOfProfileSteps++;
	if(PBevel->AddEndSmoother) LNumOfProfileSteps++;

	E3dSpline_FreeCVs(PBevel->Profile);

	PBevel->Profile->CVs_Count = LNumOfProfileSteps + 1;
	EResizeNT(E3dSplineCV, PBevel->Profile->CVs_Count,  PBevel->Profile->CVs);

// Initialize Spline CVs
//
	{
		E3dSplineCV*  LBevelV = PBevel->Profile->CVs;
		EIndex  LC = 0;const EIndex  LN = LNumOfProfileSteps + 1;for(;LC < LN;LC++, LBevelV++) E3dSplineCV_Init(LBevelV, E3dSplineLINEAR);
	}

	E3dSplineCV*  LBevelV = PBevel->Profile->CVs;
	if(PBevel->AddStartSmoother) LBevelV++;
	LStartBevelV = LBevelV;
	LRadius = PBevel->Radius;

// Do N+1 steps, so we have valid coordinates for the StartSmoother, even if NumOfRoundingSteps == 1 and
// also because _GetUs needs the last CV (0.0, 0.0) to compute the correct length of the bevel profile
//
	EIndex  LC = 0;const EIndex  LN = PBevel->NumOfRoundingSteps + 1;for(;LC < LN;LC++)
	{
		LBevelV->Position.X = -E3d_Cos(LAngle) * LRadius;
		LBevelV->Position.Y = (1.0 - E3d_Sin(LAngle)) * PBevel->Radius;
		LBevelV->Position.Z = 0.0;

//printf("%s() %f,%f\n", __FUNCTION__, LBevelV->Position.X, LBevelV->Position.Y);fflush(stdout);

		LAngle += LBevelStep, LBevelV++;
	}
	LEndBevelV = LBevelV - 2;

	if(PBevel->AddStartSmoother)
	{
// Subtract a percentage of the first bevel-step for the position of the "smoother" Edge
//
		LBevelV = PBevel->Profile->CVs;
		*LBevelV = *LStartBevelV;
		LStartBevelV->Position.Y -= (LStartBevelV[0].Position.Y - LStartBevelV[1].Position.Y) * E3dSMOOTHER;
	}

	if(PBevel->AddEndSmoother)
	{
		LBevelV = PBevel->Profile->CVs + LNumOfProfileSteps - 1;
		LBevelV->Position.X = LEndBevelV->Position.X * E3dSMOOTHER;
		LBevelV->Position.Y = 0.0;
		LBevelV->Position.Z = 0.0;
	}

//printf("%s() %s:%d  PBevel->Profile %p\n", __FUNCTION__, __FILE__, __LINE__, PBevel->Profile);fflush(stdout);
	E3dSpline_CreateLinearSegments(PBevel->Profile);


/*
	{
		E3dCoordinate*  LBev = PBevel->Profile->SLinearSegments;

		EIndex  LC = 0;const EIndex  LN = PBevel->Profile->CVs_Count;for(;LC < LN;LC++, LBev += 3)
		{
printf("%s() %f,%f,%f\n", __FUNCTION__, LBev[0], LBev[1], LBev[2]);fflush(stdout);
		}
	}
*/


/*
	LBevelV = PBevel->Profile->CVs;
	EIndex  LC = 0;const EIndex  LN = PBevel->Profile->CVs_Count;for(;LC < LN;LC++, LBevelV++)
	{
printf("%s() %f,%f\n", __FUNCTION__, LBevelV->Position.X, LBevelV->Position.Y);fflush(stdout);
	}
*/
}


void E3d_BevelInit(E3dBevel* PBevel)
{
	PBevel->Type = 0;
	PBevel->Radius = 0.1;

//	E3dSpline*	LSpline;E_New(E3dSpline, LSpline);E_AssignNoChkNoFree(LSpline, PBevel->Profile);

	PBevel->Profile = NULL;
	PBevel->NumOfRoundingSteps = 0;
	PBevel->Invert = FALSE;
	PBevel->AddStartSmoother = PBevel->AddEndSmoother = TRUE;
}


void E3d_BevelCopy(const E3dBevel* PSBevel, E3dBevel* PDBevel)
{
#define _Copy(m) PDBevel->m = PSBevel->m
	_Copy(Type);
	_Copy(Radius);
	_Copy(NumOfRoundingSteps);
	_Copy(Invert);
	_Copy(AddStartSmoother);
	_Copy(AddEndSmoother);

	PDBevel->Profile = NULL;
	if(PSBevel->Profile)
	{
		E_Assign(E3dObject_Copy((E3dObject*)(PSBevel->Profile), 0, NULL), PDBevel->Profile);
		if(PDBevel->Profile)
		{
			E3dSpline_CreateLinearSegments(PDBevel->Profile);
		}
	}
	else PDBevel->Profile = NULL;

#undef _Copy
}


void E3d_BevelClear(E3dBevel* PBevel)
{
//printf("%s() %s:%d  PBevel->Profile <%p:%s>\n", __FUNCTION__, __FILE__, __LINE__, PBevel->Profile, PBevel->Profile ? PBevel->Profile->Name : " ");fflush(stdout);
	E_DeleteChk(PBevel->Profile);
}




//========================================
// Set Vertex coordinates for Extrude
//========================================
static int _SetVertices(E3dGeometry* PProfileGeometry,  E3dVertex* PVertices, const E3dVertexSize PVertexSize,  E3dInfo* PInfo, const EIndex PNumOfProfilePoints)
{
	E3dSpline*  LProfileSpline = NULL;
	E3dFace*    LProfileFace = NULL;
	E3dMatrix   LMatrix;
	E3dCoord3*  LPoints;
	EBool       LUsingDefaultAxis;

//printf("%s() %s:%d  PNumOfProfilePoints %d\n", __FUNCTION__, __FILE__, __LINE__, (int)PNumOfProfilePoints);fflush(stdout);

	if(EObject_IsSubClass(PProfileGeometry, E3dSpline))
	{
		LProfileSpline = (E3dSpline*)PProfileGeometry;
	}
	else if(EObject_IsSubClass(PProfileGeometry, E3dFace))
	{
		LProfileFace = (E3dFace*)PProfileGeometry;
	}
	else return(0);


	E3dCoord3  LAxis = PInfo->Axis;
	if(LAxis.X == 0.0 && LAxis.Y == 0.0 && LAxis.Z == 0.0) LAxis.Y = 1.0;
	E3d_Normalize3_v(LAxis);


	ENewN(PNumOfProfilePoints, LPoints);

// If the Extrude is not along the Z axis, transform the profile coordinates accordingly
//
	if(LAxis.X == 0.0 && LAxis.Y == 0.0 && LAxis.Z == 1.0)
	{
		E3dCoord3*  LPoint = LPoints;

		if(LProfileSpline)
		{
			E3dCoordinate*  LLinearSegment = LProfileSpline->SLinearSegments;if(LLinearSegment)
			{
				const int       LDimensions = LProfileSpline->Dimensions;
				EIndex  LVC = 0;for(;LVC < PNumOfProfilePoints;LVC++, LLinearSegment += LDimensions, LPoint++)
				{
					LPoint->X = LLinearSegment[E3dX];
					LPoint->Y = LLinearSegment[E3dY];
					LPoint->Z = LLinearSegment[E3dZ];
				}
			}
			else if(LProfileSpline->CVs_Count > 0)
			{
				const E3dSplineCV*  LCV = LProfileSpline->CVs;
				*LPoint = LCV->Position;
			}
		}
		else if(LProfileFace)
		{
			if(LProfileFace->Vertices)
			{
				EIndex  LVC = 0;const E3dVertex*  LProfileVertex = LProfileFace->Vertices;const E3dVertexSize  LProfileVertexSize = LProfileFace->VertexSize;for(;LVC < PNumOfProfilePoints;LVC++, EStruct_Inc(LProfileVertex, LProfileVertexSize), LPoint++)
				{
					*LPoint = LProfileVertex->Position;
				}
			}
			else
			{
printf("%s() %s:%d  %sIMPLEMENT ME%s  LProfileFace has no Vertices!\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout);
			}
		}

		LUsingDefaultAxis = TRUE;
	}
	else
	{
		E3dCoord3*  LPoint = LPoints;

// Rotate contour into the proper coordinate system for extrusion
//
		E3dMatrix_LoadIdentity(LMatrix);  // E3d_VectorsGetRotMatrix only fills out the top-left 3x3 elements
		const E3dCoord3  LZAxis = { 0.0, 0.0, 1.0 };
		E3d_VectorsGetRotMatrix(&LAxis,  &LZAxis,  LMatrix);

		if(LProfileSpline)
		{
			const int  LDimensions = LProfileSpline->Dimensions;
			EIndex  LVCnt = 0;const E3dCoordinate*  LLinearSegment = LProfileSpline->SLinearSegments;for(;LVCnt < PNumOfProfilePoints;LVCnt++, LLinearSegment += LDimensions, LPoint++)
			{
				E3dMatrix_Transform3x4_sp(LMatrix,  LLinearSegment[E3dX], LLinearSegment[E3dY], LLinearSegment[E3dZ],  LPoint);
			}
		}
		else if(LProfileFace)
		{
			EIndex  LVCnt = 0;const E3dVertex*  LProfileVertex = LProfileFace->Vertices;const E3dVertexSize  LProfileVertexSize = LProfileFace->VertexSize;for(;LVCnt < PNumOfProfilePoints;LVCnt++, EStruct_Inc(LProfileVertex, LProfileVertexSize), LPoint++)
			{
				E3dMatrix_Transform3x4_vp(LMatrix, LProfileVertex->Position,  LPoint);
			}
		}
		LUsingDefaultAxis = FALSE;
	}


	EIndex  LNumOfPathPoints = PInfo->PathSteps + 1;

	{
		E3dCoordinate  LT;
		EIndex         LUN = LNumOfPathPoints;
//		E3dCoordinate  LStartBevelWidth = _BevelWidth(&PInfo->StartBevel), LEndBevelWidth = _BevelWidth(&PInfo->EndBevel);



		if(PInfo->CenterAlongPath)
		{
			const E3dCoordinate  LTotalDepth = E3d_Abs(PInfo->Depth);
			LT = -LTotalDepth * 0.5;
		}
		else LT = PInfo->Offset;


		E3dCoordinate  LSideLength = E3d_Abs(PInfo->Depth);

		if(PInfo->StartBevel.Type)
		{
			if(PInfo->IncludeBevelsInLength)
			{
				LT += PInfo->StartBevel.Radius;
				LSideLength -= PInfo->StartBevel.Radius;
			}
		}


		E3dVertex*  LVertex = PVertices;

// Starting bevel
//
		if(PInfo->StartBevel.Type)
		{
			E3dBevel*      LBevel = &PInfo->StartBevel;
			E3dCoord3      LSplineNormal = { 0.0, 0.0, 1.0 };
			E3dCoord3*     LPoint = LPoints;
			E3dCoordinate  LNormalDir;
			int            LDir;

			if(LBevel->Invert) { LNormalDir = 1.0;LDir = -1; }
			else { LNormalDir = 1.0;LDir = 1; }

			if(LProfileSpline)
			{
				E3dExtrude_Bevel(LPoint, PNumOfProfilePoints, LProfileSpline->Closed, LSplineNormal, LBevel,  LVertex, PVertexSize,  PNumOfProfilePoints, LT,  LNormalDir, LDir);

printf("%s() %s:%d  PNumOfProfilePoints %d  %d\n", __FUNCTION__, __FILE__, __LINE__, (int)PNumOfProfilePoints, (int)LBevel->Profile->SNumOfLinearSegments);fflush(stdout);

				LVertex = EStruct_ByIndex(LVertex, PVertexSize,  PNumOfProfilePoints * LBevel->Profile->SNumOfLinearSegments);
			}
			else if(LProfileFace)
			{
printf("%s() %s:%d  PNumOfProfilePoints %d  %d\n", __FUNCTION__, __FILE__, __LINE__, (int)PNumOfProfilePoints, (int)LBevel->Profile->SNumOfLinearSegments);fflush(stdout);
				{
					E3dSpline*    LProfileContour = LProfileFace->Exterior;
					const EIndex  LVN = LProfileContour->Closed ? LProfileContour->SNumOfLinearSegments : LProfileContour->SNumOfLinearSegments + 1;

//					const EIndex  LVN = LProfileFace->Exterior->SNumOfLinearSegments;

					E3dExtrude_Bevel(LPoint, LVN, TRUE, LSplineNormal, LBevel,  LVertex, PVertexSize,  PNumOfProfilePoints, LT,  LNormalDir, LDir);
					LPoint += LVN;LVertex = EStruct_ByIndex(LVertex, PVertexSize,  LVN);
				}

				EIndex  LHC = 0;const EIndex  LHN = LProfileFace->Holes_Count;for(;LHC < LHN;LHC++)
				{
					E3dSpline*    LProfileContour = LProfileFace->Holes[LHC];
					const EIndex  LVN = LProfileContour->Closed ? LProfileContour->SNumOfLinearSegments : LProfileContour->SNumOfLinearSegments + 1;

					E3dExtrude_Bevel(LPoint, LVN, TRUE, LSplineNormal, LBevel,  LVertex, PVertexSize,  PNumOfProfilePoints, LT,  LNormalDir, LDir);
					LPoint += LVN;LVertex = EStruct_ByIndex(LVertex, PVertexSize,  LVN);
				}
				LVertex = EStruct_ByIndex(PVertices, PVertexSize, LBevel->Profile->SNumOfLinearSegments * PNumOfProfilePoints);
			}

			LNumOfPathPoints += LBevel->Profile->SNumOfLinearSegments;
		}



// Sides
//
		if(PInfo->EndBevel.Type)
		{
			if(PInfo->IncludeBevelsInLength)
			{
				LSideLength -= PInfo->EndBevel.Radius;
			}
		}


		const E3dCoordinate  LTStep = 1.0 / (E3dCoordinate)(PInfo->PathSteps) * LSideLength;
		EIndex  LUCnt = 0;for(;LUCnt < LUN;LUCnt++)
		{
			E3dCoord3*  LPoint = LPoints;

			if(PInfo->Profile3D)
			{
				EIndex  LVCnt = 0;for(;LVCnt < PNumOfProfilePoints;LVCnt++, LPoint++, EStruct_Inc(LVertex, PVertexSize))
				{
					LVertex->Position.X = LPoint->X;LVertex->Position.Y = LPoint->Y;LVertex->Position.Z = LPoint->Z + LT;
				}
			}
			else
			{
				EIndex  LVCnt = 0;for(;LVCnt < PNumOfProfilePoints;LVCnt++, LPoint++, EStruct_Inc(LVertex, PVertexSize))
				{
					LVertex->Position.X = LPoint->X;LVertex->Position.Y = LPoint->Y;LVertex->Position.Z = LT;
				}
			}
			if(LUCnt < LUN - 1) LT += LTStep;
		}



// End bevel
//
		if(PInfo->EndBevel.Type)
		{
			E3dBevel*         LBevel = &PInfo->EndBevel;
			const E3dCoord3*  LPoint = LPoints;
			E3dCoord3         LSplineInvNormal = { 0.0, 0.0, -1.0 };
			E3dCoordinate     LNormalDir;
			int               LDir;

			if(LBevel->Invert) { LNormalDir = -1.0;LDir = 1; }
			else { LNormalDir = -1.0;LDir = -1; }

			if(LProfileSpline)
			{
				E3dExtrude_Bevel(LPoints, PNumOfProfilePoints, LProfileSpline->Closed, LSplineInvNormal, LBevel,  LVertex, PVertexSize,  PNumOfProfilePoints, LT,  LNormalDir, LDir);LVertex = EStruct_ByIndex(LVertex, PVertexSize, PNumOfProfilePoints * LBevel->Profile->SNumOfLinearSegments);
			}
			else if(LProfileFace)
			{
				{
					E3dSpline*    LProfileContour = LProfileFace->Exterior;
					const EIndex  LVN = LProfileContour->Closed ? LProfileContour->SNumOfLinearSegments : LProfileContour->SNumOfLinearSegments + 1;
					E3dExtrude_Bevel(LPoint, LVN, TRUE, LSplineInvNormal, LBevel,  LVertex, PVertexSize,  PNumOfProfilePoints, LT,  LNormalDir, LDir);LPoint += LVN;LVertex = EStruct_ByIndex(LVertex, PVertexSize,  LVN);
				}

				EIndex  LHC = 0;const EIndex  LHN = LProfileFace->Holes_Count;for(;LHC < LHN;LHC++)
				{
					E3dSpline*    LProfileContour = LProfileFace->Holes[LHC];
					const EIndex  LVN = LProfileContour->Closed ? LProfileContour->SNumOfLinearSegments : LProfileContour->SNumOfLinearSegments + 1;

					E3dExtrude_Bevel(LPoint, LVN, TRUE, LSplineInvNormal, LBevel,  LVertex, PVertexSize,  PNumOfProfilePoints, LT,  LNormalDir, LDir);LPoint += LVN;LVertex = EStruct_ByIndex(LVertex, PVertexSize,  LVN);
				}
			}

			LNumOfPathPoints += LBevel->Profile->SNumOfLinearSegments;
		}
	}


	{
		const E3dCoordinate  LStartBevelRadius = PInfo->StartBevel.Type ? PInfo->StartBevel.Radius : 0.0;
		const E3dCoordinate  LEndBevelRadius   = PInfo->EndBevel.Type   ? PInfo->EndBevel.Radius : 0.0;
		const E3dCoordinate  LDepthAbs = E3d_Abs(PInfo->Depth);

		const EBool          LIncludeBevelsInLength = PInfo->IncludeBevelsInLength;
		const E3dCoordinate  LDepth = LIncludeBevelsInLength ? LDepthAbs : LDepthAbs + (LStartBevelRadius + LEndBevelRadius);
		E3dCoordinate        LZStart = PInfo->CenterAlongPath ? -LDepthAbs / 2.0 : 0.0;

		if(!LIncludeBevelsInLength) LZStart -= LStartBevelRadius;

		const EIndex  LStartBevelUPoints = PInfo->StartBevel.Type ? PInfo->StartBevel.Profile->SNumOfLinearSegments : 0;
		const EIndex  LEndBevelUPoints   = PInfo->EndBevel.Type   ? PInfo->EndBevel.Profile->SNumOfLinearSegments : 0;

		if(PInfo->Twist != 0.0)
		{
			const E3dCoordinate  LTwistAngle = PInfo->Twist;
			E3dCoordinate        LTDepth = LDepth, LTZStart = LZStart;
			EIndex               LTUPoints = LNumOfPathPoints;
			EIndex               LLastUPoints;

//			EIndex  LDebugUC = 0;

			E3dVertex*           LVertex = PVertices;

			if(!PInfo->TwistBevels)
			{
				LTDepth  -= LStartBevelRadius + LEndBevelRadius;
				LTZStart += LStartBevelRadius;

				const EIndex  LUStart = LStartBevelUPoints;
//				LDebugUC += LUStart;

				LTUPoints -= LUStart + LEndBevelUPoints;
				LVertex = EStruct_ByIndex(LVertex, PVertexSize, LUStart * PNumOfProfilePoints);
				LLastUPoints = LEndBevelUPoints;
			}
			else LLastUPoints = 0;
			

			{
				EIndex  LUC = 0;for(;LUC < LTUPoints;LUC++)
				{
					const E3dCoordinate  LT = (LVertex->Position.Z - LTZStart) / LTDepth;
					const E3dCoordinate  LAngle = LT * LTwistAngle;
					E3dMatrix_LoadIdentity(LMatrix);E3dMatrix_RotateDeg(LMatrix, 'z', LAngle);

					EIndex  LVC = 0;for(;LVC < PNumOfProfilePoints;LVC++, EStruct_Inc(LVertex, PVertexSize))
					{
						E3dMatrix_Transform3x3_v(LMatrix, LVertex->Position);
					}
				}
			}

// We need to apply the final twist to the end bevel (we just won't change it through the U steps)
//
			if(LLastUPoints)
			{
				EIndex  LUC = 0;for(;LUC < LLastUPoints;LUC++)
				{
					EIndex  LVC = 0;for(;LVC < PNumOfProfilePoints;LVC++, EStruct_Inc(LVertex, PVertexSize))
					{
						E3dMatrix_Transform3x3_v(LMatrix, LVertex->Position);
					}
				}
			}

//printf("%s() %s:%d  LTUPoints %d  LDebugUC %d  LNumOfPathPoints %d\n", __FUNCTION__, __FILE__, __LINE__, LTUPoints, LDebugUC, LNumOfPathPoints);fflush(stdout);
		}


		if(PInfo->Scale.X != 1.0 || PInfo->Scale.Y != 1.0)
		{
			const E3dCoord2  LScaleV = PInfo->Scale;
			E3dCoordinate    LTDepth = LDepth, LTZStart = LZStart;
			EIndex           LTUPoints = LNumOfPathPoints;
			EIndex           LLastUPoints;

			EIndex  LDebugUC = 0;

			E3dVertex*       LVertex = PVertices;
			if(!PInfo->ScaleBevels)
			{
				LTDepth  -= LStartBevelRadius + LEndBevelRadius;
				LTZStart += LStartBevelRadius;

				const EIndex  LUStart = LStartBevelUPoints;

				LDebugUC += LUStart;

				LTUPoints -= LUStart + LEndBevelUPoints;
				LVertex = EStruct_ByIndex(LVertex, PVertexSize, LUStart * PNumOfProfilePoints);
				LLastUPoints = LEndBevelUPoints;
			}
			else LLastUPoints = 0;
			

			E3dCoord2            LScale;E_SetXY_s(LScale,  1.0, 1.0);
			const E3dCoordinate  LTTX = 1.0 - LScaleV.X;
			const E3dCoordinate  LTTY = 1.0 - LScaleV.Y;

			{
				E3dCoordinate  LU = 0.0;const E3dCoordinate  LUStep = 1.0 / (E3dCoordinate)(LTUPoints - 1);
				EIndex  LUC = 0;for(;LUC < LTUPoints;LUC++, LDebugUC++, LU += LUStep)
				{
					if(PInfo->Profile3D)
					{
						LScale.X = LScaleV.X * LU + (1.0 - LU);
						LScale.Y = LScaleV.Y * LU + (1.0 - LU);
						EIndex  LVC = 0;for(;LVC < PNumOfProfilePoints;LVC++, EStruct_Inc(LVertex, PVertexSize))
						{
							LVertex->Position.X *= LScale.X;
							LVertex->Position.Y *= LScale.X;
							LVertex->Position.Z *= LScale.X;
						}
					}
					else
					{
						const E3dCoordinate  LT = (LVertex->Position.Z - LTZStart) / LTDepth;
						LScale.X = (1.0 - LT) * LTTX + LScaleV.X;
						LScale.Y = (1.0 - LT) * LTTY + LScaleV.Y;
						EIndex  LVC = 0;for(;LVC < PNumOfProfilePoints;LVC++, EStruct_Inc(LVertex, PVertexSize))
						{
							LVertex->Position.X *= LScale.X;
							LVertex->Position.Y *= LScale.Y;
						}
					}
				}
			}

// We need to apply the final scale to the end bevel (just don't change it through the U steps)
// We also need to scale down along Z.
//
			if(LLastUPoints)
			{
				const E3dCoordinate  LZBase = LVertex->Position.Z;

				EIndex  LUC = 0;for(;LUC < LLastUPoints;LUC++, LDebugUC++)
				{
					const E3dCoordinate  LZ = (LVertex->Position.Z - LZBase) * LScale.X + LZBase;

// Keep extrapolating LScale to have a smooth transition into the scaled down bevel.
//
					const E3dCoordinate  LT = (LZ - LTZStart) / LTDepth;
					LScale.X = (1.0 - LT) * LTTX + LScaleV.X;
					LScale.Y = (1.0 - LT) * LTTY + LScaleV.Y;

					EIndex  LVC = 0;for(;LVC < PNumOfProfilePoints;LVC++, EStruct_Inc(LVertex, PVertexSize))
					{
						LVertex->Position.X *= LScale.X;
						LVertex->Position.Y *= LScale.Y;
						LVertex->Position.Z = LZ;
					}
				}
			}

//printf("%s() %s:%d  LTUPoints %d  LDebugUC %d  LNumOfPathPoints %d\n", __FUNCTION__, __FILE__, __LINE__, LTUPoints, LDebugUC, LNumOfPathPoints);fflush(stdout);

		}
	}


	const EIndex  LVN = LNumOfPathPoints * PNumOfProfilePoints;

// If Depth is negative, invert along Z
//
	if(PInfo->Depth < 0.0)
	{
		EIndex  LVC = 0;E3dVertex*  LVertex = PVertices;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, PVertexSize)) LVertex->Position.Z *= -1.0;
	}


// If the Extrude is not along the Z axis, transform the contour coordinates accordingly
//
	if(!LUsingDefaultAxis)
	{
// Rotate extruded vertices back to the proper coordinate system
//
		E3dMatrix_LoadIdentity(LMatrix);  // E3d_VectorsGetRotMatrix only fills out the top-left 3x3 elements
		const E3dCoord3  LZAxis = { 0.0, 0.0, 1.0 };
		E3d_VectorsGetRotMatrix(&LZAxis,  &LAxis,  LMatrix);

		EIndex  LVC = 0;E3dVertex*  LVertex = PVertices;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, PVertexSize))
		{
			E3dMatrix_Transform3x4_v(LMatrix, LVertex->Position);
		}
	}

	EFree(LPoints);

	return(LNumOfPathPoints * PNumOfProfilePoints);
}


//========================================
// Bevel callback handler
//========================================
static void _DCB_BevelRadius(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dBevelPanel*  LBevelPanel = (E3dBevelPanel*)PClientData;
	E3dBevel*       LBevel = LBevelPanel->Bevel;

	LBevel->Radius = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
	if(LBevel->NumOfRoundingSteps) E3d_BevelCreateSegments(LBevel);
	else LBevel->Type = 0;
	if(LBevelPanel->Callback) LBevelPanel->Callback(LBevelPanel, LBevelPanel->ClientData);
}


//========================================
// Bevel callback handler
//========================================
static void _DCB_BevelSteps(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dBevelPanel*  LBevelPanel = (E3dBevelPanel*)PClientData;
	E3dBevel*       LBevel = LBevelPanel->Bevel;


	LBevel->NumOfRoundingSteps = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

//printf("%s() %d %p\n", __FUNCTION__, LBevel->NumOfProfileSteps, LBevel);fflush(stdout);
	if(LBevel->NumOfRoundingSteps) E3d_BevelCreateSegments(LBevel);
	else
	{
		E3dSpline_FreeRenderData(LBevel->Profile);
		E3dSpline_FreeCVs(LBevel->Profile);
		LBevel->Type = 0;
	}

	if(LBevelPanel->Callback) LBevelPanel->Callback(LBevelPanel, LBevelPanel->ClientData);
}


//========================================
// Bevel callback handler
//========================================
static void _DCB_Invert(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dBevelPanel*  LBevelPanel = (E3dBevelPanel*)PClientData;
	E3dBevel*       LBevel = LBevelPanel->Bevel;


	LBevel->Invert = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

//printf("%s() %d %p\n", __FUNCTION__, LBevel->NumOfProfileSteps, LBevel);fflush(stdout);
	if(LBevel->NumOfRoundingSteps) E3d_BevelCreateSegments(LBevel);
	else
	{
		if(LBevel->Profile)
		{
			E3dSpline_FreeRenderData(LBevel->Profile);
			E3dSpline_FreeCVs(LBevel->Profile);
		}
		LBevel->Type = 0;
	}

	if(LBevelPanel->Callback) LBevelPanel->Callback(LBevelPanel, LBevelPanel->ClientData);
}


//========================================
// Bevel callback handler
//========================================
static void _DCB_AddStartSmoother(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dBevelPanel*  LBevelPanel = (E3dBevelPanel*)PClientData;
	E3dBevel*       LBevel = LBevelPanel->Bevel;


	LBevel->AddStartSmoother = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

//printf("%s() %d %p\n", __FUNCTION__, LBevel->NumOfProfileSteps, LBevel);fflush(stdout);
	if(LBevel->NumOfRoundingSteps) E3d_BevelCreateSegments(LBevel);
	else
	{
		if(LBevel->Profile)
		{
			E3dSpline_FreeRenderData(LBevel->Profile);
			E3dSpline_FreeCVs(LBevel->Profile);
		}
		LBevel->Type = 0;
	}

	if(LBevelPanel->Callback) LBevelPanel->Callback(LBevelPanel, LBevelPanel->ClientData);
}


//========================================
// Bevel callback handler
//========================================
static void _DCB_AddEndSmoother(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dBevelPanel*  LBevelPanel = (E3dBevelPanel*)PClientData;
	E3dBevel*       LBevel = LBevelPanel->Bevel;


	LBevel->AddEndSmoother = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

//printf("%s() %d %p\n", __FUNCTION__, LBevel->NumOfProfileSteps, LBevel);fflush(stdout);
	if(LBevel->NumOfRoundingSteps) E3d_BevelCreateSegments(LBevel);
	else
	{
		if(LBevel->Profile)
		{
			E3dSpline_FreeRenderData(LBevel->Profile);
			E3dSpline_FreeCVs(LBevel->Profile);
		}
		LBevel->Type = 0;
	}

	if(LBevelPanel->Callback) LBevelPanel->Callback(LBevelPanel, LBevelPanel->ClientData);
}


//========================================
// Create BevelInfo drawer
//========================================
Widget E3d_BevelCreateDrawer(Widget PParentW, const char* PName, E3dBevelPanel* PBevelPanel, E3dBevelPanelCallback PCallback, EPointer PClientData, E3dBevel* PBevel)
{
	Widget    LIndentWs[7];int  LIndentN;        // For indenting the ScaleWidgets
	Widget    LTraverseWs[7];int  LTraverseN;    // For TextField traversal
	EArg      Args[256];int  ArgCnt;

	XeDrawer*  LDrawer = &PBevelPanel->Drawer;Widget  LDrawerW = XeDrawer_NewVa(PParentW, LDrawer, PName, 0, 0);
	EXeStart;
	EXeSetArg(XeNheightAdjustable, FALSE);
	Xe_SetValues(LDrawer->OpenCloseButtonW,  Args, ArgCnt);

	PBevelPanel->Callback   = PCallback;
	PBevelPanel->ClientData = PClientData;
	PBevelPanel->Bevel      = PBevel;

	Widget     LW;

	{
		EXeStart;
		EXeSetArg(XeNorientation, XeVERTICAL);
		EXeSetArg(XeNspacingY, 2);
//		EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
		EXeSetArg(XeNheightAdjustable, FALSE);
		Widget  LMatrixW = XeWidget_New("MatrixW", XeMatrixWidget_class, LDrawerW,  Args, ArgCnt);
		LDrawer->DrawerW = LMatrixW;

		LIndentN = 0;
		LTraverseN = 0;
		EXeStart;
		EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 4);
		EXeSetArg(XeNorientation, XeHORIZONTAL);
		LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PBevelPanel->RadiusW = LW = XeScaleNew_Coord(LMatrixW, "Radius",  Args, ArgCnt);
		XeScale_SetValueAndRanges_Coord(LW,  PBevel->Radius,  0.0001, 1000000.0,   0.0001, 10.0,  FALSE);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_BevelRadius, (EPointer)PBevelPanel);

		EXeStart;
		EXeSetArg(XeNtextFieldDigits, 5);
		EXeSetArg(XeNminimum, 0);EXeSetArg(XeNmaximum, 50);
		EXeSetArg(XeNscaleMaximum, 100);
		EXeSetArg(XeNvalue, PBevel->NumOfRoundingSteps);
		EXeSetArg(XeNorientation, XeHORIZONTAL);
		LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = PBevelPanel->RoundingStepsW = LW = XeWidget_ManagedNew("Steps", XeScaleWidget_class, LMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_BevelSteps, (EPointer)PBevelPanel);

		Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 2);
	}

// Set up TextField traversal
//
	Xe_TraversalRing(LTraverseWs, LTraverseN);

	{
		EXeStart;
		EXeSetArg(XeNorientation, XeHORIZONTAL);
		EXeSetArg(XeNspacingX, 8);
		EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
		EXeSetArg(XeNheightAdjustable, FALSE);
		Widget  LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LDrawer->DrawerW,  Args, ArgCnt);

		EXeStart;
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNset, PBevel->Invert);
		XeToolTip8(
		 "Make a bevel for an exterior contour go outward and for a hole, inward.");
		PBevelPanel->InvertW = LW = XeWidget_ManagedNew("Invert", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Invert, (EPointer)PBevelPanel);

		EXeStart;
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNset, PBevel->AddStartSmoother);
		XeToolTip8(
		 "Add an extra segment to make sure that when the normals are averaged, the cap looks flat.");
		PBevelPanel->AddStartSmootherW = LW = XeWidget_ManagedNew("Cap smoother", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_AddStartSmoother, (EPointer)PBevelPanel);

		EXeStart;
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNset, PBevel->AddEndSmoother);
		XeToolTip8(
		 "Add an extra segment to make sure that when the normals are averaged, the sides look flat.");
		PBevelPanel->AddEndSmootherW = LW = XeWidget_ManagedNew("Side smoother", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_AddEndSmoother, (EPointer)PBevelPanel);
	}

	return(LDrawerW);
}


//========================================
// Update BevelInfo drawer
//========================================
void E3d_BevelPanelUpdate(E3dBevelPanel* PBevelPanel, E3dBevel* PBevel)
{
	PBevelPanel->Bevel = PBevel;

	XeScale_SetValue_Coord(PBevelPanel->RadiusW, PBevel->Radius, FALSE);
	XeScale_SetValue(PBevelPanel->RoundingStepsW, PBevel->NumOfRoundingSteps, FALSE);

	XeToggleButton_SetState(PBevelPanel->InvertW, PBevel->Invert, FALSE);
	XeToggleButton_SetState(PBevelPanel->AddStartSmootherW, PBevel->AddStartSmoother, FALSE);
	XeToggleButton_SetState(PBevelPanel->AddEndSmootherW, PBevel->AddEndSmoother, FALSE);
}


/*
static E3dCoordinate _BevelWidth(E3dBevel* PBevel)
{
	if(PBevel->Profile && E3d_SplineNumOfLinearSegmentPoints(PBevel->Profile))
	{
	 E3dCoordinate*  LCoord = PBevel->Profile->SLinearSegments;
	 E3dCoordinate   LMin = LCoord[0], LMax = LCoord[0];
	 const int       LDimensions = PBevel->Profile->Dimensions;

	 EIndex  LC = 0;const EIndex  LN = E3d_SplineNumOfLinearSegmentPoints(PBevel->Profile);for(;LC < LN;LC++, LCoord += LDimensions)
	 {
		 if(LMin > LCoord[0]) LMin = LCoord[0];
		 if(LMax < LCoord[0]) LMax = LCoord[0];
	 }
	 return(LMax - LMin);
	}
	else return(0.0);
}
*/


void E3dExtrude_CheckProfiles(E3dObject* PObject, E3dExtrudeSplineInfoCommon* PInfo)
{
//printf("%s() %s:%d  <%s>  ProfileGeometries %p  Profiles %p\n", __FUNCTION__, __FILE__, __LINE__, PObject->Name, PInfo->ProfileGeometries, PInfo->Profiles);fflush(stdout);
	if(PInfo->Profiles == NULL && PInfo->ProfileGeometries)
	{
// Update profile extra information
//
		PInfo->Profiles = E3dExtrude_ProfilesCreate(PInfo->ProfileGeometries, PInfo->ProfileGeometries_Count);

// Add output to the Profile Geometries
//
		EIndex  LC = 0;for(;LC < PInfo->ProfileGeometries_Count;LC++)
		{
			if(PInfo->ProfileGeometries[LC])
			{
			 EObject*  LProfileG = PInfo->ProfileGeometries[LC];
			 if(EObject_IsSubClass(LProfileG, E3dGeometry)) E3dBaseObject_AddOutput((E3dBaseObject*)LProfileG, &PInfo->InputInterface);
			}
		}
	}
}


void E3dExtrude_CheckBevelsAndProfiles(E3dObject* PObject, E3dExtrudeSplineInfoCommon* PInfo)
{
// Update Bevels
//
	{
		E3dBevel*  LBevel = &PInfo->StartBevel;

		if(LBevel->NumOfRoundingSteps) E3d_BevelCreateSegments(LBevel);
		else
		{
			if(LBevel->Profile) { E3dSpline_FreeRenderData(LBevel->Profile);E3dSpline_FreeCVs(LBevel->Profile); }
			LBevel->Type = 0;
		}
	}

	{
		E3dBevel*	LBevel = &PInfo->EndBevel;

		if(LBevel->NumOfRoundingSteps) E3d_BevelCreateSegments(LBevel);
		else
		{
			if(LBevel->Profile) { E3dSpline_FreeRenderData(LBevel->Profile);E3dSpline_FreeCVs(LBevel->Profile); }
			LBevel->Type = 0;
		}
	}

	E3dExtrude_CheckProfiles(PObject, PInfo);
}


//========================================================================================================
// Get 2D bbox of PProfile, to calculate the UV distribution between the caps and the sides
//========================================================================================================
E3dCoord2* E3d_UVLayout(E3dExtrudeSplineInfoCommon* PInfo, const E3dExtrudeSplineProfile* PProfile, E3dFace* PProfileFace, E3dSpline* PPathSpline, const E3dCoordinate PDepth,  E3dCoord2* PProfile2DBBoxMin, E3dCoord2* PProfile2DBBoxMax, E3dCoordinate** PUsP, EIndex* PNumOfUsP, const char PCloseEnds)
{
	E3dCoord2*     L2DVertices = NULL;
	E3dCoordinate  LCapULength = 0.0, LCapVLength = 0.0;

	PInfo->SidesBelowInUVSpace = TRUE;
	const EBool  LSidesBelowInUVSpace = PInfo->SidesBelowInUVSpace;

	if(PCloseEnds)
	{
		if(PProfileFace->Vertices == NULL) E3dFace_Tesselate(PProfileFace);

		L2DVertices = E3dFace_Flatten(PProfileFace);if(L2DVertices)
		{
			const E3dCoord2*  L2DVertex = L2DVertices;

			PProfile2DBBoxMin->X = PProfile2DBBoxMax->X = L2DVertex->X;
			PProfile2DBBoxMin->Y = PProfile2DBBoxMax->Y = L2DVertex->Y;
			EIndex  LVC = 1;const EIndex  LVN = PProfileFace->Vertices_Count;for(;LVC < LVN;LVC++, L2DVertex++)
			{
				if(PProfile2DBBoxMin->X > L2DVertex->X) PProfile2DBBoxMin->X = L2DVertex->X;
				else if(PProfile2DBBoxMax->X < L2DVertex->X) PProfile2DBBoxMax->X = L2DVertex->X;
				if(PProfile2DBBoxMin->Y > L2DVertex->Y) PProfile2DBBoxMin->Y = L2DVertex->Y;
				else if(PProfile2DBBoxMax->Y < L2DVertex->Y) PProfile2DBBoxMax->Y = L2DVertex->Y;
			}
		}
		const E3dCoordinate  LCapXLength = PProfile2DBBoxMax->X - PProfile2DBBoxMin->X;
		const E3dCoordinate  LCapYLength = PProfile2DBBoxMax->Y - PProfile2DBBoxMin->Y;

		if(LSidesBelowInUVSpace)
		{
			if(LCapXLength < LCapYLength)  // Orient cap Polygons in UV space with optimal usage
			{
				LCapULength = LCapYLength;
				LCapVLength = LCapXLength;
				PInfo->CapUVOrientation = E3dXY_TO_VU;
			}
			else
			{
				LCapULength = LCapXLength;
				LCapVLength = LCapYLength;
				PInfo->CapUVOrientation = E3dXY_TO_UV;
			}
		}
		else
		{
			if(LCapXLength < LCapYLength)  // Orient cap Polygons in UV space with optimal usage
			{
				LCapULength = LCapXLength;
				LCapVLength = LCapYLength;
				PInfo->CapUVOrientation = E3dXY_TO_UV;
			}
			else
			{
				LCapULength = LCapYLength;
				LCapVLength = LCapXLength;
				PInfo->CapUVOrientation = E3dXY_TO_VU;
			}
		}
	}

	E3dCoordinate  LSidesTotalULength;E3dCoordinate*  LUs = _GetUs((E3dExtrudeInfoCommon*)PInfo,  PPathSpline, PDepth, PInfo->PathSteps,  PNumOfUsP, &LSidesTotalULength);
	E3dCoordinate  LSidesTotalVLength = 0.0;
	EBool          LHaveVisibleProfileSegment = FALSE;

	{
		E3dSpline*  LContour = PProfileFace->Exterior;
		LSidesTotalVLength += LContour->Length;

		EIndex  LC = 0;const EIndex  LN = LContour->CVs_Count;const E3dSplineCV*  LCV = LContour->CVs;for(;LC < LN;LC++, LCV++)
		{
			if((LContour->Closed || LC < LN - 1) && (LCV->SegmentFlags & E3dSegmentVISIBLE)) { LHaveVisibleProfileSegment = TRUE;break; }
		}
	}

	EIndex	LHC = 0;const EIndex  LHN = PProfileFace->Holes_Count;for(;LHC < LHN;LHC++)
	{
		E3dSpline*  LContour = PProfileFace->Holes[LHC];
		LSidesTotalVLength = LContour->Length;

		EIndex  LC = 0;const EIndex  LN = LContour->CVs_Count;const E3dSplineCV*  LCV = LContour->CVs;for(;LC < LN;LC++, LCV++)
		{
			if((LContour->Closed || LC < LN - 1) && (LCV->SegmentFlags & E3dSegmentVISIBLE)) { LHaveVisibleProfileSegment = TRUE;break; }
		}
	}

	const E3dCoordinate  LSidesSmallerExtent = LSidesTotalULength < LSidesTotalVLength ? LSidesTotalULength : LSidesTotalVLength;
	if(LSidesBelowInUVSpace)
	{
		if(LSidesTotalULength > LSidesTotalVLength) PInfo->SidesUVOrientation = E3dXY_TO_UV;
		else PInfo->SidesUVOrientation = E3dXY_TO_VU;
	}
	else
	{
		if(LSidesTotalULength < LSidesTotalVLength) PInfo->SidesUVOrientation = E3dXY_TO_UV;
		else PInfo->SidesUVOrientation = E3dXY_TO_VU;
	}

	EBool  LSides = LHaveVisibleProfileSegment;


	E3dCoordinate  LTotalULength = 0.0;

	if(LSides && !PInfo->SidesBelowInUVSpace) LTotalULength += LSidesTotalULength;
	if(PCloseEnds & ExtrCLOSE_START) LTotalULength += LCapULength;
	if(PCloseEnds & ExtrCLOSE_END)   LTotalULength += LCapULength;

	const E3dCoordinate  LUMargin  = E3dU_MARGIN  * LTotalULength;
	const E3dCoordinate  LUSpacing = E3dU_SPACING * LTotalULength;
	PInfo->UMargin  = LUMargin;
	PInfo->USpacing = LUSpacing;

	E3dCoordinate  LU = LUMargin;
	EBool          LAddUSpacing = FALSE, LAddVSpacing = FALSE;

	PInfo->NormStartCapUStart = LU;

	if(PCloseEnds & ExtrCLOSE_START)
	{
		PInfo->NormStartCapUStart = LU;LU += LCapULength;
		LAddUSpacing = TRUE;
	}

	if(LSides && !LSidesBelowInUVSpace)
	{
		if(LAddUSpacing) LU += LUSpacing;

		PInfo->NormSidesUStart = LU;
		LU += LSidesTotalULength;

		LAddUSpacing = TRUE;
	}

	if(PCloseEnds & ExtrCLOSE_END)
	{
		if(LAddUSpacing) LU += LUSpacing;
		PInfo->NormEndCapUStart = LU;
		LU += LCapULength;
	}

	LTotalULength = LU + LUMargin;

	PInfo->SidesULength = LSidesTotalULength;

// Normalize Us
//
	PInfo->NormCapULength      = LCapULength / LTotalULength;
	PInfo->NormStartCapUStart /= LTotalULength;
	PInfo->NormEndCapUStart   /= LTotalULength;

	const EIndex  LSidesGapCount = PProfileFace->Holes_Count;

	if(LSidesBelowInUVSpace)
	{
// The sides are below the caps in UV space
//

//printf("%s() %s:%d LSidesBelowInUVSpace TRUE\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
		E3dCoordinate  LTotalVLength = 0.0;

		if(PCloseEnds) LTotalVLength += LCapVLength;
		if(LSides) LTotalVLength += LSidesSmallerExtent;

		const E3dCoordinate  LVMargin  = E3dV_MARGIN  * LTotalVLength;
		const E3dCoordinate  LVSpacing = E3dV_SPACING * LTotalVLength;
		E3dCoordinate        LV = LVMargin;

		if(PCloseEnds)
		{
			PInfo->NormCapVStart = LV;LV += LCapVLength;
			LAddVSpacing = TRUE;
		}

		if(LSides)
		{
			if(LAddVSpacing) LV += LVSpacing;
			PInfo->NormSidesVStart = LV;LV += LSidesSmallerExtent;
		}

		LTotalVLength = LV + LVMargin;

		PInfo->NormCapVStart  /= LTotalVLength;
		PInfo->NormCapVLength  = LCapVLength / LTotalVLength;

		PInfo->NormSidesUStart   = E3dU_MARGIN;
		PInfo->NormSidesVStart  /= LTotalVLength;

		if(PInfo->SidesUVOrientation == E3dXY_TO_UV)
		{
//printf("%s() %s:%d UV\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
// U will be along the extrusion path, V will be along the profile contours

			PInfo->NormSidesULength  = 1.0 - (E3dU_MARGIN * 2.0);
			PInfo->NormSidesVLength = LSidesSmallerExtent / LTotalVLength - ((E3dCoordinate)LSidesGapCount * E3dV_SPACING);
		}
		else
		{
// U will be along the profile contours, V will be along the extrusion path
			PInfo->NormSidesULength = 1.0 - (E3dU_MARGIN * 2.0 + (E3dCoordinate)LSidesGapCount * E3dU_SPACING);
			PInfo->NormSidesVLength  = LSidesSmallerExtent / LTotalVLength;
//printf("%s() %s:%d VU  gaps %d  %f\n", __FUNCTION__, __FILE__, __LINE__, (int)LSidesGapCount, PInfo->NormSidesULength);fflush(stdout);

//PInfo->NormSidesULength   0.859375
		}
	}
	else
	{
// The sides are to the left and right of the caps in UV space.
//
//printf("%s() %s:%d LSidesBelowInUVSpace FALSE\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
		PInfo->NormCapVLength = LCapVLength / (E3dV_MARGIN * LCapVLength + LCapVLength + E3dV_MARGIN * LCapVLength);
		PInfo->NormCapVStart =  E3dV_MARGIN;

		PInfo->NormSidesUStart  /= LTotalULength;
		PInfo->NormSidesVStart   = E3dV_MARGIN;

		if(PInfo->SidesUVOrientation == E3dXY_TO_UV)
		{
// U will be along the extrusion path, V will be along the profile contours
//printf("%s() %s:%d  E3dXY_TO_UV\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
			PInfo->NormSidesULength  = LSidesTotalULength / LTotalULength;
			PInfo->NormSidesVLength  = 1.0 - (E3dV_MARGIN * 2.0 + (E3dCoordinate)LSidesGapCount * E3dV_SPACING);
		}
		else
		{
// U will be along the profile contours, V will be along the extrusion path
//printf("%s() %s:%d  E3dXY_TO_VU  PInfo->CloseEnd %d\n", __FUNCTION__, __FILE__, __LINE__, PInfo->CloseEnd);fflush(stdout);
			PInfo->NormSidesULength  = LSidesTotalULength / LTotalULength - ((E3dCoordinate)LSidesGapCount * E3dU_SPACING);
			PInfo->NormSidesVLength = 1.0 - (E3dV_MARGIN * 2.0);
		}
	}

	*PUsP = LUs;
	return(L2DVertices);
}


typedef struct
{
	Widget            CloseStartW, CloseSidesW, CloseEndW, TrianglesOnCapsW, TrianglesOnSidesW, CenteredW, IncludeBevelsInLengthW,
	                  Profile3DW,
	                  PathStepsW, DepthW, OffsetW, TwistW, ScaleXW, ScaleYW,
	                  SmoothingAngleW;

	E3dpAxisSelector  AxisSelector;

	E3dBevelPanel     StartBevelPanel, EndBevelPanel;
} E3dResourcesGUI;


//========================================
// AxisSelector callback
//========================================
static void _CB_AxisSelector(ECoreObject* PStruct, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*    LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*         LObject = LPanel->Object;
	E3dInfo*           LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);
	E3dpAxisSelector*  LAxisSelector = (E3dpAxisSelector*)PStruct;

	LInfo->Axis = LAxisSelector->Axis;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_Centered(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->CenterAlongPath = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ObjectComponent_class);if(LDrawer)
	{
		E3dResourcesGUI*  LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);
		if(LInfo->CenterAlongPath) Xe_SetValuesVa(LResourcesGUI->OffsetW,  XeNsensitive, FALSE,  NULL);else Xe_SetValuesVa(LResourcesGUI->OffsetW,  XeNsensitive, TRUE,  NULL);
	}

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_IncludeBevelsInLength(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->IncludeBevelsInLength = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_Depth(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->Depth = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3duf_SHAPE|E3dUF_CALL_OUTPUTS, LPanel);
}


static void _DCB_Offset(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->Offset = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3duf_SHAPE|E3dUF_CALL_OUTPUTS, LPanel);
}


static void _DCB_Twist(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->Twist = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3duf_SHAPE|E3dUF_CALL_OUTPUTS, LPanel);
}


static void _DCB_ScaleX(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->Scale.X = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3duf_SHAPE|E3dUF_CALL_OUTPUTS, LPanel);
}


static void _DCB_ScaleY(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->Scale.Y = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3duf_SHAPE|E3dUF_CALL_OUTPUTS, LPanel);
}


//================================================================
// Callback for Profile-Spline changes
// This function should Not call any 3DWindow redrawing
// functions, just return with the appropriate change-flags
// (e.g. E3dChgMESHSHAPE) if a redraw is needed!
//================================================================
static EChangeFlags _InputChanged(EInterface* PInterface, EInterfaceCallbackStruct* PCBS)
{
	E3dObject*    LObject  = (E3dObject*)PInterface->Object;
	E3dInfo*      LInfo    = (E3dInfo*)PInterface->ObjectComponent;
	EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass(LObject);

//{ char  s[256];printf("%s() %s:%d [%s]\n", __FUNCTION__, __FILE__, __LINE__, E3d_SceneChangeFlagsToString(LChanged, s));fflush(stdout); }

// Only respond if the Object is instantiated (e.g. not in the undo stack after being deleted)
//
	if(LObject->Models_Count)
	{
//{ char  s[256];printf("%s() %s:%d [%s]\n", __FUNCTION__, __FILE__, __LINE__, E3d_SceneChangeFlagsToString(LChanged, s));fflush(stdout); }
// E3dFIXME - only update the Profile entry for the Profile Geometry that actually changed
//
		E3dExtrude_ProfilesUpdateFaces(LInfo->Profiles, LInfo->ProfileGeometries, LInfo->ProfileGeometries_Count);
		E3dObject_CustomUpdate(LObject, LInfo, PCBS->CallOutputs ? E3dUF_ALL : E3dUF_ALL - E3dUF_CALL_OUTPUTS, &LChanged);
	}

	return(LChanged);
}


//================================================
// Constructor
//================================================
static int _Init(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dInfo*  LThis = (E3dInfo*)PInfo;


	EList_Init(LThis->ProfileGeometries);
	LThis->Profiles = NULL;

	LThis->ProfileFollowing = E3dSTRICT;

	LThis->Profile3D = FALSE;
	LThis->Depth  = 1.0;
	LThis->Offset = 0.0;
	LThis->Twist = 0.0;LThis->TwistBevels = FALSE;
	E_SetXY_s(LThis->Scale,  1.0, 1.0);LThis->ScaleBevels = FALSE;
	LThis->PathSteps = 1;
	LThis->Axis.X = 0.0;LThis->Axis.Y = 0.0;LThis->Axis.Z = 1.0;  // Extrude along the Z axis
	LThis->SmoothingAngle = 60.0;

	LThis->CenterAlongPath = TRUE;

	LThis->TrianglesOnCaps  = FALSE;
	LThis->TrianglesOnSides = FALSE;
	LThis->CloseStart       = TRUE;
	LThis->CloseEnd         = TRUE;
	LThis->CloseSides       = TRUE;
	LThis->CreateUVs        = TRUE;
	LThis->CreateTangents   = TRUE;

	LThis->IncludeBevelsInLength = TRUE;

	E3d_BevelInit(&LThis->StartBevel);
	E3d_BevelInit(&LThis->EndBevel);

// Initialize InputInterface
//
	{
		EInterface*  LInput = &LThis->InputInterface;E_InterfaceInit(LInput);
		LInput->InputChanged    = _InputChanged;
		LInput->Object          = (ECoreObject*)PObject;
		LInput->ObjectComponent = (ECoreObject*)PInfo;
	}

	return(E_SUCCESS);
}


//========================================
// Copy-constructor
//========================================
static void _Copy(const E3dObject* PSObject, const E3dObjectComponent* PSInfo,  E3dObject* PDObject, E3dObjectComponent* PDInfo)
{
	const E3dInfo*  LSInfo = (E3dInfo*)PSInfo;
	E3dInfo*        LDInfo = (E3dInfo*)PDInfo;

	_Init(PDObject, PDInfo);

#define _CopyField(f)  LDInfo->f = LSInfo->f
	_CopyField(ProfileFollowing);
	_CopyField(Profile3D);
	_CopyField(Depth);
	_CopyField(Offset);
	_CopyField(Twist);_CopyField(TwistBevels);
	_CopyField(Scale);_CopyField(ScaleBevels);
	_CopyField(PathSteps);
	_CopyField(Axis);
	_CopyField(SmoothingAngle);
	_CopyField(TrianglesOnCaps);
	_CopyField(TrianglesOnSides);
	_CopyField(CloseStart);
	_CopyField(CloseEnd);
	_CopyField(CloseSides);
	_CopyField(CreateUVs);
	_CopyField(CreateTangents);
	_CopyField(CenterAlongPath);
	_CopyField(IncludeBevelsInLength);

/*
	_CopyField(StartBevel.NumOfRoundingSteps);
	_CopyField(StartBevel.Radius);
	_CopyField(EndBevel.NumOfRoundingSteps);
	_CopyField(EndBevel.Radius);
*/
#undef _CopyField

	if(LSInfo->ProfileGeometries_Count)
	{
		ENewNClr(LSInfo->ProfileGeometries_Count, LDInfo->ProfileGeometries);
		EIndex  LC = 0;for(;LC < LSInfo->ProfileGeometries_Count;LC++)
		{
			EObject*  LProfileG = LSInfo->ProfileGeometries[LC];if(LProfileG)
			{
				E_AssignToMember(LProfileG, LDInfo,ProfileGeometries[LC]);
				if(EObject_IsSubClass(LProfileG, E3dGeometry)) E3dBaseObject_AddOutput((E3dBaseObject*)LProfileG, &LDInfo->InputInterface);
			}
		}

		LDInfo->Profiles = E3dExtrude_ProfilesCreate(LDInfo->ProfileGeometries, LSInfo->ProfileGeometries_Count);

		LDInfo->ProfileGeometries_Count = LSInfo->ProfileGeometries_Count;
	}

	{
		E3dBevel*  LBevel = &LDInfo->StartBevel;
		E3d_BevelCopy(&LSInfo->StartBevel, &LDInfo->StartBevel);
		if(LBevel->NumOfRoundingSteps) E3d_BevelCreateSegments(LBevel);else LBevel->Type = 0;
	}

	{
		E3dBevel*  LBevel = &LDInfo->EndBevel;
		E3d_BevelCopy(&LSInfo->EndBevel,   &LDInfo->EndBevel);
		if(LBevel->NumOfRoundingSteps) E3d_BevelCreateSegments(LBevel);else LBevel->Type = 0;
	}
}


//========================================================
// "Extrude" ObjectComponent class destructor
//========================================================
static int _Free(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dInfo*  LInfo = (E3dInfo*)PInfo;


//printf("%s() %s:%d  %p %p %d\n", __FUNCTION__, __FILE__, __LINE__, LInfo->Profiles, LInfo->ProfileGeometries, (int)(LInfo->ProfileGeometries_Count));fflush(stdout);
	E3dExtrude_ProfilesFree(PObject, (E3dExtrudeSplineInfoCommon*)PInfo);

	E3d_BevelClear(&LInfo->StartBevel);
	E3d_BevelClear(&LInfo->EndBevel);

	E_InterfaceClear(&LInfo->InputInterface);

	return(E_SUCCESS);
}


static E3dInfo*         _TmpInfo = NULL;
static E3dObjectPanel*  _TmpPanel = NULL;

#include "ExtrudeCommon.h"

static void _RequestProfilePick(void)
{
	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color = Xe_ForegroundRGBA8pColor;

	E3dp_PickRequest(
	 XeString_NewVa(TRUE,
	 XeFONT, XeFontDEFAULT,  XeCOLOR_RGBA, LLabelRGBAp8Color,  XeTEXT, "Extrude:    ",
	 E3dp_LMouseButtonS,     XeTEXT, ": pick Spline or Face as the new profile  ",
	 E3dp_RMouseButtonEscS,  XeTEXT, ": cancel",
	 0),
	 E3dItemSPLINE|E3dItemFACE|E3dItemMODEL,  _PickCB_Profile, NULL,  0, FALSE);
}


static void _DCB_ProfilePick(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	_TmpInfo = LInfo;
	_TmpPanel = LPanel;
	_RequestProfilePick();
}


static void _DCB_ProfileShow(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);
	EChangeFlags     LChanged = 0;

	EIndex  LC = 0, LFound = 0;const EIndex  LN = LInfo->ProfileGeometries_Count;for(;LC < LN;LC++)
	{
		E3dObject*  LProfileObject = (E3dObject*)LInfo->ProfileGeometries[LC];if(LProfileObject)
		{
			LFound++;
			if(E3dSel_ObjectSelect(E3d_Scene, LProfileObject, FALSE, FALSE)) LChanged |= E3dChgSELECTION;
		}
	}

	if(LFound)
	{
		E3dWindows_RedrawByChangeFlags(LChanged);
		E3dTransformPanel_Update(NULL, E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
		E3dp_PrintMessage(ELogMESSAGE, 3000, "Extrude: selected profile Geometries: %d", LFound);
	}
	else E3dp_PrintMessage(ELogMESSAGE, 3000, "Extrude: no profile Geometries!");
}


static void _DCB_Profile3D(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->Profile3D = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


//================================================================
// GUIProc for "Extrude"
//================================================================
void _GUI(E3dObject* PObject, E3dObjectComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dResourcesGUI*         LResourcesGUI;
	E3dInfo*                 LInfo = (E3dInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

				E3d_BevelPanelUpdate(&LResourcesGUI->StartBevelPanel, &LInfo->StartBevel);
				E3d_BevelPanelUpdate(&LResourcesGUI->EndBevelPanel, &LInfo->EndBevel);

				XeToggleButton_SetState(LResourcesGUI->CloseStartW, LInfo->CloseStart, FALSE);
				XeToggleButton_SetState(LResourcesGUI->CloseEndW,   LInfo->CloseEnd, FALSE);
				XeToggleButton_SetState(LResourcesGUI->CloseSidesW, LInfo->CloseSides, FALSE);

				XeToggleButton_SetState(LResourcesGUI->TrianglesOnCapsW,  LInfo->TrianglesOnCaps, FALSE);
				XeToggleButton_SetState(LResourcesGUI->TrianglesOnSidesW, LInfo->TrianglesOnSides, FALSE);
				XeToggleButton_SetState(LResourcesGUI->CenteredW,         LInfo->CenterAlongPath, FALSE);
				XeToggleButton_SetState(LResourcesGUI->Profile3DW,        LInfo->Profile3D, FALSE);

				XeToggleButton_SetState(LResourcesGUI->IncludeBevelsInLengthW, LInfo->IncludeBevelsInLength, FALSE);
				XeScale_SetValue_f32(LResourcesGUI->SmoothingAngleW, LInfo->SmoothingAngle, FALSE);

				XeScale_SetValue_Coord(LResourcesGUI->DepthW,  LInfo->Depth, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->OffsetW, LInfo->Offset, FALSE);
				if(LResourcesGUI->TwistW)  XeScale_SetValue_Coord(LResourcesGUI->TwistW,  LInfo->Twist,   FALSE);
				if(LResourcesGUI->ScaleXW) XeScale_SetValue_Coord(LResourcesGUI->ScaleXW, LInfo->Scale.X, FALSE);
				if(LResourcesGUI->ScaleYW) XeScale_SetValue_Coord(LResourcesGUI->ScaleYW, LInfo->Scale.Y, FALSE);
				XeScale_SetValue(LResourcesGUI->PathStepsW, LInfo->PathSteps, FALSE);

				E3dpAxisSelector_Set(&LResourcesGUI->AxisSelector,  LInfo->Axis.X, LInfo->Axis.Y, LInfo->Axis.Z);
			}
			else
			{
				Widget  LTopMatrixW = LDrawer->Drawer.DrawerW, LW;
				Widget  LIndentWs[16];int  LIndentN;      // For indenting the ScaleWidgets
				Widget  LTraverseWs[16];int  LTraverseN;  // For TextField traversal
				EArg   Args[256];int  ArgCnt;


				ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

// Options
//
				{
					EXeStart;
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNspacingX, 12);
					EXeSetArg(XeNcolumns, 3);EXeSetArg(XeNrows, 2);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("LMatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->CloseStart);
					LResourcesGUI->CloseStartW = LW = XeWidget_ManagedNew("Close start", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CloseStart, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->CloseEnd);
					LResourcesGUI->CloseEndW = LW = XeWidget_ManagedNew("Close end", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CloseEnd, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->TrianglesOnCaps);
					XeToolTip8("Triangulate start and end caps.");
					LResourcesGUI->TrianglesOnCapsW = LW = XeWidget_ManagedNew("Triangles", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_TrianglesOnCaps, (EPointer)LPanel);


					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->CloseSides);
					LResourcesGUI->CloseSidesW = LW = XeWidget_ManagedNew("Close sides", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_CloseSides, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->TrianglesOnSides);
					LResourcesGUI->TrianglesOnSidesW = LW = XeWidget_ManagedNew("Side triangles", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_TrianglesOnSides, (EPointer)LPanel);

					EXeStart;
					LW = XeWidget_ManagedWindowlessNew("Pad", XeCoreWidget_class, LMatrixW,  Args, ArgCnt);



					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->CenterAlongPath);
					XeToolTip8("Center along extrusion direction");
					LResourcesGUI->CenteredW = LW = XeWidget_ManagedNew("Centered", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Centered, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->IncludeBevelsInLength);
					XeToolTip8("Include bevels in the length of extrusion ('Depth').");
					LResourcesGUI->IncludeBevelsInLengthW = LW = XeWidget_ManagedNew("Include bevels", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_IncludeBevelsInLength, (EPointer)LPanel);
				}


// Profile
//
				{
					EXeStart;
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginSizeY, 5);
					EXeSetArg(XeNspacingX, 0);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNmarginSizeX, 2);EXeSetArg(XeNmarginSizeY, 3);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeWidget_ManagedNew("Profile", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNspacingLeft, 6);
					EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNlabelType, XeIMAGE);EXeSetArg(XeNimageSizeYPtr, &Xe_IconSize);EXeSetImages2(_Xe_Theme->Image_Find);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Select the Splines and/or Faces that are used as profiles for this extrusion (so you can find them).");
					LW = XeWidget_ManagedNew("Show", XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_ProfileShow, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNlabelType, XeIMAGE);EXeSetArg(XeNimageSizeYPtr, &Xe_IconSize);EXeSetImages2(_Xe_Theme->Image_Replace);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Pick a new profile Spline or Face.");
					LW = XeWidget_ManagedNew("Pick", XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_ProfilePick, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNspacingLeft, 6);
					EXeSetArg(XeNset, LInfo->Profile3D);
					XeToolTip8("Consider the Z coordinate of the profile Spline / Face.\nIf this is OFF, the caps will be flat, even if the profile Spline/Face has curvature in the Z direction.");
					LResourcesGUI->Profile3DW = LW = XeWidget_ManagedNew("3D", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Profile3D, (EPointer)LPanel);
				}


				LTraverseN = 0;

				EXeStart;
				EXeSetArg(XeNshadowThickness, 2);
				EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeWidget_ManagedNew("", XeSeparatorWidget_class, LTopMatrixW,  Args, ArgCnt);
				E3d_BevelCreateDrawer(LTopMatrixW, "Bevel start", &LResourcesGUI->StartBevelPanel, _CB_Bevel, (EPointer)LPanel, &LInfo->StartBevel);
				LTraverseWs[LTraverseN++] = LResourcesGUI->StartBevelPanel.RadiusW;
				LTraverseWs[LTraverseN++] = LResourcesGUI->StartBevelPanel.RoundingStepsW;

				EXeStart;
				EXeSetArg(XeNshadowThickness, 2);
				EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeWidget_ManagedNew("", XeSeparatorWidget_class, LTopMatrixW,  Args, ArgCnt);
				E3d_BevelCreateDrawer(LTopMatrixW, "Bevel end", &LResourcesGUI->EndBevelPanel, _CB_Bevel, (EPointer)LPanel, &LInfo->EndBevel);
				LTraverseWs[LTraverseN++] = LResourcesGUI->EndBevelPanel.RadiusW;
				LTraverseWs[LTraverseN++] = LResourcesGUI->EndBevelPanel.RoundingStepsW;

				EXeStart;
				EXeSetArg(XeNshadowThickness, 2);
				EXeSetArg(XeNshadowType, XeShadowETCHED_IN);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeWidget_ManagedNew("", XeSeparatorWidget_class, LTopMatrixW,  Args, ArgCnt);

				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->DepthW = LW = XeScaleNew_Coord(LTopMatrixW, "Depth",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Depth,  -1000000.0, 1000000.0,  -10.0, 10.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Depth, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->OffsetW = LW = XeScaleNew_Coord(LTopMatrixW, "Offset",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Offset,  -1000000.0, 1000000.0,  -10.0, 10.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Offset, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->TwistW = LW = XeScaleNew_Coord(LTopMatrixW, "Twist",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Twist,  -7200.0, 7200.0,  -360.0, 360.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Twist, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->ScaleXW = LW = XeScaleNew_Coord(LTopMatrixW, "Scale X",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Scale.X,  0.0001, 1000.0,  0.1, 10.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_ScaleX, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 4);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->ScaleYW = LW = XeScaleNew_Coord(LTopMatrixW, "Scale Y",  Args, ArgCnt);
				XeScale_SetValueAndRanges_Coord(LW,  LInfo->Scale.Y,  0.0001, 1000.0,  0.1, 10.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_ScaleY, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 5);
				EXeSetArg(XeNminimum, 1);EXeSetArg(XeNmaximum, 500);
				EXeSetArg(XeNscaleMaximum, 100);
				EXeSetArg(XeNvalue, LInfo->PathSteps);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->PathStepsW = LW = XeWidget_ManagedNew("Path steps", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_PathSteps, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 5);EXeSetArg(XeNdecimalPoints, 2);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SmoothingAngleW = LW = XeScale_Newf32(LTopMatrixW, "Smoothing angle",  Args, ArgCnt);
				XeScale_SetValueAndRange_f32(LW,  LInfo->SmoothingAngle,  0.0, 180.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SmoothingAngle, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				EXeSetArg(XeNmarginSizeX, 2);EXeSetArg(XeNmarginSizeY, 3);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeWidget_ManagedNew("Extrude along:", XeLabelWidget_class, LTopMatrixW,  Args, ArgCnt);

				E3dpAxisSelector_New(&LResourcesGUI->AxisSelector, NULL, LTopMatrixW, TRUE,  _CB_AxisSelector, (EPointer)LPanel,  LInfo->Axis.X, LInfo->Axis.Y, LInfo->Axis.Z);

				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 2);

// Set up TextField traversal
//
				Xe_TraversalRing(LTraverseWs, LTraverseN);
			}
			if(LInfo->CenterAlongPath) Xe_SetValuesVa(LResourcesGUI->OffsetW,  XeNsensitive, FALSE,  NULL);else Xe_SetValuesVa(LResourcesGUI->OffsetW,  XeNsensitive, TRUE,  NULL);
		break;
	}
}


//========================================
// Rebuild() (create/change) method
//========================================
static E3dObject* _Update(E3dObject* PObject, E3dObjectComponent* PInfo, const EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	E3dScene*  LScene = E3d_Scene;
	E3dMesh*   LMesh;
	E3dInfo*   LInfo;
	EBool      LNewObject;

//printf("%s() %s:%d  <%s>\n", __FUNCTION__, __FILE__, __LINE__, PObject ? PObject->Name : "-");fflush(stdout);

	if(PObject == NULL)
	{
		E_New(E3dMesh, LMesh);
		PObject = (E3dObject*)LMesh;
		LInfo = (E3dInfo*)E3dObjectComponent_Add((E3dObject*)LMesh, _ObjectComponent_class);
		if(PInfo && PInfo->Class == _ObjectComponent_class) _Copy(PObject, PInfo, PObject, (E3dObjectComponent*)LInfo);

// Add output to the Profile Geometries
//
		EIndex  LProfileC = 0;for(;LProfileC < LInfo->ProfileGeometries_Count;LProfileC++)
		{
			EObject*  LProfileG = LInfo->ProfileGeometries[LProfileC];

			if(EObject_IsSubClass(LProfileG, E3dGeometry)) E3dBaseObject_AddOutput((E3dBaseObject*)LProfileG, &LInfo->InputInterface);
		}

		LNewObject = TRUE;
	}
	else
	{
		if(!EObject_IsSubClass(PObject, E3dMesh)) return(NULL);

		LMesh = (E3dMesh*)PObject;
		LInfo = (E3dInfo*)E3dObjectComponent_ByClass((E3dObject*)LMesh, _ObjectComponent_class);

		if(PCheckOnly)
		{
			E3dExtrude_CheckBevelsAndProfiles((E3dObject*)LMesh, (E3dExtrudeSplineInfoCommon*)LInfo);
		}

		LNewObject = FALSE;
	}

//printf("%s() %s:%d  <%s> %p %p\n", __FUNCTION__, __FILE__, __LINE__, PObject ? PObject->Name : "-", LInfo->ProfileGeometries, LInfo->Profiles);fflush(stdout);


	if(LInfo->ProfileGeometries_Count && LInfo->ProfileGeometries == NULL) { LInfo->ProfileGeometries_Count = 0;return(NULL); }
	if(PCheckOnly || LInfo->ProfileGeometries_Count == 0) return(NULL);


// If any of the profiles is missing, bail
//
	{
		EIndex  LProfileC = 0;for(;LProfileC < LInfo->ProfileGeometries_Count;LProfileC++)
		{
			if(LInfo->ProfileGeometries[LProfileC] == NULL) return(NULL);
		}
	}

	if(LMesh)
	{
		E3dMesh_FreeCaches(LMesh);

		if(PUpdateFlags & E3dUF_EDGES) E3dMesh_EdgesFree(LMesh);

// Calculate number of Vertices
//
		EIndex  LPathStepsWithBevel = LInfo->PathSteps;
		if(LInfo->StartBevel.Type) LPathStepsWithBevel += LInfo->StartBevel.Profile->SNumOfLinearSegments;
		if(LInfo->EndBevel.Type) LPathStepsWithBevel += LInfo->EndBevel.Profile->SNumOfLinearSegments;
		const EIndex	LNumOfPathPoints = LPathStepsWithBevel + 1;

		EIndex  LNumOfVertices = 0;

		{
			EIndex  LProfileC = 0;for(;LProfileC < LInfo->ProfileGeometries_Count;LProfileC++)
			{
				E3dExtrudeSplineProfile*  LProfile = LInfo->Profiles[LProfileC];

				LNumOfVertices += LNumOfPathPoints * LProfile->NumOfPoints;
			}
		}

		if(LNumOfVertices == 0)
		{
			E3dMesh_PolyGroupsFree(LMesh);
			E3dMesh_VerticesFree(LMesh);
			E3dObject_UpdateForDisplay(PObject, ~E3dUF_TRIANGULATE, LScene);
			return(PObject);
		}

// Allocate Vertices for the Mesh
//
		if(LNumOfVertices != LMesh->Vertices_Count)
		{
			E3dMesh_VerticesFree(LMesh);
			E3dMesh_VerticesAdd(LMesh, LNumOfVertices);
		}

		const EIndex  LNumOfPolyGroups = LInfo->ProfileGeometries_Count;

// Free unused PolyGroups
//
		if(LMesh->PolyGroups_Count > LNumOfPolyGroups)
		{
			EIndex  LProfileC = LNumOfPolyGroups;for(;LProfileC < LMesh->PolyGroups_Count;LProfileC++) E3dMesh_PolyGroupDel(LMesh, LMesh->PolyGroups[LProfileC]);
		}

		{
			E3dVertex*  LVertex = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;
			EIndex      LVertexOffset = 0;
			int         LPolyGroupC = 0;

			uint8_t     LCloseEnds = 0;
			if(LInfo->CloseStart) LCloseEnds |= ExtrCLOSE_START;
			if(LInfo->CloseEnd) LCloseEnds |= ExtrCLOSE_END;

			E3dMesh_EdgesFree(LMesh);  // Edges will be built in E3dExtrude_CreatePolygons

			EIndex  LProfileC = 0;for(;LProfileC < LInfo->ProfileGeometries_Count;LProfileC++)
			{
				E3dExtrudeSplineProfile*  LProfile = LInfo->Profiles[LProfileC];
				EObject*                  LProfileGeo = LInfo->ProfileGeometries[LProfileC];
				E3dFace*                  LProfileFace = LProfile->Face ? LProfile->Face : E_DynamicCast(LProfileGeo, E3dFace);
				E3dPolyGroup*             LPolyGroup;
				E3dPolygon*               LPolygons;
				const EIndex              LNumOfProfileHoles = LProfile->NumOfHoles;
				const EIndex              LNumOfProfilePoints = LProfile->NumOfPoints, LNumOfProfileSegments = LProfile->NumOfSegments;

//printf("%s() %s:%d  LNumOfProfilePoints %d  LNumOfProfileSegments %d\n", __FUNCTION__, __FILE__, __LINE__, (int)LNumOfProfilePoints, (int)LNumOfProfileSegments);fflush(stdout);

				if(LMesh->PolyGroups_Count <= LPolyGroupC) LPolyGroup = E3dMesh_PolyGroupAdd(LMesh);
				else LPolyGroup = LMesh->PolyGroups[LPolyGroupC];
				LPolyGroupC++;
				E3dPolyGroup_FreeContents(LPolyGroup);
				E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);
				E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_UV, 0,  NULL);    // Make sure that the VertexNodes have UVs (evaluator coordinates)

				EIndex  LNumOfPolygons = LPathStepsWithBevel * LNumOfProfileSegments;

				if(LInfo->TrianglesOnSides) LNumOfPolygons *= 2;

				if(LProfileFace->PolyGroup.Polygons_Count == 0) E3dFace_Tesselate(LProfileFace);
				E3dPolyGroup*  lPolyGroup = &LProfileFace->PolyGroup;
				const EIndex   lTriangles_Count = E3dPolyGroup_CountTriangles(lPolyGroup);

				if(LCloseEnds & ExtrCLOSE_START)
				{
					if(LInfo->TrianglesOnCaps) LNumOfPolygons += lTriangles_Count;
					else LNumOfPolygons += lPolyGroup->Polygons_Count;
				}

				if(LCloseEnds & ExtrCLOSE_END)
				{
					if(LInfo->TrianglesOnCaps) LNumOfPolygons += lTriangles_Count;
					else LNumOfPolygons += lPolyGroup->Polygons_Count;
				}

// Allocate Polygons for the PolyGroup
//
				if(LNumOfPolygons == LPolyGroup->Polygons_Count) LPolygons = LPolyGroup->Polygons;
				else
				{
					if(LPolyGroup->Polygons) E3dPolyGroup_FreePolygons(LPolyGroup);
					if((LPolygons = E3dPolygons_New(LNumOfPolygons)) != NULL)
					{
						LPolyGroup->Polygons = LPolygons;LPolyGroup->Polygons_Count = LNumOfPolygons;
					}
				}

				if(LNumOfProfilePoints)
				{
					E3dCoord2i_v(PProfile2DBBoxMin,  0.0, 0.0);
					E3dCoord2i_v(PProfile2DBBoxMax,  0.0, 0.0);
					E3dCoordinate*  LUs;EIndex  LNumOfUs;
					E3dCoord2*      L2DVertices = E3d_UVLayout((E3dExtrudeSplineInfoCommon*)LInfo, LProfile, LProfileFace,  NULL, E3d_Abs(LInfo->Depth),  &PProfile2DBBoxMin, &PProfile2DBBoxMax,  &LUs, &LNumOfUs,  LCloseEnds);

					E3dPolygon*     LPolygon = LPolyGroup->Polygons;
					if(EObject_IsSubClass(LProfileGeo, E3dSpline))
					{
						E3dCoordinate   LTotalVLength;EIndex	LTotalNumOfVs;
						E3dCoordinate*  LVs;

						if(LInfo->SidesUVOrientation == E3dXY_TO_UV)
						{
							LVs = E3dExtrude_FaceGetVs(LProfileFace, E3dV_SPACING / LInfo->NormSidesVLength,  &LTotalVLength, &LTotalNumOfVs);
							_ScaleAndBias(LVs, LTotalNumOfVs,  LInfo->NormSidesVLength, LInfo->NormSidesVStart);
						}
						else
						{
							LVs = E3dExtrude_FaceGetVs(LProfileFace, E3dU_SPACING / LInfo->NormSidesULength,  &LTotalVLength, &LTotalNumOfVs);
							_ScaleAndBias(LVs, LTotalNumOfVs,  LInfo->NormSidesULength, LInfo->NormSidesUStart);
						}

//printf("%s() %s:%d  LNumOfUs %d   %d %d %d\n", __FUNCTION__, __FILE__, __LINE__, LNumOfUs, ((E3dSpline*)LProfileGeo)->SNumOfLinearSegments, LNumOfProfilePoints, LNumOfProfileSegments);fflush(stdout);
//printf("%s() %s:%d LInfo->PathSteps %d LPathStepsWithBevel %d\n", __FUNCTION__, __FILE__, __LINE__, LInfo->PathSteps, LPathStepsWithBevel);fflush(stdout);

						LVertex = EStruct_ByIndex(LVertex, LVertexSize, _SetVertices((E3dGeometry*)LProfileGeo,  LVertex, LVertexSize,  LInfo, LNumOfProfilePoints));

						if(LInfo->CloseSides) LPolygon = E3dExtrude_CreatePolygons(LMesh, (E3dExtrudeSplineInfoCommon*)LInfo, (E3dSpline*)LProfileGeo, NULL,  LUs, LNumOfUs,  LVs, LNumOfProfilePoints, LNumOfProfileSegments,  LInfo->PathSteps, LPathStepsWithBevel, LVertexOffset, LPolygon, LPolyGroup,  LInfo->Depth >= 0.0, FALSE);

						EFree(LVs);
					}
					else if(EObject_IsSubClass(LProfileGeo, E3dFace))
					{
// If we just created a new Object (e.g. a Mesh), assign the Material of the Face to the PolyGroups.
// Otherwise leave the Material alone, because the user might have set it.
//
						if(LNewObject)
						{
							E_AssignToMember(LProfileFace->PolyGroup.Material, LPolyGroup,Material);
						}
						E3dM_ObjectUpdateDrawingMaterial((E3dObject*)LMesh, LPolyGroup->Material, LPolyGroup,DrawingMaterial);

						LVertex = EStruct_ByIndex(LVertex, LVertexSize, _SetVertices((E3dGeometry*)LProfileFace,  LVertex, LVertexSize,  LInfo, LNumOfProfilePoints));

						{
							E3dCoordinate   LTotalVLength;EIndex  LTotalNumOfVs;
							E3dCoordinate*  LVs;

							if(LInfo->SidesUVOrientation == E3dXY_TO_UV)
							{
								LVs = E3dExtrude_FaceGetVs(LProfileFace, E3dV_SPACING / LInfo->NormSidesVLength,  &LTotalVLength, &LTotalNumOfVs);
								_ScaleAndBias(LVs, LTotalNumOfVs,  LInfo->NormSidesVLength, LInfo->NormSidesVStart);
							}
							else
							{
								LVs = E3dExtrude_FaceGetVs(LProfileFace, E3dU_SPACING / LInfo->NormSidesULength,  &LTotalVLength, &LTotalNumOfVs);
								_ScaleAndBias(LVs, LTotalNumOfVs,  LInfo->NormSidesULength, LInfo->NormSidesUStart);

/*
{
printf("%s() %s:%d LVs %d:", __FUNCTION__, __FILE__, __LINE__, (int)LTotalNumOfVs);fflush(stdout);
EIndex  LC = 0;for(;LC < LTotalNumOfVs;LC++)
{
	printf("  %f", LVs[LC]);fflush(stdout);
}
printf("\n");fflush(stdout);
}
*/
							}

							E3dCoordinate*  LVPtr = LVs;

							EIndex  LVOffset = LVertexOffset;

							{
								E3dSpline*    LProfileSpline = LProfileFace->Exterior;
								const EIndex  LContourNumOfVSegments = LProfileSpline->SNumOfLinearSegments;
								EIndex        LContourNumOfVPoints = LContourNumOfVSegments;if(!LProfileSpline->Closed) LContourNumOfVPoints++;

								if(LInfo->CloseSides) LPolygon = E3dExtrude_CreatePolygons(LMesh, (E3dExtrudeSplineInfoCommon*)LInfo, LProfileFace->Exterior, NULL,  LUs, LNumOfUs,  LVPtr, LNumOfProfilePoints, LContourNumOfVSegments,  LInfo->PathSteps, LPathStepsWithBevel,  LVOffset, LPolygon, LPolyGroup,  LInfo->Depth >= 0.0, FALSE);
								LVPtr    += LContourNumOfVPoints;
								LVOffset += LContourNumOfVPoints;

//printf("%s() %s:%d   <%s> LContourNumOfVSegments %d  LContourNumOfVPoints %d\n", __FUNCTION__, __FILE__, __LINE__, LProfileFace->Name, (int)LContourNumOfVSegments, (int)LContourNumOfVPoints);fflush(stdout);
							}

							if(LNumOfProfileHoles)
							{
								EIndex  LHC = 0;for(;LHC < LNumOfProfileHoles;LHC++)
								{
									E3dSpline*    LProfileSpline = LProfileFace->Holes[LHC];
			            const EIndex  LContourNumOfVSegments = LProfileSpline->SNumOfLinearSegments;
									EIndex        LContourNumOfVPoints = LContourNumOfVSegments;if(!LProfileSpline->Closed) LContourNumOfVPoints++;

									if(LInfo->CloseSides) LPolygon = E3dExtrude_CreatePolygons(LMesh, (E3dExtrudeSplineInfoCommon*)LInfo,  LProfileFace->Holes[LHC], NULL,
									                                     LUs, LNumOfUs,  LVPtr, LNumOfProfilePoints, LContourNumOfVSegments,
									                                     LInfo->PathSteps, LPathStepsWithBevel,  LVOffset, LPolygon, LPolyGroup,  LInfo->Depth >= 0.0, TRUE);
									LVPtr    += LContourNumOfVPoints;
									LVOffset += LContourNumOfVPoints;

//printf("%s() %s:%d   <%s> LContourNumOfVSegments %d  LContourNumOfVPoints %d\n", __FUNCTION__, __FILE__, __LINE__, LProfileFace->Name, (int)LContourNumOfVSegments, (int)LContourNumOfVPoints);fflush(stdout);
								}
							}

							EFree(LVs);
						}
					}

//printf("%s() %s:%d LVertexOffset %d, LNumOfPathPoints %d, LNumOfProfilePoints %d\n", __FUNCTION__, __FILE__, __LINE__, LVertexOffset, LNumOfPathPoints, LNumOfProfilePoints);fflush(stdout);

					if(LCloseEnds) LPolygon = E3dExtrude_CloseEnds((E3dExtrudeSplineInfoCommon*)LInfo,  LProfile, LProfileFace,  L2DVertices, &PProfile2DBBoxMin, &PProfile2DBBoxMax,  LMesh, LPolyGroup, LPolygon, LVertexOffset, LNumOfPathPoints, LNumOfProfilePoints, LCloseEnds, LInfo->Depth >= 0.0);
					EFreeChk(L2DVertices);
					EFree(LUs);

					LPolyGroup->SmoothingAngle = LInfo->SmoothingAngle;
					if(LInfo->SmoothingAngle > 0.0)
					{
						LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;

						LPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;
					}
					else
					{
						LPolyGroup->VertexNormalType = 0;
						LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;
					}

					LPolyGroup->Polygons_Count = LPolygon - LPolyGroup->Polygons;
				}

				LVertexOffset += LNumOfPathPoints * LNumOfProfilePoints;
			}
		}

#if E3dEXTRUDESPLINE_CREATE_EDGES
		const EBool   LCreateEdges = !LInfo->TrianglesOnSides;
#else
		const EBool   LCreateEdges = FALSE;
#endif

//E3dMesh_EdgeFlagsUpdate(LMesh);  // For debugging (to show selected Edges).

		EChangeFlags  LChanged;

		if(LCreateEdges)
		{
			LChanged = E3dObject_UpdateForDisplay(PObject, PUpdateFlags & (~(E3dUF_TRIANGULATE|E3dUF_EDGES)), LScene);
		}
		else
		{
			LChanged = E3dObject_UpdateForDisplay(PObject, PUpdateFlags & (~E3dUF_TRIANGULATE), LScene);
		}

//{ char	s[256];printf("%s() %s:%d  [%s]\n", __FUNCTION__, __FILE__, __LINE__, E3d_SceneChangeFlagsToString(LChanged, s));fflush(stdout); }

		if(PTriggeredChangesP) *PTriggeredChangesP |= LChanged;
	}

//	if(LNewModel) E3dNodeHrc_UpdateMaterialsForDisplay((E3dNode*)PModel);
	return(PObject);
}


//================================================
// Menu/ToolBar callback
//================================================
static void _MCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*    LScene = E3d_Scene;
	const EBool  LShowOptions = ((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS;

	{
		E3dObject**  LObjects;const EIndex  LObjects_Count = E3dScene_GetObjects(E3d_Scene, &LObjects, E3dItemMESH, E3dSELECTED, TRUE);

		if(LObjects) { E3d_ExtrudeMesh(LObjects, LObjects_Count,  LShowOptions);return; }
	}

	E3dObject**  LObjects;const EIndex  LObjects_Count = E3dScene_GetObjects(E3d_Scene, &LObjects, E3dItemSPLINE|E3dItemFACE, E3dSELECTED, TRUE);if(LObjects)
	{
		E3dInfo  LTInfo;E3dObjectComponent_InitStatic(&LTInfo, _ObjectComponent_class);
		EBool    LSuccess = FALSE;

		LTInfo.ProfileGeometries = (EObject**)LObjects;LTInfo.ProfileGeometries_Count = LObjects_Count;

		E3dModel*  LOrigModel = E3dObject_GetSelectingModel(LObjects[0], FALSE);

// Check if all Objects have the same Model
//
		{
			EIndex  LC = 1;for(;LC < LObjects_Count;LC++)
			{
				E3dModel*  LSelectingModel = E3dObject_GetSelectingModel(LObjects[LC], FALSE);
				
				if(LSelectingModel != LOrigModel) { LOrigModel = NULL;break; }
			}
		}

		E3dNode*  LParentNode = NULL;

		if(LOrigModel == NULL)
		{
			EListAI(, E3dModel*, LModels);

			EIndex  LC = 0;for(;LC < LObjects_Count;LC++)
			{
				E3dModel*  LSelectingModel = E3dObject_GetSelectingModel(LObjects[LC], FALSE);
				
				EList_PointerAddAChk(LModels, 16, LSelectingModel);
			}

			LParentNode = E3dNodeHrc_NearestCommonAncestor((E3dNode**)LModels, LModels_Count);
			EList_FreeAChk(LModels);
		}

		E3dObject*  LObject = E3dObject_CustomNew(LScene, LParentNode, LOrigModel, NULL, (E3dObjectComponent*)(&LTInfo), E3dUF_ALL, _NewObjectAddingMode, (E3dp_Prefs.Duplicate_Flags & E3dCopyMODEL_UNIQUE_NAME) != 0,  NULL);if(LObject)
		{
			((E3dGeometry*)LObject)->AutoLOD = FALSE;

			LSuccess = TRUE;
		  E3dScene_ObjectsUpdateModelList(LScene);
			E3dScene_LayoutOnSchematics(E3d_Scene, TRUE);
		}

		EList_ObjectsFree(LTInfo.ProfileGeometries);

		if(LSuccess)
		{
			if(LShowOptions) E3dObject_Edit(LObject, _ObjectComponent_class);

			E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE|E3dChgHIERARCHY);
		}
	}
	else
	{
		E3dp_PrintMessage(ELogWARNING, 5000, "Select Splines, Face or Meshes to extrude");
	}
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* LPlugin)
{
	_Plugin = LPlugin;

	{
		E3dObjectComponentClass  LExtrudeClass=
		{
			.Name=        "Extrude",
			.StructSize=  sizeof(E3dInfo),
			.Resources=   _ExtrudeResources,

			.Init=        _Init,
			.Copy=        _Copy,
			.Free=        _Free,
			.Update=      _Update,
			.GUI=         _GUI
		};
		_ObjectComponent_class = E3dObjectComponent_ClassRegister(&LExtrudeClass);
	}

	if(_ObjectComponent_class)
	{
		EpM_CreateIconStates3_sp(Extrude);E_Assign(_Image_Extrude, _ObjectComponent_class->Icon);
		Xe_PushButtonAddImg3(_ToolBarButtons[_ToolBarButtons_Count++], Extrude, "Tool->Surface", "Extrude", '\0', NULL, NULL, TRUE, "Extrude a Spline or a Face along the selected axis to create a surface.\nRight button: dialog box for this tool", _MCB, (EPointer)0);
		_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Surface", "Extrude", '\0', NULL, NULL, TRUE, NULL, _MCB, (EPointer)0);
	}

	E3d_ExtrudeAlongSplineInit();
	E3d_ExtrudeMeshInit();

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		int  LC;

		for(LC = 0;LC < _MenuButtons_Count;LC++) { XeItemFree(_MenuButtons[LC]); } _MenuButtons_Count = 0;
		for(LC = 0;LC < _ToolBarButtons_Count;LC++) { XeItemFree(_ToolBarButtons[LC]); } _ToolBarButtons_Count = 0;
	}

	if(_ObjectComponent_class) { E3dObjectComponent_ClassDeactivate(_ObjectComponent_class);_ObjectComponent_class = NULL; }

	E3d_ExtrudeAlongSplineExit();
	E3d_ExtrudeMeshExit();

	EpIconFree3(Extrude);

	return(E_SUCCESS);
}
