//================================================================
// Interactive revolve tool
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 03:14:42
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#include <assert.h>
#include <math.h>
#include <stdio.h>

#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Label.h>
#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/PushButton.h>
#include <Xe/ToggleButton.h>

#include <E3D/E3D.h>
#include <E3D/Matrix.h>
#include <E3D/Select.h>

#include <E3D/GL/Draw.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/Panels.h>
#include <E3D/UI/StatusPanel.h>
#include <E3D/UI/WidgetCombos.h>


char*  E_PluginInfo = "Create a Mesh surface by revolving a Spline.";

static EPlugin*  _Plugin = NULL;


static const char*  E3dpLng_Revolve_Pick_a_Spline_or_a_Face_as_the_new_profile = "Pick a Spline or a Face as the new profile";


#define E3dCREATE_EDGES  1

typedef struct
{
	Widget            TrianglesW, TangentsW, UVsW,
	                  SmoothingAngleW,
	                  StartAngleScaleW, EndAngleScaleW,
	                  SidesScaleW;
	E3dpAxisSelector  AxisSelector;
} E3dResourcesGUI;


static EguiItem  _MenuButton = NULL, _ToolBarButton = NULL;

EpIconDef3(Revolve)


static const char  _NewObjectAddingMode = E3dNEW_INSTANCE;
//static const char  _NewObjectAddingMode = E3dNEW_MODEL;

static void _Copy(const E3dObject* PSObject, const E3dObjectComponent* PSInfo,  E3dObject* PDObject, E3dObjectComponent* PDInfo);

enum
{
	CR_TRIANGLES = XeCR_CUSTOM0,
	CR_TANGENTS,
	CR_UVS,
	CR_CHANGE_SMOOTHING_ANGLE,
	CR_SET_START_ANGLE,
	CR_SET_END_ANGLE,
	CR_SET_SIDES,
	CR_SET_CLOSE
};


#define RevCLOSE              0x01
#define RevAROUND_WORLD_AXIS  0x02
#define RevTRIANGLES          0x04
#define RevCREATE_TANGENTS    0x08


typedef struct
{
	E3dObjectComponentCore;

// Resources
//
	E3dGeometry*   InputGeometry;
	E3dCoordinate  StartAngle, EndAngle;
	float          SmoothingAngle;
	E3dCoord3      Axis;
	int            Sides;
	EBool          Triangles, CreateTangents, CreateUVs, UseProfileSegmentVisibility;
	EBool          Full, AroundWorldAxis, Close;

// Internal variables
//
	EInterface       InputInterface;     // This is added to the profile Geometry as an output node
	E3dGLCoordinate  GLAxis[2][3];
	EBool            GLAxisUptodate;
} E3dInfo;



#define ROffset(field)  (void*)(EPtrOffset(E3dInfo, field))
static EResource  _Resources[]=
{
	{ "Sides",            EResINT32,           ROffset(Sides) },
	{ "Axis",             E3dCOORDINATE_TYPE,  ROffset(Axis),      .Ext.Count= 3 },
//{ "AxX",              E3dCOORDINATE_TYPE,  ROffset(Axis.X) },
//{ "AxY",              E3dCOORDINATE_TYPE,  ROffset(Axis.Y) },
//{ "AxZ",              E3dCOORDINATE_TYPE,  ROffset(Axis.Z) },

	{ "ProfileGeometry",  EResOBJECT,          ROffset(InputGeometry) },

	{ "StartAngle",       E3dCOORDINATE_TYPE,  ROffset(StartAngle) },
	{ "EndAngle",         E3dCOORDINATE_TYPE,  ROffset(EndAngle) },
	{ "Tris",             EResBOOLEAN,         ROffset(Triangles) },
	{ "Tangents",         EResBOOLEAN,         ROffset(CreateTangents) },
	{ "UVs",              EResBOOLEAN,         ROffset(CreateUVs) },
	{ "UseProfileSegmentVisibility", EResBOOLEAN,  ROffset(UseProfileSegmentVisibility) },
	{ "SmoothingAngle",   EResFLOAT32,         ROffset(SmoothingAngle) },
	{ }
};
#undef ROffset



static E3dObjectComponentClass*  _ObjectComponent_class = NULL;

//========================================
// Set vertex coordinates
//========================================
static int* _SetVertices(E3dMesh* PMesh, E3dInfo* PInfo, E3dCoord3** PVertexTangentsRet, const EBool PSetVertices)
{
	E3dMatrix       LSplineToRotMatrix;
	E3dMatrix       LVertexMatrix;
	E3dCoord3*      LVertexTangents = NULL;
	E3dCoordinate*  LLinSegments;
	int*            LVertexReindexingTable = NULL;
	EBool           LFull = PInfo->Full, LUsingDefaultAxis;

	const E3dCoordinate  LStartAngle = PInfo->StartAngle, LEndAngle = PInfo->EndAngle;

	E3dSpline*  LProfileSpline = E_DynamicCast(PInfo->InputGeometry, E3dSpline);if(LProfileSpline == NULL) return(NULL);
	if(LProfileSpline == NULL || LProfileSpline->SLinearSegments == NULL) return(NULL);

	const int  LVPoints = E3d_SplineNumOfLinearSegmentPoints(LProfileSpline);
	const int  LSides = PInfo->Sides;

	int  LUPoints;
	if(LFull) LUPoints = LSides;
	else LUPoints = LSides + 1;

	E3dCoord3		LAxis = PInfo->Axis;
	if((LAxis.X == 0.0) && (LAxis.Y == 0.0) && (LAxis.Z == 0.0)) LAxis.Y = 1.0;
	E3d_Normalize3_v(LAxis);


	E3dMatrix_LoadIdentity(LVertexMatrix);
	E3dMatrix_LoadIdentity(LSplineToRotMatrix);

	const int		LDimensions = LProfileSpline->Dimensions;

// If the revolution is not around the Y axis, transform the contour coordinates accordingly
//
	if(LAxis.X == 0.0 && LAxis.Y == 1.0 && LAxis.Z == 0.0)
	{
		LLinSegments = LProfileSpline->SLinearSegments;

		LUsingDefaultAxis = TRUE;
	}
	else
	{
		E3dCoordinate*  LLinSegmentIn;
		E3dCoord3       LBiAxis, LTrAxis;

		switch(E3d_PrincipalAxis_v(LAxis))
		{
			case E3dX:          LBiAxis.X = 0.0;LBiAxis.Y = 1.0;LBiAxis.Z = 0.0;break;
			case E3dY:          LBiAxis.X = 1.0;LBiAxis.Y = 0.0;LBiAxis.Z = 0.0;break;
			case E3dZ:default:  LBiAxis.X = 0.0;LBiAxis.Y = 1.0;LBiAxis.Z = 0.0;break;
		}

		E3d_Cross3_vvv(LBiAxis,  LAxis,  LTrAxis);E3d_Normalize3_v(LTrAxis);

		E3d_Cross3_vvs(LAxis,  LTrAxis,  LSplineToRotMatrix[M00], LSplineToRotMatrix[M10], LSplineToRotMatrix[M20]);
		LSplineToRotMatrix[M01] = LAxis.X;  LSplineToRotMatrix[M11] = LAxis.Y;  LSplineToRotMatrix[M21] = LAxis.Z;
		LSplineToRotMatrix[M02] = LTrAxis.X;LSplineToRotMatrix[M12] = LTrAxis.Y;LSplineToRotMatrix[M22] = LTrAxis.Z;

		E3dMatrix_Invert3x3(LSplineToRotMatrix, LVertexMatrix);

		ENewN(LDimensions * LVPoints, LLinSegments);

		LLinSegmentIn = LProfileSpline->SLinearSegments;
		EIndex  LC = 0;E3dCoordinate*  LLinSegment = LLinSegments;for(;LC < LVPoints;LC++, LLinSegmentIn += LDimensions, LLinSegment += LDimensions)
		{
			E3dMatrix_Transform3x3_aa(LSplineToRotMatrix,  LLinSegmentIn,  LLinSegment);
		}

		LUsingDefaultAxis = FALSE;
	}

	if(PSetVertices)
	{
		{
			ENewN(PMesh->Vertices_Count, LVertexReindexingTable);
			int*  LIndexPtr = LVertexReindexingTable;

// See if we need to merge Vertices too close to the Y axis
//
			{
				EIndex  LIndex = 0;
				EIndex  LUC = 0;for(;LUC < LUPoints;LUC++)
				{
					EIndex  LVC = 0;for(LVC = 0;LVC < LVPoints;LVC++, LIndex++, LIndexPtr++)
					{
						*LIndexPtr = LIndex;
					}
				}
			}


#define EPSILON	0.0001
			EIndex  LVC = 0;E3dCoordinate*  LLinSegment = LLinSegments;for(;LVC < LVPoints;LVC++, LLinSegment += LDimensions)
			{
				const EIndex  LIndex = LVC;
				LIndexPtr = LVertexReindexingTable + LIndex;

				E3dCoordinate  LX = LLinSegment[E3dX];
				if(LX < 0.0) LX = 0.0;

				if(LX < EPSILON)
				{
					EIndex  LUC = 0;for(;LUC < LUPoints;LUC++, LIndexPtr += LVPoints) *LIndexPtr = LIndex;
				}
			}
		}

		const E3dCoordinate  LAngleStep = (LEndAngle - LStartAngle) / ((E3dCoordinate)LSides);

		{
			E3dCoordinate   LAngle = LStartAngle;

			EIndex  LUC = 0;E3dVertex*  LVertex = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;for(;LUC < LUPoints;LUC++, LAngle += LAngleStep)
			{
				EIndex  LVC = 0;E3dCoordinate*  LLinSegment = LLinSegments;for(;LVC < LVPoints;LVC++, LLinSegment += LDimensions)
				{
					E3dCoordinate  LX = LLinSegment[E3dX];
					if(LX < 0.0) LX = 0.0;

					const E3dCoordinate  LA = LAngle * E3dDEGREES_TO_RADIANS;
					LVertex->Position.X = E3d_Cos(LA) * LX;
					LVertex->Position.Y = LLinSegment[E3dY];
					LVertex->Position.Z = -E3d_Sin(LA) * LX;
					EStruct_Inc(LVertex, LVertexSize);
				}
			}
		}

		if(PInfo->CreateTangents)
		{
			ENewN(PMesh->Vertices_Count, LVertexTangents);

			E3dCoord3*  LVertexTangent = LVertexTangents;
			E3dCoordinate  LAngle = LStartAngle + 90.0;
			EIndex  LUC = 0;for(;LUC < LUPoints;LUC++, LAngle += LAngleStep)
			{
				EIndex  LVC = 0;for(;LVC < LVPoints;LVC++)
				{
					const E3dCoordinate  LA = LAngle*E3dDEGREES_TO_RADIANS;
					LVertexTangent->X = E3d_Cos(LA);
					LVertexTangent->Y = 0.0;
					LVertexTangent->Z = -E3d_Sin(LA);
					LVertexTangent++;
				}
			}
		}

// If the revolution is not around the Y axis, transform the Vertex coordinates accordingly
//
		if(!LUsingDefaultAxis)
		{
			const EIndex  LVN = PMesh->Vertices_Count;
			E3dVertex*    LVertex = PMesh->Vertices;const E3dVertexSize  LVertexSize = PMesh->VertexSize;

			{
				EIndex  LVC = 0;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LVertexSize))
				{
					E3dMatrix_Transform3x3_v(LVertexMatrix, LVertex->Position);
				}
			}

// Transform tangents
//
			if(LVertexTangents)
			{
				E3dCoord3*	LVertexTangent = LVertexTangents;

				EIndex  LVC = 0;for(;LVC < LVN;LVC++, EStruct_Inc(LVertex, LVertexSize), LVertexTangent++)
				{
					E3dMatrix_Transform3x3_p(LVertexMatrix, LVertexTangent);
				}
			}
		}
	}


// Update GLAxis
	{
		E3dCoordinate   LYMin, LYMax;
		E3dCoordinate*  LLinSegment = LLinSegments;

		LYMin = LYMax = LLinSegment[E3dY];

		EIndex  LC = 1;for(;LC < LVPoints;LC++, LLinSegment += LDimensions)
		{
			const E3dCoordinate  LY = LLinSegment[E3dY];
			if(LYMin > LY) LYMin = LY;
			if(LYMax < LY) LYMax = LY;
		}

		{
			const E3dCoordinate  LAxisLength = LYMax - LYMin;

			E3dMatrix_Transform3x3_sa(LVertexMatrix, 0.0, LYMin - LAxisLength * 0.16, 0.0,  PInfo->GLAxis[0]);
			E3dMatrix_Transform3x3_sa(LVertexMatrix, 0.0, LYMax + LAxisLength * 0.16, 0.0,  PInfo->GLAxis[1]);
		}

		PInfo->GLAxisUptodate = TRUE;
	}

// If we're not using the default axis, this array was allocated here, so we must free it
//
	if(!LUsingDefaultAxis) EFree(LLinSegments);

	if(PVertexTangentsRet) *PVertexTangentsRet = LVertexTangents;

	return(LVertexReindexingTable);
}


#define _M_Add2Tris(mV0, mV1, mV2, mV3)\
{\
	const E3dVertexIndex  l0 = LVertexReindexingTable[mV0];\
	const E3dVertexIndex  l1 = LVertexReindexingTable[mV1];\
	const E3dVertexIndex  l2 = LVertexReindexingTable[mV2];\
	const E3dVertexIndex  l3 = LVertexReindexingTable[mV3];\
	if(l0 == l1)\
	{\
		if(l2 != l3)\
		{\
			E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, l0, l2, l3);\
			LPolygon++;\
		}\
	}\
	else if(l2 == l3)\
	{\
		E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, l0, l1, l2);\
		LPolygon++;\
	}\
	else\
	{\
		E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LVertexReindexingTable[mV0], LVertexReindexingTable[mV1], LVertexReindexingTable[mV2]);\
		LPolygon++;\
		E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LVertexReindexingTable[mV2], LVertexReindexingTable[mV3], LVertexReindexingTable[mV0]);\
		LPolygon++;\
	}\
}


// Edges around the profile Spline's starting point, when the Spline's starting point touches the revolve axis and the quads degenerate to triangles
//
#define _EdgesT1(mVC)\
do {\
	if(l3 > l0) { LEdge->Start = l0;LEdge->End = l3; }\
	else { LEdge->Start = l3;LEdge->End = l0; }\
/*printf("%s() %s:%d  T1  %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, LEdge->Start, LEdge->End);fflush(stdout);*/\
	if(lUC % LUEdgeVisibilityStep != 0) LEdge->Flags &= ~E3dEdgeVISIBLE;\
	LEdge++;\
\
	if((!lPathClosed) && lUC == lPathLastSegment)\
	{\
		const EIndex  l3b = l3 + lProfilePointCount;\
		if(l3b > l0) { LEdge->Start = l0;LEdge->End = l3b; }\
		else { LEdge->Start = l3b;LEdge->End = l0; }\
/*printf("%s() %s:%d  T1b %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, LEdge->Start, LEdge->End);fflush(stdout);*/\
		LEdge++;\
	}\
\
	if(l3 > l2) { LEdge->Start = l2;LEdge->End = l3; }\
	else { LEdge->Start = l3;LEdge->End = l2; }\
	if(!LVEdgesVisibility[mVC]) LEdge->Flags &= ~E3dEdgeVISIBLE;\
/*printf("%s() %s:%d  t1  %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, LEdge->Start, LEdge->End);fflush(stdout);*/\
	LEdge++;\
} while(0)


// Edges around the profile Spline's end point, when the Spline's end point touches the revolve axis and the quads degenerate to triangles
//
#define _EdgesT2(mVC)\
do {\
	if(l2 > l0) { LEdge->Start = l0;LEdge->End = l2; }\
	else { LEdge->Start = l2;LEdge->End = l0; }\
/*printf("%s() %s:%d  T2  %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, LEdge->Start, LEdge->End);fflush(stdout);*/\
	if(lUC % LUEdgeVisibilityStep != 0) LEdge->Flags &= ~E3dEdgeVISIBLE;\
	LEdge++;\
\
	if((!lPathClosed) && lUC == lPathLastSegment)\
	{\
		const EIndex  l0b = l0 + lProfilePointCount;\
		if(l2 > l0b) { LEdge->Start = l0b;LEdge->End = l2; }\
		else { LEdge->Start = l2;LEdge->End = l0b; }\
/*printf("%s() %s:%d  T2b %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, LEdge->Start, LEdge->End);fflush(stdout);*/\
		LEdge++;\
	}\
\
	if(l1 > l0) { LEdge->Start = l0;LEdge->End = l1; }\
	else { LEdge->Start = l1;LEdge->End = l0; }\
	if(!LVEdgesVisibility[mVC]) LEdge->Flags &= ~E3dEdgeVISIBLE;\
/*printf("%s() %s:%d  t2  %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, LEdge->Start, LEdge->End);fflush(stdout);*/\
	LEdge++;\
} while(0)


// Edges for quads
//
static inline E3dEdge* _EdgesQ(E3dEdge* PEdge,  const EIndex PUC, const EBool PUClosed, const EIndex PULastSegment, const EIndex PUEdgeVisibilityStep,  const EIndex PVC, const EBool PVClosed, const EIndex PVLastSegment, const EIndex PVPointCount, const EBool* PVEdgesVisibility,  const E3dVertexIndex PV0, const E3dVertexIndex PV1, const E3dVertexIndex PV2, const E3dVertexIndex PV3)
{
// Along V Edge
//
	PEdge->Start = PV0;PEdge->End = PV3;
//printf("%s() %s:%d  Q   %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, PV0, PV3);fflush(stdout);
	if(PUC % PUEdgeVisibilityStep != 0) PEdge->Flags &= ~E3dEdgeVISIBLE;
	PEdge++;

	if((!PUClosed) && PUC == PULastSegment)
	{
		if(PV3 > PV0) { PEdge->Start = PV0 + PVPointCount;PEdge->End = PV3 + PVPointCount; }
		else { PEdge->Start = PV3 + PVPointCount;PEdge->End = PV0 + PVPointCount; }
//printf("%s() %s:%d  Qb  %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, PEdge->Start, PEdge->End);fflush(stdout);
		PEdge++;
	}

// Along U Edge
//
	if(PV1 > PV0) { PEdge->Start = PV0;PEdge->End = PV1; }
	else { PEdge->Start = PV1;PEdge->End = PV0; }
	if(!PVEdgesVisibility[PVC]) PEdge->Flags &= ~E3dEdgeVISIBLE;
//printf("%s() %s:%d  q   %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, PEdge->Start, PEdge->End);fflush(stdout);
	PEdge++;

	if((!PVClosed) && PVC == PVLastSegment)
	{
		if(PV1 > PV0) { PEdge->Start = PV0 + 1;PEdge->End = PV1 + 1; }
		else { PEdge->Start = PV1 + 1;PEdge->End = PV0 + 1; }
		PEdge++;
	}

	return(PEdge);
}


/*
#define _EdgesQ()\
do {\
	LEdge->Start = l0;LEdge->End = l3;\
@*printf("%s() %s:%d  Q   %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, l0, l3);fflush(stdout);@/\
	if(lUC % LUEdgeVisibilityStep != 0) LEdge->Flags &= ~E3dEdgeVISIBLE;\
	LEdge++;\
\
	if((!lPathClosed) && lUC == lPathLastSegment)\
	{\
		if(l3 > l0) { LEdge->Start = l0 + lProfilePointCount;LEdge->End = l3 + lProfilePointCount; }\
		else { LEdge->Start = l3 + lProfilePointCount;LEdge->End = l0 + lProfilePointCount; }\
@*printf("%s() %s:%d  Qb  %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, LEdge->Start, LEdge->End);fflush(stdout);@/\
		LEdge++;\
	}\
\
	if(l1 > l0) { LEdge->Start = l0;LEdge->End = l1; }\
	else { LEdge->Start = l1;LEdge->End = l0; }\
@*printf("%s() %s:%d  q   %2d-%2d\n", __FUNCTION__, __FILE__, __LINE__, LEdge->Start, LEdge->End);fflush(stdout);@/\
	LEdge++;\
} while(0)
*/


#define _M_AddQuad(mV0, mV1, mV2, mV3)\
{\
	const E3dVertexIndex  l0 = LVertexReindexingTable[mV0];\
	const E3dVertexIndex  l1 = LVertexReindexingTable[mV1];\
	const E3dVertexIndex  l2 = LVertexReindexingTable[mV2];\
	const E3dVertexIndex  l3 = LVertexReindexingTable[mV3];\
	if(l0 == l1)\
	{\
		if(l2 != l3)\
		{\
			E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, l0, l2, l3);\
			LPolygon++;\
			_EdgesT1(lVC + 1);\
		}\
	}\
	else if(l2 == l3)\
	{\
		E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, l0, l1, l2);\
		LPolygon++;\
		_EdgesT2(lVC);\
	}\
	else\
	{\
		E3dITriangle*  lITriangle;\
\
		E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, l0, l1, l2, l3);\
		ENewN(2, LPolygon->ITriangles);lITriangle = LPolygon->ITriangles;\
		LPolygon->ITriangles_Count = 2;\
		lITriangle->VertexNodes[0] = 0;lITriangle->VertexNodes[1] = 1;lITriangle->VertexNodes[2] = 2;lITriangle++;\
		lITriangle->VertexNodes[0] = 2;lITriangle->VertexNodes[1] = 3;lITriangle->VertexNodes[2] = 0;\
		LPolygon++;\
		LEdge = _EdgesQ(LEdge,  lUC, lPathClosed, lPathLastSegment, LUEdgeVisibilityStep,  lVC, lProfileClosed, lProfileLastSegment, lProfilePointCount, LVEdgesVisibility,  l0, l1, l2, l3);\
	}\
}


#define _M_Add2TrisUV(mV0, mV1, mV2, mV3)\
{\
	const E3dVertexIndex  l0 = LVertexReindexingTable[mV0];\
	const E3dVertexIndex  l1 = LVertexReindexingTable[mV1];\
	const E3dVertexIndex  l2 = LVertexReindexingTable[mV2];\
	const E3dVertexIndex  l3 = LVertexReindexingTable[mV3];\
	if(l0 == l1)\
	{\
		if(l2 != l3)\
		{\
			E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, l0, l2, l3);\
			E3dVertexNode*  lVertexNode = LPolygon->VertexNodes;\
			E3dCoord2*  lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
			lUV->X = lU;lUV->Y = LVTable[lVC];             EStruct_Inc(lUV, LVertexNodeSize);\
			lUV->X = lU + LUStep;lUV->Y = LVTable[lVC + 1];EStruct_Inc(lUV, LVertexNodeSize);\
			lUV->X = lU;lUV->Y = LVTable[lVC + 1];\
			LPolygon++;\
		}\
	}\
	else if(l2 == l3)\
	{\
		E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, l0, l1, l2);\
		E3dVertexNode*  lVertexNode = LPolygon->VertexNodes;\
		E3dCoord2*  lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
		lUV->X = lU;lUV->Y = LVTable[lVC];         EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU + LUStep;lUV->Y = LVTable[lVC];EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU + LUStep;lUV->Y = LVTable[lVC + 1];\
		LPolygon++;\
	}\
	else\
	{\
		E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LVertexReindexingTable[mV0], LVertexReindexingTable[mV1], LVertexReindexingTable[mV2]);\
		E3dVertexNode*  lVertexNode = LPolygon->VertexNodes;\
		E3dCoord2*  lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
		lUV->X = lU;lUV->Y = LVTable[lVC];         EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU + LUStep;lUV->Y = LVTable[lVC];EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU + LUStep;lUV->Y = LVTable[lVC + 1];\
		LPolygon++;\
		E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, LVertexReindexingTable[mV2], LVertexReindexingTable[mV3], LVertexReindexingTable[mV0]);\
		lVertexNode = LPolygon->VertexNodes;\
		lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
		lUV->X = lU + LUStep;lUV->Y = LVTable[lVC + 1];EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU;lUV->Y = LVTable[lVC + 1];         EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU;lUV->Y = LVTable[lVC];\
		LPolygon++;\
	}\
}


#define _M_AddQuadUV(mV0, mV1, mV2, mV3)\
{\
	const E3dVertexIndex  l0 = LVertexReindexingTable[mV0];\
	const E3dVertexIndex  l1 = LVertexReindexingTable[mV1];\
	const E3dVertexIndex  l2 = LVertexReindexingTable[mV2];\
	const E3dVertexIndex  l3 = LVertexReindexingTable[mV3];\
	if(l0 == l1)\
	{\
		if(l2 != l3)\
		{\
			E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, l0, l2, l3);\
			E3dVertexNode*  lVertexNode = LPolygon->VertexNodes;\
			E3dCoord2*  lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
			lUV->X = lU;lUV->Y = LVTable[lVC];             EStruct_Inc(lUV, LVertexNodeSize);\
			lUV->X = lU + LUStep;lUV->Y = LVTable[lVC + 1];EStruct_Inc(lUV, LVertexNodeSize);\
			lUV->X = lU;lUV->Y = LVTable[lVC + 1];\
			LPolygon++;\
			_EdgesT1(lVC + 1);\
		}\
	}\
	else if(l2 == l3)\
	{\
		E3dPolygon_SetAsTriangle(LPolygon, LVertexNodeSize, l0, l1, l2);\
		E3dVertexNode*  lVertexNode = LPolygon->VertexNodes;\
		E3dCoord2*  lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
		lUV->X = lU;lUV->Y = LVTable[lVC];         EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU + LUStep;lUV->Y = LVTable[lVC];EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU + LUStep;lUV->Y = LVTable[lVC + 1];\
		LPolygon++;\
		_EdgesT2(lVC);\
	}\
	else\
	{\
		E3dITriangle*  lITriangle;\
\
		E3dPolygon_SetVa(LPolygon, LVertexNodeSize, 4, l0, l1, l2, l3);\
		ENewN(2, LPolygon->ITriangles);lITriangle = LPolygon->ITriangles;\
		LPolygon->ITriangles_Count = 2;\
		lITriangle->VertexNodes[0] = 0;lITriangle->VertexNodes[1] = 1;lITriangle->VertexNodes[2] = 2;lITriangle++;\
		lITriangle->VertexNodes[0] = 2;lITriangle->VertexNodes[1] = 3;lITriangle->VertexNodes[2] = 0;\
		E3dVertexNode*  lVertexNode = LPolygon->VertexNodes;\
		E3dCoord2*  lUV = (E3dCoord2*)EOffsetToPtr(lVertexNode, LUVOffsetInVtxNode);\
		lUV->X = lU;lUV->Y = LVTable[lVC];EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU + LUStep;lUV->Y = LVTable[lVC];    EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU + LUStep;lUV->Y = LVTable[lVC + 1];EStruct_Inc(lUV, LVertexNodeSize);\
		lUV->X = lU;lUV->Y = LVTable[lVC + 1];\
		LPolygon++;\
		LEdge = _EdgesQ(LEdge,  lUC, lPathClosed, lPathLastSegment, LUEdgeVisibilityStep,  lVC, lProfileClosed, lProfileLastSegment, lProfilePointCount, LVEdgesVisibility,  l0, l1, l2, l3);\
	}\
}


//========================================
// Update method
//========================================
static E3dObject* _Update(E3dObject* PObject, E3dObjectComponent* PInfoV, const EUpdateFlags PUpdateFlags, EChangeFlags* PTriggeredChangesP, const EBool PCheckOnly)
{
	E3dScene*      LScene = E3d_Scene;
	E3dMesh*       LMesh;
	E3dPolyGroup*  LPolyGroup;
	E3dInfo*       LInfo;

	if(PObject == NULL)
	{
		E_New(E3dMesh, LMesh);
		PObject = (E3dObject*)LMesh;

		LInfo = (E3dInfo*)E3dObjectComponent_Add(PObject, _ObjectComponent_class);
		if(PInfoV && PInfoV->Class == _ObjectComponent_class)
		{
			_Copy(PObject, PInfoV, PObject, (E3dObjectComponent*)LInfo);
		}

		E3dBaseObject_AddOutput((E3dBaseObject*)LInfo->InputGeometry, &LInfo->InputInterface);

		if((LPolyGroup = E3dMesh_PolyGroupAdd(LMesh)) == NULL) { E_Delete(LMesh);return(NULL); }
		LPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;
	}
	else
	{
		if(!EObject_IsSubClass(PObject, E3dMesh)) return(NULL);

		LMesh = (E3dMesh*)PObject;
		LInfo = (E3dInfo*)E3dObjectComponent_ByClass(PObject, _ObjectComponent_class);

		if(LMesh->PolyGroups_Count == 0)
		{
			LPolyGroup = E3dMesh_PolyGroupAdd(LMesh);
		}
		else LPolyGroup = LMesh->PolyGroups[0];

		E3dM_ObjectUpdateDrawingMaterial((E3dObject*)LMesh, LPolyGroup->Material, LPolyGroup,DrawingMaterial);

// When a Scene is loaded, Update methods are called to create this link
// The Geometry will be up-to-date, so if the link wasn't there yet, just exit
// We should probably add a 3rd argument to this function ("Action=LINK"), so
// even input-less ComponentClasses don't rebuild their Geometry.
//
		if(PCheckOnly && LInfo->InputGeometry)
		{
			if(E3dBaseObject_AddOutput((E3dBaseObject*)LInfo->InputGeometry, &LInfo->InputInterface) == E_SUCCESS)
			{
				return(PObject);
			}
		}
	}


	const E3dSpline*  LProfileSpline = (E3dSpline*)(LInfo->InputGeometry);

	if(LProfileSpline == NULL || LProfileSpline->SNumOfLinearSegments < 1) return(NULL);

	const E3dCoordinate  LStartAngle = LInfo->StartAngle;
	const E3dCoordinate  LEndAngle   = LInfo->EndAngle;

	const EBool          LFull = (LEndAngle - LStartAngle) == 360.0 ? TRUE : FALSE;

	LInfo->Full = LFull;

	const int            LVPoints  = E3d_SplineNumOfLinearSegmentPoints(LProfileSpline);
	const int            LSegments = LProfileSpline->SNumOfLinearSegments;

	const int            LSides = LInfo->Sides;

	if(PCheckOnly || LInfo->InputGeometry == NULL) return(NULL);


	const EIndex  LUPoints = LFull ? LSides : LSides + 1;

	EIndex  LNumOfVertices = LUPoints * LVPoints;
	EIndex  LNumOfEdges    = LUPoints * LSegments + LVPoints * LSides;
	EIndex  LNumOfPolygons = LSides * LSegments;

	if(LInfo->Triangles) LNumOfPolygons *= 2;

//printf("LUPoints %d, LVPoints %d, LNumOfVertices %d, LNumOfPolygons %d\n", LUPoints, LVPoints, LNumOfVertices, LNumOfPolygons);fflush(stdout);


	if(LMesh)
	{
#if E3dCREATE_EDGES
		const EBool  LCreateEdges = !LInfo->Triangles;
		LMesh->SaveEdges = TRUE;
#else
		const EBool  LCreateEdges = FALSE;
#endif


		E3dMesh_FreeCaches(LMesh);
		E3dMesh_EdgesFree(LMesh);

		int  LUVOffsetInVtxNode = -1;

		E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_Normal, 0,  NULL);

// Make sure that the VertexNodes have UV
//
		if(LInfo->CreateUVs)
		{
			LUVOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(LPolyGroup,  E3dVtxAttribute_UV, 0);

			if(LUVOffsetInVtxNode < 0)
			{
				E3dPolyGroup_VertexAttributeAdd(LPolyGroup,  NULL, E3dVtxAttribute_UV, 0,  NULL);
				LUVOffsetInVtxNode = E3dPolyGroup_VertexAttributeOffset(LPolyGroup,  E3dVtxAttribute_UV, 0);
			}
		}
		else E3dPolyGroup_VertexAttributeDel(LPolyGroup,  E3dVtxAttribute_UV, 0);


// Allocate Vertices for the Mesh
//
		if(LNumOfVertices != LMesh->Vertices_Count)
		{
			E3dMesh_VerticesFree(LMesh);
			E3dMesh_VerticesAdd(LMesh, LNumOfVertices);
		}

		EIndex  LUEdgeVisibilityStep = 1;

		if(LCreateEdges)
		{
// Allocate Edges for the Mesh
//
			if(LNumOfEdges != LMesh->Edges_Count) E3dMesh_EdgesNew(LMesh, LNumOfEdges, TRUE);

// Find integer divider of LSegments, to determine which V Edges will be visible
//
			if(LSides >= 8)
			{
				LUEdgeVisibilityStep = 1;
				{
					EIndex  LC = LSides / 4;for(;LC > 0;LC--)
					{
						if(LSides % LC == 0) { LUEdgeVisibilityStep = LC;break; }
					}
				}

				if(LUEdgeVisibilityStep == 1)
				{
					if(LSides % 3 == 0) LUEdgeVisibilityStep = 3;
					else if(LSides % 2 == 0) LUEdgeVisibilityStep = 2;
				}
//printf("%s() %s:%d  LUEdgeVisibilityStep %d\n", __FUNCTION__, __FILE__, __LINE__, (int)LUEdgeVisibilityStep);fflush(stdout);
			}
			else LUEdgeVisibilityStep = 1;
		}


// Allocate Polygons for the Mesh
//
		{
			E3dPolygon*  LPolygons;

			if(LPolyGroup->Polygons) E3dPolyGroup_FreePolygons(LPolyGroup);

			LPolygons = E3dPolygons_New(LNumOfPolygons);if(LPolygons)
			{
				LPolyGroup->Polygons = LPolygons;LPolyGroup->Polygons_Count = LNumOfPolygons;
			}
		}


		E3dCoord3*  LVertexTangents = NULL;
		int*        LVertexReindexingTable = _SetVertices(LMesh, LInfo, &LVertexTangents, TRUE);


// Build the Polygons
//
		EIndex  LNumOfVertexNodes;
		if(LInfo->Triangles) LNumOfVertexNodes = LNumOfPolygons * 3;
		else
		{
// E3dFIXME
// We should take into account Vertex welding at the revolve axis
//
			LNumOfVertexNodes = LNumOfPolygons * 4;
		}


		const E3dCoordinate  LUStep =  1.0 / (E3dCoordinate)LSides;
//		const E3dCoordinate  LVStep = -1.0 / (E3dCoordinate)LSegments;


		E3dCoordinate*  LVTable = NULL;


		{
			E3dPolygon*  LPolygon = LPolyGroup->Polygons;
			E3dEdge*     LEdge    = LMesh->Edges;

			{
// Build list of which Edges along the profile Spline should be visible.
// The Edges crossing CVs will be visible. The ones between CVs won't be.
//
				EBool*  LVEdgesVisibility = NULL;
				if(LProfileSpline)
				{
//printf("%s() %s:%d  %lld: ", __FUNCTION__, __FILE__, __LINE__, (long long)(LSegments + 1));fflush(stdout);
					ENewN(LSegments + 1, LVEdgesVisibility);
					EIndex  LCVC = 0, LI = 0;const EIndex  LCVN = LProfileSpline->CVs_Count;const E3dSplineCV*  LCV = LProfileSpline->CVs;for(;LCVC < LCVN;LCVC++, LCV++)
					{
						EIndex  LN = LCV->NumOfLinearSegments - 1;if(LN < 0) LN = 0;
//printf("%s() %s:%d  CV %3d  lin %3d\n", __FUNCTION__, __FILE__, __LINE__, (int)LCVC, (int)LN);fflush(stdout);
//printf(" 1");fflush(stdout);
						LVEdgesVisibility[LI++] = TRUE;
						EIndex  LC = 0;for(;LC < LN;LC++)
						{
//printf(" 0");fflush(stdout);
							LVEdgesVisibility[LI++] = FALSE;
						}
					}
				}

				const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

// Look-up table for V parameter, using arc-length
//
				ENewN(LSegments + 1, LVTable);

				E3dSpline_GetDrawingArcLengthParameters(LProfileSpline, 0.0, TRUE,  LVTable, NULL);

				EBool*  LProfileSegmentVisFlags = NULL;

// If UseProfileSegmentVisibility is enabled, only generate Polygons for visible segments of the profile Spline.
// This means we need to generate an EBool array that indicates which piece-wise linear segments are visible
//
				if(LInfo->UseProfileSegmentVisibility)
				{
					if(LProfileSpline && LProfileSpline->CVs_Count > 1)
					{
						ENewN(LSegments, LProfileSegmentVisFlags);
						EIndex  LCVC = 0, LI = 0;const EIndex  LCVN = LProfileSpline->Closed ? LProfileSpline->CVs_Count : LProfileSpline->CVs_Count - 1;const E3dSplineCV*  LCV = LProfileSpline->CVs;for(;LCVC < LCVN;LCVC++, LCV++)
						{
							const EBool  LSegmentVisible = LCV->SegmentFlags & E3dSegmentVISIBLE ? TRUE : FALSE;
							EIndex  LC = 0;const EIndex  LN = LCV->NumOfLinearSegments;for(;LC < LN;LC++)
							{
								LProfileSegmentVisFlags[LI++] = LSegmentVisible;
							}
						}
					}
				}

				if(LInfo->CreateUVs)
				{
					if(LInfo->Triangles)
					{
						E3d_ToroidBuildPolygons(_M_Add2TrisUV,  0,  LSides, LFull,  LVPoints, LSegments, LProfileSpline->Closed, LProfileSegmentVisFlags);
					}
					else
					{
						E3d_ToroidBuildPolygons(_M_AddQuadUV,   0,  LSides, LFull,  LVPoints, LSegments, LProfileSpline->Closed, LProfileSegmentVisFlags);
					}
				}
				else
				{
					if(LInfo->Triangles)
					{
						E3d_ToroidBuildPolygons(_M_Add2Tris,    0,  LSides, LFull,  LVPoints, LSegments,  LProfileSpline->Closed, LProfileSegmentVisFlags);
					}
					else
					{
						E3d_ToroidBuildPolygons(_M_AddQuad,     0,  LSides, LFull,  LVPoints, LSegments,  LProfileSpline->Closed, LProfileSegmentVisFlags);
					}
				}

				EFreeChk(LVEdgesVisibility);
				EFreeChk(LProfileSegmentVisFlags);
			}

			LNumOfPolygons = LPolygon - LPolyGroup->Polygons;
			if(LNumOfPolygons < LPolyGroup->Polygons_Count)
			{
				EResizeNT(E3dPolygon, LNumOfPolygons,  LPolyGroup->Polygons);

				LPolyGroup->Polygons_Count = LNumOfPolygons;
			}

			LMesh->Edges_Count = LEdge - LMesh->Edges;
		}

		EFree(LVertexReindexingTable);

// Set up normals
//
		LPolyGroup->SmoothingAngle = LInfo->SmoothingAngle;


// Create tangents for anisotropic shaders
//  
		if(LInfo->CreateTangents)
		{
			int  LTangentOffset = E3dPolyGroup_VertexAttributeOffset(LPolyGroup, E3dVtxAttribute_Tangent, 0);

			if(LTangentOffset < 0)
			{
//printf("%s()-01 LPolyGroup->VertexNodeSize %d\n", __FUNCTION__, LPolyGroup->VertexNodeSize);fflush(stdout);
				E3dPolyGroup_VertexAttributeAdd(LPolyGroup, NULL, E3dVtxAttribute_Tangent, 0,  NULL);
//printf("%s()-02 LPolyGroup->VertexNodeSize %d\n", __FUNCTION__, LPolyGroup->VertexNodeSize);fflush(stdout);
				LTangentOffset = E3dPolyGroup_VertexAttributeOffset(LPolyGroup, E3dVtxAttribute_Tangent, 0);
			}

			if(LTangentOffset > -1)
			{
				if(LNumOfVertexNodes)
				{
					const E3dVertexNodeSize  LVertexNodeSize = LPolyGroup->VertexNodeSize;

					E3dPolygon*  LPolygon = LPolyGroup->Polygons;
					EIndex  LPC = 0;for(;LPC < LNumOfPolygons;LPC++, LPolygon++)
					{
						E3dVertexNode*  LVertexNode = LPolygon->VertexNodes;
						E3dCoord3*      LTangent = (E3dCoord3*)EOffsetToPtr(LVertexNode, LTangentOffset);
						EIndex  LVC = 0;const EIndex  LVN = LPolygon->NumOfVertices;for(;LVC < LVN;LVC++, EStruct_Inc(LVertexNode, LVertexNodeSize), EStruct_Inc(LTangent, LVertexNodeSize))
						{
							*LTangent = LVertexTangents[LVertexNode->VertexID];
						}
					}
				}
			}
		}
		else E3dPolyGroup_VertexAttributeDel(LPolyGroup,  E3dVtxAttribute_Tangent, 0);


		EFreeChk(LVertexTangents);

		EFreeChk(LVTable);

		if(LInfo->SmoothingAngle > 0.0)
		{
			LPolyGroup->Flags |= E3dUSE_VERTEX_NORMALS;

			LPolyGroup->VertexNormalType = E3dNormalFROM_SMOOTHING_ANGLE;
		}
		else
		{
			LPolyGroup->VertexNormalType = 0;
			LPolyGroup->Flags &= ~E3dUSE_VERTEX_NORMALS;
		}

		EChangeFlags  LChanged;
		if(LCreateEdges) LChanged = E3dObject_UpdateForDisplay(PObject, PUpdateFlags & ~E3dUF_EDGES, LScene);
		else LChanged = E3dObject_UpdateForDisplay(PObject, PUpdateFlags, LScene);

		if(PTriggeredChangesP) *PTriggeredChangesP |= LChanged;
	}

	return(PObject);
}


static void _CustomObjectChanged(E3dObject* PObject, E3dObjectComponent* PInfo, const EChangeFlags PUpdateFlags, E3dObjectPanel* PPanel)
{
	EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass(PObject);

	E3dObject_CustomUpdate(PObject, PInfo, PUpdateFlags, &LChanged);
	E3dWindows_RedrawByChangeFlags(LChanged);
}


//========================================
// Dialog box callback handlers
//========================================
static void _DCB_Triangles(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->Triangles = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


//========================================
// Dialog box callback handlers
//========================================
static void _DCB_Tangents(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->CreateTangents = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_UVs(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->CreateUVs = ((XeToggleButtonCallbackStruct*)PCallData)->Set;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_SmoothingAngle(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	{
		E3dMesh*       LMesh = (E3dMesh*)LObject;
		E3dPolyGroup*  LPolyGroup = LMesh->PolyGroups[0];

		LPolyGroup->SmoothingAngle = LInfo->SmoothingAngle = ((XeScaleCallbackStruct*)PCallData)->Value.f32;

		_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
	}
}


static void _DCB_StartAngle(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dInfo*                 LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ObjectComponent_class);
	E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	LInfo->StartAngle = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
	if(LInfo->StartAngle > LInfo->EndAngle)
	{
		LInfo->EndAngle = LInfo->StartAngle;
		XeScale_SetValue_Coord(LResourcesGUI->EndAngleScaleW, LInfo->EndAngle, FALSE);
	}
	if(LInfo->StartAngle < (LInfo->EndAngle - 360.0))
	{
		LInfo->EndAngle = LInfo->StartAngle + 360.0;
		XeScale_SetValue_Coord(LResourcesGUI->EndAngleScaleW, LInfo->EndAngle, FALSE);
	}

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_EndAngle(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*               LObject = LPanel->Object;
	E3dInfo*                 LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);
	XeComponentClassDrawer*  LDrawer = XeInfoPanel_GetFolderByClass((XeInfoPanel*)LPanel, _ObjectComponent_class);
	E3dResourcesGUI*         LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

	LInfo->EndAngle = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
	if(LInfo->StartAngle > LInfo->EndAngle)
	{
		LInfo->StartAngle = LInfo->EndAngle;
		XeScale_SetValue_Coord(LResourcesGUI->StartAngleScaleW, LInfo->StartAngle, FALSE);
	}
	if(LInfo->StartAngle < (LInfo->EndAngle-360.0))
	{
		LInfo->StartAngle = LInfo->EndAngle-360.0;
		XeScale_SetValue_Coord(LResourcesGUI->StartAngleScaleW, LInfo->StartAngle, FALSE);
	}

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static void _DCB_Sides(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	LInfo->Sides = ((XeScaleCallbackStruct*)PCallData)->Value.i32;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


//========================================
// AxisSelector callback
//========================================
static void _CB_AxisSelector(ECoreObject* PStruct, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*    LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*         LObject = LPanel->Object;
	E3dInfo*           LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);
	E3dpAxisSelector*  LAxisSelector = (E3dpAxisSelector*)PStruct;

	LInfo->Axis = LAxisSelector->Axis;

	_CustomObjectChanged(LObject, (E3dObjectComponent*)LInfo, E3dUF_ALL, LPanel);
}


static E3dInfo*         _TmpInfo = NULL;
static E3dObjectPanel*  _TmpPanel = NULL;

#define _PickInputGeometryTypeWarningStr  E3dpLng_Revolve_Pick_a_Spline_or_a_Face_as_the_new_profile
#define _PickInputGeometryObjectTypes     E3dItemSPLINE

#include <E3D/PickInput.ci>

static void _RequestProfilePick(void)
{
	const E3dpPrefs*  LPrefs = &E3dp_Prefs;
	const uint32_t    LLabelRGBAp8Color = Xe_ForegroundRGBA8pColor;
	E3dp_PickRequest(
	 XeString_NewVa(TRUE,
	 XeFONT, XeFontDEFAULT,  XeCOLOR_RGBA, LLabelRGBAp8Color,  XeTEXT, _ObjectComponent_class->Name,  XeTEXT, ":    ",
	 E3dp_LMouseButtonS,     XeTEXT, ": pick Spline or Face as the new profile  ",
	 E3dp_RMouseButtonEscS,  XeTEXT, ": cancel",
	 0),
	 E3dItemSPLINE|E3dItemFACE|E3dItemMODEL,  _PickCB_InputGeometry, NULL,  0, FALSE);
}


static void _DCB_ProfileShow(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);
	EChangeFlags     LChanged = 0;

	if(LInfo->InputGeometry)
	{
		if(E3dSel_ObjectSelect(E3d_Scene, (E3dObject*)LInfo->InputGeometry, FALSE, FALSE)) LChanged |= E3dChgSELECTION;

		E3dWindows_RedrawByChangeFlags(LChanged);
		E3dTransformPanel_Update(NULL, E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
		E3dp_PrintMessage(ELogMESSAGE, 3000, "Revolve: selected profile Geometry <%s>", LInfo->InputGeometry->Name);
	}
	else E3dp_PrintMessage(ELogMESSAGE, 3000, "Revolve: no profile Geometry!");
}


static void _DCB_ProfilePick(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dObjectPanel*  LPanel = (E3dObjectPanel*)PClientData;
	E3dObject*       LObject = LPanel->Object;
	E3dInfo*         LInfo = (E3dInfo*)E3dObjectComponent_ByClass(LObject, _ObjectComponent_class);

	_TmpInfo = LInfo;
	_TmpPanel = LPanel;
	_RequestProfilePick();
}


//========================================
// GUI
//========================================
static void _GUI(E3dObject* PObject, E3dObjectComponent* PInfo, void* PPanel, void* PDrawer, const int PAction)
{
	E3dObjectPanel*          LPanel = (E3dObjectPanel*)PPanel;
	XeComponentClassDrawer*  LDrawer = (XeComponentClassDrawer*)PDrawer;
	E3dResourcesGUI*         LResourcesGUI;
	E3dInfo*                 LInfo = (E3dInfo*)PInfo;

	switch(PAction)
	{
		case E_UPDATE:
			if(LDrawer->ResourcesGUI)
			{
				LResourcesGUI = (E3dResourcesGUI*)(LDrawer->ResourcesGUI);

				XeToggleButton_SetState(LResourcesGUI->TrianglesW, LInfo->Triangles, FALSE);
				XeToggleButton_SetState(LResourcesGUI->TangentsW, LInfo->CreateTangents, FALSE);
				XeToggleButton_SetState(LResourcesGUI->UVsW, LInfo->CreateUVs, FALSE);

				XeScale_SetValue_f32(LResourcesGUI->SmoothingAngleW, LInfo->SmoothingAngle, FALSE);

				XeScale_SetValue_Coord(LResourcesGUI->StartAngleScaleW, LInfo->StartAngle, FALSE);
				XeScale_SetValue_Coord(LResourcesGUI->EndAngleScaleW, LInfo->EndAngle, FALSE);

				XeScale_SetValue(LResourcesGUI->SidesScaleW, LInfo->Sides, FALSE);

//printf("%s() %f,%f,%f\n", __FUNCTION__, LInfo->Axis.X, LInfo->Axis.Y, LInfo->Axis.Z);fflush(stdout);
				E3dpAxisSelector_Set(&LResourcesGUI->AxisSelector,  LInfo->Axis.X, LInfo->Axis.Y, LInfo->Axis.Z);
			}
			else
			{
				Widget  LTopMatrixW = LDrawer->Drawer.DrawerW, LW;
				Widget  LIndentWs[7];int  LIndentN;      // For indenting the ScaleWidgets
				Widget  LTraverseWs[7];int  LTraverseN;  // For TextField traversal
				EArg    Args[256];int  ArgCnt;


				ENew(LResourcesGUI);LDrawer->ResourcesGUI = LResourcesGUI;

// Profile
//
				{
					EXeStart;
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNmarginSizeY, 5);
					EXeSetArg(XeNspacingX, 0);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
					EXeSetArg(XeNfontList, Xe_LabelFontList);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNmarginSizeX, 2);EXeSetArg(XeNmarginSizeY, 3);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeWidget_ManagedNew("Profile", XeLabelWidget_class, LMatrixW,  Args, ArgCnt);

					EXeStart;
					EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNspacingLeft, 6);
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);EXeSetArg(XeNactiveColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNlabelType, XeIMAGE);EXeSetArg(XeNimageSizeYPtr, &Xe_IconSize);EXeSetImages2(_Xe_Theme->Image_Find);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Select the Spline that is used as the profile (so you can find it).");
					LW = XeWidget_ManagedNew("Show", XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_ProfileShow, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);EXeSetArg(XeNactiveColorPtr, &Xe_BackgroundRGBAiColor);
					EXeSetArg(XeNmarginSizeX, 0);
					EXeSetArg(XeNshadowThickness, 0);EXeSetArg(XeNlabelType, XeIMAGE);EXeSetArg(XeNimageSizeYPtr, &Xe_IconSize);EXeSetImages2(_Xe_Theme->Image_Replace);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					XeToolTip8("Pick a new profile Spline.");
					LW = XeWidget_ManagedNew("Pick", XePushButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNactivateCallback, _DCB_ProfilePick, (EPointer)LPanel);
				}


				{
					EXeStart;
					EXeSetArg(XeNmarginSizeY, 0);
					EXeSetArg(XeNorientation, XeHORIZONTAL);
					EXeSetArg(XeNspacingX, 12);
					EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
					EXeSetArg(XeNheightAdjustable, FALSE);
					Widget  LMatrixW = XeWidget_ManagedNew("MatrixW", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);


					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->Triangles);
					LResourcesGUI->TrianglesW = LW = XeWidget_ManagedNew("Triangles", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Triangles, (EPointer)LPanel);


					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->CreateTangents);
					XeToolTip8(
"Create geometric tangents along the rotation.\n\
Geometric tangents are used by anisotropic\nshaders (as opposed to texture-space tangents,\n\
which are used for normal-mapping).");
					LResourcesGUI->TangentsW = LW = XeWidget_ManagedNew("Tangents", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_Tangents, (EPointer)LPanel);

					EXeStart;
					EXeSetArg(XeNshadowThickness, 0);
					EXeSetArg(XeNhighlightThickness, 0);
					EXeSetArg(XeNset, LInfo->CreateUVs);
					XeToolTip8("Create 'evaluator coordinates' U and V.\n\
U increases along the circumference, V along the profile spline.\n\
UVs can be used for among many things, UV texture-mapping.");
					LResourcesGUI->UVsW = LW = XeWidget_ManagedNew("UVs", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
					XeAddCallback(LW, XeNvalueChangedCallback, _DCB_UVs, (EPointer)LPanel);
				}

				LTraverseN = 0;
				LIndentN = 0;
				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 6);EXeSetArg(XeNdecimalPoints, 2);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeToolTip8("The normal smoothing angle for the PolyGroup created.");
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SmoothingAngleW = LW = XeWidget_ManagedNew("Smoothing angle:", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
				XeScale_SetValueAndRange_f32(LW,  LInfo->SmoothingAngle,  0.0, 180.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_SmoothingAngle, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 6);EXeSetArg(XeNdecimalPoints, 2);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->StartAngleScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "Start angle",  Args, ArgCnt);
				XeScale_SetValueAndRange_Coord(LW,  LInfo->StartAngle,  -360.0, 360.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_StartAngle, (EPointer)LPanel);

				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 6);EXeSetArg(XeNdecimalPoints, 2);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->EndAngleScaleW = LW = XeScaleNew_Coord(LTopMatrixW, "End angle",  Args, ArgCnt);
				XeScale_SetValueAndRange_Coord(LW,  LInfo->EndAngle,  -360.0, 360.0,  FALSE);
				XeAddCallback(LW, XeNvalueChangedCallback, _DCB_EndAngle, (EPointer)LPanel);


				EXeStart;
				EXeSetArg(XeNtextFieldDigits, 6);
				EXeSetArg(XeNminimum, 3);EXeSetArg(XeNmaximum, 200);
				EXeSetArg(XeNscaleMaximum, 50);
				EXeSetArg(XeNvalue, LInfo->Sides);
				EXeSetArg(XeNorientation, XeHORIZONTAL);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeToolTip8("The number of revolution steps (\"U\" subdivisions).\n\
To set the number of \"V\" subdivisions, edit the Spline's attributes.");
				LIndentWs[LIndentN++] = LTraverseWs[LTraverseN++] = LResourcesGUI->SidesScaleW = LW = XeWidget_ManagedNew("Sides:", XeScaleWidget_class, LTopMatrixW,  Args, ArgCnt);
				XeAddCallback(LW, XeNvalueChangedCallback,_DCB_Sides, (EPointer)LPanel);

				Xe_WidgetsIndentFromLeft(LIndentWs, LIndentN, 2);


				EXeStart;
				EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
				EXeSetArg(XeNforegroundColorPtr, &Xe_ForegroundRGBAiColor);
				EXeSetArg(XeNfontList, Xe_LabelFontList);
				EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
				EXeSetArg(XeNmarginSizeX, 2);EXeSetArg(XeNmarginSizeY, 3);
				EXeSetArg(XeNheightAdjustable, FALSE);
				XeWidget_ManagedNew("Revolve around:", XeLabelWidget_class, LTopMatrixW,  Args, ArgCnt);
				E3dpAxisSelector_New(&LResourcesGUI->AxisSelector, NULL, LTopMatrixW, TRUE,  _CB_AxisSelector, (EPointer)LPanel,  LInfo->Axis.X, LInfo->Axis.Y, LInfo->Axis.Z);


// Set up TextField traversal
//
				Xe_TraversalRing(LTraverseWs, LTraverseN);
			}
		break;
	}
}


//================================================================
// Callback for Profile-Spline changes
// This function should Not call any 3DWindow redrawing function,
// just return with TRUE if a redraw is needed!
//================================================================
static EChangeFlags _InputChanged(EInterface* PInterface, EInterfaceCallbackStruct* PCBS)
{
	E3dObject*    LObject  = (E3dObject*)PInterface->Object;
	E3dInfo*      LInfo    = (E3dInfo*)PInterface->ObjectComponent;
	EChangeFlags  LChanged = E3dObject_GetChangeFlagsByClass(LObject);

// Only respond if the Object is instantiated (e.g. not in the undo stack after being deleted)
//
	if(LObject->Models_Count)
	{
		E3dObject_CustomUpdate(LObject, LInfo, PCBS->CallOutputs ? E3dUF_ALL : E3dUF_ALL - E3dUF_CALL_OUTPUTS, &LChanged);
	}

	return(LChanged);
}


//================================================
// Menu/ToolBar callback
//================================================
static void _MCB(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*   LScene = E3d_Scene;
	E3dObjItem  LItem;

	if(E3dScene_GetSingleSelection(LScene, &LItem, E3dItemSPLINE, E3dVISIBLE) != E3dItemNONE)
	{
		E3dObject*  LInputObject = LItem.Object;
		EBool       LSuccess = FALSE;

		E3dInfo  LTInfo;E3dObjectComponent_InitStatic(&LTInfo, _ObjectComponent_class);
		E_Assign(LInputObject, LTInfo.InputGeometry);

		E3dModel*  LOrigModel = E3dObject_GetSelectingModel(LInputObject, FALSE);
		if(LOrigModel == NULL) LOrigModel = LInputObject->Models[0];

		E3dObject*  LObject = E3dObject_CustomNew(LScene, NULL, LOrigModel, NULL, (E3dObjectComponent*)(&LTInfo), E3dUF_ALL, _NewObjectAddingMode, (E3dp_Prefs.Duplicate_Flags & E3dCopyMODEL_UNIQUE_NAME) != 0,  NULL);if(LObject)
		{
			((E3dGeometry*)LObject)->AutoLOD = FALSE;

			LSuccess = TRUE;
		  E3dScene_ObjectsUpdateModelList(LScene);
			E3dScene_LayoutOnSchematics(LScene, TRUE);
		}

		E_Delete(LTInfo.InputGeometry);

		if(LSuccess)
		{
			if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) E3dObject_Edit(LObject, _ObjectComponent_class);

			E3dWindows_RedrawByChangeFlags(E3dChgMESHSHAPE|E3dChgHIERARCHY);
		}
	}
	else E3dp_PrintMessage(ELogWARNING, 5000, "Select a Spline to revolve");
}


//================================================
// Constructor
//================================================
static int _Init(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dInfo*  LThis = (E3dInfo*)PInfo;

	LThis->InputGeometry = NULL;
	LThis->StartAngle = 0.0;LThis->EndAngle = 360.0;
	LThis->Sides = 24;
	E_SetXYZ_s(LThis->Axis,  0.0, 1.0, 0.0);		// The Y axis
	LThis->AroundWorldAxis = TRUE;
	LThis->Close = FALSE;
	LThis->Triangles = FALSE;
	LThis->CreateTangents = TRUE;
	LThis->UseProfileSegmentVisibility = FALSE;
	LThis->CreateUVs = TRUE;
	LThis->SmoothingAngle = 60.0;

	{
		EInterface*  LInput = &LThis->InputInterface;E_InterfaceInit(LInput);
		LInput->InputChanged    = _InputChanged;
		LInput->Object          = (ECoreObject*)PObject;
		LInput->ObjectComponent = (ECoreObject*)PInfo;
	}

	LThis->Full = FALSE;

	LThis->GLAxis[0][E3dX] = LThis->GLAxis[0][E3dY] = LThis->GLAxis[0][E3dZ] = 0.0;
	LThis->GLAxis[1][E3dX] = LThis->GLAxis[1][E3dY] = LThis->GLAxis[1][E3dZ] = 0.0;
	LThis->GLAxisUptodate = FALSE;
	return(E_SUCCESS);
}


//========================================
// Copy-constructor
//========================================
static void _Copy(const E3dObject* PSObject, const E3dObjectComponent* PSInfo,  E3dObject* PDObject, E3dObjectComponent* PDInfo)
{
	const E3dInfo*  LSInfo = (E3dInfo*)PSInfo;
	E3dInfo*        LDInfo = (E3dInfo*)PDInfo;

	_Init(PDObject, PDInfo);

#define _CopyField(f)  LDInfo->f = LSInfo->f
	_CopyField(StartAngle);
	_CopyField(EndAngle);
	_CopyField(Sides);
	_CopyField(Axis);
	_CopyField(AroundWorldAxis);
	_CopyField(Close);
	_CopyField(Triangles);
	_CopyField(CreateTangents);
	_CopyField(CreateUVs);
	_CopyField(SmoothingAngle);
	_CopyField(GLAxis[0][E3dX]);_CopyField(GLAxis[0][E3dY]);_CopyField(GLAxis[0][E3dZ]);
	_CopyField(GLAxis[1][E3dX]);_CopyField(GLAxis[1][E3dY]);_CopyField(GLAxis[1][E3dZ]);
	_CopyField(GLAxisUptodate);
#undef _CopyField

	E_AssignToMember(LSInfo->InputGeometry, LDInfo,InputGeometry);
	if(LSInfo->InputGeometry)
	{
		E3dBaseObject_AddOutput((E3dBaseObject*)LDInfo->InputGeometry, &LDInfo->InputInterface);
	}
}


//================================================
// Destructor
//================================================
static int _Free(E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dInfo*  LInfo = (E3dInfo*)PInfo;

	E3dBaseObject_RemoveOutput((E3dBaseObject*)LInfo->InputGeometry, &LInfo->InputInterface);

	E_DeleteChk(LInfo->InputGeometry);

	E_InterfaceClear(&LInfo->InputInterface);

	return(E_SUCCESS);
}


//========================================
// Draw revolution axis
//========================================
static E3dHWDrawFlags _GLDrawControls(void* P3DWin, const E3dDrawContext* PDrawContext, const E3dDisplayFlags PDisplayFlags,  E3dModel* PModel, E3dObject* PObject, E3dObjectComponent* PInfo)
{
	E3dInfo*     LInfo = (E3dInfo*)PInfo;
	const float  LAAMultF = (float)sqrt(E3dp_Prefs.AntiAliasing);

	if(LInfo && PModel->Selection)
	{
#ifdef EGLES
if(E3d_Verbosity & E3dVbOPENGL) { printf("%s() %s:%d  %sGLES IMPLEMENT ME%s\n", __FUNCTION__, __FILE__, __LINE__, EYellowCode, ENormalCode);fflush(stdout); }
#else
		E3glDisable(GL_LIGHTING);
#endif

		E3dGL_TexturesDisable();

		if(!LInfo->GLAxisUptodate) _SetVertices((E3dMesh*)PObject, LInfo, NULL, FALSE);

		EcM_glColorRGBAf_s(E3dp_NormalDrawContext.NormalRGBAfColor);

		E3glLineWidth(E3dp_Prefs.LineWidth * 2.0 * LAAMultF);

		E3glBegin(GL_LINES);
		 E3dGL_Vertex3v(LInfo->GLAxis[0]);E3dGL_Vertex3v(LInfo->GLAxis[1]);
		E3glEnd();

		E3glLineWidth(E3dp_Prefs.LineWidth * LAAMultF);
	}

	return(0);
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	_Plugin = PPlugin;

	{
		E3dObjectComponentClass  LObjectComponentClass =
		{
			.Name=            "Revolve",
			.StructSize=      sizeof(E3dInfo),
			.Resources=       _Resources,

			.Init=            _Init,
			.Copy=            _Copy,
			.Free=            _Free,
			.Update=          _Update,
			.GUI=             _GUI,
			.GLDrawControls=  _GLDrawControls
		};
		_ObjectComponent_class = E3dObjectComponent_ClassRegister(&LObjectComponentClass);
	}

	if(_ObjectComponent_class)
	{
	 EpM_CreateIconStates3_sp(Revolve);E_Assign(_Image_Revolve, _ObjectComponent_class->Icon);
	 Xe_PushButtonAddImg3(_ToolBarButton, Revolve, "Tool->Surface", "Revolve", '\0', NULL, NULL, TRUE, "Revolve a Spline around the selected axis to create a surface.\nRight button: dialog box for this tool", _MCB, (EPointer)0);

	 _MenuButton = XePushButton_Add("Menu->Surface", "Revolve", '\0', NULL, NULL, TRUE, NULL, _MCB, (EPointer)0);
	}
	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		XeItemFree(_MenuButton);
		XeItemFree(_ToolBarButton);
	}

	if(_ObjectComponent_class) { E3dObjectComponent_ClassDeactivate(_ObjectComponent_class);_ObjectComponent_class = NULL; }

	EpIconFree3(Revolve);

	return(E_SUCCESS);
}
