//================================================================
// EQUINOX-3D plugin
//
// - Reset Transforms
// - Freeze Transforms
// - Freeze Scaling
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 17:39:42
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#include <stdio.h>

#include <ELists.h>
#include <EMalloc.h>
#include <EPlugins.h>

#include <Xe/Matrix.h>
#include <Xe/Scale.h>
#include <Xe/Shell.h>
#include <Xe/PushButton.h>
#include <Xe/ToggleButton.h>

#include <Xe/Dialog.h>

#include <E3D/E3D.h>
#include <E3D/Matrix.h>

#include <E3D/UI/Panel.h>
#include <E3D/UI/StatusPanel.h>  // For E3dp_PrintMessage

#include <E3D/TagOps.h>


char*  E_PluginInfo = " - Reset Transforms\n - Freeze Transforms\n - Freeze Scaling\n - Origin to center";


static EguiItem  _MenuButtons[16];static int  _MenuButtons_Count = 0;
static EguiItem  _ToolBarButtons[16];static int  _ToolBarButtons_Count = 0;


static EPlugin*         _Plugin = NULL;


static EOperatorClass*  _OperatorClass = NULL;


//========================================
// Freeze transforms on selection
//========================================
static void _CB_FreezeToWorld(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	E3dNode**     LRootNodes = LScene->RootNodes;
	EChangeFlags  LChanged = 0;

	EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*  LRootNode = LRootNodes[LC];
		E3dNode*  LNode = LRootNode;
		EBool     LHrcChanged = FALSE;

		do
		{
			if(LNode->Selection != E3dNodeSelNONE)
			{
				E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
				{
					E3dObject**  LObjects = LModel->Objects;
					E3dMatrix    LMatrix;E3dMatrix_Copy(LNode->LocalToWorldMatrix, LMatrix);

					EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
					{
						E3dObject*  LObject = LObjects[LObjC];

						if(EObject_IsSubClass(LObject, E3dGeometry))
						{
							E3dGeometry*  LGeometry = (E3dGeometry*)LObject;

							E3dGeometry_Transform(LGeometry, LMatrix);
							LHrcChanged = TRUE;
							LChanged |= E3dObject_GetChangeFlagsByClass(LObject)|E3dChgTRANSFORM;
						}
					}

					E3dModel_ResetTransforms((E3dModel*)LNode);
				}
			}

			LNode = LNode->Next;
		} while(LNode);

		if(LHrcChanged)
		{
			E3dNodeHrc_UpdateForDisplay(LRootNode, E3duf_SHAPE);
		}
	}

	E3dWindows_RedrawByChangeFlags(LChanged);
	E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  0);
}


//========================================
// Freeze transforms on selections
// - Branch freeze to parent of branch
//========================================
static void _CB_FreezeBranchToItsParent(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	E3dNode**     LRootNodes = LScene->RootNodes;
	EIndex        LBranchesFroze = 0;
	EChangeFlags  LChanged = 0;

	EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*      LRootNode;
		EChangeFlags  LChgF = 0;
		EBool         LHrcChanged = FALSE;
		E3dNode*      LNode = LRootNode = LRootNodes[LC];

		for(;LNode;)
		{
// Take care of this branch in one sitting
//
			if(LNode->Selection == E3dNodeSelBRANCH_ROOT)
			{
				E3dNode*   LBranchRootNode = LNode;
				E3dNode*   LNodeNextAfterBranch = E3dNode_BranchGetNodeAfter(LBranchRootNode);
				E3dMatrix  LBranchParentWorldToLocalMatrix;

				if(LRootNode->Parent)
				{
					E3dMatrix_Invert3x4(LRootNode->Parent->LocalToWorldMatrix, LBranchParentWorldToLocalMatrix);
				}
				else E3dMatrix_LoadIdentity(LBranchParentWorldToLocalMatrix);


//printf("%s() %s:%d [%s] [%s]\n", __FUNCTION__, __FILE__, __LINE__, LNode->Name, LNodeNextAfterBranch ? LNodeNextAfterBranch->Name : "-");fflush(stdout);

// E3dNodeHrc_UpdateConnections can handle branches (the branch root doesn't have to be a hierarchy root).
// We just have to call it on the hierarchy root once we're done.
//
				E3dNodeHrc_UpdateConnections(LBranchRootNode);
				E3dNodeHrc_UpdateMatrices(LBranchRootNode, FALSE);

				for(;LNode;LNode = LNode->Next)
				{
					E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);

//					E3dMatrix  LMatrix;E3dMatrix_Mult4x4_3(LBranchParentWorldToLocalMatrix, LNode->LocalToWorldMatrix, LMatrix);
					E3dMatrix  LMatrix;E3dMatrix_Copy(LNode->LocalToWorldMatrix, LMatrix);

					if(LModel)
					{
						E3dObject**  LObjects = LModel->Objects;

						EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
						{
							E3dObject*  LObject = LObjects[LObjC];

							if(EObject_IsSubClass(LObject, E3dGeometry))
							{
								E3dGeometry*  LGeometry = (E3dGeometry*)LObject;

								E3dGeometry_Transform(LGeometry, LMatrix);
								LHrcChanged = TRUE;
								LChgF |= E3dObject_GetChangeFlagsByClass(LObject)|E3dChgTRANSFORM;
								LChanged |= LChgF;
							}
						}

						E3dModel_ResetTransforms(LModel);
					}
				}
				LNode = LNodeNextAfterBranch;

// We must restore connections in the hierarchy, so if there are other selected branches, they will work right
//
//printf("%s() %s:%d %s{%s\n", __FUNCTION__, __FILE__, __LINE__, EGreenCode, ENormalCode);fflush(stdout);
				E3dNodeHrc_UpdateConnections(LRootNode);


				LBranchesFroze++;
			}
			else LNode = LNode->Next;
		}

		if(LHrcChanged)
		{
			E3dNodeHrc_UpdateMatrices(LRootNode, TRUE);
			E3dNodeHrc_CallNodeInfoUpdateProcs(LRootNode, LChgF, FALSE);
			E3dNodeHrc_UpdateForDisplay(LRootNode, E3duf_SHAPE);
		}
	}

	if(LBranchesFroze == 0) E3dp_PrintMessage(ELogWARNING, 5000, "No branches selected\n");
	else
	{
		if(LBranchesFroze == 1) E3dp_PrintMessage(ELogMESSAGE, 5000, "Froze: 1 branch");
		else E3dp_PrintMessage(ELogMESSAGE, 5000, "Froze: %d branches", LBranchesFroze);
	}

	E3dWindows_RedrawByChangeFlags(LChanged);
	E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  0);
}


//========================================
// Freeze transforms on selections
// - Branch freeze to root of branch
//========================================
static void _CB_FreezeBranchToItsRoot(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	E3dNode**     LRootNodes = LScene->RootNodes;
	EChangeFlags  LChanged = 0;


	EIndex  LBranchesFroze = 0;
	EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*  LRootNode = LRootNodes[LC];
		E3dNode*  LNode = LRootNode;
		EChangeFlags	LChgF = 0;

		do
		{
//printf("In\n");fflush(stdout);
// Take care of this branch in one sitting
//
			if(LNode->Selection == E3dNodeSelBRANCH_ROOT)
			{
				if(LNode->Child)
				{
					E3dNode*  LBranchRootNode = LNode;
					E3dNode*  LNodeNextAfterBranch;

					if((LNodeNextAfterBranch = E3dNode_BranchGetLastNode(LBranchRootNode)) != NULL) LNodeNextAfterBranch = LNodeNextAfterBranch->Next;
					else LNodeNextAfterBranch = NULL;

					E3dTransformSet  LTransformSetBackup;

					E3dModel*  LBranchRootModel = E_DynamicCast(LBranchRootNode, E3dModel);if(LBranchRootModel)
					{
						memcpy(&LTransformSetBackup, &LBranchRootModel->TransformSet, sizeof(E3dTransformSet));  // Back up transforms of branch root

						EList_Init(LBranchRootModel->TransformSet.Transforms);

						E3dTransformSet_Reset(&LBranchRootModel->TransformSet, FALSE);
					}

					E3dNodeHrc_UpdateMatrices(LBranchRootNode, FALSE);


					LNode = LNode->Child;
					do
					{
						E3dMatrix  LMatrix;E3dMatrix_Copy(LNode->LocalToWorldMatrix, LMatrix);

						E3dModel*  LModel = E_DynamicCast(LNode, E3dModel);if(LModel)
						{
							E3dObject**  LObjects = LModel->Objects;
							EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
							{
								E3dObject*  LObject = LObjects[LObjC];

								if(EObject_IsSubClass(LObject, E3dGeometry))
								{
									E3dGeometry*  LGeometry = (E3dGeometry*)LObject;

									E3dGeometry_Transform(LGeometry, LMatrix);
									E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);
									LChgF |= E3dObject_GetChangeFlagsByClass(LObject)|E3dChgTRANSFORM;
									LChanged |= LChgF;
								}
							}

							E3dModel_ResetTransforms(LModel);
						}

						if((LNode = LNode->Next) == NULL) break;
					} while(LNode != LNodeNextAfterBranch);
					LBranchesFroze++;

					if(LBranchRootModel) { memcpy(&LBranchRootModel->TransformSet, &LTransformSetBackup, sizeof(E3dTransformSet)); } // Back up transforms of branch root

					LNode = LNodeNextAfterBranch;
				}
				else LNode = LNode->Next;
			}
			else LNode = LNode->Next;
		} while(LNode);

		if(LChgF)
		{
			E3dNodeHrc_UpdateConnections(LRootNode);
			E3dNodeHrc_UpdateMatrices(LRootNode, TRUE);
			E3dNodeHrc_CallNodeInfoUpdateProcs(LRootNode, LChgF, FALSE);
			E3dNodeHrc_UpdateForDisplay(LRootNode, LChgF);
		}
	}


	if(LBranchesFroze == 0) E3dp_PrintMessage(ELogWARNING, 5000, "No branches selected\n");
	else
	{
		if(LBranchesFroze == 1) E3dp_PrintMessage(ELogMESSAGE, 5000, "Froze: 1 branch");
		else E3dp_PrintMessage(ELogMESSAGE, 5000, "Froze: %d branches", LBranchesFroze);
	}

	E3dWindows_RedrawByChangeFlags(LChanged);
	E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
}


//========================================
// Freeze rotation on selection
//========================================
static void _CB_FreezeRotation(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	E3dNode**     LRootNodes = LScene->RootNodes;
	EIndex        LTotalNumOfGeometriesChanged = 0;
	EChangeFlags  LChanged = 0;


	EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*  LRootNode = LRootNodes[LC];
		EIndex    LHrcNumOfGeometriesChanged = 0;

		E3dNode*  LNode = LRootNode;
		for(;LNode;)
		{
			if(LNode->Selection == E3dNodeSelBRANCH_ROOT)
			{
				LHrcNumOfGeometriesChanged += E3dNodeHrc_FreezeRotation(LNode);

				LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNode) | E3dChgTRANSFORM;

				LNode = E3dNode_BranchGetNodeAfter(LNode);
			}
			else LNode = LNode->Next;
		}

	 LTotalNumOfGeometriesChanged += LHrcNumOfGeometriesChanged;
	 if(LHrcNumOfGeometriesChanged) E3dNodeHrc_UpdateForDisplay(LRootNode, E3duf_SHAPE);



/*
	 if(LRootNode->Selection == E3dNodeSelBRANCH_ROOT)
	 {
		 LNumGeometriesChanged += E3dNodeHrc_FreezeRotation(LRootNode);

		 LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LRootNode) | E3dChgTRANSFORM;

		 if(LNumGeometriesChanged) E3dNodeHrc_UpdateForDisplay(LRootNode, E3duf_SHAPE);

		 LTotalNumOfGeometriesChanged += LNumGeometriesChanged;
	 }
*/
	}

	E3dWindows_RedrawByChangeFlags(LChanged);
	E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);

	switch(LTotalNumOfGeometriesChanged)
	{
		case 0:   break;
		case 1:   E3dp_PrintMessage(ELogMESSAGE, 5000, "Froze rotation and scaling on 1 Geometry");break;
		default:  E3dp_PrintMessage(ELogMESSAGE, 5000, "Froze rotation and scaling on %d Geometries", LTotalNumOfGeometriesChanged);break;
	}
}


//========================================
// Freeze scaling on selection
//========================================
static void _CB_FreezeScaling(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*     LScene = E3d_Scene;
	E3dNode**     LRootNodes = LScene->RootNodes;
	EIndex        LTotalNumOfGeometriesChanged = 0;
	EChangeFlags  LChanged = 0;


	EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*  LRootNode = LRootNodes[LC];
		EIndex    LHrcNumOfGeometriesChanged = 0;

		E3dNode*  LNode = LRootNode;for(;LNode;)
		{
			if(LNode->Selection == E3dNodeSelBRANCH_ROOT)
			{
//printf("%s() %s:%d\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);
				LHrcNumOfGeometriesChanged += E3dNodeHrc_FreezeScaling(LNode);

				LChanged |= E3dNodeHrc_GetTransformChangeFlagsByObjects(LNode) | E3dChgTRANSFORM;

				LNode = E3dNode_BranchGetNodeAfter(LNode);
			}
			else LNode = LNode->Next;
		}

		LTotalNumOfGeometriesChanged += LHrcNumOfGeometriesChanged;
		if(LHrcNumOfGeometriesChanged) E3dNodeHrc_UpdateForDisplay(LRootNode, E3duf_SHAPE);
	}

	E3dWindows_RedrawByChangeFlags(LChanged);
	E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);

	switch(LTotalNumOfGeometriesChanged)
	{
		case 0:   E3dp_PrintMessage(ELogMESSAGE, 5000, "Nothing changed");break;
		case 1:   E3dp_PrintMessage(ELogMESSAGE, 5000, "Froze scaling on 1 Geometry");break;
		default:  E3dp_PrintMessage(ELogMESSAGE, 5000, "Froze scaling on %d Geometries", LTotalNumOfGeometriesChanged);break;
	}
}


enum
{
	CR_SET_ORIGIN_AT_CENTER = XeCR_CUSTOM0,
	CR_SET_ORIGIN_AT_BOTTOM,
	CR_SET_ORIGIN_AT,
	CR_Y_CENTER
};

static Widget         _DialogW = NULL;
static EPluginLock*   _DialogLock = NULL;


static Widget         _YOriginAtCenterBW = NULL, _YOriginAtBottomBW = NULL, _YOriginAtBW = NULL;

static E3dCoordinate  _YCenter = 0.0;
static short          _YCentering = E3dCENTER;


//========================================================================
// "Move the Node's local origin to the centroid of its Geometries
//========================================================================
static void _OriginToCenter(void)
{
	E3dScene*     LScene = E3d_Scene;
	E3dNode**     LRootNodes = LScene->RootNodes;
	E3dNode*      LRootNode;
	EChangeFlags  LChanged = 0;


	EIndex  LC = 0;const EIndex  LN = LScene->RootNodes_Count;for(;LC < LN;LC++)
	{
		E3dNode*      LNode = LRootNode = LRootNodes[LC];
		EChangeFlags  LChgF = 0;


		if(LNode->Selection != E3dNodeSelNONE)
		{
			E3dModel*	LModel = E_DynamicCast(LNode, E3dModel);

//printf("%s()-00 [%s] {%s}\n", __FUNCTION__, LNode->Name, LNode->Class->Name);fflush(stdout);
			if(LModel)
			{
				E3dMatrix      LMatrix;
				E3dCoord3      LBBMin, LBBMax;
				E3dCoordinate  LXOffset, LYOffset, LZOffset;


//printf("%s()-01 [%s]\n", __FUNCTION__, LNode->Name);fflush(stdout);
				const EBool  LNodeSelected = E3dNode_IsSelected(LNode);

				E3dModel_GetBoundingBox(LModel, &LBBMin, &LBBMax, E3dBB_SELECTED_GEOMETRY);

// Get center of bounding box of all selected Geometries on the Model
//
				LXOffset = (LBBMax.X - LBBMin.X) * 0.5 + LBBMin.X;


				switch(_YCentering)
				{
					default:      LYOffset = (LBBMax.Y - LBBMin.Y) * 0.5 + LBBMin.Y;break;
					case E3dMIN:  LYOffset = LBBMin.Y;break;
					case E3dVAL:  LYOffset = _YCenter;break;
				}

				LZOffset = (LBBMax.Z - LBBMin.Z) * 0.5 + LBBMin.Z;

				E3dMatrix_Copy(LModel->LocalToWorldMatrix, LMatrix);
				LMatrix[M30] = 0.0;
				LMatrix[M31] = 0.0;
				LMatrix[M32] = 0.0;
				E3dMatrix_Translate(LMatrix, LXOffset, LYOffset, LZOffset);
				LModel->TransformSet.Translation.X += LMatrix[M30];
				LModel->TransformSet.Translation.Y += LMatrix[M31];
				LModel->TransformSet.Translation.Z += LMatrix[M32];

				E3dMatrix_LoadIdentity(LMatrix);
				do
				{
					E3dObject**  LObjects = LModel->Objects;
					EIndex  LObjC = 0;const EIndex  LObjN = LModel->Objects_Count;for(;LObjC < LObjN;LObjC++)
					{
						E3dObject*  LObject = LObjects[LObjC];

						if(EObject_IsSubClass(LObject, E3dGeometry))
						{
							E3dGeometry*  LGeometry = (E3dGeometry*)LObject;


							if(LNodeSelected || ((LNode->Selection == E3dNodeSelOBJECT) && (LGeometry->Selection == E3dObjectSelOBJECT)))
							{
								E3dGeometry_GetBoundingBox(LGeometry, E3dBB_ALL,  &LBBMin, &LBBMax);
//printf("[%s] BB: %f,%f,%f-%f,%f,%f\n", LGeometry->Name, LBBMin.X, LBBMin.Y, LBBMin.Z, LBBMax.X, LBBMax.Y, LBBMax.Z);fflush(stdout);
								LMatrix[M30] = -LXOffset;
								LMatrix[M31] = -LYOffset;
								LMatrix[M32] = -LZOffset;
								E3dGeometry_Transform(LGeometry, LMatrix);

								LChgF |= E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);
								LChgF |= E3dObject_GetChangeFlagsByClass(LObject) | E3dChgTRANSFORM;
								LChanged |= LChgF;
							}
						}
					}
					LNode = LNode->Next;
				} while(LNode);
			}
		}

		if(LChgF)
		{
			E3dNodeHrc_UpdateMatrices(LRootNode, TRUE);
			E3dNodeHrc_CallNodeInfoUpdateProcs(LRootNode, LChgF, FALSE);
		}
	}

	if(LChanged)
	{
		E3dWindows_RedrawByChangeFlags(LChanged);
		E3dTransformPanel_Update(NULL,  E3dDoALL, E3dDoALL, E3dDoALL,  -1,  E3dTrPanelALL);
	}
}


//========================================
// Dialog box callback
//========================================
static void _DCB_OriginToCenterClose(Widget PW, EPointer PClientData, EPointer PCallData)
{
	Xe_WidgetUnmap(_DialogW);
	E_PluginLockDel(_Plugin, &_DialogLock, FALSE);
}


//========================================
// Dialog box callback
//========================================
static void _DCB_OriginToCenter(Widget PW, EPointer PClientData, EPointer PCallData)
{
	switch((EPointerInt)PClientData)
	{
		case XeCR_OK:
			_OriginToCenter();
			Xe_WidgetUnmap(_DialogW);
			E_PluginLockDel(_Plugin, &_DialogLock, FALSE);
		break;

		case CR_SET_ORIGIN_AT_CENTER:
			if(((XeToggleButtonCallbackStruct*)PCallData)->Set) _YCentering = E3dCENTER;
		break;

		case CR_SET_ORIGIN_AT_BOTTOM:
			if(((XeToggleButtonCallbackStruct*)PCallData)->Set) _YCentering = E3dMIN;
		break;

		case CR_SET_ORIGIN_AT:
			if(((XeToggleButtonCallbackStruct*)PCallData)->Set) _YCentering = E3dVAL;
		break;

		case CR_Y_CENTER:
			_YCenter = ((XeScaleCallbackStruct*)PCallData)->Value.m_Coordinate;
			_YCentering = E3dVAL;
			XeToggleButton_SetState(_YOriginAtCenterBW, FALSE, FALSE);
			XeToggleButton_SetState(_YOriginAtBottomBW, FALSE, FALSE);
			XeToggleButton_SetState(_YOriginAtBW, TRUE, FALSE);
		break;
	}
}


static int _DialogLockDel(EPlugin* PPlugin, EPluginLock* PLock)
{
	Xe_WidgetUnmap(_DialogW);

	return(E_SUCCESS);
}


//========================================
// Popup settings dialog box
//========================================
static void _OriginToCenterPopupDialog(void)
{
	if(_DialogLock == NULL) _DialogLock = E_PluginLockAdd(_Plugin, _DialogLockDel);

	if(_DialogW == NULL)
	{
		EArg  Args[256];int  ArgCnt;

		EXeStart;
		EXeSetArg(XeNallowShellResize, TRUE);
		EXeSetArg(XeNbuttonsStretch, FALSE);
		EXeSetArg(XeNworkAreaType, XeMATRIX);
		EXeSetArg(XeNwmDecorations, XeWMDecorCLOSE|XeWMDecorBORDER|XeWMDecorRESIZEH|XeWMDecorTITLE);
		EXeSetArg(XeNtitle, "Origin to center v0.4.0");EXeSetArg(XeNiconName, "Origin to ctr");
		Widget  LDialogW = XeDialog_New(XeDIALOG_TEMPLATE, "Origin to ctr", E3dp_TopShellW,  Args, ArgCnt);if(LDialogW == NULL) return;
		XeAddCallback(LDialogW, XeNwmCallback, _DCB_OriginToCenterClose, (EPointer)NULL);

		_DialogW = LDialogW;

		Widget  LMessageBoxW = XeDialog_GetChild(LDialogW, XeDIALOG_MAIN);
		EXeStart;
		EXeSetArg(XeNorientation, XeVERTICAL);
		EXeSetArg(XeNworkAreaMarginSizeX, 6);EXeSetArg(XeNworkAreaMarginSizeY, 6);
		Xe_SetValues(LMessageBoxW, Args, ArgCnt);
		{ Widget  LW = XeDialog_AddButton("OK",    XePushButton_class,   LMessageBoxW, TRUE);XeAddCallback(LW, XeNactivateCallback, _DCB_OriginToCenter, (EPointer)XeCR_OK); }

		Widget  LTopMatrixW = XeDialog_GetChild(LDialogW, XeDIALOG_WORK_AREA);
		EXeStart;
		EXeSetArg(XeNorientation, XeVERTICAL);
		EXeSetArg(XeNspacingY, 4);
		Xe_SetValues(LTopMatrixW, Args, ArgCnt);

		EXeStart;
		EXeSetArg(XeNradioBehavior, TRUE);
		EXeSetArg(XeNorientation, XeVERTICAL);
		EXeSetArg(XeNspacingY, 2);
		EXeSetArg(XeNmarginSizeX, 0);EXeSetArg(XeNmarginSizeY, 0);
		EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
		Widget  LMatrixW = XeWidget_ManagedNew("", XeMatrixWidget_class, LTopMatrixW,  Args, ArgCnt);

		Widget  LW;

		EXeStart;
		EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
		EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNfontList, Xe_LabelFontList);
		EXeSetArg(XeNset, (EBool)(_YCentering  == E3dCENTER));
		EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
		LW = _YOriginAtCenterBW = XeWidget_ManagedNew("Y origin at Y center", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_OriginToCenter, (EPointer)CR_SET_ORIGIN_AT_CENTER);

		EXeStart;
		EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
		EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNfontList, Xe_LabelFontList);
		EXeSetArg(XeNset, (EBool)(_YCentering == E3dMIN));
		EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
		LW = _YOriginAtBottomBW = XeWidget_ManagedNew("Y origin at bottom", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_OriginToCenter, (EPointer)CR_SET_ORIGIN_AT_BOTTOM);

		EXeStart;
		EXeSetArg(XeNbackgroundColorPtr, &Xe_BackgroundRGBAiColor);
		EXeSetArg(XeNselectColorPtr, &Xe_HighlightRGBAiColor);
		EXeSetArg(XeNshadowThickness, 0);
		EXeSetArg(XeNhighlightThickness, 0);
		EXeSetArg(XeNfontList, Xe_LabelFontList);
		EXeSetArg(XeNset, (EBool)(_YCentering == E3dVAL));
		EXeSetArg(XeNwidthAdjustable, FALSE);EXeSetArg(XeNalignmentX, XeAlignBEGINNING);
		LW = _YOriginAtBW = XeWidget_ManagedNew("Y origin at:", XeToggleButton_class, LMatrixW,  Args, ArgCnt);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_OriginToCenter, (EPointer)CR_SET_ORIGIN_AT);


		EXeStart;
		EXeSetArg(XeNscaleLengthInPixels, 120);
		EXeSetArg(XeNtextFieldDigits, 8);EXeSetArg(XeNdecimalPoints, 5);
		EXeSetArg(XeNorientation, XeHORIZONTAL);
		LW = XeScaleNew_Coord(LTopMatrixW, "", Args, ArgCnt);
		XeScale_SetValueAndRanges_Coord(LW,  _YCenter,  -1000.0, 1000.0,  -10.0, 10.0,  FALSE);
		XeAddCallback(LW, XeNvalueChangedCallback, _DCB_OriginToCenter, (EPointer)CR_Y_CENTER);

		XeShell_Raise(_DialogW);
		XeShell_CenterToWidget(_DialogW, E3dp_TopShellW);
	}
	else XeShell_Raise(_DialogW);
}


//========================================================================
// "Move the Model's local origin to the centroid of its Geometries
//========================================================================
static void _CB_OriginToCenter(Widget PW, EPointer PClientData, EPointer PCallData)
{
	if(((XePushButtonCallbackStruct*)PCallData)->Reason == XeCR_ACTIVATE_OPTIONS) { _OriginToCenterPopupDialog();return; }
	_OriginToCenter();
}


typedef struct
{
	EOperatorCore;

// E3dOperatorSnapPoints part
//
	E3dCoord3           GridSize;
	unsigned int        TransformAxes;
	E3dShapeSnapshot**  ShapeSnapshots;EIndex  NumOfShapeSnapshots;
} E3dOperatorSnapPoints;


//================================================
// Snap tags to grid Operator proc
//================================================
static EChangeFlags _SnapPoints_Proc(EOps* POps, EOperator* POperator, const int POperation)
{
	E3dScene*               LScene = E3d_Scene;
	E3dOperatorSnapPoints*  LOperator = (E3dOperatorSnapPoints*)POperator;

	const EIndex  LShapeN = LOperator->NumOfShapeSnapshots;
	switch(POperation)
	{
		case EOpDO:
			{
				EListAI(, EOpNode, LOpNodes);

				E3dCoordinate  LGridSizeX = LOperator->GridSize.X, LGridSizeY = LOperator->GridSize.Y, LGridSizeZ = LOperator->GridSize.Z;
				EBool          LGridSnapX = FALSE, LGridSnapY = FALSE, LGridSnapZ = FALSE;


				if(LOperator->TransformAxes & E3dDoX) LGridSnapX = TRUE;
				if(LOperator->TransformAxes & E3dDoY) LGridSnapY = TRUE;
				if(LOperator->TransformAxes & E3dDoZ) LGridSnapZ = TRUE;

				EChangeFlags  LChanged = 0;

				EIndex  LObjC = 0;for(LObjC = 0;LObjC < LShapeN;LObjC++)
				{
					E3dShapeSnapshot*  LShapeSnapshot = LOperator->ShapeSnapshots[LObjC];
					E3dObject*         LObject = LShapeSnapshot->Object;
					const EBool        LLocalToWorldMatrixIsNotIdentity = LShapeSnapshot->LocalToWorldMatrixIsNotIdentity;

					E3dMatrix  LLocalToWorldMatrix;E3dMatrix_LoadIdentity(LLocalToWorldMatrix);
					E3dMatrix  LWorldToLocalMatrix;E3dMatrix_LoadIdentity(LWorldToLocalMatrix);

					if(LLocalToWorldMatrixIsNotIdentity)
					{
						E3dMatrix_Copy(LShapeSnapshot->LocalToWorldMatrix, LLocalToWorldMatrix);
						E3dMatrix_Invert3x4(LLocalToWorldMatrix, LWorldToLocalMatrix);
					}

					if(EObject_IsSubClass(LObject, E3dSpline))
					{
						const E3dSplineCVsSnapshot*  LSnapshot = (E3dSplineCVsSnapshot*)LShapeSnapshot;
						const E3dSnapshotSplineCV*   LSnapshotCV = LSnapshot->SnapshotCVs;
						E3dSpline*                   LSpline = (E3dSpline*)LObject;
						E3dSplineCV*                 LCVs = LSpline->CVs;

						EIndex  LC = 0;EIndex  LVN = LSnapshot->CVs_Count;for(;LC < LVN;LC++, LSnapshotCV++)
						{
							E3dSplineCV*  LCV = LCVs + LSnapshotCV->Index;
							E3dCoord3     LPosition, LPrevious, LNext;

							if(LLocalToWorldMatrixIsNotIdentity)
							{
								E3dMatrix_Transform3x4_vv(LLocalToWorldMatrix, LCV->Position,  LPosition);
								E3dMatrix_Transform3x4_vv(LLocalToWorldMatrix, LCV->Previous,  LPrevious);
								E3dMatrix_Transform3x4_vv(LLocalToWorldMatrix, LCV->Next,      LNext);
							}
							else
							{
								LPosition = LCV->Position;
								LPrevious = LCV->Previous;
								LNext     = LCV->Next;
							}

							const E3dCoord3  LOldPosition = LPosition;

							if(LGridSnapX) LPosition.X = E3d_SnapCoordinateTo(LPosition.X, LGridSizeX);
							if(LGridSnapY) LPosition.Y = E3d_SnapCoordinateTo(LPosition.Y, LGridSizeY);
							if(LGridSnapZ) LPosition.Z = E3d_SnapCoordinateTo(LPosition.Z, LGridSizeZ);


							if(LLocalToWorldMatrixIsNotIdentity)
							{
								E3dMatrix_Transform3x4_vv(LWorldToLocalMatrix,  LPosition,  LCV->Position);

// Apply the differentials to the tangent end points.
// We always update these, as the user might switch back-and-forth between e.g. Bezier and linear Spline type.
//
								{
									const E3d_Sub3i_vv(LDiff, LPosition, LOldPosition);

									E3d_Add3_vvv(LPrevious, LDiff, LPrevious);
									E3d_Add3_vvv(LNext,     LDiff, LNext);

									E3dMatrix_Transform3x4_vv(LWorldToLocalMatrix,  LPrevious,  LCV->Previous);
									E3dMatrix_Transform3x4_vv(LWorldToLocalMatrix,  LNext,      LCV->Next);
								}
							}
							else
							{
								LCV->Position = LPosition;

// Apply the differentials to the tangent end points.
// We always update these, as the user might switch back-and-forth between e.g. Bezier and linear Spline type.
//
								{
									const E3d_Sub3i_vv(LDiff, LPosition, LOldPosition);

									if(LGridSnapX) { LPrevious.X += LDiff.X;LNext.X += LDiff.X; }
									if(LGridSnapY) { LPrevious.Y += LDiff.Y;LNext.Y += LDiff.Y; }
									if(LGridSnapZ) { LPrevious.Z += LDiff.Z;LNext.Z += LDiff.Z; }

									LCV->Previous = LPrevious;
									LCV->Next     = LNext;
								}
							}
						}

						LChanged |= E3dSpline_Changed(LSpline, E3dChgSPLINESHAPE,  &LOpNodes, &LOpNodes_Count, &LOpNodes_CountAllocated);
					}
					else if(EObject_IsSubClass(LObject, E3dMesh))
					{
						const E3dMeshVerticesSnapshot*  LSnapshot = (E3dMeshVerticesSnapshot*)LShapeSnapshot;
						const E3dSnapshotVertex*        LSnapshotVertex = LSnapshot->SnapshotVertices;
						E3dMesh*                        LMesh = (E3dMesh*)LObject;
						E3dVertex*                      LVertices = LMesh->Vertices;const E3dVertexSize  LVertexSize = LMesh->VertexSize;

						EIndex  LC = 0;EIndex  LVN = LSnapshot->NumOfPoints;for(;LC < LVN;LC++, LSnapshotVertex++)
						{
							E3dVertex*  LVertex = EStruct_ByIndex(LVertices, LVertexSize, LSnapshotVertex->Index);
							E3dCoord3   L;

							if(LLocalToWorldMatrixIsNotIdentity) E3dMatrix_Transform3x4_vv(LLocalToWorldMatrix, LVertex->Position,  L);
							else L = LVertex->Position;

							if(LGridSnapX) L.X = E3d_SnapCoordinateTo(L.X, LGridSizeX);
							if(LGridSnapY) L.Y = E3d_SnapCoordinateTo(L.Y, LGridSizeY);
							if(LGridSnapZ) L.Z = E3d_SnapCoordinateTo(L.Z, LGridSizeZ);

							if(LLocalToWorldMatrixIsNotIdentity) E3dMatrix_Transform3x4_vv(LWorldToLocalMatrix,  L,  LVertex->Position);
							else LVertex->Position = L;
						}


						E3dObject_UpdateForDisplay(LObject, E3duf_SHAPE, LScene);

						EOpNode*  LOpNode = EList_StructAddAClr(LOpNodes, 4);if(LOpNode)
						{
							LOpNode->Object  = (ECoreObject*)LObject;
							LOpNode->Changed = E3dChgMESHSHAPE;
						}

						LChanged |= E3dChgMESHSHAPE;
					}
				}

				if(LOpNodes)
				{
					LChanged |= E3dObjects_CallOutputs(ECbINTERNAL|ECbUSER, LOpNodes, LOpNodes_Count);
					EFree(LOpNodes);
				}

				return(LChanged);
			}

		case EOpUNDO:
		return(_E3d_UndoPointsTransform(LOperator->ShapeSnapshots, LOperator->NumOfShapeSnapshots));
	}
	return(0);
}


static void _SnapPoints_Init(EOperator* POperator)
{
	E3dOperatorSnapPoints*  LOperator = (E3dOperatorSnapPoints*)POperator;

	LOperator->GridSize.X = 1.0;
	LOperator->GridSize.Y = 1.0;
	LOperator->GridSize.Z = 1.0;
	LOperator->TransformAxes = 0;
	LOperator->NumOfShapeSnapshots = 0;
	LOperator->ShapeSnapshots = NULL;
}


static int _SnapPoints_Free(EOperator* POperator)
{
	E3dOperatorSnapPoints*	LOperator = (E3dOperatorSnapPoints*)POperator;

	E3d_ShapeSnapshotsFree(&LOperator->ShapeSnapshots, &LOperator->NumOfShapeSnapshots);

	return(E_SUCCESS);
}


//========================================
// Snap tagged points to the grid
//========================================
static void _CB_SnapTags(Widget PW, EPointer PClientData, EPointer PCallData)
{
	E3dScene*  LScene = E3d_Scene;
	EIndex     LShapeSnapshots_Count;E3dShapeSnapshot**  LShapeSnapshots = E3dScene_CollectTransformPoints(LScene, &LShapeSnapshots_Count, FALSE, TRUE);

	if(LShapeSnapshots_Count > 0)
	{
		E3dpWindow*             L3DWindow = E3dp_Main3DWindow;
		E3dGridSettings*        LGridSettings = &L3DWindow->GridSettings;
		E3dOperatorSnapPoints*  LOperator;
		unsigned int            LTranslateAxes = 0;

// Store Point (Vertex, Spline key etc.) positions for undo
//
		LOperator = (E3dOperatorSnapPoints*)EOp_OperatorNew(_OperatorClass);
		EObject_SetName(LOperator, EStrDup("Snap points to grid"), FALSE);

		E3dTransformPanel*  LTransformPanel = &E3dp_TransformPanel;
		if(LTransformPanel->TransformMode != E3dTmLOCAL)
		{
			EIndex  LC = 0;for(;LC < LShapeSnapshots_Count;LC++)
			{
				E3dShapeSnapshot*  LShapeSnapshot = LShapeSnapshots[LC];
				E3dObject*         LObject = LShapeSnapshot->Object;if(LObject)
				{
					E3dModel*  LModel = E3dObject_GetSelectingModel(LObject, TRUE);if(LModel)
					{
						E3dMatrix_Copy(LModel->LocalToWorldMatrix, LShapeSnapshot->LocalToWorldMatrix);
						LShapeSnapshot->LocalToWorldMatrixIsNotIdentity = TRUE;
					}
				}
			}
		}

		if(LGridSettings->GridSnapToX) LTranslateAxes |= E3dDoX;
		if(LGridSettings->GridSnapToY) LTranslateAxes |= E3dDoY;
		if(LGridSettings->GridSnapToZ) LTranslateAxes |= E3dDoZ;
		LOperator->GridSize       = LGridSettings->GridSnapSize;
		LOperator->TransformAxes  = LTranslateAxes;
		LOperator->ShapeSnapshots = LShapeSnapshots;LOperator->NumOfShapeSnapshots = LShapeSnapshots_Count;

		E3dp_UndoStackAddOperator((EOperator*)LOperator, E3dLng_Snap_tagged_points_to_grid);

		E3dWindows_RedrawByChangeFlags(_SnapPoints_Proc(NULL, (EOperator*)LOperator, EOpDO));
	}
}


//========================================
// Entry point of the plugin
//========================================
int E_PluginInit(EPlugin* PPlugin)
{
	EOperatorClass  LOperatorClassDesc[] =
	{
		{
			.Name=        "Snap points to grid",
			.StructSize=  sizeof(E3dOperatorSnapPoints),
			.Init=        _SnapPoints_Init,
			.Free=        _SnapPoints_Free,
			.Proc=        _SnapPoints_Proc
		}
	};

	_Plugin = PPlugin;

	_OperatorClass = EOp_OperatorClassRegister(LOperatorClassDesc);

	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Transform", "Freeze to world", '\0', NULL, NULL, FALSE, NULL, _CB_FreezeToWorld, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Transform", "Freeze branch to its parent", '\0', NULL, NULL, FALSE, NULL, _CB_FreezeBranchToItsParent, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Transform", "Freeze branch to its root", '\0', NULL, NULL, FALSE, NULL, _CB_FreezeBranchToItsRoot, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Transform", "Freeze rotation on branch", '\0', NULL, NULL, FALSE, NULL, _CB_FreezeRotation, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Transform", "Freeze scaling on branch", '\0', NULL, NULL, FALSE, NULL, _CB_FreezeScaling, (EPointer)0);
	_MenuButtons[_MenuButtons_Count++] = XePushButton_Add("Menu->Transform", "Snap tagged points to grid", '\0', NULL, NULL, FALSE, NULL, _CB_SnapTags, (EPointer)0);


	_ToolBarButtons[_ToolBarButtons_Count++] = XePushButton_Add("Tool->Transform", "Freeze to world", '\0', NULL, NULL, FALSE, "Perform the local-to-world transformations on the Geometries of the selected Models,\nthen reset the transformation values of the Models.", _CB_FreezeToWorld, (EPointer)0);
	_ToolBarButtons[_ToolBarButtons_Count++] = XePushButton_Add("Tool->Transform", "Freeze rotation", '\0', NULL, NULL, FALSE, "'Bake' rotations and scaling into the Geometries of the selected Models, then\nreset the rotation and scaling values of the Models (leaving only translation).", _CB_FreezeRotation, (EPointer)0);
	_ToolBarButtons[_ToolBarButtons_Count++] = XePushButton_Add("Tool->Transform", "Freeze scaling", '\0', NULL, NULL, FALSE, "'Bake' scaling into the Geometries of the selected Models,\nthen reset the scaling values of the Models.", _CB_FreezeScaling, (EPointer)0);
	_ToolBarButtons[_ToolBarButtons_Count++] = XePushButton_Add("Tool->Transform", "Origin to ctr", '\0', NULL, NULL, FALSE, "Move the local origin of Model(s) to the center of their selected Geometries.", _CB_OriginToCenter, (EPointer)0);

	return(E_SUCCESS);
}


//========================================
// Exit method of the plugin
//========================================
int E_PluginExit(const EBool PExiting)
{
	if(!PExiting || E3dp_GUIFreeOnExit)
	{
		{ int  LC = 0;for(;LC < _MenuButtons_Count;LC++) { XeItemFree(_MenuButtons[LC]); } _MenuButtons_Count = 0; }
		{ int  LC = 0;for(;LC < _ToolBarButtons_Count;LC++) { XeItemFree(_ToolBarButtons[LC]); } _ToolBarButtons_Count = 0; }
	}

	if(_OperatorClass) { EOp_OperatorClassDeactivate(_OperatorClass);_OperatorClass = NULL; }

	return(E_SUCCESS);
}
