//========================================================================
// E3D library
//
// - 3D (file, network etc.) I/O
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-20 19:07:34
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//========================================================================
#ifndef	_E3D_IO_h
#define	_E3D_IO_h

#include <float.h>

#ifndef _E3D_Scene_h
#include <E3D/Scene.h>
#endif


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#define E3dFileFormatNONE            NULL
#define E3dFileFormatFROM_EXTENSION  NULL  // Get format from filename extension
#define E3dFileFormatAUTOMATIC       NULL  // Automatic format recognition (for reading)

// Filter types
//
#define	E3dDataTypeNONE               0
#define E3dDataTypeMODEL              1
#define E3dDataTypeHIERARCHY         (1<<1)
#define E3dDataTypeSCENE             (1<<2)
#define E3dDataTypeMODEL_ANIMATION   (1<<3)
#define E3dDataTypeVERTEX_ANIMATION  (1<<4)
#define E3dDataTypeMATERIAL          (1<<5)
#define E3dDataTypeLIGHT             (1<<6)


typedef struct E3dIOModule_S  E3dIOModule;

typedef struct
{
	E3dIOModule*    IOModule;
	const char*     FileName;FILE*  File;       // For I/O to and from file
	uint8_t*        Buffer;size_t  BufferSize;  // For I/O to and from memory

	EBool           ClearSceneBefore;  // This allows Scene readers to overwrite (turn off) the DeleteAllBeforeReadingScene flag, as some file formats have multiple hierarchies, but no Lights, Cameras etc.
	EBool           ReadSession;
} E3dIOStruct;

typedef EBool  (*E3dCheckProc)(E3dIOStruct* PIOStruct);
typedef int    (*E3dReadProc)(E3dIOStruct* PIOStruct, void** PObjectRet);
typedef int    (*E3dWriteProc)(E3dIOStruct* PIOStruct, const unsigned char PType);

typedef int    (*E3dNodeReadFunc)(E3dIOStruct* PIOStruct, E3dNode** PONodeRet);
typedef int    (*E3dNodeWriteFunc)(E3dIOStruct* PIOStruct, E3dNode*, const unsigned char PType);

typedef int    (*E3dSceneReadFunc)(E3dIOStruct* PIOStruct, E3dScene** PSceneRet);
typedef int    (*E3dSceneWriteFunc)(E3dIOStruct* PIOStruct, E3dScene*, const unsigned char PType);

#define E3dIOModuleCore\
	EIOModuleCore;\
	unsigned int    StructSize;\
	EResource*      Resources;\
	unsigned int    DataTypes;           /* E3dDataTypeMODEL, E3dDataTypeSCENE etc...    */\
	E3dCheckProc    Check;               /* Function to check if data is in this format  */\
	E3dReadProc     Read;                /* Function to read 3D data                     */\
	E3dWriteProc    Write;               /* Function to write 3D data                    */\
	void            (*GUIForLoad)(E3dIOModule*, void*, const uint16_t);\
	void            (*GUIForSave)(E3dIOModule*, void*, const uint16_t)

struct E3dIOModule_S
{
	E3dIOModuleCore;
};

EList(extern, E3dIOModule*, E3d_3DModelFileFormats);
EList(extern, E3dIOModule*, E3d_3DHierarchyFileFormats);
EList(extern, E3dIOModule*, E3d_3DSceneFileFormats);

extern EpImage*  _Image_File3D;  // Default 3D file icon for XeFileSelectionBox

extern int            E3d_3DIOReadStatus;

extern E3dIOModule*   E3d_IOModuleByFileNameExtension(const char* PExtension,  E3dIOModule** PFormats, const EIndex PNIOModules,  const uint32_t PFlags);

extern const char*    E3d_GetBaseFileName(const char* PFileName, char* PNameRet, const int PMaxLen);


extern void*      E3d_3DModelIOModuleAdd(E3dIOModule* PTemplate);
extern void       E3d_3DModelIOModuleRemove(void* PFileFormat);
extern void       E3d_3DModelIOModuleDeactivate(void* PFileFormat);

extern void*      E3d_3DHierarchyIOModuleAdd(E3dIOModule* PTemplate);
extern void       E3d_3DHierarchyIOModuleRemove(void* PFileFormat);
extern void       E3d_3DHierarchyIOModuleDeactivate(void* PFileFormat);

extern void*      E3d_3DSceneIOModuleAdd(E3dIOModule* PTemplate);
extern void       E3d_3DSceneIOModuleRemove(void* PFileFormat);
extern void       E3d_3DSceneIOModuleDeactivate(void* PFileFormat);

extern void       E3d_IOModulesFree(void);


extern int       E3dNode_LoadFromFile(const char* PFileName, E3dIOModule* PFileFormat, E3dIOModule** PFileFormatRet, E3dNode** PNode);
extern int       E3dNode_SaveToFile(const char* PFileName, char* PNameReturn, const int PNameMaxLen, E3dNode* PNode, E3dIOModule* PFileFormat, E3dIOModule** PFileFormatRet);

extern int       E3d_HierarchyReadFromFile(const char* PFileName, E3dIOModule* PFileFormat, E3dIOModule** PFileFormatRet, E3dNode** PNodePtr);
extern int       E3d_HierarchyWriteToFile(const char* PFileName, char* PNameReturn, int PNameMaxLen, E3dNode* PNode, E3dIOModule* PFileFormat, E3dIOModule** PFileFormatRet);

extern int       E3d_SceneReadFromFile(const char* PFileName, E3dIOModule* PFileFormat, E3dIOModule** PFileFormatRet, E3dScene** PScene, EBool* PClearSceneBeforeP, const EBool PReadSession);
extern int       E3d_SceneWriteToFile(const char* PFileName, char* PNameReturn, int PNameMaxLen, E3dScene* PModel, E3dIOModule* PFileFormat, E3dIOModule** PFileFormatRet);

extern E3dGeometry*  E3d_3DIconReadGeometry(const char* PFileName);
extern E3dNode*      E3d_3DIconReadNode(const char* PFileName);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
} // Close scope of 'extern "C"' declaration
#endif


#endif	// _E3D_IO_h
