//========================================================================
// E3D library
//
// Matrix operations
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-10 20:49:49
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//========================================================================
#ifndef _E3D_Matrix_h
#define _E3D_Matrix_h


#ifndef _E3D_Types_h
#include <E3D/Types.h>
#endif


//----------------------------------------------------------------
// Macro functions
//----------------------------------------------------------------

//
// Suffixes for input and output arguments
//
// I : implicit (mX, mY, mZ declared and set by the caller)
// V : values   (InX, InY, InZ)
// S : struct  (In.X, In.Y, In.Z)
// P : struct pointer  (In->X, In->Y, In->Z)
//
// If there is only one suffix, the input and output will be the same variable
//

#ifdef E3dSIMD
#define _M_ClearWs(m)	m.unused = 0.0
#define _M_ClearWp(m)	m->unused = 0.0
#else
#define _M_ClearWs(m)
#define _M_ClearWp(m)
#endif

//========================================================
// Transform [X, Y] with a Matrix's left 2x2 part
//========================================================
#define E3dMatrix_Transform2x2_s(mMatrix,  mInOut) do\
{\
	const E3dCoordinate  lX = mInOut.X, lY = mInOut.Y;\
	mInOut.X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10];\
	mInOut.Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11];\
} while(0)

//========================================================
// Transform [X, Y] with a Matrix's left 2x2 part
//========================================================
#define E3dMatrix_Transform2x2_p(mMatrix,  mInOut) do\
{\
	const E3dCoordinate  lX = mInOut->X, lY = mInOut->Y;\
	mInOut->X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10];\
	mInOut->Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11];\
} while(0)

//========================================================
// Transform [X, Y] with a Matrix's left 2x2 part
//========================================================
#define E3dMatrix_Transform2x2_ss(mMatrix,  mIn_X, mIn_Y,  mOut_X, mOut_Y) do\
{\
	const E3dCoordinate  lX = mIn_X, lY = mIn_Y;\
	mOut_X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10];\
	mOut_Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11];\
} while(0)



//========================================================
// Transform [X, Y] with a Matrix's left 2x4 part
//========================================================
#define E3dMatrix_Transform2x4_s(mMatrix,  mIO_X, mIO_Y) do\
{\
	const E3dCoordinate  mX = mIO_X, mY = mIO_Y;\
	mIO_X = mX * (mMatrix)[M00] + mY * (mMatrix)[M10] + (mMatrix)[M30];\
	mIO_Y = mX * (mMatrix)[M01] + mY * (mMatrix)[M11] + (mMatrix)[M31];\
} while(0)



#define E3dMatrix_Transform3x3_s(mMatrix,  mIO_X, mIO_Y, mIO_Z) do\
{\
	const E3dCoordinate  lX = mIO_X, lY = mIO_Y, lZ = mIO_Z;\
	mIO_X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	mIO_Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	mIO_Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
} while(0)

#define E3dMatrix_Transform3x3_a(mMatrix,  mIO) do\
{\
	const E3dCoordinate  lX = mIO[0], lY = mIO[1], lZ = mIO[2];\
	mIO[0] = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	mIO[1] = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	mIO[2] = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
} while(0)

#define E3dMatrix_Transform3x3_v(mMatrix,  mInOut) do\
{\
	const E3dCoordinate  lX = mInOut.X, lY = mInOut.Y, lZ = mInOut.Z;\
	mInOut.X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	mInOut.Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	mInOut.Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWs(mInOut);\
} while(0)

#define E3dMatrix_Transform3x3_p(mMatrix,  mInOut) do\
{\
	const E3dCoordinate  lX = mInOut->X, lY = mInOut->Y, lZ = mInOut->Z;\
	mInOut->X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	mInOut->Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	mInOut->Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWp(mInOut);\
} while(0)

#define E3dMatrix_Transform3x3_sv(mMatrix,  mX, mY, mZ,  mOut) do\
{\
	const E3dCoordinate  lX = mX, lY = mY, lZ = mZ;\
	mOut.X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	mOut.Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	mOut.Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWs(mOut);\
} while(0)

#define E3dMatrix_Transform3x3_ss(mMatrix,  mX, mY, mZ,  mOut_X, mOut_Y, mOut_Z) do\
{\
	const E3dCoordinate  lX = mX, lY = mY, lZ = mZ;\
	mOut_X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	mOut_Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	mOut_Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
} while(0)

#define E3dMatrix_Transform3x3_sp(mMatrix,  mX, mY, mZ,  mOut) do\
{\
	const E3dCoordinate  lX = mX, lY = mY, lZ = mZ;\
	mOut->X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	mOut->Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	mOut->Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWp(mOut);\
} while(0)

#define E3dMatrix_Transform3x3_vs(mMatrix,  mIn,  mOut_X, mOut_Y, mOut_Z) do\
{\
	const E3dCoordinate  lX = mIn.X, lY = mIn.Y, lZ = mIn.Z;\
	mOut_X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	mOut_Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	mOut_Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
} while(0)

#define E3dMatrix_Transform3x3_vv(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn.X, lY = mIn.Y, lZ = mIn.Z;\
	(mOut).X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	(mOut).Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	(mOut).Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWs(mOut);\
} while(0)

#define E3dMatrix_Transform3x3_vp(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn.X, lY = mIn.Y, lZ = mIn.Z;\
	(mOut)->X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	(mOut)->Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	(mOut)->Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWs(mOut);\
} while(0)

#define E3dMatrix_Transform3x3_pv(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn->X, lY = mIn->Y, lZ = mIn->Z;\
	(mOut).X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	(mOut).Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	(mOut).Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWs(mOut);\
} while(0)

#define E3dMatrix_Transform3x3_pp(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn->X, lY = mIn->Y, lZ = mIn->Z;\
	(mOut)->X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	(mOut)->Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	(mOut)->Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWs(mOut);\
} while(0)

#define E3dMatrix_Transform3x3_sa(mMatrix,  mIn_X, mIn_Y, mIn_Z,  mOut) do\
{\
	const E3dCoordinate  lX = mIn_X, lY = mIn_Y, lZ = mIn_Z;\
	(mOut)[E3dX] = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	(mOut)[E3dY] = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	(mOut)[E3dZ] = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWs(mOut);\
} while(0)

#define E3dMatrix_Transform3x3_aa(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn[E3dX], lY = mIn[E3dY], lZ = mIn[E3dZ];\
	(mOut)[E3dX] = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20];\
	(mOut)[E3dY] = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21];\
	(mOut)[E3dZ] = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22];\
	_M_ClearWs(mOut);\
} while(0)



//========================================================
// Transform [X, Y, Z] with a Matrix's left 3x4 part
//  Implicit input (mX, mY, mZ)
//========================================================
#define E3dMatrix_Transform3x4_Is(mMatrix,  mOut_X, mOut_Y, mOut_Z) do\
{\
	mOut_X = mX * (mMatrix)[M00] + mY * (mMatrix)[M10] + mZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mOut_Y = mX * (mMatrix)[M01] + mY * (mMatrix)[M11] + mZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mOut_Z = mX * (mMatrix)[M02] + mY * (mMatrix)[M12] + mZ * (mMatrix)[M22] + (mMatrix)[M32];\
} while(0)

//========================================================
// Transform [X, Y, Z] with a Matrix's left 3x4 part
//========================================================
#define E3dMatrix_Transform3x4_s(mMatrix,  mIO_X, mIO_Y, mIO_Z) do\
{\
	const E3dCoordinate  lX = mIO_X, lY = mIO_Y, lZ = mIO_Z;\
	mIO_X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mIO_Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mIO_Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
} while(0)

#define E3dMatrix_Transform3x4_v(mMatrix,  mIO) do\
{\
	const E3dCoordinate  lX = mIO.X, lY = mIO.Y, lZ = mIO.Z;\
	mIO.X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mIO.Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mIO.Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
	_M_ClearWs(mIO);\
} while(0)

#define E3dMatrix_Transform3x4_p(mMatrix,  mIO) do\
{\
	const E3dCoordinate  lX = mIO->X, lY = mIO->Y, lZ = mIO->Z;\
	mIO->X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mIO->Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mIO->Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
	_M_ClearWs(mIO);\
} while(0)


#define E3dMatrix_Transform3x4_ss(mMatrix,  mIn_X, mIn_Y, mIn_Z,  mOut_X, mOut_Y, mOut_Z) do\
{\
	const E3dCoordinate  lX = mIn_X, lY = mIn_Y, lZ = mIn_Z;\
	mOut_X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mOut_Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mOut_Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
} while(0)

#define E3dMatrix_Transform3x4_vs(mMatrix,  mIn,  mOut_X, mOut_Y, mOut_Z) do\
{\
	const E3dCoordinate  lX = mIn.X, lY = mIn.Y, lZ = mIn.Z;\
	mOut_X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mOut_Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mOut_Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
} while(0)

#define E3dMatrix_Transform3x4_vv(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn.X, lY = mIn.Y, lZ = mIn.Z;\
	mOut.X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mOut.Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mOut.Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
	_M_ClearWs(mOut);\
} while(0)

#define E3dMatrix_Transform3x4_vp(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn.X, lY = mIn.Y, lZ = mIn.Z;\
	mOut->X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mOut->Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mOut->Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
	_M_ClearWp(mOut);\
} while(0)

#define E3dMatrix_Transform3x4_pv(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn->X, lY = mIn->Y, lZ = mIn->Z;\
	mOut.X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mOut.Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mOut.Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
	_M_ClearWs(mOut);\
} while(0)

#define E3dMatrix_Transform3x4_pp(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn->X, lY = mIn->Y, lZ = mIn->Z;\
	mOut->X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mOut->Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mOut->Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
	_M_ClearWp(mOut);\
} while(0)

#define E3dMatrix_Transform3x4_sp(mMatrix,  mX, mY, mZ,  mOut) do\
{\
	const E3dCoordinate  lX = mX, lY = mY, lZ = mZ;\
	(mOut)->X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	(mOut)->Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	(mOut)->Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
	_M_ClearWp(mOut);\
} while(0)

#define E3dMatrix_Transform3x4_sv(mMatrix,  mX, mY, mZ,  mOut) do\
{\
	const E3dCoordinate  lX = mX, lY = mY, lZ = mZ;\
	(mOut).X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	(mOut).Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	(mOut).Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
	_M_ClearWs(mOut);\
} while(0)

#define E3dMatrix_Transform3x4_ps(mMatrix,  mIn,  mOut_X, mOut_Y, mOut_Z) do\
{\
	const E3dCoordinate  lX = mIn->X, lY = mIn->Y, lZ = mIn->Z;\
	mOut_X = lX * (mMatrix)[M00] + lY * (mMatrix)[M10] + lZ * (mMatrix)[M20] + (mMatrix)[M30];\
	mOut_Y = lX * (mMatrix)[M01] + lY * (mMatrix)[M11] + lZ * (mMatrix)[M21] + (mMatrix)[M31];\
	mOut_Z = lX * (mMatrix)[M02] + lY * (mMatrix)[M12] + lZ * (mMatrix)[M22] + (mMatrix)[M32];\
} while(0)



//========================================================================
// Transform a homogenous 3D position [X, Y, Z, 1.0] with a 4x4 Matrix	
//========================================================================
#define E3dMatrix_Transform4x4_hs(mMatrix,  mOut_X, mOut_Y, mOut_Z, mOut_W) do\
{\
	mOut_X = mX * mMatrix[M00] + mY * mMatrix[M10] + mZ * mMatrix[M20] + mMatrix[M30],\
	mOut_Y = mX * mMatrix[M01] + mY * mMatrix[M11] + mZ * mMatrix[M21] + mMatrix[M31],\
	mOut_Z = mX * mMatrix[M02] + mY * mMatrix[M12] + mZ * mMatrix[M22] + mMatrix[M32],\
	mOut_W = mX * mMatrix[M03] + mY * mMatrix[M13] + mZ * mMatrix[M23] + mMatrix[M33];\
} while(0)

#define E3dMatrix_Transform4x4_hv(mMatrix,  mInOut) do\
{\
	const E3dCoordinate  lX = (mInOut).X, lY = (mInOut).Y, lZ = (mInOut).Z;\
\
	(mInOut).X = lX * mMatrix[M00] + lY * mMatrix[M10] + lZ * mMatrix[M20] + mMatrix[M30],\
	(mInOut).Y = lX * mMatrix[M01] + lY * mMatrix[M11] + lZ * mMatrix[M21] + mMatrix[M31],\
	(mInOut).Z = lX * mMatrix[M02] + lY * mMatrix[M12] + lZ * mMatrix[M22] + mMatrix[M32],\
	(mInOut).W = lX * mMatrix[M03] + lY * mMatrix[M13] + lZ * mMatrix[M23] + mMatrix[M33];\
} while(0)

#define E3dMatrix_Transform4x4_hss(mMatrix,  mX, mY, mZ,  mOut_X, mOut_Y, mOut_Z, mOut_W) do\
{\
  const E3dCoordinate  lX = mX, lY = mY, lZ = mZ;\
\
	mOut_X = lX * mMatrix[M00] + lY * mMatrix[M10] + lZ * mMatrix[M20] + mMatrix[M30];\
	mOut_Y = lX * mMatrix[M01] + lY * mMatrix[M11] + lZ * mMatrix[M21] + mMatrix[M31];\
	mOut_Z = lX * mMatrix[M02] + lY * mMatrix[M12] + lZ * mMatrix[M22] + mMatrix[M32];\
	mOut_W = lX * mMatrix[M03] + lY * mMatrix[M13] + lZ * mMatrix[M23] + mMatrix[M33];\
} while(0)

#define E3dMatrix_Transform4x4_hsv(mMatrix,  mX, mY, mZ,  mOut) do\
{\
	const E3dCoordinate  lX = mX, lY = mY, lZ = mZ;\
\
	mOut.X = lX * mMatrix[M00] + lY * mMatrix[M10] + lZ * mMatrix[M20] + mMatrix[M30];\
	mOut.Y = lX * mMatrix[M01] + lY * mMatrix[M11] + lZ * mMatrix[M21] + mMatrix[M31];\
	mOut.Z = lX * mMatrix[M02] + lY * mMatrix[M12] + lZ * mMatrix[M22] + mMatrix[M32];\
	mOut.W = lX * mMatrix[M03] + lY * mMatrix[M13] + lZ * mMatrix[M23] + mMatrix[M33];\
} while(0)

#define E3dMatrix_Transform4x4_hvs(mMatrix,  mIn,  mOut_X, mOut_Y, mOut_Z, mOut_W) do\
{\
	const E3dCoordinate  lX = mIn.X, lY = mIn.Y, lZ = mIn.Z;\
\
	mOut_X = lX * mMatrix[M00] + lY * mMatrix[M10] + lZ * mMatrix[M20] + mMatrix[M30];\
	mOut_Y = lX * mMatrix[M01] + lY * mMatrix[M11] + lZ * mMatrix[M21] + mMatrix[M31];\
	mOut_Z = lX * mMatrix[M02] + lY * mMatrix[M12] + lZ * mMatrix[M22] + mMatrix[M32];\
	mOut_W = lX * mMatrix[M03] + lY * mMatrix[M13] + lZ * mMatrix[M23] + mMatrix[M33];\
} while(0)

#define E3dMatrix_Transform4x4_hvv(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn.X, lY = mIn.Y, lZ = mIn.Z;\
\
	mOut.X = lX * mMatrix[M00] + lY * mMatrix[M10] + lZ * mMatrix[M20] + mMatrix[M30];\
	mOut.Y = lX * mMatrix[M01] + lY * mMatrix[M11] + lZ * mMatrix[M21] + mMatrix[M31];\
	mOut.Z = lX * mMatrix[M02] + lY * mMatrix[M12] + lZ * mMatrix[M22] + mMatrix[M32];\
	mOut.W = lX * mMatrix[M03] + lY * mMatrix[M13] + lZ * mMatrix[M23] + mMatrix[M33];\
} while(0)

#define E3dMatrix_Transform4x4_hpv(mMatrix,  mIn,  mOut) do\
{\
	const E3dCoordinate  lX = mIn->X, lY = mIn->Y, lZ = mIn->Z;\
\
	mOut.X = lX * mMatrix[M00] + lY * mMatrix[M10] + lZ * mMatrix[M20] + mMatrix[M30];\
	mOut.Y = lX * mMatrix[M01] + lY * mMatrix[M11] + lZ * mMatrix[M21] + mMatrix[M31];\
	mOut.Z = lX * mMatrix[M02] + lY * mMatrix[M12] + lZ * mMatrix[M22] + mMatrix[M32];\
	mOut.W = lX * mMatrix[M03] + lY * mMatrix[M13] + lZ * mMatrix[M23] + mMatrix[M33];\
} while(0)


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

typedef struct
{
	int         Pointer, AllocatedDepth;
	E3dMatrix*  Matrices;
} E3dMatrixStack;


//================================================
// Rotate Matrix around the X axis - radians
//================================================
#define E3d_MatrixRotateX_rad(mMatrix, mAngle) do\
{\
	const E3dCoordinate  lCosA = E3d_Cos(-mAngle), lSinA = E3d_Sin(-mAngle);\
	const E3dCoordinate  lMSinA = -lSinA;\
	E3dMatrix            lTMatrix;\
\
	lTMatrix[M10] = mMatrix[M10] * lCosA + mMatrix[M20] * lSinA;lTMatrix[M11] = mMatrix[M11] * lCosA + mMatrix[M21] * lSinA;lTMatrix[M12] = mMatrix[M12] * lCosA + mMatrix[M22] * lSinA;\
	lTMatrix[M20] = mMatrix[M10] * lMSinA + mMatrix[M20] * lCosA;lTMatrix[M21] = mMatrix[M11] * lMSinA + mMatrix[M21] * lCosA;lTMatrix[M22] = mMatrix[M12] * lMSinA + mMatrix[M22] * lCosA;\
\
	mMatrix[M10] = lTMatrix[M10];mMatrix[M11] = lTMatrix[M11];mMatrix[M12] = lTMatrix[M12];\
	mMatrix[M20] = lTMatrix[M20];mMatrix[M21] = lTMatrix[M21];mMatrix[M22] = lTMatrix[M22];\
} while(0)


//================================================
// Rotate Matrix around the X axis - radians
//================================================
#define E3d_MatrixRotateY_rad(mMatrix, mAngle) do\
{\
	const E3dCoordinate  lCosA = E3d_Cos(-mAngle), lSinA = E3d_Sin(-mAngle);\
	const E3dCoordinate  lMSinA = -lSinA;\
	E3dMatrix            lTMatrix;\
\
	lTMatrix[M00] = mMatrix[M00] * lCosA + mMatrix[M20] * lMSinA;lTMatrix[M01] = mMatrix[M01] * lCosA + mMatrix[M21] * lMSinA;lTMatrix[M02] = mMatrix[M02] * lCosA + mMatrix[M22] * lMSinA;\
	lTMatrix[M20] = mMatrix[M00] * lSinA + mMatrix[M20] * lCosA; lTMatrix[M21] = mMatrix[M01] * lSinA + mMatrix[M21] * lCosA; lTMatrix[M22] = mMatrix[M02] * lSinA + mMatrix[M22] * lCosA;\
\
	mMatrix[M00] = lTMatrix[M00];mMatrix[M01] = lTMatrix[M01];mMatrix[M02] = lTMatrix[M02];\
	mMatrix[M20] = lTMatrix[M20];mMatrix[M21] = lTMatrix[M21];mMatrix[M22] = lTMatrix[M22];\
} while(0)


//================================================
// Rotate Matrix around the X axis - radians
//================================================
#define E3d_MatrixRotateZ_rad(mMatrix, mAngle) do\
{\
	const E3dCoordinate  lCosA = E3d_Cos(-mAngle), lSinA = E3d_Sin(-mAngle);\
	const E3dCoordinate  lMSinA = -lSinA;\
	E3dMatrix            lTMatrix;\
\
	lTMatrix[M00] = mMatrix[M00] * lCosA + mMatrix[M10] * lSinA;lTMatrix[M01] = mMatrix[M01] * lCosA + mMatrix[M11] * lSinA;lTMatrix[M02] = mMatrix[M02] * lCosA + mMatrix[M12] * lSinA;\
	lTMatrix[M10] = mMatrix[M00] * lMSinA + mMatrix[M10] * lCosA;lTMatrix[M11] = mMatrix[M01] * lMSinA + mMatrix[M11] * lCosA;lTMatrix[M12] = mMatrix[M02] * lMSinA + mMatrix[M12] * lCosA;\
\
	mMatrix[M00] = lTMatrix[M00];mMatrix[M01] = lTMatrix[M01];mMatrix[M02] = lTMatrix[M02];\
	mMatrix[M10] = lTMatrix[M10];mMatrix[M11] = lTMatrix[M11];mMatrix[M12] = lTMatrix[M12];\
} while(0)



//================================================
// Matrix constructors
//================================================
static inline void E3dMatrix3x3_VectorsToRows3x3_ppp(E3dMatrix PMatrix, const E3dCoord3* PV0, const E3dCoord3* PV1, const E3dCoord3* PV2)
{
	PMatrix[M00] = PV0->X;PMatrix[M01] = PV0->Y;PMatrix[M02] = PV0->Z;
	PMatrix[M10] = PV1->X;PMatrix[M11] = PV1->Y;PMatrix[M12] = PV1->Z;
	PMatrix[M20] = PV2->X;PMatrix[M21] = PV2->Y;PMatrix[M22] = PV2->Z;
}


static inline void E3dMatrix3x3_VectorsToCols3x3_ppp(E3dMatrix PMatrix, const E3dCoord3* PV0, const E3dCoord3* PV1, const E3dCoord3* PV2)
{
	PMatrix[M00] = PV0->X;PMatrix[M10] = PV0->Y;PMatrix[M20] = PV0->Z;
	PMatrix[M01] = PV1->X;PMatrix[M11] = PV1->Y;PMatrix[M21] = PV1->Z;
	PMatrix[M02] = PV2->X;PMatrix[M12] = PV2->Y;PMatrix[M22] = PV2->Z;
}



//----------------------------------------------------------------
// Global variables
//----------------------------------------------------------------
extern E3dMatrix      E3d_IdentityMatrix;


//----------------------------------------------------------------
// Function prototypes
//----------------------------------------------------------------
extern void           E3dMatrix_Print3x3(const E3dMatrix PMatrix);
extern void           E3dMatrix_Print4x4(const E3dMatrix PMatrix);

extern void           E3dMatrix_LoadIdentity(E3dMatrix PDestMtx);
extern void           E3dMatrix_Copy(const E3dMatrix PSrcMtx, E3dMatrix PDestMtx);

extern E3dCoordinate  E3dMatrix_Determinant3x3(const E3dMatrix PSrc);
extern void           E3dMatrix_Transpose3x3(const E3dMatrix PM, E3dMatrix PDst);

extern void           E3dMatrix_TransposeAdjoint3x3(const E3dMatrix PM, E3dMatrix PDst);
extern void           E3dMatrix_Adjoint3x3(const E3dMatrix PSrc, E3dMatrix PDst);
extern void           E3dMatrix_Invert3x3(const E3dMatrix PM, E3dMatrix PDst);

extern void           E3dMatrix_Adjoint3x4(const E3dMatrix PM, E3dMatrix PDst);
extern void           E3dMatrix_Invert3x4(const E3dMatrix PM, E3dMatrix PDst);

extern void           E3dMatrix_Transpose4x4(const E3dMatrix PM, E3dMatrix PDst);
extern EBool          E3d_Matrix4x4Invert(const E3dMatrix PSMatrix, E3dMatrix PDMatrix);

extern void           E3dMatrix_Copy3x3(const E3dMatrix PSrcMtx, E3dMatrix PDestMtx);
extern void           E3dMatrix_Copy3x4(const E3dMatrix PSrcMtx, E3dMatrix PDestMtx);

extern void           E3dMatrix_Transform3DDiv(const E3dMatrix PMatrix,  E3dCoord3* PPos);

extern void           E3dMatrix_Mult4x4(const E3dMatrix PSrcMtx, E3dMatrix PDestMtx);
extern void           E3dMatrix_Mult3x3_3(const E3dMatrix PMatrix1, const E3dMatrix PMatrix2, E3dMatrix PDestMtx);
extern void           E3dMatrix_Mult4x4_3(const E3dMatrix PMatrix1, const E3dMatrix PMatrix2, E3dMatrix PDestMtx);

extern EBool          E3dMatrix_Equal(const E3dMatrix PMtx1, const E3dMatrix PMtx2);
extern void           E3dMatrix_LoadTranslation(E3dMatrix PMatrix, const E3dCoordinate PX, const E3dCoordinate PY, const E3dCoordinate PZ);
extern void           E3dMatrix_Translate(E3dMatrix PMatrix, const E3dCoordinate PX, const E3dCoordinate PY, const E3dCoordinate PZ);
extern void           E3dMatrix_Scale(E3dMatrix PMatrix, const E3dCoordinate PX, const E3dCoordinate PY, const E3dCoordinate PZ);


extern void           E3dMatrix_Rotate2DRad(E3dMatrix PMatrix, const E3dCoordinate PAngle);
extern void           E3dMatrix_RotateDeg(E3dMatrix PMatrix, const int PAxis, const E3dCoordinate PAngle);
extern void           E3dMatrix_RotateRad(E3dMatrix PMatrix, const int PAxis, const E3dCoordinate PAngle);

extern void           E3dMatrix_RotateSinCos(E3dMatrix PMatrix, const int PAxis, const E3dCoordinate PSinA, const E3dCoordinate PCosA);
extern void           E3dMatrix_RotateXYZ(E3dMatrix PMatrix, const E3dCoord3* PAngles, const char POrder);
extern void           E3dMatrix_RotateAxisAngle(E3dMatrix PMatrix, const E3dCoord3* PAxis, const E3dCoordinate PAngle);

extern void           E3dMatrix_ToEuler(const E3dMatrix PMatrix, const char PRotOrder, E3dCoord3* PAngles);

extern void           E3dMatrix_SetLookAtInv(E3dMatrix PMatrix, const E3dCoord3* LPosition, const E3dCoord3* PAimPoint, const E3dCoord3* PUpVector);
extern void           E3dMatrix_LookAt(E3dMatrix PMatrix, const E3dCoord3* LPosition, const E3dCoord3* PAimPoint, const E3dCoord3* PUpVector);
extern void           E3dMatrix_SetLookAt(E3dMatrix PMatrix, const E3dCoord3* LPosition, const E3dCoord3* PAimPoint, const E3dCoord3* PUpVector);

extern void           E3dMatrix_SetOrtho(E3dMatrix PMatrix, const E3dCoordinate PLeft, const E3dCoordinate PRight, const E3dCoordinate PBottom, const E3dCoordinate PTop,  const E3dCoordinate PZNear, const E3dCoordinate PZFar);
extern void           E3dMatrix_SetFrustum(E3dMatrix PMatrix,  const E3dCoordinate PLeft, const E3dCoordinate PRight, const E3dCoordinate PBottom, const E3dCoordinate PTop,  const E3dCoordinate PZNear, const E3dCoordinate PZFar);
extern void           E3dMatrix_SetPerspective(E3dMatrix PMatrix, const E3dCoordinate PFieldOfView, const E3dCoordinate PAspectRatio, const E3dCoordinate PZNear, const E3dCoordinate PZFar);


extern void           E3dMatrix_ToQuaternion(const E3dMatrix PMatrix, E3dCoord4* PQuaternion);
extern void           E3dQuaternion_ToMatrix(const E3dCoord4* PQuaternion, E3dMatrix PMatrix);
extern void           E3d_AxisAngleToMatrix(const E3dCoord4* PAA, E3dMatrix PMatrix);

extern void           E3d_VectorsGetRotMatrix(const E3dCoord3* P1, const E3dCoord3* P2, E3dMatrix PMatrix);
extern void           E3d_NormalGetDefRotMatrix(const E3dCoord3* PNormal, E3dMatrix PMatrix);
extern void           E3d_TangentAndNormalToRotationMatrix(const E3dCoord3* PTangent, const E3dCoord3* PNormal, E3dMatrix PMatrix);


extern void           E3dEulerRotation_Init(E3dEulerRotation* PEulerRotation);
extern void           E3dRotation_Init(E3dRotation* PRotation);
extern void           E3dRotation_Set(E3dRotation* PRotation, const E3dCoordinate PX, const E3dCoordinate PY, const E3dCoordinate PZ, const E3dCoordinate PW);

extern EBool          E3d_UnProject(const E3dCoordinate PWindowX, const E3dCoordinate PWindowY, const E3dCoordinate PWindowZ,
                                    const GLint PViewport[4], E3dMatrix PProjectionMatrix, E3dMatrix PModelViewMatrix,
                                    E3dCoord3* PPositionP);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_Matrix_h
