//================================================================
// E3D
//
// Rendering-related header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 06:28:41
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Raytracing_h
#define _E3D_Raytracing_h

#include <E3D/Rendering.h>

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#ifndef _E3dRayTracer
#define _E3dRayTracer
typedef struct E3dRayTracer_S         E3dRayTracer;
typedef struct E3dRayTracer_thread_S  E3dRayTracer_thread;
#endif


typedef struct
{
	E3dCoordinate        CDFPosition;

	const E3dRTriangle*  RTriangle;
	E3dCoordinate        Alpha, SinAlpha, CosAlpha;
	E3dCoord3            TA, TB, TC, TU;
	E3dCoordinate        Product, TriangleApparentSolidAngle;  // Apparent solid angle from sampled point
	EBool                Set;
} E3dLightTriangleCDFEntry;



typedef struct
{
	union
	{
		E3dCoordinate  SplitValue;  // Interior: split plane position
		const void**   Primitives;  // Leaf: list of primitives
		const void*    Primitive;   // Leaf: direct index of one primitive
	};
	unsigned int    Flags;        // 2 bits: Split-axis / IsLeaf (0:X, 1:Y, 2:Z, 3:IsLeaf); 30bits: NumOfPrimitives (leaf) or index of right child (left child comes immediately after the current KDNode)
} E3dKDTreeNode;


typedef struct
{
	int             RefCnt;
	E3dBBox         BBox;

	float           CostRatio;     // Node traversal cost divided by primitive intersection cost
	float           EmptyBonus;    // Empty bonus
	EIndex	        NextFreeNodeIndex, NumOfAllocatedNodes;
	EIndex          TotalNumOfPrimitives;
	int             MaxDepth, MaxLeafSize;
	E3dKDTreeNode*  Nodes;

// These hold temporary data, to keep the function parameter-count reasonable.
//
	const void**    Primitives;
	const E3dBBox*  AllBounds;

// Statistics
//
	int             DepthLimitReached, NumOfBadSplits, NumOfClips, NumOfNullClips, NumOfBadClips;
	E3dNode*        VisNode;    // For visualizing the KDTree

	ETime           BuildStartMilliseconds, BuildEndMilliseconds;
} E3dKDTree;


typedef struct
{
	E3dBBox                BBox;
//	E3dBBoxMinSpan         BBox;
	union
	{
	 const E3dRTriangle**  Primitives;  // Leaf: list of primitives
	 const E3dRTriangle*   Primitive;   // Leaf: direct index of one primitive
	};
	unsigned int           Flags;       // 2 bits: Split-axis / IsLeaf (0:X, 1:Y, 2:Z, 3:IsLeaf); 30bits: NumOfPrimitives (leaf) or index of right child (left child comes immediately after the current BVHNode)
	EBool                  SiblingOverlaps;	// If TRUE, siblings overlap along split axis, so we can't skip the far child if an intersection is found in the near child
} E3dBVHNode;


typedef struct
{
	int                    RefCnt;
	E3dBBox                BBox;

	EIndex                 NextFreeNodeIndex, NumOfAllocatedNodes;
	EIndex                 TotalNumOfPrimitives;
	int                    MaxDepth, MaxLeafSize;
	E3dBVHNode*            Nodes;

	const E3dRTriangle**   Primitives;
	E3dBBoxCenterExtents*  AllBounds;

	E3dNode*               VisNode;  // For visualizing the BVH

	ETime                  BuildStartMilliseconds, BuildEndMilliseconds;
} E3dBVH;


#define E3dMAX_KDSTACK 64

typedef struct
{
	EIndex  TriQueries,
          TriQueriesAfterExcludeFace,
          TriQueriesAfterMailbox,
          RedundantTriQueries;
} E3dRayStats;

typedef struct
{
	E3dBBox    BBox;
	EInt32_3   GridCount;
	E3dVoxel*  Voxels;
	E3dCoord3  VoxelSize, VoxelSizeInv;
} E3dUniformGrid;



// Ray-traversal accelerator with:
// - Uniform-grid
// - KDTree
// - Bounding Volume Hierarchy (BVH)
//
typedef struct
{
	EObjectCore(EObject);

	EList(, E3dRTriangle*, Triangles);  // Used temporarily, to build the KD-Tree, BVH etc., then freed

// This is distinct from Triangles_Count. The latter will be 0, after the Triangles array is freed.
// This one is needed to let the per-thread Accelerator know how many triangle mail-boxes to allocate.
//
	EIndex           TotalNumOfTriangles;

	E3dBBox          BBox;

	E3dUniformGrid   UniformGrid;
	E3dKDTree*       KDTree;
	E3dBVH*          BVH;

	E3dRayStats      RayStats;  // Debugging / profiling
} E3dAccelerator;


// Per-thread (thread-writeable) version of E3dAccelerator
//
typedef struct
{
	E3dUniformGrid        UniformGrid;
	E3dKDTree*            KDTree;
	E3dBVH*               BVH;

// Thread-writeable members
//
	E3dCoordinate         EntryT, ExitT;  // Temporary data: ray T at entering the scene and at exiting
	E3dRTriangleCore*     TestRayLastHit;
	EList(, E3dRTriangleMailbox, TriangleMailboxes);
	EIndex                RayCounter;

	E3dRayStats           RayStats;       // Debugging / profiling
} E3dRWAccelerator;


//
// Data writable by individual rendering threads
//
#define E3dRayTracer_thread_core\
	E3dRenderer_thread_core;\
\
	E3dStackRay*               RayStack;\
	E3dStackRay                Ray;\
	E3dLightTriangleCDFEntry*  LightTriangleCDF;\
\
	E3dRWAccelerator           Accelerator, ShadowAccelerator;\
\
	const E3dRTriangle*        FirstHitRTriangle;\
	E3dCoord2                  FirstHitUV;\
	E3dCoordinate              FirstHitT;\
	E3dSamplingCache           SamplingCache

struct E3dRayTracer_thread_S
{
	E3dRayTracer_thread_core;
};


#ifdef USEOpenGL
#define E3dRayTracerGLFields\
	E3dGPUFrameBuffer*  GPUFloatFrameBuffer;      /* Used for HDR tonemapping on the GPU, SSAA etc. */\
	E3dGPUFrameBuffer*  GPUFrameBuffer            /* Used for HDR tonemapping on the GPU, SSAA etc. */
#else
#define E3dRayTracerGLFields
#endif


//
// Ray-tracer
//
#define E3dRayTracerCore\
	E3dRendererCore;                        /* It's a subclass of E3dRenderer  */\
\
	int              NSubPixels;            /* Subpixel samples for renderer-native anti-aliasing    */\
\
/* For anti-aliasing, path-tracing etc. */\
\
	E3dCoord2*         SubPixelSamples;EIndex  NumOfSubPixelSamplesPerTile;\
	int                NumOfSubPixelSampleTiles;\
\
	E3dShadePointFunc  ShadePoint;\
\
	int                MaxRayDepth;\
	float              RayLumaMin;\
	E3dCoordinate      LightmapRayOffset;\
\
	EInt32_3           GridCount;\
	int                VoxelLookAhead;\
	EBool              ShowVoxelGrid, ShowKDTree,  ShowBVHInternal, ShowBVHLeaf;\
\
	EBool              VolumeRendering;\
	int                RayMarchSteps;      /* Starting (minimum) steps (adaptive)                    */\
	int                RayMarchSubdivMax;  /* Max. adaptive subdivision depth for ray-march steps    */\
	float              RayMarchThreshold;  /* Min. difference between intensities of line-integral steps that causes a split  */\
\
	E3dAccelerator*    Accelerator;\
	E3dAccelerator*    ShadowAccelerator;\
\
	E3dGPUAA           OverlayAA;          /* Anti-aliasing for 3D overlays (wireframe, Vertices, selected Polygons etc.) */\
\
	E3dRayTracerGLFields

struct E3dRayTracer_S
{
	E3dRayTracerCore;
};



#define E3dPathTracer_core\
	E3dRayTracerCore;\
\
/* Resources  */\
\
	int               NumOfPathSamples;     /* If > 0, use path-tracing with this many rays per pixel  */\
	EBool             DirectIllumination;\
	int               MaxIndirectRayDepth;  /* This is separate from the classical ray-tracer's MaxRayDepth. For example lots of glass may require a large MaxRayDepth, but we don't want a lot of path-tracer bounces.  */\
\
\
	EIndex            NumOfPathTracerLightsPerList, PathTracerLightLists_Count


typedef struct
{
	E3dPathTracer_core;
} E3dPathTracer;



#define E3dPathTracer_thread_core\
	E3dRayTracer_thread_core;\
/* Pre-allocated lists for ShaderLights, used in the path-tracer	*/\
/* This helps to avoid billions of malloc/frees while computing indirect illumination.	*/\
\
	E3dShaderLight**  PathTracerLightLists;EIndex  PathTracerLightLists_Count, NumOfPathTracerLightsPerList;\
	E3dIRay*          PathRays

typedef struct
{
 E3dPathTracer_thread_core;
} E3dPathTracer_thread;




typedef const E3dRTriangle*  (*E3dRayIntersectProc)(E3dRWAccelerator* LPccelerator, E3dGridRay* PRay, E3dCoord2* PUVRet, const void* PExcludeFace);



#define E3dAccKDTREE    0x01
#define E3dAccBVH       0x02
#define E3dUNIFORMGRID  0x04

#define  E3dRAYTRACER_ACCELERATOR  E3dAccKDTREE

#if E3dRAYTRACER_ACCELERATOR == E3dAccKDTREE
	#define E3dAccelerator_IsValid(m)   (m->KDTree != NULL)
	#define E3d_RayIntersectScene       E3d_RayIntersectKDTree
	#define E3d_RayIntersectSceneFront  E3d_RayIntersectKDTreeBackFaceCull
#else
	#if E3dRAYTRACER_ACCELERATOR == E3dAccBVH
		#define E3dAccelerator_IsValid(m)   (m->BVH != NULL)
		#define E3d_RayIntersectScene       E3d_RayIntersectBVH
		#define E3d_RayIntersectSceneFront  E3d_RayIntersectBVH
	#else
		#define E3dAccelerator_IsValid(m)   (m->UniformGrid.Voxels != NULL)
		#define E3d_RayIntersectScene       E3d_RayIntersectVoxels
		#define E3d_RayIntersectSceneFront  E3d_RayIntersectVoxelsBackFaceCull
	#endif
#endif



typedef struct
{
	E3dObjectComponentCore;

	EList(, E3dRPolyGroup*, RPolyGroups);
	E3dAccelerator*  Accelerator;

	E3dModel*        Model;         // If not NULL, the transforms of this Model are applied (to RPolyGroups), otherwise everything is in Object coordinates
	uint16_t         ContentFlags;  // The flags used when E3dObject_GetRenderTriangles() was called (E3dGrtINVERTED etc.)
} E3dAcceleratorObjectComponent;


#define E3d_SnapToSurfaceObjectGetRenderTrianglesContentFlags  (E3dGrtDO_INVISIBLE|E3dGrtDO_WIREFRAME)
extern E3dObjectComponentClass*                                E3dAcceleratorObjectComponent_class;


extern void  E3dAccelerators_Init(void);



//------------------------------------------------
// Uniform grids
//------------------------------------------------

//
// Voxel-walking 3D-DDA state.
//
typedef struct
{
	EInt32_3   GridPosition;
	EInt32_3   GridStep,
	           GridOut;
	E3dCoord3  GridNext,
	           GridDelta;
} E3dRayGridState;



static inline EBool E3dAccelerator_Valid(E3dAccelerator* PAccelerator)
{
	if(PAccelerator == NULL || PAccelerator->UniformGrid.Voxels || PAccelerator->BVH || PAccelerator->KDTree) return(TRUE);
	else return(FALSE);
}

//----------------------------------------
// Rendering/Raytracing.c
//----------------------------------------
extern E3dMaterial*  E3d_RendererMaterialFindByName(E3dRenderer* PRenderer, const char* PName);


//----------------------------------------
// Panel/Renderer.c
//----------------------------------------
extern void          E3d_TrayTracerInitThreads(E3dWindow* P3DWindow, E3dRayTracer* PRenderer, E3dScene* PScene, const int PNumOfThreads);
extern void          E3dRayTracer_UpdateAASamples(E3dRayTracer* PRenderer);


//----------------------------------------
// Accelerators/Accelerator.c
//----------------------------------------
extern int   E3dAccelerator_ResetForRendering(const E3dAccelerator* PAccelerator, E3dRWAccelerator* PRWAccelerator);
extern void  E3dAccelerator_CopyToThread(E3dAccelerator* PSrcAccelerator, E3dRWAccelerator* PDstAccelerator);

extern void  E3d_RWAcceleratorFreeContents(E3dRWAccelerator* PAccelerator);
extern void  E3dAccelerator_FreeContents(E3dAccelerator* PAccelerator);
extern void  E3d_RayStatsClear(E3dRayStats* PStats);


extern void  E3dAccelerator_CollectPrimitives(E3dAccelerator* PAccelerator,  E3dRPolyGroup* const* PRPolyGroups, const EIndex PRPolyGroups_Count,  const EBool PIncludeNonShadowCasters);
extern void  E3dAccelerator_Build(E3dAccelerator* PAccelerator,  E3dRPolyGroup* const* PRPolyGroups, const EIndex PRPolyGroups_Count,  const uint16_t PAcceleratorTypes);

//----------------------------------------
// Accelerators/Voxel.c
//----------------------------------------
extern void  E3d_RayDDAInit(const E3dRWAccelerator* PAccelerator, const E3dRay* PRay, E3dRayGridState* PState);
extern void  E3d_RayDDAInitStartVoxel(const E3dRWAccelerator* PAccelerator, const E3dRay* PRay, E3dRayGridState* PState);
extern void  E3d_RayDDAInitStartSet(const E3dRWAccelerator* PAccelerator, const E3dRay* PRay, E3dRayGridState* PState);

extern void                 E3dAccelerator_BuildUniformGrid(E3dRayTracer* PRenderer, E3dAccelerator* PAccelerator);

extern const E3dRTriangle*  E3d_RayIntersectVoxels(E3dRWAccelerator* PAccelerator, E3dGridRay* PRay, E3dCoord2* PUVRet, const void* PExcludeFace);
extern const E3dRTriangle*  E3d_RayIntersectVoxelsBackFaceCull(E3dRWAccelerator* PAccelerator, E3dGridRay* PRay, E3dCoord2* PUVRet, const void* PExcludeFace);

extern EBool                E3d_ShadowRayIntersectVoxels(E3dRWAccelerator* PAccelerator, E3dGridRay* PRay, const void* PExcludeFace, const E3dCoordinate PLightDistance);
extern EBool                E3d_ShadowRayIntersectVoxelsStartSet(E3dRWAccelerator* PAccelerator, E3dGridRay* PRay, E3dRayGridState* PState, const void* PExcludeFace, const E3dCoordinate PDistance);;

//----------------------------------------
// Accelerators/BVH.c
//----------------------------------------
extern void  E3dAccelerator_BuildBVH(E3dAccelerator* PAccelerator);
extern void  E3dBVH_Free(E3dBVH* PBVH);
extern void  E3dBVH_Visualize(E3dScene* PScene, E3dRayTracer* PRenderer);

extern const E3dRTriangle*  E3d_RayIntersectBVH(E3dRWAccelerator* PAccelerator, E3dGridRay* PRay, E3dCoord2* PUVRet, const void* PExcludeFace);
extern EBool                E3d_ShadowRayIntersectBVH(E3dRWAccelerator* PAccelerator, E3dGridRay* PRay, const void* PExcludeFace, const E3dCoordinate PMaxDistance);

//----------------------------------------
// Accelerators/KDTree.c
//----------------------------------------
extern E3dKDTree*  E3dKDTree_Build(const void** PPrimitives, const EIndex PPrimitives_Count, const E3dBBox* PAllBounds, const E3dBBox PTreeBBox);
extern void        E3dAccelerator_BuildKDTree(E3dAccelerator* PAccelerator);
extern void        E3dKDTree_Free(E3dKDTree* PTree);

extern void        E3dKDTree_Visualize(E3dScene* PScene, E3dKDTree* PTree);
extern void        E3dKDTree_SaveASCII(E3dKDTree* PKDTree, const char* PFileName);

extern const E3dRTriangle*  E3d_RayIntersectKDTree(E3dRWAccelerator* PAccelerator, E3dGridRay* PRay, E3dCoord2* PUVRet, const void* PExcludeFace);
extern const E3dRTriangle*  E3d_RayIntersectKDTreeBackFaceCull(E3dRWAccelerator* PAccelerator, E3dGridRay* PRay, E3dCoord2* PUVRet, const void* PExcludeFace);
extern EBool                E3d_ShadowRayIntersectKDTree(E3dRWAccelerator* PAccelerator, E3dGridRay* PRay, const void* PExcludeFace, const E3dCoordinate PDistance);


//----------------------------------------
// Panel/Renderer.c
//----------------------------------------
extern void          E3d_TrayTracerInitThreads(E3dWindow* P3DWindow, E3dRayTracer* PRenderer, E3dScene* PScene, const int PNumOfThreads);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif


#endif  // _E3D_Raytracing_h
