//================================================================
// E3D
//
// Rendering-related header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 19:14:16
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Rendering_h
#define _E3D_Rendering_h

#include <assert.h>
#include <unistd.h>
#include <stdint.h>    // For uint32_t etc. for E_Random

#ifndef _EClass_h
#include <EClass.h>
#endif

#ifndef _EThreads_h
#include <EThreads.h>
#endif

#ifndef _E3D_Node_h
#include <E3D/Node.h>
#endif

#ifndef _E3D_Camera_h
#include <E3D/Camera.h>
#endif

#ifndef _E3D_SL_h
#include <E3D/SL.h>
#endif

#ifndef _E3D_Window_h
#include <E3D/UI/Window.h>
#endif

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#define E3dRenderDEBUG  1

// Default SSAA subpixel count (4x means 2x resolution off-screen buffer in X and Y directions)
#ifdef E3dGL_HAS_FBO
#define E3d_DefaultSSAANSamples  4
#else
#define E3d_DefaultSSAANSamples  1
#endif




//------------------------------------------------
// Fast nearest neighbour query with KD-trees
//------------------------------------------------
#define E3dKDTreePointCore\
	union { E3dCoord3	Position;E3dCoordinate	Position3[3]; };\
	unsigned short  Plane        /* For the kd-tree  */

typedef struct
{
	E3dKDTreePointCore;
} E3dKDTreePoint;

typedef struct _E3dNearestKDTreePointList
{
	const E3dKDTreePoint**  Points;EIndex  Points_Count;  // This is an array of pointers, so we don't need to specialize to RGBPhoton or SpectralPhoton
	float*                  DistanceSq;                     // Squared distance array
	EIndex                  HeapNMax;                       // Heap maximum size
	E3dCoordinate           QueryPosition[3];
	short                   HaveHeap;                       // Is the heap built?
} E3dNearestKDTreePointList;


#define E3dKDTREE_POINTMAP_MAX_THREAD_DEPTH  5

typedef struct E3dKDTreePointMap_S  E3dKDTreePointMap;

typedef struct
{
	int   NextThreadIndex;
} E3dKDTreePointMapBalancing;

typedef struct
{
	EThreadCore;

	E3dKDTreePointMap*  PointMap;
	E3dBBox             BBox;
	E3dKDTreePoint**    PBal;
	E3dKDTreePoint**    POrg;
	int                 Index;
	int                 Start, End;
} E3dKDPtreePointMapBalancingThread;


#define E3dKDTreePointMapCore\
	E3dKDTreePoint*  Points;       /* Actual point array used after _PhotonmapBalance() call                                    */\
	E3dKDTreePoint*  TmpPoints;    /* Temporary point array if _PhotonmapBalance() is called, this array is freed and not used  */\
\
	EIndex      Points_Count, Points_CountAllocated;  /* Number of points                           */\
	int         BytesPerPoint;\
	E3dBBox     BBox;                                 /* Bounding box                               */\
	ETime       BalancingStartMicroseconds;\
	E3dKDTreePointMapBalancing  Balancing;\
	EThreadSet  ThreadSet;                            /* For balancing the KDTree                   */\
	EBool       Balanced;                             /* Is KDTreePoints array a balanced kd-tree?  */

struct E3dKDTreePointMap_S
{
	E3dKDTreePointMapCore;
};



// Points arranged in a KD-tree, for fast nearest-neighbour queries
//
extern E3dKDTreePointMap*          E3d_KDTreePointMapInit(E3dKDTreePointMap* PPointMap);
extern E3dKDTreePointMap*          E3d_KDTreePointMapNew(const int PStructSize);
extern void                        E3d_KDTreePointMapFreeContents(E3dKDTreePointMap* PPointMap);

extern E3dNearestKDTreePointList*  E3d_NearestKDTreePointsNew(const int PMaxCount);
extern void                        E3d_NearestKDTreePointsFree(E3dNearestKDTreePointList* PNearestPhotons);

extern EBool                       E3d_KDTreePointMapBuildTree(E3dKDTreePointMap* PPointMap, void (*PAbortCallbackFunc)(EPointer PClientData), const EPointer PAbortCallbackClientData, const EBool PShowProgressBar);

extern void                        E3d_NearestPointsJensen_iterative(const E3dKDTreePointMap* PPointMap, E3dNearestKDTreePointList* PNearestPoints);




#define _E3dM_Modulate3(mColor, mFactor, mChannelMask, mClampFunc) do\
{\
	float*  lColor = (float*)(mColor);\
	switch(mChannelMask)\
	{\
		case EpRED:\
			{\
				const float  lV = mFactor * LTextureColor.R;\
				lColor[0] += lV;\
				lColor[1] += lV;\
				lColor[2] += lV;\
\
				mClampFunc\
			}\
		break;\
\
		case EpGREEN:\
			{\
				const float  lV = mFactor * LTextureColor.G;\
				lColor[0] += lV;\
				lColor[1] += lV;\
				lColor[2] += lV;\
\
				mClampFunc\
			}\
		break;\
\
		case EpBLUE:\
			{\
				const float  lV = mFactor * LTextureColor.B;\
				lColor[0] += lV;\
				lColor[1] += lV;\
				lColor[2] += lV;\
\
				mClampFunc\
			}\
		break;\
\
		case EpALPHA:\
			{\
				const float  lV = mFactor * LTextureColor.A;\
				lColor[0] += lV;\
				lColor[1] += lV;\
				lColor[2] += lV;\
\
				mClampFunc\
			}\
		break;\
\
		case EpRED | EpGREEN | EpBLUE:\
			{\
				const float  lV = mFactor;\
				lColor[0] += LTextureColor.R * lV;\
				lColor[1] += LTextureColor.G * lV;\
				lColor[2] += LTextureColor.B * lV;\
\
				mClampFunc\
			}\
		break;\
\
		case EpRED | EpGREEN | EpBLUE | EpALPHA:\
			{\
				const float  lV = mFactor * LTextureColor.A;\
				lColor[0] += LTextureColor.R * lV;\
				lColor[1] += LTextureColor.G * lV;\
				lColor[2] += LTextureColor.B * lV;\
\
				mClampFunc\
			}\
		break;\
\
		case EpINTENSITY:\
			{\
				const float  lV = mFactor * (LTextureColor.R + LTextureColor.G + LTextureColor.B) * 0.333333;\
				lColor[0] += lV;\
				lColor[1] += lV;\
				lColor[2] += lV;\
\
				mClampFunc\
			}\
		break;\
	}\
} while(0)


//================================================================================================
// Add Emission (direct radiance) of area light, when directly looking into an area light.
//
// Emission colors are in radiance (not total power of light).
// Also, luckily, we only need to compute radiant exitance (radiosity) because:
// - the cosine term due to the viewing angle is cancelled out by looking at a larger area at lower angles.
// - the distance inverse-square term (due to the solid angle) is cancelled out by the same solid angle looking at a larger area of the area light from farther away.
//================================================================================================
static inline void _E3d_AddEmission(const E3dSurfacePoint* PPoint, const E3dMaterial* PMaterial, const float* PEmission,  const int PSpectralIndex, const int PNSpectralSamples,  float* PColor)
{
	if(PMaterial->Light.CutoffAngle == 0.0 || E3d_Dot3_vv(PPoint->Normal, PPoint->PointToViewer) > PMaterial->Light.CutoffAngleCos)
	{
		E3dM_SpectralLoop(PSpectralIndex, PNSpectralSamples) PColor[LSI] += PEmission[LSI];
	}
}


//================================================
// Beer-Lambert law of absorption
//================================================
#define _E3dM_Extinction(mSpectralIndex, mNumOfSpectralSamples,  mVolumeShader, mDistance, mColor) do\
{\
	const E3dShader*  lVolumeShader = mVolumeShader;\
	if(lVolumeShader && lVolumeShader->HasVolumeAbsorption)\
	{\
		const E3dCoordinate  lDistance = mDistance;\
		const float*         lExtinction = lVolumeShader->Volume.Extinction.Samples;\
\
		E3dM_SpectralLoop(mSpectralIndex, mNumOfSpectralSamples)  mColor[LSI] *= E3d_Exp(-lDistance * lExtinction[LSI]);\
	}\
} while(0)


#define _E3d_MIPMap(mRay, mRTriangle, m) do\
{\
	LMipRay.Direction = mRay->PixelCornerDirections[m];E3d_RayIntersectTriangle_UVKnownHit(&LMipRay, mRTriangle, &LUVs[m]);\
\
	E3dAnisoPoint*  lAnisoPoint = LAnisoPoints + m;\
	E3dCoord3*      lPosition = &lAnisoPoint->Position;\
/* Set hit position, for procedural texture-mappers	*/\
	{\
		const E3dCoordinate	lMT = LMipRay.LastT;\
		E3d_AddScaled3_vvp(LMipRay.Origin, LMipRay.Direction, lMT,  lPosition);\
	}\
} while(0)


//==========================================================================================
// Normalize texture coordinates (convert to 0.0...1.0 range) by applying tiling etc.
//==========================================================================================
#define _M_TexcoordsNormalize(mMapper) do\
{\
	const E3dCoordinate  LSRepeatsF = (E3dCoordinate)(mMapper->SCount), LTRepeatsF = (E3dCoordinate)(mMapper->TCount);\
\
	if((mMapper->Class == E3dTextureMapperFreehand_class) || ((LS >= 0.0f) && (LT >= 0.0f) && (LS <= LSRepeatsF) && (LT <= LTRepeatsF)))\
/*	if(0)*/\
	{\
printf("%s() %s:%d  _M_TexcoordsNormalize()\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout);\
		if(LS >= 0.0) LS -= E3d_Floor(LS);  /* Modulo...	*/\
		else LS += (1.0 - E3d_Ceil(LS));\
\
		if(LT >= 0.0) LT -= E3d_Floor(LT);  /* Modulo...	*/\
		else LT += (1.0 - E3d_Ceil(LT));\
	}\
	else LS = E3dST_OFF_TEXTURE;\
} while(0)


#define E3d_Span4_vv(mS, mm, mD) do\
{\
	E3dCoordinate  lSpan;\
\
	mD.mm = E3d_Abs(mS[0].mm - mS[1].mm);\
	lSpan = E3d_Abs(mS[0].mm - mS[3].mm);if(mD.mm < lSpan) mD.mm = lSpan;\
	lSpan = E3d_Abs(mS[0].mm - mS[2].mm);if(mD.mm < lSpan) mD.mm = lSpan;\
	lSpan = E3d_Abs(mS[1].mm - mS[2].mm);if(mD.mm < lSpan) mD.mm = lSpan;\
	lSpan = E3d_Abs(mS[1].mm - mS[3].mm);if(mD.mm < lSpan) mD.mm = lSpan;\
	lSpan = E3d_Abs(mS[2].mm - mS[3].mm);if(mD.mm < lSpan) mD.mm = lSpan;\
} while(0)


// Temporary solution to RendererClass re-activation...
// We'll always allocate this size, regardless of StructSize, so if we need to deactivate/reactivate
// a Class and its Renderers and the new StructSize is greater, we don't need to go throught all the
// references of Renderers.
// Eventually we'll do this though
//
#define E3dRENDERER_MAX_SIZE  262144


#define E3dAE_OFF            0x00
#define E3dAE_SPOT_METERING  0x01


// Anti-aliasing
//
enum
{
	E3dAA_NONE = 0,
	E3dAA_02x,
	E3dAA_04x,
	E3dAA_06x,
	E3dAA_09x,
	E3dAA_02xR,
	E3dAA_03xR,
	E3dAA_04xR,
	E3dAA_05xR,
	E3dAA_07xR,
	E3dAA_09xR,
	E3dAA_16xR,
	E3dAA_25xR,
	E3dAA_36xR,
	E3dAA_49xR,
	E3dAA_64xR,
	E3dAA_ADAPTIVE_2x2,
	E3dAA_ADAPTIVE_3x3,
	E3dAA_ADAPTIVE_4x4
};


enum
{
	E3dUNIFORM = 1,
	E3dSTRATIFIED
};


#ifndef _E3dScene
#define _E3dScene
typedef struct E3dScene_S  E3dScene;
#endif


#ifndef _E3dRenderer
#define _E3dRenderer
typedef struct E3dRenderer_S         E3dRenderer;
typedef struct E3dRenderer_thread_S  E3dRenderer_thread;
#endif

#ifndef _E3dRayTracer
#define _E3dRayTracer
typedef struct E3dRayTracer_S         E3dRayTracer;
typedef struct E3dRayTracer_thread_S  E3dRayTracer_thread;
#endif

typedef struct E3dRendererSet_S  E3dRendererSet;


enum
{
	E3dGET = 1,
	E3dSET
};

enum
{
	E3dSHADER_ID = 1,
	E3dVERTEX_PROGRAM_RESOURCES,
	E3dVERTEX_PROGRAM_SOURCE,
	E3dFRAGMENT_PROGRAM_RESOURCES,
	E3dFRAGMENT_PROGRAM_SOURCE,
};



#define E3dPhotonCore\
	E3dKDTreePointCore;\
	char            Phi, Theta  /* quantized angle  */

typedef struct
{
	E3dPhotonCore;
} E3dCorePhoton;


typedef struct
{
	E3dPhotonCore;

	float  PowerRGB[3];
} E3dRGBPhoton;

typedef struct
{
	E3dPhotonCore;

	float           Power;
	unsigned short  SpectralIndex;
} E3dSpectralPhoton;


#ifndef _E3dLightInstance
#define _E3dLightInstance
typedef struct E3dLightInstance_S  E3dLightInstance;
#endif


// Renderers
//
typedef int    (*E3dRenderFunc)(E3dWindow* P3DWindow, E3dRenderer* PRenderer, E3dScene* PScene,  E3dCamera* PCamera, E3dModel* PCameraModel);
typedef int    (*E3dRendererInfoFunc)(E3dRenderer* PRenderer, E3dGeometry* PGeometry, E3dShadeable* PShadeable, int, ...);
typedef EBool  (*E3dRendererUpdateFunc)(E3dRenderer*, E3dRenderer*);

typedef int    (*E3dTraceRayFunc)(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, E3dStackRay* PRay, int PRayIter, E3dShaderResult* PResult);

typedef void   (*E3dShadePointFunc)(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dRTriangle* PRTriangle, E3dSurfacePoint* PPoint, E3dStackRay* PRay,  const E3dMaterial* PSurfaceMaterial, const E3dShader* PVolumeShader, const E3dShader* PShaderBeyondSurface,  E3dShaderResult* PResultTmp, E3dShaderResult* PResult);

typedef void*  (*E3dBindMaterialFunc)(E3dGPUContext* PGPUContext,  E3dMaterial* PMaterial, E3dScene* PScene, E3dRenderer* PRenderer, E3dModel* PModel, E3dGeometry* PGeometry, E3dShadeable* PShadeable,  const E3dLightInstance* PLightInstances, const int PNLightInstances);
typedef void   (*E3dUnbindMaterialFunc)(E3dGPUContext* PGPUContext,  E3dMaterial* PMaterial, void* PHwShader, E3dRenderer* PRenderer);


typedef struct E3dRendererClass_S  E3dRendererClass;



typedef void  (*E3dShaderClassRendererProc)(E3dShaderClass*, E3dShaderClassRendererBinding*);

#define E3dRendererClassCore\
	EBaseClassCore(E3dRenderer);\
	void				(*ThreadInit)(const E3dRenderer* PRenderer, E3dRenderer_thread* PThread);	/* Thread clone-constructor, used by multi-threaded rendering	*/\
	void				(*ThreadFree)(E3dRenderer_thread* PThread);	/* Thread clone destructor, used by multi-threaded rendering	*/\
\
	void				(*FreeCaches)(E3dRenderer*);        /* Free cached data (accelerators, triangles etc.)           */\
\
	E3dRendererUpdateFunc       Update;             /* Called when Resources change                              */\
	E3dRendererInfoFunc         Info;               /* Get source code of a shader for this renderer class etc.  */\
	E3dRenderFunc               Render;\
	E3dBindMaterialFunc         BindMaterial;\
	E3dUnbindMaterialFunc       UnbindMaterial;\
	ECoreObjectGUIFunc          GUI;                /* Update/free GUI panel for renderer instance               */\
	unsigned int                Flags;\
	unsigned int                ShaderClassRendererBindingSize;\
	unsigned int                ShaderVariantSize;  /* Size of structure representing a shader variant for this renderer class  */\
	E3dShaderClassRendererProc  ShaderClassFree;    /* Called when a ShaderClass is freed                                       */\
	E3dShaderDevice*            ShaderDevice

struct E3dRendererClass_S
{
 E3dRendererClassCore;
};

typedef struct
{
	EBaseClassCore(E3dRenderer);

	void             (*ThreadInit)(const E3dRenderer* PRenderer, E3dRenderer_thread* PRenderer_thread);  /* Thread clone-constructor, used by multi-threaded rendering  */\
	void             (*ThreadCopy)(const E3dRenderer* PSrcRenderer, E3dRenderer* PDstRenderer);
	void             (*ThreadFree)(E3dRenderer_thread* PThread);
	void             (*FreeCaches)(E3dRenderer* PRenderer);
	E3dRendererUpdateFunc       Update;
	E3dRenderFunc               Render;
	E3dBindMaterialFunc		      BindMaterial;
	E3dUnbindMaterialFunc       UnbindMaterial;
	ECoreObjectGUIFunc          GUI;
	unsigned int                Flags;
	E3dShaderClassRendererProc  ShaderClassFree;
} E3dRendererClassDescriptor;





// Renderer flags
//
#define E3dPROGRESSIVE_REFINEMENT  0x0008
#define E3dHDR_IMAGE               0x0010



// Display flags for 3DWindows
// Display-mode flags
//
#define E3dDF_NONE                  0x00000000
#define E3dDF_WIREFRAME_UNSELECTED  0x00000001  // Draw Edges of unselected objects
#define E3dDF_WIREFRAME_SELECTED    0x00000002  // Draw Edges of selected objects
#define E3dDF_SOLID                 0x00000010  // Draw surfaces
#define E3dDF_Z_BUFFER              0x00000100  // Use Z-buffer

#define E3dDF_COLOR                 0x00000400  // Color surfaces flat
#define E3dDF_LIGHTING              0x00000800  // Use lighting

#define E3dDF_CONTROLS              0x00002000  // Drawing controls (we are after tone-mapping, so call GLDrawControls)

#define E3dDF_MODEMASK              0x00003FFF



// Display modes for type-selective window redrawing and picking
//
#define E3dDM_Wireframe           (E3dDF_WIREFRAME_UNSELECTED | E3dDF_WIREFRAME_SELECTED)
#define E3dDM_WireAndZ            (E3dDM_Wireframe | E3dDF_Z_BUFFER)
#define E3dDM_HiddenLine          (E3dDF_SOLID | E3dDM_Wireframe | E3dDF_Z_BUFFER)
#define E3dDM_ColorWire           (E3dDF_WIREFRAME_SELECTED | E3dDF_Z_BUFFER | E3dDF_COLOR | E3dDF_Z_BUFFER)
#define E3dDM_ShadedWire          (E3dDF_WIREFRAME_SELECTED | E3dDF_Z_BUFFER | E3dDF_LIGHTING | E3dDF_Z_BUFFER)
#define E3dDM_Solid               (E3dDF_WIREFRAME_SELECTED | E3dDF_SOLID | E3dDF_Z_BUFFER)
#define E3dDM_SolidAndWire        (E3dDM_Wireframe | E3dDF_SOLID | E3dDF_Z_BUFFER)
#define E3dDM_ColorSolid          (E3dDF_WIREFRAME_SELECTED | E3dDF_SOLID | E3dDF_COLOR | E3dDF_COLOR | E3dDF_Z_BUFFER)
#define E3dDM_ColorSolidAndWire   (E3dDM_Wireframe | E3dDF_SOLID | E3dDF_COLOR | E3dDF_Z_BUFFER)
#define E3dDM_ShadedSolid         (E3dDF_WIREFRAME_SELECTED | E3dDF_SOLID | E3dDF_LIGHTING | E3dDF_Z_BUFFER)
#define E3dDM_ShadedSolidAndWire  (E3dDM_Wireframe | E3dDF_SOLID | E3dDF_LIGHTING | E3dDF_Z_BUFFER)


// Display flags for 3D windows
// Node flags
//
#define E3dDF_SELECTED   0x01000000
#define E3dDF_LOCKED     0x02000000
#define E3dDF_PICK       0x04000000

#define E3dDF_NODEFLAGS  0x07000000


#define E3dDF_ALL  (E3dDF_MODEMASK|E3dDF_NODEFLAGS)



enum
{
	E3dAreaLightsNONE = 0,
	E3dAreaLightsHEMISPHERE,
	E3dAreaLightsSAMPLE_AT_VTX,
	E3dAreaLightsSAMPLE,
	E3dAreaLightsSAMPLE_FOR_PHOTONS
};


enum
{
	E3dSampleRANDOM = 1,
	E3dSampleHALTON
};


typedef struct
{
	EInt32_2	Size;
	EInt32_2	RegionStart, RegionSize;
} E3dImageSpecs;




#define E3dLOD_MULT	(1.0 / 200.0)

// Renderer states
//
enum
{
	E3dIDLE = 0,
	E3dPREPARING_FRAME,
	E3dRENDERING,
	E3dPAUSED,    // Incomplete frame, Renderer stopped, but able to resume
	E3dSTOPPED    // Incomplete frame, Renderer stopped, not able to resume (but may not have to rebuild acceleration structures etc.)
};



// E3dLights can be instantiated from Models and
// this struct is used to store unique instances with spectrally sampled colors (like in E3dScene)
//
typedef struct 
{
	E3dCoord3  Position;                              // World position

	int        Type;                                  // Light source type
	float      Power[EcMAX_NUM_OF_SPECTRAL_SAMPLES];  // Spectral color of the light source

	float      SpotAngle, SpotPenumbraAngle, SpotCutoffAngleCos, SpotExponent,
	           ConstantAttenuation,
	           LinearAttenuation,
	           QuadraticAttenuation,
	           ShadowBrightness;       // If a point is in shadow from this light, its diffuse color will be multiplied by this value
} E3dRLightInstance;


typedef struct
{
	E3dCoord3  Position;
} E3dAnisoPoint;

typedef struct
{
	E3dCoord2      UVs[E3dMAX_TEXTURE_ANISOTROPY];
	E3dAnisoPoint  Points[E3dMAX_TEXTURE_ANISOTROPY];
	E3dST          STs[E3dMAX_TEXTURE_ANISOTROPY];
} E3dAniso;


enum
{
	E3dRADIANCE = 1,
	E3dIRRADIANCE_DIFFUSE,
	E3dIRRADIANCE_SHI,     // Spherical harmonics irradiance map
	E3dIRRADIANCE_SHR,     // Spherical harmonics radiance map / light field
	E3dIRRADIANCE_SHO,     // Spherical harmonics occlusion map
	E3dIRRADIANCE_WAVELET
};


typedef struct
{
	int      NumOfSubFrames;    // Render this many sub-frames and average them, e.g. for motion blur
	uint8_t  StartPixelSize;    // 2, 4, 8, 16 etc. for progressive, pixelized rendering
	uint8_t  PixelMeaning;      // E3dRADIANCE (normal rendering), E3dIRRADIANCE_DIFFUSE, E3dIRRADIANCE_SHI etc.
} E3dRendererSettables;


typedef struct
{
	E3dCamera*    Camera;               // Currently used Camera. This is a COPY (not a reference) of the active Camera, to make sure
	                                    // that it does not change while a frame is being rendered, even in interactive/multi-threaded mode.
	E3dModel*     CameraModel;          // The Model that instantiates the Camera used - COPY

	EChangeFlags  SceneChangeFlags;     // What changed in the Scene
	EChangeFlags  RendererChangeFlags;  // What changed in this Renderer
} E3dRenderQueueEntry;



typedef struct
{
	float     Radius, Power, Strength;  // User-settable values

	float     Work_Power;               // Power that matches the kernel (kernel needs to be rebuilt if this changes)
	float     Work_Strength;            // Strength that matches the kernel (kernel needs to be rebuilt if this changes)
#ifdef USEOpenGL
	GLuint    GLKernelTexture;
	EInt32_2  GLFBOSize;
#endif
} E3dBloom;

extern EArg  E3dBloomUI_Radius[];
extern EArg  E3dBloomUI_Strength[];
extern EArg  E3dBloomUI_Power[];


#define E3dMAX_REFINEMENT_PASSES  8   // Supports starting rendering from 256x256 macro-pixels


//--------------------------------
// Lightmap rendering
//--------------------------------
typedef struct
{
	E3dTextureMapper*  TextureMapper;  /* Current texture-mapper and its:   */\
	E3dModel*          Model;          /* ... Model, ...                    */\
	E3dGeometry*       Geometry;       /* ... Geometry and ...              */\
	EObject*           Shadeable;      /* ... Shadeable (e.g. PolyGroup) for rendering light-maps  */\
} E3dLightMap;


#define E3dDrawNODE           0x01
#define E3dDrawNODE_CUSTOM    0x02
#define E3dDrawNODE_CONTROLS  0x04

typedef struct
{
	E3dNode*           Node;
	E3dNodeComponent*  NodeComponent;
	uint32_t           DrawFlags;  // E3dDrawNODE etc.
} E3dDrawItem;


typedef struct
{
	uint16_t  MSAA;                 // Number of MSAA samples
	uint16_t  SSAA;                 // Supersampling / full-screen antialiasing (render at higher res, filter down), offscreen render target resolution subpixels. E.g. 4 means 2x resolution in both directions
	EBool     LineAA, PolygonAA;    // Traditional, blending-based line and polygon AA (e.g. on SGI)
} E3dGPUAA;


#define E3dRendererCore\
	EObjectCore(E3dRenderer);\
\
	E3dCamera*  Camera;       /* Currently used Camera. This must be a COPY, not a reference to an  */\
	                          /* active Camera to make sure that it does not change while a frame   */\
	                          /* is being rendered, even in interactive/multi-threaded mode.        */\
	E3dModel*   CameraModel;  /* The Model that instantiates the Camera used                        */\
\
\
	E3dRendererSettables  User, Work;        /* "User" contains fields settable by the user. Those are copied over to "Work" when a frame starts. This is used for settings that could not be safely changed during rendering, and those that we want to reset after the frame started (e.g. we don't want "PixelMeaning" to stick when rendering irradiance maps). */\
\
	int       NumOfSpectralSamples;          /* E.g. 3 for RGB                  */\
	int       ColorModel;                    /* EpRGB or EpSPECTRAL             */\
	float     WavelengthMin, WavelengthMax;  /* Wavelength range in nanometers  */\
	float     InvNumOfSpectralSamples;\
	float*    SpectralToR;                   /* RGB weights                     */\
	float*    SpectralToG;\
	float*    SpectralToB;\
	float     RGBtoXYZ[16];                  /* RGB to CIE XYZ Matrix    */\
	float     XYZtoRGB[16];                  /* CIE XYZ to RGB Matrix    */\
\
	E3dMaterial*  CameraIsInThisMaterial;    /* The Camera is inside this Material. Needed for volume effects */\
/* Point / spot etc. lights that are visible and have a non-black color */\
\
	EBool     UsePointLights;\
	EList(, E3dRLightInstance, VisibleLights);\
\
	EIndex       FrameCnt;            /* Frames rendered since this Renderer was created.        */\
	EIndex       WorkingFrameCnt;     /* Frame currently being rendered (FrameCnt may be ahead)  */\
\
	E3dRendererSet*    RendererSet;\
	E3dRendererClass*  PreVizRendererClass;  /* E.g. wireframe to see the outline for long renders  */\
\
\
	E3dLightMap      LightMap;\
\
/* Image settings */\
\
	E3dImageSpecs    ImageSpecs;          /* Copied from the RendererSet when starting a render, so changes won't interfere with a running Renderer */\
\
	uint8_t          AntiAliasingType;    /* 0, E3dUNIFORM, E3dSTRATIFIED     */\
	int              NAnisotropicTextureSamples;  /* For anisotropic texture sampling  */\
\
	E3dBloom         Bloom;\
\
	EpToneMapper     HDRToneMapper;\
	EpExposure       Exposure;\
	int              ExposureMode;\
	EBool            CreateHDRImage;             /* Create High Dynamic Range image   */\
	EBool            SaveDepthImage;\
	EBool            SaveShadowImage;\
\
	double           SizeFactor;                 /* atan(FieldOfView / 2) * WindowSizeY, for LODs    */\
\
	EcRGBAf32Color   BackgroundRGBAfColor;\
\
	float            AmbientLightColor[EcMAX_NUM_OF_SPECTRAL_SAMPLES];\
\
\
	EIndex           NumOfExtraShaderLights;     /* Other than AreaLight samples       */\
\
	EList(, E3dRPolyGroup*, RPolyGroups);\
	EListA(, E3dMaterial*, Materials);           /* This lookup table allows us to only duplicate each Material once, for rendering (rather than for every Shadeable*. We identify Materials by their ID field. */\
\
	EList(, E3dAreaLight*, AreaLights);\
	EList(, E3dRTriangle*, LightTriangles);\
\
	uint32_t         UseZBuffer:1, UseSolidSurfaces:1, UseFlatColor:1, UseLighting:1;\
\
	uint8_t          AreaLightType;\
\
	EBool            TexturesOn, ShadowsOn, ReflectionsOn, TransparencyOn, FogOn;\
\
\
/* Internal variables */\
\
	E3dDisplayFlags  HiddenObjectDisplayFlags;\
\
	EIndex_2*        PixelRemappingGrid;             /* Image->Size.X * Size.Y grid with pseudo-randomly remapped pixel coordinates for frameless rendering  */\
	EIndex_2*        PixelRemappingGridProgressive;  /* Image->Size.X * Size.Y grid with pseudo-randomly remapped pixel coordinates for frameless rendering  */\
\
	EpImage*         Image0;          /* Double buffering : the two buffers                         */\
	EpImage*         Image1;\
\
	EpImage*         ImageRend;       /* Double buffering : the image being rendered (back-buffer)  */\
	EpImage*         ImageDraw;       /* Double buffering : the image being shown (front-buffer)    */\
\
	EpImage*         AccumImage;      /* Temporal sub-frames are accumulated into this Image        */\
\
	EpImage*         ShadowImage;\
	EpImage*         ShadowAccumImage;\
\
	EpImage*         DepthImage;\
\
	EChangeFlags     SceneChangeFlags;      /* What changed in the Scene      - last change only    */\
	EChangeFlags     RendererChangeFlags;   /* What changed in this renderer  - last change only    */\
\
	E3dRenderQueueEntry  InactiveQueue;    /* Render queue for caching changes while Renderer is inactive (i.e. the Ray-tracer is inactive when the current renderer is Cg)  */\
	E3dRenderQueueEntry  Queue;            /* Render queue - only one entry (last change)    */\
\
	EIndex               NumOfPixelsComputed, NumOfPixelsDrawn;\
\
	E3dShow              Show;             /* What components to display */\
\
/* Internal variables */\
	int          State;                /* E3dIDLE, E3dPREPARING_FRAME, E3dRENDERING, E3dPAUSED, E3dSTOPPED */\
	EBool        BuildingAccelerator,\
	             GoRender, FirstIteration, ShowProgress,\
	             GoSequenceRender;\
	ETime        StartMilliseconds, FinishMilliseconds,\
	             LastRedrawMilliseconds, LastStatsUpdateMilliseconds;\
	ETime        PrevFrameMicroSeconds, FrameTimeMicroseconds, LastDisplayedFrameTimeMicroseconds;\
\
	ETime        Start_TimeStamp;      /* The last time a frame started. Used to detect whether a change should trigger a re-render. */\
	int          SubFrame;             /* Current sub-frame        */\
	EIndex       NumOfShaderLights;\
\
	EListA(, E3dDrawItem, LastDrawItems);  /* Items to draw after others, possibly in a sorted order (e.g. transparent objects, overdraw skeletons with no Z test,  etc.). Lazily grown array. */\
\
/* Multi-threading */\
\
	EThreadSet       ThreadSet;\
	int              NumOfThreadLocks;\
	pthread_mutex_t  Workload_mutex, Progress_mutex, Exposure_mutex,\
	                 Debug_mutex;      /* Used for printfs etc. */\
\
	EIndex           CurrentPixelSizeR;\
	EIndex           NewPixelsPerPass[E3dMAX_REFINEMENT_PASSES];int  Pass;\
	int              AASubFrame;       /* Current anti-aliasing sub-frame            */\
	EIndex_2         ImageSizeP;       /* Macro pixels, depends on CurrentPixelSizeR */\
	EIndex           NextPixel, PixelsPerPacket, EndPixel;\
	EBool            AreaLightsEnabled

struct E3dRenderer_S
{
	E3dRendererCore;
};



#define E3dGPURendererCore\
	E3dRendererCore;\
	E3dGPUAA  AntiAliasing

typedef struct
{
	E3dGPURendererCore;
} E3dGPURenderer;

extern E3dRendererClass  E3dGPURenderer_class[];



//#define E3dRENDERER_THREAD_MAX_STATESTACK_SIZE  16

#ifdef E3dRENDERER_THREAD_MAX_STATESTACK_SIZE
	#define E3dRendererThreadStateStack\
		char  StateStr[E3dRENDERER_THREAD_MAX_STATESTACK_SIZE][1024];\
		int   StatePtr
#else
	#define E3dRendererThreadStateStack
#endif  // E3dRENDERER_THREAD_MAX_STATESTACK_SIZE


//
// Data writable by individual rendering threads
//
#define E3dRenderer_thread_core\
	EThreadCore;\
\
	E3dWindow*       L3DWindow;\
	E3dRenderer*     Renderer;\
	E3dScene*        Scene;\
	E3dCamera*       Camera;E3dModel*  CameraModel;\
	E3dTraceRayFunc  TraceRayProc;\
	E3dRendererThreadStateStack;\
\
	E3dLightMap      LightMap;\
\
	E3dCoord2        StratifiedUV;                             /* Anti-aliased renderers pass down stratified 0...1 U/V values  */\
/* For MIP-mapping / anisotropic texture filtering	*/\
	E3dAniso          Aniso;\
\
	E3dShaderLight*   ShaderLights;EIndex  NumOfShaderLights;  /* During rendering, we collect lights here that affect shaders (visible from a given point)  */\
	EIndex            NumOfLightSamplesFacingThisWay,          /* These are used for determining light occlusion ratio    */\
	                  NumOfUnshadowedLightSamples;\
	ERandomState      RandomState[5];\
	ERandomState      LightSamplingRandomState[5]



struct E3dRenderer_thread_S
{
	E3dRenderer_thread_core;
};



// RendererSets group varous Renderer instances (of different Classes) per view port (3DWindow)
//
struct E3dRendererSet_S
{
	EObjectCore(EObject);

	char*          FileName;            // For saving image(s)/animations
	double         StartFrame, EndFrame, FrameStep;

	E3dImageSpecs  ImageSpecs;          // Changeable by user

	EpImage*       ScaledBackgroundImage;

	EList(, E3dRenderer*, Renderers);   // Set of Renderers (one for each registered RendererClass)
	E3dRenderer*   Renderer;            // The current Renderer
	E3dRenderer*   LastRenderer;        // The user can toggle between this one and Renderer (middle button)
	EBool          InteractiveRedraw;
};


extern EResource           E3dRendererSet_resources[];

EList(extern, E3dRendererClass*, E3dRenderer_Classes);

extern E3dRendererClass    E3dRenderer_class[];
extern E3dRendererClass    E3dRayTracer_class[];

extern E3dRendererClass*   E3dWireframeRenderer_class;
extern E3dRendererClass*   E3dWireAndZRenderer_class;
extern E3dRendererClass*   E3dHiddenLineRenderer_class;
extern E3dRendererClass*   E3dShadedSolidRenderer_class;
extern E3dRendererClass*   E3dGLSLRenderer_class;



extern const char*    E3dGL_SemanticCULL_FACE_ENABLE;
extern const char*    E3dGL_SemanticDEPTH_TEST_ENABLE;
extern const char*    E3dGL_SemanticBLEND_SRC_FACTOR;
extern const char*    E3dGL_SemanticBLEND_DST_FACTOR;



#include <math.h>
#include <E3D/Math.h>


#define _E3d_ShadeSurface(mShadeSurfaceFunc,  mSurfacePoint,  mSurfaceMaterial, mVolumeShader, mShaderBeyondSurface,  mLightSamples, mNLightSamples,  mResult) do\
{\
	E3dSurfacePoint*  lSurfacePoint = mSurfacePoint;\
	const E3dShader*  lSurfaceShader = mSurfaceMaterial->Shader;\
	uint8_t  lTmpShader[lSurfaceShader->Class->StructSize];const E3dShader*  lTextureModulatedSurfaceShader = E3dShader_ModulateWithTextures(mSurfaceMaterial,lSurfaceShader, (E3dShader*)&lTmpShader, lSurfacePoint->TextureColors);\
	mShadeSurfaceFunc(lSurfacePoint,  mSurfaceMaterial, lTextureModulatedSurfaceShader, mVolumeShader, mShaderBeyondSurface,  mLightSamples, mNLightSamples,  mResult);\
} while(0)


#define E3dRenderer_ApplyChange(m, mChangeFlags, mRendererChangeFlags) do\
{\
	if(LSettingsPanel->ApplyChangesToAllWindows)\
	{\
		EIndex  lPC = 0;for(;lPC < E3dp_Panels_Count;lPC++)\
		{\
			E3dPanel*  l3DPanel = E3dp_Panels[lPC];EIndex  lWC = 0;const EIndex  lWN = l3DPanel->Windows_Count;for(;lWC < lWN;lWC++)\
			{\
				E3dWindow*  l3DWindow = l3DPanel->Windows[lWC];E3dpWindow*  l3DpWindow = E_DynamicCast(l3DWindow, E3dpWindow);if(l3DpWindow && l3DpWindow != LEdited3DWindow)\
				{\
					if(l3DWindow->StatusFlags & E3dWF_CREATED)\
					{\
						if(l3DpWindow->RendererSet->Renderer && LEdited3DWindow->RendererSet->Renderer)\
						{\
							const E3dRenderer*  lSrcRenderer = LEdited3DWindow->RendererSet->Renderer;\
							E3dRenderer*        lDstRenderer = l3DpWindow->RendererSet->Renderer;\
							lDstRenderer->m = lSrcRenderer->m;\
						}\
					}\
				}\
			}\
		}\
		E3dWindow_Redraw((E3dWindow*)LEdited3DWindow,  mChangeFlags, mRendererChangeFlags,  TRUE);\
		E3dWindows_RedrawByChangeFlagsExt(mChangeFlags, (E3dWindow*)LEdited3DWindow, FALSE, TRUE);\
	}\
	else E3dWindow_Redraw((E3dWindow*)LEdited3DWindow,  mChangeFlags, mRendererChangeFlags,  TRUE);\
} while(0)


static inline void E3d_RTriangleComputeTextureSpaceTangent(const E3dRTriangle* PRTriangle, E3dST* PST, const E3dTxMapperIndex PTextureMapperIndex)
{
	const E3dRPolyGroup*     LRPolyGroup = PRTriangle->RPolyGroup;
	const E3dVertexNodeSize  LRVertexNodeSize = LRPolyGroup->RVertexNodeSize;
	const uint8_t*           LRVertexNodes8 = (const uint8_t*)PRTriangle->RVertexNodes;
	const E3dST*             LST0 = (const E3dST*)(LRVertexNodes8 + LRVertexNodeSize * 0 + LRPolyGroup->ST0Offset) + PTextureMapperIndex;

/*
	const E3dST*    LST1 = (const E3dST*)(LRVertexNodes8 + LRVertexNodeSize * 1 + LRPolyGroup->ST0Offset) + PTxMapperIndex;
	const E3dST*    LST2 = (const E3dST*)(LRVertexNodes8 + LRVertexNodeSize * 2 + LRPolyGroup->ST0Offset) + PTxMapperIndex;

	E3dCoordinate  LS0 = LST0->S, LT0 = LST0->T,
	               LS1 = LST1->S, LT1 = LST1->T,
	               LS2 = LST2->S, LT2 = LST2->T,
	               LTgX, LTgY, LTgZ,
	               LDeltaS1, LDeltaT1, LDeltaS2, LDeltaT2,
	               LA, LB;


	LDeltaS1 = LS1 - LS0;
	LDeltaT1 = LT1 - LT0;
	LDeltaS2 = LS2 - LS0;
	LDeltaT2 = LT2 - LT0;

	LA = (LDeltaS1 - LDeltaT1 * LDeltaS2 / LDeltaT2);
	if(LA != 0.0f)
	{
		LA = 1.0f / LA;
	}
	LB = (LDeltaS2 - LDeltaT2 * LDeltaS1 / LDeltaT1);
	if(LB != 0.0f)
	{
		LB = 1.0f / LB;
	}

	LTgX = LA * PRTriangle->Edge1.X + LB * PRTriangle->Edge2.X;
	LTgY = LA * PRTriangle->Edge1.Y + LB * PRTriangle->Edge2.Y;
	LTgZ = LA * PRTriangle->Edge1.Z + LB * PRTriangle->Edge2.Z;

	E3d_Normalize3_v(LTgX, LTgY, LTgZ);


	PST->Tangent.X = LTgX;
	PST->Tangent.Y = LTgY;
	PST->Tangent.Z = LTgZ;
*/

	PST->Tangent = LST0->Tangent;
//printf("%s() %f,%f,%f\n", __FUNCTION__, PST->Tangent.X, PST->Tangent.Y, PST->Tangent.Z);fflush(stdout);
}


static inline void E3d_MapTextureNoNormalize(const E3dRTriangle* PRTriangle, const E3dCoord3* LPosition, const E3dCoord3* PNormal, const E3dCoord2* PUV, const E3dMaterial* PMaterial, const EIndex PTxMapperIndex, E3dST* PST)
{
	const E3dRPolyGroup*     LRPolyGroup = PRTriangle->RPolyGroup;
	const E3dVertexNodeSize  LRVertexNodeSize = LRPolyGroup->RVertexNodeSize;
	const E3dCoordinate      LU = PUV->X, LV = PUV->Y, LNUV = 1.0 - LU - LV;

	{
		const E3dTextureMapper*       LTextureMapper = LRPolyGroup->TxMappers[PTxMapperIndex];
		const E3dTextureMapperClass*  LTextureMapperClass = LTextureMapper->Class;

		if(LTextureMapperClass->MapPoint)
		{
			LTextureMapperClass->MapPoint(LTextureMapper, LPosition, PNormal, (E3dST*)PST);
		}
		else
		{
			const uint8_t*  LRVertexNodes8 = (const uint8_t*)PRTriangle->RVertexNodes;
			const E3dST*    LST0 = (const E3dST*)(LRVertexNodes8 + LRVertexNodeSize * 0 + LRPolyGroup->ST0Offset) + PTxMapperIndex;
			const E3dST*    LST1 = (const E3dST*)(LRVertexNodes8 + LRVertexNodeSize * 1 + LRPolyGroup->ST0Offset) + PTxMapperIndex;
			const E3dST*    LST2 = (const E3dST*)(LRVertexNodes8 + LRVertexNodeSize * 2 + LRPolyGroup->ST0Offset) + PTxMapperIndex;

			PST->S = LNUV * LST0->S + LST1->S * LU + LST2->S * LV;
			PST->T = LNUV * LST0->T + LST1->T * LU + LST2->T * LV;

			{
				EIndex  LTxC = 0;const EIndex  LTxN = PMaterial->Textures_Count;for(;LTxC < LTxN;LTxC++)
				{
					E3d2DTexture*  LTexture = PMaterial->Textures[LTxC];
					if(LRPolyGroup->TextureMapperIndices[LTxC] == PTxMapperIndex && LTexture->DoNormalMap && LTexture->NormalMapType == E3dTANGENT_SPACE) { E3d_RTriangleComputeTextureSpaceTangent(PRTriangle, (E3dST*)PST, PTxMapperIndex);break; }
				}
			}
		}
	}
}


//
// Do not normalize texcoords here (clamp for repeat etc.) because differential rays may be used for mip-mapping etc. and they need the un-clamped texcoords.
// Clamping must be done just before sampling the texture.
//
static inline void E3d_MapTextures(const E3dRTriangle* PRTriangle, const E3dCoord3* LPosition, const E3dCoord3* PNormal, const E3dCoord2* PUV, const E3dMaterial* PMaterial, E3dSTWithDiff* PST)
{
	const E3dRPolyGroup*     LRPolyGroup = PRTriangle->RPolyGroup;
	const E3dVertexNodeSize  LRVertexNodeSize = LRPolyGroup->RVertexNodeSize;
	const E3dCoordinate      LU = PUV->X, LV = PUV->Y, LNUV = 1.0 - LU - LV;
	E3dSTWithDiff*           LST = PST;

	const E3dTxMapperIndex  LTxMN = LRPolyGroup->TxMappers_Count;
#ifdef E3dRenderDEBUG
	if(LTxMN > 0)  // We should not have to do this check. The render prep should ensure that PolyGroups with Textures have the corresponding TextureMappers
#endif  // E3dRenderDEBUG
	{
		EBool  LNeedTxTangents[LTxMN];EClr_s(LNeedTxTangents);

		{
			EIndex  LTxC = 0;const EIndex  LTxN = PMaterial->Textures_Count;for(;LTxC < LTxN;LTxC++)
			{
				E3d2DTexture*  LTexture = PMaterial->Textures[LTxC];
				if(LTexture->DoNormalMap && LTexture->NormalMapType == E3dTANGENT_SPACE) LNeedTxTangents[LRPolyGroup->TextureMapperIndices[LTxC]] = TRUE;
			}
		}

		E3dTxMapperIndex  LTxMC = 0;for(;LTxMC < LTxMN;LTxMC++, LST++)
		{
			const E3dTextureMapper*       LTextureMapper = LRPolyGroup->TxMappers[LTxMC];
			const E3dTextureMapperClass*  LTextureMapperClass = LTextureMapper->Class;

			if(LTextureMapperClass->MapPoint)
			{
//printf("%s() %s:%d  %s %p\n", __FUNCTION__, __FILE__, __LINE__, LTextureMapperClass->Name, LTextureMapperClass->MapPoint);fflush(stdout);
				LTextureMapperClass->MapPoint(LTextureMapper,  LPosition, PNormal,  (E3dST*)LST);
_E3dCheckST_p(LST);
			}
			else
			{
				const uint8_t*           LRVertexNodes8 = (const uint8_t*)PRTriangle->RVertexNodes;
				const E3dVertexNodeSize  LST0Offset = LRPolyGroup->ST0Offset;
				const E3dST*             LST0 = (const E3dST*)(LRVertexNodes8 + LRVertexNodeSize * 0 + LST0Offset) + LTxMC;
				const E3dST*             LST1 = (const E3dST*)(LRVertexNodes8 + LRVertexNodeSize * 1 + LST0Offset) + LTxMC;
				const E3dST*             LST2 = (const E3dST*)(LRVertexNodes8 + LRVertexNodeSize * 2 + LST0Offset) + LTxMC;

_E3dCheckST_p(LST0);
_E3dCheckST_p(LST1);
_E3dCheckST_p(LST2);

				LST->S = LNUV * LST0->S + LST1->S * LU + LST2->S * LV;
				LST->T = LNUV * LST0->T + LST1->T * LU + LST2->T * LV;

				if(LNeedTxTangents[LTxMC]) E3d_RTriangleComputeTextureSpaceTangent(PRTriangle, (E3dST*)LST, LTxMC);
			}

// Indicate that no MIP-mapping should be used
//
			LST->dS = 0.0;
			LST->dT = 0.0;
		}
	}
}


#define _E3dM_SampleTextures(mMaterial, mRTriangle, mSurfacePoint) do\
{\
	EcRGBAf32Color*           lColor = (mSurfacePoint)->TextureColors;\
	const E3dRPolyGroup*      lRPolyGroup = mRTriangle->RPolyGroup;\
	const E3dTxMapperIndex*   lTextureMapperIndices = lRPolyGroup->TextureMapperIndices;\
	E3dTextureMapper* const*  lTextureMappers = lRPolyGroup->TxMappers;\
\
	EIndex  lTxC = 0;const EIndex  lTxN = (mMaterial)->Textures_Count;for(;lTxC < lTxN;lTxC++, lColor++)\
	{\
		const E3d2DTexture*      lTexture = (mMaterial)->Textures[lTxC];\
		const E3dTxMapperIndex   lMapperIndex = lTextureMapperIndices[lTxC];\
		const E3dTextureMapper*  lMapper = lTextureMappers[lMapperIndex];\
\
/*printf("%s() %s:%d  %p %p\n", __FUNCTION__, __FILE__, __LINE__, lTexture, lTexture->Sample);fflush(stdout);*/\
		lTexture->Sample(lTexture, (mSurfacePoint)->ST + lMapperIndex,  lMapper->TilingS, lMapper->TilingT,  mSurfacePoint, (float*)lColor);\
	}\
} while(0)


//========================================================
// Simplified version of _E3dM_SampleTextures
// It doesn't do normal mapping etc.
// E3dSPECTRALFIXME
//========================================================
static inline void _E3d_LightSampleTextures(const E3dMaterial* PMaterial, E3dTextureMapper* const* PTextureMappers, const E3dTxMapperIndex* PTextureMapperIndices, E3dSurfacePoint* PPoint)
{
	EcRGBAf32Color*  LColor = PPoint->TextureColors;

	EIndex  LTxC = 0;const EIndex  LTxN = PMaterial->Textures_Count;for(;LTxC < LTxN;LTxC++, LColor++)
	{
		float*                  LResult = (float*)LColor;
		const E3dTxMapperIndex  LMapperIndex = PTextureMapperIndices[LTxC];

		const E3dCoordinate  LFS = PPoint->ST[LMapperIndex].S;

		if(LFS != E3dST_OFF_TEXTURE)
		{
			const E3d2DTexture*      LTexture = PMaterial->Textures[LTxC];
			const E3dTextureMapper*  LTextureMapper = PTextureMappers[LMapperIndex];
			const E3dImage*          LImage = LTexture->Image;
			const EIndex             LSizeX = LImage->Size.X, LSizeY = LImage->Size.Y;
			const E3dCoordinate      LFX = LFS * (E3dCoordinate)LSizeX;
			const E3dCoordinate      LFY = PPoint->ST[LMapperIndex].T * (E3dCoordinate)LSizeY;

// Bi-linear interpolation
//
			if(LTexture->DoBilinear)
			{
//printf("%s() %s:%d  %d %s %dx%d\n", __FUNCTION__, __FILE__, __LINE__, LTxC, LTexture->FileName,  LSizeX, LSizeY);fflush(stdout);
				_E3d_Bilinear0(LFX, LFY,  LTextureMapper->TilingS, LTextureMapper->TilingT);
				_E3d_ImageColorBilinear_AUTO_RGBAf32(LImage,  LT, LFraction, LFractionInv, LNeighbor,  LResult);
			}
			else
			{
				_E3d_ImageColorNearest_AUTO_RGBAf32(LImage,  LFX, LFY,  LTextureMapper->TilingS, LTextureMapper->TilingT,  LResult);
			}
		}
		else
		{
			_E3dM_NoTextureCoverage(LResult);
		}
	}
}


//========================================================
//========================================================
static inline const E3dShader* _E3d_LightShaderGetAtPoint(E3dSurfacePoint* PPoint, const E3dRTriangle* PRTriangle, const E3dCoord2* PUV, const E3dMaterial* PMaterial, const E3dShader* PSrcShader, E3dShader* PDstShader)
{
	switch(PMaterial->Textures_Count)
	{
		case 0:
		return(PSrcShader);

		case 1:
			{
				E3dSTWithDiff*        LST = PPoint->ST;
				const E3dRPolyGroup*  LRPolyGroup = PRTriangle->RPolyGroup;

				E3d_MapTextures(PRTriangle,  &PPoint->Position, &PPoint->Normal,  PUV, PMaterial, PPoint->ST);
				_E3d_LightSampleTextures(PMaterial,  LRPolyGroup->TxMappers, LRPolyGroup->TextureMapperIndices,  PPoint);

// With 1 texture, we can avoid calling _E3dShader_ModulateWithTextures if the texture is not mapped at the given point
//
				if(LST->S != E3dST_OFF_TEXTURE)
				{
					E3dShader_ModulateWithTextures(PMaterial, PSrcShader, PDstShader, PPoint->TextureColors);
					return(PDstShader);
				}
				else return(PSrcShader);
			}
		break;

		default:
			{
				const E3dRPolyGroup*  LRPolyGroup = PRTriangle->RPolyGroup;
				E3d_MapTextures(PRTriangle,  &PPoint->Position, &PPoint->Normal,  PUV, PMaterial, PPoint->ST);
				_E3d_LightSampleTextures(PMaterial,  LRPolyGroup->TxMappers, LRPolyGroup->TextureMapperIndices,  PPoint);
				E3dShader_ModulateWithTextures(PMaterial, PSrcShader, PDstShader, PPoint->TextureColors);
			}
		return(PDstShader);
	}
}


//========================================================
// Sign-agnostic Fresnel term (PNdotV may be negative)
// Schlick's approximation
//========================================================
static inline float E3d_Fresnel(const float PCurrentRefractiveIndex, const float PRefractiveIndex, const float PNdotV)
{
/*
	const float  Leta = PCurrentRefractiveIndex / PRefractiveIndex;
	const float  LT0 = Leta - 1.0, LT1 = Leta + 1.0,
	             LR0 = LT0 * LT0 / (LT1 * LT1);
*/

	const float  LT0 = PCurrentRefractiveIndex - PRefractiveIndex, LT1 = PCurrentRefractiveIndex + PRefractiveIndex, LT2 = LT0 / LT1,
	             LR0 = LT2 * LT2;  // Reflectivity at normal incidence


//printf("%s() %s:%d  %f %f %f\n", __FUNCTION__, __FILE__, __LINE__, LT0, LT1, LR0);fflush(stdout);

// Simplified Fresnel term (Schlick's approximation)
//
	const float  c = 1.0 - fabsf(PNdotV);
	return(LR0 + (1.0 - LR0) * c * c * c * c * c);

// The full Fresnel term
//
/*
	const float  c = 1.0 - fabsf(PNdotV);
	float        g = E3d_Sqrt(PRefractiveIndex * PRefractiveIndex + c * c - 1.0), gPc = g + c, gMc = g - c;

	return(0.5 * gMc * gMc / (gPc * gPc) * (1.0 + (c * gPc - 1.0) * (c * gPc - 1.0) / ((c * gMc + 1.0) * (c * gMc + 1.0))));
*/
}


//========================================================
// Fresnel term for dielectrics
//========================================================
static inline void E3d_DielectricFresnel(const float* PEtas_i, const float* PEtas_t,
 const int PNSpectralSamples,
 const E3dCoordinate PCosI,
 float* PReflectivity)
{
	const E3dCoordinate  L2    = E3d_Sqrt(E3d_Max(0.0, 1.0 - PCosI * PCosI));
	const E3dCoordinate  LCosI = E3d_Abs(PCosI);

	int  LC = 0;for(;LC < PNSpectralSamples;LC++)
	{
		const float          LEtaI = PEtas_i[LC];
		const float          LEtaT = PEtas_t[LC];
		const E3dCoordinate  LSinT = LEtaI / LEtaT * L2;	// Compute sin t with Snell's law

		if(LSinT >= 1.0)
		{
// Handle total internal reflection
//
			PReflectivity[LC] = 1.0;
		}
		else
		{
			const E3dCoordinate  LCosT = E3d_Sqrt(E3d_Max(0.0, 1.0 - LSinT * LSinT));

			const E3dCoordinate  LEtaTxLCosI = LEtaT * LCosI,
			                     LEtaIxLCosI = LEtaI * LCosI,
			                     LEtaIxLCosT = LEtaI * LCosT,
			                     LEtaTxLCosT = LEtaT * LCosT;

			const E3dCoordinate  LRParl = (LEtaTxLCosI - LEtaIxLCosT) / (LEtaTxLCosI + LEtaIxLCosT);
			const E3dCoordinate  LRPerp = (LEtaIxLCosI - LEtaTxLCosT) / (LEtaIxLCosI + LEtaTxLCosT);

			PReflectivity[LC] = (LRParl * LRParl + LRPerp * LRPerp) * 0.5;
		}
	}
}


//========================================================
// Fresnel term for dielectrics - enter/exit check
//========================================================
static inline void E3d_DielectricFresnelDirCheck(const float* PEtas_i, const float* PEtas_t,
 const int LNumOfSpectralSamples,
 const E3dCoordinate PCosI,
 float* PReflectivity)
{
	const EBool   LEntering = PCosI > 0.0;
	const float*  LEtas_i;
	const float*  LEtas_t;

//assert(PCosi >= -1.0 && PCosi <= 1.0);
	if(LEntering) { LEtas_i = PEtas_i;LEtas_t = PEtas_t; }
	else { LEtas_i = PEtas_t;LEtas_t = PEtas_i; }

	{
		const E3dCoordinate  L2 = E3d_Sqrt(E3d_Max(0.0, 1.0 - PCosI * PCosI));
		const E3dCoordinate  LCosI = E3d_Abs(PCosI);

		int  LC = 0;for(;LC < LNumOfSpectralSamples;LC++)
		{
			const float          LEtaI = LEtas_i[LC];
			const float          LEtaT = LEtas_t[LC];
			const E3dCoordinate  LSinT = LEtaI / LEtaT * L2;	// Compute sin t with Snell's law

			if(LSinT >= 1.0)
			{
// Handle total internal reflection
//
				PReflectivity[LC] = 1.0;
			}
			else
			{
				const E3dCoordinate  LCosT = E3d_Sqrt(E3d_Max(0.0, 1.0 - LSinT * LSinT));

				const E3dCoordinate  LEtaTxLCosI = LEtaT * LCosI,
				                     LEtaIxLCosI = LEtaI * LCosI,
				                     LEtaIxLCosT = LEtaI * LCosT,
				                     LEtaTxLCosT = LEtaT * LCosT;

				const E3dCoordinate  LRParl = (LEtaTxLCosI - LEtaIxLCosT) / (LEtaTxLCosI + LEtaIxLCosT);
				const E3dCoordinate  LRPerp = (LEtaIxLCosI - LEtaTxLCosT) / (LEtaIxLCosI + LEtaTxLCosT);

				PReflectivity[LC] = (LRParl * LRParl + LRPerp * LRPerp) * 0.5;
				}
		}
	}
}


//========================================================
// Fresnel term for conductors (metal)
//========================================================
static inline void E3d_MetalFresnel(const float* PEtas,
 const float* Pks,
 const int PNSpectralSamples,
 const E3dCoordinate PCosI,
 float* PReflectivity)
{
	const E3dCoordinate  LCosI2 = PCosI * PCosI;

	int  LC = 0;for(;LC < PNSpectralSamples;LC++)
	{
		const E3dCoordinate  LEta = PEtas[LC], Lk = Pks[LC],
		                     LTmpF = LEta * LEta + Lk * Lk, LTmp  = LTmpF * LCosI2,
		                     L2EtaCosI = 2.0 * LEta * PCosI;

		const E3dCoordinate  LRparl2 = (LTmp  - (L2EtaCosI) + 1.0)    / (LTmp  + (L2EtaCosI) + 1.0);
		const E3dCoordinate  LRperp2 = (LTmpF - (L2EtaCosI) + LCosI2) / (LTmpF + (L2EtaCosI) + LCosI2);

		E3dCoordinate        LF = (LRparl2 + LRperp2) * 0.5;if(LF < 0.0) LF = 0.0;

		PReflectivity[LC] = LF;
	}
}



static const float	_Ones[] = { 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0 };


//========================================================
// Dielectric Fresnel term for shaders
//========================================================
static inline void E3d_ShaderFresnel(E3dSurfacePoint* PPoint, const E3dShader* PVolumeShader, const E3dShader* PShaderBeyondSurface, const int PSpectralIndex, const int PNSpectralSamples, E3dShaderResult* PResult)
{
	const E3dCoord3      LNormal = PPoint->Normal;
	const E3dCoord4      LPointToViewer = PPoint->PointToViewer;
	const E3dCoordinate  LNdotV = E3d_Dot3_vv(LNormal, LPointToViewer);

	if(0)
	{
// Old Fresnel term (Schlick's approximation)
//
		const float  LCurrentRefractiveIndex = PVolumeShader ? PVolumeShader->RefractiveIndex.Samples[PSpectralIndex] : 1.0;
		const float  LNextRefractiveIndex    = PShaderBeyondSurface ? PShaderBeyondSurface->RefractiveIndex.Samples[PSpectralIndex] : 1.0;
		const float  LFresnel = E3d_Fresnel(LCurrentRefractiveIndex, LNextRefractiveIndex, LNdotV);
		const float  LNFresnel = 1.0 - LFresnel;

		{
			E3dM_SpectralLoop(PSpectralIndex, PNSpectralSamples)
			{
				PResult->Reflective[LSI] *= LFresnel;PResult->Transparent[LSI] *= LNFresnel;
			}
		}
	}
	else
	{
// Physically accurate dielectric Fresnel term (i.e. glass)
//
		float  LFresnel[PNSpectralSamples];
		E3d_DielectricFresnel(PVolumeShader ? PVolumeShader->RefractiveIndex.Samples + PSpectralIndex : _Ones, PShaderBeyondSurface ? PShaderBeyondSurface->RefractiveIndex.Samples + PSpectralIndex : _Ones,
		 PNSpectralSamples,
		 LNdotV,
		 LFresnel);

		{
			E3dM_SpectralLoop(PSpectralIndex, PNSpectralSamples)
			{
				const float  LFresnelV = LFresnel[LSI];
				PResult->Reflective[LSI] *= LFresnelV;PResult->Transparent[LSI] *= 1.0 - LFresnelV;
			}
		}
	}
}


//================================================================================================
// Intersect Ray with a bounding box and return the t parameter at the entry and exit points
//================================================================================================
static inline EBool E3d_RayIntersectBBox(const E3dGridRay* PRay, const E3dBBox* PBBox, const E3dCoordinate* PRayDirInv)
{
	const E3dCoord3     LBBoxMin = PBBox->Min, LBBoxMax = PBBox->Max;
	const E3d_Sub3i_vv  (LOriginToBoxMin,  PRay->Origin, LBBoxMin);
	E3dCoordinate       LLMin = -1.0, LLMax = -1.0;

	if(PRay->Direction.X != 0.0)
	{
		const E3dCoordinate  LTmp1 =                             -LOriginToBoxMin.X  * PRayDirInv[0];
		const E3dCoordinate  LTmp2 = ((LBBoxMax.X - LBBoxMin.X) - LOriginToBoxMin.X) * PRayDirInv[0];

		if(LTmp1 > LTmp2)
		{
			if(LTmp1 < 0.0) return(FALSE);
			LLMin = LTmp2;
			LLMax = LTmp1;
		}
		else
		{
			if(LTmp2 < 0.0) return(FALSE);
			LLMin = LTmp1;
			LLMax = LTmp2;
		}
	}

	if(PRay->Direction.Y != 0.0)
	{
		E3dCoordinate  LTmp1 =                              -LOriginToBoxMin.Y * PRayDirInv[1];
		E3dCoordinate  LTmp2 = ((LBBoxMax.Y - LBBoxMin.Y) - LOriginToBoxMin.Y) * PRayDirInv[1];
		if(LTmp1 > LTmp2) EM_Swap(E3dCoordinate,  LTmp1, LTmp2);
		if(LTmp1 > LLMin) LLMin = LTmp1;
		if(LTmp2 < LLMax || LLMax < 0.0) LLMax = LTmp2;
		if(LLMax < 0.0) return(FALSE);
	}

	if(PRay->Direction.Z != 0.0)
	{
		E3dCoordinate  LTmp1 =                              -LOriginToBoxMin.Z * PRayDirInv[2];
		E3dCoordinate  LTmp2 = ((LBBoxMax.Z - LBBoxMin.Z) - LOriginToBoxMin.Z) * PRayDirInv[2];
		if(LTmp1 > LTmp2) EM_Swap(E3dCoordinate,  LTmp1, LTmp2);
		if(LTmp1 > LLMin) LLMin = LTmp1;
		if(LTmp2 < LLMax || LLMax < 0.0) LLMax = LTmp2;	
	} 

	if((LLMin <= LLMax) && (LLMax >= 0.0)) return(TRUE);
	else return(FALSE);
}


//================================================================================================
// Intersect Ray with a bounding box and return the t parameter at the entry and exit points
// PEntryT < 0.0 means we're inside the BBox (the entry point is behind us)
//================================================================================================
static inline EBool E3d_RayIntersectBBoxInOut(const E3dRay* PRay, const E3dBBox* PBBox, const E3dCoordinate* PRayDirInv, E3dCoordinate* PEntryT, E3dCoordinate* PExitT)
{
	const E3dCoord3     LBBoxMin = PBBox->Min, LBBoxMax = PBBox->Max;
	const E3d_Sub3i_vv  (LOriginToBoxMin,  PRay->Origin, LBBoxMin);
	E3dCoordinate       LLMin = -1.0, LLMax = -1.0;

	if(PRay->Direction.X != 0.0)
	{
		E3dCoordinate  LTmp1 =                             -LOriginToBoxMin.X  * PRayDirInv[0];
		E3dCoordinate  LTmp2 = ((LBBoxMax.X - LBBoxMin.X) - LOriginToBoxMin.X) * PRayDirInv[0];
		if(LTmp1 > LTmp2) EM_Swap(E3dCoordinate,  LTmp1, LTmp2);
		LLMin = LTmp1;
		LLMax = LTmp2;
		if(LLMax < 0.0) return(FALSE);
	}

	if(PRay->Direction.Y != 0.0)
	{
		E3dCoordinate  LTmp1 =                              -LOriginToBoxMin.Y * PRayDirInv[1];
		E3dCoordinate  LTmp2 = ((LBBoxMax.Y - LBBoxMin.Y) - LOriginToBoxMin.Y) * PRayDirInv[1];
		if(LTmp1 > LTmp2) EM_Swap(E3dCoordinate,  LTmp1, LTmp2);
		if(LTmp1 > LLMin) LLMin = LTmp1;
		if(LTmp2 < LLMax || LLMax < 0.0) LLMax = LTmp2;
		if(LLMax < 0.0) return(FALSE);
	}

	if(PRay->Direction.Z != 0.0)
	{
		E3dCoordinate  LTmp1 =                              -LOriginToBoxMin.Z * PRayDirInv[2];
		E3dCoordinate  LTmp2 = ((LBBoxMax.Z - LBBoxMin.Z) - LOriginToBoxMin.Z) * PRayDirInv[2];
		if(LTmp1 > LTmp2) EM_Swap(E3dCoordinate,  LTmp1, LTmp2);
		if(LTmp1 > LLMin) LLMin = LTmp1;
		if(LTmp2 < LLMax || LLMax < 0.0) LLMax = LTmp2;	
	} 

	if(LLMin <= LLMax && LLMax >= 0.0)
	{
		*PEntryT = LLMin;
		*PExitT = LLMax;
		return(TRUE);
	}
	else return(FALSE);
}


//================================================================================================
// Intersect Ray with a bounding box and return the t parameter at the entry and exit points
// PEntryT < 0.0 means we're inside the BBox (the entry point is behind us)
//================================================================================================
static inline EBool E3d_RayIntersectBBoxMaxDist(const E3dGridRay* PRay, const E3dBBox* PBBox, const E3dCoordinate* PRayDirInv, E3dCoordinate* PEntryT, E3dCoordinate* PExitT, const E3dCoordinate LMaxDistance)
{
	const E3dCoord3     LBBoxMin = PBBox->Min, LBBoxMax = PBBox->Max;
	const E3d_Sub3i_vv  (LOriginToBoxMin,  PRay->Origin, LBBoxMin);
	E3dCoordinate       LLMin = -1.0, LLMax = -1.0;

	if(PRay->Direction.X != 0.0)
	{
		E3dCoordinate  LTmp1 =                             -LOriginToBoxMin.X  * PRayDirInv[0];
		E3dCoordinate  LTmp2 = ((LBBoxMax.X - LBBoxMin.X) - LOriginToBoxMin.X) * PRayDirInv[0];
		if(LTmp1 > LTmp2) EM_Swap(E3dCoordinate,  LTmp1, LTmp2);
		LLMin = LTmp1;
		LLMax = LTmp2;
		if(LLMax < 0.0 || LLMin > LMaxDistance) return(FALSE);
	}

	if(PRay->Direction.Y != 0.0)
	{
		E3dCoordinate  LTmp1 =                              -LOriginToBoxMin.Y * PRayDirInv[1];
		E3dCoordinate  LTmp2 = ((LBBoxMax.Y - LBBoxMin.Y) - LOriginToBoxMin.Y) * PRayDirInv[1];
		if(LTmp1 > LTmp2) EM_Swap(E3dCoordinate,  LTmp1, LTmp2);
		if(LTmp1 > LLMin) LLMin = LTmp1;
		if(LTmp2 < LLMax || LLMax < 0.0) LLMax = LTmp2;
		if(LLMax < 0.0 || LLMin > LMaxDistance) return(FALSE);
	}

	if(PRay->Direction.Z != 0.0)
	{
		E3dCoordinate  LTmp1 =                              -LOriginToBoxMin.Z * PRayDirInv[2];
		E3dCoordinate  LTmp2 = ((LBBoxMax.Z - LBBoxMin.Z) - LOriginToBoxMin.Z) * PRayDirInv[2];
		if(LTmp1 > LTmp2) EM_Swap(E3dCoordinate,  LTmp1, LTmp2);
		if(LTmp1 > LLMin) LLMin = LTmp1;
		if(LTmp2 < LLMax || LLMax < 0.0) LLMax = LTmp2;	
	} 

	if(LLMin <= LLMax && LLMax >= 0.0 && LLMin <= LMaxDistance)
	{
		*PEntryT = LLMin;
		*PExitT = LLMax;
		return(TRUE);
	}
	else return(FALSE);
}


static inline void E3d_ClosestPointOnTriangle(const E3dRTriangle* PRTriangle, const E3dCoord3* PSrcPosition, E3dCoord3* PDstPosition, E3dCoord2* PUV)
{
	const E3dCoord3  LEdge0 = PRTriangle->Edge1;
	const E3dCoord3  LEdge1 = PRTriangle->Edge2;
	const E3dCoord3  LVertex0 = PRTriangle->Vertex0;
	const E3d_Sub3i_vp(v0, LVertex0, PSrcPosition);

	const E3dCoordinate  a = E3d_Dot3_vv(LEdge0, LEdge0);
	const E3dCoordinate  b = E3d_Dot3_vv(LEdge0, LEdge1);
	const E3dCoordinate  c = E3d_Dot3_vv(LEdge1, LEdge1);
	const E3dCoordinate  d = E3d_Dot3_vv(LEdge0, v0);
	const E3dCoordinate  e = E3d_Dot3_vv(LEdge1, v0);


/*
 vector3 edge0 = triangle[1] - triangle[0];
 vector3 edge1 = triangle[2] - triangle[0];
 vector3 v0 = triangle[0] - sourcePosition;

 E3dCoordinate a = edge0.dot(edge0);
 E3dCoordinate b = edge0.dot(edge1);
 E3dCoordinate c = edge1.dot(edge1);
 E3dCoordinate d = edge0.dot(v0);
 E3dCoordinate e = edge1.dot(v0);
*/

	const E3dCoordinate  det = a * c - b * b;
	E3dCoordinate        s = b * e - c * d;
	E3dCoordinate        t = b * d - a * e;

	if(s + t < det)
	{
		if(s < 0.0)
		{
			if(t < 0.0)
			{
				if(d < 0.0)
				{
					s = E3d_ClampMinMax_Coordinate(-d / a, 0.0, 1.0);
					t = 0.0;
				}
				else
				{
					s = 0.0;
					t = E3d_ClampMinMax_Coordinate(-e / c, 0.0, 1.0);
				}
			}
			else
			{
				s = 0.0;
				t = E3d_ClampMinMax_Coordinate(-e / c,  0.0, 1.0);
			}
		}
		else if (t < 0.0)
		{
			s = E3d_ClampMinMax_Coordinate(-d / a, 0.0, 1.0);
			t = 0.0;
		}
		else
		{
			const E3dCoordinate invDet = 1.0 / det;
			s *= invDet;
			t *= invDet;
		}
	}
	else
	{
		if(s < 0.0)
		{
		 const E3dCoordinate  tmp0 = b + d;
		 const E3dCoordinate  tmp1 = c + e;
		 if(tmp1 > tmp0)
		 {
			 const E3dCoordinate  numer = tmp1 - tmp0;
			 const E3dCoordinate  denom = a - 2.0 * b + c;
			 s = E3d_ClampMinMax_Coordinate(numer / denom,  0.0, 1.0);
			 t = 1.0 - s;
		 }
		 else
		 {
			 t = E3d_ClampMinMax_Coordinate(-e/c, 0.0, 1.0);
			 s = 0.0;
		 }
		}
		else if(t < 0.0)
		{
			if(a + d > b + e)
			{
				const E3dCoordinate  numer = c + e - b - d;
				const E3dCoordinate  denom = a - 2 * b + c;
				s = E3d_ClampMinMax_Coordinate(numer / denom, 0.0, 1.0);
				t = 1 - s;
			}
			else
			{
				s = E3d_ClampMinMax_Coordinate(-e / c, 0.0, 1.0);
				t = 0.0;
			}
		}
		else
		{
			const E3dCoordinate  numer = c + e - b - d;
			const E3dCoordinate  denom = a - 2 * b + c;
			s = E3d_ClampMinMax_Coordinate(numer / denom, 0.0, 1.0);
			t = 1.0 - s;
		}
	}

// return triangle[0] + s * edge0 + t * edge1;

// Return barycentric coordinates
//
	PUV->X = s;
	PUV->Y = t;
	PDstPosition->X = LVertex0.X + s * LEdge0.X + t * LEdge1.X;
	PDstPosition->Y = LVertex0.Y + s * LEdge0.Y + t * LEdge1.Y;
	PDstPosition->Z = LVertex0.Z + s * LEdge0.Z + t * LEdge1.Z;
}





//----------------------------------------
// Material.c
//----------------------------------------
extern void     E3d_RGBToSpectrumO(const EcRGBf32Color* PRGB, float* PSpectrum, const int PNSpectralSamples, const float PScaler);


//----------------------------------------
// Rendering/Rendering.c
//----------------------------------------
extern int                E3d_ShadowPixelType, E3d_ShadowComponentsPerPixel, E3d_ShadowPixelComponentType;
extern void               E3dp_RenderingHWInfoInit(E3dGPUInfo* PHardwareInfo);
extern void               E3d_RenderingInit(void);

extern void               _E3d_ImageSpecsCopy(E3dImageSpecs* PSrc, E3dImageSpecs* PDst);

extern void               E3d_RendererUpdateSpectrum(E3dRenderer* PRenderer);

extern E3dRendererClass*  E3dRenderer_ClassFindByName(const char* PName);
extern E3dRendererClass*  E3dRenderer_ClassRegister(E3dRendererClassDescriptor* PTemplate);
extern E3dRendererClass*  E3dRenderer_ClassRegisterInactive(const char* PName);
extern void               E3dRenderer_ClassDeactivate(E3dRendererClass* PClass);
extern void               E3dRenderer_ClassesFree(void);


extern E3dRenderer*     E3dRenderer_New(E3dRendererClass* PRendererClass);
extern E3dRenderer*     E3dRenderer_CloneNew(E3dRendererClass* PRendererClass);

extern void             E3dRender_QueueClear(E3dRenderQueueEntry* PEntry);
extern void             E3dRenderer_ClearChangeQueues(E3dRenderer* PRenderer);

extern void             E3dRenderer_ImagesFree(E3dRenderer* PRenderer);
extern void             E3dRenderer_FreeCaches(E3dRenderer* PRenderer);

extern void             E3dRenderer_Finish(E3dRenderer* PRenderer, const long long PPollMicroseconds, const EBool PSyncGUI);
extern EBool            E3dRenderer_Stop(E3dRenderer* PRenderer);
extern void             E3d_RenderersStop(void);
extern void             E3dRenderers_FreeCaches(void);
extern void             E3dRenderers_MarkChanges(const EChangeFlags PSceneChangeFlags);

extern void             E3dRendererSet_Init(E3dRendererSet* PRendererSet);
extern E3dRendererSet*  E3dRendererSet_New(void);
extern int              E3dRendererSet_InitToRayTracer(E3dRendererSet* PRendererSet);
extern E3dRenderer*     E3dRendererSet_RendererByClass(const E3dRendererSet* PRendererSet, const E3dRendererClass* PRendererClass);

extern void             E3dRendererSet_CreateRenderers(E3dRendererSet* PRendererSet);
extern E3dRenderer*     E3dRendererSet_RendererByClassName(E3dRendererSet* PRendererSet, const char* PRendererClassName);
extern void             E3dRendererSet_MarkChanges(E3dRendererSet* PRendererSet, E3dRenderer* PExceptThisRenderer, const EChangeFlags PSceneChangedFlags);
extern void             E3dRenderer_CheckImageBuffers(E3dRenderer* PRenderer);
extern void             E3d_RendererUpdatePixelRemappingGrids(E3dRenderer* PRenderer);

extern EBool            E3dRenderer_NeedsToneMapping(const E3dRenderer* PRenderer);
extern E3dDisplayFlags  E3dRenderer_DisplayFlags(E3dRenderer* PRenderer);

extern EBool            E3d_MultiThreadedRendererRedraw(E3dWindow* P3DWindow, E3dRayTracer* PRenderer, E3dScene* PScene);



extern int         E3d_TraceRayOpaque(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, E3dStackRay* PRay, int PRayIter, E3dShaderResult* PColor);
extern int         E3d_TraceRay(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, E3dStackRay* PRay, int PRayIter, E3dShaderResult* PColor);

extern int         E3d_SceneRenderFrame(E3dWindow* P3DWindow, E3dScene* PScene, E3dRenderer* PRenderer,  E3dCamera* PCamera, E3dModel* PCameraModel,  const EBool PSwapBuffers);


extern void        E3dRenderer_FreeLightSamples(E3dRenderer* PRenderer);

extern E3dCoord2*  E3d_CreateStratifiedSamples2D(const EIndex PNumOfSamples, const char PType, ERandomState* PState, E3dCoord2* PSamples);


extern void        E3d_HemisphereSampleCosineWeightTBN(const E3dCoordinate PU, const E3dCoordinate PV, const E3dCoord3* PTangent, const E3dCoord3* PBitangent, const E3dCoord3* PNormal, E3dCoord3* PResult);
extern E3dCoord3*  E3d_CreateSphereSamples(const EIndex PNumOfSphereSamples, const char PType, ERandomState* PRandomState, const EBool PCosineWeight);


extern char*       E3d_DisplayFlagsToString(const E3dDisplayFlags PFlags, char* PString);
extern char*       E3d_ItemFlagsToString(const E3dDisplayFlags PFlags, char* PString);


extern void        E3dGPUAA_Init(E3dGPUAA* PAA);


//----------------------------------------
// Rendering/Raytracing.c
//----------------------------------------
extern int         E3dRenderer_NSubPixels(const E3dRenderer* PRenderer);
extern EBool       E3d_SceneCollectAreaLights(const E3dScene* PScene, E3dRenderer* PRenderer, const EBool PForce);
extern void        E3dScene_SampleLights(const E3dScene* PScene, E3dRenderer* PRenderer, const EChangeFlags PSceneChangeFlags, const int PAreaLightType, const EIndex PNumOfSamplesPerLight, const EBool PForce);



//----------------------------------------
// Accelerator.c
//----------------------------------------
typedef struct
{
 EIndex  TotalNumOfPolygons, TotalNumOfTriangles, GlobalTriangleIndex;
} E3dGetRenderTrianglesStats;

extern EObjectClass  E3dAccelerator_class[];

// Flags for E3dObject_GetRenderTriangles
//
#define E3dGrtINVERTED             0x0001
#define E3dGrtDO_INVISIBLE         0x0002
#define E3dGrtDO_WIREFRAME         0x0004
#define E3dGrtSHOW_FACES           0x0008
#define E3dGrtAREALIGHTS_ONLY      0x0010
#define E3dGrtUSE_LOD              0x0020
#define E3dGrtPREP_FOR_RAYTRACING  0x0040  // Prep RTriangles for CPU ray-tracing

extern int     E3dObject_GetRenderTriangles(E3dRenderer* PRenderer, E3dModel* PModel, E3dObject* PObject,  const uint16_t PFlags,  const E3dCoordinate* PVertexMatrix, const E3dCoordinate* PNormalMatrix,  E3dRPolyGroup*** PRenderPolyGroupsP, EIndex* PNumOfRenderPolyGroupsP, E3dGetRenderTrianglesStats* PStats);

#define E3dWORLD  0x01
#define E3dVIEW   0x02
extern int     E3dNode_GetRenderTriangles(E3dRenderer* PRenderer, E3dNode* PNode, const uint16_t PFlags, const uint8_t PTransformTo, const EBool PAreaLightsOnly, E3dRPolyGroup*** PRenderPolyGroupsP, EIndex* PNumOfRenderPolyGroupsP, E3dGetRenderTrianglesStats* PStats);

extern EIndex  E3dScene_GetRenderTriangles(E3dScene* PScene, E3dRenderer* PRenderer, const uint16_t PFlags, const EBool PFreeOld, const EBool PAreaLightsOnly);
extern EBool   E3d_RenderTrianglesFree(E3dRenderer* PRenderer);

extern int     E3dObject_RPolyGroupsUpdate(E3dRenderer* PRenderer, const E3dModel* PModel, const E3dObject* PObject,  const E3dVertex* PVertices, const E3dVertexSize PVertexSize,  const E3dCoordinate* PVertexMatrix, const E3dCoordinate* PNormalMatrix, E3dRPolyGroup** PRPolyGroups, const EIndex PMaxNumOfRPolyGroups, EIndex* PPGC, EIndex* PGlobalTriangleIndex);

extern int     E3dNode_RPolyGroupsUpdate(E3dRenderer* PRenderer, E3dNode* PNode,  const E3dVertex* PVertices, const E3dVertexSize PVertexSize, EIndex* PPGC, EIndex* PGlobalTriangleIndex);

extern int     E3dObject_RPolyGroupsUpdateMaterials(E3dNode* PNode, E3dObject* PObject, E3dRenderer* PRenderer, const E3dCoordinate* PWorldToViewerMatrix, const E3dCoordinate PSizeFactor, E3dRPolyGroup** PRPolyGroups, const EIndex PNumOfRPolyGroups, EIndex* PPGCP);
extern int     E3dScene_RPolyGroupsUpdateMaterials(E3dScene* PScene, E3dRayTracer* PRenderer);

extern void    E3d_RendererBuildAccelerator(E3dRayTracer* PRenderer);


//----------------------------------------
// Rendering/Lights.c
//----------------------------------------
#define E3dSAMPLING_TILES_PER_SET  42

extern void    E3d_SamplingTilesCreate(void);
extern void    E3d_SamplingTilesFree(void);

extern EIndex  _E3d_AddPointLights(const E3dRenderer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dSurfacePoint* PPoint,  const int PSpectralIndex, const int PNSpectralSamples,  E3dShaderLight* PShaderLight);
extern EIndex  _E3d_AddPointLightsSh(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracerThread, const E3dSurfacePoint* PPoint, const void* PExcludeFace,  const int PSpectralIndex, const int PNSpectralSamples,  E3dShaderLight* PShaderLight);

extern EIndex  E3d_AreaLightNumOfElements(const E3dAreaLight* PAreaLight);

extern EIndex  _E3d_AddAreaLightSh(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dSurfacePoint* PPoint, const void* PExcludeFace, const E3dAreaLight* PLight, const float PShadowBrightness, E3dGridRay* PShadowRay, const EInt32_3* PStartVoxelC,  const int PSpectralIndex, const int PNSpectralSamples,  const EIndex PShaderLightStartIndex);

extern EIndex  _E3d_AddAreaLights(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dSurfacePoint* PPoint, const int PSpectralIndex, const int PNSpectralSamples, const EIndex PShaderLightStartIndex);
extern EIndex  _E3d_AddAreaLightsSh(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dSurfacePoint* PPoint, const void* PExcludeFace, const int PSpectralIndex, const int PNSpectralSamples, const EIndex PShaderLightStartIndex);
extern EIndex  _E3d_AddAreaLightsOneSampleSh(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dSurfacePoint* PPoint, const void* PExcludeFace,  const int PSpectralIndex, const int PNSpectralSamples,  E3dShaderLight* PShaderLight);

extern EIndex  _E3d_AddAreaLights_Volume(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dSurfacePoint* PPoint,  const int PSpectralIndex, const int PNSpectralSamples,  const EIndex PLightSampleStartIndex, const EIndex PSampleNumDivider);
extern EIndex  _E3d_AddAreaLightsSh_Volume(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dSurfacePoint* PPoint, const void* PExcludeFace,  const int PSpectralIndex, const int PNSpectralSamples,  const EIndex PLightSampleStartIndex, const EIndex PSampleNumDivider);
extern EIndex  _E3d_AddAreaLightsOneSample_Volume(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dSurfacePoint* PPoint, const void* PExcludeFace,  const int PSpectralIndex, const int PNSpectralSamples,  E3dShaderLight* PShaderLight);
extern EIndex  _E3d_AddAreaLightsOneSampleSh_Volume(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PRayTracer_thread, const E3dSurfacePoint* PPoint, const void* PExcludeFace,  const int PSpectralIndex, const int PNSpectralSamples,  E3dShaderLight* PShaderLight);


extern void    E3d_ShadePoint(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PThread, const E3dRTriangle* PRTriangle, E3dSurfacePoint* PPoint, E3dStackRay* PRay,  const E3dMaterial* PSurfaceMaterial, const E3dShader* PVolumeShader, const E3dShader* PShaderBeyondSurface,  E3dShaderResult* PResultTmp, E3dShaderResult* PResult);
extern void    E3d_ShadePointShadows(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PThread, const E3dRTriangle* PRTriangle, E3dSurfacePoint* PPoint, E3dStackRay* PRay,  const E3dMaterial* PSurfaceMaterial, const E3dShader* PVolumeShader, const E3dShader* PShaderBeyondSurface,  E3dShaderResult* PResultTmp, E3dShaderResult* PResult);
extern void    E3d_ShadePointShadowsNoBRDF(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PThread, const E3dRTriangle* PRTriangle, E3dSurfacePoint* PPoint, E3dStackRay* PRay,  const E3dMaterial* PSurfaceMaterial, const E3dShader* PVolumeShader, const E3dShader* PShaderBeyondSurface,  E3dShaderResult* PResultTmp, E3dShaderResult* PResult);
extern void    E3d_ShadePointShadows_Volume(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PThread, const E3dRTriangle* PRTriangle, E3dSurfacePoint* PPoint, E3dStackRay* PRay,  const E3dMaterial* PSurfaceMaterial, const E3dShader* PVolumeShader, const E3dShader* PShaderBeyondSurface,  E3dShaderResult* PResultTmp, E3dShaderResult* PResult);

extern void    E3d_ShadeSurfaceIrradiance(const E3dShaderLight* PLights, const int PNLights,  E3dShaderResult* PResult);

extern void    (*E3d_ShadeSurfaceSphericalHarmonicsIrradianceFunc)(const E3dShaderLight* PLights, const int PNLights,  E3dShaderResult* PResult);
extern void    (*E3d_ShadePointSphericalHarmonicsIrradianceFunc)(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PThread, const E3dRTriangle* PRTriangle, E3dSurfacePoint* PSurfacePoint, E3dStackRay* PRay,  const E3dMaterial* PMaterial, const E3dShader* PVolumeShader, const E3dShader* PShaderBeyondSurface,  E3dShaderResult* PResultTmp, E3dShaderResult* PResult);

extern void    (*E3d_ShadeSurfaceSphericalHarmonicsRadianceFunc)(const E3dShaderLight* PLights, const int PNLights,  E3dShaderResult* PResult);
extern void    (*E3d_ShadePointSphericalHarmonicsRadianceFunc)(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PThread, const E3dRTriangle* PRTriangle, E3dSurfacePoint* PSurfacePoint, E3dStackRay* PRay,  const E3dMaterial* PMaterial, const E3dShader* PVolumeShader, const E3dShader* PShaderBeyondSurface,  E3dShaderResult* PResultTmp, E3dShaderResult* PResult);

extern void    (*E3d_ShadeSurfaceSphericalHarmonicsOcclusionFunc)(const E3dShaderLight* PLights, const int PNLights,  E3dShaderResult* PResult);
extern void    (*E3d_ShadePointSphericalHarmonicsOcclusionFunc)(const E3dRayTracer* PRenderer, E3dRayTracer_thread* PThread, const E3dRTriangle* PRTriangle, E3dSurfacePoint* PSurfacePoint, E3dStackRay* PRay,  const E3dMaterial* PMaterial, const E3dShader* PVolumeShader, const E3dShader* PShaderBeyondSurface,  E3dShaderResult* PResultTmp, E3dShaderResult* PResult);

extern void    E3d_ShadeSurfaceWavelets(const E3dShaderLight* PLights, const int PNLights,  E3dShaderResult* PResult);


extern E3dShaderClassRendererBinding*  E3dShader_ClassAddRenderer(E3dShaderClass* PShaderClass, E3dRendererClass* PRendererClass, const E3dShaderClassRendererBinding* PDescriptor);
extern E3dShaderClassRendererBinding*  E3dShader_ClassGetRendererBinding(const E3dShaderClass* PShaderClass, const E3dRendererClass* PRendererClass);
extern void                            E3dShader_ClassRemoveRenderer(E3dShaderClass* PShaderClass, const E3dRendererClass* PRendererClass);


// Language-agnostic shader cores for built-in shaders
extern const char*  Material_Constant_core_sl;
extern const char*  Material_Lambert_core_sl;
extern const char*  Material_Phong_core_sl;


extern EResource  E3dSL_ConstantResources[];
extern EResource  E3dSL_LambertResources[];
extern EResource  E3dSL_PhongResources[];


extern E3dShaderVariant*  (*E3dShader_GenerateCode)(const E3dGeometry* PGeometry, const E3dShadeable* PShadeable, E3dRenderer* PRenderer,
 E3dShaderClassRendererBinding* PShaderClassBinding,
 const E3dShaderPassesMask PPassesMask,
 const E3dLightInstance* PLights, const int PNumOfLights, E3dShadowMap** PShadowMaps,
 EResource** PFrgResourcesPtr, EIndex* PNumOfFrgResourcesPtr,
 EResource** PVtxResourcesPtr, EIndex* PNumOfVtxResourcesPtr,
 const EBool PAddComments);

extern void   (*E3dShader_UpdateNormalMap)(E3dRenderer* PRenderer, E3dMaterial* PMaterial, E3dShader* PShader, E3d2DTexture* PTexture, E3dTextureMapper* PMapper);


extern void  E3dGL_HierarchyDraw(E3dWindow* P3DWindow, E3dRenderer* PRenderer, const uint32_t PSubCameraID, const E3dDrawContext* PDrawContext, const E3dDisplayFlags PDisplayFlags,  E3dNode* PRootNode, const E3dMatrix PWorldToViewerMatrix, const unsigned char POpaqueTransparent, const EBool PDrawCamera);
extern void  E3dGL_WindowSetViewTransform(E3dWindow* P3DWindow, E3dRenderer* PRenderer, const uint32_t PSubCameraID, const EIndex PAAMult,  E3dMatrix PWorldToViewerMatrix);

//----------------------------------------
// GL/DrawMesh.c
//----------------------------------------
extern int  E3dGL_DrawConstantColor(E3dGPUContext* PGPUContext, E3dRenderer* PRenderer, const E3dDisplayFlags PDisplayFlags,  E3dScene* PScene, E3dModel* PModel, E3dGeometry* PGeometry, E3dShadeable* PShadeable);
extern int  E3dGL_DrawLambert(E3dGPUContext* PGPUContext, E3dRenderer* PRenderer, const E3dDisplayFlags PDisplayFlags,  E3dScene* PScene, E3dModel* PModel, E3dGeometry* PGeometry, E3dShadeable* PShadeable);
extern int  E3dGL_DrawPhong(E3dGPUContext* PGPUContext, E3dRenderer* PRenderer, const E3dDisplayFlags PDisplayFlags,  E3dScene* PScene, E3dModel* PModel, E3dGeometry* PGeometry, E3dShadeable* PShadeable);

//----------------------------------------
// GL/FrameBuffer.c
//----------------------------------------
enum
{
	E3dBloom2D_POWER = 1,
	E3dBloomSEPARABLE_POWER,
	E3dBloomSEPARABLE_GAUSSIAN,
};

extern void  E3dGL_UpdateBloomKernel(E3dRenderer* PRenderer, E3dBloom* PBloom, const int PBloomRadius);

//----------------------------------------
// E3D/Panel/Renderer.c
//----------------------------------------
extern void  E3dRenderer_UpdateStats(E3dRenderer* PRenderer);
extern void  E3d_DrawContextUpdateFromRenderer(const E3dRenderer* PRenderer, E3dDrawContext* PDrawContext);





#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_Rendering_h
