//================================================================
// E3D
//
// Shading language (GLSL, Cg etc.) common definitions
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 15:09:36
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_SL_h
#define _E3D_SL_h

#include <E3D/Light.h>  // For E3dShadowMap

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


// Global state for error reporting
//
typedef struct
{
	char*              StateString;
	E3dMaterial*       Material;
	E3dShader*         Shader;
	E3dShaderVariant*  ShaderVariant;
	char*              ParameterName;

	const char*        File;
	const char*        Function;int	Line;
	E3dRenderer*       Renderer;
} E3dSLDebugRec;



typedef struct E3dShaderDevice_s  E3dShaderDevice;

#ifndef _E3dLightInstance
#define _E3dLightInstance
typedef struct E3dLightInstance_S  E3dLightInstance;
#endif


typedef char*  (*E3dGetShaderSignatureFunc)(const E3dShaderDevice* PShaderDevice, E3dShaderClassRendererBinding* PShaderClassBinding, const E3dGeometry* PGeometry, const E3dShadeable* PShadeable, E3dRenderer* PRenderer,  const E3dShaderPassesMask PPassesMask,  const E3dLightInstance* PLights, const int PNLights, E3dShadowMap** PShadowMaps,  char* PID);
typedef void   (*E3dShaderVariantInitFunc)(E3dShaderVariant* PShaderVariant);
typedef int    (*E3dVertexProgramInitFunc)(E3dStageProgram*  PVertexProgram, E3dMaterial* PMaterial, EResource* PResources, E3dTextureMapper* const* PTxMappers, const E3dTxMapperIndex* PTextureMapperIndices, const int PNumOfTextures);
typedef void   (*E3dStageProgramFreeFunc)(E3dStageProgram* PProgram);
typedef void   (*E3dShaderCheckNeededTermsFunc)(const E3dShadeable* PShadeable, const E3dShaderClassRendererBinding* PShaderClassBinding, E3dShaderVariant* PShaderVariant, E3dSLCodeGenParams* PCodeGenParams);

#define E3dVERTEX_PROGRAM    0x01
#define E3dFRAGMENT_PROGRAM  0x02


#define E3dSLMAX_NUM_OF_LIGHTS              1  // Max # of lights per shader pass. Must be <= E3dMAX_NUM_OF_LIGHTS
#define E3dSLMAX_NUM_OF_2DTEXTURES         16  // Max # of texture samplers. Used for statically allocated arrays. E3dGL_MAX_TEXTURE_IMAGE_UNITS provides the actual number.
#define E3dSLMAX_NUM_OF_PARAMETERS        256
#define E3dSLMAX_NUM_OF_ARRAY_PARAMETERS   32
#define E3dSLMAX_NUM_OF_SKINNING_JOINTS    30

#define E3dSL_COMPILE_ERROR         0x01
#define E3dSL_SYMBOL_RESOLVE_ERROR  0x02
#define E3dSL_ALWAYS                0x04

//------------------------------------------------
// Shader target back-end description
// This struct is specified under a RendererClass and it describes how to generate Shader code for that RendererClass.
//------------------------------------------------
struct E3dShaderDevice_s
{
	EBool                          CodeGeneratorInitialized;

	char*                          LanguageVersionString;

	const char*                    FragmentFunctionPrefix;
	const char*                    FragmentHeader;
	const char*                    FragmentEntryPointName;

	EBool                          ParametersGlobal;                 // Vertex/fragment program uniforms are global variables (GLSL)
	EBool                          SupportsConstForLocalVariables;

	const char*                    VertexHeader;

	const char*                    Precision;

	const char*                    VertexAttributeQualifier;
	const char*                    VertexAttributePrefix;
	const char*                    VertexProgramOutParamQualifier;
	const char*                    VUParamQualifier;
	const char*                    BuiltIn_VertexPositionInParamName;const char*  BuiltIn_VertexPositionOutParamName;
	const char*                    BuiltIn_VertexColorInParamName;
	const char*                    BuiltIn_VertexNormalInParamName;const char*  BuiltIn_VertexNormalOutParamName;
	const char*                    BuiltIn_VertexModelViewProjectionMatrixParamName;
	const char*                    BuiltIn_VertexTexCoordInParamName;

	const char*                    BuiltIn_FragmentOutColorName;

	const char*                    VFParamPrefix;                    // Prefix for vertex-fragment interface parameters (a.k.a. interpolants / varying params)
	const char*                    VUParamPrefix;                    // Prefix for vertex program uniform parameters
	const char*                    FUParamQualifier;                 // Fragment program uniform parameter qualifier. E.g. "uniform"
	const char*                    FUParamPrefix;                    // Prefix for fragment program uniform parameters

	EBool                          HasSwizzle;                       // Language supports a.xyz etc.
	char                           XName, YName, ZName, WName,       // Vector component names
																 RName, GName, BName, AName;

	const char*                    Param_DefFormatString;
	const char*                    Sampler2D_DefFormatString;
	const char*                    Sampler2DShadow_DefFormatString;
	EBool                          ShadowMappingAutoColor;           // If TRUE, looking up a texel in a shadow map will already do the disntance check

	EBool                          Sampler_PassCoordinatesAsVector;
	EBool                          VectorOperators;                  // If TRUE, we can use standard operators on vectors (+,-,* etc.)
	const char*                    WorldShadingNormalString;
	int                            MaxNumOfTextures;
	E3dGetShaderSignatureFunc      GetShaderSignature;
	EResResourceHandlerProc        RT_sampler2D, RT_sampler2DShadow, RT_sampler3D, RT_samplerCUBE, RT_float3x3, RT_float3x4;  // Resource translators for special data types

	EBool                          AccumulateLightPerShadeCall;

	EpPixelFormat                  NormalMapPixelFormat;

	E3dShaderVariantInitFunc       ShaderVariantInit;
	E3dVertexProgramInitFunc       VertexProgramInit;
	E3dStageProgramFreeFunc        VertexProgramFree;
	int                            VertexProgramSize;

	int                            FragmentProgramSize;
	E3dShaderCheckNeededTermsFunc  ShaderCheckNeededTerms;
};



//
// Shading-language-independent macros
//
#define E3dSLM_ShaderResolveUniforms(mMaterial,  mTxMappers, mTextureMapperIndices) do\
{\
E3dSL_ClearDebugRec();\
\
	{ char  lParamName[256];sprintf(lParamName, "%sEyeLocalPosition", LFUParamPrefix);E3dSLM_GetNamedParameterFrag(PMaterial, lParamName, LThisShaderVariant->Param_EyeLocalPosition, FALSE); }\
	{ char  lParamName[256];sprintf(lParamName, "%sEyeWorldPosition", LFUParamPrefix);E3dSLM_GetNamedParameterFrag(PMaterial, lParamName, LThisShaderVariant->Param_EyeWorldPosition, FALSE); }\
\
	{\
		int  lTxC = 0;for(;lTxC < LNumOfTextures;lTxC++)\
		{\
			const E3dTxMapperIndex        lMapperIndex = mTextureMapperIndices[lTxC];\
			const E3dTextureMapper*       lTextureMapper = mTxMappers[lMapperIndex];\
			const E3dTextureMapperClass*  lTextureMapperClass = lTextureMapper->Class;\
			if(lTextureMapperClass == E3dTextureMapperCubeMapReflection_class || lTextureMapperClass == E3dTextureMapperCubeMapRefraction_class)\
			{\
				if(lTextureMapperClass == E3dTextureMapperCubeMapReflection_class && lTextureMapper->BRDFSamplesSqrt > 0 && lTextureMapper->EnvLightFactor != 0.0)\
				{\
					{\
						char  lParamName[256];sprintf(lParamName, "%sTexture%dBRDFSamplesSqrt", LFUParamPrefix, lTxC);\
						E3dSLM_GetNamedParameterFrag(PMaterial, lParamName, LThisShaderVariant->TextureBRDFSamplesSqrtParam[lTxC], TRUE);\
/*printf("%s() %s:%d  %s %d\n", __FUNCTION__, __FILE__, __LINE__, lParamName, LThisShaderVariant->TextureBRDFSamplesSqrtParam[lTxC]);fflush(stdout);*/\
					}\
					if(lTextureMapper->EnvLightFactor != 1.0)\
					{\
						char  lParamName[256];sprintf(lParamName, "%sTexture%dEnvLightFactor", LFUParamPrefix, lTxC);\
						E3dSLM_GetNamedParameterFrag(PMaterial, lParamName, LThisShaderVariant->TextureEnvLightFactorParam[lTxC], TRUE);\
					}\
				}\
				break;\
			}\
		}\
	}\
\
	ENewNClr(LNumOfTextures, LThisShaderVariant->TextureBlendingParameters);\
\
/*printf("%s() %s %d\n", __FUNCTION__, LShaderVariant->Signature, LNumOfTextures);fflush(stdout);*/\
\
	{\
		int  lTxC = 0;for(;lTxC < LNumOfTextures;lTxC++)\
		{\
			const E3dTxMapperIndex  lMapperIndex = mTextureMapperIndices[lTxC];\
			if(mTxMappers[lMapperIndex]->Class == E3dTextureMapperCubeMapRefraction_class)\
			{\
				char  lName[256];sprintf(lName, "%sEta%d", LFUParamPrefix, (int)lTxC);\
				E3dSLM_GetNamedParameterFrag(PMaterial, lName, LThisShaderVariant->Eta[lTxC], TRUE);\
			}\
\
			const EIndex  lNumOfBlendersPerTexture = LThisShaderVariant->NumOfBlendersPerTexture;\
			ENewNClr(lNumOfBlendersPerTexture + 1, LThisShaderVariant->TextureBlendingParameters[lTxC]);    /* +1 for Alpha_Factor */\
\
/*printf("%s() %s:%d  TextureBlenderLists %p\n", __FUNCTION__, __FILE__, __LINE__, LShaderVariant->TextureBlenderLists);fflush(stdout);*/\
			if(LShaderVariant->TextureBlenderLists)\
			{\
				E3dSLParameter*           lSLParameters = LThisShaderVariant->TextureBlendingParameters[lTxC];\
				char                      lTmpStr[256];\
				const E3dTextureBlender*  lTransparentBlender = NULL;\
\
				E3dTextureBlender*        lBlender = LShaderVariant->TextureBlenderLists[lTxC];\
				EIndex  lBC = 0;for(;lBC < lNumOfBlendersPerTexture;lBC++, lBlender++)\
				{\
					if(lBlender->Factor != 0.0 && lBlender->Channels)\
					{\
						sprintf(lTmpStr, "%s%d%s_Factor", LFUParamPrefix, (int)lTxC, lBlender->ShaderResource->Name);\
						E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLParameters[lBC], TRUE);\
/*printf(" %s() %s:%d  %2d  %f %04x  %s  %d\n", __FUNCTION__, __FILE__, __LINE__, lBC, lBlender->Factor, lBlender->Channels, lTmpStr, lSLParameters[lBC]);fflush(stdout);*/\
					}\
					else lSLParameters[lBC] = E3dSLParameterNONE;\
					if((EPointerInt)(lBlender->ShaderResource->Address) == EPtrOffset(E3dShader, Transparent)) lTransparentBlender = lBlender;\
				}\
\
				if(lTransparentBlender && lTransparentBlender->Factor != 0.0)\
				{\
					sprintf(lTmpStr, "%s%dAlpha_Factor", LFUParamPrefix, (int)lTxC);\
					E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLParameters[lBC], TRUE);\
				}\
				else lSLParameters[lBC] = E3dSLParameterNONE;\
			}\
		}\
	}\
\
	{\
		int  lC = 0;E3dSLLight*  lSLLight = LThisShaderVariant->Lights;for(;lC < LNLightsForPass;lC++, lSLLight++)\
		{\
			char         lTmpStr[256];\
			const char*  lCoordSpaceStr = LCodeGenParams->UseWorldSpace ? "World" : "Local";\
\
			if(LCodeGenParams->NeedLightColor)\
			{\
				sprintf(lTmpStr, "%sLight%d%sPosition", LFUParamPrefix, (int)lC, lCoordSpaceStr);E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLLight->Param_Position, TRUE);\
				sprintf(lTmpStr, "%sLight%dColor", LFUParamPrefix, (int)lC);E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLLight->Param_Color, TRUE);\
				if(LCodeGenParams->NeedLightDiffuseSpecularFactors) { sprintf(lTmpStr, "%sLight%dDiffuseSpecularFactors", LFUParamPrefix, (int)lC);E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLLight->Param_DiffuseSpecularFactors, TRUE); }\
\
				switch(LCodeGenParams->LightTypes[lC])\
				{\
					case E3dLightSPOT:\
/* Spotlight params */\
						sprintf(lTmpStr, "%sLight%d%sDirection", LFUParamPrefix, (int)lC, lCoordSpaceStr);\
						E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLLight->Param_SpotDirection, TRUE);\
\
						sprintf(lTmpStr, "%sLight%dSpotExponent", LFUParamPrefix, (int)lC);\
						E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLLight->Param_SpotExponent, TRUE);\
\
						sprintf(lTmpStr, "%sLight%dSpotCutoffAngleCos", LFUParamPrefix, (int)lC);\
						E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLLight->Param_SpotCutoffAngleCos, TRUE);\
\
						sprintf(lTmpStr, "%sLight%dSpotAngle", LFUParamPrefix, (int)lC);\
						E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLLight->Param_SpotAngle, TRUE);\
\
						sprintf(lTmpStr, "%sLight%dSpotPenumbraAngle", LFUParamPrefix, (int)lC);\
						E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLLight->Param_SpotPenumbraAngle, TRUE);\
					break;\
				}\
			}\
\
			if(LShadowMaps[lC])\
			{\
				sprintf(lTmpStr, "%sShadowMap%d", LFUParamPrefix, (int)lC);\
\
				E3dSL_DebugRec.ParameterName = lTmpStr;\
				E3dSLM_GetNamedParameterFrag(PMaterial, lTmpStr, lSLLight->Param_ShadowMap, TRUE);\
			}\
		}\
	}\
\
	if(LShader->IsTransparent) { char  lParamName[256];sprintf(lParamName, "%sAlpha", LFUParamPrefix);E3dSLM_GetNamedParameterFrag(PMaterial, lParamName, LThisShaderVariant->AlphaParam, FALSE); }\
	if(LCodeGenParams->HasFresnel) { char  lParamName[256];sprintf(lParamName, "%sFresnel0_RGB", LFUParamPrefix);E3dSLM_GetNamedParameterFrag(PMaterial, lParamName, LThisShaderVariant->Fresnel0_RGBParam, FALSE); }\
\
	if(LShaderClassBinding->Resources)\
	{\
		int  lC = 0;EResource*  lResource = LShaderClassBinding->Resources;for(;lResource->Type;lC++, lResource++)\
		{\
			if(lResource->Semantic == NULL)\
			{\
				char  lName[256];sprintf(lName, "%s%s", LFUParamPrefix, lResource->Name);\
				E3dSL_ClearDebugRec();\
				E3dSL_Location();\
				E3dSLM_GetNamedParameterFrag(mMaterial, lName, LThisShaderVariant->FragmentUniformParams[lC], TRUE);\
/*printf("%s() %s:%d  %s %d\n", __FUNCTION__, __FILE__, __LINE__, lResource->Name, LThisShaderVariant->FragmentUniformParams[lC]);fflush(stdout);*/\
\
/* For each texture, resolve the blending weight parameter for this shader resource	*/\
				int  lTxC = 0;for(;lTxC < LNumOfTextures;lTxC++)\
				{\
/*				LTexture = (mMaterial)->Textures[lTxC];*/\
				}\
			}\
		}\
	}\
\
	int  lTxC = 0;for(;lTxC < LNumOfTextures;lTxC++)\
	{\
		E3d2DTexture*  lTexture = (mMaterial)->Textures[lTxC];\
\
/* E3dSLTEXTURE	*/\
		const E3dTxMapperIndex  lMapperIndex = mTextureMapperIndices[lTxC];\
		if(E3dSL_IsTextureUsed(mMaterial, lTxC, mTxMappers[lMapperIndex], LShaderVariant))\
		{\
			char  lName[256];sprintf(lName, "%sTexture%d", LFUParamPrefix, lTxC);\
			E3dSL_Location();\
			E3dSLM_GetNamedParameterFrag(mMaterial, lName, LThisShaderVariant->TextureParam[lTxC], TRUE);\
			E3dSL_ClearDebugRec();\
		}\
		else LThisShaderVariant->TextureParam[lTxC] = 0;\
		if(lTexture->DoNormalMap && !(lTexture->ImageAsNormalMap))\
		{\
			char  lName[256];sprintf(lName, "%sNormalMap%d", LFUParamPrefix, lTxC);\
			E3dSL_Location();\
			E3dSLM_GetNamedParameterFrag(mMaterial, lName, LThisShaderVariant->NormalMapParam[lTxC], TRUE);\
			E3dSL_ClearDebugRec();\
		}\
	}\
} while(0)


extern E3dSLDebugRec  E3dSL_DebugRec;          // Struct for debugging shaders
extern EBool          E3dSL_ProceduralPlanar;


static inline void E3dSL_ClearDebugRec(void)
{
	memset(&E3dSL_DebugRec, 0, sizeof(E3dSLDebugRec));
}


#define E3dSL_Location() do\
{\
	E3dSL_DebugRec.Function = __FUNCTION__;E3dSL_DebugRec.File = __FILE__;E3dSL_DebugRec.Line = __LINE__;\
} while(0)


extern const char  E3dSLSemantic_Position[];
extern const char  E3dSLSemantic_Normal[];
extern const char  E3dSLSemantic_Tangent[];
extern const char  E3dSLSemantic_Bitangent[];

extern const char  E3dSLSemantic_PointToEye[];
extern const char  E3dSLSemantic_EYE_Position[];
extern const char  E3dSLSemantic_Fresnel[];

extern const char  E3dSLSemantic_LightColor[];
extern const char  E3dSLSemantic_LightDiffuseSpecularFactors[];

extern char* (*E3dShaderVariant_GetSignature)(const E3dShaderDevice* PShaderDevice, E3dShaderClassRendererBinding* PShaderClassBinding,  const E3dGeometry* PGeometry, const E3dShadeable* PShadeable,  E3dRenderer* PRenderer,
 const E3dShaderPassesMask PPassesMask,
 const E3dLightInstance* PLights, const int PNLights, E3dShadowMap** PShadowMaps,
 char* PID);

extern EBool  (*E3dSL_IsTextureUsed)(const E3dMaterial* PMaterial, const EIndex PTxC, const E3dTextureMapper* PMapper, const E3dShaderVariant* PShaderVariant);
extern void   (*E3dSL_ShaderCheckNeededTerms)(const E3dShadeable* PShadeable, const E3dShaderClassRendererBinding* PRendererBinding, E3dShaderVariant* PShaderVariant, E3dSLCodeGenParams* PCodeGenParams);
extern void   (*E3dSL_FrgSaveSource)(const E3dMaterial* PMaterial, const E3dShaderVariant* PShaderVariant, const char* PSuffix);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_SL_h
