//================================================================
// E3D library
//
// Scene management
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 19:23:18
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Scene_h
#define _E3D_Scene_h


#ifndef _ELists_h
#include <ELists.h>
#endif

#ifndef _EUndo_h
#include <EUndo.h>
#endif

#ifndef _E3D_Animation_h
#include <E3D/Animation.h>
#endif

#ifndef _E3D_Item_h
#include <E3D/Item.h>
#endif

#ifndef _E3D_Ray_h
#include <E3D/Ray.h>
#endif

#ifndef _E3D_Shape_h
#include <E3D/Shape.h>
#endif


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

#define E3dSELECTED                 0x01
#define E3dUNSELECTED               0x02
#define E3dVISIBLE                  0x04
#define E3dHIDDEN                   0x08
#define E3dEXCLUDE_MODEL_SELECTION  0x10  // For E3dScene_GetObjects(). Don't consider an Object selected if one of its referencing Nodes is selected, but the Object itself is not


// How to add a newly created Object to the Scene
//
enum
{
	E3dNEW_INSTANCE = 1,  // Just add new Object to Model
	E3dNEW_MODEL,         // Add new Model with Object and add it to parent of Model
	E3dNEW_HRC            // Add new Model with Object and add it to Scene after root of Model
};



// Node-change Scene callback-struct
//
typedef struct
{
	E3dCallbackStructCore;

	E3dNode**  Nodes;EIndex  NumOfNodes;
	void*      Value;
} E3dSceneNodesCallbackStruct;


// Node-change Scene callback-struct
//
typedef struct
{
	E3dCallbackStructCore;

	E3dObject*  Object;
	EObject*    SubObject;  // E.g a PolyGroup when Object is a Mesh

	EIndex      ContourIndex, PointIndex;
} E3dSceneObjectCallbackStruct;


// Material-change Scene callback-struct
//
typedef struct
{
	E3dCallbackStructCore;

	E3dMaterial**  Materials;EIndex  NumOfMaterials;
} E3dSceneMaterialsCallbackStruct;


#ifndef _E3dScene
#define _E3dScene
typedef struct E3dScene_S	  E3dScene;
#endif

#ifndef _E3dLightInstance
#define _E3dLightInstance
typedef struct E3dLightInstance_S  E3dLightInstance;
#endif

typedef struct
{
	E3dShadowMap*  ShadowMap;
	E3dCoordinate  LODSizeFactor;
	E3dMatrix      LightViewMatrix, LightViewMatrixInv,
	               LightFrustumMatrix, LightFrustumMatrixInv,
	               ShadowProjectionMatrix;         // World-space shadow Matrix
	EBool          Valid;
} E3dShadowMapT;

// E3dLights can be instantiated from Models and this
// struct is used to store unique instances (in E3dScene)
//
struct E3dLightInstance_S
{
	E3dModel*      Model;          // Instantiating Model
	E3dLight*      Light;          // Pointer to the instantiated Light
	E3dCoord3      Position;       // World position
	E3dCoord3      Direction;      // In world space (spot lights / directional lights)

	int            Type;           // Light source type
	EcRGBf32Color  Color;          // Color of the light source
	float          DiffuseFactor, SpecularFactor;

	float          SpotAngle, SpotPenumbraAngle, SpotExponent,
	               ConstantAttenuation,
	               LinearAttenuation,
	               QuadraticAttenuation,
	               ShadowBrightness;  // If a point is in shadow from this light, its diffuse color will be multiplied by this value

	float          SpotCutoffAngleCos;

	E3dShadowMapT  ShadowMap;
};


typedef struct
{
	E3dMaterial*  Material;
	EListA(, EObject*, Shadeables);
} E3dMaterialAndShadeable;


typedef struct
{
	char*  Name;
} E3dLayer;


struct E3dScene_S
{
	char*          Name;
	char*          URL;

	EListA(, E3dNode*,      RootNodes);
	EListA(, E3dMaterial*,  Materials);
	EListA(, E3d2DTexture*, Textures);

// Point / spot etc. lights that are visible and have a non-black color
//
	EList(, E3dLightInstance, VisibleLights);

	EList(, ECallbackRec, Callbacks);

#ifdef USEOpenGL
	EList(, E3dGLLight, GLLights);
#endif // USEOpenGL


// For Matrix-stack operations
//
	int            MatrixPtr, MatrixScalePtr;

	E3dCoord3      CurrentScale;
	E3dCoord3      ScaleStack[E3dMAX_HIERARCHY_DEPTH];
	E3dMatrix      MatrixStack[E3dMAX_HIERARCHY_DEPTH];


// For animation
//
	EList(, E3dAnimStack*, AnimStacks);
	E3dAnimStack*  CurrentAnimStack;

	EListA(, E3dAnimation*, GlobalAnimations);  // Animations that are not specific to a certain AnimStack. E.g. rigid body physics

// Change-tracking for animation
//
	E3dChangeList  ChangeList;
};


typedef EChangeFlags (*E3dSceneFunction)(E3dScene* PScene);

//----------------------------------------------------------------
// Global variables
//----------------------------------------------------------------
extern E3dScene*  E3d_Scene;

extern E3dMatrix  E3d_IdentityMatrix;


//----------------------------------------------------------------
// Rendering/Rendering.c
//----------------------------------------------------------------
// 3D Icons
//
extern E3dMesh*    E3d_3DCrossMesh;
extern E3dObject*  E3d_ManipulatorRot;
extern E3dObject*  E3d_ManipulatorTrn;


//----------------------------------------------------------------
// Scene.c
//----------------------------------------------------------------
extern EBool        E3d_Exiting;

extern const void*  E3d_Classes[];extern const int  E3d_Classes_Count;

extern EIndex       E3dScene_CollectClassesAndObjects(E3dScene* PScene,  E3dNode** PNodeBranches, const EIndex PNNodeBranches,  EResContext* PContext, const uint16_t PAddToAllObjects,  E3dObject*** PObjectsP);
extern ECoreClass*  E3d_ClassFindByName(const char* PClassName);

EList(extern, E3dLayer, E3d_Layers);
extern int  E3d_CurrentLayerIndex;

extern void         E3dScene_Init(E3dScene* PScene);
extern E3dScene*    E3dScene_New(const char* PName);
extern void         E3dScene_Free(E3dScene* PScene);

extern EBool         E3dScene_HasSelectedAnimationChannel(const E3dScene* PScene);

extern void          E3dLightInstance_Clear(E3dLightInstance* PLightInstance);


extern char*         E3dScene_GetUniqueNodeName(E3dScene* PScene, const char* PBaseName);
extern char*         E3dScene_GetUniqueObjectName(E3dScene* PScene, const char* PBaseName);


extern E3dLight*     E3dScene_LightAdd(E3dScene* PScene, int PType);


extern EBool         E3dScene_NodeHrcAdd(E3dScene* PScene, E3dNode* PRootNode, const EUpdateFlags PNotifyFlags);
extern EBool         E3dScene_NodeHrcInsert(E3dScene* PScene, E3dNode* PRootNode, const EIndex PHere, const EBool PNotify);
extern EBool         E3dScene_NodeHrcRemove(E3dScene* PScene, E3dNode* PRootNode, EChangeFlags* PChangedRet, const EBool PNotify);
 
extern EBool         E3dScene_MaterialAdd(E3dScene* PScene, E3dMaterial* PMaterial);
extern EIndex        E3dScene_MaterialsAdd(E3dScene* PScene, E3dMaterial** PMaterials, const EIndex PCount);

extern EIndex        E3dScene_MaterialsAddFromObject(E3dScene* PScene, E3dObject* PObject);
extern EIndex        E3dScene_MaterialsAddFromNodeHrc(E3dScene* PScene, E3dNode* PRootNode);
extern EBool         E3dScene_MaterialRemove(E3dScene* PScene, E3dMaterial* PMaterial);
extern EIndex        E3dScene_MaterialsRemove(E3dScene* PScene, E3dMaterial** PMaterials, const EIndex PCount);

extern E3dMaterial*  E3dScene_MaterialCopy(E3dScene* PScene, const E3dRenderer* PRenderer, const E3dMaterial* PMaterial, const EBool PCopyCaches);

extern EIndex        E3dScene_MaterialCollectShadeables(E3dScene* PScene, const E3dMaterial* PMaterial, E3dShadeable*** PShadeablesPtr);
extern EIndex        E3dScene_CollectMaterialsAndShadeables(E3dScene* PScene, E3dMaterialAndShadeable** PList);

extern EBool         E3dScene_TextureAdd(E3dScene* PScene, E3d2DTexture* PTexture);
extern EBool         E3dScene_TextureRemove(E3dScene* PScene, E3d2DTexture* PTexture);

extern int           E3dScene_TextureCollectShadeables(E3dScene* PScene, E3d2DTexture* PTexture, E3dShadeableRec*** PShadeablesPtr);

extern int           E3dScene_ImageLoad(E3dScene* PScene, const char* PPath,  E3dImage** PImageRet);
extern int           E3dScene_TextureImageLoad(E3dScene* PScene, E3d2DTexture* PTexture, const char* PPath,  const double PFrame,  EBool PForceReRead,  E3dImage** PImageRet);

extern EBool         E3dScene_NodeHrcHasTextures(const E3dNode* PRootNode);
extern EIndex        E3dScene_NodeHrcReadTextures(E3dScene* PScene, E3dNode* PNode, const char* PPath, const double PFrame);

extern void          E3dMaterial_UpdateTextureImages(E3dScene* PScene, E3dMaterial* PMaterial, const char* PPath, const double PFrame);

extern E3dObject*    E3dObject_Copy(const E3dObject* PObject, const ECopyFlags PFlags, E3dScene* PScene);
extern void          E3dObject_CopyTo(const E3dObject* PObject, E3dObject* PNewObject, const ECopyFlags PFlags, E3dScene* PScene);

extern EChangeFlags  E3dObject_UpdateForDisplay(E3dObject* PObject, const EUpdateFlags PUpdateFlags, E3dScene* PScene);

extern E3dObject*    E3dObject_CustomNew(E3dScene* PScene, E3dNode* PParent, E3dModel* POrigModel, const char* PNewModelName, E3dObjectComponent* PInfo, const EUpdateFlags PUpdateFlags, const char PAddMode, const EBool PCreateUniqueName,  E3dModel** PNewModelP);
extern E3dObject*    E3dObject_CustomNewInit(E3dScene* PScene, void* PComponentTemplate, const EChangeFlags PUpdateFlags, const EBool PCreateUniqueName);

extern E3dModel*     E3dModel_CustomNew(E3dScene* PScene, E3dNodeComponentClass* PClass, const EBool PCreateUniqueName);
extern E3dModel*     E3d_ModelCustomObjectNewInit(E3dScene* PScene, void* PComponentTemplate, E3dObject** PObjectRet, const EChangeFlags PUpdateFlags, const char* PName, const EBool PCreateUniqueName);
extern E3dModel*     E3dModel_CustomObjectNew(E3dScene* PScene, E3dObjectComponentClass* PClass, E3dObject** PObjectRet, const EBool PCreateUniqueName);


extern int               E3dScene_ScriptFunctionRegister(const char* PName, const char* PUIName, const char* PToolTip, void* PFunc,  EScriptFunction** PFuncStruct);
extern int               E3dScene_ScriptFunctionRemove(EScriptFunction* PFuncStruct);
extern EScriptFunction*  E3dScene_ScriptFunctionByName(const char* PName);


extern EChangeFlags  E3dSpline_Changed(E3dSpline* PSpline, const EChangeFlags PChangeFlags,  EOpNode** POpNodesP, EIndex* POpNodes_CountP, EIndex* POpNodes_CountAllocatedP);


extern void          E3d_Init(void);
extern void          E3d_Free(void);
extern void          E3dScene_Clear(E3dScene* PScene, const EBool PAll, const EBool PRemoveResidentAnims);
extern void          E3dScene_InvalidateShadowCaches(E3dScene* PScene);
extern void          E3dScene_RefreshLights(E3dScene* PScene, E3dRenderer* PRenderer);

extern void          E3dLayer_Init(E3dLayer* PLayer);
extern void          E3dLayerFree_Contents(E3dLayer* PLayer);
extern void          E3dLayers_Free(E3dLayer** PLayers, EIndex* PLayers_Count);
extern void          (*E3d_LayerFromScene)(E3dScene* PScene,  const uint8_t PAction, const int PIndex, const int POldLayerCount, const int* PReindexingTable,  E3dLayer* PLayer);

extern EBool         E3d_IsObject(ECoreObject* PObject);
extern EBool         E3dObject_TypePass(const ECoreObject* PObject, const uint32_t PObjectTypeFlags);

extern EChangeFlags  E3d_SceneFromLayer(const int PLayerIndex, E3dScene* PScene);


extern EIndex        E3dScene_GetAnimatedObjects(E3dScene* PScene, const uint32_t PObjectTypeFlags, E3dAnimatedObject** PListRet);
extern void          E3d_AnimatedObjectsFree(E3dAnimatedObject* PAO, const EIndex PNAO);

extern EIndex        E3dSpline_CVInsertBetween(E3dSpline* PSpline, E3dMatrix PWorldToLocalMatrix,  E3dCoord3* P3Position, E3dRay* PRay, E3dSplineCV* PCV);

extern EBool         E3dScene_ObjectIsAnimated(E3dScene* PScene, ECoreObject* PObject);


//--------------------------------
// NodeHrc.c
//--------------------------------
extern E3dNode*      E3dNodeHrc_BranchCopy(E3dScene* PScene, E3dNode* PRootNode, const EFlags PCopyFlags, const EBool PInstance, const uint32_t PObjectTypeFlags, const EBool PVisibleOnly, const EBool PUpdate);
extern EChangeFlags  E3dNodeHrc_CloneInfo(E3dScene* PScene, E3dNode* PRootNode, E3dNode* PNewRootNode, const EFlags PFlags);



//--------------------------------
// Surface.c
//--------------------------------
typedef struct
{
	E3dObject*          Object;
	E3dModel*           Model;

	EBool               Active;
	E3dCoordinate       MinDistance;
} E3dSnapToSurface;


extern void  E3d_SnapToSurfaceInit(E3dSnapToSurface* PThis);
extern void  E3d_SnapToSurfaceCopyContents(const E3dSnapToSurface* PSrc, E3dSnapToSurface* PDst);
extern void  E3d_SnapToSurfaceFreeContents(E3dSnapToSurface* PThis);
extern void  E3d_SnapToSurfaceSnapPoints(E3dSnapToSurface* PSnap, const E3dModel* PPathModel, const E3dMatrix PWorldToLocalMatrix, const EIndex PUPoints, const uint8_t PNormalSource, const E3dCoord3* PNormal,  const E3dCoord3* PPositions, const E3dCoord3* PTangents, const EBool PPathClosed,  E3dCoordinate** PMatricesP, EIndex* PMatrices_CountAllocatedP);




//--------------------------------
// GeometryAdv.c
//--------------------------------
extern E3dMaterial*  E3dGeometry_GetMaterial(const E3dGeometry* PGeometry, const EBool PSelectedSubGeo);

//--------------------------------
// Model.c
//--------------------------------
extern E3dModel*  E3d_VectorTest(E3dModel* PModel, E3dGeometry* PGeometry, E3dGeometry** PGeometryRet,  const E3dCoordinate PX1, const E3dCoordinate PY1, const E3dCoordinate PZ1,  const E3dCoordinate PX2, const E3dCoordinate PY2, const E3dCoordinate PZ2,  const EBool PCreateArrow);
extern void       E3d_ModelsShadowMatrix(E3dShadowMap* PShadowMap, E3dLightInstance* PLight, E3dModel** PModels, const EIndex PNumOfModels, E3dMatrix PFrustumMatrix, E3dMatrix PReceiverSTMatrix);


//----------------------------------------
// Adv/SceneAdv.c
//----------------------------------------
extern void E3d_SceneAdvInit(void);



#ifdef USEOpenGL
//----------------------------------------
// GL/Scene.c
//----------------------------------------
extern void    E3dGL_SceneRefreshLights(E3dScene* PScene);
extern void    E3dScene_GLDefineLights(E3dScene* PScene);


//----------------------------------------
// GL/Texture.c
//----------------------------------------
extern int      E3dGL_SceneTextureGetGLImage(E3dScene* PScene, E3d2DTexture* PTexture, const GLenum PGLTarget);
extern EBool    E3dGL_SceneTextureUpdateGLImage(E3dScene* PScene, E3d2DTexture* PTexture, E3dTextureMapper* PMapper);

#endif // USEOpenGL



extern EIndex         E3dScene_GetObjects(E3dScene* PScene, E3dObject*** PObjectsRet, const uint32_t PObjectTypeFlags, const E3dItemFlags PFlags, const EBool PAddReferences);
extern EIndex         E3dScene_GetInstances(E3dScene* PScene, E3dObjectInstance*** PInstances, const uint32_t PObjectTypeFlags, const E3dItemFlags PFlags, const EBool PAddReferences);

extern E3dNode*       E3dScene_NodeFindByName(E3dScene* PScene, const char* PName);
extern E3dNode*       E3dScene_NodeFindByComponentClass(E3dScene* PScene, const E3dNodeComponentClass* PClass);
extern E3dMaterial*   E3dScene_MaterialFindByName(E3dScene* PScene, const char* PName);


extern EIndex         E3dScene_GetSelectedNodes(E3dScene* PScene, E3dNode*** PNodesRet, EIndex* PNumOfNodesRet, const int PSelectionTypes, const EBool PAddReferencesw);
extern EIndex         E3dScene_GetNodes(E3dScene* PScene, E3dNode*** PNodesRet, const EBool PAddReferences);

extern int            E3dScene_GetSingleSelection(E3dScene* PScene, void* PSelectedItem, const int PTarget, const E3dItemFlags PFlags);
extern EChangeFlags   E3dScene_UnselectAll(E3dScene* PScene, const unsigned int PFlags, const EBool PNotify);
extern EBool          E3dScene_UnhighlightAll(E3dScene* PScene, const uint32_t PFlags);

extern void           E3dScene_CallbackAdd(E3dScene* PScene, const EChangeFlags PEventMask, ECallbackFunc PCallbackFunc, EPointer PClientData);
extern EIndex         E3dScene_CallbackRemove(E3dScene* PScene, ECallbackFunc PCallbackFunc, EPointer PClientData);

extern void           E3dScene_WalkNodes(E3dScene* PScene, void (*PFunct)(E3dNode*, EPointer), EPointer PClientData);
extern void           E3dScene_WalkObjects(E3dScene* PScene, void (*PFunct)(E3dModel*, E3dObject*, EPointer), EPointer PClientData);


extern E3dShapeSnapshot**  E3dScene_CollectTransformPoints(E3dScene* PScene, EIndex* PNumOfShapeSnapshotsRet, const EBool PAllPoints, const EBool PVisibleOnly);

extern int            E3dScene_CollectNodeNames(E3dScene* PScene, char*** PNamesRet);

extern void           E3dScene_AnimStackAppend(E3dScene* PScene, E3dAnimStack* PAnimStack);
extern void           E3dScene_AnimStackRemove(E3dScene* PScene, E3dAnimStack* PAnimStack);

extern void           E3dScene_AddDefaultAnimStack(E3dScene* PScene);

extern void           E3dScene_GlobalAnimationAppend(E3dScene* PScene, E3dAnimation* PAnimation);
extern E3dAnimation*  E3dScene_GlobalAnimationAdd(E3dScene* PScene, E3dAnimationClass* PClass);
extern EBool          E3dScene_GlobalAnimationRemove(E3dScene* PScene, E3dAnimation* PAnimation, const EBool PRemoveResidentAnim);

extern void           E3dScene_AnimationAppend(E3dScene* PScene, E3dAnimation* PAnimation);
extern E3dAnimation*  E3dScene_AnimationAdd(E3dScene* PScene, E3dAnimationClass* PClass);
extern E3dKeyFrameAnimation*  E3dScene_KeyFrameAnimationAdd(E3dScene* PScene, ECoreObject* PTargetObject, const char* PAnimName, const EResource* PTargetMember0);
extern EBool          E3dScene_AnimationRemove(E3dScene* PScene, E3dAnimation* PAnimation, const EBool PRemoveResidentAnim);

extern void           E3dScene_ObjectsUpdateModelList(E3dScene* PScene);

extern void           E3dGL_MatrixPush(E3dScene* PScene);
extern void           E3dGL_MatrixPop(E3dScene* PScene);


typedef struct
{
	EBool       Active;
	E3dSpline*  OrigSpline;
	E3dModel*   Model;
} E3dScalingBySpline;



//----------------------------------------
// Visual debugging
//----------------------------------------
//----------------------------------------
// Panel/Panel.c
//----------------------------------------
extern void          E3d_VisualizeBBoxWithMeshEdges(const E3dBBox* PBBox, E3dMesh* PMesh, const char* PModelName);
extern void          E3d_VisualizeBBoxMinSpanWithMeshEdges(E3dScene* PScene, const E3dBBoxMinSpan* PBBox, E3dMesh* PMesh, const char* PModelName);

extern void          E3d_VisualizeVectorsWithMeshEdges(const E3dCoord3* PSamples, const EIndex PNumOfSphereSamples, const char* PModelName);
extern void          E3d_VisualizeRaysWithMeshEdges(const E3dRay* PSamples, const EIndex PCount, const char* PModelName, const E3dCoordinate PLength);
extern E3dPointSet*  E3d_VisualizePointsWithPointSet(const E3dCoord3* PSamples, const EIndex PCount, const char* PModelName);
extern void          E3d_VisualizeColorPointsWithPointSet(E3dPoint* PPoints, const EIndex PCount, const char* PModelName);


//----------------------------------------
// Panel/Undo.c
//----------------------------------------
extern void    E3d_SceneFreeUndoStack(E3dScene* PScene);
extern void    E3d_SceneAddOps(E3dScene* PScene, EOps* POps, EBool PDuplicate);
extern void    E3d_SceneAddOps1Operator(E3dScene* PScene, EOperator* POperator, int LSize, EBool PDuplicate);

extern void    E3d_SceneUndo(E3dScene* PScene);
extern void    E3d_SceneRedo(E3dScene* PScene);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _E3D_Scene_h
