//================================================================
// E3D library
//
// Splines header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2018-Apr-01 09:32:26
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _E3D_Spline_h
#define _E3D_Spline_h

#ifndef _EMalloc_h
#include <EMalloc.h>
#endif

#ifndef _E3D_Geometry_h
#include <E3D/Geometry.h>
#endif


#define	E3dCVALLOC_INCREMENT  4

#define E3dSPLINE_SHOW_TANGENTS_AT_END  1

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#define E3dBezierDEFAULT	0

// Continuity constraint
//
// Geometric continuity: in terms of the directions of tangent vectors only 
//   G0: endpoints meet 
//   G1: direction of tangent vectors agree 
//
// Parametric continuity: defined in terms of direction and magnitude of tangent vectors 
//   C0: endpoints meet 
//   C1: tangent vectors (first derivative: velocity) are equal 
//   C2: derivative of tangents (second derivative: acceleration) are equal 
//
#define E3dBezierC0  1
#define E3dBezierC1  2
#define E3dBezierC2  3
#define E3dBezierG1  4
#define E3dBezierG2  5


// Sub-CV identifiers for TranslatePanel.c
//
enum
{
	E3dCV_NONE = 0,
	E3dCV_IN_TANGENT,
	E3dCV_POSITION,
	E3dCV_OUT_TANGENT
};

//
// Saveables
//

// CV flags
//
#define E3dCV_TAGGED                    0x0001

// Segment flags
//
#define E3dSegmentVISIBLE               0x0001
#define E3dSegmentSELECTED              0x0002
#define E3dSegmentARC_LENGTH_TESSELATE  0x0004  // Uniformly subdivide segment by length (vs. natural parameterization)

#define E3dSegmentDEFAULT  (E3dSegmentVISIBLE|E3dSegmentARC_LENGTH_TESSELATE)



//
// Runtime-only
//

// CV flags
//
#define E3dCV_PREV_ACTIVE  0x0100  // On a Bezier or Hermite curve, the 'Previous' position is active (for moving etc.)
#define E3dCV_ACTIVE       0x0200
#define E3dCV_NEXT_ACTIVE  0x0400  // On a Bezier or Hermite curve, the 'Next' position is active (for moving etc.)

// Temporary flags, used to mark CVs for faster/more memory-efficient algorithms
//
#define E3dCV_MARK0        0x1000
//--------------------------------

typedef struct
{
// User-settable values
//

// CV
//
	E3dCoord3       Position, Previous, Next, Normal;
	uint16_t        CVFlags;
	uint16_t				Constraint;            // For tangents

// Segment
//
	uint16_t        SegmentShape;          // Segment shape E3dSEGMENT_BEZIER etc.
	uint16_t        SegmentFlags;
	int32_t         SegmentNumOfLinSteps;  // User-settable number of linear segment approximations for this segment.

	EcRGBA8Color    Color;
	float           Thickness;

// Internal values
//
	E3dCoordinate*  LinearSegments;       // Piece-wise linear segment points, including the last one (the beginning point, if the Spline is closed and this CV is for the last segment), so the number of valid entries here always (NumOfLinearSegments + 1) * Dimensions
	EIndex          NumOfLinearSegments;  // Number of linear segments for this segment.

	E3dCoordinate   SegmentLength;
	E3dCoord4*      StepsForLength;int  NumOfStepsForLength;  // Linear segments used for (approximating) the length of this segment.
                                                            // The W component is the cumulative length at the end of a linear segment
} E3dSplineCV;



// This structure is used for taking "snapshots" of SplineCVs for interactive
// transforming and for Undo/Redo of transforming and deleting SplineCVs
//
typedef struct
{
	E3dCoord3  Position, Previous, Next;
	int        Index;    // Index in the original array
	uint16_t   CVFlags;
	uint16_t   SegmentFlags;
} E3dSnapshotSplineCV;



enum
{
	E3dSplineNONE = 0,
	E3dSplineLINEAR,
	E3dSplineBEZIER,
	E3dSplineHERMITE,
	E3dSplineBSPLINE,
	E3dSplineNURBS,
	E3dSplineCARDINAL,
	E3dSplineSTEP,
	E3dSplineARC    // Circular arc
};


#define caseE3dSplineHAS_TANGENTS\
	case E3dSplineBEZIER:\
	case E3dSplineHERMITE

#define caseE3dSplineGOES_THROUGH_CVS\
	case E3dSplineBEZIER:\
	case E3dSplineCARDINAL:\
	case E3dSplineLINEAR:\
	case E3dSplineHERMITE

#define caseE3dSplineGOES_THROUGH_CVS\
	case E3dSplineBEZIER:\
	case E3dSplineCARDINAL:\
	case E3dSplineLINEAR:\
	case E3dSplineHERMITE

#define caseE3dSplineINDIRECT_CVS\
	case E3dSplineCARDINAL:\
	case E3dSplineBSPLINE

#define caseE3dSplineNO_INDIRECT_CVS\
	case E3dSplineBEZIER:\
	case E3dSplineLINEAR:\
	case E3dSplineHERMITE

// Spline Flags
#define E3dSplfSHOW_CVS           0x0001  // Show CVs
#define E3dSplfSHOW_LINES         0x0002  // Show tangents (Bezier) and polyline (B-spline and Cardinal)
#define E3dSplfSEGMENT_COLORS     0x0004  // Segments have their own colors
#define E3dSplfSEGMENT_THICKNESS  0x0008  // Segments have their own relative thickness
#define E3dSplfSEGMENT_VISIBIITY  0x0010  // Some segments are visible, others aren't

#define E3dSplfDEFAULT  (E3dSplfSHOW_CVS|E3dSplfSHOW_LINES)

typedef struct
{
	E3dGeometryCore;

	int             Dimensions;

	EcRGBAf32Color  RGBAfColor;EBool  UseCustomColor;
	float           Thickness;

	EList(, E3dSplineCV, CVs);             // Control Vertices


	EList(, E3dObject*, Faces);            // Faces that use this Spline as a contour, if any

	EObject*        AnimationChannel;      // If this Spline defines an animation curve, this points to the Channel

	E3dCoordinate   Length;                // Total length of the Spline
	int             NumOfLinSteps;         // Number of linear segments between 2 CVs, for drawing the Spline.
	int             NumOfStepsForLength;   // # of steps (linear segments) used for determining (approximating) the length of segments

	E3dCoordinate*  SLinearSegments;       // Piece-wise linear segment points, including the last one (the beginning point, if the Spline is closed), so the number of valid entries here is always (SNumOfLinearSegments + 1) * Dimensions
	EIndex          SNumOfLinearSegments;  // Total number of piece-wise linear segments.

	int             Phantom;               // Automatic phantom key (extremity) (CARDINAL, B-SPLINE)
	float           Tension;               // Tension of the curve (CARDINAL)
	int             Constraint;            // Default key constraint (BEZIER)
	int             SplineType;


	EBool           UniformLinSteps;       // Use Steps field of Spline for all segments
	EBool           Closed;                // Curve is closed (wrap-around)

	uint16_t        Flags;                 // Or-ed flags of E3dSplfSHOW_CVS etc.
	float           VectorDrawScale;       // For drawing Spline CV normals and tangents
} E3dSpline;





// Coordinate modes for E3dVectorInfo
//
enum
{
	E3dVTM_P1_P2 = 0,
	E3dVTM_P1_DIR,
	E3dVTM_P1_DIR_LENGTH
};


typedef struct
{
	E3dObjectComponentCore;

// E3dVectorInfo part
//
	E3dCoord3  P0, P1;
	char       CoordinateMode;
	EBool      ArrowOn;
} E3dVectorInfo;



extern E3dObjectClass   E3dSpline_class[];


//--------------------------------------------------------
// Inline functions
//--------------------------------------------------------

static inline EIndex	E3d_SplineNumOfLinearSegmentPoints(const E3dSpline* PSpline)
{
	return(PSpline->Closed ? PSpline->SNumOfLinearSegments : PSpline->SNumOfLinearSegments + 1);
}


static inline void E3d_SplineCVFreeContents(E3dSplineCV* PCV)
{
//printf("%s() %s:%d  %p  %p\n", __FUNCTION__, __FILE__, __LINE__, PCV, PCV->LinearSegments);fflush(stdout);
	if(PCV->LinearSegments) { EFree(PCV->LinearSegments);PCV->LinearSegments = NULL;PCV->NumOfLinearSegments = 0; }
	if(PCV->StepsForLength) { EFree(PCV->StepsForLength);PCV->StepsForLength = NULL;PCV->NumOfStepsForLength = 0; }
}


static inline void E3d_SplineCVCopy(const E3dSplineCV* PSrcCV, E3dSplineCV* PDstCV)
{
#define _Copy(m)	PDstCV->m = PSrcCV->m;
	_Copy(Position);
	_Copy(Previous);
	_Copy(Next);
	_Copy(Normal);
	_Copy(CVFlags);
	_Copy(Constraint);

	_Copy(SegmentShape);
	_Copy(SegmentFlags);
	_Copy(SegmentNumOfLinSteps);

	_Copy(NumOfStepsForLength);
#undef _Copy
}


static inline void E3dSplineCV_SegmentInfoCopy(const E3dSplineCV* PSrcCV, E3dSplineCV* PDstCV)
{
#define _Copy(m)  PDstCV->m = PSrcCV->m;
	_Copy(Constraint);

	_Copy(SegmentShape);
	_Copy(SegmentFlags);
	_Copy(SegmentNumOfLinSteps);

	_Copy(NumOfStepsForLength);
#undef _Copy
}

enum
{
	E3dSplineNormalAUTO = 0,
	E3dSplineNormalCV,
	E3dSplineNormalFIXED
};

#define E3dSplineAT_CVs  -1

#ifndef _E3D_Math_h
#include <E3D/Math.h>
#endif

//================================================================================
// Compute the tangent at, and the angle between two consecutive line segments
//================================================================================
static inline int E3d_ComputeTangent(const EIndex PUC, const EIndex PUN, const E3dCoordinate* PPathPoints, const int PDimensions, const EPointerInt PStride, const EBool PClosed,  E3dCoord3* PTangentP, E3dCoordinate* PAngleCosineP)
{
	if(PUN > 1)
	{
		const E3dCoordinate*  LPathPoint = PPathPoints + PUC * PDimensions;
		const E3dCoordinate*  LPathPrevPoint = (const E3dCoordinate*)((const uint8_t*)LPathPoint - PStride);
		const E3dCoordinate*  LPathNextPoint = (const E3dCoordinate*)((const uint8_t*)LPathPoint + PStride);
		EBool                 LCheckLengths = TRUE;

		if(PUC == 0)
		{
			if(PClosed) LPathPrevPoint = PPathPoints + (PUN - 1) * PDimensions;
			else { LPathPrevPoint = LPathPoint;LCheckLengths = FALSE; }
		}
		else
		{
			if(PClosed) { if(PUC == PUN - 1) LPathNextPoint = PPathPoints; }
			else if(PUC == PUN) { LPathNextPoint = LPathPoint;LCheckLengths = FALSE; }
		}

		if(LCheckLengths)
		{
			const E3d_Sub3i_aa(LPosToPrev,  LPathPrevPoint, LPathPoint);const E3dCoordinate  LPosToPrevLenSq = E3d_Dot3_vv(LPosToPrev, LPosToPrev);
			const E3d_Sub3i_aa(LPosToNext,  LPathNextPoint, LPathPoint);const E3dCoordinate  LPosToNextLenSq = E3d_Dot3_vv(LPosToNext, LPosToNext);
			if(LPosToPrevLenSq != LPosToNextLenSq && LPosToPrevLenSq > 0.0 && LPosToNextLenSq > 0.0)
			{
				const E3dCoordinate  LPosToPrevLen = E3d_Sqrt(LPosToPrevLenSq);const  E3d_Div3i_vs(LNormalizedPosToPrev,  LPosToPrev, LPosToPrevLen);
				const E3dCoordinate  LPosToNextLen = E3d_Sqrt(LPosToNextLenSq);const  E3d_Div3i_vs(LNormalizedPosToNext,  LPosToNext, LPosToNextLen);

				E3d_Sub3_vvp(LNormalizedPosToNext, LNormalizedPosToPrev,  PTangentP);

				goto TangentDone;
			}
		}

// Compute tangent of the curve at the current point
//
		PTangentP->X = LPathNextPoint[E3dX] - LPathPrevPoint[E3dX];
		PTangentP->Y = LPathNextPoint[E3dY] - LPathPrevPoint[E3dY];
		PTangentP->Z = LPathNextPoint[E3dZ] - LPathPrevPoint[E3dZ];

TangentDone:
		E3d_Normalize3_p(PTangentP);  // Normalize tangent vector
		return(E_SUCCESS);
	}
	else return(E_NO_DATA);
}






//--------------------------------------------------------
// Function prototypes
//--------------------------------------------------------
extern void           E3dSpline_Print(E3dSpline* PSpline);

extern E3dSpline*     E3dSpline_FromPositions(const E3dCoord3* PCVPositions, const EIndex PNumOfCVs, const EBool PClosed);

extern E3dCoordinate*  E3dSpline_GetUsAtCVs(E3dSpline* PSpline);

extern int            E3dSpline_GetTransforms(E3dSpline* PSpline, const E3dCoordinate* PMatrix,  const E3dCoordinate* PUs, const EIndex PUCount,  const uint8_t PNormalSource, const E3dCoord3* PNormal,  E3dCoordinate** PMatrices, EIndex* PMatrices_CountAllocated,  E3dCoord3** PPositions, E3dCoord3** PTangents, E3dCoord3** PNormalsP,  E3dCoordinate** PAngleCosinesP);

extern E3dSpline*     E3dSpline_ConvertToLinear(E3dSpline* PSpline, E3dSpline* PDSpline);

extern int            E3dSpline_Invert(E3dSpline* PSpline);
extern EIndex         E3dSpline_NumOfSelectedCVs(const E3dSpline* PSpline);

extern int            E3dSpline_CopyContents(const E3dSpline* PSrc, E3dSpline* PDst);

extern void	          E3dSplineCV_Init(E3dSplineCV* PCV, const int PSegmentType);
extern void           E3dSplineCVs_Init(E3dSplineCV* PCVs, const int PSegmentType, const EIndex PN);

extern E3dSplineCV*   E3dSplineCVs_New(const int PType, const EIndex PNumOfCVs);

extern void           E3dSpline_SetCVs(E3dSpline* PSpline, E3dSplineCV* PCVs, const EIndex PCount);
extern int            E3dSpline_CollectCVs(E3dSpline* PSpline, void** PTransformedPointsRet, unsigned int* PPointSizeRet, const EBool PAllCVs);

extern void*          E3dSpline_AddCV(E3dSpline* PSpline);

extern void           E3dSpline_FreeCVs(E3dSpline* PSpline);

extern E3dCoordinate  E3dSpline_GetDrawingArcLengthParameters(const E3dSpline* PSpline, const E3dCoordinate PGapSize, const EBool PInvert,  E3dCoordinate* PTable, EIndex* PNumOfGapsP);

extern EBool          E3dSpline_GetBoundingBox(E3dSpline* PSpline, const unsigned int PWhatToInclude,  E3dCoord3* PBBMin, E3dCoord3* PBBMax);
extern EBool          E3dSpline_GetTransformedBoundingBox(E3dSpline* PSpline, const E3dMatrix PMatrix,  E3dCoord3* PBBMin, E3dCoord3* PBBMax);
extern EBool          E3dSpline_GetPerspectiveProjectedBoundingBox(E3dSpline* PSpline, const E3dMatrix PMatrix, const unsigned int PWhatToInclude,  E3dCoord2* PBBMin, E3dCoord2* PBBMax,  E3dCoord4* PMaxProjectedXYAbsValPoint);

extern EBool          E3dSpline_CreateLinearSegments(E3dSpline* PSpline);
extern void           E3dSpline_FreeRenderData(E3dSpline* PSpline);

extern E3dCoordinate  E3dSpline_UpdateSegmentLengths(E3dSpline* PSpline, EIndex PSteps);

extern int            E3dSpline_GetXYRange(const E3dSpline* PSpline,  E3dCoord2* PXY0, E3dCoord2* PXY1);

extern int            E3dSpline_GetYValueFromX(E3dSpline* PSpline, const E3dCoordinate PX, E3dCoordinate* PValue);

extern int            E3dSpline_Get3DPosition(E3dSpline* PSpline, E3dCoordinate PTime,  E3dCoord3* PPositionP, E3dCoord3* PTangentP, E3dCoord3* PNormalP);

extern int            E3dSpline_GetClosestPointToPoint(const E3dSpline* PSpline,  const E3dCoord3* PPoint,  E3dCoord3* PPositionR, E3dCoord3* PTangentR, E3dCoord3* PNormalR);
extern EChangeFlags   E3dSpline_Unselect(E3dSpline* PSpline, const unsigned int PFlags);

extern void           E3dSpline_Transform(E3dSpline* PSpline, const E3dMatrix PMatrix);

extern void           E3d_CreateVectorArrow(E3dSplineCV* PCV,  const E3dCoordinate PX2, const E3dCoordinate PY2, const E3dCoordinate PZ2);

extern E3dCoord2*     E3d_CreateUnitArc(const EIndex PSides, const E3dCoordinate PStartAngle, const E3dCoordinate PEndAngle,  E3dCoord2* PPoints);

extern int            E3dSplines_ReadFromCSV(const char* PBuffer, const EIndex PBufferSize,  E3dSpline*** PSplinesP, EIndex* PSplines_CountP);

//--------------------------------------------------------
// Face.c
//--------------------------------------------------------



extern EChangeFlags   E3dSpline_CallOutputs(E3dSpline* PSpline, const EChangeFlags PChangeFlags, const EBool PClearFaceCaches);
extern EChangeFlags   E3dSpline_FullUpdate(E3dSpline* PSpline);


//--------------------------------------------------------
// SplineOps.c
//--------------------------------------------------------
extern void           E3d_SplineOpsInit(void);
extern void           E3d_SplineOpsFree(void);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif  // _E3D_Spline_h
