//========================================================================
// E library
//
// - Memory management and memory-tracker
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 22:07:51
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//========================================================================
#ifndef _EMalloc_h
#define _EMalloc_h


#include <sys/types.h>
#include <string.h>

#ifdef __APPLE__
	#include <stdlib.h>
#else
	#include <malloc.h>
#endif

#include <ETypes.h>

#include <stdint.h>  // For uint32_t etc. for E_Random


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif


#define EMemDEBUG  0  // Do memory profiling

#if EMemDEBUG
extern char*  EMem_Name;
extern EBool  EMem_Verbose;
extern char*  EMem_File;
extern int    EMem_Line;
extern char*  EMem_Function;

extern EIndex	EMem_BlocksPrintStart;

#define E_MemDebug\
	EMem_Function = (char*)__FUNCTION__;\
	EMem_File = (char*)__FILE__;\
	EMem_Line = __LINE__


#define	EMemALLOC_INCREMENT	4096	// For EMem_Blocks


typedef struct
{
	char           Name[256];
	unsigned long  Size;
	void*          Address;
	EBool          Active;

	ECallerInfo    Caller;
} EMemEntry;


extern EMemEntry*    EMem_Blocks;extern unsigned int  EMemBlocks_Count, EMemBlocks_CountAllocated;
extern void*         EMem_Block;

extern void   EPrintUnfreedMemory(void);

extern void*  EMalloc(const size_t PBSize);
extern void*  ECalloc(const size_t PNMemb, const size_t PBSize);
extern void*  ERealloc(void* PBlock, const size_t PBSize);
extern void   EFree(void* PBlock);

#else // EMemDEBUG

#define EMalloc(size)          malloc(size)
#define ECalloc(nmemb, size)   calloc(nmemb, size)
#define ERealloc(block, size)  realloc(block, size)
#define EFree(block)           free(block)

#define E_MemDebug

#endif // EMemDEBUG


#define EClr_p(pDst)             do { memset(pDst, 0, sizeof(*pDst)); } while(0)
#define EClr_s(pDst)             do { memset(&pDst, 0, sizeof(pDst)); } while(0)
#define EClr_a(pDst)             do { memset(pDst, 0, sizeof(pDst)); } while(0)
#define EClrN(pDst, pCount)      do { memset(pDst, 0, sizeof((pDst)[0]) * pCount); } while(0)

#define EMallocClr(pSize, pDst)          do { E_MemDebug;pDst = EMalloc(pSize);memset(pDst, 0, pSize); } while(0)
#define EMallocT(pType, pSize, pDst)     do { E_MemDebug;pDst = (pType)EMalloc(pSize); } while(0)
#define EMallocTClr(pType, pSize, pDst)  do { const size_t  lSize = pSize;E_MemDebug;pDst = (pType)EMalloc(lSize);memset(pDst, 0, lSize); } while(0)
#define ENew(pDst)                       do { E_MemDebug;const EIndex  lSize = sizeof(*(pDst));pDst = (typeof(pDst))EMalloc(lSize); } while(0)
#define ENewT(pType, pDst)               do { E_MemDebug;const EIndex  lSize = sizeof(pType);pDst = (pType*)EMalloc(lSize); } while(0)
#define ENewNT(pType, pCount, pDst)      do { E_MemDebug;pDst = (pType*)EMalloc(sizeof(pType) * (pCount)); } while(0)  // The while is just a trick to "swallow" the semicolon at the instantiation of the macro

#define ENewClr(pDst)                    do { E_MemDebug;const EIndex  lSize = sizeof(*(pDst));pDst = (typeof(pDst))EMalloc(lSize);memset(pDst, 0, lSize); } while(0)

#define ENewN(pCount, pDst)              do { E_MemDebug;pDst = (typeof(pDst))EMalloc(sizeof(pDst[0]) * (pCount)); } while(0)	// The while is just a trick to "swallow" the semicolon at the instantiation of the macro
#define ENewNClr(pCount, pDst)           do { E_MemDebug;const EIndex  lSize = sizeof(pDst[0]) * (pCount);pDst = (typeof(pDst[0])*)EMalloc(lSize);memset(pDst, 0, lSize); } while(0)

#define EBufferAI(pQualifier, pType, pList)   pQualifier pType*  pList = NULL;pQualifier EIndex  pList##_Size = 0, pList##_SizeAllocated = 0
#define EBufferAIS(pQualifier, pType, pInitialSize, pList)  pQualifier pType*  pList = (pQualifier pType*)EMalloc(sizeof(pType*) * pInitialSize);pQualifier EIndex  pList##_Size = 0, pList##_SizeAllocated = pInitialSize


//================================================
// Resize memory buffer
//================================================
#define EResizeNT(pType, pCount,  pBuffer) do\
{\
	const EIndex  llCount = pCount;if(llCount)\
	{\
		E_MemDebug;\
		pType*  lTBuffer = (pType*)ERealloc(pBuffer, sizeof(pType) * llCount);\
		if(lTBuffer) pBuffer = lTBuffer;\
	}\
	else\
	{\
		EFreeClr(pBuffer);\
	}\
} while(0)


#define EResizeN(pCount, pBuffer) do\
{\
	const EIndex  llCount = pCount;if(llCount)\
	{\
		E_MemDebug;\
		typeof(pBuffer[0])*  lTBuffer = (typeof(pBuffer[0])*)ERealloc(pBuffer, sizeof(typeof(pBuffer[0])) * llCount);\
		if(lTBuffer) pBuffer = lTBuffer;\
	}\
	else\
	{\
		EFreeClr(pBuffer);\
	}\
} while(0)


#define EList_Resize(pBuffer, pCount) do\
{\
	EResizeNT(typeof(pBuffer[0]), pCount, pBuffer);\
	pBuffer##_Count = pCount;\
} while(0)


#define EList_ResizeN(pBuffer, pCount) do\
{\
	EResizeNT(typeof(pBuffer[0]), pCount, pBuffer);\
} while(0)


#define EList_SizeCheckA(pBuffer, pNewNElements) do\
{\
	const EIndex  lN = pNewNElements;if(pBuffer##_CountAllocated < lN)\
	{\
		EResizeNT(typeof(pBuffer[0]), lN, pBuffer);pBuffer##_CountAllocated = lN;\
	}\
} while(0)


#define EList_Copy(pSrc, pDst) do\
{\
	const EIndex  lCount = pSrc##_Count;\
	EResizeNT(typeof(pDst[0]), lCount, pDst);\
	memcpy(pDst, pSrc, sizeof(pDst[0]) * lCount);\
	pDst##_Count = lCount;\
} while(0)


//================================================================
// Resize memory buffer and clear the newly added segment
//================================================================
#define EResizeNClr(pType, pOldCount, pNewCount,  pBuffer) do\
{\
	const EIndex  lOldCount = pOldCount;\
	const EIndex  lNewCount = pNewCount;if(lNewCount)\
	{\
		E_MemDebug;\
		pType*  lBuffer = (pType*)ERealloc(pBuffer, sizeof(pType) * lNewCount);if(lBuffer)\
		{\
			pBuffer = lBuffer;\
			if(lNewCount > lOldCount) memset(lBuffer + lOldCount, 0, sizeof(pType) * (lNewCount - lOldCount));\
		}\
	}\
	else EFreeClr(pBuffer);\
} while(0)


//================================================================
// Resize memory buffer
// - Expose temporary variable, so the outcome can be queried
//================================================================
#define EResizeNN(pType, pCount,  pBuffer, pTBuffer) do\
{\
	const EIndex  lCount = pCount;if(lCount)\
	{\
		E_MemDebug;\
		pTBuffer = (pType*)ERealloc(pBuffer, sizeof(pType) * lCount);if(pTBuffer) pBuffer = pTBuffer;\
	}\
	else\
	{\
		EFreeClr(pBuffer);pTBuffer = NULL;\
	}\
} while(0)


//================================================
// Resize memory buffer and return new count
//================================================
#define EResizeNC(pType, pCount,  pBuffer, pCountOut) do\
{\
	const EIndex  llCount = pCount;if(llCount)\
	{\
		E_MemDebug;\
		pType*  pTBuffer = (pType*)ERealloc(pBuffer, sizeof(pType) * llCount);\
		if(pTBuffer) { pBuffer = pTBuffer;pCountOut = llCount; }\
	}\
	else\
	{\
		EFreeClr(pBuffer);\
		pCountOut = llCount;\
	}\
} while(0)


#define EFreeClr(p)     do { EFree(p);p = NULL; } while(0)
#define EFreeChk(p)     do { if(p) EFree(p); } while(0)
#define EFreeChkClr(p)  do { if(p) { EFree(p);p = NULL; } } while(0)


#define ECopyN(pSrc, pCount,  pDst) do\
{\
	const EIndex  lCount = pCount;\
	E_MemDebug;\
	if(lCount)\
	{\
		const size_t  lSize = sizeof(pSrc[0]) * lCount;\
		pDst = (typeof(&pSrc[0]))EMalloc(lSize);if(pDst)\
		{\
			memcpy(pDst, pSrc, lSize);\
		}\
	}\
	else pDst = NULL;\
} while(0)

#define ECopyNChk(pType, pCount, pSrc, pDst) do\
{\
	E_MemDebug;\
	EFreeChk(pDst);\
	if(pCount)\
	{\
		pDst = (pType*)EMalloc(sizeof(pType) * (pCount));if(pDst)\
		{\
			memcpy(pDst, pSrc, sizeof(pType) * pCount);\
		}\
	}\
	else pDst = NULL;\
} while(0)

#define ECopyN_p(pType, pName, pSrc, pDst) do\
{\
	if(pDst->pName) { EFree(pDst->pName);pDst->pName = NULL;pDst->pName##_Count = 0; }\
	const EIndex  lCount = pSrc->pName##_Count;if(lCount)\
	{\
		pDst->pName = (pType*)EMalloc(sizeof(pType) * lCount);if(pDst->pName)\
		{\
			memcpy(pDst->pName, pSrc->pName, sizeof(pType) * lCount);\
			pDst->pName##_Count = lCount;\
		}\
	}\
} while(0)


#define EList_Copy_p(pType, pName,  pSrc, pDst) do\
{\
	if(pDst->pName.Items) { EFree(pDst->pName.Items);pDst->pName.Items = NULL;pDst->pName.Count = 0; }\
	const EIndex  lCount = pSrc->pName.Count;if(lCount)\
	{\
		pDst->pName.Items = (pType*)EMalloc(sizeof(pType) * lCount);if(pDst->pName.Items)\
		{\
			memcpy(pDst->pName.Items, pSrc->pName.Items, sizeof(pType) * lCount);\
			pDst->pName.Count = lCount;\
		}\
	}\
} while(0)


extern char*      EStrDup(const char* PString);
extern uint8_t*   EStrDupU(const uint8_t* PString);
extern char*      EStrNDup(const char* PString, const EIndex PLen);
extern wchar_t*   EStrDup8toW(const char* PString);
extern char*      EStrDupWto8(const wchar_t* PString);
extern wchar_t*   EStrDupW(const wchar_t* PString);
extern uint16_t*  EStrDupW16(const uint16_t* PString);
extern char*      EStrDupWto8(const wchar_t* PString);
extern char*      EStrDupWtoUTF8(const wchar_t* PString);

extern size_t     EStrLenW(const wchar_t* PString);
extern size_t     EStrLenW16(const uint16_t* PString);
extern size_t     EStrLenW16Unaligned(const uint8_t* PString);
extern size_t     EStrLenW32Unaligned(const uint8_t* PString);

extern char*      EStrDupSuffix(const char* PString, const char* PSuffix);



//================================================
// Generate random bits
//================================================
static inline uint32_t _RandomBits(ERandomState* PState)
{
	const uint64_t  LSum = (uint64_t)2111111111ul * (uint64_t)PState[3] +
	                       (uint64_t)1492 * (uint64_t)(PState[2]) +
	                       (uint64_t)1776 * (uint64_t)(PState[1]) +
	                       (uint64_t)5115 * (uint64_t)(PState[0]) +
	                       (uint64_t)PState[4];
	PState[3] = PState[2];PState[2] = PState[1];PState[1] = PState[0];
	PState[4] = (uint32_t)(LSum >> 32);            // Carry
	PState[0] = (uint32_t)(LSum);                  // Low 32 bits of sum
	return(PState[0]);
} 


//================================================
// Random number between 0.0 and 1.0
//================================================
static inline double E_RandomD(ERandomState* PState)
{
	return((double)_RandomBits(PState) * (1.0 / (65536.0 * 65536.0)));
}


//================================================
// Random number between 0.0 and 0.9999999...
//================================================
static inline double E_RandomDN(ERandomState* PState)
{
	return((double)_RandomBits(PState) * (1.0 / (65536.0 * 65536.0 + 1.0)));
}


//========================================================
// Returns integer random number in desired interval
//========================================================
static inline int E_IRandom(const int min, const int max, ERandomState* PState)
{
// Output random integer in the interval min <= x <= max
// Relative error on frequencies < 2^-32
//
	if(max <= min)
	{
		if(max == min) return min; else return 0x80000000;
	}

// Multiply interval with random and truncate
//
	const int  r = (int)((max - min + 1) * E_RandomD(PState)) + min; 
	if(r > max) return(max);
	else return(r);
}



//extern double  E_RandomD(ERandomState* PState);
//extern double  E_RandomDN(ERandomState* PState);


#define EStackAlloc(pSize, pBuffer)    ((pSize) <= sizeof(pBuffer) ? (EPointer)(pBuffer) : EMalloc(pSize))
#define EStackFree(pPointer, pBuffer)  { if((pPointer) != ((EPointer)(pBuffer))) EFree(pPointer); }

extern ECodePoint  E_CodePoint;

extern void        E_MallocInit(void);
extern void        E_MallocFree(void);

extern void        E_RandomInit(const uint32_t seed, ERandomState* PState);

extern int         Easprintf(char** PResult, const char* PFormatStr, ...);

extern long long   E_GetMemoryUse(long long* PTotalRAMRet);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif	// _EMalloc_h
