//================================================================================================
// Profiling
//
// AUTHOR:  Gabor Nagy
// DATE:    2014-Jul-16 22:05:07
//
// Copyright 2014-2018 Gabor Nagy. All rights reserved.
//================================================================================================

#ifndef _ETiming_h
#define _ETiming_h


// We mostly use macros here, so they can be "compiled away" when profiling is disabled
//

#ifndef TIMER_ENABLE
	#define TIMER_ENABLE  0
#endif

#ifndef TIMER_USE_AUTO_SCOPE
	#define TIMER_USE_AUTO_SCOPE  1
#endif

#ifndef TIMER_PRINTOUT_ENABLE
	#define TIMER_PRINTOUT_ENABLE  1
#endif

#if TIMER_ENABLE

#ifndef ETm_Malloc

#define ETm_Malloc(m)                     malloc(m)
#define ETm_MallocT(mType, mSize, mDst)   do { mDst = (mType)ETm_Malloc(mSize); } while(0)

#define ETm_Realloc(m, mSize)             realloc(m, mSize)
#define ETm_Resize(mType, mCount, mBuffer) mBuffer = (mType*)realloc(mBuffer, sizeof(mType) * (mCount))

#define ETm_New(mType, mDst)              do { mDst = (mType*)ETm_Malloc(sizeof(mType)); } while(0)
#define ETm_NewClr(mType, mDst)           do { const int  lByteSize = sizeof(mType);mDst = (mType*)ETm_Malloc(lByteSize);memset(mDst, 0, lByteSize); } while(0)

#define ETm_NewN(mType, mCount, mDst)     do { mDst = (mType*)ETm_Malloc(sizeof(mType) * (mCount)); } while(0)
#define ETm_NewNClr(mType, mCount, mDst)  do { const int  lByteSize = sizeof(mType) * (mCount);mDst = (mType*)ETm_Malloc(lByteSize);memset(mDst, 0, lByteSize); } while(0)

#define ETm_Free(m)                       free(m)
#define ETm_FreeClr(m)                    do { free(m);m = NULL; } while(0)
#define ETm_FreeChk(m)                    if(m) ETm_Free(m);
#define ETm_FreeChkClr(m)                 if(m) { ETm_Free(m);m = NULL; }
#define ETm_DeleteChkClr(m)               if(m) { delete m;m = NULL; }
#define ETm_DeleteChk(m)                  if(m) delete m;


#define ETm_PointerAdd(mList, mPointer)\
do {\
	ETm_Resize(typeof((mList)[0]), mList##_Count + 1, mList);\
	mList[mList##_Count] = mPointer;\
	mList##_Count += 1;\
} while(0)


#define ETmList_IntAdd(mList, mValue)\
do {\
	ETm_Resize(int, mList##_Count + 1, mList);\
	mList[mList##_Count] = mValue;\
	mList##_Count += 1;\
} while(0)


#define ETmList_StructAdd(mType, mList, mAllocIncrement,  mDst)\
{\
	if(mList##_Count >= mList##_CountAllocated)\
	{\
		const int  lNewAllocCount = mList##_Count + mAllocIncrement;\
		ETm_Resize(mType, lNewAllocCount, mList);\
		mList##_CountAllocated = lNewAllocCount;\
	}\
	mDst = (mType*)mList + mList##_Count;\
}

#define ETmList_StructAddAClr(mType, mList, mAllocIncrement,  mDst)\
{\
	if(mList##_Count >= mList##_CountAllocated)\
	{\
		const int  lNewAllocCount = mList##_Count + mAllocIncrement;\
		ETm_Resize(mType, lNewAllocCount, mList);\
		mList##_CountAllocated = lNewAllocCount;\
	}\
	memset(mList + mList##_Count, 0, sizeof(mType));\
	mDst = mList + mList##_Count;\
	mList##_Count += 1;\
}



#ifndef TIMER_DRAW_HIERARCHY_LINKS
	#define TIMER_DRAW_HIERARCHY_LINKS  0
#endif

#else

#define TIMER_DRAW_HIERARCHY_LINKS  0

#endif


#define TIMER_MAX_NAMELEN        64
#define TIMER_INITIAL_ALLOC      64
#define TIMER_ALLOC_INCREMENT    64
#ifndef TIMER_NAME_COLUMN_WIDTH
#define TIMER_NAME_COLUMN_WIDTH  24
#endif

typedef struct ETmTimerEntry_S  ETmTimerEntry;

typedef uint64_t  TIMERTime;

struct ETmTimerEntry_S
{
	char            Name[TIMER_MAX_NAMELEN];  // We use static buffers here to avoid the CPU overhead of malloc/free (we're trying to measure performance after all)
  const char*     FunctionName;const char*  FileName;int  LineNumber;  // Location of invocation
	char            Bracket;                  // Opening character for bracketing when printing the name (e.g. '[' means  [name])

	const char*     FWName;                   // Fly-weight name : pointer to the Name member of the first ETmTimerEntry that has this name. After a capture is done, we set and use these for quick (pointer) comparisons.

	TIMERTime       Start, End;

	int             ParentIndex;              // The parent is set as we build the hierarchy. We use and index because Entries may be reallocated during the build.
	int             ChildIndex;               // Child and RightSibling are computed after the profile stack is built
	int             RightSiblingIndex;

	int             Level;
};


typedef struct
{
	const char*     Name;
	int*            EntryIndices;int  EntryIndices_Count;
	TIMERTime       TotalTime, SelfTime;  // Total time spent in unit. Time spend in unit, excluding time in children
} ETmProfilingUnit;


typedef struct
{
	int             Verbosity;

	ETmTimerEntry*  Entries;
	int             Entries_Count, Entries_CountAllocated;

	int             CurrentParentIndex;
	int             CurrentLevel;

	TIMERTime       PrevTime;


// Profiling info
//
	ETmProfilingUnit*  ProfilingUnits;int  ProfilingUnits_Count, ProfilingUnits_CountAllocated;
} ETmTimerStack;



	#ifdef _WIN32
		#define TimingWINDOWS  1
	#else
		#ifdef _WIN64
			#define TimingWINDOWS  1
		#endif
	#endif

	#ifndef TextColorNormal
		#ifdef TimingWINDOWS
			#define  TextColorNormal   do { HANDLE  hOut = GetStdHandle(STD_OUTPUT_HANDLE);SetConsoleTextAttribute(hOut, FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_BLUE); } while(0)
			#define  TextColorRed      do { HANDLE  hOut = GetStdHandle(STD_OUTPUT_HANDLE);SetConsoleTextAttribute(hOut, FOREGROUND_RED|FOREGROUND_INTENSITY); } while(0)
			#define  TextColorGreen    do { HANDLE  hOut = GetStdHandle(STD_OUTPUT_HANDLE);SetConsoleTextAttribute(hOut, FOREGROUND_GREEN|FOREGROUND_INTENSITY); } while(0)
			#define  TextColorBlue     do { HANDLE  hOut = GetStdHandle(STD_OUTPUT_HANDLE);SetConsoleTextAttribute(hOut, FOREGROUND_BLUE|FOREGROUND_INTENSITY); } while(0)
			#define  TextColorCyan     do { HANDLE  hOut = GetStdHandle(STD_OUTPUT_HANDLE);SetConsoleTextAttribute(hOut, FOREGROUND_GREEN|FOREGROUND_BLUE|FOREGROUND_INTENSITY); } while(0)
			#define  TextColorMagenta  do { HANDLE  hOut = GetStdHandle(STD_OUTPUT_HANDLE);SetConsoleTextAttribute(hOut, FOREGROUND_RED|FOREGROUND_BLUE|FOREGROUND_INTENSITY); } while(0)
			#define  TextColorYellow   do { HANDLE  hOut = GetStdHandle(STD_OUTPUT_HANDLE);SetConsoleTextAttribute(hOut, FOREGROUND_RED|FOREGROUND_GREEN|FOREGROUND_INTENSITY); } while(0)
		#else
			static const char  NormalCode[]  = { 0x1B, '[' ,'m', 0 };
			static const char  RedCode[]     = { 0x1B, '[', '3', '1', 'm', 0 };
			static const char  GreenCode[]   = { 0x1B, '[', '3', '2', 'm', 0 };
			static const char  YellowCode[]  = { 0x1B, '[', '3', '3', 'm', 0 };
			static const char  BlueCode[]    = { 0x1B, '[', '1', 'm', 0 };
			static const char  MagentaCode[] = { 0x1B, '[', '3', '5', 'm', 0 };
			static const char  CyanCode[]    = { 0x1B, '[', '3', '6', 'm', 0 };

			#define  TextColorNormal   printf("%s", NormalCode)
			#define  TextColorRed      printf("%s", RedCode)
			#define  TextColorGreen    printf("%s", GreenCode)
			#define  TextColorBlue     printf("%s", BlueCode)
			#define  TextColorCyan     printf("%s", CyanCode)
			#define  TextColorMagenta  printf("%s", MagentaCode)
			#define  TextColorYellow   printf("%s", YellowCode)
		#endif
	#endif

	#ifndef TIMER_IGNORE_BELOW_THRESHOLD
	#define TIMER_IGNORE_BELOW_THRESHOLD 1
	#endif

	#ifndef TIMER_WARNING_THRESHOLD
	#define TIMER_WARNING_THRESHOLD  0.001  // In seconds (1ms) - yellow messages
	#endif

	#ifndef TIMER_ALERT_THRESHOLD
	#define TIMER_ALERT_THRESHOLD    0.005  // In seconds (5ms) - red messages
	#endif


//================================================
// Get system up-time in seconds
//================================================
	static inline double _GetSeconds(void)
	{
		return((double)E_GetMicroseconds() / 1000000.0);
	}


#define _GetTime() E_GetMicroseconds()


	static inline void _EntrySet(ETmTimerStack* PTimerStack, const char* PName, const TIMERTime PNow)
	{
		if(PTimerStack->Entries_Count >= PTimerStack->Entries_CountAllocated)
		{
			int  LNewAllocCount = PTimerStack->Entries_Count * 2;
			if(LNewAllocCount == 0)	LNewAllocCount = TIMER_INITIAL_ALLOC;
			ETm_Resize(ETmTimerEntry, LNewAllocCount, PTimerStack->Entries);
			PTimerStack->Entries_CountAllocated = LNewAllocCount;
		}
		ETmTimerEntry*  LEntry = PTimerStack->Entries + PTimerStack->Entries_Count;

		LEntry->FunctionName      = NULL;
		LEntry->FileName          = NULL;
		LEntry->FWName            = NULL;
		LEntry->LineNumber        = 0;
 		LEntry->Start             = PTimerStack->PrevTime;
		LEntry->End               = PNow;
		LEntry->ParentIndex       = PTimerStack->CurrentParentIndex;
		LEntry->Level             = PTimerStack->CurrentLevel;
		LEntry->ChildIndex        = -1;
		LEntry->RightSiblingIndex = -1;
		LEntry->Bracket           = '[';

		const int  LL = (int)strlen(PName);
		if(LL >= TIMER_MAX_NAMELEN) { memcpy(LEntry->Name, PName, TIMER_MAX_NAMELEN - 1);LEntry->Name[TIMER_MAX_NAMELEN - 1] = 0; }
		else memcpy(LEntry->Name, PName, LL + 1);
		PTimerStack->Entries_Count++;
	}


	#define TIMER_INIT(mTimerStack)\
		do {\
			extern EBool  mTimerStack##_Initialized;if(!mTimerStack##_Initialized)\
			{\
				memset(&mTimerStack, 0, sizeof(ETmTimerStack));\
				mTimerStack.CurrentParentIndex = -1;\
				mTimerStack##_Initialized = TRUE;\
			}\
			mTimerStack.PrevTime = _GetTime();mTimerStack.Entries_Count = 0;\
		} while(0)

	#define TIMER_RESET(mTimerStack)\
		do {\
			mTimerStack.PrevTime = _GetTime();\
		} while(0)

	#define TimerPrefix  do { TextColorCyan;printf("[Profiler]");fflush(stdout);TextColorNormal; } while(0)
	#define TimerPrefixF do { TextColorCyan;fprintf(PStream, "[Profiler]");TextColorNormal; } while(0)


	#if TIMER_IGNORE_BELOW_THRESHOLD

		#define TIMESTAMP(mTimerStack, m)\
		do {\
			const TIMERTime  lNow = _GetTime(), lTimePassed = lNow - mTimerStack.PrevTime;\
			if(lTimePassed >= TIMER_ALERT_THRESHOLD)\
			{\
				_EntrySet(mTimerStack, m);\
\
				if(TIMER_PRINTOUT_ENABLE)\
				{\
					TimerPrefix;printf("%s() %s:%d  ", __FUNCTION__, __FILE__, __LINE__);\
					TextColorRed;\
					printf("[%s] % 7.3lf ms\n", m, lTimePassed * 1000.0);fflush(stdout);\
					TextColorNormal;\
				}\
			}\
			else if(lTimePassed >= TIMER_WARNING_THRESHOLD)\
			{\
				_EntrySet(mTimerStack, m);\
\
				if(TIMER_PRINTOUT_ENABLE)\
				{\
					TimerPrefix;printf("%s() %s:%d  ", __FUNCTION__, __FILE__, __LINE__);\
					TextColorYellow;\
					printf("[%s] % 7.3lf ms\n", m, lTimePassed * 1000.0);fflush(stdout);\
					TextColorNormal;\
				}\
			}\
			mTimerStack.PrevTime = lNow;\
		} while(0)

	#else

		#define TIMESTAMP(mTimerStack, m)\
		do {\
			const TIMERTime  lNow = _GetTime(), lTimePassed = lNow - mTimerStack.PrevTime;\
			if(lTimePassed >= TIMER_ALERT_THRESHOLD)\
			{\
				_EntrySet(mTimerStack, m);\
\
				if(TIMER_PRINTOUT_ENABLE)\
				{\
					TimerPrefix;printf("%s() %s:%d  ", __FUNCTION__, __FILE__, __LINE__);\
					TextColorRed;\
					printf("[%s] % 7.3lf ms\n", m, lTimePassed * 1000.0);fflush(stdout);\
					TextColorNormal;\
				}\
			}\
			else if(lTimePassed >= TIMER_WARNING_THRESHOLD)\
			{\
				_EntrySet(mTimerStack, m);\
\
				if(TIMER_PRINTOUT_ENABLE)\
				{\
					TimerPrefix;printf("%s() %s:%d  ", __FUNCTION__, __FILE__, __LINE__);\
					TextColorYellow;\
					printf("[%s] % 7.3lf ms\n", m, lTimePassed * 1000.0);fflush(stdout);\
					TextColorNormal;\
				}\
			}\
			else\
			{\
				_EntrySet(mTimerStack, m);\
\
				if(TIMER_PRINTOUT_ENABLE)\
				{\
					TimerPrefix;printf("%s() %s:%d  [%s] % 7.3lf ms\n", __FUNCTION__, __FILE__, __LINE__, #m, lTimePassed * 1000.0);fflush(stdout);\
				}\
			}\
			mTimerStack.PrevTime = lNow;\
		} while(0)

	#endif


	static inline void TimeStamp_PUSH(ETmTimerStack* PTimerStack, const char* PName, const char* PFunctionName, const char* PFileName, const int PLine,  const char PBracket)
	{
		const TIMERTime  lNow = _GetTime();

		if(PTimerStack->Verbosity)
		{
			TimerPrefix;printf("%s:%d", __FILE__, __LINE__);
			int  LC = 0;for(;LC < PTimerStack->CurrentLevel;LC++) printf(" ");
			printf("PUSH [%s]\n", PName);
			fflush(stdout);
		}

		{
			_EntrySet(PTimerStack, PName, lNow);

			PTimerStack->PrevTime = lNow;

			ETmTimerEntry*  LEntry = PTimerStack->Entries + PTimerStack->Entries_Count - 1;
			LEntry->FunctionName    = PFunctionName;
			LEntry->FileName        = PFileName;
			LEntry->LineNumber      = PLine;
			LEntry->Bracket         = PBracket;

			PTimerStack->CurrentParentIndex = PTimerStack->Entries_Count - 1;
			PTimerStack->CurrentLevel += 1;

			LEntry->Start  = lNow;
			LEntry->End    = lNow;  // The last child that "pops" back to this entry will set this
		}
	}


	#define TIMER_PUSH(mTimerStack, mName)\
	do {\
		TimeStamp_PUSH(&(mTimerStack), mName, __FUNCTION__, __FILE__, __LINE__, '[');\
	} while(0)

	#define TIMER_PUSH_WITH_OPTIONS(mTimerStack, mName, mBracket)\
	do {\
		TimeStamp_PUSH(&(mTimerStack), mName, __FUNCTION__, __FILE__, __LINE__, mBracket);\
	} while(0)


//========================================================================================
// Append string to the current timer Entry's Name
// This is very useful if you want to start a timer, then
// use a conditional or a loop to determine a name suffix
// For example:
//  TIMER_PUSH(My_TimerStack, "Picking fruit:")   // The timer starts here
//  switch(MyStruct->FruitType)
//  {
//   case MyBANANA:  TimeStamp_NAME_APPEND("Banana");/* Do some timed stuff */;break;
//   case MyORANGE:  TimeStamp_NAME_APPEND("Orange");/* Do some timed stuff */;break;
//  }
//  TIMER_POP(My_TimerStack)   // The timer ends here
//========================================================================================
	static inline void TimeStamp_NAME_APPEND(ETmTimerStack* PTimerStack, const char* PString)
	{
		ETmTimerEntry*  lEntry = PTimerStack->Entries + PTimerStack->Entries_Count - 1;
		const int  LL = (int)strlen(lEntry->Name);
		const int  LL1 = (int)strlen(PString);
		const int  LL2 = LL + LL1 > TIMER_MAX_NAMELEN - 1 ? LL1 : TIMER_MAX_NAMELEN - 1 - LL;
		memcpy(lEntry->Name + LL, PString, LL2);lEntry->Name[LL + LL2] = 0;
	}


	static inline void Timestamp_POP(ETmTimerStack* PTimerStack)
	{
		const TIMERTime  lNow = _GetTime();

		if(PTimerStack->CurrentLevel > 0)
		{

			PTimerStack->CurrentLevel -= 1;
			const ETmTimerEntry*  lEntry = PTimerStack->Entries + PTimerStack->Entries_Count - 1;
			ETmTimerEntry*        lParent = PTimerStack->Entries + PTimerStack->CurrentParentIndex;
			if(lEntry != lParent)    // PTimerStack->CurrentParent has descendants : update its End time
			{
				lParent->End = lNow;
			}
			else lParent->End = lNow;


			if(PTimerStack->Verbosity)
			{
				TimerPrefix;printf("%s:%d ", __FILE__, __LINE__);
				int  LC = 0;for(;LC < PTimerStack->CurrentLevel;LC++) printf(" ");
//				printf("POP [%s -> %s]\n", lParent->Name, lParent->ParentIndex >= 0 ? PTimerStack->Entries[lParent->ParentIndex].Name : "-");
				printf("POP\n");
			}

			PTimerStack->CurrentParentIndex = lParent->ParentIndex;
		}
		else
		{
			TimerPrefix;printf("%s:%d  ", __FILE__, __LINE__);
			TextColorRed;
			printf("TIMER STACK UNDERFLOW\n");fflush(stdout);
			TextColorNormal;
		}
	}

	#define TIMER_POP(mTimerStack)\
	do {\
		Timestamp_POP(&(mTimerStack));\
	} while(0)

	static void	_UnitsClear(ETmTimerStack* PTimerStack)
	{
		if(PTimerStack->ProfilingUnits)
		{
			int  LC = 0;const int  LN = PTimerStack->ProfilingUnits_Count;ETmProfilingUnit*  LUnit = PTimerStack->ProfilingUnits;for(;LC < LN;LC++)
			{
				ETm_FreeChk(LUnit->EntryIndices);
				memset(LUnit, 0, sizeof(ETmProfilingUnit));
			}
			PTimerStack->ProfilingUnits_Count = 0;
		}
	}


	static inline void _UnitsFree(ETmTimerStack* PTimerStack)
	{
		if(PTimerStack->ProfilingUnits)
		{
			int  LC = 0;const int  LN = PTimerStack->ProfilingUnits_Count;ETmProfilingUnit*  LUnit = PTimerStack->ProfilingUnits;for(;LC < LN;LC++)
			{
				ETm_FreeChk(LUnit->EntryIndices);
			}
			ETm_FreeClr(PTimerStack->ProfilingUnits);PTimerStack->ProfilingUnits_Count = PTimerStack->ProfilingUnits_CountAllocated = 0;
		}
	}


//================================================================================================
// Set FWName fields to point to the Name field of the first Entry that has the same name.
// This will allow us to replace string comparisons with pointer comparisons.
//================================================================================================
/*
	static inline void _SetFWNames(ETmTimerStack* PTimerStack)
	{
		int  LC = 0;const int  LN = PTimerStack->Entries_Count;ETmTimerEntry*  LEntry = PTimerStack->Entries;for(;LC < LN;LC++, LEntry++)
		{
			if(LEntry->FWName == NULL)
			{
				const char*  LName = LEntry->Name;
				LEntry->FWName = LName;
// Find other Entries with the same name and set their FWName as well
//
				int  LC1 = LC + 1;ETmTimerEntry*  LEntry1 = PTimerStack->Entries;for(;LC1 < LN;LC1++, LEntry1++)
				{
					if(LEntry1->FWName == NULL && strcmp(LName, LEntry1->Name) == 0) LEntry1->FWName = LName;
				}
			}
		}
	}
*/


//================================================================================================
// Process timing information in a TimerStack.
// To make timings more accurate and minimize interference with the code being profiled,
// data collection overhead during processing is kept to an absolute minimum.
// You can shift that processing to a "safe" time, by calling this function, for example at the
// end of a frame in a video game, or after data for several frames are collected.
//================================================================================================
	static inline void TimerStack_CollectProfilingInfo(ETmTimerStack* PTimerStack)
	{
// Fill out hierarchy links : Child and RightSibling.
// This will allow us to more efficiently parse the hierarchy.
//
		{
			int  LC = 0;const int  LN = PTimerStack->Entries_Count;ETmTimerEntry*  LEntry = PTimerStack->Entries;for(;LC < LN;LC++, LEntry++)
			{
				const int  LParentIndex = LEntry->ParentIndex;if(LParentIndex)
				{
					ETmTimerEntry*  LParent = PTimerStack->Entries + LParentIndex;
					if(LParent->ChildIndex == -1) LParent->ChildIndex = LC;
				}
			}
		}


		_UnitsClear(PTimerStack);


		{
			int  LC = 0;const int  LN = PTimerStack->Entries_Count;ETmTimerEntry*  LEntry = PTimerStack->Entries;for(;LC < LN;LC++, LEntry++)
			{
				if(LEntry->FWName == NULL)
				{
					ETmProfilingUnit*  LNewProfilingUnit;

					ETmList_StructAddAClr(ETmProfilingUnit, PTimerStack->ProfilingUnits, TIMER_ALLOC_INCREMENT,  LNewProfilingUnit);

					ETmList_IntAdd(LNewProfilingUnit->EntryIndices, LC);

					const char*  LName = LEntry->Name;
					LNewProfilingUnit->Name = LName;
					LEntry->FWName = LName;
// Find other Entries with the same name and set their FWName as well
//
					int  LC1 = LC + 1;ETmTimerEntry*  LEntry1 = PTimerStack->Entries + LC1;for(;LC1 < LN;LC1++, LEntry1++)
					{
						if(LEntry1->FWName == NULL && strcmp(LName, LEntry1->Name) == 0)
						{
							LEntry1->FWName = LName;
							ETmList_IntAdd(LNewProfilingUnit->EntryIndices, LC1);
						}
					}
				}
			}
		}

// Add up total time for each ProfilingUnit
//
		{
			int  LC = 0;const int  LN = PTimerStack->ProfilingUnits_Count;ETmProfilingUnit*  LUnit = PTimerStack->ProfilingUnits;for(;LC < LN;LC++, LUnit++)
			{
				TIMERTime  LUnitTotalTime = 0;

				int  LC1 = 0;const int  LN1 = LUnit->EntryIndices_Count;for(;LC1 < LN1;LC1++)
				{
					ETmTimerEntry*  LEntry = PTimerStack->Entries + LUnit->EntryIndices[LC1];
					LUnitTotalTime += LEntry->End - LEntry->Start;
				}
				LUnit->TotalTime = LUnitTotalTime;
			}
		}
	}

	static inline void _PadString(const char* PStr, const int PLengthToPadTo)
	{
		const int  LL = (int)strlen(PStr);
		if(LL < PLengthToPadTo)
		{
			int  LC = 0;const int  LN = PLengthToPadTo - LL;for(;LC < LN;LC++) printf(" ");
		}
	}

#define TimerPrefixf  do { if(LPrintPrefix) TimerPrefixF; } while(0)

	static inline void TimerStack_OutputProfilingInfo(ETmTimerStack* PTimerStack, FILE* PStream)
	{
		const double  LTimeDivider = (double)1000000.0;  // We're using microseconds
		const EBool   LPrintPrefix = PStream == stdout;

		TimerPrefixf;fprintf(PStream, "Total entries : %d\n", PTimerStack->Entries_Count);fflush(stdout);

printf("%s() %s:%d %d\n", __FUNCTION__, __FILE__, __LINE__, PTimerStack->ProfilingUnits_Count);fflush(stdout);

		int  LC = 0;const int  LN = PTimerStack->ProfilingUnits_Count;ETmProfilingUnit*  LUnit = PTimerStack->ProfilingUnits;for(;LC < LN;LC++, LUnit++)
		{
			ETmTimerEntry*  LEntry = PTimerStack->Entries + LUnit->EntryIndices[0];

			TimerPrefixf;fprintf(PStream, " ");

			if(LEntry->Level > 0)
			{
#if TIMER_DRAW_HIERARCHY_LINKS
				{ int  LCI = 0;const int  LNI = LEntry->Level - 1;for(;LCI < LNI;LCI++) printf(" "); }

				TextColorGreen;
				const ETmProfilingUnit*  LNextUnit = LC <  LN - 1 ? LUnit + 1 : NULL;

				if(LNextUnit)
				{
					if(LNextUnit->NextSiblingIndex >= 0) printf("");
					else fprintf(PStream, "");
				}
				else fprintf(PStream, "");
				TextColorNormal;
#else
				{ int  LCI = 0;const int  LNI = LEntry->Level;for(;LCI < LNI;LCI++) printf(" "); }
#endif

			}
//			printf("[%s]", LUnit->Name);_PadString(LUnit->Name, TIMER_NAME_COLUMN_WIDTH - LEntry->Level);

			char  LClosingBracket = ' ';
			switch(LEntry->Bracket)
			{
				case '[':  LClosingBracket = ']';break;
				case '{':  LClosingBracket = '}';break;
				case '|':  LClosingBracket = '|';break;
				case ':':  LClosingBracket = ':';break;
				case '#':  LClosingBracket = '#';break;
				case '%':  LClosingBracket = '%';break;
				case '*':  LClosingBracket = '*';break;
				case '-':  LClosingBracket = '-';break;
				case '=':  LClosingBracket = '=';break;
				case '<':  LClosingBracket = '>';break;
				case '(':  LClosingBracket = ')';break;
				case '"':  LClosingBracket = '"';break;
				case '/':  LClosingBracket = '/';break;
				case '\'':  LClosingBracket = '\'';break;
			}

			fprintf(PStream, "%c%s%c", LEntry->Bracket, LUnit->Name, LClosingBracket);_PadString(LUnit->Name, TIMER_NAME_COLUMN_WIDTH);
			fprintf(PStream, " invocations:%5d  total time: % 5.2lfms", LUnit->EntryIndices_Count, (double)(LUnit->TotalTime) / LTimeDivider * 1000.0);

			fprintf(PStream, "  from: %s() %s:%d\n", LEntry->FunctionName, LEntry->FileName, LEntry->LineNumber);fflush(stdout);
			fflush(PStream);
		}
	}

	static inline void TimerStack_PrintProfilingInfo(ETmTimerStack* PTimerStack)
	{
		TimerStack_OutputProfilingInfo(PTimerStack, stdout);
	}


#if TIMER_USE_AUTO_SCOPE
namespace Timer
{
	struct Scope
	{
		ETmTimerStack*  TimerStack;
		const char*     Name;

		Scope(ETmTimerStack* PTimerStack, const char* PName,  const char* PFunctionName, const char* PFileName, const int PLine) : TimerStack(PTimerStack), Name(PName)
		{
			TimeStamp_PUSH(PTimerStack, PName, PFunctionName, PFileName, PLine);
//			TIMER_PUSH(PTimerStack, PName);
		};

		~Scope(void) { Timestamp_POP(TimerStack); };
	};

};

	#define TIMER_SCOPE(mTimerStack, mName) Timer::Scope  TimerScopeStart_##__LINE__(&(mTimerStack), mName, __FUNCTION__, __FILE__, __LINE__)

#else

	#define TIMER_SCOPE(mTimerStack, mName) do { printf("%s() %s:%d profiler auto-scoping is disabled (TIMER_USE_AUTO_SCOPE = 0)\n", __FUNCTION__, __FILE__, __LINE__);fflush(stdout); } while(0)

#endif


#else

	#define TIMER_INIT(m)                                          do { } while(0)
	#define TIMER_RESET(m)                                         do { } while(0)
	#define TIMESTAMP(mTimerStack, mName)                          do { } while(0)
	#define TIMER_PUSH(mTimerStack, m)                             do { } while(0)
	#define TIMER_PUSH_WITH_OPTIONS(mTimerStack, mName, mBracket)  do { } while(0)
	#define TIMER_POP(mTimerStack)                                 do { } while(0)
	#define TIMER_SCOPE(mTimerStack, m)                            do { } while(0)

#endif  // TIMER_ENABLE


#endif  // _ETiming_h
