//========================================================================================================
// Generic Tool Class
//
// Tools are the fundamental building blocks for function-trees, such as:
// - image-processing trees
// - shader trees
// etc.
//
// An ETool is an EObject with inputs, outputs and a "signal-propagation" mechanism
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 04:40:33
//
// Copyright (C) 1994-2021 By Gabor Nagy. All rights reserved.
//========================================================================================================
#ifndef _ETool_h
#define _ETool_h


#ifndef _EClass_h
#include <EClass.h>
#endif

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

typedef struct EToolClass_S*  EToolClassPtr;
typedef struct EToolClass_S   EToolClass;
typedef struct ETool_S        ETool;


#define E_ToolGetCurrentTime  E_GetMicroseconds


typedef struct EToolInterface_S  EToolInterface;

typedef struct
{
	EEntityCore;

	EToolInterface*  Interface0;
	EToolInterface*  Interface1;
} EToolConnection;




#define E_IN   0x01
#define E_OUT  0x02

enum
{
	EToolSUCCESS = 0,
	EToolALREADY_CONNECTED,
	EToolIF0_ALREADY_HAS_AN_INPUT,
	EToolIF1_ALREADY_HAS_AN_INPUT,
	EToolDIRECTION_MISMATCH,
	EToolTYPE_MISMATCH
};


struct EToolInterface_S
{
	EResource          Resource;          // Name, Type structoffset etc. of resource in the current Tool
	EChangeFlags       Changed;
	ETime              ChangedTime;
	ETool*             Tool;              // Pointer back to the Tool to which interface belongs

	EToolConnection*   InputConnection;   // There can be only one input connection per interface
	EToolInterface*    InputInterface;    // There can be only one input connection per interface
	EChangeFlags       InputChanged;      // If the Tool of this interface is inactive when a signal comes in, these are updated, 
	ETime              InputChangedTime;  // so when the Tool is activated, it will know whether to update itself and how.

	EList(, EToolConnection*, OutputConnections);
	unsigned char      Direction;         // E_IN, E_OUT or both: tells whether this interface is an input, an output or both
};


typedef struct
{
	ETool*             Tool;  // Pointer the Tool to which Connections go

	EList(, EToolConnection*, Connections);
} EToolConnectionList;


#define EToolCore\
	EObjectCore(ETool);\
	EList(, EToolInterface*, Interfaces);\
	EList(, EToolConnectionList, ToolConnectionLists);  /* This is a private member used for an optimization: it lists Connections per target tool, */\
	                                                    /* so signals and connection changes can be propagated to other tools in a single call      */\
	EBool                 CacheResult, Active



#define EToolClassCore\
	EBaseClassCore(ETool);\
	int   (*Execute)(ETool* PTool, const EBool PGUIJustPoppedUp);\
	void  (*FreeCache)(ETool* PTool);\
	int   (*ConnectionChanged)(ETool* PTool);\
	int   (*InputChanged)(ETool* PTool);\
	void  (*ActivateDeactivate)(ETool* PTool, const EBool PActivate)


struct EToolClass_S
{
	EToolClassCore;
};


struct ETool_S
{
	EToolCore;
};

#define EM_ToolClassInit(mDescriptor)\
do {\
	EBaseClass_Init(mDescriptor);\
	LClass->ActivateDeactivate = mDescriptor->ActivateDeactivate;\
} while(0)

#define EM_ToolClassInitNew(mDescriptor)\
do {\
	EBaseClass_InitNew(mDescriptor);\
	LClass->ActivateDeactivate = mDescriptor->ActivateDeactivate;\
} while(0)

extern EToolClass  ETool_class[];

extern ETool*           ETool_New(void* PCl);
extern EToolInterface*  ETool_InterfaceNew(ETool* PTool, const EResource* PResource);
EToolInterface*         ETool_AddInterface(ETool* PTool, const EResource* PResource, const char PDirection);
extern void             EToolInterface_Free(EToolInterface* PInterface);

extern EToolConnection*  EToolConnection_New(void);
extern void              EToolConnection_Free(EToolConnection* PConnection);
extern int               ETool_InterfacesConnect(EToolInterface* PInterface0, EToolInterface* PInterface1);
extern void              ETool_InterfacesDisconnect(EToolConnection* PConnection);

extern void  ETool_Execute(ETool* PTool, const EBool PPropagateResult, const EBool PGUIJustPoppedUp);
extern void  ETool_PropagateResult(ETool* PTool);


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif

#endif // _ETool_h
