//================================================================
// lib2D
//
// - Drawing
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 08:16:24
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================
#ifndef _EImageDraw_h
#define _EImageDraw_h


#include <stdio.h>

#ifndef _EImage_h
#include <Image/Image.h>
#endif

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif



// Used for drawing polygons
//
typedef struct
{
	EIndex  Start, End;
} EpPolyXRange;


#define EpRegionINSIDE             0x01
#define EpRegionBOUNDARY           0x02
#define EpRegionOUTSIDE            0x04
#define EpRegionALL                0x07


//========================================================
// Collect starting and ending X positions of a 2D line
//========================================================
static inline char Ep_PLine(EpPolyXRange* PEpPolyXRanges,  const EIndex PX1, const EIndex PY1,  const EIndex PX2, const EIndex PY2,  const EIndex PImageSizeX, const EIndex PImageSizeY)
{
	char    LXU;
	EIndex  LYD = PY2 - PY1, LYD2;
	EIndex  LErr = 0;

	EIndex  LXD = PX2 - PX1;
	EIndex  LX = PX1;

	if(PY1 == PY2) return(0);  // Horizontal line

	if(LYD < 0) LYD = -LYD;

	if(LXD < 0)
	{
		LXD = -LXD;LXU = -1;
	}
	else LXU = 1;

// This line makes some edge lines better.
// It's not essential.
//
	if(LYD > LXD) LErr = LYD >> 1;

//Printf(" %d to %d   LXD:%d  LXU:%d   LYD:%d\n", PX1, PX2, LXD, LXU, LYD);fflush(stdout);

// Record x values along line
//
	LYD2 = LYD - 1;

	if(LXU == 1)
	{
		if(PY1 > PY2)
		{
			EIndex  LY = PY1;for(; LY > PY2 - 1; LY--)
			{
				if(LY >= 0 && LY < PImageSizeY)
				{
					PEpPolyXRanges[LY].Start = LX;
				}
				LErr += LXD;
				if(LErr > LYD2) do
				{
					LErr -= LYD;LX++;
				} while(LErr >= LYD);
			}
			return(0);
		}
		else
		{
			EIndex  LY = PY1;for(;LY < PY2+1; LY++)
			{
				if(LY >= 0 && LY < PImageSizeY)
				{
					PEpPolyXRanges[LY].End = LX;
				}
				LErr += LXD;
				if(LErr > LYD2) do
				{
					LErr -= LYD;LX++;
				} while(LErr >= LYD);
			}
			return(0);
		}
	}
	else
	{
		if(PY1 > PY2)
		{
			EIndex  LY = PY1;for(;LY > PY2 - 1;LY--)
			{
				if(LY >= 0 && LY < PImageSizeY)
				{
					PEpPolyXRanges[LY].Start = LX;
				}
				LErr += LXD;
				if(LErr > LYD2) do
				{
					LErr -= LYD;LX--;
				} while(LErr >= LYD);
			}
			return(0);
		}
		else
		{
			EIndex  LY = PY1;for(;LY < PY2 + 1;LY++)
			{
				if(LY >= 0 && LY < PImageSizeY)
				{
					PEpPolyXRanges[LY].End = LX;
				}
				LErr += LXD;
				if(LErr > LYD2) do
				{
					LErr -= LYD;LX--;
				} while(LErr >= LYD);
			}
		}
	}

	return(0);
}


//========================================================================================================================
// Clip coordinate for drawing (destination buffer only, no source)
//
// Arguments
//  mType      Data type of the coordinate
//  mDst       Starting position in the destination
//  mDstSize   Size of the destination buffer (e.g. width of the destination image)
//  mDrawSize  Size of region drawn
//  mDoDraw    Flag to be set to FALSE, if the coordinate is out of range
//
// Description
//  Clips a coordinate, used for copying. It will adjust mSrc, mCopySize and mDst.
//  If the coordinates are completely out of range, this version sets the flag mDoDraw to FALSE.
//  This macro may be used for multiple dimensions / coordinates.
//  For example:
//   EBool  LDoDraw = TRUE;Ep_Clip1D(int,  LSrcX, LSrcImageSizeX,  LDstX, LDstImageSizeX,  LCopySizeX,  LDoDraw);if(LDoDraw)
//   {
//     Ep_Clip1D(int,  LSrcY, LSrcImageSizeY,  LDstY, LDstImageSizeY,  LCopySizeY);if(LDoDraw)...
//========================================================================================================================
#define Ep_Clip1D(mType,  mDst, mDstSize,  mDrawSize,  mDoDraw) do\
{\
	if(mDst < 0)\
	{\
		if(mDst + mDrawSize <= 0) (mDoDraw) = FALSE;\
		else\
		{\
			const mType  lDiff = -mDst;\
			mDrawSize -= lDiff;if(mDrawSize < 1) (mDoDraw) = FALSE;\
			else { mDst = 0;if(mDst + mDrawSize >= mDstSize) mDrawSize = mDstSize - mDst; }\
		}\
	}\
	else\
	{\
		if(mDst >= mDstSize) (mDoDraw) = FALSE;\
		else if(mDst + mDrawSize >= mDstSize) mDrawSize = mDstSize - mDst;\
	}\
} while(0)


//========================================================================================================================
// Clip coordinate for drawing into a destination: supports explicit destination clip region (in addition to the destination size)
//
// Arguments
//  mType         Data type of the coordinate
//  mDst          Starting position in the destination
//  mDstSize      Size of the destination buffer (e.g. width of the destination image)
//  mDstClipMin   Clip region minimum
//  mDstClipSize  Clip region length. If 0, no explicint clipping (equivalent to Ep_Clip1D)
//  mDrawSize     Size of region drawn
//  mDoDraw       Flag to be set to FALSE, if the coordinate is out of range
//
// Description
//  Clips a coordinate, used for copying. It will adjust mSrc, mCopySize and mDst.
//  If the coordinates are completely out of range, this version sets the flag mDoDraw to FALSE.
//  This macro may be used for multiple dimensions / coordinates.
//  For example:
//   EBool  LDoDraw = TRUE;Ep_Clip1DC(int,  LSrcX, LSrcImageSizeX, LClipX, LClipSizeX,  LDstX, LDstImageSizeX,  LCopySizeX,  LDoDraw);if(LDoDraw)
//   {
//     Ep_Clip1DC(int,  LSrcY, LSrcImageSizeY, LClipY, LClipSizeY,  LDstY, LDstImageSizeY,  LCopySizeY);if(LDoDraw)...
//========================================================================================================================
#define Ep_Clip1DC(mType,  mDst, mDstSize, mDstClipMin, mDstClipSize,  mDrawSize,  mDoDraw) do\
{\
	if(mDst < mDstClipMin)\
	{\
		if(mDst + mDrawSize <= mDstClipMin) (mDoDraw) = FALSE;\
		else\
		{\
			const mType  lDstMax = mDstClipSize > 0 ? mDstClipMin + mDstClipSize : mDstSize;\
			if(mDst >= lDstMax) (mDoDraw) = FALSE;\
			else\
			{\
				mDst = mDstClipMin;\
				const mType  lDiff = mDstClipMin - mDst;\
				mDrawSize -= lDiff;if(mDst + mDrawSize >= lDstMax) mDrawSize = lDstMax - mDst;\
			}\
		}\
	}\
	else\
	{\
		const mType  lDstMax = mDstClipSize > 0 ? mDstClipMin + mDstClipSize : mDstSize;\
		if(mDst >= lDstMax) (mDoDraw) = FALSE;\
		else if(mDst + mDrawSize >= lDstMax) mDrawSize = lDstMax - mDst;\
	}\
} while(0)


//========================================================================================================================
// Clip coordinate for src->dst copying
//
// Arguments
//  mType      Data type of the coordinate
//  mSrc       Starting position in the source
//  mSrcSize   Size of the source buffer (e.g. width of the source image)
//  mDst       Starting position in the destination
//  mDstSize   Size of the destination buffer (e.g. width of the destination image)
//  mCopySize  Size of region copied
//
// Description
//  Clips a coordinate, used for copying. It will adjust mSrc, mCopySize and mDst.
//  If the coordinates are completely out of range, this version causes the calling function to return with E_CLIPPED.
//  This macro may be used for multiple dimensions / coordinates.
//  For example:
//   Ep_Clip1DCopy_r(int,  LSrcX, LSrcImageSizeX,  LDstX, LDstImageSizeX,  LCopySizeX)
//   Ep_Clip1DCopy_r(int,  LSrcY, LSrcImageSizeY,  LDstY, LDstImageSizeY,  LCopySizeY)
//========================================================================================================================
#define Ep_Clip1DCopy_r(mType,  mSrc, mSrcSize,  mDst, mDstSize,  mCopySize) do\
{\
	if(mSrc < 0)\
	{\
		if(mSrc + mCopySize <= 0) return(E_CLIPPED);\
		else\
		{\
			const mType  lDiff = -mSrc;\
			mDst += lDiff;mCopySize -= lDiff;if(mCopySize < 1 || mDst >= mDstSize) return(E_CLIPPED);\
			else { mSrc = 0;if(mSrc + mCopySize >= mSrcSize) mCopySize = mSrcSize - mSrc; }\
		}\
	}\
	else\
	{\
		if(mSrc >= mSrcSize) return(E_CLIPPED);\
		else if(mSrc + mCopySize >= mSrcSize) mCopySize = mSrcSize - mSrc;\
	}\
\
	if(mDst < 0)\
	{\
		if(mDst + mCopySize <= 0) return(E_CLIPPED);\
		else\
		{\
			const mType  lDiff = -mDst;\
			mSrc += lDiff;mCopySize -= lDiff;if(mCopySize < 1 || mSrc >= mSrcSize) return(E_CLIPPED);\
			else { mDst = 0;if(mDst + mCopySize >= mDstSize) mCopySize = mDstSize - mDst; }\
		}\
	}\
	else\
	{\
		if(mDst >= mDstSize) return(E_CLIPPED);\
		else if(mDst + mCopySize >= mDstSize) mCopySize = mDstSize - mDst;\
	}\
} while(0)


//========================================================================================================================
// Clip coordinate for src->dst copying: supports explicit destination clip region (in addition to the destination size)
//
// Arguments
//  mType         Data type of the coordinate
//  mSrc          Starting position in the source
//  mSrcSize      Size of the source buffer (e.g. width of the source image)
//  mDst          Starting position in the destination
//  mDstSize      Size of the destination buffer (e.g. width of the destination image)
//  mDstClipMin   Clip region minimum
//  mDstClipSize  Clip region length. If 0, no explicint clipping (equivalent to Ep_Clip1DCopy_r)
//  mCopySize     Size of region copied
//
// Description
//  Clips a coordinate, used for copying. It will adjust mSrc, mCopySize and mDst.
//  If the coordinates are completely out of range, this version causes the calling function to return with E_CLIPPED.
//  This macro may be used for multiple dimensions / coordinates.
//  For example:
//   Ep_Clip1DCopyC_r(int,  LSrcX, LSrcImageSizeX, LClipX, LClipSizeX,  LDstX, LDstImageSizeX,  LCopySizeX);
//   Ep_Clip1DCopyC_r(int,  LSrcY, LSrcImageSizeY, LClipY, LClipSizeY,  LDstY, LDstImageSizeY,  LCopySizeY);
//========================================================================================================================
#define Ep_Clip1DCopyC_r(mType,  mSrc, mSrcSize,  mDst, mDstSize, mDstClipMin, mDstClipSize,  mCopySize) do\
{\
	if(mSrc < 0)\
	{\
		if(mSrc + mCopySize <= 0) return(E_CLIPPED);\
		else\
		{\
			const mType  lDiff = -mSrc;\
			mDst += lDiff;mCopySize -= lDiff;if(mCopySize < 1 || mDst >= mDstSize) return(E_CLIPPED);\
			else { mSrc = 0;if(mSrc + mCopySize >= mSrcSize) mCopySize = mSrcSize - mSrc; }\
		}\
	}\
	else\
	{\
		if(mSrc >= mSrcSize) return(E_CLIPPED);\
		else if(mSrc + mCopySize >= mSrcSize) mCopySize = mSrcSize - mSrc;\
	}\
\
	if(mDst < mDstClipMin)\
	{\
		if(mDst + mCopySize <= mDstClipMin) return(E_CLIPPED);\
		else\
		{\
			const mType  lDiff = mDstClipMin - mDst;\
			mSrc += lDiff;\
			if(mSrc >= mSrcSize) return(E_CLIPPED);\
			else\
			{\
				mCopySize -= lDiff;\
				const mType  lDstMax = mDstClipSize > 0 ? mDstClipMin + mDstClipSize : mDstSize;mDst = mDstClipMin;\
				if(mDst + mCopySize >= lDstMax) mCopySize = lDstMax - mDst;\
			}\
		}\
	}\
	else\
	{\
		const mType  lDstMax = mDstClipSize > 0 ? mDstClipMin + mDstClipSize : mDstSize;\
		if(mDst >= lDstMax) return(E_CLIPPED);\
		else if(mDst + mCopySize >= lDstMax) mCopySize = lDstMax - mDst;\
	}\
} while(0)


extern void       Ep_DrawPoints2DRGBAp8(EpImage* PImage,  const EIndex_2* PPoints, const int PCount,  const EcRGBAp8Color PSColor);
extern void       Ep_DrawPoints2DSetA8(EpImage* PImage,  const EIndex_2* PPoints, const int PCount,  const uint8_t PA);
extern void       Ep_DrawPoints2DRGBAi(EpImage* PImage,  const EIndex_2* PPoints, const int PCount,  const EcRGBAiColor* PSColor, const uint8_t PAlphaOp);
extern void       Ep_DrawColoredPoints2DRGBAi(EpImage* PImage,  const EpColoredPixelRGBAi* PPoints, const int PCount, const uint8_t PAlphaOp);

extern void       Ep_DrawHLineRGBAp8(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PLength,  const EcRGBAp8Color PSColor);
extern void       Ep_DrawHLineRGBAf(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PLength,  const EcRGBAf32Color PSColor);

extern void       Ep_DrawVLineRGBAp8(EpImage* PImage,  EIndex PX, EIndex PY,  EIndex PLength,  const EcRGBAp8Color PSColor);
extern void       Ep_DrawVLineRGBAf(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PLength,  const EcRGBAf32Color PSColor);

extern void       Ep_DrawHLineBlendRGBA8(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PLength,  const EcRGBA8Color* PColor);
extern void       Ep_DrawHLineBlendRGBAf(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PLength,  const EcRGBAf32Color* PColor);

extern void       Ep_DrawVLineBlendRGBA8(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PLength,  const EcRGBA8Color* PColor);
extern void       Ep_DrawVLineBlendRGBAf(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PLength,  const EcRGBAf32Color* PColor);

extern void       Ep_DrawRectangleRGBA8(EpImage* PImage,  const EcRGBA8Color PColor,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EIndex PThickness);
extern void       Ep_DrawRectangleRGBAp8(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EcRGBAp8Color PSColor);
extern void       Ep_DrawRectangleRGBAi(EpImage* PImage,  const EcRGBAiColor* PColor,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EIndex PThickness);
extern void       Ep_DrawRectangleRGBf(EpImage* PImage, const EcRGBf32Color* PColor, const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EIndex PThickness);
extern void       Ep_DrawRectangleRGBAf(EpImage* PImage, const EcRGBAf32Color* PColor, const EIndex PX, const EIndex PY, const EIndex PSizeX, const EIndex PSizeY);
extern void       Ep_DrawRectangleGrayf32(EpImage* PImage,  const float PColor,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EIndex PThickness);

extern void       Ep_DrawRectangleRoundedRGBAf32
(
	EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,
	const int16_t PCornerTopLeftRadius, const int16_t PCornerTopRightRadius,
	const int16_t PCornerBottomLeftRadius, const int16_t PCornerBottomRightRadius,
	const EBool PDrawCorners,
	const EcRGBAf32Color* PColor,
	const uint16_t PInsideBlendMode, const uint16_t POutsideBlendMode
);

extern void       Ep_FillRectangleRGBAp8(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EcRGBAp8Color PColor);
extern void       Ep_FillRectangleRoundedRGBAi(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,
 const EIndex PCornerTopLftRadius, const EIndex PCornerTopRgtRadius,
 const EIndex PCornerBtmLftRadius, const EIndex PCornerBtmRgtRadius,
 const uint8_t PRegions,
 const EcRGBAiColor* PColor,
 const EcRGBAiColor* POutsideColor,
 const uint16_t PTargetBlendFunction);

extern void       Ep_FillRectangleRGBAi(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EcRGBAiColor* PColor);

extern void       Ep_FillRectangleBlendRGBA8_RGBAi(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EcRGBAiColor* PColor);
extern void       Ep_FillRectangleRGBA8_RGBAf(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EcRGBAf32Color* PColor);
extern void       Ep_FillRectangleRGBAf_RGBAf(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const EcRGBAf32Color* PColor);

extern uint32_t*  Ep_FillRectangle(EcRGBAp8Color* PRGBAPixels,  const EIndex PImageSizeX, const EIndex PImageSizeY,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,  const unsigned short PRed, const unsigned short PGreen, const unsigned short PBlue, const unsigned short PAlpha,  const short PChannels);

extern void       Ep_FillGradientRectangleRGBAf(EpImage* PImage,  const int POrientation,  const EIndex PX, const EIndex PY, const EIndex PSizeX, const EIndex PSizeY,  const EcRGBAf32Color* PColor0, const EcRGBAf32Color* PColor1);
extern void       Ep_FillGradientRectangleBlendRGBAf(EpImage* PImage,  const int POrientation,  const EIndex PX, const EIndex PY, const EIndex PSizeX, const EIndex PSizeY,  const EcRGBAf32Color* PColor0, const EcRGBAf32Color* PColor1);
extern void       Ep_FillGradientRectangleExpRGBAf(EpImage* PImage, const int POrientation,  const EIndex PX, const EIndex PY, const EIndex PSizeX, const EIndex PSizeY,  const EcRGBAf32Color* PColor0, const EcRGBAf32Color* PColor1,  const float PExponent);

extern int        Ep_DrawPolygon(EpGray8Pixel* PDGray8Image,  const EIndex PImageSizeX, const EIndex PImageSizeY,  const EIndex PNumOfPoints, const EIndex_2* PPoints, const EpGray8Pixel PColor);

extern int        EpImage_DrawDisc(EpImage* PImage, const EcRGBAf32Color* PColor,  const EIndex POriginX, const EIndex POriginY, const float PRadius, const float PSoftness, 	const EBool PAntiAliasing);

extern void EpImage_DrawQuarterCircleRGBA8
(
	EpImage* PImage, EcRGBA8Color* PDstPixels,  const EcRGBAf32Color* PStartColor, const EcRGBAf32Color* PEndColor, const float PThickness,
	const EIndex POriginX, const EIndex POriginY, const EIndex PRadius, const uint8_t PQuadrant,
	const uint16_t PInsideBlendMode, const uint16_t POutsideBlendMode,
	const EBool PAntiAliasing
);

extern void EpImage_DrawQuarterDiscRGBA8
(
	EpImage* PImage, EcRGBA8Color* PDstPixels,  const EcRGBAiColor* PInsideColor, const EcRGBAiColor* POutsideColor,
	const EIndex POriginX, const EIndex POriginY, const EIndex PRadius,
	const uint8_t PQuadrant,
	const uint8_t PRegions,
	const uint8_t PAlphaSource,
	const EBool PMultiplyByAlpha,
	const EBool PAntiAliasing
);

extern void EpImage_DrawQuarterDiscBandsRGBA8
(
	EpImage* PImage, EcRGBA8Color* PDstPixels,  const EIndex POriginX, const EIndex POriginY, const EIndex PRadius, const EpBand* PBands,
	const uint8_t PQuadrant,
	const uint16_t PInsideBlendMode, const uint16_t POutsideBlendMode,
	const EBool PAntiAliasing
);

extern void EpImage_RectangleDrawCornerBands
(
	EpImage* PImage, const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,
	const EIndex PTopLftRadius, const EpBand* PTopLftBands,  const EIndex PTopRgtRadius, const EpBand* PTopRgtBands,
	const EIndex PBtmLftRadius, const EpBand* PBtmLftBands,  const EIndex PBtmRgtRadius, const EpBand* PBtmRgtBands,
	const uint16_t PInsideBlendMode, const uint16_t POutsideBlendMode
);

extern int EpImage_RoundOffCorners(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EIndex PSizeX, const EIndex PSizeY,
	const int PCornerTopLeftRadius, const int PCornerTopRightRadius,
	const int PCornerBottomLeftRadius, const int PCornerBottomRightRadius,
	const EcRGBAiColor* POutsideColor, const EcRGBAiColor* PColor, const uint8_t PRegions, const EBool PAntiAliasing);


// One vertex at any of the various stages of the pipeline
//
typedef struct EpAAVertex
{
	unsigned int  Header;
	float         X, Y, Z, W;
	float         R, G, B, A;
} EpAAVertex;

extern void  Ep_AASetupLine(EpImage* PImage, EpAAVertex* PV1, EpAAVertex* PV2);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif


#endif // _EImageDraw_h
