//================================================================================
// lib2D image library
//
// Image header file
//
// AUTHOR:  Gabor Nagy
// DATE:    2019-Jun-23 19:07:30
//
// Copyright (C) 1994-2021 by Gabor Nagy. All rights reserved.
//================================================================================
#ifndef _EImage_h
#define _EImage_h

#include <stdio.h>

#ifndef _EClass_h
#include <EClass.h>
#endif

#include <ELog.h>

#include <Color/Color.h>


#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
extern "C" {
#endif

typedef uint32_t  EpChannelFlags;


#define EpROTATE_RIGHT    0x0001    // By 90 degrees
#define EpROTATE_LEFT     0x0002    // By 90 degrees
#define EpFLIP_H          0x0004
#define EpFLIP_V          0x0008

typedef struct
{
	int  Type;             // Pixel semantic: EpGRAY, EpRGBA, EpSPECTRAL etc.
	int  NComponents;      // Number of components per pixel. E.g. 4 for RGBA
	int  ComponentType;    // EResUINT8, EResUINT16, etc.
} EpPixelFormat;

static inline void EpPixelFormat_Set(EpPixelFormat* PFormat,  const uint16_t PType, const uint16_t PNComponents, const uint16_t PComponentType)
{
	PFormat->Type = PType;
	PFormat->NComponents = PNComponents;
	PFormat->ComponentType = PComponentType;
}

typedef float     EpCoordinate;

typedef struct
{
	EpCoordinate  X, Y;
} EpCoord2;

typedef struct
{
	EpCoordinate  X, Y, Z;
} EpCoord3;

typedef struct
{
	float  V, W;
} EpWeightedValue;

typedef struct
{
	EIndex_2  Position, Size;
} EpRegion;

typedef struct
{
	EIndex_2      Position;
	EcRGBAiColor  Color;
} EpColoredPixelRGBAi;

#define Ep_AddPointRGBAi(mPoints, mCounter,  mX, mY,  mColor) do\
{\
	EpColoredPixelRGBAi*  lPoint = (mPoints) + mCounter;\
	lPoint->Position.X = mX;\
	lPoint->Position.Y = mY;\
	lPoint->Color = mColor;\
	mCounter++;\
} while(0)

// Channel masks
//
#define	EpRED        0x0001
#define	EpGREEN      0x0002
#define	EpBLUE       0x0004
#define	EpALPHA      0x0008
#define	EpINTENSITY  0x0010
#define	EpZ          0x0020

#define	EpR          EpRED
#define	EpG          EpGREEN
#define	EpB          EpBLUE
#define	EpA          EpALPHA

#define EpChRGBA  (0x0F)
#define EpChALL   (0x3F)

#define EpAUTO  -1

typedef enum
{
	EpTOP_TO_BOTTOM = 1,
	EpBOTTOM_TO_TOP
} EpOrientation;


#define EpRGBf_MAX  1000000.0


typedef float  (*EpConvertComponentProc)(const float PComponent, const void* PStruct);


// Operations for Images with alpha channels
// 0: overwrite (no blending)
enum
{
	EpSRCxALPHA_plus_DSTxONE_MINUS_ALPHA = 1,  // Traditional alpha-blending (src * A + dst * (1 - A))
	EpSRC_plus_DSTxONE_MINUS_ALPHA,            // Alpha-blending with premultiplied-alpha source image (src + dst * (1 - A)). This makes interpolating RGBA colors more correct
	EpSRC_plus_DSTxALPHA,                      // Alpha-blending with premultiplied-alpha source image (src + dst * A). This makes interpolating RGBA colors more correct, and it makes blending with the target easier when A > 1. For realistic shadow casting and relighting the background.
	EpFOR_COMPOSITABLE_IMAGE,                  // When used with EpACCUMULATEOPACITY, it produces an intermediate Image that, can be composited on another Image with EpSRCxALPHA_plus_DSTxONE_MINUS_ALPHA, resulting in the proper blending, as if the background was composited first, then the decal on top.
	EpFOR_COMPOSITABLE_IMAGE_PREMULTALPHA,     // When used with EpACCUMULATEOPACITY, it produces an intermediate Image that, can be composited on another Image with EpSRC_plus_DSTxONE_MINUS_ALPHA, resulting in the proper blending, as if the background was composited first, then the decal on top.
	EpSRC                                      // Simply overwrite the destination
};

enum
{
	EpTO_BLACK_PLUS_ALPHA = 1,
	EpTO_WHITE_PLUS_ALPHA
};


// PAlphaSource values for EpImage_Composite
enum
{
	EpFROM_SOURCE0 = 1,
	EpFROM_SOURCE1,
	EpADD,               // Simply add up alpha values
	EpACCUMULATEOPACITY  // Physically accurate way: multiply together transparency values (1-A)
};


typedef struct
{
	EcRGBf32Color           WhiteBalance;float  ColorSaturation;
	float                   Bias, Gain, Gamma;
	EBool                   AdaptiveExposure;  // Non-linear tone-mapping: out = exp(-input);
	EBool                   FastExposure;

	EpConvertComponentProc  ToneMapR, ToneMapG, ToneMapB;

	float                   InvGamma;
} EpExposure;


typedef struct
{
	EpConvertComponentProc  ConvertComponentProc;
	float*                  ResponseTablef32;  // For fast tone-mapping
	uint16_t*               ResponseTable16;
	uint8_t*                ResponseTable8;
	float                   Min, Mult;
	int                     Type;
} EpToneMapperChannel;


typedef struct
{
	EpToneMapperChannel*  Channels;
	EIndex                NumOfChannels;

	float                 ColorSaturation;

	void*                 Struct;
	int                   TableSize;  // Look-up table resolution for fast tone-mapping

	int                   MaxVal16;   // Maximum input value in 16-bit images (e.g. a camera might output 10-bit RAW, in 16 bit values, where only the lowest 10 bits in a 16-bit pixe component are used)
	EBool                 Uptodate;

	pthread_mutex_t       Update_mutex;
} EpToneMapper;



#define EpZERO_BLANKED  1

#define EpMAX_COLORNUM  4096



typedef uint8_t  EpGray8Pixel;

typedef struct
{
	uint8_t  R, G;
} EpRG8Pixel;


typedef struct
{
	uint8_t  I, A;  // Intensity + Alpha
} EpGrayA8Pixel;


typedef uint16_t  EpGray16Pixel;

typedef struct
{
	unsigned short  I, A;  // Intensity + Alpha
} EpGrayA16Pixel;


typedef float  EpGrayf32Pixel;

typedef struct
{
	float  I, A;  // Intensity + Alpha
} EpGrayAf32Pixel;

typedef struct
{
	float  R, G;
} EpRGf32Pixel;


typedef EcRGBAf32Color   EpRGBAf32Pixel;
typedef EcRGBA16Color    EpRGBA16Pixel;

typedef EcRGBf32Color    EpRGBf32Pixel;
typedef EcRGB16Color     EpRGB16Pixel;

typedef EcRGBA8Color     EpRGBA8Pixel;
typedef EcRGBA8Color     EpRGBX8Pixel;
typedef EcRGB8Color      EpRGB8Pixel;

typedef uint8_t          EpColorIndex8Pixel;
typedef uint16_t         EpColorIndex16Pixel;
typedef uint32_t         EpPRGBA8Pixel;


//------------------------------------------------
// Dithering algorithms
//------------------------------------------------
enum
{
	EpDITHER_NONE = 0,
	EpDITHER_BAYER44,
	EpDITHER_BAYER88,
	EpDITHER_MATDASH,
	EpDITHER_LINES,
	EpDITHER_HALFTONE,
	EpDITHER_FLOYD_STEINBERG,
	EpDITHER_FAX
};


// Pixel semantics
//
enum
{
	EpGRAY = 1,
	EpGRAYA,
	EpCOLORINDEX,
	EpRG,
	EpRGB,
	EpRGBA,
	EpSPECTRAL,
	EpDEPTH,

	EpPixelTypeALL = 4096
};


//------------------------------------------------------------------------------------------------------------------------
// Pixel format IDs.
// These uniquely identify some common pixel formats, in a more compact way than the fully generic values in EpPixelBuffer.
// Used for capturing images (e.g. from cameras), transferring images trough a network etc.
//------------------------------------------------------------------------------------------------------------------------
enum
{
	EpPixelFormatGRAY8 = 1,
	EpPixelFormatGRAY16,
	EpPixelFormatGRAYf32,
	EpPixelFormatGRAYA8,
	EpPixelFormatGRAYA16,
	EpPixelFormatGRAYAf32,
	EpPixelFormatCOLORINDEX8,
	EpPixelFormatCOLORINDEX16,
	EpPixelFormatRG8,
	EpPixelFormatRG16,
	EpPixelFormatRGf32,
	EpPixelFormatRGB8,
	EpPixelFormatRGB16,
	EpPixelFormatRGBf32,
	EpPixelFormatRGBA8,
	EpPixelFormatRGBA16,
	EpPixelFormatRGBAf32,
	EpPixelFormatRAW_G,            // Just get the green channels from the RAW image, without de-bayering. Camera calibration / response curve acquisition, will be more accurate when not affected by demosaicing-artifacts.
	EpPixelFormatRAW8,             // 8-bit Bayer (unspecified filter layout. e.g. the app requesting doesn't know).
	EpPixelFormatBGGR8,            // 8-bit Bayer
	EpPixelFormatGBRG8,            // 8-bit Bayer
	EpPixelFormatGRBG8,            // 8-bit Bayer
	EpPixelFormatRGGB8,            // 8-bit Bayer
	EpPixelFormatRAW10,            // 10-bit Bayer packed into uint16s (unspecified filter layout. e.g. the app requesting doesn't know)
	EpPixelFormatRAW10_PACK160,    // 10-bit Bayer Ximea 16 pixels x 10 bits = 160 bits are grouped in 20 bytes (60% more bandwidth efficient)  https://www.ximea.com/support/wiki/allprod/Transport_Data_Packing
	EpPixelFormatBGGR10,           // 10-bit Bayer
	EpPixelFormatGBRG10,           // 10-bit Bayer
	EpPixelFormatGRBG10,           // 10-bit Bayer
	EpPixelFormatRGGB10,           // 10-bit Bayer
	EpPixelFormatRAW12,            // 12-bit Bayer packed into uint16s (unspecified filter layout. e.g. the app requesting doesn't know)
	EpPixelFormatBGGR12,           // 12-bit Bayer
	EpPixelFormatGBRG12,           // 12-bit Bayer
	EpPixelFormatGRBG12,           // 12-bit Bayer
	EpPixelFormatRGGB12,           // 12-bit Bayer
	EpPixelFormatRAW16,            // 16-bit Bayer packed into uint16s (unspecified filter layout. e.g. the app requesting doesn't know)
	EpPixelFormatBGR8,
	EpPixelFormatRGBX8,
	EpPixelFormatYUYV8,
	EpPixelFormatMJPEG,
	EpPixelFormatH264,
	EpPixelFormatZ_RGB_5,          // Z-pixmap type, 5 bits/channel:     XBBBBBGGGGGRRRRR
	EpPixelFormatZ_RGB_565,        // Z-pixmap type, 5-6-5 bits/channel: BBBBBGGGGGGRRRRR
	EpPixelFormatZ_RGB_4,          // Z-pixmap type, 4 bits/channel:     XXXXBBBBGGGGRRRR
	EpPixelFormatZ_RGB_4_D,        // Z-pixmap type, 4 bits/channel:     XXXXBBBBGGGGRRRR dithered
	EpPixelFormatZ_RGB_332,        // Z-pixmap type, 3-3-2 bits/channel: BBGGGRRR
	EpPixelFormatZ_Red_8,          // Z-pixmap type, 8 bits/pixel: Red channel only
	EpPixelFormatZ_Green_8,        // Z-pixmap type, 8 bits/pixel: Green channel only
	EpPixelFormatZ_Blue_8,         // Z-pixmap type, 8 bits/pixel: Blue channel only
	EpPixelFormatZ_Alpha_8,        // Z-pixmap type, 8 bits/pixel: Alpha channel only
	EpPixelFormatZ_ColorIndex,     // Z-pixmap type, Depth bits/pixel: Colorindex
	EpPixelFormatP_RGBA,           // Separate bitplanes, Depth/4 bits/pixel: Red, Green, Blue, Alpha
	EpPixelFormatP_RGB,            // Separate bitplanes, Depth/3 bits/pixel: Red, Green, Blue
	EpPixelFormatP_Gray,           // Separate bitplanes, Depth bits/pixel: Grayscale
	EpPixelFormatP_ColorIndex,     // Separate bitplanes, Depth bits/pixel: Colorindex
	EpPixelFormatZ_JPEG_RGB_8,     // Z-pixmap type, 8 bits/channel: Red, Green, Blue compressed with JPEG
	EpPixelFormatYCbCr420
};


#define EpPixelFormatCOUNT     EpPixelFormatYCbCr420

#define	EpN_ImgHeaderSize  268



#define EpPALETTEFORMAT_ARGB16  0x01  // 16 bits (4 bits per channel): Alpha, Blue, Green, Red
#define EpPALETTEFORMAT_ARGB32  0x02  // 32 bits (8 bits per channel): Alpha, Blue, Green, Red
#define EpPALETTEFORMAT_ARGB64  0x04  // 64 bits (16 bits per channel): Alpha, Blue, Green, Red
#define EpPALETTEFORMAT_RGBA32  0x08  // 32 bits (8 bits per channel): Red, Green, Blue, Alpha


#define EpM_GLVertex2i(v)  glVertex2i((v)[0], (v)[1])


#define EpGLPIXELFORMAT  GL_RGBA
#define EpGLPIXELTYPE    GL_UNSIGNED_BYTE



//----------------------------------------------------------------
// Image structure for inlining into C/C++ source
//----------------------------------------------------------------
#define _EpCImage  1
typedef struct
{
	int       SizeX, SizeY,
            ComponentsPerPixel, ComponentType;
	uint8_t*  ImageBuffer_u8;
} EpCImage;

typedef EpCImage*              EpCImagePtr;
typedef struct EpImageClass_S  EpImageClass;
typedef struct EpImage_S       EpImage;
typedef EpImage*               EpImagePtr;
EListA_t(EpImagePtr);


#define EpImageClassCore\
	EBaseClassCore(EpImage)

struct EpImageClass_S
{
	EpImageClassCore;
};



typedef struct EpImageComponent_S  EpImageComponent;

typedef int  (*EpImageEditFunc)(EpImage* PImage, EpImageComponent* PComponent, int);
typedef int  (*EpImageFunc)(EpImage* PImage, EpImageComponent* PComponent);
typedef int  (*EpImageCopyFunc)(EpImage* PSrcImage, EpImageComponent* PSrcComponent, EpImage* PDstImage, EpImageComponent* PDstComponent);
typedef int  (*EpImageChangeFunc)(EpImage* PImage, EpImageComponent* PComponent, const EChangeFlags PChangeFlags);


typedef struct EpImageComponentClass_S  EpImageComponentClass;

struct EpImageComponentClass_S
{
	ECoreClassCore(EpImageComponent);

	EpImageFunc        Init;    // Constructor
	EpImageCopyFunc    Copy;    // Copy-constructor
	EpImageFunc        Free;    // Cestructor
	EpImageEditFunc    Edit;
	EpImageChangeFunc  Change;
};


#define EpImageComponentCore\
	EpImageComponentClass*  Class


// General ImageComponent structure (these are the first fields common in all ImageComponent structures)
//
struct EpImageComponent_S
{
	EpImageComponentCore;
};

#define EpPxbPIXELDATA_STATIC        0x01  // The pixel data is statically allocated. Don't free (e.g. inlined Images)
#define EpPxbNAME_STATIC             0x02  // The name data is statically allocated. Don't free (e.g. built-in names)
#define EpPxbCOMPONENT_NAMES_STATIC  0x04  // The name data is statically allocated. Don't free (e.g. built-in names)

typedef struct
{
	char*           Name;

	char**          ComponentNames;

	EpPixelFormat   PixelFormat;
	int             BytesPerPixel;
	uint16_t        Flags;             // EpPxbPIXELDATA_STATIC, EpPxbNAME_STATIC
	union
	{
		int8_t*    i8;
		uint8_t*   u8;
		int16_t*   i16;
		uint16_t*  u16;
		float*     f32;
	};

// This counter allows for identifying which derived PixelBuffers need updating. E.g. if an RGBAf32 PixelBuffer
// has a higher ChangeCount than an RGBA8 PixelBuffer, it means that tonemapping needs to be (re)done.
//
	EChange             Change;
} EpPixelBuffer;



typedef enum
{
	EpRLE = 1
} EpCompression;


#define EpImageCore\
	EObjectCore(EpImage);\
	EIndex            ID;                  /* Unique ID of this Image. This ensures that we can identify Images clearly, even if it we free Images and a freed pointer gets reused for a new allocation.  */\
\
	EIndex_2          Size;\
\
	char              LineAlignment,\
	                  PaletteFormat;\
\
	int               QFactor;             /* Quality-factor for lossy compression algorithms like JPEG  */\
	EBool             Compress;            /* Compression in destination file  */\
\
	EList(, EpPixelBuffer*, PixelBuffers);\
\
	EcRGBAiColor*     RGBAiColors;         /* Color palette for color-indexed images           */\
	int               NumOfColors;\
	int               NumOfValidPixels;    /* For read, if a file is not a complete image      */\
\
	EList(, EpImageComponent*, Components);    /* Runtime-typed ObjectComponents  */\
	EIndex            ChangeCount;\
\
	char*             FileName;\
	char*             Comment;\
	ETime             Loaded_ms            /* Last time the image was loaded (e.g. from a file). UNIX time, in ms. -1 : was not loaded */


//------------------------------------------------------------------------
// Base image for file-IO and image manipulating functions
//------------------------------------------------------------------------
struct EpImage_S
{
	EpImageCore;
};


typedef struct
{
	EIndex    OriginalImageID;
	EListA(, EpImage*, Images);
} EpScaledImage;


//----------------------------------------
// Special Image file format IDs
//----------------------------------------
#define EpFileFormatNONE            NULL
#define EpFileFormatFROM_EXTENSION  NULL  // Get format from filename extension (saving)
#define EpFileFormatAUTOMATIC       NULL  // Automatic format recognition (reading)


typedef struct EpImageFile_S      EpImageFile;
typedef struct EpImageIOModule_S  EpImageIOModule;

typedef int  (*EpLoadProc)(EpImageFile* PImageFile, EpImage* PImage);
typedef int  (*EpSaveProc)(EpImageFile* PImageFile, EpImage* PImage, const EpToneMapper* PToneMapper);
typedef int  (*EpIOOpenProc)(EpImageFile* PImageFile, EpImage* PImage);
typedef int  (*EpIOCloseProc)(EpImageFile* PImageFile);


#define EpImageIOModuleCore\
	EIOModuleCore;\
	unsigned int   StructSize;\
	EResource*     Resources;\
	void           (*Init)(EpImageIOModule*);\
	void           (*Free)(EpImageIOModule*);\
	EpIOOpenProc   Open;                /* Function to open an image file and read its header, or close it  */\
	EBool          NeedOpenFile;        /* Need an open FILE* descriptor in LImageFile for Open()     */\
	EpIOCloseProc  Close;               /* Function to close image file and free allocated resources  */\
	EpLoadProc     Load;                /* Function to read an image file                             */\
	EpSaveProc     Save;                /* Function to write an image to a file                       */\
	void           (*GUIForLoad)(void* PIOModule, void* PW, const uint16_t PAction);\
	void           (*GUIForSave)(void* PIOModule, void* PW, const uint16_t PAction)


struct EpImageIOModule_S
{
	EpImageIOModuleCore;
};


// Flags for EpImageFile
//
#define EpIOFlagKEEP_DATA_FOR_LOSSLESS    0x0001
#define EpIOFlagKEEP_DATA_FOR_LOSSY       0x0002
#define EpIOFlagNO_COPY                   0x0004
#define EpIOFlagCHANNELS_TO_PIXELBUFFERS  0x0008  // Load multiple channels (e.g. from a multi-channel EXR) to separate PixelBuffers

typedef struct
{
	char*     Name;
	uint32_t  Target;  // EpRED, EpGREEN etc.
} EpChannel;

struct EpImageFile_S
{
	char*             Name;  // Name of the file
	EpImageIOModule*  Format;
	void*             Handle;
	FILE*             File;   // If not NULL, an open file. Otherwise check MemoryBuffer or Name

	uint16_t          Flags;  // EpIOFlagKEEP_DATA_FOR_LOSSLESS, etc.
	EBool             UseMemoryBuffer;
	uint8_t*          MemoryBuffer;EIndex  MemoryBufferSize;

// If not empty, it specifies where to load channels from a multi-channel file (e.g. EXR).
// If the EpIOFlagCHANNELS_TO_PIXELBUFFERS is set, the channels (other than R,G,B,A) will be loaded into separate PixelBuffers
//
	EList(, EpChannel, ChannelTargets);

	int               IOStatus;
	EIndex_2          RegionStart, RegionSize;
	EIndex_2          DstPosition;
	int               Verbosity;  // Verbosity level for debugging etc.

	EIndex            NumOfFrames;

	EIndex            RequestedFrame;  // For Read() methods in formats that support multi-frame files (e.g. multi-page TIFF)
};




//
// This ImageComponent is used for preserving the original data for lossy file formats, so we don't have to re-encode them
//
typedef struct
{
	EpImageComponentCore;

	EpImageIOModule*  IOModule;  // The IOModule that was used to load the image (e.g. JPEG)
	uint8_t*          Data;EIndex  Data_Bytes;
}	EpImageFileInfo;


//
// This ImageComponent is used for storing sub-images on Images, such as MIP levels
//
typedef struct
{
	EpImageComponentCore;

	EList(, EpImage*, SubImages);
	EIndex  CurrentSubImageIndex;
}	EpSubImagesInfo;


typedef struct
{
	EcRGBAf32Color  StartColor, EndColor;
} EpBand;


// Plugins
//
#define	EpPLUGIN_ALLOC_STEP  8


// Scaling filter types
//
enum
{
	EpFilterNONE = 0,
	EpFilterIMPULSE,
	EpFilterBOX,
	EpFilterTRIANGLE,
	EpFilterQUADRATIC,
	EpFilterMITCHELL,
	EpFilterGAUSSIAN,
	EpFilterSOLID_CENTER_WITH_GAUSSIAN_FALLOFF,
	EpFilterQUICK      // Quick 25%, 1/3 and 50% scale with simple pixel averaging (good for anti-aliasing)
};


// Image intensity metering modes
//
enum
{
	EpMeterAVERAGE,
	EpMeter16PIXELMAX,
	EpMeterMIDDLE_SPOT,
	EpMeter4SPOT,
	EpMeter5SPOT
};


// What changed in an input of a Tool, for undo/redo etc.
//
#define EpCHG_POSITION  0x0000001
#define EpCHG_SIZE      0x0000002
#define	EpCHG_PIXELS    0x0000004

#define	EpCHG_SUBIMAGE  0x2000000  // If only a part of the image changed, this is OR-ed to the flags

#define EpCHG_ALL       0xFFFFFFFF



//--------------------------------
// Macros
//--------------------------------


//================================================================================================
// Create an EpImage from a "CStruct" image (inlined image: mostly used for icons)
// There are no checks, so be careful with 3-component (RGB) etc. images!
//================================================================================================
#define EpM_CImageToRGBA8Image_sp(mCImage, pImage) do\
{\
	if(pImage == NULL)\
	{\
		E_NewAssign(EpImage, pImage);\
		(pImage)->Size.X = (mCImage).SizeX;(pImage)->Size.Y = (mCImage).SizeY;\
		EpImage_SetPixels(pImage,  EpRGBA, 0, EResUINT8,  (mCImage).ImageBuffer_u8, TRUE);\
	}\
/*\
	else\
	{\
		printf("%s() %s:%d %s is not NULL\n", __FUNCTION__, __FILE__, __LINE__, #pImage);fflush(stdout);\
	}\
*/\
} while(0)


#define EpM_CImageToRGBA8Image_pp(mCImage, pImage) do\
{\
	if(pImage == NULL)\
	{\
		E_NewAssign(EpImage, pImage);\
		(pImage)->Size.X = (mCImage)->SizeX;(pImage)->Size.Y = (mCImage)->SizeY;\
		EpImage_SetPixels(pImage,  EpRGBA, 0, EResUINT8,  (mCImage)->ImageBuffer_u8, TRUE);\
	}\
/*\
	else\
	{\
		printf("%s() %s:%d %s is not NULL\n", __FUNCTION__, __FILE__, __LINE__, #pImage);fflush(stdout);\
	}\
*/\
} while(0)


//================================================================================================
// Create an EpImage from a "CStruct" image (inlined image: mostly used for icons)
// There are no checks, so be careful with 3-component (RGB) etc. images!
//================================================================================================
#define EpM_CImageToRGBA8Image_ps(mCImage, pImage) do\
{\
	EpImage_Init(&pImage);\
	(pImage).Size.X = (mCImage)->SizeX;(pImage).Size.Y = (mCImage)->SizeY;\
	EpImage_SetPixels(&(pImage), EpRGBA, 0, EResUINT8,  (mCImage)->ImageBuffer_u8, TRUE);\
} while(0)




#define EpM_CreateIconStates1_sp(mName) do\
{\
	EpM_CImageToRGBA8Image_sp(Image_##mName,         _Image_##mName);        EM_ObjectDebugNameM_p(_Image_##mName);\
} while(0)


#define EpM_CreateIconStates2_sp(mName) do\
{\
	EpM_CImageToRGBA8Image_sp(Image_##mName,          _Image_##mName);         EM_ObjectDebugNameM_p(_Image_##mName);\
	EpM_CImageToRGBA8Image_sp(Image_##mName##_Active, _Image_##mName##_Active);EM_ObjectDebugNameM_p(_Image_##mName##_Active);\
} while(0)


#define EpM_CreateIconStates3_sp(mName) do\
{\
	EpM_CImageToRGBA8Image_sp(Image_##mName,          _Image_##mName);         EM_ObjectDebugNameM_p(_Image_##mName);\
	EpM_CImageToRGBA8Image_sp(Image_##mName##_Active, _Image_##mName##_Active);EM_ObjectDebugNameM_p(_Image_##mName##_Active);\
	EpM_CImageToRGBA8Image_sp(Image_##mName##_Armed,  _Image_##mName##_Armed); EM_ObjectDebugNameM_p(_Image_##mName##_Armed);\
} while(0)


#define EpM_CreateIconStates3_pp(mName) do\
{\
	EpM_CImageToRGBA8Image_pp(Image_##mName,          _Image_##mName);         EM_ObjectDebugNameM_p(_Image_##mName);\
	EpM_CImageToRGBA8Image_pp(Image_##mName##_Active, _Image_##mName##_Active);EM_ObjectDebugNameM_p(_Image_##mName##_Active);\
	EpM_CImageToRGBA8Image_pp(Image_##mName##_Armed,  _Image_##mName##_Armed); EM_ObjectDebugNameM_p(_Image_##mName##_Armed);\
} while(0)


#define EpM_CreateIconStates5_sp(mName) do\
{\
	EpM_CImageToRGBA8Image_sp(Image_##mName,            _Image_##mName);           EM_ObjectDebugNameM_p(_Image_##mName);\
	EpM_CImageToRGBA8Image_sp(Image_##mName##_Active,   _Image_##mName##_Active);  EM_ObjectDebugNameM_p(_Image_##mName##_Active);\
	EpM_CImageToRGBA8Image_sp(Image_##mName##_Armed,    _Image_##mName##_Armed);   EM_ObjectDebugNameM_p(_Image_##mName##_Armed);\
	EpM_CImageToRGBA8Image_sp(Image_##mName##_Running,  _Image_##mName##_Running); EM_ObjectDebugNameM_p(_Image_##mName##_Running);\
	EpM_CImageToRGBA8Image_sp(Image_##mName##_Big,      _Image_##mName##_Big);     EM_ObjectDebugNameM_p(_Image_##mName##_Big);\
} while(0)


//
// See also
//  EpImage_New
//
#define EpImage_NewAssign(pSizeX, pSizeY,  pPixelType, pComponentsPerPixel, pPixelComponentType,  pClear,  pDst) do\
{\
	EpImage*  lImage;ENew(lImage);if(lImage)\
	{\
		EM_ObjectInit(EpImage, lImage, EpImage_class);\
\
		if(pSizeX > 0 && pSizeY > 0)\
		{\
			lImage->Size.X = pSizeX;\
			lImage->Size.Y = pSizeY;\
\
			if(pPixelType && EpImage_PixelBufferAddA(lImage,  pPixelType, pComponentsPerPixel, pPixelComponentType,  EpALLOC_PIXELS|(pClear ? EpCLEAR : 0)) != E_SUCCESS)\
			{\
				E_Delete(lImage);\
			}\
		}\
	}\
	E_Assign(lImage, pDst);\
} while(0)


//================================================================================================
// Image reader template function
//================================================================================================
#define EpM_ImageLoad(pImageFile, pSrcComponentType, pSrcNumOfComponents, pScanlineReadProc, pPixelReadProc,  pPixelType, pPixelComponents, pPixelComponentType,  pPixelCType,  pImage) do\
{\
	if(pImageFile->RegionSize.X)\
	{\
		pPixelCType*  lTargetRow;\
		const EIndex  lRegionSizeX = pImageFile->RegionSize.X, lRegionSizeY = pImageFile->RegionSize.Y;\
\
		LSrcY = pImageFile->RegionStart.Y;\
\
		EpImage_PixelBufferAddA(pImage,  pPixelType, pPixelComponents, pPixelComponentType,  EpCHECK_IF_EXISTS|EpDELETE_OTHERS);\
		EpImage_Resize(pImage,  lRegionSizeX, lRegionSizeY,  FALSE);\
		EpPixelBuffer*  lDstPixelBuffer = EpImage_PixelBufferGetByFormat(pImage,  pPixelType, pPixelComponents, pPixelComponentType,  FALSE,  NULL);if(lDstPixelBuffer)\
		{\
			pPixelCType*  lDstPixels = (pPixelCType*)lDstPixelBuffer->u8;\
			lTargetRow = lDstPixels;\
			EIndex  lY = 0;for(;lY < lRegionSizeY;LSrcY++, lY++, lTargetRow += lRegionSizeX)\
			{\
				pScanlineReadProc(lY, pSrcComponentType, pSrcNumOfComponents, pPixelReadProc, pPixelCType, lRegionSizeX);\
				M_ProgressUpdate(lY,  LDstEndY, LProgressMask);\
			}\
			pImage->NumOfValidPixels = lRegionSizeX * lRegionSizeY;\
			EChange_Bump(&lDstPixelBuffer->Change);\
			LResult = E_SUCCESS;\
		}\
		else LResult = E_NO_MEMORY;\
	}\
	else\
	{\
		EpImage_PixelBufferAddA(pImage,  pPixelType, pPixelComponents, pPixelComponentType,  EpCHECK_IF_EXISTS|EpDELETE_OTHERS);\
		EpImage_Resize(pImage,  LSizeX, LSizeY,  FALSE);\
		EpPixelBuffer*  lDstPixelBuffer = EpImage_PixelBufferGetByFormat(pImage, pPixelType, pPixelComponents, pPixelComponentType,  FALSE,  NULL);if(lDstPixelBuffer)\
		{\
			pPixelCType*  lDstPixels = (pPixelCType*)lDstPixelBuffer->u8;\
			EIndex  lY = 0;LSrcY = 0;for(;lY < LDstEndY;lY++, LSrcY += LYDir)\
			{\
				pScanlineReadProc(lY, pSrcComponentType, pSrcNumOfComponents, pPixelReadProc, pPixelCType, LSizeX);\
				M_ProgressUpdate(lY, LDstEndY,  LProgressMask);\
			}\
			pImage->NumOfValidPixels = LSizeX * LSizeY;\
			EChange_Bump(&lDstPixelBuffer->Change);\
			LResult = E_SUCCESS;\
		}\
		else LResult = E_NO_MEMORY;\
	}\
} while(0)


#define Ep_DoAllFormats(mProcedure) do\
{\
	if(LSizeX > 0 && LSizeY > 0)\
	{\
		mProcedure(EpGray8Pixel,        EpGRAY,  0, EResUINT8);\
		mProcedure(EpGrayA8Pixel,       EpGRAYA, 0, EResUINT8);\
\
		mProcedure(EpGray16Pixel,       EpGRAY,  0, EResUINT16);\
		mProcedure(EpGrayA16Pixel,      EpGRAYA, 0, EResUINT16);\
\
		mProcedure(EpGrayf32Pixel,      EpGRAY,  0, EResFLOAT32);\
		mProcedure(EpGrayAf32Pixel,     EpGRAYA, 0, EResFLOAT32);\
\
		mProcedure(EpColorIndex8Pixel,  EpCOLORINDEX, 0, EResUINT8);\
		mProcedure(EpColorIndex16Pixel, EpCOLORINDEX, 0, EResUINT16);\
\
		mProcedure(EpRGBA8Pixel,        EpRGB,  4, EResUINT8);\
		mProcedure(EpRGBA8Pixel,        EpRGBA, 0, EResUINT8);\
\
		mProcedure(EpRGB16Pixel,        EpRGB,  0, EResUINT16);\
		mProcedure(EpRGBA16Pixel,       EpRGBA, 0, EResUINT16);\
\
		mProcedure(EpRGBf32Pixel,       EpRGB,  0, EResFLOAT32);\
		mProcedure(EpRGBAf32Pixel,      EpRGBA, 0, EResFLOAT32);\
	}\
} while(0)



//
// Icon definition macros
//

// Neutral
//
#define EpIconDef(mQualifier, m)\
extern EpCImage    Image_##m;\
mQualifier EpImagePtr  _Image_##m = NULL;


// Neutral, Active
//
#define EpIconDef2(m)\
extern EpCImage    Image_##m,         Image_##m##_Active;\
static EpImagePtr  _Image_##m = NULL, _Image_##m##_Active = NULL;


// Neutral, Active, Armed
//
#define EpIconDef3(m)\
extern EpCImage    Image_##m,         Image_##m##_Active,         Image_##m##_Armed;\
static EpImagePtr  _Image_##m = NULL, _Image_##m##_Active = NULL, _Image_##m##_Armed = NULL;


// Neutral, Active, Armed
//
#define EpIconDef3_g(m)\
extern EpCImage   Image_##m,         Image_##m##_Active,         Image_##m##_Armed;\
EpImagePtr        _Image_##m = NULL, _Image_##m##_Active = NULL, _Image_##m##_Armed = NULL;


// Neutral, Active, Armed
//
#define EpIconDef3_p(m)\
extern EpCImagePtr  Image_##m,         Image_##m##_Active,         Image_##m##_Armed;\
static EpImagePtr   _Image_##m = NULL, _Image_##m##_Active = NULL, _Image_##m##_Armed = NULL;


// Neutral, Active, Armed, Running, Big
//
#define XeIconDef5(m)\
extern EpCImage    Image_##m,         Image_##m##_Active,         Image_##m##_Armed,         Image_##m##_Running,         Image_##m##_Big;\
static EpImagePtr  _Image_##m = NULL, _Image_##m##_Active = NULL, _Image_##m##_Armed = NULL, _Image_##m##_Running = NULL, _Image_##m##_Big = NULL;



#define EpIconFree(m) do\
{\
	E_DeleteChk(_Image_##m);\
} while(0)

#define EpIconFree2(m) do\
{\
	E_DeleteChk(_Image_##m);\
	E_DeleteChk(_Image_##m##_Active);\
} while(0)

#define EpIconFree3(m) do\
{\
	E_DeleteChk(_Image_##m);\
	E_DeleteChk(_Image_##m##_Active);\
	E_DeleteChk(_Image_##m##_Armed);\
} while(0)

#define EpIconFree5(m) do\
{\
	E_DeleteChk(_Image_##m);\
	E_DeleteChk(_Image_##m##_Active);\
	E_DeleteChk(_Image_##m##_Armed);\
	E_DeleteChk(_Image_##m##_Running);\
	E_DeleteChk(_Image_##m##_Big);\
} while(0)


#define EpM_ColorSaturationRGB8(pSPixel, pDPixel,  pSaturation) do\
{\
	int        lR = pSPixel->R, lG = pSPixel->G, lB = pSPixel->B;\
	const int  lBW = (LRIT * lR + LGIT * lG + LBIT * lB) >> 8;\
\
	lR = lBW + pSaturation * lR;lG = lBW + pSaturation * lG;lB = lBW + pSaturation * lB;\
	if(lR < 0) lR = 0;else if(lR > 255) lR = 255;\
	if(lG < 0) lG = 0;else if(lG > 255) lG = 255;\
	if(lB < 0) lB = 0;else if(lB > 255) lB = 255;\
	pDPixel->R = lR;\
	pDPixel->G = lG;\
	pDPixel->B = lB;\
} while(0);

#define EpM_ColorSaturationRGB16(pSPixel, pDPixel,  pSaturation) do\
{\
	float        lR = pSPixel->R, lG = pSPixel->G, lB = pSPixel->B;\
	const float  lBW = (LRT * lR + LGT * lG + LBT * lB) * EInv65536N;\
\
	lR = lBW + pSaturation * lR;lG = lBW + pSaturation * lG;lB = lBW + pSaturation * lB;\
	if(lR < 0) lR = 0;else if(lR > 65535) lR = 65535;\
	if(lG < 0) lG = 0;else if(lG > 65535) lG = 65535;\
	if(lB < 0) lB = 0;else if(lB > 65535) lB = 65535;\
	pDPixel->R = lR;\
	pDPixel->G = lG;\
	pDPixel->B = lB;\
} while(0);

#define EpM_ColorSaturationRGBf32(pSPixel, pDPixel,  pSaturation) do\
{\
	float        lR = pSPixel->R, lG = pSPixel->G, lB = pSPixel->B;\
	const float  lBW = (LRT * lR + LGT * lG + LBT * lB);\
\
	lR = lBW + pSaturation * lR;lG = lBW + pSaturation * lG;lB = lBW + pSaturation * lB;\
	if(lR < 0.0f) lR = 0.0f;\
	if(lG < 0.0f) lG = 0.0f;\
	if(lB < 0.0f) lB = 0.0f;\
	pDPixel->R = lR;\
	pDPixel->G = lG;\
	pDPixel->B = lB;\
} while(0);


//================================================
// externs
//================================================

//--------------------------------
// Image.c
//--------------------------------
extern EpImageClass            EpImage_class[];
extern EpImageComponentClass*  EpSubImagesInfo_class;
extern EpImage*                EpIcon_FileImage;  // Generic image file icon

extern const char*             EpStringMIPLevelNames[];

extern EpImageComponentClass*  EpGeoTIFFInfo_lass;
extern EpImageComponentClass*  EpEXIFInfo_Class;


extern EpCoordinate        Ep_SnapCoordinateTo(EpCoordinate PF, EpCoordinate PS);

extern void                EpImage_Init(EpImage* PImage);
extern EpImage*            EpImage_New(const EIndex PSizeX, const EIndex PSizeY,  const int PPixelType, const int PPixelComponents, const int PPixelComponentType,  const EBool PClear);
extern EpImage*            EpImage_NewC(EpImageClass* PClass, const EIndex PSizeX, const EIndex PSizeY,  const int PPixelType, const int PComponentsPerPixel, const int PPixelComponentType,  const EBool PClear);

extern void                EpImage_FreeContents(EpImage* PImage);
extern void                EpImage_CopyContents(EpImage* PImageS, EpImage* PImageD);

extern int                 EpImage_Copy(const EpImage* PSImage, EpImage* PDImage);

extern EpImageComponentClass*  EpImageComponent_ClassFindByName(char* PName);
extern EpImageComponentClass*  EpImageComponent_ClassRegister(EpImageComponentClass* PClassDescriptor);
extern void                    EpImageComponent_ClassDeactivate(EpImageComponentClass* PClass);
extern void                    EpImageComponent_ClassesFree(void);
extern int                     EpImageComponent_Remove(EpImage* PImage, EpImageComponent* PInfo);
extern int                     EpImageComponent_RemoveAll(EpImage* PImage);
extern EpImageComponent*       EpImageComponent_Add(EpImage* PImage, EpImageComponentClass* PClass);
extern EpImageComponent*       EpImageComponent_Copy(EpImage* PSImage, EpImageComponent* PSInfo, EpImage* PDImage);
extern EIndex                  EpImageComponent_CopyAll(EpImage* PSImage, EpImage* PDImage, const EBool PAllowMultiples);

extern EpImageComponent*       EpImageComponent_ByClass(const EpImage* PImage, EpImageComponentClass* PClass);

extern void                    EpToneMapper_Init(EpToneMapper* PToneMapper);
extern void                    EpToneMapper_FreeContents(EpToneMapper* PToneMapper);
extern void                    EpToneMapper_Clear(EpToneMapper* PToneMapper);

extern EBool                   EpToneMapper_UpdateRGBA(EpToneMapper* PToneMapper, const EpImage* PImage,
                                EpConvertComponentProc PConvertComponentProcR,
                                EpConvertComponentProc PConvertComponentProcG,
                                EpConvertComponentProc PConvertComponentProcB,
                                EpConvertComponentProc PConvertComponentProcA,
                                void* PStruct, const int PTableSize);


extern void             EpExposure_Init(EpExposure* PExposure);
extern void             EpExposure_Update(EpExposure* PExposure, EpToneMapper* PToneMapper, const EpImage* PImage);

extern float            Ep_ToneMapComponent(const float PComponent, const void* PExposure);
extern float            Ep_ToneMapR(const float PComponent, const void* PExposure);
extern float            Ep_ToneMapG(const float PComponent, const void* PExposure);
extern float            Ep_ToneMapB(const float PComponent, const void* PExposure);
extern float            Ep_ToneMapAdaptiveExposure(const float PComponent, const void* PExposure);
extern float            Ep_ToneMapAdaptiveExposureR(const float PComponent, const void* PExposure);
extern float            Ep_ToneMapAdaptiveExposureG(const float PComponent, const void* PExposure);
extern float            Ep_ToneMapAdaptiveExposureB(const float PComponent, const void* PExposure);

extern void             Ep_Init(void);
extern void             Ep_Free(void);

extern int              EpImage_Gain(const EpImage* PSImage, EpImage* PDImage, const float PGain);
extern EpImage*         EpImage_GetSubImage(EpImage* PImage);

extern int              Ep_MipLevels(const EIndex PSizeX, const EIndex PSizeY);


extern const char  Ep_R[], Ep_G[], Ep_B[], Ep_A[],  Ep_X[], Ep_Y[], Ep_Z[];

extern const char*  Ep_RGB[];
extern const char*  Ep_RGBA[];
extern const char*  Ep_XYZ[];


//--------------------------------
// Camera.c
//--------------------------------
extern int              EpCameraInterface_ClassesFree(void);


//--------------------------------
// Pixels.c
//--------------------------------
extern int                 Ep_Verbosity;

extern EBool               EpImage_HasContent(const EpImage* PSImage);
extern EBool               EpImage_ArePixelsSignedValues(const EpImage* PImage);

extern int                 EpImage_ComponentsPerPixel(const EpImage* PImage);

extern int                 Ep_PixelTypeToFormat(const int PPixelType, const int PComponentType);

extern int                 EpImage_SetPixels(EpImage* PImage, const int PPixelType, const int PComponentsPerPixel, const int PPixelComponentType,  void* PPixels, const EBool PStatic);


// Actions for EpImage_PixelBufferAddA
//
#define  EpCLEAR                0x0001    // Clear new PixelBuffer to black
#define  EpCHECK_IF_EXISTS      0x0002    // Check if a PixelBuffer with the given pixel format already exists
#define  EpALLOC_PIXELS         0x0004    // Check if a PixelBuffer with the given pixel format already exists
#define  EpDELETE_OTHERS        0x0008    // Delete PixelBuffers other than the once specified
#define  EpCREATE_FROM_RICHEST  0x0010    // Create new PixelBuffer from richest existing PixelBuffer

extern int             Ep_PixelFormatParamsToComponentsInfo(const int PPixelType, const int PComponentType,  int* PComponentsPerPixel, int* PBytesPerComponent);

extern EpPixelBuffer*  EpImage_PixelBufferAdd(EpImage* PImage);
extern int             EpImage_PixelBufferAddA(EpImage* PImage,  const int PPixelType, const int PPixelComponents, const int PPixelComponentType,  const uint16_t PFlags);
extern EpPixelBuffer*  EpImage_PixelBufferGetByFormat(EpImage* PImage,  const int PPixelType, const int PComponentsPerPixel, const int PComponentType,  const EBool PAdd,  int* PResult);
extern int             EpImage_PixelBufferDel(EpImage* PImage, EpPixelBuffer* PPixelBuffer);

extern void            EpImage_FreePixelBuffers(EpImage* PImage);

extern EpPixelBuffer*  EpImage_PixelBufferGetByTemplate(const EpImage* PImage, const EpPixelBuffer* PTemplate);
extern EpPixelBuffer*  EpImage_PixelBufferGetByComponenType(const EpImage* PImage, const int PType);
extern EpPixelBuffer*  EpImage_PixelBufferGetRichest(const EpImage* PImage);

extern char*           Ep_PixelTypeToString(const int PPixelType, char* PStr);
extern char*           Ep_PixelFormatToString(const int PPixelFormat, char* PStr);
extern int             Ep_PixelFormatPixelType(const int PPixelFormat);

extern char*           EpPixelBuffer_FormatToString(const EpPixelBuffer* PPixelBuffer, char* PStr);
extern void            EpPixelBuffer_Free(EpPixelBuffer* PPixelBuffer);

extern int             Ep_PixelTypeToRichness(const int PPixelType);

extern char*           EpPixelFormat_ToString(const EpPixelFormat* PFormat,  char* PStr);

extern char*           EpImage_PixelFormatsToString(const EpImage* PImage, char* PStr);

extern void            Ep_RGB8ToRGBX8(const EpRGB8Pixel*  PSrcImage, EpRGBA8Pixel* PDstImage, const EIndex PNumOfPixels);
extern void            Ep_RGBX8ToRGB8(const EpRGBA8Pixel* PSrcImage, EpRGB8Pixel*  PDstImage, const EIndex PNumOfPixels);

extern int             EpImage_HDRToSDR(EpImage* PSImage, EpImage* PDImage, const EpToneMapper* PToneMapper, const EBool PForce, const EBool PRemoveOrigBuffers);

extern int             EpImage_RGBAf32ToRGBA8(EpImage* PSImage, EpImage* PDImage,  const EpToneMapper* PToneMapper, const EBool PForce);
extern int             EpImage_RGBA16ToRGBA8(EpImage* PImage,   EpImage* PDImage,  const EpToneMapper* PToneMapper, const EBool PForce);
extern int             EpImage_RGBA8ToRGBf32(EpImage* PImage);

extern int             EpImage_RGBf32ToRGBX8(EpImage* PSImage, EpImage* PDImage,  const EpToneMapper* PToneMapper, const EBool PForce);
extern int             EpImage_RGB16ToRGBX8(EpImage* PSImage,  EpImage* PDImage,  const EpToneMapper* PToneMapper, const EBool PForce);

extern int             EpImage_RGB8ToRGBX8(EpImage* PImage, const EpRGB8Pixel* PRGB8Pixels);
extern EpRGB8Pixel*    EpImage_RGBX8ToRGB8(EpImage* PImage, EpRGB8Pixel* PRGB8Pixels);

extern int             EpImage_RGf32ToRG8(EpImage* PSImage, EpImage* PDImage, const EpToneMapper* PToneMapper, const EBool PForce);

extern void            EpImage_ColorIndex8ToRGBX8(EpImage* PImage);

extern int             EpImage_GrayA16ToGrayA8(EpImage* PSImage, EpImage* PDImage,  const EpToneMapper* PToneMapper, const EBool PForce);
extern int             EpImage_GrayAf32ToGrayA8(EpImage* PSImage,  EpImage* PDImage,  const EpToneMapper* PToneMapper, const EBool PForce);
extern int             EpImage_Grayf32ToGray8(EpImage* PSImage,  EpImage* PDImage,  const EpToneMapper* PToneMapper, const EBool PForce);
extern int             EpImage_Gray16ToGray8(EpImage* PSImage, EpImage* PDImage,  const EpToneMapper* PToneMapper, const EBool PForce);

extern EpImage*        EpImage_FromCImage(const EpCImage* PCImage);
extern EpImage*        EpImage_FromCImageC(const EpCImage* PCImage, EpImageClass* PClass);

extern int             EpImage_CopyPixelBuffers(const EpImage* PSImage, EpImage* PDImage,  const EBool PAllocatePixelData, const EBool PCopyPixelData);
extern int             EpImage_CopyPixels(const EpImage* PSImage, EpImage* PDImage);


extern void           EpImage_MoveContents(EpImage* PSImage, EpImage* PDImage);

extern int            EpImage_Resize(EpImage* PImage,  const EIndex PNewSizeX, const EIndex PNewSizeY,  const EIndex PClear);

extern int            EpImage_Clear(EpImage* PImage, EpImage* PResultImage, const EcRGBAiColor* PColor, const EpChannelFlags PChannels);

extern EBool          EpImage_HasIntensity(const EpImage* PImage);
extern EBool          EpImage_HasRed(const EpImage* PImage);
extern EBool          EpImage_HasGreen(const EpImage* PImage);
extern EBool          EpImage_HasBlue(const EpImage* PImage);
extern EBool          EpImage_HasAlpha(const EpImage* PImage);

extern EBool          EpImage_IsHDR(const EpImage* PImage);

extern int            EpImage_AlphaAdd(EpImage* PSImage, EpImage* PDImage, const int PAlpha);
extern void           EpImage_AlphaDel(EpImage* PImage);

extern int            EpImage_GetPixelColor(const EpImage* PImage,  const EIndex PX, const EIndex PY,  EcRGBAiColor* PColor);
extern int            EpImage_GetPixelColorF(const EpImage* PImage,  const EIndex PX, const EIndex PY,  EcRGBAf32Color* PColor);
extern int            EpImage_GetPixelColorAvg(EpImage* PSImage,  const EIndex PPX, const EIndex PPY,  const EIndex PSampeKernelSize, EcRGBAf32Color* PRGBAf32Color);
extern int            EpImage_SetPixelColorRGBAi(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EcRGBAiColor* PColor, const uint32_t PBlendMode, const EpChannelFlags PChannels);
extern int            EpImage_SetPixelColorRGBAf(EpImage* PImage,  const EIndex PX, const EIndex PY,  const EcRGBAf32Color* PColor, const uint32_t PBlendMode, const float PGamma, const EpChannelFlags PChannels);


extern int            EpImage_CreateRGBAf32(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);
extern int            EpImage_CreateRGBA16(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);
extern int            EpImage_CreateRGBA8(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);

extern int            EpImage_CreateRGBf32(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);
extern int            EpImage_CreateRGBf32Mult(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers, const float PMult);
extern int            EpImage_CreateRGB16(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);
extern int            EpImage_CreateRGBX8(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);

extern int            EpImage_CreateGrayAf32(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);
extern int            EpImage_CreateGrayA16(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);
extern int            EpImage_CreateGrayA8(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);

extern int            EpImage_CreateGrayf32Mult(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers, const float PMult);
extern int            EpImage_CreateGrayf32(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);
extern int            EpImage_CreateGray16(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);
extern int            EpImage_CreateGray8(const EpImage* PSImage, EpImage* PDImage, const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);

extern int            EpImage_CreatePixelFormat(const EpImage* PSImage, EpImage* PDImage,  const int PPixelType, const int PComponentsPerPixel, const int PComponentType,  const EBool PUpdate, const EBool PDeleteOriginalPixelBuffers);

extern void           Ep_RowConvertToFloats(const EpImage* PSImage, EpImage* POutImage, float* PRow, const EIndex PY);
extern void           Ep_RowConvertFromFloats(EpImage* POutImage, const float* PRow, const EIndex PY);

extern EIndex         EpImage_CorrectPixelValues(EpImage* PImage);

extern int            EpImage_Convolve(EpImage* PSImage, EpImage* PDImage, const float* PKernel, const EIndex PKernelSizeX, const EIndex PKernelSizeY);

extern int            EpImage_MultiplyColorByAlpha(EpImage* PImage);


//--------------------------------
// Scale.c
//--------------------------------
extern void      Ep_RGB8ImageScale50Percent(EpRGB8Pixel* PSrcImage, EpRGB8Pixel* PDstImage,  const EIndex PSizeX, const EIndex PSizeY);
extern void      EpImage_ScaleQuick50Percent(const EpImage* PSImage, EpImage* PDImage);

extern EBool     EpImage_Scale(const EpImage* PImageS, EpImage* PImageD,  const EIndex PNewSizeX, const EIndex PNewSizeY,  int PFilterType, double PBlurFactor, void (*LShowProgressProc)(float, void*), void* PShowProgressProcData);
extern EBool     EpImage_ScaleAlpha(EpImage* PImageS, EpImage* PImageD,  const EIndex PNewSizeX, const EIndex PNewSizeY,  const int PFilterType, double PBlurFactor, void (*LShowProgressProc)(float, void*), void* PShowProgressProcData);

extern void      Ep_ScaleFree(void);


//--------------------------------
// IO.c
//--------------------------------
EList(extern, EpImageIOModule*, EpImage_FileFormats);
extern ELogFunc     Ep_LogFunc;

extern EpImageComponentClass*  EpImageFileInfo_class;

extern int          EpImage_IOReadStatus;

extern char*        EpImage_IOPluginPath;

EList(extern, char*, Ep_Extensions);

extern void         (*EpProgressIndicator_SetFunc)(float PValue, void (*PAbortCallback)(EPointer), EPointer PClientData);

extern void         EpProgressIndicator_SetMessage(char* PMessage, ...);
extern void         EpProgressIndicator_Set(const float PValue, void (*PAbortCallback)(EPointer), EPointer PClientData);

extern const char*  Ep_GetBaseFileName(const char* PFileName, char* PNameRet, const int PMaxLen);

extern EpImageIOModule*  Ep_GetImageFormatByName(const char* PName);
extern EpImageIOModule*  Ep_GetImageFormatByExtension(const char* PExtension);
extern uint32_t*         Ep_ABGRToRGBA(uint32_t* PABGRImage,  const EIndex PSizeX, const EIndex PSizeY);

extern EpImageIOModule*  EpImage_IOModuleAdd(EpImageIOModule* PDescriptor);
extern void              EpImage_IOModuleRemove(EpImageIOModule* PFileFormat, const EBool PFree);
extern void              EpImage_IOModuleDeactivate(EpImageIOModule* PFileFormat);

extern EpImageFile*      EpImage_FileOpen(const char* PFileName, EpImageIOModule* PFileFormat, EpImageIOModule** PFileFormatRet, EpImage* PImageD);
extern void              EpImage_FileClose(EpImageFile* PImageFile);

extern void              EpImageFile_Init(EpImageFile* PImageFile);

extern int               EpImage_FileReadRegion(EpImageFile* PImageFile,  const EIndex PSrcX, const EIndex PSrcY,  const EIndex PSizeX, const EIndex PSizeY,  const EIndex PDstX, const EIndex PDstY,  EpImage* PImage);
extern int               EpImage_ReadFromFile(const char* PFileName, EpImageIOModule* PFileFormat, EpImageIOModule** PFileFormatRet, EpImage* PImageD);
extern int               EpImage_ReadFromFileAdv(const char* PFileName, EpImageIOModule* PFileFormat, EpImageIOModule** PFileFormatRet, EpImageFile* PImageFile, EpImage* PImageD);

extern int               EpImage_ReadFromOpenFile(EpImageFile* PImageFile, EpImage* PImage, const EIndex PFrame);

extern int               EpImage_ReadFromMemory(EpImageIOModule* PFileFormat,  const uint8_t* PBuffer, const EIndex PBufferSize,  const uint16_t PFlags,  EpImage* PImageD);

extern int               EpImage_ReadNumberedFile(const char* PBaseFileName, const int PFrameNumber, const char* PFileExtension, int* PDigitsPtr, EpImageIOModule* PFileFormat, EpImageIOModule** PFileFormatRet, EpImage* PImage);

extern int               EpImage_SaveToFile(EpImage* PImage, const EpToneMapper* PToneMapper,  const char* PFileName, char* PNameReturn, const int PNameMaxLen, EpImageIOModule* PFileFormat);
extern int               EpImage_SaveToOpenFile(EpImage* PImage, FILE* PFile, EpImageIOModule* PFileFormat);
extern int               EpImage_SaveToMemory(EpImage* PImage, const char* PFileName, EpImageIOModule* PFileFormat, const EpToneMapper* PToneMapper,  uint8_t** PBufferP, EIndex* PBufferSizeP);

extern void              EpImage_FileFree(EpImageFile* PImageFile);


typedef void  (*EpCacheAreaReadTileProc)(const char* PPath,  const EIndex PTotalSamplesPerRow, const EIndex PTileSamplesPerRow,  const EIndex PTotalSamplesPerCol, const EIndex PTileSamplesPerCol,  EIndex PLonI, EIndex PLatI,  const EIndex PUStartCell, const EIndex PVStartCell,  const EIndex PSrcUCells, const EIndex PSrcVCells,  const EIndex PUStep, const EIndex PVStep, const EIndex PDstUCells, const EIndex PDstUCell, const EIndex PDstVCell, void* PBuffer);

extern void  Ep_CacheArea(const char* PPath, const EIndex PTotalSamplesPerRow, const EIndex PTotalSamplesPerCol, const EIndex PTileSamplesPerRow, const EIndex PTileSamplesPerCol,
              EpCacheAreaReadTileProc PLoadSubTile,
              const EIndex PStartUCell, const EIndex PSrcTotalUCells, const EIndex PUStep,  const EIndex PStartVCell, const EIndex PSrcTotalVCells, const EIndex PVStep,
              void* PBuffer, void (*PProgressFunc)(const double PFractionComplete));


extern EBool  Ep_IOModuleSetResource1b(EpImageIOModule* PObject, const char* PName, const EBool PValue);


//--------------------------------
// ToneMap.c
//--------------------------------
extern int   EpToneMapper_Apply(EpToneMapper* PToneMapper,  const EpImage* PSImage, EpImage* PDImage, const EpRegion* PRegion, const uint32_t PChannels, const EIndex PNThreads,  void* PObject);
extern void  (*Ep_ToneMapFree)(void);

//----------------------------------------
// esrc/Tools/Image/Plugins.c
//----------------------------------------
extern int     EpImage_IOPluginsLoad(const char* PAppName, const char* PCommandLine, const char* PFileName, const int PVerbosity);
extern int     Ep_PluginsLoad(const char* PPath, const char* PFileName);
extern void    Ep_PluginsFree(void);

#if defined (__cplusplus) && (!(defined (__CPLUSLIBS)))
}	// Close scope of 'extern "C"' declaration
#endif


#endif // _EImage_h
